/* Generated by Cython 0.29.5 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "name": "ultron.tradingday.Calendar",
        "sources": [
            "ultron/tradingday/Calendar.pyx"
        ]
    },
    "module_name": "ultron.tradingday.Calendar"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_5"
#define CYTHON_HEX_VERSION 0x001D05F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ultron__tradingday__Calendar
#define __PYX_HAVE_API__ultron__tradingday__Calendar
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "ultron/tradingday/Calendar.pyx",
  "ultron/tradingday/Calendar.pxd",
  "stringsource",
  "ultron/tradingday/Date.pxd",
  "ultron/tradingday/Period.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6ultron_10tradingday_4Date_Date;
struct __pyx_obj_6ultron_10tradingday_6Period_Period;
struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl;
struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar;
struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl;
struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl;
struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar;
struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl;
struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl;
struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl;

/* "ultron/tradingday/Enums/_Weekdays.pxd":3
 * # -*- coding: utf-8 -*-
 * 
 * cdef public enum Weekdays:             # <<<<<<<<<<<<<<
 *     Sunday = 1
 *     Monday = 2
 */
enum Weekdays {
  Sunday = 1,
  Monday = 2,
  Tuesday = 3,
  Wednesday = 4,
  Thursday = 5,
  Friday = 6,
  Saturday = 7
};

/* "ultron/tradingday/Enums/_TimeUnits.pxd":3
 * # -*- coding: utf-8 -*-
 * 
 * cdef public enum TimeUnits:             # <<<<<<<<<<<<<<
 *     BDays = 0
 *     Days = 1
 */
enum TimeUnits {
  BDays = 0,
  Days = 1,
  Weeks = 2,
  Months = 3,
  Years = 4
};

/* "ultron/tradingday/Enums/_Months.pxd":4
 * 
 * from ultron.tradingday.Enums._Months cimport Months as mt
 * cdef public enum Months:             # <<<<<<<<<<<<<<
 *     January = 1
 *     February = 2
 */
enum Months {
  January = 1,
  February = 2,
  March = 3,
  April = 4,
  May = 5,
  June = 6,
  July = 7,
  August = 8,
  September = 9,
  October = 10,
  November = 11,
  December = 12
};

/* "ultron/tradingday/Enums/_BizDayConventions.pxd":3
 * # -*- coding: utf-8 -*-
 * 
 * cdef public enum BizDayConventions:             # <<<<<<<<<<<<<<
 *     Following = 0
 *     ModifiedFollowing = 1
 */
enum BizDayConventions {
  Following = 0,
  ModifiedFollowing = 1,
  Preceding = 2,
  ModifiedPreceding = 3,
  Unadjusted = 4,
  HalfMonthModifiedFollowing = 5,
  Nearest = 6
};
struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween;
struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate;
struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_advanceDate;
struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_holDatesList;

/* "ultron/tradingday/Calendar.pxd":23
 *     cpdef isEndOfMonth(self, Date d)
 *     cpdef endOfMonth(self, Date d)
 *     cpdef bizDaysBetween(self, Date fromDate, Date toDate, bint includeFirst=*, bint includeLast=*)             # <<<<<<<<<<<<<<
 *     cpdef adjustDate(self, Date d, int c=*)
 *     cpdef advanceDate(self, Date d, Period period, int c=*, bint endOfMonth=*)
 */
struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween {
  int __pyx_n;
  int includeFirst;
  int includeLast;
};

/* "ultron/tradingday/Calendar.pxd":24
 *     cpdef endOfMonth(self, Date d)
 *     cpdef bizDaysBetween(self, Date fromDate, Date toDate, bint includeFirst=*, bint includeLast=*)
 *     cpdef adjustDate(self, Date d, int c=*)             # <<<<<<<<<<<<<<
 *     cpdef advanceDate(self, Date d, Period period, int c=*, bint endOfMonth=*)
 *     cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=*)
 */
struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate {
  int __pyx_n;
  int c;
};

/* "ultron/tradingday/Calendar.pxd":25
 *     cpdef bizDaysBetween(self, Date fromDate, Date toDate, bint includeFirst=*, bint includeLast=*)
 *     cpdef adjustDate(self, Date d, int c=*)
 *     cpdef advanceDate(self, Date d, Period period, int c=*, bint endOfMonth=*)             # <<<<<<<<<<<<<<
 *     cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=*)
 *     cpdef bizDatesList(self, Date fromDate, Date toDate)
 */
struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_advanceDate {
  int __pyx_n;
  int c;
  int endOfMonth;
};

/* "ultron/tradingday/Calendar.pxd":26
 *     cpdef adjustDate(self, Date d, int c=*)
 *     cpdef advanceDate(self, Date d, Period period, int c=*, bint endOfMonth=*)
 *     cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=*)             # <<<<<<<<<<<<<<
 *     cpdef bizDatesList(self, Date fromDate, Date toDate)
 */
struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_holDatesList {
  int __pyx_n;
  int includeWeekEnds;
};

/* "ultron/tradingday/Date.pxd":3
 * # -*- coding: utf-8 -*-
 * 
 * cdef class Date(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef public int __serialNumber__
 */
struct __pyx_obj_6ultron_10tradingday_4Date_Date {
  PyObject_HEAD
  struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *__pyx_vtab;
  int __serialNumber__;
  int _year;
  int _month;
  int _day;
};


/* "ultron/tradingday/Period.pxd":3
 * # -*- coding: utf-8 -*-
 * 
 * cdef class Period(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef public int _length
 */
struct __pyx_obj_6ultron_10tradingday_6Period_Period {
  PyObject_HEAD
  struct __pyx_vtabstruct_6ultron_10tradingday_6Period_Period *__pyx_vtab;
  int _length;
  int _units;
};


/* "ultron/tradingday/Calendar.pxd":7
 * 
 * 
 * cdef class CalendarImpl(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef bint isBizDay(self, Date date)
 */
struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl {
  PyObject_HEAD
  struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_vtab;
};


/* "ultron/tradingday/Calendar.pxd":13
 * 
 * 
 * cdef class Calendar(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef public CalendarImpl _impl
 */
struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar {
  PyObject_HEAD
  struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *__pyx_vtab;
  struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *_impl;
  PyObject *name;
};


/* "ultron/tradingday/Calendar.pyx":735
 * 
 * 
 * cdef class ChinaSseImpl(CalendarImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */
struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
};


/* "ultron/tradingday/Calendar.pyx":903
 * cdef ChinaSseImpl _sseImpl = ChinaSseImpl()
 * 
 * cdef class ChinaIBImpl(CalendarImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */
struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
};


/* "ultron/tradingday/Calendar.pyx":917
 *             return isinstance(right, ChinaIBImpl)
 * 
 * cdef class NullCalendar(CalendarImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */
struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
};


/* "ultron/tradingday/Calendar.pyx":931
 *             return isinstance(right, NullCalendar)
 * 
 * cdef class ChinaCFFEXImpl(CalendarImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */
struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
};


/* "ultron/tradingday/Calendar.pyx":979
 * ]
 * 
 * cdef class WestenImpl(CalendarImpl):             # <<<<<<<<<<<<<<
 *     cdef bint isWeekEnd(self, int weekDay):
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 */
struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
};


/* "ultron/tradingday/Calendar.pyx":986
 *         return EasterMonday[year - 1901]
 * 
 * cdef class TargetImpl(WestenImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */
struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl __pyx_base;
};



/* "ultron/tradingday/Date.pxd":3
 * # -*- coding: utf-8 -*-
 * 
 * cdef class Date(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef public int __serialNumber__
 */

struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date {
  int (*dayOfMonth)(struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  int (*dayOfYear)(struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  int (*year)(struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  int (*month)(struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  int (*weekday)(struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  PyObject *(*toDateTime)(struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  PyObject *(*_calculate_date)(struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int, int, int);
};
static struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *__pyx_vtabptr_6ultron_10tradingday_4Date_Date;


/* "ultron/tradingday/Period.pxd":3
 * # -*- coding: utf-8 -*-
 * 
 * cdef class Period(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef public int _length
 */

struct __pyx_vtabstruct_6ultron_10tradingday_6Period_Period {
  struct __pyx_obj_6ultron_10tradingday_6Period_Period *(*normalize)(struct __pyx_obj_6ultron_10tradingday_6Period_Period *, int __pyx_skip_dispatch);
  int (*length)(struct __pyx_obj_6ultron_10tradingday_6Period_Period *, int __pyx_skip_dispatch);
  int (*units)(struct __pyx_obj_6ultron_10tradingday_6Period_Period *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6ultron_10tradingday_6Period_Period *__pyx_vtabptr_6ultron_10tradingday_6Period_Period;


/* "ultron/tradingday/Calendar.pyx":169
 *         pass
 * 
 * cdef class CalendarImpl(object):             # <<<<<<<<<<<<<<
 *     cdef bint isBizDay(self, Date date):
 *         pass
 */

struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl {
  int (*isBizDay)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *);
  int (*isWeekEnd)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, int);
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_vtabptr_6ultron_10tradingday_8Calendar_CalendarImpl;


/* "ultron/tradingday/Calendar.pyx":10
 * from ultron.tradingday.Period cimport Period
 * 
 * cdef class Calendar(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, str holCenter):
 *         holCenter = holCenter.lower()
 */

struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar {
  PyObject *(*isBizDay)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  PyObject *(*isHoliday)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  PyObject *(*isWeekEnd)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, int, int __pyx_skip_dispatch);
  PyObject *(*isEndOfMonth)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  PyObject *(*endOfMonth)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  PyObject *(*bizDaysBetween)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween *__pyx_optional_args);
  PyObject *(*adjustDate)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate *__pyx_optional_args);
  PyObject *(*advanceDate)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, struct __pyx_obj_6ultron_10tradingday_6Period_Period *, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_advanceDate *__pyx_optional_args);
  PyObject *(*holDatesList)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_holDatesList *__pyx_optional_args);
  PyObject *(*bizDatesList)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *__pyx_vtabptr_6ultron_10tradingday_8Calendar_Calendar;


/* "ultron/tradingday/Calendar.pyx":735
 * 
 * 
 * cdef class ChinaSseImpl(CalendarImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */

struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaSseImpl {
  struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaSseImpl;


/* "ultron/tradingday/Calendar.pyx":903
 * cdef ChinaSseImpl _sseImpl = ChinaSseImpl()
 * 
 * cdef class ChinaIBImpl(CalendarImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */

struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaIBImpl {
  struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
  int (*isBizDay)(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch);
  int (*isWeekEnd)(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *, int, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaIBImpl;


/* "ultron/tradingday/Calendar.pyx":917
 *             return isinstance(right, ChinaIBImpl)
 * 
 * cdef class NullCalendar(CalendarImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */

struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_NullCalendar {
  struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_vtabptr_6ultron_10tradingday_8Calendar_NullCalendar;


/* "ultron/tradingday/Calendar.pyx":931
 *             return isinstance(right, NullCalendar)
 * 
 * cdef class ChinaCFFEXImpl(CalendarImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */

struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl {
  struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl;


/* "ultron/tradingday/Calendar.pyx":979
 * ]
 * 
 * cdef class WestenImpl(CalendarImpl):             # <<<<<<<<<<<<<<
 *     cdef bint isWeekEnd(self, int weekDay):
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 */

struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_WestenImpl {
  struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_base;
  int (*easterMonday)(struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *, int);
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_vtabptr_6ultron_10tradingday_8Calendar_WestenImpl;


/* "ultron/tradingday/Calendar.pyx":986
 *         return EasterMonday[year - 1901]
 * 
 * cdef class TargetImpl(WestenImpl):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         pass
 */

struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_TargetImpl {
  struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_WestenImpl __pyx_base;
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_vtabptr_6ultron_10tradingday_8Calendar_TargetImpl;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* IncludeStringH.proto */
#include <string.h>

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__Months(enum Months value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isHoliday(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isWeekEnd(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, int __pyx_v_weekday, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isEndOfMonth(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_endOfMonth(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_adjustDate(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_advanceDate(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, struct __pyx_obj_6ultron_10tradingday_6Period_Period *__pyx_v_period, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_advanceDate *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_holDatesList(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_holDatesList *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_bizDatesList(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_12CalendarImpl_isBizDay(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_12CalendarImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_v_self, CYTHON_UNUSED int __pyx_v_weekDay); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_12ChinaSseImpl_isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_12ChinaSseImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self, int __pyx_v_weekDay); /* proto*/
int __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isBizDay(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date, int __pyx_skip_dispatch); /* proto*/
int __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isBizDay__pyx_wrap_1(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date); /* proto*/
int __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, int __pyx_v_weekDay, int __pyx_skip_dispatch); /* proto*/
int __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isWeekEnd__pyx_wrap_1(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, int __pyx_v_weekDay); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_12NullCalendar_isBizDay(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_12NullCalendar_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self, int __pyx_v_weekDay); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_isBizDay(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self, int __pyx_v_weekDay); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_10WestenImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_v_self, int __pyx_v_weekDay); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_10WestenImpl_easterMonday(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_v_self, int __pyx_v_year); /* proto*/
static int __pyx_f_6ultron_10tradingday_8Calendar_10TargetImpl_isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date); /* proto*/

/* Module declarations from 'ultron.tradingday.Date' */
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_4Date_Date = 0;

/* Module declarations from 'ultron.tradingday.Period' */
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_6Period_Period = 0;

/* Module declarations from 'ultron.tradingday.Enums._Weekdays' */

/* Module declarations from 'ultron.tradingday.Enums._TimeUnits' */

/* Module declarations from 'ultron.tradingday.Enums._Months' */

/* Module declarations from 'ultron.tradingday.Enums._BizDayConventions' */

/* Module declarations from 'ultron.tradingday.Calendar' */
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl = 0;
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_8Calendar_Calendar = 0;
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaSseImpl = 0;
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaIBImpl = 0;
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_8Calendar_NullCalendar = 0;
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl = 0;
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_8Calendar_WestenImpl = 0;
static PyTypeObject *__pyx_ptype_6ultron_10tradingday_8Calendar_TargetImpl = 0;
static PyObject *__pyx_v_6ultron_10tradingday_8Calendar_sse_holDays = 0;
static PyObject *__pyx_v_6ultron_10tradingday_8Calendar_sse_working_weekends = 0;
static PyObject *__pyx_v_6ultron_10tradingday_8Calendar_ib_working_weekends = 0;
static struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_6ultron_10tradingday_8Calendar__sseImpl = 0;
static int __pyx_v_6ultron_10tradingday_8Calendar_EasterMonday[0x12B];
static PyObject *__pyx_v_6ultron_10tradingday_8Calendar__holDict = 0;
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_CalendarImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, PyObject *); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaSseImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *, PyObject *); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaIBImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *, PyObject *); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_NullCalendar__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *, PyObject *); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaCFFEXImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *, PyObject *); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_WestenImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *, PyObject *); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_TargetImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "ultron.tradingday.Calendar"
extern int __pyx_module_is_main_ultron__tradingday__Calendar;
int __pyx_module_is_main_ultron__tradingday__Calendar = 0;

/* Implementation of 'ultron.tradingday.Calendar' */
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_c[] = "c";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_impl[] = "_impl";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_null[] = "null";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_lower[] = "lower";
static const char __pyx_k_month[] = "month";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_period[] = "period";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_toDate[] = "toDate";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_Calendar[] = "Calendar";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_china_ib[] = "china.ib";
static const char __pyx_k_fromDate[] = "fromDate";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_isBizDay[] = "isBizDay";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_china_sse[] = "china.sse";
static const char __pyx_k_holCenter[] = "holCenter";
static const char __pyx_k_isHoliday[] = "isHoliday";
static const char __pyx_k_isWeekEnd[] = "isWeekEnd";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_TargetImpl[] = "TargetImpl";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_WestenImpl[] = "WestenImpl";
static const char __pyx_k_adjustDate[] = "adjustDate";
static const char __pyx_k_endOfMonth[] = "endOfMonth";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_ChinaIBImpl[] = "ChinaIBImpl";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_advanceDate[] = "advanceDate";
static const char __pyx_k_china_cffex[] = "china.cffex";
static const char __pyx_k_includeLast[] = "includeLast";
static const char __pyx_k_CalendarImpl[] = "CalendarImpl";
static const char __pyx_k_ChinaSseImpl[] = "ChinaSseImpl";
static const char __pyx_k_NullCalendar[] = "NullCalendar";
static const char __pyx_k_bizDatesList[] = "bizDatesList";
static const char __pyx_k_holDatesList[] = "holDatesList";
static const char __pyx_k_includeFirst[] = "includeFirst";
static const char __pyx_k_isEndOfMonth[] = "isEndOfMonth";
static const char __pyx_k_nullcalendar[] = "nullcalendar";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_westernStyle[] = "westernStyle";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_ChinaCFFEXImpl[] = "ChinaCFFEXImpl";
static const char __pyx_k_bizDaysBetween[] = "bizDaysBetween";
static const char __pyx_k_includeWeekEnds[] = "includeWeekEnds";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_pyx_unpickle_TargetImpl[] = "__pyx_unpickle_TargetImpl";
static const char __pyx_k_pyx_unpickle_WestenImpl[] = "__pyx_unpickle_WestenImpl";
static const char __pyx_k_pyx_unpickle_ChinaIBImpl[] = "__pyx_unpickle_ChinaIBImpl";
static const char __pyx_k_pyx_unpickle_CalendarImpl[] = "__pyx_unpickle_CalendarImpl";
static const char __pyx_k_pyx_unpickle_ChinaSseImpl[] = "__pyx_unpickle_ChinaSseImpl";
static const char __pyx_k_pyx_unpickle_NullCalendar[] = "__pyx_unpickle_NullCalendar";
static const char __pyx_k_ultron_tradingday_Calendar[] = "ultron.tradingday.Calendar";
static const char __pyx_k_pyx_unpickle_ChinaCFFEXImpl[] = "__pyx_unpickle_ChinaCFFEXImpl";
static const char __pyx_k_0_is_not_a_valid_description_of[] = "{0} is not a valid description of a holiday center";
static const char __pyx_k_unknown_business_day_convention[] = "unknown business-day convention";
static const char __pyx_k_Incompatible_checksums_s_vs_0xd4[] = "Incompatible checksums (%s vs 0xd41d8cd = ())";
static PyObject *__pyx_kp_s_0_is_not_a_valid_description_of;
static PyObject *__pyx_n_s_Calendar;
static PyObject *__pyx_n_s_CalendarImpl;
static PyObject *__pyx_n_s_ChinaCFFEXImpl;
static PyObject *__pyx_n_s_ChinaIBImpl;
static PyObject *__pyx_n_s_ChinaSseImpl;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd4;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_NullCalendar;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_TargetImpl;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_WestenImpl;
static PyObject *__pyx_n_s_adjustDate;
static PyObject *__pyx_n_s_advanceDate;
static PyObject *__pyx_n_s_bizDatesList;
static PyObject *__pyx_n_s_bizDaysBetween;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_kp_s_china_cffex;
static PyObject *__pyx_kp_s_china_ib;
static PyObject *__pyx_kp_s_china_sse;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_endOfMonth;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fromDate;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_holCenter;
static PyObject *__pyx_n_s_holDatesList;
static PyObject *__pyx_n_s_impl;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_includeFirst;
static PyObject *__pyx_n_s_includeLast;
static PyObject *__pyx_n_s_includeWeekEnds;
static PyObject *__pyx_n_s_isBizDay;
static PyObject *__pyx_n_s_isEndOfMonth;
static PyObject *__pyx_n_s_isHoliday;
static PyObject *__pyx_n_s_isWeekEnd;
static PyObject *__pyx_n_s_lower;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_month;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_null;
static PyObject *__pyx_n_s_nullcalendar;
static PyObject *__pyx_n_s_period;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_CalendarImpl;
static PyObject *__pyx_n_s_pyx_unpickle_ChinaCFFEXImpl;
static PyObject *__pyx_n_s_pyx_unpickle_ChinaIBImpl;
static PyObject *__pyx_n_s_pyx_unpickle_ChinaSseImpl;
static PyObject *__pyx_n_s_pyx_unpickle_NullCalendar;
static PyObject *__pyx_n_s_pyx_unpickle_TargetImpl;
static PyObject *__pyx_n_s_pyx_unpickle_WestenImpl;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_target;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_toDate;
static PyObject *__pyx_n_s_ultron_tradingday_Calendar;
static PyObject *__pyx_kp_s_unknown_business_day_convention;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_westernStyle;
static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar___init__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, PyObject *__pyx_v_holCenter); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_2isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4isHoliday(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_6isWeekEnd(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, int __pyx_v_weekday); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_8isEndOfMonth(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_10endOfMonth(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_12bizDaysBetween(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_v_includeFirst, int __pyx_v_includeLast); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_14adjustDate(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_v_c); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_16advanceDate(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, struct __pyx_obj_6ultron_10tradingday_6Period_Period *__pyx_v_period, int __pyx_v_c, int __pyx_v_endOfMonth); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_18holDatesList(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_v_includeWeekEnds); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_20bizDatesList(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_22__richcmp__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_24__deepcopy__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_memo); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_26__reduce__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_28__setstate__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_5_impl___get__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self); /* proto */
static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_5_impl_2__set__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_5_impl_4__del__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4name___get__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self); /* proto */
static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4name_2__set__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4name_4__del__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12CalendarImpl___reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12CalendarImpl_2__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl_2__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl_4__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl_6__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_2isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_4isWeekEnd(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, int __pyx_v_weekDay); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_6__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_8__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_10__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar_2__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar_4__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar_6__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_2__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_4__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_6__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10WestenImpl___reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10WestenImpl_2__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl_2__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl_4__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl_6__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar___pyx_unpickle_CalendarImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_2__pyx_unpickle_ChinaSseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_4__pyx_unpickle_ChinaIBImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_6__pyx_unpickle_NullCalendar(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8__pyx_unpickle_ChinaCFFEXImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10__pyx_unpickle_WestenImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12__pyx_unpickle_TargetImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_CalendarImpl(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_Calendar(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_ChinaSseImpl(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_ChinaIBImpl(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_NullCalendar(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_WestenImpl(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_TargetImpl(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyString_Type_lower = {0, &__pyx_n_s_lower, 0, 0, 0};
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_9;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_11;
static PyObject *__pyx_int_12;
static PyObject *__pyx_int_13;
static PyObject *__pyx_int_14;
static PyObject *__pyx_int_15;
static PyObject *__pyx_int_16;
static PyObject *__pyx_int_17;
static PyObject *__pyx_int_18;
static PyObject *__pyx_int_19;
static PyObject *__pyx_int_20;
static PyObject *__pyx_int_21;
static PyObject *__pyx_int_22;
static PyObject *__pyx_int_23;
static PyObject *__pyx_int_24;
static PyObject *__pyx_int_25;
static PyObject *__pyx_int_26;
static PyObject *__pyx_int_27;
static PyObject *__pyx_int_28;
static PyObject *__pyx_int_29;
static PyObject *__pyx_int_30;
static PyObject *__pyx_int_31;
static PyObject *__pyx_int_1991;
static PyObject *__pyx_int_1992;
static PyObject *__pyx_int_1993;
static PyObject *__pyx_int_1994;
static PyObject *__pyx_int_1995;
static PyObject *__pyx_int_1996;
static PyObject *__pyx_int_1997;
static PyObject *__pyx_int_1998;
static PyObject *__pyx_int_1999;
static PyObject *__pyx_int_2000;
static PyObject *__pyx_int_2001;
static PyObject *__pyx_int_2002;
static PyObject *__pyx_int_2003;
static PyObject *__pyx_int_2004;
static PyObject *__pyx_int_2005;
static PyObject *__pyx_int_2006;
static PyObject *__pyx_int_2007;
static PyObject *__pyx_int_2008;
static PyObject *__pyx_int_2009;
static PyObject *__pyx_int_2010;
static PyObject *__pyx_int_2011;
static PyObject *__pyx_int_2012;
static PyObject *__pyx_int_2013;
static PyObject *__pyx_int_2014;
static PyObject *__pyx_int_2015;
static PyObject *__pyx_int_2016;
static PyObject *__pyx_int_2017;
static PyObject *__pyx_int_2018;
static PyObject *__pyx_int_2019;
static PyObject *__pyx_int_2020;
static PyObject *__pyx_int_2021;
static PyObject *__pyx_int_2022;
static PyObject *__pyx_int_2023;
static PyObject *__pyx_int_222419149;
static int __pyx_k_;
static int __pyx_k__3;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__76;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_tuple__78;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__80;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__94;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__97;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_tuple__99;
static PyObject *__pyx_tuple__100;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__102;
static PyObject *__pyx_tuple__103;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__105;
static PyObject *__pyx_tuple__106;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__109;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__111;
static PyObject *__pyx_tuple__112;
static PyObject *__pyx_tuple__113;
static PyObject *__pyx_tuple__114;
static PyObject *__pyx_tuple__115;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_tuple__117;
static PyObject *__pyx_tuple__118;
static PyObject *__pyx_tuple__119;
static PyObject *__pyx_tuple__120;
static PyObject *__pyx_tuple__121;
static PyObject *__pyx_tuple__122;
static PyObject *__pyx_tuple__123;
static PyObject *__pyx_tuple__124;
static PyObject *__pyx_tuple__125;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__127;
static PyObject *__pyx_tuple__128;
static PyObject *__pyx_tuple__129;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__131;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__133;
static PyObject *__pyx_tuple__134;
static PyObject *__pyx_tuple__135;
static PyObject *__pyx_tuple__136;
static PyObject *__pyx_tuple__137;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__139;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__141;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__144;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__146;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__148;
static PyObject *__pyx_tuple__149;
static PyObject *__pyx_tuple__150;
static PyObject *__pyx_tuple__151;
static PyObject *__pyx_tuple__152;
static PyObject *__pyx_tuple__153;
static PyObject *__pyx_tuple__154;
static PyObject *__pyx_tuple__155;
static PyObject *__pyx_tuple__156;
static PyObject *__pyx_tuple__157;
static PyObject *__pyx_tuple__158;
static PyObject *__pyx_tuple__159;
static PyObject *__pyx_tuple__160;
static PyObject *__pyx_tuple__161;
static PyObject *__pyx_tuple__162;
static PyObject *__pyx_tuple__163;
static PyObject *__pyx_tuple__164;
static PyObject *__pyx_tuple__165;
static PyObject *__pyx_tuple__166;
static PyObject *__pyx_tuple__167;
static PyObject *__pyx_tuple__168;
static PyObject *__pyx_tuple__169;
static PyObject *__pyx_tuple__170;
static PyObject *__pyx_tuple__171;
static PyObject *__pyx_tuple__172;
static PyObject *__pyx_tuple__173;
static PyObject *__pyx_tuple__174;
static PyObject *__pyx_tuple__175;
static PyObject *__pyx_tuple__176;
static PyObject *__pyx_tuple__177;
static PyObject *__pyx_tuple__178;
static PyObject *__pyx_tuple__179;
static PyObject *__pyx_tuple__180;
static PyObject *__pyx_tuple__181;
static PyObject *__pyx_tuple__182;
static PyObject *__pyx_tuple__183;
static PyObject *__pyx_tuple__184;
static PyObject *__pyx_tuple__185;
static PyObject *__pyx_tuple__186;
static PyObject *__pyx_tuple__187;
static PyObject *__pyx_tuple__188;
static PyObject *__pyx_tuple__189;
static PyObject *__pyx_tuple__190;
static PyObject *__pyx_tuple__191;
static PyObject *__pyx_tuple__192;
static PyObject *__pyx_tuple__193;
static PyObject *__pyx_tuple__194;
static PyObject *__pyx_tuple__195;
static PyObject *__pyx_tuple__196;
static PyObject *__pyx_tuple__197;
static PyObject *__pyx_tuple__198;
static PyObject *__pyx_tuple__199;
static PyObject *__pyx_tuple__200;
static PyObject *__pyx_tuple__201;
static PyObject *__pyx_tuple__202;
static PyObject *__pyx_tuple__203;
static PyObject *__pyx_tuple__204;
static PyObject *__pyx_tuple__205;
static PyObject *__pyx_tuple__206;
static PyObject *__pyx_tuple__207;
static PyObject *__pyx_tuple__208;
static PyObject *__pyx_tuple__209;
static PyObject *__pyx_tuple__210;
static PyObject *__pyx_tuple__211;
static PyObject *__pyx_tuple__212;
static PyObject *__pyx_tuple__213;
static PyObject *__pyx_tuple__214;
static PyObject *__pyx_tuple__215;
static PyObject *__pyx_tuple__216;
static PyObject *__pyx_tuple__217;
static PyObject *__pyx_tuple__218;
static PyObject *__pyx_tuple__219;
static PyObject *__pyx_tuple__220;
static PyObject *__pyx_tuple__221;
static PyObject *__pyx_tuple__222;
static PyObject *__pyx_tuple__223;
static PyObject *__pyx_tuple__224;
static PyObject *__pyx_tuple__225;
static PyObject *__pyx_tuple__226;
static PyObject *__pyx_tuple__227;
static PyObject *__pyx_tuple__228;
static PyObject *__pyx_tuple__229;
static PyObject *__pyx_tuple__230;
static PyObject *__pyx_tuple__231;
static PyObject *__pyx_tuple__232;
static PyObject *__pyx_tuple__233;
static PyObject *__pyx_tuple__234;
static PyObject *__pyx_tuple__235;
static PyObject *__pyx_tuple__236;
static PyObject *__pyx_tuple__237;
static PyObject *__pyx_tuple__238;
static PyObject *__pyx_tuple__239;
static PyObject *__pyx_tuple__240;
static PyObject *__pyx_tuple__241;
static PyObject *__pyx_tuple__242;
static PyObject *__pyx_tuple__243;
static PyObject *__pyx_tuple__244;
static PyObject *__pyx_tuple__245;
static PyObject *__pyx_tuple__246;
static PyObject *__pyx_tuple__247;
static PyObject *__pyx_tuple__248;
static PyObject *__pyx_tuple__249;
static PyObject *__pyx_tuple__250;
static PyObject *__pyx_tuple__251;
static PyObject *__pyx_tuple__252;
static PyObject *__pyx_tuple__253;
static PyObject *__pyx_tuple__254;
static PyObject *__pyx_tuple__255;
static PyObject *__pyx_tuple__256;
static PyObject *__pyx_tuple__257;
static PyObject *__pyx_tuple__258;
static PyObject *__pyx_tuple__259;
static PyObject *__pyx_tuple__260;
static PyObject *__pyx_tuple__261;
static PyObject *__pyx_tuple__262;
static PyObject *__pyx_tuple__263;
static PyObject *__pyx_tuple__264;
static PyObject *__pyx_tuple__265;
static PyObject *__pyx_tuple__266;
static PyObject *__pyx_tuple__267;
static PyObject *__pyx_tuple__268;
static PyObject *__pyx_tuple__269;
static PyObject *__pyx_tuple__270;
static PyObject *__pyx_tuple__271;
static PyObject *__pyx_tuple__272;
static PyObject *__pyx_tuple__273;
static PyObject *__pyx_tuple__274;
static PyObject *__pyx_tuple__275;
static PyObject *__pyx_tuple__276;
static PyObject *__pyx_tuple__277;
static PyObject *__pyx_tuple__278;
static PyObject *__pyx_tuple__279;
static PyObject *__pyx_tuple__280;
static PyObject *__pyx_tuple__281;
static PyObject *__pyx_tuple__282;
static PyObject *__pyx_tuple__283;
static PyObject *__pyx_tuple__284;
static PyObject *__pyx_tuple__285;
static PyObject *__pyx_tuple__286;
static PyObject *__pyx_tuple__287;
static PyObject *__pyx_tuple__288;
static PyObject *__pyx_tuple__289;
static PyObject *__pyx_tuple__290;
static PyObject *__pyx_tuple__291;
static PyObject *__pyx_tuple__292;
static PyObject *__pyx_tuple__293;
static PyObject *__pyx_tuple__294;
static PyObject *__pyx_tuple__295;
static PyObject *__pyx_tuple__296;
static PyObject *__pyx_tuple__297;
static PyObject *__pyx_tuple__298;
static PyObject *__pyx_tuple__299;
static PyObject *__pyx_tuple__300;
static PyObject *__pyx_tuple__301;
static PyObject *__pyx_tuple__302;
static PyObject *__pyx_tuple__303;
static PyObject *__pyx_tuple__304;
static PyObject *__pyx_tuple__305;
static PyObject *__pyx_tuple__306;
static PyObject *__pyx_tuple__307;
static PyObject *__pyx_tuple__308;
static PyObject *__pyx_tuple__309;
static PyObject *__pyx_tuple__310;
static PyObject *__pyx_tuple__311;
static PyObject *__pyx_tuple__312;
static PyObject *__pyx_tuple__313;
static PyObject *__pyx_tuple__314;
static PyObject *__pyx_tuple__315;
static PyObject *__pyx_tuple__316;
static PyObject *__pyx_tuple__317;
static PyObject *__pyx_tuple__318;
static PyObject *__pyx_tuple__319;
static PyObject *__pyx_tuple__320;
static PyObject *__pyx_tuple__321;
static PyObject *__pyx_tuple__322;
static PyObject *__pyx_tuple__323;
static PyObject *__pyx_tuple__324;
static PyObject *__pyx_tuple__325;
static PyObject *__pyx_tuple__326;
static PyObject *__pyx_tuple__327;
static PyObject *__pyx_tuple__328;
static PyObject *__pyx_tuple__329;
static PyObject *__pyx_tuple__330;
static PyObject *__pyx_tuple__331;
static PyObject *__pyx_tuple__332;
static PyObject *__pyx_tuple__333;
static PyObject *__pyx_tuple__334;
static PyObject *__pyx_tuple__335;
static PyObject *__pyx_tuple__336;
static PyObject *__pyx_tuple__337;
static PyObject *__pyx_tuple__338;
static PyObject *__pyx_tuple__339;
static PyObject *__pyx_tuple__340;
static PyObject *__pyx_tuple__341;
static PyObject *__pyx_tuple__342;
static PyObject *__pyx_tuple__343;
static PyObject *__pyx_tuple__344;
static PyObject *__pyx_tuple__345;
static PyObject *__pyx_tuple__346;
static PyObject *__pyx_tuple__347;
static PyObject *__pyx_tuple__348;
static PyObject *__pyx_tuple__349;
static PyObject *__pyx_tuple__350;
static PyObject *__pyx_tuple__351;
static PyObject *__pyx_tuple__352;
static PyObject *__pyx_tuple__353;
static PyObject *__pyx_tuple__354;
static PyObject *__pyx_tuple__355;
static PyObject *__pyx_tuple__356;
static PyObject *__pyx_tuple__357;
static PyObject *__pyx_tuple__358;
static PyObject *__pyx_tuple__359;
static PyObject *__pyx_tuple__360;
static PyObject *__pyx_tuple__361;
static PyObject *__pyx_tuple__362;
static PyObject *__pyx_tuple__363;
static PyObject *__pyx_tuple__364;
static PyObject *__pyx_tuple__365;
static PyObject *__pyx_tuple__366;
static PyObject *__pyx_tuple__367;
static PyObject *__pyx_tuple__368;
static PyObject *__pyx_tuple__369;
static PyObject *__pyx_tuple__370;
static PyObject *__pyx_tuple__371;
static PyObject *__pyx_tuple__372;
static PyObject *__pyx_tuple__373;
static PyObject *__pyx_tuple__374;
static PyObject *__pyx_tuple__375;
static PyObject *__pyx_tuple__376;
static PyObject *__pyx_tuple__377;
static PyObject *__pyx_tuple__378;
static PyObject *__pyx_tuple__379;
static PyObject *__pyx_tuple__380;
static PyObject *__pyx_tuple__381;
static PyObject *__pyx_tuple__382;
static PyObject *__pyx_tuple__383;
static PyObject *__pyx_tuple__384;
static PyObject *__pyx_tuple__385;
static PyObject *__pyx_tuple__386;
static PyObject *__pyx_tuple__387;
static PyObject *__pyx_tuple__388;
static PyObject *__pyx_tuple__389;
static PyObject *__pyx_tuple__390;
static PyObject *__pyx_tuple__391;
static PyObject *__pyx_tuple__392;
static PyObject *__pyx_tuple__393;
static PyObject *__pyx_tuple__394;
static PyObject *__pyx_tuple__395;
static PyObject *__pyx_tuple__396;
static PyObject *__pyx_tuple__397;
static PyObject *__pyx_tuple__398;
static PyObject *__pyx_tuple__399;
static PyObject *__pyx_tuple__400;
static PyObject *__pyx_tuple__401;
static PyObject *__pyx_tuple__402;
static PyObject *__pyx_tuple__403;
static PyObject *__pyx_tuple__404;
static PyObject *__pyx_tuple__405;
static PyObject *__pyx_tuple__406;
static PyObject *__pyx_tuple__407;
static PyObject *__pyx_tuple__408;
static PyObject *__pyx_tuple__409;
static PyObject *__pyx_tuple__410;
static PyObject *__pyx_tuple__411;
static PyObject *__pyx_tuple__412;
static PyObject *__pyx_tuple__413;
static PyObject *__pyx_tuple__414;
static PyObject *__pyx_tuple__415;
static PyObject *__pyx_tuple__416;
static PyObject *__pyx_tuple__417;
static PyObject *__pyx_tuple__418;
static PyObject *__pyx_tuple__419;
static PyObject *__pyx_tuple__420;
static PyObject *__pyx_tuple__421;
static PyObject *__pyx_tuple__422;
static PyObject *__pyx_tuple__423;
static PyObject *__pyx_tuple__424;
static PyObject *__pyx_tuple__425;
static PyObject *__pyx_tuple__426;
static PyObject *__pyx_tuple__427;
static PyObject *__pyx_tuple__428;
static PyObject *__pyx_tuple__429;
static PyObject *__pyx_tuple__430;
static PyObject *__pyx_tuple__431;
static PyObject *__pyx_tuple__432;
static PyObject *__pyx_tuple__433;
static PyObject *__pyx_tuple__434;
static PyObject *__pyx_tuple__435;
static PyObject *__pyx_tuple__436;
static PyObject *__pyx_tuple__437;
static PyObject *__pyx_tuple__438;
static PyObject *__pyx_tuple__439;
static PyObject *__pyx_tuple__440;
static PyObject *__pyx_tuple__441;
static PyObject *__pyx_tuple__442;
static PyObject *__pyx_tuple__443;
static PyObject *__pyx_tuple__444;
static PyObject *__pyx_tuple__445;
static PyObject *__pyx_tuple__446;
static PyObject *__pyx_tuple__447;
static PyObject *__pyx_tuple__448;
static PyObject *__pyx_tuple__449;
static PyObject *__pyx_tuple__450;
static PyObject *__pyx_tuple__451;
static PyObject *__pyx_tuple__452;
static PyObject *__pyx_tuple__453;
static PyObject *__pyx_tuple__454;
static PyObject *__pyx_tuple__455;
static PyObject *__pyx_tuple__456;
static PyObject *__pyx_tuple__457;
static PyObject *__pyx_tuple__458;
static PyObject *__pyx_tuple__459;
static PyObject *__pyx_tuple__460;
static PyObject *__pyx_tuple__461;
static PyObject *__pyx_tuple__462;
static PyObject *__pyx_tuple__463;
static PyObject *__pyx_tuple__464;
static PyObject *__pyx_tuple__465;
static PyObject *__pyx_tuple__466;
static PyObject *__pyx_tuple__467;
static PyObject *__pyx_tuple__468;
static PyObject *__pyx_tuple__469;
static PyObject *__pyx_tuple__470;
static PyObject *__pyx_tuple__471;
static PyObject *__pyx_tuple__472;
static PyObject *__pyx_tuple__473;
static PyObject *__pyx_tuple__474;
static PyObject *__pyx_tuple__475;
static PyObject *__pyx_tuple__476;
static PyObject *__pyx_tuple__477;
static PyObject *__pyx_tuple__478;
static PyObject *__pyx_tuple__479;
static PyObject *__pyx_tuple__480;
static PyObject *__pyx_tuple__481;
static PyObject *__pyx_tuple__482;
static PyObject *__pyx_tuple__483;
static PyObject *__pyx_tuple__484;
static PyObject *__pyx_tuple__485;
static PyObject *__pyx_tuple__486;
static PyObject *__pyx_tuple__487;
static PyObject *__pyx_tuple__488;
static PyObject *__pyx_tuple__489;
static PyObject *__pyx_tuple__490;
static PyObject *__pyx_tuple__491;
static PyObject *__pyx_tuple__492;
static PyObject *__pyx_tuple__493;
static PyObject *__pyx_tuple__494;
static PyObject *__pyx_tuple__495;
static PyObject *__pyx_tuple__496;
static PyObject *__pyx_tuple__497;
static PyObject *__pyx_tuple__498;
static PyObject *__pyx_tuple__499;
static PyObject *__pyx_tuple__500;
static PyObject *__pyx_tuple__501;
static PyObject *__pyx_tuple__502;
static PyObject *__pyx_tuple__503;
static PyObject *__pyx_tuple__504;
static PyObject *__pyx_tuple__505;
static PyObject *__pyx_tuple__506;
static PyObject *__pyx_tuple__507;
static PyObject *__pyx_tuple__508;
static PyObject *__pyx_tuple__509;
static PyObject *__pyx_tuple__510;
static PyObject *__pyx_tuple__511;
static PyObject *__pyx_tuple__512;
static PyObject *__pyx_tuple__513;
static PyObject *__pyx_tuple__514;
static PyObject *__pyx_tuple__515;
static PyObject *__pyx_tuple__516;
static PyObject *__pyx_tuple__517;
static PyObject *__pyx_tuple__518;
static PyObject *__pyx_tuple__519;
static PyObject *__pyx_tuple__520;
static PyObject *__pyx_tuple__521;
static PyObject *__pyx_tuple__522;
static PyObject *__pyx_tuple__523;
static PyObject *__pyx_tuple__524;
static PyObject *__pyx_tuple__525;
static PyObject *__pyx_tuple__526;
static PyObject *__pyx_tuple__527;
static PyObject *__pyx_tuple__528;
static PyObject *__pyx_tuple__529;
static PyObject *__pyx_tuple__530;
static PyObject *__pyx_tuple__531;
static PyObject *__pyx_tuple__532;
static PyObject *__pyx_tuple__533;
static PyObject *__pyx_tuple__534;
static PyObject *__pyx_tuple__535;
static PyObject *__pyx_tuple__536;
static PyObject *__pyx_tuple__537;
static PyObject *__pyx_tuple__538;
static PyObject *__pyx_tuple__539;
static PyObject *__pyx_tuple__540;
static PyObject *__pyx_tuple__541;
static PyObject *__pyx_tuple__542;
static PyObject *__pyx_tuple__543;
static PyObject *__pyx_tuple__544;
static PyObject *__pyx_tuple__545;
static PyObject *__pyx_tuple__546;
static PyObject *__pyx_tuple__547;
static PyObject *__pyx_tuple__548;
static PyObject *__pyx_tuple__549;
static PyObject *__pyx_tuple__550;
static PyObject *__pyx_tuple__551;
static PyObject *__pyx_tuple__552;
static PyObject *__pyx_tuple__553;
static PyObject *__pyx_tuple__554;
static PyObject *__pyx_tuple__555;
static PyObject *__pyx_tuple__556;
static PyObject *__pyx_tuple__557;
static PyObject *__pyx_tuple__558;
static PyObject *__pyx_tuple__559;
static PyObject *__pyx_tuple__560;
static PyObject *__pyx_tuple__561;
static PyObject *__pyx_tuple__562;
static PyObject *__pyx_tuple__563;
static PyObject *__pyx_tuple__564;
static PyObject *__pyx_tuple__565;
static PyObject *__pyx_tuple__566;
static PyObject *__pyx_tuple__567;
static PyObject *__pyx_tuple__568;
static PyObject *__pyx_tuple__569;
static PyObject *__pyx_tuple__570;
static PyObject *__pyx_tuple__571;
static PyObject *__pyx_tuple__572;
static PyObject *__pyx_tuple__573;
static PyObject *__pyx_tuple__574;
static PyObject *__pyx_tuple__576;
static PyObject *__pyx_tuple__578;
static PyObject *__pyx_tuple__580;
static PyObject *__pyx_tuple__582;
static PyObject *__pyx_tuple__584;
static PyObject *__pyx_tuple__586;
static PyObject *__pyx_codeobj__575;
static PyObject *__pyx_codeobj__577;
static PyObject *__pyx_codeobj__579;
static PyObject *__pyx_codeobj__581;
static PyObject *__pyx_codeobj__583;
static PyObject *__pyx_codeobj__585;
static PyObject *__pyx_codeobj__587;
/* Late includes */

/* "ultron/tradingday/Calendar.pyx":11
 * 
 * cdef class Calendar(object):
 *     def __init__(self, str holCenter):             # <<<<<<<<<<<<<<
 *         holCenter = holCenter.lower()
 *         try:
 */

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_holCenter = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_holCenter,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_holCenter)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 11, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_holCenter = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 11, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_holCenter), (&PyString_Type), 1, "holCenter", 1))) __PYX_ERR(0, 11, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar___init__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), __pyx_v_holCenter);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar___init__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, PyObject *__pyx_v_holCenter) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_holCenter);

  /* "ultron/tradingday/Calendar.pyx":12
 * cdef class Calendar(object):
 *     def __init__(self, str holCenter):
 *         holCenter = holCenter.lower()             # <<<<<<<<<<<<<<
 *         try:
 *             self._impl = _holDict[holCenter]()
 */
  __pyx_t_1 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyString_Type_lower, __pyx_v_holCenter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_holCenter, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "ultron/tradingday/Calendar.pyx":13
 *     def __init__(self, str holCenter):
 *         holCenter = holCenter.lower()
 *         try:             # <<<<<<<<<<<<<<
 *             self._impl = _holDict[holCenter]()
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "ultron/tradingday/Calendar.pyx":14
 *         holCenter = holCenter.lower()
 *         try:
 *             self._impl = _holDict[holCenter]()             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             raise ValueError("{0} is not a valid description of a holiday center".format(holCenter))
 */
      if (unlikely(__pyx_v_6ultron_10tradingday_8Calendar__holDict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 14, __pyx_L3_error)
      }
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_6ultron_10tradingday_8Calendar__holDict, __pyx_v_holCenter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 14, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl))))) __PYX_ERR(0, 14, __pyx_L3_error)
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_v_self->_impl);
      __Pyx_DECREF(((PyObject *)__pyx_v_self->_impl));
      __pyx_v_self->_impl = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "ultron/tradingday/Calendar.pyx":13
 *     def __init__(self, str holCenter):
 *         holCenter = holCenter.lower()
 *         try:             # <<<<<<<<<<<<<<
 *             self._impl = _holDict[holCenter]()
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "ultron/tradingday/Calendar.pyx":15
 *         try:
 *             self._impl = _holDict[holCenter]()
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             raise ValueError("{0} is not a valid description of a holiday center".format(holCenter))
 *         self.name = holCenter
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 15, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);

      /* "ultron/tradingday/Calendar.pyx":16
 *             self._impl = _holDict[holCenter]()
 *         except KeyError:
 *             raise ValueError("{0} is not a valid description of a holiday center".format(holCenter))             # <<<<<<<<<<<<<<
 *         self.name = holCenter
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_is_not_a_valid_description_of, __pyx_n_s_format); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 16, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_v_holCenter) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_holCenter);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 16, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 16, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 16, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "ultron/tradingday/Calendar.pyx":13
 *     def __init__(self, str holCenter):
 *         holCenter = holCenter.lower()
 *         try:             # <<<<<<<<<<<<<<
 *             self._impl = _holDict[holCenter]()
 *         except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "ultron/tradingday/Calendar.pyx":17
 *         except KeyError:
 *             raise ValueError("{0} is not a valid description of a holiday center".format(holCenter))
 *         self.name = holCenter             # <<<<<<<<<<<<<<
 * 
 *     cpdef isBizDay(self, Date d):
 */
  __Pyx_INCREF(__pyx_v_holCenter);
  __Pyx_GIVEREF(__pyx_v_holCenter);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_holCenter;

  /* "ultron/tradingday/Calendar.pyx":11
 * 
 * cdef class Calendar(object):
 *     def __init__(self, str holCenter):             # <<<<<<<<<<<<<<
 *         holCenter = holCenter.lower()
 *         try:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_holCenter);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":19
 *         self.name = holCenter
 * 
 *     cpdef isBizDay(self, Date d):             # <<<<<<<<<<<<<<
 *         return self._impl.isBizDay(d)
 * 
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_3isBizDay(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("isBizDay", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isBizDay); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_3isBizDay)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_d)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_d));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":20
 * 
 *     cpdef isBizDay(self, Date d):
 *         return self._impl.isBizDay(d)             # <<<<<<<<<<<<<<
 * 
 *     cpdef isHoliday(self, Date d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_self->_impl->__pyx_vtab)->isBizDay(__pyx_v_self->_impl, __pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":19
 *         self.name = holCenter
 * 
 *     cpdef isBizDay(self, Date d):             # <<<<<<<<<<<<<<
 *         return self._impl.isBizDay(d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.isBizDay", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_3isBizDay(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_2isBizDay[] = "Calendar.isBizDay(self, Date d)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_3isBizDay(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isBizDay (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_d), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "d", 0))) __PYX_ERR(0, 19, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_2isBizDay(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_v_d));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_2isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("isBizDay", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isBizDay(__pyx_v_self, __pyx_v_d, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.isBizDay", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":22
 *         return self._impl.isBizDay(d)
 * 
 *     cpdef isHoliday(self, Date d):             # <<<<<<<<<<<<<<
 *         return not self._impl.isBizDay(d)
 * 
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5isHoliday(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isHoliday(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("isHoliday", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isHoliday); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5isHoliday)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_d)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_d));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":23
 * 
 *     cpdef isHoliday(self, Date d):
 *         return not self._impl.isBizDay(d)             # <<<<<<<<<<<<<<
 * 
 *     cpdef isWeekEnd(self, int weekday):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_self->_impl->__pyx_vtab)->isBizDay(__pyx_v_self->_impl, __pyx_v_d) != 0))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":22
 *         return self._impl.isBizDay(d)
 * 
 *     cpdef isHoliday(self, Date d):             # <<<<<<<<<<<<<<
 *         return not self._impl.isBizDay(d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.isHoliday", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5isHoliday(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_4isHoliday[] = "Calendar.isHoliday(self, Date d)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5isHoliday(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isHoliday (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_d), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "d", 0))) __PYX_ERR(0, 22, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4isHoliday(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_v_d));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4isHoliday(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("isHoliday", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isHoliday(__pyx_v_self, __pyx_v_d, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.isHoliday", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":25
 *         return not self._impl.isBizDay(d)
 * 
 *     cpdef isWeekEnd(self, int weekday):             # <<<<<<<<<<<<<<
 *         return self._impl.isWeekEnd(weekday)
 * 
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_7isWeekEnd(PyObject *__pyx_v_self, PyObject *__pyx_arg_weekday); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isWeekEnd(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, int __pyx_v_weekday, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("isWeekEnd", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isWeekEnd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_7isWeekEnd)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_weekday); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":26
 * 
 *     cpdef isWeekEnd(self, int weekday):
 *         return self._impl.isWeekEnd(weekday)             # <<<<<<<<<<<<<<
 * 
 *     cpdef isEndOfMonth(self, Date d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_self->_impl->__pyx_vtab)->isWeekEnd(__pyx_v_self->_impl, __pyx_v_weekday)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":25
 *         return not self._impl.isBizDay(d)
 * 
 *     cpdef isWeekEnd(self, int weekday):             # <<<<<<<<<<<<<<
 *         return self._impl.isWeekEnd(weekday)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.isWeekEnd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_7isWeekEnd(PyObject *__pyx_v_self, PyObject *__pyx_arg_weekday); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_6isWeekEnd[] = "Calendar.isWeekEnd(self, int weekday)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_7isWeekEnd(PyObject *__pyx_v_self, PyObject *__pyx_arg_weekday) {
  int __pyx_v_weekday;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isWeekEnd (wrapper)", 0);
  assert(__pyx_arg_weekday); {
    __pyx_v_weekday = __Pyx_PyInt_As_int(__pyx_arg_weekday); if (unlikely((__pyx_v_weekday == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 25, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.isWeekEnd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_6isWeekEnd(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((int)__pyx_v_weekday));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_6isWeekEnd(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, int __pyx_v_weekday) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("isWeekEnd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isWeekEnd(__pyx_v_self, __pyx_v_weekday, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.isWeekEnd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":28
 *         return self._impl.isWeekEnd(weekday)
 * 
 *     cpdef isEndOfMonth(self, Date d):             # <<<<<<<<<<<<<<
 *         return d.month() != self.adjustDate(d + 1).month()
 * 
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_9isEndOfMonth(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isEndOfMonth(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("isEndOfMonth", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isEndOfMonth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_9isEndOfMonth)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_d)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_d));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":29
 * 
 *     cpdef isEndOfMonth(self, Date d):
 *         return d.month() != self.adjustDate(d + 1).month()             # <<<<<<<<<<<<<<
 * 
 *     cpdef endOfMonth(self, Date d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_d->__pyx_vtab)->month(__pyx_v_d, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Add(((PyObject *)__pyx_v_d), __pyx_int_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 29, __pyx_L1_error)
  __pyx_t_4 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->adjustDate(__pyx_v_self, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_3), 0, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_month); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":28
 *         return self._impl.isWeekEnd(weekday)
 * 
 *     cpdef isEndOfMonth(self, Date d):             # <<<<<<<<<<<<<<
 *         return d.month() != self.adjustDate(d + 1).month()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.isEndOfMonth", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_9isEndOfMonth(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_8isEndOfMonth[] = "Calendar.isEndOfMonth(self, Date d)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_9isEndOfMonth(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isEndOfMonth (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_d), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "d", 0))) __PYX_ERR(0, 28, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_8isEndOfMonth(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_v_d));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_8isEndOfMonth(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("isEndOfMonth", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isEndOfMonth(__pyx_v_self, __pyx_v_d, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.isEndOfMonth", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":31
 *         return d.month() != self.adjustDate(d + 1).month()
 * 
 *     cpdef endOfMonth(self, Date d):             # <<<<<<<<<<<<<<
 *         return self.adjustDate(Date.endOfMonth(d), BizDayConventions.Preceding)
 * 
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_11endOfMonth(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_endOfMonth(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate __pyx_t_5;
  __Pyx_RefNannySetupContext("endOfMonth", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_endOfMonth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_11endOfMonth)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_d)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_d));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":32
 * 
 *     cpdef endOfMonth(self, Date d):
 *         return self.adjustDate(Date.endOfMonth(d), BizDayConventions.Preceding)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bizDaysBetween(self, Date fromDate, Date toDate, bint includeFirst=True, bint includeLast=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_endOfMonth); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_d)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_d));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 32, __pyx_L1_error)
  __pyx_t_5.__pyx_n = 1;
  __pyx_t_5.c = Preceding;
  __pyx_t_2 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->adjustDate(__pyx_v_self, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1), 0, &__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":31
 *         return d.month() != self.adjustDate(d + 1).month()
 * 
 *     cpdef endOfMonth(self, Date d):             # <<<<<<<<<<<<<<
 *         return self.adjustDate(Date.endOfMonth(d), BizDayConventions.Preceding)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.endOfMonth", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_11endOfMonth(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_10endOfMonth[] = "Calendar.endOfMonth(self, Date d)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_11endOfMonth(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("endOfMonth (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_d), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "d", 0))) __PYX_ERR(0, 31, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_10endOfMonth(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_v_d));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_10endOfMonth(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("endOfMonth", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar_8Calendar_endOfMonth(__pyx_v_self, __pyx_v_d, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.endOfMonth", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":34
 *         return self.adjustDate(Date.endOfMonth(d), BizDayConventions.Preceding)
 * 
 *     cpdef bizDaysBetween(self, Date fromDate, Date toDate, bint includeFirst=True, bint includeLast=False):             # <<<<<<<<<<<<<<
 *         cdef int wd = 0
 *         cdef Date d
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_13bizDaysBetween(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween *__pyx_optional_args) {
  int __pyx_v_includeFirst = ((int)1);
  int __pyx_v_includeLast = ((int)0);
  int __pyx_v_wd;
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("bizDaysBetween", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_includeFirst = __pyx_optional_args->includeFirst;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_includeLast = __pyx_optional_args->includeLast;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_bizDaysBetween); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_13bizDaysBetween)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_includeFirst); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 34, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_includeLast); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 34, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[5] = {__pyx_t_6, ((PyObject *)__pyx_v_fromDate), ((PyObject *)__pyx_v_toDate), __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[5] = {__pyx_t_6, ((PyObject *)__pyx_v_fromDate), ((PyObject *)__pyx_v_toDate), __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 34, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_fromDate));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_fromDate));
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, ((PyObject *)__pyx_v_fromDate));
          __Pyx_INCREF(((PyObject *)__pyx_v_toDate));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_toDate));
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, ((PyObject *)__pyx_v_toDate));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":35
 * 
 *     cpdef bizDaysBetween(self, Date fromDate, Date toDate, bint includeFirst=True, bint includeLast=False):
 *         cdef int wd = 0             # <<<<<<<<<<<<<<
 *         cdef Date d
 * 
 */
  __pyx_v_wd = 0;

  /* "ultron/tradingday/Calendar.pyx":38
 *         cdef Date d
 * 
 *         if fromDate != toDate:             # <<<<<<<<<<<<<<
 *             if fromDate < toDate:
 *                 d = fromDate
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_fromDate), ((PyObject *)__pyx_v_toDate), Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_9) {

    /* "ultron/tradingday/Calendar.pyx":39
 * 
 *         if fromDate != toDate:
 *             if fromDate < toDate:             # <<<<<<<<<<<<<<
 *                 d = fromDate
 *                 while d < toDate:
 */
    __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_fromDate), ((PyObject *)__pyx_v_toDate), Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_9) {

      /* "ultron/tradingday/Calendar.pyx":40
 *         if fromDate != toDate:
 *             if fromDate < toDate:
 *                 d = fromDate             # <<<<<<<<<<<<<<
 *                 while d < toDate:
 *                     if self.isBizDay(d):
 */
      __Pyx_INCREF(((PyObject *)__pyx_v_fromDate));
      __pyx_v_d = __pyx_v_fromDate;

      /* "ultron/tradingday/Calendar.pyx":41
 *             if fromDate < toDate:
 *                 d = fromDate
 *                 while d < toDate:             # <<<<<<<<<<<<<<
 *                     if self.isBizDay(d):
 *                         wd += 1
 */
      while (1) {
        __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_v_toDate), Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 41, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (!__pyx_t_9) break;

        /* "ultron/tradingday/Calendar.pyx":42
 *                 d = fromDate
 *                 while d < toDate:
 *                     if self.isBizDay(d):             # <<<<<<<<<<<<<<
 *                         wd += 1
 *                     d += 1
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isBizDay(__pyx_v_self, __pyx_v_d, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 42, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_9) {

          /* "ultron/tradingday/Calendar.pyx":43
 *                 while d < toDate:
 *                     if self.isBizDay(d):
 *                         wd += 1             # <<<<<<<<<<<<<<
 *                     d += 1
 *                 if self.isBizDay(toDate):
 */
          __pyx_v_wd = (__pyx_v_wd + 1);

          /* "ultron/tradingday/Calendar.pyx":42
 *                 d = fromDate
 *                 while d < toDate:
 *                     if self.isBizDay(d):             # <<<<<<<<<<<<<<
 *                         wd += 1
 *                     d += 1
 */
        }

        /* "ultron/tradingday/Calendar.pyx":44
 *                     if self.isBizDay(d):
 *                         wd += 1
 *                     d += 1             # <<<<<<<<<<<<<<
 *                 if self.isBizDay(toDate):
 *                     wd += 1
 */
        __pyx_t_1 = PyNumber_InPlaceAdd(((PyObject *)__pyx_v_d), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 44, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_d, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
        __pyx_t_1 = 0;
      }

      /* "ultron/tradingday/Calendar.pyx":45
 *                         wd += 1
 *                     d += 1
 *                 if self.isBizDay(toDate):             # <<<<<<<<<<<<<<
 *                     wd += 1
 *             elif fromDate > toDate:
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isBizDay(__pyx_v_self, __pyx_v_toDate, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 45, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_9) {

        /* "ultron/tradingday/Calendar.pyx":46
 *                     d += 1
 *                 if self.isBizDay(toDate):
 *                     wd += 1             # <<<<<<<<<<<<<<
 *             elif fromDate > toDate:
 *                 d = toDate
 */
        __pyx_v_wd = (__pyx_v_wd + 1);

        /* "ultron/tradingday/Calendar.pyx":45
 *                         wd += 1
 *                     d += 1
 *                 if self.isBizDay(toDate):             # <<<<<<<<<<<<<<
 *                     wd += 1
 *             elif fromDate > toDate:
 */
      }

      /* "ultron/tradingday/Calendar.pyx":39
 * 
 *         if fromDate != toDate:
 *             if fromDate < toDate:             # <<<<<<<<<<<<<<
 *                 d = fromDate
 *                 while d < toDate:
 */
      goto __pyx_L4;
    }

    /* "ultron/tradingday/Calendar.pyx":47
 *                 if self.isBizDay(toDate):
 *                     wd += 1
 *             elif fromDate > toDate:             # <<<<<<<<<<<<<<
 *                 d = toDate
 *                 while d < fromDate:
 */
    __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_fromDate), ((PyObject *)__pyx_v_toDate), Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_9) {

      /* "ultron/tradingday/Calendar.pyx":48
 *                     wd += 1
 *             elif fromDate > toDate:
 *                 d = toDate             # <<<<<<<<<<<<<<
 *                 while d < fromDate:
 *                     if self.isBizDay(d):
 */
      __Pyx_INCREF(((PyObject *)__pyx_v_toDate));
      __pyx_v_d = __pyx_v_toDate;

      /* "ultron/tradingday/Calendar.pyx":49
 *             elif fromDate > toDate:
 *                 d = toDate
 *                 while d < fromDate:             # <<<<<<<<<<<<<<
 *                     if self.isBizDay(d):
 *                         wd += 1
 */
      while (1) {
        __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_v_fromDate), Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 49, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (!__pyx_t_9) break;

        /* "ultron/tradingday/Calendar.pyx":50
 *                 d = toDate
 *                 while d < fromDate:
 *                     if self.isBizDay(d):             # <<<<<<<<<<<<<<
 *                         wd += 1
 *                     d += 1
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isBizDay(__pyx_v_self, __pyx_v_d, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 50, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_9) {

          /* "ultron/tradingday/Calendar.pyx":51
 *                 while d < fromDate:
 *                     if self.isBizDay(d):
 *                         wd += 1             # <<<<<<<<<<<<<<
 *                     d += 1
 *                 if self.isBizDay(fromDate):
 */
          __pyx_v_wd = (__pyx_v_wd + 1);

          /* "ultron/tradingday/Calendar.pyx":50
 *                 d = toDate
 *                 while d < fromDate:
 *                     if self.isBizDay(d):             # <<<<<<<<<<<<<<
 *                         wd += 1
 *                     d += 1
 */
        }

        /* "ultron/tradingday/Calendar.pyx":52
 *                     if self.isBizDay(d):
 *                         wd += 1
 *                     d += 1             # <<<<<<<<<<<<<<
 *                 if self.isBizDay(fromDate):
 *                     wd += 1
 */
        __pyx_t_1 = PyNumber_InPlaceAdd(((PyObject *)__pyx_v_d), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 52, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_d, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
        __pyx_t_1 = 0;
      }

      /* "ultron/tradingday/Calendar.pyx":53
 *                         wd += 1
 *                     d += 1
 *                 if self.isBizDay(fromDate):             # <<<<<<<<<<<<<<
 *                     wd += 1
 *             if self.isBizDay(fromDate) and not includeFirst:
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isBizDay(__pyx_v_self, __pyx_v_fromDate, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 53, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_9) {

        /* "ultron/tradingday/Calendar.pyx":54
 *                     d += 1
 *                 if self.isBizDay(fromDate):
 *                     wd += 1             # <<<<<<<<<<<<<<
 *             if self.isBizDay(fromDate) and not includeFirst:
 *                 wd -= 1
 */
        __pyx_v_wd = (__pyx_v_wd + 1);

        /* "ultron/tradingday/Calendar.pyx":53
 *                         wd += 1
 *                     d += 1
 *                 if self.isBizDay(fromDate):             # <<<<<<<<<<<<<<
 *                     wd += 1
 *             if self.isBizDay(fromDate) and not includeFirst:
 */
      }

      /* "ultron/tradingday/Calendar.pyx":47
 *                 if self.isBizDay(toDate):
 *                     wd += 1
 *             elif fromDate > toDate:             # <<<<<<<<<<<<<<
 *                 d = toDate
 *                 while d < fromDate:
 */
    }
    __pyx_L4:;

    /* "ultron/tradingday/Calendar.pyx":55
 *                 if self.isBizDay(fromDate):
 *                     wd += 1
 *             if self.isBizDay(fromDate) and not includeFirst:             # <<<<<<<<<<<<<<
 *                 wd -= 1
 *             if self.isBizDay(toDate) and not includeLast:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isBizDay(__pyx_v_self, __pyx_v_fromDate, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_10 = ((!(__pyx_v_includeFirst != 0)) != 0);
    __pyx_t_9 = __pyx_t_10;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_9) {

      /* "ultron/tradingday/Calendar.pyx":56
 *                     wd += 1
 *             if self.isBizDay(fromDate) and not includeFirst:
 *                 wd -= 1             # <<<<<<<<<<<<<<
 *             if self.isBizDay(toDate) and not includeLast:
 *                 wd -= 1
 */
      __pyx_v_wd = (__pyx_v_wd - 1);

      /* "ultron/tradingday/Calendar.pyx":55
 *                 if self.isBizDay(fromDate):
 *                     wd += 1
 *             if self.isBizDay(fromDate) and not includeFirst:             # <<<<<<<<<<<<<<
 *                 wd -= 1
 *             if self.isBizDay(toDate) and not includeLast:
 */
    }

    /* "ultron/tradingday/Calendar.pyx":57
 *             if self.isBizDay(fromDate) and not includeFirst:
 *                 wd -= 1
 *             if self.isBizDay(toDate) and not includeLast:             # <<<<<<<<<<<<<<
 *                 wd -= 1
 *         return wd
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isBizDay(__pyx_v_self, __pyx_v_toDate, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 57, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L17_bool_binop_done;
    }
    __pyx_t_10 = ((!(__pyx_v_includeLast != 0)) != 0);
    __pyx_t_9 = __pyx_t_10;
    __pyx_L17_bool_binop_done:;
    if (__pyx_t_9) {

      /* "ultron/tradingday/Calendar.pyx":58
 *                 wd -= 1
 *             if self.isBizDay(toDate) and not includeLast:
 *                 wd -= 1             # <<<<<<<<<<<<<<
 *         return wd
 * 
 */
      __pyx_v_wd = (__pyx_v_wd - 1);

      /* "ultron/tradingday/Calendar.pyx":57
 *             if self.isBizDay(fromDate) and not includeFirst:
 *                 wd -= 1
 *             if self.isBizDay(toDate) and not includeLast:             # <<<<<<<<<<<<<<
 *                 wd -= 1
 *         return wd
 */
    }

    /* "ultron/tradingday/Calendar.pyx":38
 *         cdef Date d
 * 
 *         if fromDate != toDate:             # <<<<<<<<<<<<<<
 *             if fromDate < toDate:
 *                 d = fromDate
 */
  }

  /* "ultron/tradingday/Calendar.pyx":59
 *             if self.isBizDay(toDate) and not includeLast:
 *                 wd -= 1
 *         return wd             # <<<<<<<<<<<<<<
 * 
 *     cpdef adjustDate(self, Date d, int c=BizDayConventions.Following):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_wd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":34
 *         return self.adjustDate(Date.endOfMonth(d), BizDayConventions.Preceding)
 * 
 *     cpdef bizDaysBetween(self, Date fromDate, Date toDate, bint includeFirst=True, bint includeLast=False):             # <<<<<<<<<<<<<<
 *         cdef int wd = 0
 *         cdef Date d
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.bizDaysBetween", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_13bizDaysBetween(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_12bizDaysBetween[] = "Calendar.bizDaysBetween(self, Date fromDate, Date toDate, bool includeFirst=True, bool includeLast=False)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_13bizDaysBetween(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate = 0;
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate = 0;
  int __pyx_v_includeFirst;
  int __pyx_v_includeLast;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bizDaysBetween (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fromDate,&__pyx_n_s_toDate,&__pyx_n_s_includeFirst,&__pyx_n_s_includeLast,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fromDate)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_toDate)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bizDaysBetween", 0, 2, 4, 1); __PYX_ERR(0, 34, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_includeFirst);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_includeLast);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bizDaysBetween") < 0)) __PYX_ERR(0, 34, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fromDate = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)values[0]);
    __pyx_v_toDate = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)values[1]);
    if (values[2]) {
      __pyx_v_includeFirst = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_includeFirst == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L3_error)
    } else {
      __pyx_v_includeFirst = ((int)1);
    }
    if (values[3]) {
      __pyx_v_includeLast = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_includeLast == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L3_error)
    } else {
      __pyx_v_includeLast = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bizDaysBetween", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 34, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.bizDaysBetween", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fromDate), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "fromDate", 0))) __PYX_ERR(0, 34, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_toDate), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "toDate", 0))) __PYX_ERR(0, 34, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_12bizDaysBetween(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), __pyx_v_fromDate, __pyx_v_toDate, __pyx_v_includeFirst, __pyx_v_includeLast);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_12bizDaysBetween(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_v_includeFirst, int __pyx_v_includeLast) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween __pyx_t_2;
  __Pyx_RefNannySetupContext("bizDaysBetween", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.includeFirst = __pyx_v_includeFirst;
  __pyx_t_2.includeLast = __pyx_v_includeLast;
  __pyx_t_1 = __pyx_vtabptr_6ultron_10tradingday_8Calendar_Calendar->bizDaysBetween(__pyx_v_self, __pyx_v_fromDate, __pyx_v_toDate, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.bizDaysBetween", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":61
 *         return wd
 * 
 *     cpdef adjustDate(self, Date d, int c=BizDayConventions.Following):             # <<<<<<<<<<<<<<
 * 
 *         cdef Date d1
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_15adjustDate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_adjustDate(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate *__pyx_optional_args) {
  int __pyx_v_c = __pyx_k_;
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d1 = 0;
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d2 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("adjustDate", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_c = __pyx_optional_args->c;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_adjustDate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_15adjustDate)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_c); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 61, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_d), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_d), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 61, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_d));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_d));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_d));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":66
 *         cdef Date d2
 * 
 *         if c == BizDayConventions.Unadjusted:             # <<<<<<<<<<<<<<
 *             return d
 *         d1 = d
 */
  __pyx_t_8 = ((__pyx_v_c == Unadjusted) != 0);
  if (__pyx_t_8) {

    /* "ultron/tradingday/Calendar.pyx":67
 * 
 *         if c == BizDayConventions.Unadjusted:
 *             return d             # <<<<<<<<<<<<<<
 *         d1 = d
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_d));
    __pyx_r = ((PyObject *)__pyx_v_d);
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":66
 *         cdef Date d2
 * 
 *         if c == BizDayConventions.Unadjusted:             # <<<<<<<<<<<<<<
 *             return d
 *         d1 = d
 */
  }

  /* "ultron/tradingday/Calendar.pyx":68
 *         if c == BizDayConventions.Unadjusted:
 *             return d
 *         d1 = d             # <<<<<<<<<<<<<<
 * 
 *         if c == BizDayConventions.Following or c == BizDayConventions.ModifiedFollowing or \
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_d));
  __pyx_v_d1 = __pyx_v_d;

  /* "ultron/tradingday/Calendar.pyx":70
 *         d1 = d
 * 
 *         if c == BizDayConventions.Following or c == BizDayConventions.ModifiedFollowing or \             # <<<<<<<<<<<<<<
 *                 c == BizDayConventions.HalfMonthModifiedFollowing:
 *             while self.isHoliday(d1):
 */
  switch (__pyx_v_c) {
    case Following:
    case ModifiedFollowing:
    case HalfMonthModifiedFollowing:

    /* "ultron/tradingday/Calendar.pyx":72
 *         if c == BizDayConventions.Following or c == BizDayConventions.ModifiedFollowing or \
 *                 c == BizDayConventions.HalfMonthModifiedFollowing:
 *             while self.isHoliday(d1):             # <<<<<<<<<<<<<<
 *                 d1 += 1
 *             if c == BizDayConventions.ModifiedFollowing or c == BizDayConventions.HalfMonthModifiedFollowing:
 */
    while (1) {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isHoliday(__pyx_v_self, __pyx_v_d1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 72, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!__pyx_t_8) break;

      /* "ultron/tradingday/Calendar.pyx":73
 *                 c == BizDayConventions.HalfMonthModifiedFollowing:
 *             while self.isHoliday(d1):
 *                 d1 += 1             # <<<<<<<<<<<<<<
 *             if c == BizDayConventions.ModifiedFollowing or c == BizDayConventions.HalfMonthModifiedFollowing:
 *                 if d1.month() != d.month():
 */
      __pyx_t_1 = PyNumber_InPlaceAdd(((PyObject *)__pyx_v_d1), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 73, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_d1, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
      __pyx_t_1 = 0;
    }

    /* "ultron/tradingday/Calendar.pyx":74
 *             while self.isHoliday(d1):
 *                 d1 += 1
 *             if c == BizDayConventions.ModifiedFollowing or c == BizDayConventions.HalfMonthModifiedFollowing:             # <<<<<<<<<<<<<<
 *                 if d1.month() != d.month():
 *                     return self.adjustDate(d, BizDayConventions.Preceding)
 */
    switch (__pyx_v_c) {
      case ModifiedFollowing:
      case HalfMonthModifiedFollowing:

      /* "ultron/tradingday/Calendar.pyx":75
 *                 d1 += 1
 *             if c == BizDayConventions.ModifiedFollowing or c == BizDayConventions.HalfMonthModifiedFollowing:
 *                 if d1.month() != d.month():             # <<<<<<<<<<<<<<
 *                     return self.adjustDate(d, BizDayConventions.Preceding)
 *                 if c == BizDayConventions.HalfMonthModifiedFollowing:
 */
      __pyx_t_8 = ((((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_d1->__pyx_vtab)->month(__pyx_v_d1, 0) != ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_d->__pyx_vtab)->month(__pyx_v_d, 0)) != 0);
      if (__pyx_t_8) {

        /* "ultron/tradingday/Calendar.pyx":76
 *             if c == BizDayConventions.ModifiedFollowing or c == BizDayConventions.HalfMonthModifiedFollowing:
 *                 if d1.month() != d.month():
 *                     return self.adjustDate(d, BizDayConventions.Preceding)             # <<<<<<<<<<<<<<
 *                 if c == BizDayConventions.HalfMonthModifiedFollowing:
 *                     if d.dayOfMonth() <= 15 < d1.dayOfMonth():
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_9.__pyx_n = 1;
        __pyx_t_9.c = Preceding;
        __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->adjustDate(__pyx_v_self, __pyx_v_d, 0, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L0;

        /* "ultron/tradingday/Calendar.pyx":75
 *                 d1 += 1
 *             if c == BizDayConventions.ModifiedFollowing or c == BizDayConventions.HalfMonthModifiedFollowing:
 *                 if d1.month() != d.month():             # <<<<<<<<<<<<<<
 *                     return self.adjustDate(d, BizDayConventions.Preceding)
 *                 if c == BizDayConventions.HalfMonthModifiedFollowing:
 */
      }

      /* "ultron/tradingday/Calendar.pyx":77
 *                 if d1.month() != d.month():
 *                     return self.adjustDate(d, BizDayConventions.Preceding)
 *                 if c == BizDayConventions.HalfMonthModifiedFollowing:             # <<<<<<<<<<<<<<
 *                     if d.dayOfMonth() <= 15 < d1.dayOfMonth():
 *                         return self.adjustDate(d, BizDayConventions.Preceding)
 */
      __pyx_t_8 = ((__pyx_v_c == HalfMonthModifiedFollowing) != 0);
      if (__pyx_t_8) {

        /* "ultron/tradingday/Calendar.pyx":78
 *                     return self.adjustDate(d, BizDayConventions.Preceding)
 *                 if c == BizDayConventions.HalfMonthModifiedFollowing:
 *                     if d.dayOfMonth() <= 15 < d1.dayOfMonth():             # <<<<<<<<<<<<<<
 *                         return self.adjustDate(d, BizDayConventions.Preceding)
 *         elif c == BizDayConventions.Preceding or c == BizDayConventions.ModifiedPreceding:
 */
        __pyx_t_8 = (((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_d->__pyx_vtab)->dayOfMonth(__pyx_v_d, 0) <= 15);
        if (__pyx_t_8) {
          __pyx_t_8 = (15 < ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_d1->__pyx_vtab)->dayOfMonth(__pyx_v_d1, 0));
        }
        __pyx_t_10 = (__pyx_t_8 != 0);
        if (__pyx_t_10) {

          /* "ultron/tradingday/Calendar.pyx":79
 *                 if c == BizDayConventions.HalfMonthModifiedFollowing:
 *                     if d.dayOfMonth() <= 15 < d1.dayOfMonth():
 *                         return self.adjustDate(d, BizDayConventions.Preceding)             # <<<<<<<<<<<<<<
 *         elif c == BizDayConventions.Preceding or c == BizDayConventions.ModifiedPreceding:
 *             while self.isHoliday(d1):
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_9.__pyx_n = 1;
          __pyx_t_9.c = Preceding;
          __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->adjustDate(__pyx_v_self, __pyx_v_d, 0, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_r = __pyx_t_1;
          __pyx_t_1 = 0;
          goto __pyx_L0;

          /* "ultron/tradingday/Calendar.pyx":78
 *                     return self.adjustDate(d, BizDayConventions.Preceding)
 *                 if c == BizDayConventions.HalfMonthModifiedFollowing:
 *                     if d.dayOfMonth() <= 15 < d1.dayOfMonth():             # <<<<<<<<<<<<<<
 *                         return self.adjustDate(d, BizDayConventions.Preceding)
 *         elif c == BizDayConventions.Preceding or c == BizDayConventions.ModifiedPreceding:
 */
        }

        /* "ultron/tradingday/Calendar.pyx":77
 *                 if d1.month() != d.month():
 *                     return self.adjustDate(d, BizDayConventions.Preceding)
 *                 if c == BizDayConventions.HalfMonthModifiedFollowing:             # <<<<<<<<<<<<<<
 *                     if d.dayOfMonth() <= 15 < d1.dayOfMonth():
 *                         return self.adjustDate(d, BizDayConventions.Preceding)
 */
      }

      /* "ultron/tradingday/Calendar.pyx":74
 *             while self.isHoliday(d1):
 *                 d1 += 1
 *             if c == BizDayConventions.ModifiedFollowing or c == BizDayConventions.HalfMonthModifiedFollowing:             # <<<<<<<<<<<<<<
 *                 if d1.month() != d.month():
 *                     return self.adjustDate(d, BizDayConventions.Preceding)
 */
      break;
      default: break;
    }

    /* "ultron/tradingday/Calendar.pyx":70
 *         d1 = d
 * 
 *         if c == BizDayConventions.Following or c == BizDayConventions.ModifiedFollowing or \             # <<<<<<<<<<<<<<
 *                 c == BizDayConventions.HalfMonthModifiedFollowing:
 *             while self.isHoliday(d1):
 */
    break;
    case Preceding:

    /* "ultron/tradingday/Calendar.pyx":80
 *                     if d.dayOfMonth() <= 15 < d1.dayOfMonth():
 *                         return self.adjustDate(d, BizDayConventions.Preceding)
 *         elif c == BizDayConventions.Preceding or c == BizDayConventions.ModifiedPreceding:             # <<<<<<<<<<<<<<
 *             while self.isHoliday(d1):
 *                 d1 -= 1
 */
    case ModifiedPreceding:

    /* "ultron/tradingday/Calendar.pyx":81
 *                         return self.adjustDate(d, BizDayConventions.Preceding)
 *         elif c == BizDayConventions.Preceding or c == BizDayConventions.ModifiedPreceding:
 *             while self.isHoliday(d1):             # <<<<<<<<<<<<<<
 *                 d1 -= 1
 *             if c == BizDayConventions.ModifiedPreceding and d1.month() != d.month():
 */
    while (1) {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isHoliday(__pyx_v_self, __pyx_v_d1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 81, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!__pyx_t_10) break;

      /* "ultron/tradingday/Calendar.pyx":82
 *         elif c == BizDayConventions.Preceding or c == BizDayConventions.ModifiedPreceding:
 *             while self.isHoliday(d1):
 *                 d1 -= 1             # <<<<<<<<<<<<<<
 *             if c == BizDayConventions.ModifiedPreceding and d1.month() != d.month():
 *                 return self.adjustDate(d, BizDayConventions.Following)
 */
      __pyx_t_1 = PyNumber_InPlaceSubtract(((PyObject *)__pyx_v_d1), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_d1, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
      __pyx_t_1 = 0;
    }

    /* "ultron/tradingday/Calendar.pyx":83
 *             while self.isHoliday(d1):
 *                 d1 -= 1
 *             if c == BizDayConventions.ModifiedPreceding and d1.month() != d.month():             # <<<<<<<<<<<<<<
 *                 return self.adjustDate(d, BizDayConventions.Following)
 *         elif c == BizDayConventions.Nearest:
 */
    __pyx_t_8 = ((__pyx_v_c == ModifiedPreceding) != 0);
    if (__pyx_t_8) {
    } else {
      __pyx_t_10 = __pyx_t_8;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_8 = ((((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_d1->__pyx_vtab)->month(__pyx_v_d1, 0) != ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_d->__pyx_vtab)->month(__pyx_v_d, 0)) != 0);
    __pyx_t_10 = __pyx_t_8;
    __pyx_L12_bool_binop_done:;
    if (__pyx_t_10) {

      /* "ultron/tradingday/Calendar.pyx":84
 *                 d1 -= 1
 *             if c == BizDayConventions.ModifiedPreceding and d1.month() != d.month():
 *                 return self.adjustDate(d, BizDayConventions.Following)             # <<<<<<<<<<<<<<
 *         elif c == BizDayConventions.Nearest:
 *             d2 = d
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_9.__pyx_n = 1;
      __pyx_t_9.c = Following;
      __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->adjustDate(__pyx_v_self, __pyx_v_d, 0, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "ultron/tradingday/Calendar.pyx":83
 *             while self.isHoliday(d1):
 *                 d1 -= 1
 *             if c == BizDayConventions.ModifiedPreceding and d1.month() != d.month():             # <<<<<<<<<<<<<<
 *                 return self.adjustDate(d, BizDayConventions.Following)
 *         elif c == BizDayConventions.Nearest:
 */
    }

    /* "ultron/tradingday/Calendar.pyx":80
 *                     if d.dayOfMonth() <= 15 < d1.dayOfMonth():
 *                         return self.adjustDate(d, BizDayConventions.Preceding)
 *         elif c == BizDayConventions.Preceding or c == BizDayConventions.ModifiedPreceding:             # <<<<<<<<<<<<<<
 *             while self.isHoliday(d1):
 *                 d1 -= 1
 */
    break;
    case Nearest:

    /* "ultron/tradingday/Calendar.pyx":86
 *                 return self.adjustDate(d, BizDayConventions.Following)
 *         elif c == BizDayConventions.Nearest:
 *             d2 = d             # <<<<<<<<<<<<<<
 *             while self.isHoliday(d1) and self.isHoliday(d2):
 *                 d1 += 1
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_d));
    __pyx_v_d2 = __pyx_v_d;

    /* "ultron/tradingday/Calendar.pyx":87
 *         elif c == BizDayConventions.Nearest:
 *             d2 = d
 *             while self.isHoliday(d1) and self.isHoliday(d2):             # <<<<<<<<<<<<<<
 *                 d1 += 1
 *                 d2 -= 1
 */
    while (1) {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isHoliday(__pyx_v_self, __pyx_v_d1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 87, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_8) {
      } else {
        __pyx_t_10 = __pyx_t_8;
        goto __pyx_L16_bool_binop_done;
      }
      __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isHoliday(__pyx_v_self, __pyx_v_d2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 87, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = __pyx_t_8;
      __pyx_L16_bool_binop_done:;
      if (!__pyx_t_10) break;

      /* "ultron/tradingday/Calendar.pyx":88
 *             d2 = d
 *             while self.isHoliday(d1) and self.isHoliday(d2):
 *                 d1 += 1             # <<<<<<<<<<<<<<
 *                 d2 -= 1
 * 
 */
      __pyx_t_1 = PyNumber_InPlaceAdd(((PyObject *)__pyx_v_d1), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 88, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_d1, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "ultron/tradingday/Calendar.pyx":89
 *             while self.isHoliday(d1) and self.isHoliday(d2):
 *                 d1 += 1
 *                 d2 -= 1             # <<<<<<<<<<<<<<
 * 
 *             if self.isHoliday(d1):
 */
      __pyx_t_1 = PyNumber_InPlaceSubtract(((PyObject *)__pyx_v_d2), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 89, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_d2, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
      __pyx_t_1 = 0;
    }

    /* "ultron/tradingday/Calendar.pyx":91
 *                 d2 -= 1
 * 
 *             if self.isHoliday(d1):             # <<<<<<<<<<<<<<
 *                 return d2
 *             else:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isHoliday(__pyx_v_self, __pyx_v_d1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_10) {

      /* "ultron/tradingday/Calendar.pyx":92
 * 
 *             if self.isHoliday(d1):
 *                 return d2             # <<<<<<<<<<<<<<
 *             else:
 *                 return d1
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_d2));
      __pyx_r = ((PyObject *)__pyx_v_d2);
      goto __pyx_L0;

      /* "ultron/tradingday/Calendar.pyx":91
 *                 d2 -= 1
 * 
 *             if self.isHoliday(d1):             # <<<<<<<<<<<<<<
 *                 return d2
 *             else:
 */
    }

    /* "ultron/tradingday/Calendar.pyx":94
 *                 return d2
 *             else:
 *                 return d1             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("unknown business-day convention")
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_d1));
      __pyx_r = ((PyObject *)__pyx_v_d1);
      goto __pyx_L0;
    }

    /* "ultron/tradingday/Calendar.pyx":85
 *             if c == BizDayConventions.ModifiedPreceding and d1.month() != d.month():
 *                 return self.adjustDate(d, BizDayConventions.Following)
 *         elif c == BizDayConventions.Nearest:             # <<<<<<<<<<<<<<
 *             d2 = d
 *             while self.isHoliday(d1) and self.isHoliday(d2):
 */
    break;
    default:

    /* "ultron/tradingday/Calendar.pyx":96
 *                 return d1
 *         else:
 *             raise ValueError("unknown business-day convention")             # <<<<<<<<<<<<<<
 *         return d1
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 96, __pyx_L1_error)
    break;
  }

  /* "ultron/tradingday/Calendar.pyx":97
 *         else:
 *             raise ValueError("unknown business-day convention")
 *         return d1             # <<<<<<<<<<<<<<
 * 
 *     cpdef advanceDate(self, Date d, Period period, int c=BizDayConventions.Following, bint endOfMonth=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_d1));
  __pyx_r = ((PyObject *)__pyx_v_d1);
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":61
 *         return wd
 * 
 *     cpdef adjustDate(self, Date d, int c=BizDayConventions.Following):             # <<<<<<<<<<<<<<
 * 
 *         cdef Date d1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.adjustDate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_d1);
  __Pyx_XDECREF((PyObject *)__pyx_v_d2);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_15adjustDate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_14adjustDate[] = "Calendar.adjustDate(self, Date d, int c=BizDayConventions.Following)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_15adjustDate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d = 0;
  int __pyx_v_c;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("adjustDate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_d,&__pyx_n_s_c,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "adjustDate") < 0)) __PYX_ERR(0, 61, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_d = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)values[0]);
    if (values[1]) {
      __pyx_v_c = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_c == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
    } else {
      __pyx_v_c = __pyx_k_;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("adjustDate", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 61, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.adjustDate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_d), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "d", 0))) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_14adjustDate(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), __pyx_v_d, __pyx_v_c);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_14adjustDate(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, int __pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate __pyx_t_2;
  __Pyx_RefNannySetupContext("adjustDate", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.c = __pyx_v_c;
  __pyx_t_1 = __pyx_vtabptr_6ultron_10tradingday_8Calendar_Calendar->adjustDate(__pyx_v_self, __pyx_v_d, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.adjustDate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":99
 *         return d1
 * 
 *     cpdef advanceDate(self, Date d, Period period, int c=BizDayConventions.Following, bint endOfMonth=False):             # <<<<<<<<<<<<<<
 * 
 *         cdef int n
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_17advanceDate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_advanceDate(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, struct __pyx_obj_6ultron_10tradingday_6Period_Period *__pyx_v_period, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_advanceDate *__pyx_optional_args) {
  int __pyx_v_c = __pyx_k__3;
  int __pyx_v_endOfMonth = ((int)0);
  int __pyx_v_n;
  int __pyx_v_units;
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d1 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("advanceDate", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_c = __pyx_optional_args->c;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_endOfMonth = __pyx_optional_args->endOfMonth;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_advanceDate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_17advanceDate)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_c); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_endOfMonth); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[5] = {__pyx_t_6, ((PyObject *)__pyx_v_d), ((PyObject *)__pyx_v_period), __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[5] = {__pyx_t_6, ((PyObject *)__pyx_v_d), ((PyObject *)__pyx_v_period), __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 99, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_d));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_d));
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, ((PyObject *)__pyx_v_d));
          __Pyx_INCREF(((PyObject *)__pyx_v_period));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_period));
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, ((PyObject *)__pyx_v_period));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":105
 *         cdef Date d1
 * 
 *         n = period.length()             # <<<<<<<<<<<<<<
 *         units = period.units()
 * 
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_6ultron_10tradingday_6Period_Period *)__pyx_v_period->__pyx_vtab)->length(__pyx_v_period, 0);

  /* "ultron/tradingday/Calendar.pyx":106
 * 
 *         n = period.length()
 *         units = period.units()             # <<<<<<<<<<<<<<
 * 
 *         if n == 0:
 */
  __pyx_v_units = ((struct __pyx_vtabstruct_6ultron_10tradingday_6Period_Period *)__pyx_v_period->__pyx_vtab)->units(__pyx_v_period, 0);

  /* "ultron/tradingday/Calendar.pyx":108
 *         units = period.units()
 * 
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return self.adjustDate(d, c)
 *         elif units == TimeUnits.BDays:
 */
  __pyx_t_9 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_9) {

    /* "ultron/tradingday/Calendar.pyx":109
 * 
 *         if n == 0:
 *             return self.adjustDate(d, c)             # <<<<<<<<<<<<<<
 *         elif units == TimeUnits.BDays:
 *             d1 = d
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10.__pyx_n = 1;
    __pyx_t_10.c = __pyx_v_c;
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->adjustDate(__pyx_v_self, __pyx_v_d, 0, &__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":108
 *         units = period.units()
 * 
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return self.adjustDate(d, c)
 *         elif units == TimeUnits.BDays:
 */
  }

  /* "ultron/tradingday/Calendar.pyx":110
 *         if n == 0:
 *             return self.adjustDate(d, c)
 *         elif units == TimeUnits.BDays:             # <<<<<<<<<<<<<<
 *             d1 = d
 *             if n > 0:
 */
  __pyx_t_9 = ((__pyx_v_units == BDays) != 0);
  if (__pyx_t_9) {

    /* "ultron/tradingday/Calendar.pyx":111
 *             return self.adjustDate(d, c)
 *         elif units == TimeUnits.BDays:
 *             d1 = d             # <<<<<<<<<<<<<<
 *             if n > 0:
 *                 while n > 0:
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_d));
    __pyx_v_d1 = __pyx_v_d;

    /* "ultron/tradingday/Calendar.pyx":112
 *         elif units == TimeUnits.BDays:
 *             d1 = d
 *             if n > 0:             # <<<<<<<<<<<<<<
 *                 while n > 0:
 *                     d1 += 1
 */
    __pyx_t_9 = ((__pyx_v_n > 0) != 0);
    if (__pyx_t_9) {

      /* "ultron/tradingday/Calendar.pyx":113
 *             d1 = d
 *             if n > 0:
 *                 while n > 0:             # <<<<<<<<<<<<<<
 *                     d1 += 1
 *                     while self.isHoliday(d1):
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_n > 0) != 0);
        if (!__pyx_t_9) break;

        /* "ultron/tradingday/Calendar.pyx":114
 *             if n > 0:
 *                 while n > 0:
 *                     d1 += 1             # <<<<<<<<<<<<<<
 *                     while self.isHoliday(d1):
 *                         d1 += 1
 */
        __pyx_t_1 = PyNumber_InPlaceAdd(((PyObject *)__pyx_v_d1), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 114, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_d1, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "ultron/tradingday/Calendar.pyx":115
 *                 while n > 0:
 *                     d1 += 1
 *                     while self.isHoliday(d1):             # <<<<<<<<<<<<<<
 *                         d1 += 1
 *                     n -= 1
 */
        while (1) {
          __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isHoliday(__pyx_v_self, __pyx_v_d1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 115, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_9) break;

          /* "ultron/tradingday/Calendar.pyx":116
 *                     d1 += 1
 *                     while self.isHoliday(d1):
 *                         d1 += 1             # <<<<<<<<<<<<<<
 *                     n -= 1
 *             else:
 */
          __pyx_t_1 = PyNumber_InPlaceAdd(((PyObject *)__pyx_v_d1), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 116, __pyx_L1_error)
          __Pyx_DECREF_SET(__pyx_v_d1, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
          __pyx_t_1 = 0;
        }

        /* "ultron/tradingday/Calendar.pyx":117
 *                     while self.isHoliday(d1):
 *                         d1 += 1
 *                     n -= 1             # <<<<<<<<<<<<<<
 *             else:
 *                 while n < 0:
 */
        __pyx_v_n = (__pyx_v_n - 1);
      }

      /* "ultron/tradingday/Calendar.pyx":112
 *         elif units == TimeUnits.BDays:
 *             d1 = d
 *             if n > 0:             # <<<<<<<<<<<<<<
 *                 while n > 0:
 *                     d1 += 1
 */
      goto __pyx_L4;
    }

    /* "ultron/tradingday/Calendar.pyx":119
 *                     n -= 1
 *             else:
 *                 while n < 0:             # <<<<<<<<<<<<<<
 *                     d1 -= 1
 *                     while self.isHoliday(d1):
 */
    /*else*/ {
      while (1) {
        __pyx_t_9 = ((__pyx_v_n < 0) != 0);
        if (!__pyx_t_9) break;

        /* "ultron/tradingday/Calendar.pyx":120
 *             else:
 *                 while n < 0:
 *                     d1 -= 1             # <<<<<<<<<<<<<<
 *                     while self.isHoliday(d1):
 *                         d1 -= 1
 */
        __pyx_t_1 = PyNumber_InPlaceSubtract(((PyObject *)__pyx_v_d1), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_d1, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "ultron/tradingday/Calendar.pyx":121
 *                 while n < 0:
 *                     d1 -= 1
 *                     while self.isHoliday(d1):             # <<<<<<<<<<<<<<
 *                         d1 -= 1
 *                     n += 1
 */
        while (1) {
          __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isHoliday(__pyx_v_self, __pyx_v_d1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 121, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_9) break;

          /* "ultron/tradingday/Calendar.pyx":122
 *                     d1 -= 1
 *                     while self.isHoliday(d1):
 *                         d1 -= 1             # <<<<<<<<<<<<<<
 *                     n += 1
 *             return d1
 */
          __pyx_t_1 = PyNumber_InPlaceSubtract(((PyObject *)__pyx_v_d1), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 122, __pyx_L1_error)
          __Pyx_DECREF_SET(__pyx_v_d1, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
          __pyx_t_1 = 0;
        }

        /* "ultron/tradingday/Calendar.pyx":123
 *                     while self.isHoliday(d1):
 *                         d1 -= 1
 *                     n += 1             # <<<<<<<<<<<<<<
 *             return d1
 *         elif units == TimeUnits.Days or units == TimeUnits.Weeks:
 */
        __pyx_v_n = (__pyx_v_n + 1);
      }
    }
    __pyx_L4:;

    /* "ultron/tradingday/Calendar.pyx":124
 *                         d1 -= 1
 *                     n += 1
 *             return d1             # <<<<<<<<<<<<<<
 *         elif units == TimeUnits.Days or units == TimeUnits.Weeks:
 *             d1 = d + period
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_d1));
    __pyx_r = ((PyObject *)__pyx_v_d1);
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":110
 *         if n == 0:
 *             return self.adjustDate(d, c)
 *         elif units == TimeUnits.BDays:             # <<<<<<<<<<<<<<
 *             d1 = d
 *             if n > 0:
 */
  }

  /* "ultron/tradingday/Calendar.pyx":125
 *                     n += 1
 *             return d1
 *         elif units == TimeUnits.Days or units == TimeUnits.Weeks:             # <<<<<<<<<<<<<<
 *             d1 = d + period
 *             return self.adjustDate(d1, c)
 */
  switch (__pyx_v_units) {
    case Days:
    case Weeks:
    __pyx_t_9 = 1;
    break;
    default:
    __pyx_t_9 = 0;
    break;
  }
  if (__pyx_t_9) {

    /* "ultron/tradingday/Calendar.pyx":126
 *             return d1
 *         elif units == TimeUnits.Days or units == TimeUnits.Weeks:
 *             d1 = d + period             # <<<<<<<<<<<<<<
 *             return self.adjustDate(d1, c)
 *         else:
 */
    __pyx_t_1 = PyNumber_Add(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_v_period)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 126, __pyx_L1_error)
    __pyx_v_d1 = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "ultron/tradingday/Calendar.pyx":127
 *         elif units == TimeUnits.Days or units == TimeUnits.Weeks:
 *             d1 = d + period
 *             return self.adjustDate(d1, c)             # <<<<<<<<<<<<<<
 *         else:
 *             d1 = d + period
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10.__pyx_n = 1;
    __pyx_t_10.c = __pyx_v_c;
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->adjustDate(__pyx_v_self, __pyx_v_d1, 0, &__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":125
 *                     n += 1
 *             return d1
 *         elif units == TimeUnits.Days or units == TimeUnits.Weeks:             # <<<<<<<<<<<<<<
 *             d1 = d + period
 *             return self.adjustDate(d1, c)
 */
  }

  /* "ultron/tradingday/Calendar.pyx":129
 *             return self.adjustDate(d1, c)
 *         else:
 *             d1 = d + period             # <<<<<<<<<<<<<<
 *             if endOfMonth and self.isEndOfMonth(d):
 *                 return self.endOfMonth(d1)
 */
  /*else*/ {
    __pyx_t_1 = PyNumber_Add(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_v_period)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 129, __pyx_L1_error)
    __pyx_v_d1 = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "ultron/tradingday/Calendar.pyx":130
 *         else:
 *             d1 = d + period
 *             if endOfMonth and self.isEndOfMonth(d):             # <<<<<<<<<<<<<<
 *                 return self.endOfMonth(d1)
 *             return self.adjustDate(d1, c)
 */
    __pyx_t_11 = (__pyx_v_endOfMonth != 0);
    if (__pyx_t_11) {
    } else {
      __pyx_t_9 = __pyx_t_11;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isEndOfMonth(__pyx_v_self, __pyx_v_d, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __pyx_t_11;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_9) {

      /* "ultron/tradingday/Calendar.pyx":131
 *             d1 = d + period
 *             if endOfMonth and self.isEndOfMonth(d):
 *                 return self.endOfMonth(d1)             # <<<<<<<<<<<<<<
 *             return self.adjustDate(d1, c)
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->endOfMonth(__pyx_v_self, __pyx_v_d1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "ultron/tradingday/Calendar.pyx":130
 *         else:
 *             d1 = d + period
 *             if endOfMonth and self.isEndOfMonth(d):             # <<<<<<<<<<<<<<
 *                 return self.endOfMonth(d1)
 *             return self.adjustDate(d1, c)
 */
    }

    /* "ultron/tradingday/Calendar.pyx":132
 *             if endOfMonth and self.isEndOfMonth(d):
 *                 return self.endOfMonth(d1)
 *             return self.adjustDate(d1, c)             # <<<<<<<<<<<<<<
 * 
 *     cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=True):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10.__pyx_n = 1;
    __pyx_t_10.c = __pyx_v_c;
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->adjustDate(__pyx_v_self, __pyx_v_d1, 0, &__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "ultron/tradingday/Calendar.pyx":99
 *         return d1
 * 
 *     cpdef advanceDate(self, Date d, Period period, int c=BizDayConventions.Following, bint endOfMonth=False):             # <<<<<<<<<<<<<<
 * 
 *         cdef int n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.advanceDate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_d1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_17advanceDate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_16advanceDate[] = "Calendar.advanceDate(self, Date d, Period period, int c=BizDayConventions.Following, bool endOfMonth=False)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_17advanceDate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d = 0;
  struct __pyx_obj_6ultron_10tradingday_6Period_Period *__pyx_v_period = 0;
  int __pyx_v_c;
  int __pyx_v_endOfMonth;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("advanceDate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_d,&__pyx_n_s_period,&__pyx_n_s_c,&__pyx_n_s_endOfMonth,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_period)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("advanceDate", 0, 2, 4, 1); __PYX_ERR(0, 99, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_endOfMonth);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "advanceDate") < 0)) __PYX_ERR(0, 99, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_d = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)values[0]);
    __pyx_v_period = ((struct __pyx_obj_6ultron_10tradingday_6Period_Period *)values[1]);
    if (values[2]) {
      __pyx_v_c = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_c == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 99, __pyx_L3_error)
    } else {
      __pyx_v_c = __pyx_k__3;
    }
    if (values[3]) {
      __pyx_v_endOfMonth = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_endOfMonth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 99, __pyx_L3_error)
    } else {
      __pyx_v_endOfMonth = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("advanceDate", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 99, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.advanceDate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_d), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "d", 0))) __PYX_ERR(0, 99, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_period), __pyx_ptype_6ultron_10tradingday_6Period_Period, 1, "period", 0))) __PYX_ERR(0, 99, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_16advanceDate(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), __pyx_v_d, __pyx_v_period, __pyx_v_c, __pyx_v_endOfMonth);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_16advanceDate(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d, struct __pyx_obj_6ultron_10tradingday_6Period_Period *__pyx_v_period, int __pyx_v_c, int __pyx_v_endOfMonth) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_advanceDate __pyx_t_2;
  __Pyx_RefNannySetupContext("advanceDate", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.c = __pyx_v_c;
  __pyx_t_2.endOfMonth = __pyx_v_endOfMonth;
  __pyx_t_1 = __pyx_vtabptr_6ultron_10tradingday_8Calendar_Calendar->advanceDate(__pyx_v_self, __pyx_v_d, __pyx_v_period, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.advanceDate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":134
 *             return self.adjustDate(d1, c)
 * 
 *     cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=True):             # <<<<<<<<<<<<<<
 *         cdef list result = []
 *         cdef Date d = fromDate
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_19holDatesList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_holDatesList(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_holDatesList *__pyx_optional_args) {
  int __pyx_v_includeWeekEnds = ((int)1);
  PyObject *__pyx_v_result = 0;
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("holDatesList", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_includeWeekEnds = __pyx_optional_args->includeWeekEnds;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_holDatesList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_19holDatesList)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_includeWeekEnds); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_fromDate), ((PyObject *)__pyx_v_toDate), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_fromDate), ((PyObject *)__pyx_v_toDate), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 134, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_fromDate));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_fromDate));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_fromDate));
          __Pyx_INCREF(((PyObject *)__pyx_v_toDate));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_toDate));
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, ((PyObject *)__pyx_v_toDate));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":135
 * 
 *     cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=True):
 *         cdef list result = []             # <<<<<<<<<<<<<<
 *         cdef Date d = fromDate
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ultron/tradingday/Calendar.pyx":136
 *     cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=True):
 *         cdef list result = []
 *         cdef Date d = fromDate             # <<<<<<<<<<<<<<
 * 
 *         while d <= toDate:
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_fromDate));
  __pyx_v_d = __pyx_v_fromDate;

  /* "ultron/tradingday/Calendar.pyx":138
 *         cdef Date d = fromDate
 * 
 *         while d <= toDate:             # <<<<<<<<<<<<<<
 *             if self.isHoliday(d) and (includeWeekEnds or not self.isWeekEnd(d.weekday())):
 *                 result.append(d)
 */
  while (1) {
    __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_v_toDate), Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 138, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_8) break;

    /* "ultron/tradingday/Calendar.pyx":139
 * 
 *         while d <= toDate:
 *             if self.isHoliday(d) and (includeWeekEnds or not self.isWeekEnd(d.weekday())):             # <<<<<<<<<<<<<<
 *                 result.append(d)
 *             d += 1
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isHoliday(__pyx_v_self, __pyx_v_d, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_includeWeekEnds != 0);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isWeekEnd(__pyx_v_self, ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_d->__pyx_vtab)->weekday(__pyx_v_d, 0), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = ((!__pyx_t_9) != 0);
    __pyx_t_8 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_8) {

      /* "ultron/tradingday/Calendar.pyx":140
 *         while d <= toDate:
 *             if self.isHoliday(d) and (includeWeekEnds or not self.isWeekEnd(d.weekday())):
 *                 result.append(d)             # <<<<<<<<<<<<<<
 *             d += 1
 *         return result
 */
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_d)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 140, __pyx_L1_error)

      /* "ultron/tradingday/Calendar.pyx":139
 * 
 *         while d <= toDate:
 *             if self.isHoliday(d) and (includeWeekEnds or not self.isWeekEnd(d.weekday())):             # <<<<<<<<<<<<<<
 *                 result.append(d)
 *             d += 1
 */
    }

    /* "ultron/tradingday/Calendar.pyx":141
 *             if self.isHoliday(d) and (includeWeekEnds or not self.isWeekEnd(d.weekday())):
 *                 result.append(d)
 *             d += 1             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = PyNumber_InPlaceAdd(((PyObject *)__pyx_v_d), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_d, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
    __pyx_t_1 = 0;
  }

  /* "ultron/tradingday/Calendar.pyx":142
 *                 result.append(d)
 *             d += 1
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef bizDatesList(self, Date fromDate, Date toDate):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":134
 *             return self.adjustDate(d1, c)
 * 
 *     cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=True):             # <<<<<<<<<<<<<<
 *         cdef list result = []
 *         cdef Date d = fromDate
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.holDatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_19holDatesList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_18holDatesList[] = "Calendar.holDatesList(self, Date fromDate, Date toDate, bool includeWeekEnds=True)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_19holDatesList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate = 0;
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate = 0;
  int __pyx_v_includeWeekEnds;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("holDatesList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fromDate,&__pyx_n_s_toDate,&__pyx_n_s_includeWeekEnds,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fromDate)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_toDate)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("holDatesList", 0, 2, 3, 1); __PYX_ERR(0, 134, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_includeWeekEnds);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "holDatesList") < 0)) __PYX_ERR(0, 134, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fromDate = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)values[0]);
    __pyx_v_toDate = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)values[1]);
    if (values[2]) {
      __pyx_v_includeWeekEnds = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_includeWeekEnds == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 134, __pyx_L3_error)
    } else {
      __pyx_v_includeWeekEnds = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("holDatesList", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 134, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.holDatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fromDate), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "fromDate", 0))) __PYX_ERR(0, 134, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_toDate), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "toDate", 0))) __PYX_ERR(0, 134, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_18holDatesList(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), __pyx_v_fromDate, __pyx_v_toDate, __pyx_v_includeWeekEnds);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_18holDatesList(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_v_includeWeekEnds) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_holDatesList __pyx_t_2;
  __Pyx_RefNannySetupContext("holDatesList", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.includeWeekEnds = __pyx_v_includeWeekEnds;
  __pyx_t_1 = __pyx_vtabptr_6ultron_10tradingday_8Calendar_Calendar->holDatesList(__pyx_v_self, __pyx_v_fromDate, __pyx_v_toDate, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.holDatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":144
 *         return result
 * 
 *     cpdef bizDatesList(self, Date fromDate, Date toDate):             # <<<<<<<<<<<<<<
 *         cdef list result = []
 *         cdef Date d = fromDate
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_21bizDatesList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_bizDatesList(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_result = 0;
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_d = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("bizDatesList", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_bizDatesList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_21bizDatesList)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_fromDate), ((PyObject *)__pyx_v_toDate)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_fromDate), ((PyObject *)__pyx_v_toDate)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 144, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_fromDate));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_fromDate));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_fromDate));
          __Pyx_INCREF(((PyObject *)__pyx_v_toDate));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_toDate));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_toDate));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":145
 * 
 *     cpdef bizDatesList(self, Date fromDate, Date toDate):
 *         cdef list result = []             # <<<<<<<<<<<<<<
 *         cdef Date d = fromDate
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ultron/tradingday/Calendar.pyx":146
 *     cpdef bizDatesList(self, Date fromDate, Date toDate):
 *         cdef list result = []
 *         cdef Date d = fromDate             # <<<<<<<<<<<<<<
 * 
 *         while d <= toDate:
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_fromDate));
  __pyx_v_d = __pyx_v_fromDate;

  /* "ultron/tradingday/Calendar.pyx":148
 *         cdef Date d = fromDate
 * 
 *         while d <= toDate:             # <<<<<<<<<<<<<<
 *             if self.isBizDay(d):
 *                 result.append(d)
 */
  while (1) {
    __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_v_toDate), Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_7) break;

    /* "ultron/tradingday/Calendar.pyx":149
 * 
 *         while d <= toDate:
 *             if self.isBizDay(d):             # <<<<<<<<<<<<<<
 *                 result.append(d)
 *             d += 1
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self->__pyx_vtab)->isBizDay(__pyx_v_self, __pyx_v_d, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_7) {

      /* "ultron/tradingday/Calendar.pyx":150
 *         while d <= toDate:
 *             if self.isBizDay(d):
 *                 result.append(d)             # <<<<<<<<<<<<<<
 *             d += 1
 *         return result
 */
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_d)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 150, __pyx_L1_error)

      /* "ultron/tradingday/Calendar.pyx":149
 * 
 *         while d <= toDate:
 *             if self.isBizDay(d):             # <<<<<<<<<<<<<<
 *                 result.append(d)
 *             d += 1
 */
    }

    /* "ultron/tradingday/Calendar.pyx":151
 *             if self.isBizDay(d):
 *                 result.append(d)
 *             d += 1             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = PyNumber_InPlaceAdd(((PyObject *)__pyx_v_d), __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6ultron_10tradingday_4Date_Date))))) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_d, ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_t_1));
    __pyx_t_1 = 0;
  }

  /* "ultron/tradingday/Calendar.pyx":152
 *                 result.append(d)
 *             d += 1
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, right, int op):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":144
 *         return result
 * 
 *     cpdef bizDatesList(self, Date fromDate, Date toDate):             # <<<<<<<<<<<<<<
 *         cdef list result = []
 *         cdef Date d = fromDate
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.bizDatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_21bizDatesList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_20bizDatesList[] = "Calendar.bizDatesList(self, Date fromDate, Date toDate)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_21bizDatesList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate = 0;
  struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bizDatesList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fromDate,&__pyx_n_s_toDate,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fromDate)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_toDate)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bizDatesList", 1, 2, 2, 1); __PYX_ERR(0, 144, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bizDatesList") < 0)) __PYX_ERR(0, 144, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_fromDate = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)values[0]);
    __pyx_v_toDate = ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bizDatesList", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 144, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.bizDatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fromDate), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "fromDate", 0))) __PYX_ERR(0, 144, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_toDate), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "toDate", 0))) __PYX_ERR(0, 144, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_20bizDatesList(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), __pyx_v_fromDate, __pyx_v_toDate);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_20bizDatesList(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_fromDate, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_toDate) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("bizDatesList", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar_8Calendar_bizDatesList(__pyx_v_self, __pyx_v_fromDate, __pyx_v_toDate, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.bizDatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":154
 *         return result
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return self._impl == right._impl
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_23__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_23__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_22__richcmp__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((PyObject *)__pyx_v_right), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_22__richcmp__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ultron/tradingday/Calendar.pyx":155
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return self._impl == right._impl
 * 
 */
  __pyx_t_1 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_1) {

    /* "ultron/tradingday/Calendar.pyx":156
 *     def __richcmp__(self, right, int op):
 *         if op == 2:
 *             return self._impl == right._impl             # <<<<<<<<<<<<<<
 * 
 *     def __deepcopy__(self, memo):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_right, __pyx_n_s_impl); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_RichCompare(((PyObject *)__pyx_v_self->_impl), __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":155
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return self._impl == right._impl
 * 
 */
  }

  /* "ultron/tradingday/Calendar.pyx":154
 *         return result
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return self._impl == right._impl
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":158
 *             return self._impl == right._impl
 * 
 *     def __deepcopy__(self, memo):             # <<<<<<<<<<<<<<
 *         return Calendar(self.name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_25__deepcopy__(PyObject *__pyx_v_self, PyObject *__pyx_v_memo); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_24__deepcopy__[] = "Calendar.__deepcopy__(self, memo)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_25__deepcopy__(PyObject *__pyx_v_self, PyObject *__pyx_v_memo) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__deepcopy__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_24__deepcopy__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((PyObject *)__pyx_v_memo));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_24__deepcopy__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_memo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__deepcopy__", 0);

  /* "ultron/tradingday/Calendar.pyx":159
 * 
 *     def __deepcopy__(self, memo):
 *         return Calendar(self.name)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_Calendar), __pyx_v_self->name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":158
 *             return self._impl == right._impl
 * 
 *     def __deepcopy__(self, memo):             # <<<<<<<<<<<<<<
 *         return Calendar(self.name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.__deepcopy__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":161
 *         return Calendar(self.name)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         d = {}
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_27__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_26__reduce__[] = "Calendar.__reduce__(self)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_27__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_26__reduce__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_26__reduce__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self) {
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ultron/tradingday/Calendar.pyx":162
 * 
 *     def __reduce__(self):
 *         d = {}             # <<<<<<<<<<<<<<
 * 
 *         return Calendar, (self.name,), d
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_d = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ultron/tradingday/Calendar.pyx":164
 *         d = {}
 * 
 *         return Calendar, (self.name,), d             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_Calendar));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_Calendar));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_Calendar));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_d);
  __Pyx_GIVEREF(__pyx_v_d);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_d);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":161
 *         return Calendar(self.name)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         d = {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":166
 *         return Calendar, (self.name,), d
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_29__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_28__setstate__[] = "Calendar.__setstate__(self, state)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_29__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_28__setstate__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_28__setstate__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pxd":15
 * cdef class Calendar(object):
 * 
 *     cdef public CalendarImpl _impl             # <<<<<<<<<<<<<<
 *     cdef public str name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5_impl_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5_impl_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_5_impl___get__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_5_impl___get__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_impl));
  __pyx_r = ((PyObject *)__pyx_v_self->_impl);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5_impl_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5_impl_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_5_impl_2__set__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_5_impl_2__set__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl))))) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_impl);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_impl));
  __pyx_v_self->_impl = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar._impl.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5_impl_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5_impl_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_5_impl_4__del__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_5_impl_4__del__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_impl);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_impl));
  __pyx_v_self->_impl = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pxd":16
 * 
 *     cdef public CalendarImpl _impl
 *     cdef public str name             # <<<<<<<<<<<<<<
 * 
 *     cpdef isBizDay(self, Date d)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_4name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4name___get__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4name___get__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4name_2__set__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4name_2__set__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 16, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.Calendar.name.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_4name_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_4name_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4name_4__del__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_8Calendar_4name_4__del__(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":170
 * 
 * cdef class CalendarImpl(object):
 *     cdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_12CalendarImpl_isBizDay(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isBizDay", 0);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":173
 *         pass
 * 
 *     cdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_12CalendarImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_v_self, CYTHON_UNUSED int __pyx_v_weekDay) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isWeekEnd", 0);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12CalendarImpl_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_12CalendarImpl___reduce_cython__[] = "CalendarImpl.__reduce_cython__(self)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12CalendarImpl_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12CalendarImpl___reduce_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12CalendarImpl___reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_CalendarImpl); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CalendarImpl__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_CalendarImpl); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.CalendarImpl.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CalendarImpl__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12CalendarImpl_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_12CalendarImpl_2__setstate_cython__[] = "CalendarImpl.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12CalendarImpl_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12CalendarImpl_2__setstate_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12CalendarImpl_2__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CalendarImpl__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_CalendarImpl__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CalendarImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CalendarImpl__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.CalendarImpl.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":736
 * 
 * cdef class ChinaSseImpl(CalendarImpl):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl___init__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":739
 *         pass
 * 
 *     cdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         cdef int w = date.weekday()
 *         if (self.isWeekEnd(w) or date in sse_holDays) and date not in sse_working_weekends:
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_12ChinaSseImpl_isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date) {
  int __pyx_v_w;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("isBizDay", 0);

  /* "ultron/tradingday/Calendar.pyx":740
 * 
 *     cdef bint isBizDay(self, Date date):
 *         cdef int w = date.weekday()             # <<<<<<<<<<<<<<
 *         if (self.isWeekEnd(w) or date in sse_holDays) and date not in sse_working_weekends:
 *             return False
 */
  __pyx_v_w = ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_date->__pyx_vtab)->weekday(__pyx_v_date, 0);

  /* "ultron/tradingday/Calendar.pyx":741
 *     cdef bint isBizDay(self, Date date):
 *         cdef int w = date.weekday()
 *         if (self.isWeekEnd(w) or date in sse_holDays) and date not in sse_working_weekends:             # <<<<<<<<<<<<<<
 *             return False
 *         return True
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.isWeekEnd(((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_self), __pyx_v_w) != 0);
  if (!__pyx_t_2) {
  } else {
    goto __pyx_L5_next_and;
  }
  if (unlikely(__pyx_v_6ultron_10tradingday_8Calendar_sse_holDays == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 741, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PySet_ContainsTF(((PyObject *)__pyx_v_date), __pyx_v_6ultron_10tradingday_8Calendar_sse_holDays, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 741, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_L5_next_and:;
  if (unlikely(__pyx_v_6ultron_10tradingday_8Calendar_sse_working_weekends == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 741, __pyx_L1_error)
  }
  __pyx_t_3 = (__Pyx_PySet_ContainsTF(((PyObject *)__pyx_v_date), __pyx_v_6ultron_10tradingday_8Calendar_sse_working_weekends, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 741, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "ultron/tradingday/Calendar.pyx":742
 *         cdef int w = date.weekday()
 *         if (self.isWeekEnd(w) or date in sse_holDays) and date not in sse_working_weekends:
 *             return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":741
 *     cdef bint isBizDay(self, Date date):
 *         cdef int w = date.weekday()
 *         if (self.isWeekEnd(w) or date in sse_holDays) and date not in sse_working_weekends:             # <<<<<<<<<<<<<<
 *             return False
 *         return True
 */
  }

  /* "ultron/tradingday/Calendar.pyx":743
 *         if (self.isWeekEnd(w) or date in sse_holDays) and date not in sse_working_weekends:
 *             return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef bint isWeekEnd(self, int weekDay):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":739
 *         pass
 * 
 *     cdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         cdef int w = date.weekday()
 *         if (self.isWeekEnd(w) or date in sse_holDays) and date not in sse_working_weekends:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("ultron.tradingday.Calendar.ChinaSseImpl.isBizDay", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":745
 *         return True
 * 
 *     cdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_12ChinaSseImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self, int __pyx_v_weekDay) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("isWeekEnd", 0);

  /* "ultron/tradingday/Calendar.pyx":746
 * 
 *     cdef bint isWeekEnd(self, int weekDay):
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, right, int op):
 */
  switch (__pyx_v_weekDay) {
    case Saturday:
    case Sunday:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":745
 *         return True
 * 
 *     cdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":748
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, ChinaSseImpl)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl_2__richcmp__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_v_self), ((PyObject *)__pyx_v_right), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl_2__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ultron/tradingday/Calendar.pyx":749
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, ChinaSseImpl)
 * 
 */
  __pyx_t_1 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_1) {

    /* "ultron/tradingday/Calendar.pyx":750
 *     def __richcmp__(self, right, int op):
 *         if op == 2:
 *             return isinstance(right, ChinaSseImpl)             # <<<<<<<<<<<<<<
 * 
 * cdef set ib_working_weekends = {
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_right, __pyx_ptype_6ultron_10tradingday_8Calendar_ChinaSseImpl); 
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 750, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":749
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, ChinaSseImpl)
 * 
 */
  }

  /* "ultron/tradingday/Calendar.pyx":748
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, ChinaSseImpl)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaSseImpl.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_12ChinaSseImpl_4__reduce_cython__[] = "ChinaSseImpl.__reduce_cython__(self)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl_4__reduce_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl_4__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_ChinaSseImpl); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ChinaSseImpl__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_ChinaSseImpl); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaSseImpl.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ChinaSseImpl__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_12ChinaSseImpl_6__setstate_cython__[] = "ChinaSseImpl.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl_6__setstate_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12ChinaSseImpl_6__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ChinaSseImpl__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaSseImpl__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ChinaSseImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ChinaSseImpl__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaSseImpl.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":904
 * 
 * cdef class ChinaIBImpl(CalendarImpl):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl___init__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":907
 *         pass
 * 
 *     cpdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         return _sseImpl.isBizDay(date) or date in ib_working_weekends
 * 
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_3isBizDay(PyObject *__pyx_v_self, PyObject *__pyx_v_date); /*proto*/
int __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isBizDay(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("isBizDay", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isBizDay); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_3isBizDay)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_date)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_date));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 907, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 907, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":908
 * 
 *     cpdef bint isBizDay(self, Date date):
 *         return _sseImpl.isBizDay(date) or date in ib_working_weekends             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint isWeekEnd(self, int weekDay):
 */
  __pyx_t_6 = (((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_v_6ultron_10tradingday_8Calendar__sseImpl->__pyx_base.__pyx_vtab)->__pyx_base.isBizDay(((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_6ultron_10tradingday_8Calendar__sseImpl), __pyx_v_date) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  if (unlikely(__pyx_v_6ultron_10tradingday_8Calendar_ib_working_weekends == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 908, __pyx_L1_error)
  }
  __pyx_t_6 = (__Pyx_PySet_ContainsTF(((PyObject *)__pyx_v_date), __pyx_v_6ultron_10tradingday_8Calendar_ib_working_weekends, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 908, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_5 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":907
 *         pass
 * 
 *     cpdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         return _sseImpl.isBizDay(date) or date in ib_working_weekends
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ultron.tradingday.Calendar.ChinaIBImpl.isBizDay", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_3isBizDay(PyObject *__pyx_v_self, PyObject *__pyx_v_date); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_11ChinaIBImpl_2isBizDay[] = "ChinaIBImpl.isBizDay(self, Date date) -> bool";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_3isBizDay(PyObject *__pyx_v_self, PyObject *__pyx_v_date) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isBizDay (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_date), __pyx_ptype_6ultron_10tradingday_4Date_Date, 1, "date", 0))) __PYX_ERR(0, 907, __pyx_L1_error)
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_2isBizDay(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *)__pyx_v_self), ((struct __pyx_obj_6ultron_10tradingday_4Date_Date *)__pyx_v_date));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_2isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("isBizDay", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isBizDay(__pyx_v_self, __pyx_v_date, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaIBImpl.isBizDay", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

int __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isBizDay__pyx_wrap_1(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date) {
  return __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isBizDay(__pyx_v_self, __pyx_v_date, 0);
}

/* "ultron/tradingday/Calendar.pyx":910
 *         return _sseImpl.isBizDay(date) or date in ib_working_weekends
 * 
 *     cpdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 */

static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_5isWeekEnd(PyObject *__pyx_v_self, PyObject *__pyx_arg_weekDay); /*proto*/
int __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, int __pyx_v_weekDay, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("isWeekEnd", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isWeekEnd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_5isWeekEnd)) {
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_weekDay); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 910, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 910, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 910, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "ultron/tradingday/Calendar.pyx":911
 * 
 *     cpdef bint isWeekEnd(self, int weekDay):
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, right, int op):
 */
  switch (__pyx_v_weekDay) {
    case Saturday:
    case Sunday:
    __pyx_t_6 = 1;
    break;
    default:
    __pyx_t_6 = 0;
    break;
  }
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":910
 *         return _sseImpl.isBizDay(date) or date in ib_working_weekends
 * 
 *     cpdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("ultron.tradingday.Calendar.ChinaIBImpl.isWeekEnd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_5isWeekEnd(PyObject *__pyx_v_self, PyObject *__pyx_arg_weekDay); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_11ChinaIBImpl_4isWeekEnd[] = "ChinaIBImpl.isWeekEnd(self, int weekDay) -> bool";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_5isWeekEnd(PyObject *__pyx_v_self, PyObject *__pyx_arg_weekDay) {
  int __pyx_v_weekDay;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isWeekEnd (wrapper)", 0);
  assert(__pyx_arg_weekDay); {
    __pyx_v_weekDay = __Pyx_PyInt_As_int(__pyx_arg_weekDay); if (unlikely((__pyx_v_weekDay == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 910, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaIBImpl.isWeekEnd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_4isWeekEnd(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *)__pyx_v_self), ((int)__pyx_v_weekDay));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_4isWeekEnd(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, int __pyx_v_weekDay) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("isWeekEnd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isWeekEnd(__pyx_v_self, __pyx_v_weekDay, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaIBImpl.isWeekEnd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

int __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isWeekEnd__pyx_wrap_1(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, int __pyx_v_weekDay) {
  return __pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isWeekEnd(__pyx_v_self, __pyx_v_weekDay, 0);
}

/* "ultron/tradingday/Calendar.pyx":913
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, ChinaIBImpl)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_7__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_7__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_6__richcmp__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *)__pyx_v_self), ((PyObject *)__pyx_v_right), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_6__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ultron/tradingday/Calendar.pyx":914
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, ChinaIBImpl)
 * 
 */
  __pyx_t_1 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_1) {

    /* "ultron/tradingday/Calendar.pyx":915
 *     def __richcmp__(self, right, int op):
 *         if op == 2:
 *             return isinstance(right, ChinaIBImpl)             # <<<<<<<<<<<<<<
 * 
 * cdef class NullCalendar(CalendarImpl):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_right, __pyx_ptype_6ultron_10tradingday_8Calendar_ChinaIBImpl); 
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":914
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, ChinaIBImpl)
 * 
 */
  }

  /* "ultron/tradingday/Calendar.pyx":913
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, ChinaIBImpl)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaIBImpl.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_11ChinaIBImpl_8__reduce_cython__[] = "ChinaIBImpl.__reduce_cython__(self)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_8__reduce_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_8__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_ChinaIBImpl); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ChinaIBImpl__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_ChinaIBImpl); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaIBImpl.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ChinaIBImpl__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_11ChinaIBImpl_10__setstate_cython__[] = "ChinaIBImpl.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_10__setstate_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_11ChinaIBImpl_10__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ChinaIBImpl__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaIBImpl__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ChinaIBImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ChinaIBImpl__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaIBImpl.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":918
 * 
 * cdef class NullCalendar(CalendarImpl):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar___init__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":921
 *         pass
 * 
 *     cdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_12NullCalendar_isBizDay(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isBizDay", 0);

  /* "ultron/tradingday/Calendar.pyx":922
 * 
 *     cdef bint isBizDay(self, Date date):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef bint isWeekEnd(self, int weekDay):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":921
 *         pass
 * 
 *     cdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":924
 *         return True
 * 
 *     cdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_12NullCalendar_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self, int __pyx_v_weekDay) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("isWeekEnd", 0);

  /* "ultron/tradingday/Calendar.pyx":925
 * 
 *     cdef bint isWeekEnd(self, int weekDay):
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, right, int op):
 */
  switch (__pyx_v_weekDay) {
    case Saturday:
    case Sunday:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":924
 *         return True
 * 
 *     cdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":927
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, NullCalendar)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar_2__richcmp__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *)__pyx_v_self), ((PyObject *)__pyx_v_right), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar_2__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ultron/tradingday/Calendar.pyx":928
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, NullCalendar)
 * 
 */
  __pyx_t_1 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_1) {

    /* "ultron/tradingday/Calendar.pyx":929
 *     def __richcmp__(self, right, int op):
 *         if op == 2:
 *             return isinstance(right, NullCalendar)             # <<<<<<<<<<<<<<
 * 
 * cdef class ChinaCFFEXImpl(CalendarImpl):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_right, __pyx_ptype_6ultron_10tradingday_8Calendar_NullCalendar); 
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":928
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, NullCalendar)
 * 
 */
  }

  /* "ultron/tradingday/Calendar.pyx":927
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, NullCalendar)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.NullCalendar.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_12NullCalendar_4__reduce_cython__[] = "NullCalendar.__reduce_cython__(self)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar_4__reduce_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar_4__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_NullCalendar); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NullCalendar__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_NullCalendar); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.NullCalendar.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NullCalendar__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_12NullCalendar_6__setstate_cython__[] = "NullCalendar.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar_6__setstate_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12NullCalendar_6__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NullCalendar__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_NullCalendar__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NullCalendar, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NullCalendar__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.NullCalendar.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":932
 * 
 * cdef class ChinaCFFEXImpl(CalendarImpl):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl___init__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":935
 *         pass
 * 
 *     cdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         return _sseImpl.isBizDay(date)
 * 
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_isBizDay(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isBizDay", 0);

  /* "ultron/tradingday/Calendar.pyx":936
 * 
 *     cdef bint isBizDay(self, Date date):
 *         return _sseImpl.isBizDay(date)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint isWeekEnd(self, int weekDay):
 */
  __pyx_r = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_v_6ultron_10tradingday_8Calendar__sseImpl->__pyx_base.__pyx_vtab)->__pyx_base.isBizDay(((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_6ultron_10tradingday_8Calendar__sseImpl), __pyx_v_date);
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":935
 *         pass
 * 
 *     cdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         return _sseImpl.isBizDay(date)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":938
 *         return _sseImpl.isBizDay(date)
 * 
 *     cdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return _sseImpl.isWeekEnd(weekDay)
 * 
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self, int __pyx_v_weekDay) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isWeekEnd", 0);

  /* "ultron/tradingday/Calendar.pyx":939
 * 
 *     cdef bint isWeekEnd(self, int weekDay):
 *         return _sseImpl.isWeekEnd(weekDay)             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, right, int op):
 */
  __pyx_r = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_v_6ultron_10tradingday_8Calendar__sseImpl->__pyx_base.__pyx_vtab)->__pyx_base.isWeekEnd(((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_6ultron_10tradingday_8Calendar__sseImpl), __pyx_v_weekDay);
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":938
 *         return _sseImpl.isBizDay(date)
 * 
 *     cdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return _sseImpl.isWeekEnd(weekDay)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":941
 *         return _sseImpl.isWeekEnd(weekDay)
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, ChinaCFFEXImpl)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_2__richcmp__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *)__pyx_v_self), ((PyObject *)__pyx_v_right), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_2__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ultron/tradingday/Calendar.pyx":942
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, ChinaCFFEXImpl)
 * 
 */
  __pyx_t_1 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_1) {

    /* "ultron/tradingday/Calendar.pyx":943
 *     def __richcmp__(self, right, int op):
 *         if op == 2:
 *             return isinstance(right, ChinaCFFEXImpl)             # <<<<<<<<<<<<<<
 * 
 * cdef int EasterMonday[299]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_right, __pyx_ptype_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl); 
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":942
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, ChinaCFFEXImpl)
 * 
 */
  }

  /* "ultron/tradingday/Calendar.pyx":941
 *         return _sseImpl.isWeekEnd(weekDay)
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, ChinaCFFEXImpl)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaCFFEXImpl.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_4__reduce_cython__[] = "ChinaCFFEXImpl.__reduce_cython__(self)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_4__reduce_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_4__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_ChinaCFFEXImpl); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ChinaCFFEXImpl__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_ChinaCFFEXImpl); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaCFFEXImpl.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ChinaCFFEXImpl__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_6__setstate_cython__[] = "ChinaCFFEXImpl.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_6__setstate_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_6__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ChinaCFFEXImpl__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaCFFEXImpl__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ChinaCFFEXImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ChinaCFFEXImpl__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.ChinaCFFEXImpl.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":980
 * 
 * cdef class WestenImpl(CalendarImpl):
 *     cdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_10WestenImpl_isWeekEnd(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_v_self, int __pyx_v_weekDay) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("isWeekEnd", 0);

  /* "ultron/tradingday/Calendar.pyx":981
 * cdef class WestenImpl(CalendarImpl):
 *     cdef bint isWeekEnd(self, int weekDay):
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday             # <<<<<<<<<<<<<<
 * 
 *     cdef int easterMonday(self, int year):
 */
  switch (__pyx_v_weekDay) {
    case Saturday:
    case Sunday:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":980
 * 
 * cdef class WestenImpl(CalendarImpl):
 *     cdef bint isWeekEnd(self, int weekDay):             # <<<<<<<<<<<<<<
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":983
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 *     cdef int easterMonday(self, int year):             # <<<<<<<<<<<<<<
 *         return EasterMonday[year - 1901]
 * 
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_10WestenImpl_easterMonday(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_v_self, int __pyx_v_year) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("easterMonday", 0);

  /* "ultron/tradingday/Calendar.pyx":984
 * 
 *     cdef int easterMonday(self, int year):
 *         return EasterMonday[year - 1901]             # <<<<<<<<<<<<<<
 * 
 * cdef class TargetImpl(WestenImpl):
 */
  __pyx_r = (__pyx_v_6ultron_10tradingday_8Calendar_EasterMonday[(__pyx_v_year - 0x76D)]);
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":983
 *         return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday
 * 
 *     cdef int easterMonday(self, int year):             # <<<<<<<<<<<<<<
 *         return EasterMonday[year - 1901]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10WestenImpl_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_10WestenImpl___reduce_cython__[] = "WestenImpl.__reduce_cython__(self)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10WestenImpl_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_10WestenImpl___reduce_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10WestenImpl___reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_WestenImpl); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_WestenImpl__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_WestenImpl); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.WestenImpl.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_WestenImpl__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10WestenImpl_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_10WestenImpl_2__setstate_cython__[] = "WestenImpl.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10WestenImpl_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_10WestenImpl_2__setstate_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10WestenImpl_2__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_WestenImpl__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_WestenImpl__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_WestenImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_WestenImpl__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.WestenImpl.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":987
 * 
 * cdef class TargetImpl(WestenImpl):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static int __pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl___init__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl___init__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":990
 *         pass
 * 
 *     cdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         cdef int w = date.weekday()
 *         cdef int d = date.dayOfMonth()
 */

static int __pyx_f_6ultron_10tradingday_8Calendar_10TargetImpl_isBizDay(struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self, struct __pyx_obj_6ultron_10tradingday_4Date_Date *__pyx_v_date) {
  int __pyx_v_w;
  int __pyx_v_d;
  int __pyx_v_dd;
  int __pyx_v_m;
  int __pyx_v_y;
  int __pyx_v_em;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("isBizDay", 0);

  /* "ultron/tradingday/Calendar.pyx":991
 * 
 *     cdef bint isBizDay(self, Date date):
 *         cdef int w = date.weekday()             # <<<<<<<<<<<<<<
 *         cdef int d = date.dayOfMonth()
 *         cdef int dd = date.dayOfYear()
 */
  __pyx_v_w = ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_date->__pyx_vtab)->weekday(__pyx_v_date, 0);

  /* "ultron/tradingday/Calendar.pyx":992
 *     cdef bint isBizDay(self, Date date):
 *         cdef int w = date.weekday()
 *         cdef int d = date.dayOfMonth()             # <<<<<<<<<<<<<<
 *         cdef int dd = date.dayOfYear()
 *         cdef int m = date.month()
 */
  __pyx_v_d = ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_date->__pyx_vtab)->dayOfMonth(__pyx_v_date, 0);

  /* "ultron/tradingday/Calendar.pyx":993
 *         cdef int w = date.weekday()
 *         cdef int d = date.dayOfMonth()
 *         cdef int dd = date.dayOfYear()             # <<<<<<<<<<<<<<
 *         cdef int m = date.month()
 *         cdef int y = date.year()
 */
  __pyx_v_dd = ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_date->__pyx_vtab)->dayOfYear(__pyx_v_date, 0);

  /* "ultron/tradingday/Calendar.pyx":994
 *         cdef int d = date.dayOfMonth()
 *         cdef int dd = date.dayOfYear()
 *         cdef int m = date.month()             # <<<<<<<<<<<<<<
 *         cdef int y = date.year()
 *         cdef int em = self.easterMonday(y)
 */
  __pyx_v_m = ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_date->__pyx_vtab)->month(__pyx_v_date, 0);

  /* "ultron/tradingday/Calendar.pyx":995
 *         cdef int dd = date.dayOfYear()
 *         cdef int m = date.month()
 *         cdef int y = date.year()             # <<<<<<<<<<<<<<
 *         cdef int em = self.easterMonday(y)
 * 
 */
  __pyx_v_y = ((struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date *)__pyx_v_date->__pyx_vtab)->year(__pyx_v_date, 0);

  /* "ultron/tradingday/Calendar.pyx":996
 *         cdef int m = date.month()
 *         cdef int y = date.year()
 *         cdef int em = self.easterMonday(y)             # <<<<<<<<<<<<<<
 * 
 *         if (self.isWeekEnd(w)
 */
  __pyx_v_em = ((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_TargetImpl *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.easterMonday(((struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *)__pyx_v_self), __pyx_v_y);

  /* "ultron/tradingday/Calendar.pyx":998
 *         cdef int em = self.easterMonday(y)
 * 
 *         if (self.isWeekEnd(w)             # <<<<<<<<<<<<<<
 *                 or (d == 1 and m == Months.January)
 *                 or (dd == em - 3 and y >= 2000)
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_TargetImpl *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.isWeekEnd(((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v_self), __pyx_v_w) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "ultron/tradingday/Calendar.pyx":999
 * 
 *         if (self.isWeekEnd(w)
 *                 or (d == 1 and m == Months.January)             # <<<<<<<<<<<<<<
 *                 or (dd == em - 3 and y >= 2000)
 *                 or (dd == em and y >= 2000)
 */
  __pyx_t_2 = ((__pyx_v_d == 1) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L6_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_m == January) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_L6_next_or:;

  /* "ultron/tradingday/Calendar.pyx":1000
 *         if (self.isWeekEnd(w)
 *                 or (d == 1 and m == Months.January)
 *                 or (dd == em - 3 and y >= 2000)             # <<<<<<<<<<<<<<
 *                 or (dd == em and y >= 2000)
 *                 or (d == 1 and m == Months.May and y >= 2000)
 */
  __pyx_t_2 = ((__pyx_v_dd == (__pyx_v_em - 3)) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L8_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_y >= 0x7D0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_L8_next_or:;

  /* "ultron/tradingday/Calendar.pyx":1001
 *                 or (d == 1 and m == Months.January)
 *                 or (dd == em - 3 and y >= 2000)
 *                 or (dd == em and y >= 2000)             # <<<<<<<<<<<<<<
 *                 or (d == 1 and m == Months.May and y >= 2000)
 *                 or (d == 25 and m == Months.December)
 */
  __pyx_t_2 = ((__pyx_v_dd == __pyx_v_em) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L10_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_y >= 0x7D0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_L10_next_or:;

  /* "ultron/tradingday/Calendar.pyx":1002
 *                 or (dd == em - 3 and y >= 2000)
 *                 or (dd == em and y >= 2000)
 *                 or (d == 1 and m == Months.May and y >= 2000)             # <<<<<<<<<<<<<<
 *                 or (d == 25 and m == Months.December)
 *                 or (d == 26 and m == Months.December and y >= 2000)
 */
  __pyx_t_2 = ((__pyx_v_d == 1) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L12_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_m == May) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L12_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_y >= 0x7D0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_L12_next_or:;

  /* "ultron/tradingday/Calendar.pyx":1003
 *                 or (dd == em and y >= 2000)
 *                 or (d == 1 and m == Months.May and y >= 2000)
 *                 or (d == 25 and m == Months.December)             # <<<<<<<<<<<<<<
 *                 or (d == 26 and m == Months.December and y >= 2000)
 *                 or (d == 31 and m == Months.December and (y == 1998 or y == 1999 or y == 2001))):
 */
  __pyx_t_2 = ((__pyx_v_d == 25) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L15_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_m == December) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_L15_next_or:;

  /* "ultron/tradingday/Calendar.pyx":1004
 *                 or (d == 1 and m == Months.May and y >= 2000)
 *                 or (d == 25 and m == Months.December)
 *                 or (d == 26 and m == Months.December and y >= 2000)             # <<<<<<<<<<<<<<
 *                 or (d == 31 and m == Months.December and (y == 1998 or y == 1999 or y == 2001))):
 *             return False
 */
  __pyx_t_2 = ((__pyx_v_d == 26) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L17_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_m == December) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L17_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_y >= 0x7D0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_L17_next_or:;

  /* "ultron/tradingday/Calendar.pyx":1005
 *                 or (d == 25 and m == Months.December)
 *                 or (d == 26 and m == Months.December and y >= 2000)
 *                 or (d == 31 and m == Months.December and (y == 1998 or y == 1999 or y == 2001))):             # <<<<<<<<<<<<<<
 *             return False
 *         return True
 */
  __pyx_t_2 = ((__pyx_v_d == 31) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_m == December) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  switch (__pyx_v_y) {
    case 0x7CE:
    case 0x7CF:
    case 0x7D1:
    __pyx_t_2 = 1;
    break;
    default:
    __pyx_t_2 = 0;
    break;
  }
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "ultron/tradingday/Calendar.pyx":998
 *         cdef int em = self.easterMonday(y)
 * 
 *         if (self.isWeekEnd(w)             # <<<<<<<<<<<<<<
 *                 or (d == 1 and m == Months.January)
 *                 or (dd == em - 3 and y >= 2000)
 */
  if (__pyx_t_1) {

    /* "ultron/tradingday/Calendar.pyx":1006
 *                 or (d == 26 and m == Months.December and y >= 2000)
 *                 or (d == 31 and m == Months.December and (y == 1998 or y == 1999 or y == 2001))):
 *             return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":998
 *         cdef int em = self.easterMonday(y)
 * 
 *         if (self.isWeekEnd(w)             # <<<<<<<<<<<<<<
 *                 or (d == 1 and m == Months.January)
 *                 or (dd == em - 3 and y >= 2000)
 */
  }

  /* "ultron/tradingday/Calendar.pyx":1007
 *                 or (d == 31 and m == Months.December and (y == 1998 or y == 1999 or y == 2001))):
 *             return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, right, int op):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "ultron/tradingday/Calendar.pyx":990
 *         pass
 * 
 *     cdef bint isBizDay(self, Date date):             # <<<<<<<<<<<<<<
 *         cdef int w = date.weekday()
 *         cdef int d = date.dayOfMonth()
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ultron/tradingday/Calendar.pyx":1009
 *         return True
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, TargetImpl)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl_2__richcmp__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *)__pyx_v_self), ((PyObject *)__pyx_v_right), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl_2__richcmp__(CYTHON_UNUSED struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self, PyObject *__pyx_v_right, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ultron/tradingday/Calendar.pyx":1010
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, TargetImpl)
 * 
 */
  __pyx_t_1 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_1) {

    /* "ultron/tradingday/Calendar.pyx":1011
 *     def __richcmp__(self, right, int op):
 *         if op == 2:
 *             return isinstance(right, TargetImpl)             # <<<<<<<<<<<<<<
 * 
 * cdef dict _holDict = {'china.sse': ChinaSseImpl,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_right, __pyx_ptype_6ultron_10tradingday_8Calendar_TargetImpl); 
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ultron/tradingday/Calendar.pyx":1010
 * 
 *     def __richcmp__(self, right, int op):
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return isinstance(right, TargetImpl)
 * 
 */
  }

  /* "ultron/tradingday/Calendar.pyx":1009
 *         return True
 * 
 *     def __richcmp__(self, right, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return isinstance(right, TargetImpl)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.TargetImpl.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_10TargetImpl_4__reduce_cython__[] = "TargetImpl.__reduce_cython__(self)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl_4__reduce_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl_4__reduce_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_TargetImpl); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TargetImpl__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_TargetImpl); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.TargetImpl.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TargetImpl__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_10TargetImpl_6__setstate_cython__[] = "TargetImpl.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl_6__setstate_cython__(((struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10TargetImpl_6__setstate_cython__(struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TargetImpl__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_TargetImpl__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TargetImpl, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TargetImpl__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.TargetImpl.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CalendarImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_1__pyx_unpickle_CalendarImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar___pyx_unpickle_CalendarImpl[] = "__pyx_unpickle_CalendarImpl(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_6ultron_10tradingday_8Calendar_1__pyx_unpickle_CalendarImpl = {"__pyx_unpickle_CalendarImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_1__pyx_unpickle_CalendarImpl, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar___pyx_unpickle_CalendarImpl};
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_1__pyx_unpickle_CalendarImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CalendarImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CalendarImpl", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CalendarImpl", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_CalendarImpl") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CalendarImpl", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_CalendarImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar___pyx_unpickle_CalendarImpl(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar___pyx_unpickle_CalendarImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CalendarImpl", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = CalendarImpl.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = CalendarImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = CalendarImpl.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CalendarImpl__set_state(<CalendarImpl> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = CalendarImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CalendarImpl__set_state(<CalendarImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = CalendarImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CalendarImpl__set_state(<CalendarImpl> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CalendarImpl__set_state(CalendarImpl __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_CalendarImpl__set_state(((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = CalendarImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CalendarImpl__set_state(<CalendarImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CalendarImpl__set_state(<CalendarImpl> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CalendarImpl__set_state(CalendarImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CalendarImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_CalendarImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CalendarImpl__set_state(<CalendarImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CalendarImpl__set_state(CalendarImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_CalendarImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CalendarImpl__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CalendarImpl__set_state(CalendarImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CalendarImpl__set_state(CalendarImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CalendarImpl__set_state(CalendarImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CalendarImpl__set_state(<CalendarImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CalendarImpl__set_state(CalendarImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_CalendarImpl__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ChinaSseImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_3__pyx_unpickle_ChinaSseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_2__pyx_unpickle_ChinaSseImpl[] = "__pyx_unpickle_ChinaSseImpl(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_6ultron_10tradingday_8Calendar_3__pyx_unpickle_ChinaSseImpl = {"__pyx_unpickle_ChinaSseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_3__pyx_unpickle_ChinaSseImpl, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_2__pyx_unpickle_ChinaSseImpl};
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_3__pyx_unpickle_ChinaSseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChinaSseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChinaSseImpl", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChinaSseImpl", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_ChinaSseImpl") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChinaSseImpl", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_ChinaSseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_2__pyx_unpickle_ChinaSseImpl(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_2__pyx_unpickle_ChinaSseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChinaSseImpl", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaSseImpl.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = ChinaSseImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaSseImpl.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChinaSseImpl__set_state(<ChinaSseImpl> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaSseImpl), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaSseImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ChinaSseImpl__set_state(<ChinaSseImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = ChinaSseImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChinaSseImpl__set_state(<ChinaSseImpl> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaSseImpl__set_state(ChinaSseImpl __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaSseImpl__set_state(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaSseImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ChinaSseImpl__set_state(<ChinaSseImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChinaSseImpl__set_state(<ChinaSseImpl> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ChinaSseImpl__set_state(ChinaSseImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ChinaSseImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_ChinaSseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ChinaSseImpl__set_state(<ChinaSseImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaSseImpl__set_state(ChinaSseImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaSseImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChinaSseImpl__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaSseImpl__set_state(ChinaSseImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ChinaSseImpl__set_state(ChinaSseImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaSseImpl__set_state(ChinaSseImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ChinaSseImpl__set_state(<ChinaSseImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaSseImpl__set_state(ChinaSseImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_ChinaSseImpl__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ChinaIBImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_5__pyx_unpickle_ChinaIBImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_4__pyx_unpickle_ChinaIBImpl[] = "__pyx_unpickle_ChinaIBImpl(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_6ultron_10tradingday_8Calendar_5__pyx_unpickle_ChinaIBImpl = {"__pyx_unpickle_ChinaIBImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_5__pyx_unpickle_ChinaIBImpl, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_4__pyx_unpickle_ChinaIBImpl};
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_5__pyx_unpickle_ChinaIBImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChinaIBImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChinaIBImpl", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChinaIBImpl", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_ChinaIBImpl") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChinaIBImpl", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_ChinaIBImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_4__pyx_unpickle_ChinaIBImpl(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_4__pyx_unpickle_ChinaIBImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChinaIBImpl", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaIBImpl.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = ChinaIBImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaIBImpl.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChinaIBImpl__set_state(<ChinaIBImpl> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaIBImpl), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaIBImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ChinaIBImpl__set_state(<ChinaIBImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = ChinaIBImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChinaIBImpl__set_state(<ChinaIBImpl> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaIBImpl__set_state(ChinaIBImpl __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaIBImpl__set_state(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaIBImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ChinaIBImpl__set_state(<ChinaIBImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChinaIBImpl__set_state(<ChinaIBImpl> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ChinaIBImpl__set_state(ChinaIBImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ChinaIBImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_ChinaIBImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ChinaIBImpl__set_state(<ChinaIBImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaIBImpl__set_state(ChinaIBImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaIBImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChinaIBImpl__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaIBImpl__set_state(ChinaIBImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ChinaIBImpl__set_state(ChinaIBImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaIBImpl__set_state(ChinaIBImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ChinaIBImpl__set_state(<ChinaIBImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaIBImpl__set_state(ChinaIBImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_ChinaIBImpl__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_NullCalendar(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_7__pyx_unpickle_NullCalendar(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_6__pyx_unpickle_NullCalendar[] = "__pyx_unpickle_NullCalendar(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_6ultron_10tradingday_8Calendar_7__pyx_unpickle_NullCalendar = {"__pyx_unpickle_NullCalendar", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_7__pyx_unpickle_NullCalendar, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_6__pyx_unpickle_NullCalendar};
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_7__pyx_unpickle_NullCalendar(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_NullCalendar (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NullCalendar", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NullCalendar", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_NullCalendar") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NullCalendar", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_NullCalendar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_6__pyx_unpickle_NullCalendar(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_6__pyx_unpickle_NullCalendar(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NullCalendar", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = NullCalendar.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = NullCalendar.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = NullCalendar.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NullCalendar__set_state(<NullCalendar> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_NullCalendar), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = NullCalendar.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NullCalendar__set_state(<NullCalendar> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = NullCalendar.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NullCalendar__set_state(<NullCalendar> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_NullCalendar__set_state(NullCalendar __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_NullCalendar__set_state(((struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = NullCalendar.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NullCalendar__set_state(<NullCalendar> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NullCalendar__set_state(<NullCalendar> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_NullCalendar__set_state(NullCalendar __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_NullCalendar(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_NullCalendar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_NullCalendar__set_state(<NullCalendar> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NullCalendar__set_state(NullCalendar __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_NullCalendar__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NullCalendar__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NullCalendar__set_state(NullCalendar __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_NullCalendar__set_state(NullCalendar __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NullCalendar__set_state(NullCalendar __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_NullCalendar__set_state(<NullCalendar> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NullCalendar__set_state(NullCalendar __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_NullCalendar__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ChinaCFFEXImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_9__pyx_unpickle_ChinaCFFEXImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_8__pyx_unpickle_ChinaCFFEXImpl[] = "__pyx_unpickle_ChinaCFFEXImpl(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_6ultron_10tradingday_8Calendar_9__pyx_unpickle_ChinaCFFEXImpl = {"__pyx_unpickle_ChinaCFFEXImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_9__pyx_unpickle_ChinaCFFEXImpl, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_8__pyx_unpickle_ChinaCFFEXImpl};
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_9__pyx_unpickle_ChinaCFFEXImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChinaCFFEXImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChinaCFFEXImpl", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChinaCFFEXImpl", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_ChinaCFFEXImpl") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChinaCFFEXImpl", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_ChinaCFFEXImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_8__pyx_unpickle_ChinaCFFEXImpl(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_8__pyx_unpickle_ChinaCFFEXImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChinaCFFEXImpl", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaCFFEXImpl.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = ChinaCFFEXImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaCFFEXImpl.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChinaCFFEXImpl__set_state(<ChinaCFFEXImpl> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaCFFEXImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ChinaCFFEXImpl__set_state(<ChinaCFFEXImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = ChinaCFFEXImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChinaCFFEXImpl__set_state(<ChinaCFFEXImpl> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaCFFEXImpl__set_state(ChinaCFFEXImpl __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaCFFEXImpl__set_state(((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = ChinaCFFEXImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ChinaCFFEXImpl__set_state(<ChinaCFFEXImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChinaCFFEXImpl__set_state(<ChinaCFFEXImpl> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ChinaCFFEXImpl__set_state(ChinaCFFEXImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ChinaCFFEXImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_ChinaCFFEXImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ChinaCFFEXImpl__set_state(<ChinaCFFEXImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaCFFEXImpl__set_state(ChinaCFFEXImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_ChinaCFFEXImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChinaCFFEXImpl__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaCFFEXImpl__set_state(ChinaCFFEXImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ChinaCFFEXImpl__set_state(ChinaCFFEXImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaCFFEXImpl__set_state(ChinaCFFEXImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ChinaCFFEXImpl__set_state(<ChinaCFFEXImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaCFFEXImpl__set_state(ChinaCFFEXImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_ChinaCFFEXImpl__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_WestenImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11__pyx_unpickle_WestenImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_10__pyx_unpickle_WestenImpl[] = "__pyx_unpickle_WestenImpl(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_6ultron_10tradingday_8Calendar_11__pyx_unpickle_WestenImpl = {"__pyx_unpickle_WestenImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_11__pyx_unpickle_WestenImpl, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_10__pyx_unpickle_WestenImpl};
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_11__pyx_unpickle_WestenImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_WestenImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_WestenImpl", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_WestenImpl", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_WestenImpl") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_WestenImpl", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_WestenImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_10__pyx_unpickle_WestenImpl(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_10__pyx_unpickle_WestenImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_WestenImpl", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = WestenImpl.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = WestenImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = WestenImpl.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_WestenImpl__set_state(<WestenImpl> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_WestenImpl), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = WestenImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_WestenImpl__set_state(<WestenImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = WestenImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_WestenImpl__set_state(<WestenImpl> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_WestenImpl__set_state(WestenImpl __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_WestenImpl__set_state(((struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = WestenImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_WestenImpl__set_state(<WestenImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_WestenImpl__set_state(<WestenImpl> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_WestenImpl__set_state(WestenImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_WestenImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_WestenImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_WestenImpl__set_state(<WestenImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_WestenImpl__set_state(WestenImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_WestenImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_WestenImpl__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_WestenImpl__set_state(WestenImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_WestenImpl__set_state(WestenImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_WestenImpl__set_state(WestenImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_WestenImpl__set_state(<WestenImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_WestenImpl__set_state(WestenImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_WestenImpl__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_TargetImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_13__pyx_unpickle_TargetImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6ultron_10tradingday_8Calendar_12__pyx_unpickle_TargetImpl[] = "__pyx_unpickle_TargetImpl(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_6ultron_10tradingday_8Calendar_13__pyx_unpickle_TargetImpl = {"__pyx_unpickle_TargetImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_13__pyx_unpickle_TargetImpl, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_12__pyx_unpickle_TargetImpl};
static PyObject *__pyx_pw_6ultron_10tradingday_8Calendar_13__pyx_unpickle_TargetImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TargetImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TargetImpl", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TargetImpl", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_TargetImpl") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TargetImpl", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_TargetImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6ultron_10tradingday_8Calendar_12__pyx_unpickle_TargetImpl(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6ultron_10tradingday_8Calendar_12__pyx_unpickle_TargetImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TargetImpl", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = TargetImpl.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = TargetImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = TargetImpl.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TargetImpl__set_state(<TargetImpl> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_TargetImpl), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = TargetImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TargetImpl__set_state(<TargetImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = TargetImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TargetImpl__set_state(<TargetImpl> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TargetImpl__set_state(TargetImpl __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_TargetImpl__set_state(((struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = TargetImpl.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TargetImpl__set_state(<TargetImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TargetImpl__set_state(<TargetImpl> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TargetImpl__set_state(TargetImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TargetImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_TargetImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_TargetImpl__set_state(<TargetImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TargetImpl__set_state(TargetImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_6ultron_10tradingday_8Calendar___pyx_unpickle_TargetImpl__set_state(struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TargetImpl__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TargetImpl__set_state(TargetImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_TargetImpl__set_state(TargetImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TargetImpl__set_state(TargetImpl __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_TargetImpl__set_state(<TargetImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TargetImpl__set_state(TargetImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ultron.tradingday.Calendar.__pyx_unpickle_TargetImpl__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl __pyx_vtable_6ultron_10tradingday_8Calendar_CalendarImpl;

static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_CalendarImpl(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)o);
  p->__pyx_vtab = __pyx_vtabptr_6ultron_10tradingday_8Calendar_CalendarImpl;
  return o;
}

static void __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_CalendarImpl(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_6ultron_10tradingday_8Calendar_CalendarImpl[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_12CalendarImpl_1__reduce_cython__, METH_NOARGS, __pyx_doc_6ultron_10tradingday_8Calendar_12CalendarImpl___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_12CalendarImpl_3__setstate_cython__, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_12CalendarImpl_2__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl = {
  PyVarObject_HEAD_INIT(0, 0)
  "ultron.tradingday.Calendar.CalendarImpl", /*tp_name*/
  sizeof(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_CalendarImpl, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ultron_10tradingday_8Calendar_CalendarImpl, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ultron_10tradingday_8Calendar_CalendarImpl, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_Calendar __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar;

static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_Calendar(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)o);
  p->__pyx_vtab = __pyx_vtabptr_6ultron_10tradingday_8Calendar_Calendar;
  p->_impl = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)Py_None); Py_INCREF(Py_None);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_Calendar(PyObject *o) {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *p = (struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_impl);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6ultron_10tradingday_8Calendar_Calendar(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *p = (struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)o;
  if (p->_impl) {
    e = (*v)(((PyObject *)p->_impl), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6ultron_10tradingday_8Calendar_Calendar(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *p = (struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *)o;
  tmp = ((PyObject*)p->_impl);
  p->_impl = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6ultron_10tradingday_8Calendar_8Calendar__impl(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5_impl_1__get__(o);
}

static int __pyx_setprop_6ultron_10tradingday_8Calendar_8Calendar__impl(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5_impl_3__set__(o, v);
  }
  else {
    return __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5_impl_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6ultron_10tradingday_8Calendar_8Calendar_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_4name_1__get__(o);
}

static int __pyx_setprop_6ultron_10tradingday_8Calendar_8Calendar_name(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_4name_3__set__(o, v);
  }
  else {
    return __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_4name_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_6ultron_10tradingday_8Calendar_Calendar[] = {
  {"isBizDay", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_3isBizDay, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_2isBizDay},
  {"isHoliday", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_5isHoliday, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_4isHoliday},
  {"isWeekEnd", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_7isWeekEnd, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_6isWeekEnd},
  {"isEndOfMonth", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_9isEndOfMonth, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_8isEndOfMonth},
  {"endOfMonth", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_11endOfMonth, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_10endOfMonth},
  {"bizDaysBetween", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_13bizDaysBetween, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_12bizDaysBetween},
  {"adjustDate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_15adjustDate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_14adjustDate},
  {"advanceDate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_17advanceDate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_16advanceDate},
  {"holDatesList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_19holDatesList, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_18holDatesList},
  {"bizDatesList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_21bizDatesList, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_20bizDatesList},
  {"__deepcopy__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_25__deepcopy__, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_24__deepcopy__},
  {"__reduce__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_27__reduce__, METH_NOARGS, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_26__reduce__},
  {"__setstate__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_29__setstate__, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_8Calendar_28__setstate__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6ultron_10tradingday_8Calendar_Calendar[] = {
  {(char *)"_impl", __pyx_getprop_6ultron_10tradingday_8Calendar_8Calendar__impl, __pyx_setprop_6ultron_10tradingday_8Calendar_8Calendar__impl, (char *)"_impl: ultron.tradingday.Calendar.CalendarImpl", 0},
  {(char *)"name", __pyx_getprop_6ultron_10tradingday_8Calendar_8Calendar_name, __pyx_setprop_6ultron_10tradingday_8Calendar_8Calendar_name, (char *)"name: str", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ultron_10tradingday_8Calendar_Calendar = {
  PyVarObject_HEAD_INIT(0, 0)
  "ultron.tradingday.Calendar.Calendar", /*tp_name*/
  sizeof(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_Calendar, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Calendar(str holCenter)", /*tp_doc*/
  __pyx_tp_traverse_6ultron_10tradingday_8Calendar_Calendar, /*tp_traverse*/
  __pyx_tp_clear_6ultron_10tradingday_8Calendar_Calendar, /*tp_clear*/
  __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_23__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ultron_10tradingday_8Calendar_Calendar, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6ultron_10tradingday_8Calendar_Calendar, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6ultron_10tradingday_8Calendar_8Calendar_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ultron_10tradingday_8Calendar_Calendar, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaSseImpl __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaSseImpl;

static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_ChinaSseImpl(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *p;
  PyObject *o = __pyx_tp_new_6ultron_10tradingday_8Calendar_CalendarImpl(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl*)__pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaSseImpl;
  return o;
}

static PyMethodDef __pyx_methods_6ultron_10tradingday_8Calendar_ChinaSseImpl[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_5__reduce_cython__, METH_NOARGS, __pyx_doc_6ultron_10tradingday_8Calendar_12ChinaSseImpl_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_7__setstate_cython__, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_12ChinaSseImpl_6__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl = {
  PyVarObject_HEAD_INIT(0, 0)
  "ultron.tradingday.Calendar.ChinaSseImpl", /*tp_name*/
  sizeof(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_CalendarImpl, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "ChinaSseImpl()", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_3__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ultron_10tradingday_8Calendar_ChinaSseImpl, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6ultron_10tradingday_8Calendar_12ChinaSseImpl_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ultron_10tradingday_8Calendar_ChinaSseImpl, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaIBImpl __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaIBImpl;

static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_ChinaIBImpl(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *p;
  PyObject *o = __pyx_tp_new_6ultron_10tradingday_8Calendar_CalendarImpl(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl*)__pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaIBImpl;
  return o;
}

static PyMethodDef __pyx_methods_6ultron_10tradingday_8Calendar_ChinaIBImpl[] = {
  {"isBizDay", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_3isBizDay, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_11ChinaIBImpl_2isBizDay},
  {"isWeekEnd", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_5isWeekEnd, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_11ChinaIBImpl_4isWeekEnd},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_9__reduce_cython__, METH_NOARGS, __pyx_doc_6ultron_10tradingday_8Calendar_11ChinaIBImpl_8__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_11__setstate_cython__, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_11ChinaIBImpl_10__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl = {
  PyVarObject_HEAD_INIT(0, 0)
  "ultron.tradingday.Calendar.ChinaIBImpl", /*tp_name*/
  sizeof(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_CalendarImpl, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "ChinaIBImpl()", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_7__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ultron_10tradingday_8Calendar_ChinaIBImpl, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6ultron_10tradingday_8Calendar_11ChinaIBImpl_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ultron_10tradingday_8Calendar_ChinaIBImpl, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_NullCalendar __pyx_vtable_6ultron_10tradingday_8Calendar_NullCalendar;

static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_NullCalendar(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *p;
  PyObject *o = __pyx_tp_new_6ultron_10tradingday_8Calendar_CalendarImpl(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl*)__pyx_vtabptr_6ultron_10tradingday_8Calendar_NullCalendar;
  return o;
}

static PyMethodDef __pyx_methods_6ultron_10tradingday_8Calendar_NullCalendar[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_5__reduce_cython__, METH_NOARGS, __pyx_doc_6ultron_10tradingday_8Calendar_12NullCalendar_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_7__setstate_cython__, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_12NullCalendar_6__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ultron_10tradingday_8Calendar_NullCalendar = {
  PyVarObject_HEAD_INIT(0, 0)
  "ultron.tradingday.Calendar.NullCalendar", /*tp_name*/
  sizeof(struct __pyx_obj_6ultron_10tradingday_8Calendar_NullCalendar), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_CalendarImpl, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "NullCalendar()", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_3__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ultron_10tradingday_8Calendar_NullCalendar, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6ultron_10tradingday_8Calendar_12NullCalendar_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ultron_10tradingday_8Calendar_NullCalendar, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl;

static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *p;
  PyObject *o = __pyx_tp_new_6ultron_10tradingday_8Calendar_CalendarImpl(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl*)__pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl;
  return o;
}

static PyMethodDef __pyx_methods_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_5__reduce_cython__, METH_NOARGS, __pyx_doc_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_7__setstate_cython__, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_6__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl = {
  PyVarObject_HEAD_INIT(0, 0)
  "ultron.tradingday.Calendar.ChinaCFFEXImpl", /*tp_name*/
  sizeof(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_CalendarImpl, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "ChinaCFFEXImpl()", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_3__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_WestenImpl __pyx_vtable_6ultron_10tradingday_8Calendar_WestenImpl;

static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_WestenImpl(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *p;
  PyObject *o = __pyx_tp_new_6ultron_10tradingday_8Calendar_CalendarImpl(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl*)__pyx_vtabptr_6ultron_10tradingday_8Calendar_WestenImpl;
  return o;
}

static PyMethodDef __pyx_methods_6ultron_10tradingday_8Calendar_WestenImpl[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_10WestenImpl_1__reduce_cython__, METH_NOARGS, __pyx_doc_6ultron_10tradingday_8Calendar_10WestenImpl___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_10WestenImpl_3__setstate_cython__, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_10WestenImpl_2__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ultron_10tradingday_8Calendar_WestenImpl = {
  PyVarObject_HEAD_INIT(0, 0)
  "ultron.tradingday.Calendar.WestenImpl", /*tp_name*/
  sizeof(struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_CalendarImpl, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ultron_10tradingday_8Calendar_WestenImpl, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ultron_10tradingday_8Calendar_WestenImpl, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_TargetImpl __pyx_vtable_6ultron_10tradingday_8Calendar_TargetImpl;

static PyObject *__pyx_tp_new_6ultron_10tradingday_8Calendar_TargetImpl(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *p;
  PyObject *o = __pyx_tp_new_6ultron_10tradingday_8Calendar_WestenImpl(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6ultron_10tradingday_8Calendar_CalendarImpl*)__pyx_vtabptr_6ultron_10tradingday_8Calendar_TargetImpl;
  return o;
}

static PyMethodDef __pyx_methods_6ultron_10tradingday_8Calendar_TargetImpl[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_5__reduce_cython__, METH_NOARGS, __pyx_doc_6ultron_10tradingday_8Calendar_10TargetImpl_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_7__setstate_cython__, METH_O, __pyx_doc_6ultron_10tradingday_8Calendar_10TargetImpl_6__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6ultron_10tradingday_8Calendar_TargetImpl = {
  PyVarObject_HEAD_INIT(0, 0)
  "ultron.tradingday.Calendar.TargetImpl", /*tp_name*/
  sizeof(struct __pyx_obj_6ultron_10tradingday_8Calendar_TargetImpl), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6ultron_10tradingday_8Calendar_CalendarImpl, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "TargetImpl()", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_3__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6ultron_10tradingday_8Calendar_TargetImpl, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6ultron_10tradingday_8Calendar_10TargetImpl_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6ultron_10tradingday_8Calendar_TargetImpl, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_Calendar(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_Calendar},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "Calendar",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_0_is_not_a_valid_description_of, __pyx_k_0_is_not_a_valid_description_of, sizeof(__pyx_k_0_is_not_a_valid_description_of), 0, 0, 1, 0},
  {&__pyx_n_s_Calendar, __pyx_k_Calendar, sizeof(__pyx_k_Calendar), 0, 0, 1, 1},
  {&__pyx_n_s_CalendarImpl, __pyx_k_CalendarImpl, sizeof(__pyx_k_CalendarImpl), 0, 0, 1, 1},
  {&__pyx_n_s_ChinaCFFEXImpl, __pyx_k_ChinaCFFEXImpl, sizeof(__pyx_k_ChinaCFFEXImpl), 0, 0, 1, 1},
  {&__pyx_n_s_ChinaIBImpl, __pyx_k_ChinaIBImpl, sizeof(__pyx_k_ChinaIBImpl), 0, 0, 1, 1},
  {&__pyx_n_s_ChinaSseImpl, __pyx_k_ChinaSseImpl, sizeof(__pyx_k_ChinaSseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_k_Incompatible_checksums_s_vs_0xd4, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd4), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_NullCalendar, __pyx_k_NullCalendar, sizeof(__pyx_k_NullCalendar), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_TargetImpl, __pyx_k_TargetImpl, sizeof(__pyx_k_TargetImpl), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_WestenImpl, __pyx_k_WestenImpl, sizeof(__pyx_k_WestenImpl), 0, 0, 1, 1},
  {&__pyx_n_s_adjustDate, __pyx_k_adjustDate, sizeof(__pyx_k_adjustDate), 0, 0, 1, 1},
  {&__pyx_n_s_advanceDate, __pyx_k_advanceDate, sizeof(__pyx_k_advanceDate), 0, 0, 1, 1},
  {&__pyx_n_s_bizDatesList, __pyx_k_bizDatesList, sizeof(__pyx_k_bizDatesList), 0, 0, 1, 1},
  {&__pyx_n_s_bizDaysBetween, __pyx_k_bizDaysBetween, sizeof(__pyx_k_bizDaysBetween), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_kp_s_china_cffex, __pyx_k_china_cffex, sizeof(__pyx_k_china_cffex), 0, 0, 1, 0},
  {&__pyx_kp_s_china_ib, __pyx_k_china_ib, sizeof(__pyx_k_china_ib), 0, 0, 1, 0},
  {&__pyx_kp_s_china_sse, __pyx_k_china_sse, sizeof(__pyx_k_china_sse), 0, 0, 1, 0},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_endOfMonth, __pyx_k_endOfMonth, sizeof(__pyx_k_endOfMonth), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fromDate, __pyx_k_fromDate, sizeof(__pyx_k_fromDate), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_holCenter, __pyx_k_holCenter, sizeof(__pyx_k_holCenter), 0, 0, 1, 1},
  {&__pyx_n_s_holDatesList, __pyx_k_holDatesList, sizeof(__pyx_k_holDatesList), 0, 0, 1, 1},
  {&__pyx_n_s_impl, __pyx_k_impl, sizeof(__pyx_k_impl), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_includeFirst, __pyx_k_includeFirst, sizeof(__pyx_k_includeFirst), 0, 0, 1, 1},
  {&__pyx_n_s_includeLast, __pyx_k_includeLast, sizeof(__pyx_k_includeLast), 0, 0, 1, 1},
  {&__pyx_n_s_includeWeekEnds, __pyx_k_includeWeekEnds, sizeof(__pyx_k_includeWeekEnds), 0, 0, 1, 1},
  {&__pyx_n_s_isBizDay, __pyx_k_isBizDay, sizeof(__pyx_k_isBizDay), 0, 0, 1, 1},
  {&__pyx_n_s_isEndOfMonth, __pyx_k_isEndOfMonth, sizeof(__pyx_k_isEndOfMonth), 0, 0, 1, 1},
  {&__pyx_n_s_isHoliday, __pyx_k_isHoliday, sizeof(__pyx_k_isHoliday), 0, 0, 1, 1},
  {&__pyx_n_s_isWeekEnd, __pyx_k_isWeekEnd, sizeof(__pyx_k_isWeekEnd), 0, 0, 1, 1},
  {&__pyx_n_s_lower, __pyx_k_lower, sizeof(__pyx_k_lower), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_month, __pyx_k_month, sizeof(__pyx_k_month), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_null, __pyx_k_null, sizeof(__pyx_k_null), 0, 0, 1, 1},
  {&__pyx_n_s_nullcalendar, __pyx_k_nullcalendar, sizeof(__pyx_k_nullcalendar), 0, 0, 1, 1},
  {&__pyx_n_s_period, __pyx_k_period, sizeof(__pyx_k_period), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_CalendarImpl, __pyx_k_pyx_unpickle_CalendarImpl, sizeof(__pyx_k_pyx_unpickle_CalendarImpl), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_ChinaCFFEXImpl, __pyx_k_pyx_unpickle_ChinaCFFEXImpl, sizeof(__pyx_k_pyx_unpickle_ChinaCFFEXImpl), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_ChinaIBImpl, __pyx_k_pyx_unpickle_ChinaIBImpl, sizeof(__pyx_k_pyx_unpickle_ChinaIBImpl), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_ChinaSseImpl, __pyx_k_pyx_unpickle_ChinaSseImpl, sizeof(__pyx_k_pyx_unpickle_ChinaSseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_NullCalendar, __pyx_k_pyx_unpickle_NullCalendar, sizeof(__pyx_k_pyx_unpickle_NullCalendar), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_TargetImpl, __pyx_k_pyx_unpickle_TargetImpl, sizeof(__pyx_k_pyx_unpickle_TargetImpl), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_WestenImpl, __pyx_k_pyx_unpickle_WestenImpl, sizeof(__pyx_k_pyx_unpickle_WestenImpl), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_toDate, __pyx_k_toDate, sizeof(__pyx_k_toDate), 0, 0, 1, 1},
  {&__pyx_n_s_ultron_tradingday_Calendar, __pyx_k_ultron_tradingday_Calendar, sizeof(__pyx_k_ultron_tradingday_Calendar), 0, 0, 1, 1},
  {&__pyx_kp_s_unknown_business_day_convention, __pyx_k_unknown_business_day_convention, sizeof(__pyx_k_unknown_business_day_convention), 0, 0, 1, 0},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_westernStyle, __pyx_k_westernStyle, sizeof(__pyx_k_westernStyle), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 15, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 16, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ultron/tradingday/Calendar.pyx":96
 *                 return d1
 *         else:
 *             raise ValueError("unknown business-day convention")             # <<<<<<<<<<<<<<
 *         return d1
 * 
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_unknown_business_day_convention); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "ultron/tradingday/Calendar.pyx":177
 * 
 * cdef set sse_holDays = {
 *     Date(1991, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1991, 2, 15),
 *     Date(1991, 2, 18),
 */
  __pyx_tuple__4 = PyTuple_Pack(3, __pyx_int_1991, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "ultron/tradingday/Calendar.pyx":178
 * cdef set sse_holDays = {
 *     Date(1991, 1, 1),
 *     Date(1991, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(1991, 2, 18),
 *     Date(1991, 5, 1),
 */
  __pyx_tuple__5 = PyTuple_Pack(3, __pyx_int_1991, __pyx_int_2, __pyx_int_15); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "ultron/tradingday/Calendar.pyx":179
 *     Date(1991, 1, 1),
 *     Date(1991, 2, 15),
 *     Date(1991, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(1991, 5, 1),
 *     Date(1991, 10, 1),
 */
  __pyx_tuple__6 = PyTuple_Pack(3, __pyx_int_1991, __pyx_int_2, __pyx_int_18); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "ultron/tradingday/Calendar.pyx":180
 *     Date(1991, 2, 15),
 *     Date(1991, 2, 18),
 *     Date(1991, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1991, 10, 1),
 *     Date(1991, 10, 2),
 */
  __pyx_tuple__7 = PyTuple_Pack(3, __pyx_int_1991, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "ultron/tradingday/Calendar.pyx":181
 *     Date(1991, 2, 18),
 *     Date(1991, 5, 1),
 *     Date(1991, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1991, 10, 2),
 *     Date(1992, 1, 1),
 */
  __pyx_tuple__8 = PyTuple_Pack(3, __pyx_int_1991, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "ultron/tradingday/Calendar.pyx":182
 *     Date(1991, 5, 1),
 *     Date(1991, 10, 1),
 *     Date(1991, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1992, 1, 1),
 *     Date(1992, 2, 4),
 */
  __pyx_tuple__9 = PyTuple_Pack(3, __pyx_int_1991, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "ultron/tradingday/Calendar.pyx":183
 *     Date(1991, 10, 1),
 *     Date(1991, 10, 2),
 *     Date(1992, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1992, 2, 4),
 *     Date(1992, 2, 5),
 */
  __pyx_tuple__10 = PyTuple_Pack(3, __pyx_int_1992, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "ultron/tradingday/Calendar.pyx":184
 *     Date(1991, 10, 2),
 *     Date(1992, 1, 1),
 *     Date(1992, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(1992, 2, 5),
 *     Date(1992, 2, 6),
 */
  __pyx_tuple__11 = PyTuple_Pack(3, __pyx_int_1992, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "ultron/tradingday/Calendar.pyx":185
 *     Date(1992, 1, 1),
 *     Date(1992, 2, 4),
 *     Date(1992, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(1992, 2, 6),
 *     Date(1992, 5, 1),
 */
  __pyx_tuple__12 = PyTuple_Pack(3, __pyx_int_1992, __pyx_int_2, __pyx_int_5); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "ultron/tradingday/Calendar.pyx":186
 *     Date(1992, 2, 4),
 *     Date(1992, 2, 5),
 *     Date(1992, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(1992, 5, 1),
 *     Date(1992, 10, 1),
 */
  __pyx_tuple__13 = PyTuple_Pack(3, __pyx_int_1992, __pyx_int_2, __pyx_int_6); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "ultron/tradingday/Calendar.pyx":187
 *     Date(1992, 2, 5),
 *     Date(1992, 2, 6),
 *     Date(1992, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1992, 10, 1),
 *     Date(1992, 10, 2),
 */
  __pyx_tuple__14 = PyTuple_Pack(3, __pyx_int_1992, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "ultron/tradingday/Calendar.pyx":188
 *     Date(1992, 2, 6),
 *     Date(1992, 5, 1),
 *     Date(1992, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1992, 10, 2),
 *     Date(1993, 1, 1),
 */
  __pyx_tuple__15 = PyTuple_Pack(3, __pyx_int_1992, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "ultron/tradingday/Calendar.pyx":189
 *     Date(1992, 5, 1),
 *     Date(1992, 10, 1),
 *     Date(1992, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1993, 1, 1),
 *     Date(1993, 1, 25),
 */
  __pyx_tuple__16 = PyTuple_Pack(3, __pyx_int_1992, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "ultron/tradingday/Calendar.pyx":190
 *     Date(1992, 10, 1),
 *     Date(1992, 10, 2),
 *     Date(1993, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1993, 1, 25),
 *     Date(1993, 1, 26),
 */
  __pyx_tuple__17 = PyTuple_Pack(3, __pyx_int_1993, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "ultron/tradingday/Calendar.pyx":191
 *     Date(1992, 10, 2),
 *     Date(1993, 1, 1),
 *     Date(1993, 1, 25),             # <<<<<<<<<<<<<<
 *     Date(1993, 1, 26),
 *     Date(1993, 10, 1),
 */
  __pyx_tuple__18 = PyTuple_Pack(3, __pyx_int_1993, __pyx_int_1, __pyx_int_25); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "ultron/tradingday/Calendar.pyx":192
 *     Date(1993, 1, 1),
 *     Date(1993, 1, 25),
 *     Date(1993, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(1993, 10, 1),
 *     Date(1994, 2, 7),
 */
  __pyx_tuple__19 = PyTuple_Pack(3, __pyx_int_1993, __pyx_int_1, __pyx_int_26); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "ultron/tradingday/Calendar.pyx":193
 *     Date(1993, 1, 25),
 *     Date(1993, 1, 26),
 *     Date(1993, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 7),
 *     Date(1994, 2, 8),
 */
  __pyx_tuple__20 = PyTuple_Pack(3, __pyx_int_1993, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "ultron/tradingday/Calendar.pyx":194
 *     Date(1993, 1, 26),
 *     Date(1993, 10, 1),
 *     Date(1994, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 8),
 *     Date(1994, 2, 9),
 */
  __pyx_tuple__21 = PyTuple_Pack(3, __pyx_int_1994, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "ultron/tradingday/Calendar.pyx":195
 *     Date(1993, 10, 1),
 *     Date(1994, 2, 7),
 *     Date(1994, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 9),
 *     Date(1994, 2, 10),
 */
  __pyx_tuple__22 = PyTuple_Pack(3, __pyx_int_1994, __pyx_int_2, __pyx_int_8); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "ultron/tradingday/Calendar.pyx":196
 *     Date(1994, 2, 7),
 *     Date(1994, 2, 8),
 *     Date(1994, 2, 9),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 10),
 *     Date(1994, 2, 11),
 */
  __pyx_tuple__23 = PyTuple_Pack(3, __pyx_int_1994, __pyx_int_2, __pyx_int_9); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "ultron/tradingday/Calendar.pyx":197
 *     Date(1994, 2, 8),
 *     Date(1994, 2, 9),
 *     Date(1994, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 11),
 *     Date(1994, 5, 2),
 */
  __pyx_tuple__24 = PyTuple_Pack(3, __pyx_int_1994, __pyx_int_2, __pyx_int_10); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "ultron/tradingday/Calendar.pyx":198
 *     Date(1994, 2, 9),
 *     Date(1994, 2, 10),
 *     Date(1994, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(1994, 5, 2),
 *     Date(1994, 10, 3),
 */
  __pyx_tuple__25 = PyTuple_Pack(3, __pyx_int_1994, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "ultron/tradingday/Calendar.pyx":199
 *     Date(1994, 2, 10),
 *     Date(1994, 2, 11),
 *     Date(1994, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(1994, 10, 3),
 *     Date(1994, 10, 4),
 */
  __pyx_tuple__26 = PyTuple_Pack(3, __pyx_int_1994, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "ultron/tradingday/Calendar.pyx":200
 *     Date(1994, 2, 11),
 *     Date(1994, 5, 2),
 *     Date(1994, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(1994, 10, 4),
 *     Date(1995, 1, 2),
 */
  __pyx_tuple__27 = PyTuple_Pack(3, __pyx_int_1994, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "ultron/tradingday/Calendar.pyx":201
 *     Date(1994, 5, 2),
 *     Date(1994, 10, 3),
 *     Date(1994, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(1995, 1, 2),
 *     Date(1995, 1, 30),
 */
  __pyx_tuple__28 = PyTuple_Pack(3, __pyx_int_1994, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "ultron/tradingday/Calendar.pyx":202
 *     Date(1994, 10, 3),
 *     Date(1994, 10, 4),
 *     Date(1995, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(1995, 1, 30),
 *     Date(1995, 1, 31),
 */
  __pyx_tuple__29 = PyTuple_Pack(3, __pyx_int_1995, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "ultron/tradingday/Calendar.pyx":203
 *     Date(1994, 10, 4),
 *     Date(1995, 1, 2),
 *     Date(1995, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(1995, 1, 31),
 *     Date(1995, 2, 1),
 */
  __pyx_tuple__30 = PyTuple_Pack(3, __pyx_int_1995, __pyx_int_1, __pyx_int_30); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "ultron/tradingday/Calendar.pyx":204
 *     Date(1995, 1, 2),
 *     Date(1995, 1, 30),
 *     Date(1995, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(1995, 2, 1),
 *     Date(1995, 2, 2),
 */
  __pyx_tuple__31 = PyTuple_Pack(3, __pyx_int_1995, __pyx_int_1, __pyx_int_31); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "ultron/tradingday/Calendar.pyx":205
 *     Date(1995, 1, 30),
 *     Date(1995, 1, 31),
 *     Date(1995, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(1995, 2, 2),
 *     Date(1995, 2, 3),
 */
  __pyx_tuple__32 = PyTuple_Pack(3, __pyx_int_1995, __pyx_int_2, __pyx_int_1); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "ultron/tradingday/Calendar.pyx":206
 *     Date(1995, 1, 31),
 *     Date(1995, 2, 1),
 *     Date(1995, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(1995, 2, 3),
 *     Date(1995, 5, 1),
 */
  __pyx_tuple__33 = PyTuple_Pack(3, __pyx_int_1995, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "ultron/tradingday/Calendar.pyx":207
 *     Date(1995, 2, 1),
 *     Date(1995, 2, 2),
 *     Date(1995, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(1995, 5, 1),
 *     Date(1995, 10, 2),
 */
  __pyx_tuple__34 = PyTuple_Pack(3, __pyx_int_1995, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "ultron/tradingday/Calendar.pyx":208
 *     Date(1995, 2, 2),
 *     Date(1995, 2, 3),
 *     Date(1995, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1995, 10, 2),
 *     Date(1995, 10, 3),
 */
  __pyx_tuple__35 = PyTuple_Pack(3, __pyx_int_1995, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "ultron/tradingday/Calendar.pyx":209
 *     Date(1995, 2, 3),
 *     Date(1995, 5, 1),
 *     Date(1995, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1995, 10, 3),
 *     Date(1996, 1, 1),
 */
  __pyx_tuple__36 = PyTuple_Pack(3, __pyx_int_1995, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "ultron/tradingday/Calendar.pyx":210
 *     Date(1995, 5, 1),
 *     Date(1995, 10, 2),
 *     Date(1995, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(1996, 1, 1),
 *     Date(1996, 2, 19),
 */
  __pyx_tuple__37 = PyTuple_Pack(3, __pyx_int_1995, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "ultron/tradingday/Calendar.pyx":211
 *     Date(1995, 10, 2),
 *     Date(1995, 10, 3),
 *     Date(1996, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 19),
 *     Date(1996, 2, 20),
 */
  __pyx_tuple__38 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "ultron/tradingday/Calendar.pyx":212
 *     Date(1995, 10, 3),
 *     Date(1996, 1, 1),
 *     Date(1996, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 20),
 *     Date(1996, 2, 21),
 */
  __pyx_tuple__39 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_2, __pyx_int_19); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "ultron/tradingday/Calendar.pyx":213
 *     Date(1996, 1, 1),
 *     Date(1996, 2, 19),
 *     Date(1996, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 21),
 *     Date(1996, 2, 22),
 */
  __pyx_tuple__40 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_2, __pyx_int_20); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "ultron/tradingday/Calendar.pyx":214
 *     Date(1996, 2, 19),
 *     Date(1996, 2, 20),
 *     Date(1996, 2, 21),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 22),
 *     Date(1996, 2, 23),
 */
  __pyx_tuple__41 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_2, __pyx_int_21); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "ultron/tradingday/Calendar.pyx":215
 *     Date(1996, 2, 20),
 *     Date(1996, 2, 21),
 *     Date(1996, 2, 22),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 23),
 *     Date(1996, 2, 26),
 */
  __pyx_tuple__42 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_2, __pyx_int_22); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "ultron/tradingday/Calendar.pyx":216
 *     Date(1996, 2, 21),
 *     Date(1996, 2, 22),
 *     Date(1996, 2, 23),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 26),
 *     Date(1996, 2, 27),
 */
  __pyx_tuple__43 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_2, __pyx_int_23); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "ultron/tradingday/Calendar.pyx":217
 *     Date(1996, 2, 22),
 *     Date(1996, 2, 23),
 *     Date(1996, 2, 26),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 27),
 *     Date(1996, 2, 28),
 */
  __pyx_tuple__44 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_2, __pyx_int_26); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "ultron/tradingday/Calendar.pyx":218
 *     Date(1996, 2, 23),
 *     Date(1996, 2, 26),
 *     Date(1996, 2, 27),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 28),
 *     Date(1996, 2, 29),
 */
  __pyx_tuple__45 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_2, __pyx_int_27); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "ultron/tradingday/Calendar.pyx":219
 *     Date(1996, 2, 26),
 *     Date(1996, 2, 27),
 *     Date(1996, 2, 28),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 29),
 *     Date(1996, 3, 1),
 */
  __pyx_tuple__46 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_2, __pyx_int_28); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "ultron/tradingday/Calendar.pyx":220
 *     Date(1996, 2, 27),
 *     Date(1996, 2, 28),
 *     Date(1996, 2, 29),             # <<<<<<<<<<<<<<
 *     Date(1996, 3, 1),
 *     Date(1996, 5, 1),
 */
  __pyx_tuple__47 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_2, __pyx_int_29); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "ultron/tradingday/Calendar.pyx":221
 *     Date(1996, 2, 28),
 *     Date(1996, 2, 29),
 *     Date(1996, 3, 1),             # <<<<<<<<<<<<<<
 *     Date(1996, 5, 1),
 *     Date(1996, 9, 30),
 */
  __pyx_tuple__48 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_3, __pyx_int_1); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "ultron/tradingday/Calendar.pyx":222
 *     Date(1996, 2, 29),
 *     Date(1996, 3, 1),
 *     Date(1996, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1996, 9, 30),
 *     Date(1996, 10, 1),
 */
  __pyx_tuple__49 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "ultron/tradingday/Calendar.pyx":223
 *     Date(1996, 3, 1),
 *     Date(1996, 5, 1),
 *     Date(1996, 9, 30),             # <<<<<<<<<<<<<<
 *     Date(1996, 10, 1),
 *     Date(1996, 10, 2),
 */
  __pyx_tuple__50 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_9, __pyx_int_30); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "ultron/tradingday/Calendar.pyx":224
 *     Date(1996, 5, 1),
 *     Date(1996, 9, 30),
 *     Date(1996, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1996, 10, 2),
 *     Date(1997, 1, 1),
 */
  __pyx_tuple__51 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "ultron/tradingday/Calendar.pyx":225
 *     Date(1996, 9, 30),
 *     Date(1996, 10, 1),
 *     Date(1996, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1997, 1, 1),
 *     Date(1997, 2, 3),
 */
  __pyx_tuple__52 = PyTuple_Pack(3, __pyx_int_1996, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "ultron/tradingday/Calendar.pyx":226
 *     Date(1996, 10, 1),
 *     Date(1996, 10, 2),
 *     Date(1997, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 3),
 *     Date(1997, 2, 4),
 */
  __pyx_tuple__53 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "ultron/tradingday/Calendar.pyx":227
 *     Date(1996, 10, 2),
 *     Date(1997, 1, 1),
 *     Date(1997, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 4),
 *     Date(1997, 2, 5),
 */
  __pyx_tuple__54 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "ultron/tradingday/Calendar.pyx":228
 *     Date(1997, 1, 1),
 *     Date(1997, 2, 3),
 *     Date(1997, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 5),
 *     Date(1997, 2, 6),
 */
  __pyx_tuple__55 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);

  /* "ultron/tradingday/Calendar.pyx":229
 *     Date(1997, 2, 3),
 *     Date(1997, 2, 4),
 *     Date(1997, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 6),
 *     Date(1997, 2, 7),
 */
  __pyx_tuple__56 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_5); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "ultron/tradingday/Calendar.pyx":230
 *     Date(1997, 2, 4),
 *     Date(1997, 2, 5),
 *     Date(1997, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 7),
 *     Date(1997, 2, 10),
 */
  __pyx_tuple__57 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_6); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "ultron/tradingday/Calendar.pyx":231
 *     Date(1997, 2, 5),
 *     Date(1997, 2, 6),
 *     Date(1997, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 10),
 *     Date(1997, 2, 11),
 */
  __pyx_tuple__58 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "ultron/tradingday/Calendar.pyx":232
 *     Date(1997, 2, 6),
 *     Date(1997, 2, 7),
 *     Date(1997, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 11),
 *     Date(1997, 2, 12),
 */
  __pyx_tuple__59 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_10); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "ultron/tradingday/Calendar.pyx":233
 *     Date(1997, 2, 7),
 *     Date(1997, 2, 10),
 *     Date(1997, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 12),
 *     Date(1997, 2, 13),
 */
  __pyx_tuple__60 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "ultron/tradingday/Calendar.pyx":234
 *     Date(1997, 2, 10),
 *     Date(1997, 2, 11),
 *     Date(1997, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 13),
 *     Date(1997, 2, 14),
 */
  __pyx_tuple__61 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_12); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);

  /* "ultron/tradingday/Calendar.pyx":235
 *     Date(1997, 2, 11),
 *     Date(1997, 2, 12),
 *     Date(1997, 2, 13),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 14),
 *     Date(1997, 5, 1),
 */
  __pyx_tuple__62 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_13); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "ultron/tradingday/Calendar.pyx":236
 *     Date(1997, 2, 12),
 *     Date(1997, 2, 13),
 *     Date(1997, 2, 14),             # <<<<<<<<<<<<<<
 *     Date(1997, 5, 1),
 *     Date(1997, 5, 2),
 */
  __pyx_tuple__63 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_2, __pyx_int_14); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);

  /* "ultron/tradingday/Calendar.pyx":237
 *     Date(1997, 2, 13),
 *     Date(1997, 2, 14),
 *     Date(1997, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1997, 5, 2),
 *     Date(1997, 6, 30),
 */
  __pyx_tuple__64 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "ultron/tradingday/Calendar.pyx":238
 *     Date(1997, 2, 14),
 *     Date(1997, 5, 1),
 *     Date(1997, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(1997, 6, 30),
 *     Date(1997, 7, 1),
 */
  __pyx_tuple__65 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "ultron/tradingday/Calendar.pyx":239
 *     Date(1997, 5, 1),
 *     Date(1997, 5, 2),
 *     Date(1997, 6, 30),             # <<<<<<<<<<<<<<
 *     Date(1997, 7, 1),
 *     Date(1997, 10, 1),
 */
  __pyx_tuple__66 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_6, __pyx_int_30); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "ultron/tradingday/Calendar.pyx":240
 *     Date(1997, 5, 2),
 *     Date(1997, 6, 30),
 *     Date(1997, 7, 1),             # <<<<<<<<<<<<<<
 *     Date(1997, 10, 1),
 *     Date(1997, 10, 2),
 */
  __pyx_tuple__67 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_7, __pyx_int_1); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "ultron/tradingday/Calendar.pyx":241
 *     Date(1997, 6, 30),
 *     Date(1997, 7, 1),
 *     Date(1997, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1997, 10, 2),
 *     Date(1997, 10, 3),
 */
  __pyx_tuple__68 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "ultron/tradingday/Calendar.pyx":242
 *     Date(1997, 7, 1),
 *     Date(1997, 10, 1),
 *     Date(1997, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1997, 10, 3),
 *     Date(1998, 1, 1),
 */
  __pyx_tuple__69 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);

  /* "ultron/tradingday/Calendar.pyx":243
 *     Date(1997, 10, 1),
 *     Date(1997, 10, 2),
 *     Date(1997, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 1),
 *     Date(1998, 1, 2),
 */
  __pyx_tuple__70 = PyTuple_Pack(3, __pyx_int_1997, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);

  /* "ultron/tradingday/Calendar.pyx":244
 *     Date(1997, 10, 2),
 *     Date(1997, 10, 3),
 *     Date(1998, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 2),
 *     Date(1998, 1, 26),
 */
  __pyx_tuple__71 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "ultron/tradingday/Calendar.pyx":245
 *     Date(1997, 10, 3),
 *     Date(1998, 1, 1),
 *     Date(1998, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 26),
 *     Date(1998, 1, 27),
 */
  __pyx_tuple__72 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);

  /* "ultron/tradingday/Calendar.pyx":246
 *     Date(1998, 1, 1),
 *     Date(1998, 1, 2),
 *     Date(1998, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 27),
 *     Date(1998, 1, 28),
 */
  __pyx_tuple__73 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_1, __pyx_int_26); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "ultron/tradingday/Calendar.pyx":247
 *     Date(1998, 1, 2),
 *     Date(1998, 1, 26),
 *     Date(1998, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 28),
 *     Date(1998, 1, 29),
 */
  __pyx_tuple__74 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_1, __pyx_int_27); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "ultron/tradingday/Calendar.pyx":248
 *     Date(1998, 1, 26),
 *     Date(1998, 1, 27),
 *     Date(1998, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 29),
 *     Date(1998, 1, 30),
 */
  __pyx_tuple__75 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_1, __pyx_int_28); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);

  /* "ultron/tradingday/Calendar.pyx":249
 *     Date(1998, 1, 27),
 *     Date(1998, 1, 28),
 *     Date(1998, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 30),
 *     Date(1998, 2, 2),
 */
  __pyx_tuple__76 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_1, __pyx_int_29); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);

  /* "ultron/tradingday/Calendar.pyx":250
 *     Date(1998, 1, 28),
 *     Date(1998, 1, 29),
 *     Date(1998, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 2),
 *     Date(1998, 2, 3),
 */
  __pyx_tuple__77 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_1, __pyx_int_30); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);

  /* "ultron/tradingday/Calendar.pyx":251
 *     Date(1998, 1, 29),
 *     Date(1998, 1, 30),
 *     Date(1998, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 3),
 *     Date(1998, 2, 4),
 */
  __pyx_tuple__78 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);

  /* "ultron/tradingday/Calendar.pyx":252
 *     Date(1998, 1, 30),
 *     Date(1998, 2, 2),
 *     Date(1998, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 4),
 *     Date(1998, 2, 5),
 */
  __pyx_tuple__79 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);

  /* "ultron/tradingday/Calendar.pyx":253
 *     Date(1998, 2, 2),
 *     Date(1998, 2, 3),
 *     Date(1998, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 5),
 *     Date(1998, 2, 6),
 */
  __pyx_tuple__80 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);

  /* "ultron/tradingday/Calendar.pyx":254
 *     Date(1998, 2, 3),
 *     Date(1998, 2, 4),
 *     Date(1998, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 6),
 *     Date(1998, 5, 1),
 */
  __pyx_tuple__81 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_2, __pyx_int_5); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "ultron/tradingday/Calendar.pyx":255
 *     Date(1998, 2, 4),
 *     Date(1998, 2, 5),
 *     Date(1998, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(1998, 5, 1),
 *     Date(1998, 10, 1),
 */
  __pyx_tuple__82 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_2, __pyx_int_6); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);

  /* "ultron/tradingday/Calendar.pyx":256
 *     Date(1998, 2, 5),
 *     Date(1998, 2, 6),
 *     Date(1998, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1998, 10, 1),
 *     Date(1998, 10, 2),
 */
  __pyx_tuple__83 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "ultron/tradingday/Calendar.pyx":257
 *     Date(1998, 2, 6),
 *     Date(1998, 5, 1),
 *     Date(1998, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1998, 10, 2),
 *     Date(1999, 1, 1),
 */
  __pyx_tuple__84 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);

  /* "ultron/tradingday/Calendar.pyx":258
 *     Date(1998, 5, 1),
 *     Date(1998, 10, 1),
 *     Date(1998, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1999, 1, 1),
 *     Date(1999, 2, 10),
 */
  __pyx_tuple__85 = PyTuple_Pack(3, __pyx_int_1998, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "ultron/tradingday/Calendar.pyx":259
 *     Date(1998, 10, 1),
 *     Date(1998, 10, 2),
 *     Date(1999, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 10),
 *     Date(1999, 2, 11),
 */
  __pyx_tuple__86 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "ultron/tradingday/Calendar.pyx":260
 *     Date(1998, 10, 2),
 *     Date(1999, 1, 1),
 *     Date(1999, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 11),
 *     Date(1999, 2, 12),
 */
  __pyx_tuple__87 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_10); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);

  /* "ultron/tradingday/Calendar.pyx":261
 *     Date(1999, 1, 1),
 *     Date(1999, 2, 10),
 *     Date(1999, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 12),
 *     Date(1999, 2, 15),
 */
  __pyx_tuple__88 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);

  /* "ultron/tradingday/Calendar.pyx":262
 *     Date(1999, 2, 10),
 *     Date(1999, 2, 11),
 *     Date(1999, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 15),
 *     Date(1999, 2, 16),
 */
  __pyx_tuple__89 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_12); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);

  /* "ultron/tradingday/Calendar.pyx":263
 *     Date(1999, 2, 11),
 *     Date(1999, 2, 12),
 *     Date(1999, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 16),
 *     Date(1999, 2, 17),
 */
  __pyx_tuple__90 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_15); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "ultron/tradingday/Calendar.pyx":264
 *     Date(1999, 2, 12),
 *     Date(1999, 2, 15),
 *     Date(1999, 2, 16),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 17),
 *     Date(1999, 2, 18),
 */
  __pyx_tuple__91 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_16); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);

  /* "ultron/tradingday/Calendar.pyx":265
 *     Date(1999, 2, 15),
 *     Date(1999, 2, 16),
 *     Date(1999, 2, 17),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 18),
 *     Date(1999, 2, 19),
 */
  __pyx_tuple__92 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_17); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);

  /* "ultron/tradingday/Calendar.pyx":266
 *     Date(1999, 2, 16),
 *     Date(1999, 2, 17),
 *     Date(1999, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 19),
 *     Date(1999, 2, 22),
 */
  __pyx_tuple__93 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_18); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);

  /* "ultron/tradingday/Calendar.pyx":267
 *     Date(1999, 2, 17),
 *     Date(1999, 2, 18),
 *     Date(1999, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 22),
 *     Date(1999, 2, 23),
 */
  __pyx_tuple__94 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_19); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);

  /* "ultron/tradingday/Calendar.pyx":268
 *     Date(1999, 2, 18),
 *     Date(1999, 2, 19),
 *     Date(1999, 2, 22),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 23),
 *     Date(1999, 2, 24),
 */
  __pyx_tuple__95 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_22); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);

  /* "ultron/tradingday/Calendar.pyx":269
 *     Date(1999, 2, 19),
 *     Date(1999, 2, 22),
 *     Date(1999, 2, 23),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 24),
 *     Date(1999, 2, 25),
 */
  __pyx_tuple__96 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_23); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);

  /* "ultron/tradingday/Calendar.pyx":270
 *     Date(1999, 2, 22),
 *     Date(1999, 2, 23),
 *     Date(1999, 2, 24),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 25),
 *     Date(1999, 2, 26),
 */
  __pyx_tuple__97 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_24); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);

  /* "ultron/tradingday/Calendar.pyx":271
 *     Date(1999, 2, 23),
 *     Date(1999, 2, 24),
 *     Date(1999, 2, 25),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 26),
 *     Date(1999, 5, 3),
 */
  __pyx_tuple__98 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_25); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);

  /* "ultron/tradingday/Calendar.pyx":272
 *     Date(1999, 2, 24),
 *     Date(1999, 2, 25),
 *     Date(1999, 2, 26),             # <<<<<<<<<<<<<<
 *     Date(1999, 5, 3),
 *     Date(1999, 10, 1),
 */
  __pyx_tuple__99 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_2, __pyx_int_26); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);

  /* "ultron/tradingday/Calendar.pyx":273
 *     Date(1999, 2, 25),
 *     Date(1999, 2, 26),
 *     Date(1999, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 1),
 *     Date(1999, 10, 4),
 */
  __pyx_tuple__100 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);

  /* "ultron/tradingday/Calendar.pyx":274
 *     Date(1999, 2, 26),
 *     Date(1999, 5, 3),
 *     Date(1999, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 4),
 *     Date(1999, 10, 5),
 */
  __pyx_tuple__101 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);

  /* "ultron/tradingday/Calendar.pyx":275
 *     Date(1999, 5, 3),
 *     Date(1999, 10, 1),
 *     Date(1999, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 5),
 *     Date(1999, 10, 6),
 */
  __pyx_tuple__102 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);

  /* "ultron/tradingday/Calendar.pyx":276
 *     Date(1999, 10, 1),
 *     Date(1999, 10, 4),
 *     Date(1999, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 6),
 *     Date(1999, 10, 7),
 */
  __pyx_tuple__103 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);

  /* "ultron/tradingday/Calendar.pyx":277
 *     Date(1999, 10, 4),
 *     Date(1999, 10, 5),
 *     Date(1999, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 7),
 *     Date(1999, 12, 20),
 */
  __pyx_tuple__104 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);

  /* "ultron/tradingday/Calendar.pyx":278
 *     Date(1999, 10, 5),
 *     Date(1999, 10, 6),
 *     Date(1999, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(1999, 12, 20),
 *     Date(1999, 12, 31),
 */
  __pyx_tuple__105 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);

  /* "ultron/tradingday/Calendar.pyx":279
 *     Date(1999, 10, 6),
 *     Date(1999, 10, 7),
 *     Date(1999, 12, 20),             # <<<<<<<<<<<<<<
 *     Date(1999, 12, 31),
 *     Date(2000, 1, 3),
 */
  __pyx_tuple__106 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_12, __pyx_int_20); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);

  /* "ultron/tradingday/Calendar.pyx":280
 *     Date(1999, 10, 7),
 *     Date(1999, 12, 20),
 *     Date(1999, 12, 31),             # <<<<<<<<<<<<<<
 *     Date(2000, 1, 3),
 *     Date(2000, 1, 31),
 */
  __pyx_tuple__107 = PyTuple_Pack(3, __pyx_int_1999, __pyx_int_12, __pyx_int_31); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);

  /* "ultron/tradingday/Calendar.pyx":281
 *     Date(1999, 12, 20),
 *     Date(1999, 12, 31),
 *     Date(2000, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2000, 1, 31),
 *     Date(2000, 2, 1),
 */
  __pyx_tuple__108 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_1, __pyx_int_3); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);

  /* "ultron/tradingday/Calendar.pyx":282
 *     Date(1999, 12, 31),
 *     Date(2000, 1, 3),
 *     Date(2000, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 1),
 *     Date(2000, 2, 2),
 */
  __pyx_tuple__109 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_1, __pyx_int_31); if (unlikely(!__pyx_tuple__109)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);

  /* "ultron/tradingday/Calendar.pyx":283
 *     Date(2000, 1, 3),
 *     Date(2000, 1, 31),
 *     Date(2000, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 2),
 *     Date(2000, 2, 3),
 */
  __pyx_tuple__110 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_2, __pyx_int_1); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);

  /* "ultron/tradingday/Calendar.pyx":284
 *     Date(2000, 1, 31),
 *     Date(2000, 2, 1),
 *     Date(2000, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 3),
 *     Date(2000, 2, 4),
 */
  __pyx_tuple__111 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);

  /* "ultron/tradingday/Calendar.pyx":285
 *     Date(2000, 2, 1),
 *     Date(2000, 2, 2),
 *     Date(2000, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 4),
 *     Date(2000, 2, 7),
 */
  __pyx_tuple__112 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);

  /* "ultron/tradingday/Calendar.pyx":286
 *     Date(2000, 2, 2),
 *     Date(2000, 2, 3),
 *     Date(2000, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 7),
 *     Date(2000, 2, 8),
 */
  __pyx_tuple__113 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__113)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);

  /* "ultron/tradingday/Calendar.pyx":287
 *     Date(2000, 2, 3),
 *     Date(2000, 2, 4),
 *     Date(2000, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 8),
 *     Date(2000, 2, 9),
 */
  __pyx_tuple__114 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);

  /* "ultron/tradingday/Calendar.pyx":288
 *     Date(2000, 2, 4),
 *     Date(2000, 2, 7),
 *     Date(2000, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 9),
 *     Date(2000, 2, 10),
 */
  __pyx_tuple__115 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_2, __pyx_int_8); if (unlikely(!__pyx_tuple__115)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);

  /* "ultron/tradingday/Calendar.pyx":289
 *     Date(2000, 2, 7),
 *     Date(2000, 2, 8),
 *     Date(2000, 2, 9),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 10),
 *     Date(2000, 2, 11),
 */
  __pyx_tuple__116 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_2, __pyx_int_9); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);

  /* "ultron/tradingday/Calendar.pyx":290
 *     Date(2000, 2, 8),
 *     Date(2000, 2, 9),
 *     Date(2000, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 11),
 *     Date(2000, 5, 1),
 */
  __pyx_tuple__117 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_2, __pyx_int_10); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);

  /* "ultron/tradingday/Calendar.pyx":291
 *     Date(2000, 2, 9),
 *     Date(2000, 2, 10),
 *     Date(2000, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 1),
 *     Date(2000, 5, 2),
 */
  __pyx_tuple__118 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);

  /* "ultron/tradingday/Calendar.pyx":292
 *     Date(2000, 2, 10),
 *     Date(2000, 2, 11),
 *     Date(2000, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 2),
 *     Date(2000, 5, 3),
 */
  __pyx_tuple__119 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);

  /* "ultron/tradingday/Calendar.pyx":293
 *     Date(2000, 2, 11),
 *     Date(2000, 5, 1),
 *     Date(2000, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 3),
 *     Date(2000, 5, 4),
 */
  __pyx_tuple__120 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);

  /* "ultron/tradingday/Calendar.pyx":294
 *     Date(2000, 5, 1),
 *     Date(2000, 5, 2),
 *     Date(2000, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 4),
 *     Date(2000, 5, 5),
 */
  __pyx_tuple__121 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__121)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);

  /* "ultron/tradingday/Calendar.pyx":295
 *     Date(2000, 5, 2),
 *     Date(2000, 5, 3),
 *     Date(2000, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 5),
 *     Date(2000, 10, 2),
 */
  __pyx_tuple__122 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_5, __pyx_int_4); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);

  /* "ultron/tradingday/Calendar.pyx":296
 *     Date(2000, 5, 3),
 *     Date(2000, 5, 4),
 *     Date(2000, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 2),
 *     Date(2000, 10, 3),
 */
  __pyx_tuple__123 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_5, __pyx_int_5); if (unlikely(!__pyx_tuple__123)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);

  /* "ultron/tradingday/Calendar.pyx":297
 *     Date(2000, 5, 4),
 *     Date(2000, 5, 5),
 *     Date(2000, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 3),
 *     Date(2000, 10, 4),
 */
  __pyx_tuple__124 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);

  /* "ultron/tradingday/Calendar.pyx":298
 *     Date(2000, 5, 5),
 *     Date(2000, 10, 2),
 *     Date(2000, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 4),
 *     Date(2000, 10, 5),
 */
  __pyx_tuple__125 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);

  /* "ultron/tradingday/Calendar.pyx":299
 *     Date(2000, 10, 2),
 *     Date(2000, 10, 3),
 *     Date(2000, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 5),
 *     Date(2000, 10, 6),
 */
  __pyx_tuple__126 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);

  /* "ultron/tradingday/Calendar.pyx":300
 *     Date(2000, 10, 3),
 *     Date(2000, 10, 4),
 *     Date(2000, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 6),
 *     Date(2001, 1, 1),
 */
  __pyx_tuple__127 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);

  /* "ultron/tradingday/Calendar.pyx":301
 *     Date(2000, 10, 4),
 *     Date(2000, 10, 5),
 *     Date(2000, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 1),
 *     Date(2001, 1, 22),
 */
  __pyx_tuple__128 = PyTuple_Pack(3, __pyx_int_2000, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);

  /* "ultron/tradingday/Calendar.pyx":302
 *     Date(2000, 10, 5),
 *     Date(2000, 10, 6),
 *     Date(2001, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 22),
 *     Date(2001, 1, 23),
 */
  __pyx_tuple__129 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);

  /* "ultron/tradingday/Calendar.pyx":303
 *     Date(2000, 10, 6),
 *     Date(2001, 1, 1),
 *     Date(2001, 1, 22),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 23),
 *     Date(2001, 1, 24),
 */
  __pyx_tuple__130 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_1, __pyx_int_22); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);

  /* "ultron/tradingday/Calendar.pyx":304
 *     Date(2001, 1, 1),
 *     Date(2001, 1, 22),
 *     Date(2001, 1, 23),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 24),
 *     Date(2001, 1, 25),
 */
  __pyx_tuple__131 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_1, __pyx_int_23); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);

  /* "ultron/tradingday/Calendar.pyx":305
 *     Date(2001, 1, 22),
 *     Date(2001, 1, 23),
 *     Date(2001, 1, 24),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 25),
 *     Date(2001, 1, 26),
 */
  __pyx_tuple__132 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_1, __pyx_int_24); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);

  /* "ultron/tradingday/Calendar.pyx":306
 *     Date(2001, 1, 23),
 *     Date(2001, 1, 24),
 *     Date(2001, 1, 25),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 26),
 *     Date(2001, 1, 29),
 */
  __pyx_tuple__133 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_1, __pyx_int_25); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);

  /* "ultron/tradingday/Calendar.pyx":307
 *     Date(2001, 1, 24),
 *     Date(2001, 1, 25),
 *     Date(2001, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 29),
 *     Date(2001, 1, 30),
 */
  __pyx_tuple__134 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_1, __pyx_int_26); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);

  /* "ultron/tradingday/Calendar.pyx":308
 *     Date(2001, 1, 25),
 *     Date(2001, 1, 26),
 *     Date(2001, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 30),
 *     Date(2001, 1, 31),
 */
  __pyx_tuple__135 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_1, __pyx_int_29); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);

  /* "ultron/tradingday/Calendar.pyx":309
 *     Date(2001, 1, 26),
 *     Date(2001, 1, 29),
 *     Date(2001, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 31),
 *     Date(2001, 2, 1),
 */
  __pyx_tuple__136 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_1, __pyx_int_30); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);

  /* "ultron/tradingday/Calendar.pyx":310
 *     Date(2001, 1, 29),
 *     Date(2001, 1, 30),
 *     Date(2001, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2001, 2, 1),
 *     Date(2001, 2, 2),
 */
  __pyx_tuple__137 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_1, __pyx_int_31); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);

  /* "ultron/tradingday/Calendar.pyx":311
 *     Date(2001, 1, 30),
 *     Date(2001, 1, 31),
 *     Date(2001, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2001, 2, 2),
 *     Date(2001, 5, 1),
 */
  __pyx_tuple__138 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_2, __pyx_int_1); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);

  /* "ultron/tradingday/Calendar.pyx":312
 *     Date(2001, 1, 31),
 *     Date(2001, 2, 1),
 *     Date(2001, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 1),
 *     Date(2001, 5, 2),
 */
  __pyx_tuple__139 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);

  /* "ultron/tradingday/Calendar.pyx":313
 *     Date(2001, 2, 1),
 *     Date(2001, 2, 2),
 *     Date(2001, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 2),
 *     Date(2001, 5, 3),
 */
  __pyx_tuple__140 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);

  /* "ultron/tradingday/Calendar.pyx":314
 *     Date(2001, 2, 2),
 *     Date(2001, 5, 1),
 *     Date(2001, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 3),
 *     Date(2001, 5, 4),
 */
  __pyx_tuple__141 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);

  /* "ultron/tradingday/Calendar.pyx":315
 *     Date(2001, 5, 1),
 *     Date(2001, 5, 2),
 *     Date(2001, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 4),
 *     Date(2001, 5, 7),
 */
  __pyx_tuple__142 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);

  /* "ultron/tradingday/Calendar.pyx":316
 *     Date(2001, 5, 2),
 *     Date(2001, 5, 3),
 *     Date(2001, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 7),
 *     Date(2001, 10, 1),
 */
  __pyx_tuple__143 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_5, __pyx_int_4); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);

  /* "ultron/tradingday/Calendar.pyx":317
 *     Date(2001, 5, 3),
 *     Date(2001, 5, 4),
 *     Date(2001, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 1),
 *     Date(2001, 10, 2),
 */
  __pyx_tuple__144 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_5, __pyx_int_7); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);

  /* "ultron/tradingday/Calendar.pyx":318
 *     Date(2001, 5, 4),
 *     Date(2001, 5, 7),
 *     Date(2001, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 2),
 *     Date(2001, 10, 3),
 */
  __pyx_tuple__145 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);

  /* "ultron/tradingday/Calendar.pyx":319
 *     Date(2001, 5, 7),
 *     Date(2001, 10, 1),
 *     Date(2001, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 3),
 *     Date(2001, 10, 4),
 */
  __pyx_tuple__146 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);

  /* "ultron/tradingday/Calendar.pyx":320
 *     Date(2001, 10, 1),
 *     Date(2001, 10, 2),
 *     Date(2001, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 4),
 *     Date(2001, 10, 5),
 */
  __pyx_tuple__147 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);

  /* "ultron/tradingday/Calendar.pyx":321
 *     Date(2001, 10, 2),
 *     Date(2001, 10, 3),
 *     Date(2001, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 5),
 *     Date(2002, 1, 1),
 */
  __pyx_tuple__148 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);

  /* "ultron/tradingday/Calendar.pyx":322
 *     Date(2001, 10, 3),
 *     Date(2001, 10, 4),
 *     Date(2001, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2002, 1, 1),
 *     Date(2002, 1, 2),
 */
  __pyx_tuple__149 = PyTuple_Pack(3, __pyx_int_2001, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);

  /* "ultron/tradingday/Calendar.pyx":323
 *     Date(2001, 10, 4),
 *     Date(2001, 10, 5),
 *     Date(2002, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2002, 1, 2),
 *     Date(2002, 1, 3),
 */
  __pyx_tuple__150 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);

  /* "ultron/tradingday/Calendar.pyx":324
 *     Date(2001, 10, 5),
 *     Date(2002, 1, 1),
 *     Date(2002, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2002, 1, 3),
 *     Date(2002, 2, 11),
 */
  __pyx_tuple__151 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__151)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);

  /* "ultron/tradingday/Calendar.pyx":325
 *     Date(2002, 1, 1),
 *     Date(2002, 1, 2),
 *     Date(2002, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 11),
 *     Date(2002, 2, 12),
 */
  __pyx_tuple__152 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_1, __pyx_int_3); if (unlikely(!__pyx_tuple__152)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);

  /* "ultron/tradingday/Calendar.pyx":326
 *     Date(2002, 1, 2),
 *     Date(2002, 1, 3),
 *     Date(2002, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 12),
 *     Date(2002, 2, 13),
 */
  __pyx_tuple__153 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__153)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);

  /* "ultron/tradingday/Calendar.pyx":327
 *     Date(2002, 1, 3),
 *     Date(2002, 2, 11),
 *     Date(2002, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 13),
 *     Date(2002, 2, 14),
 */
  __pyx_tuple__154 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_12); if (unlikely(!__pyx_tuple__154)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__154);
  __Pyx_GIVEREF(__pyx_tuple__154);

  /* "ultron/tradingday/Calendar.pyx":328
 *     Date(2002, 2, 11),
 *     Date(2002, 2, 12),
 *     Date(2002, 2, 13),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 14),
 *     Date(2002, 2, 15),
 */
  __pyx_tuple__155 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_13); if (unlikely(!__pyx_tuple__155)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);

  /* "ultron/tradingday/Calendar.pyx":329
 *     Date(2002, 2, 12),
 *     Date(2002, 2, 13),
 *     Date(2002, 2, 14),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 15),
 *     Date(2002, 2, 18),
 */
  __pyx_tuple__156 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_14); if (unlikely(!__pyx_tuple__156)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__156);
  __Pyx_GIVEREF(__pyx_tuple__156);

  /* "ultron/tradingday/Calendar.pyx":330
 *     Date(2002, 2, 13),
 *     Date(2002, 2, 14),
 *     Date(2002, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 18),
 *     Date(2002, 2, 19),
 */
  __pyx_tuple__157 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_15); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);

  /* "ultron/tradingday/Calendar.pyx":331
 *     Date(2002, 2, 14),
 *     Date(2002, 2, 15),
 *     Date(2002, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 19),
 *     Date(2002, 2, 20),
 */
  __pyx_tuple__158 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_18); if (unlikely(!__pyx_tuple__158)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__158);
  __Pyx_GIVEREF(__pyx_tuple__158);

  /* "ultron/tradingday/Calendar.pyx":332
 *     Date(2002, 2, 15),
 *     Date(2002, 2, 18),
 *     Date(2002, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 20),
 *     Date(2002, 2, 21),
 */
  __pyx_tuple__159 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_19); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);

  /* "ultron/tradingday/Calendar.pyx":333
 *     Date(2002, 2, 18),
 *     Date(2002, 2, 19),
 *     Date(2002, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 21),
 *     Date(2002, 2, 22),
 */
  __pyx_tuple__160 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_20); if (unlikely(!__pyx_tuple__160)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__160);
  __Pyx_GIVEREF(__pyx_tuple__160);

  /* "ultron/tradingday/Calendar.pyx":334
 *     Date(2002, 2, 19),
 *     Date(2002, 2, 20),
 *     Date(2002, 2, 21),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 22),
 *     Date(2002, 5, 1),
 */
  __pyx_tuple__161 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_21); if (unlikely(!__pyx_tuple__161)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);

  /* "ultron/tradingday/Calendar.pyx":335
 *     Date(2002, 2, 20),
 *     Date(2002, 2, 21),
 *     Date(2002, 2, 22),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 1),
 *     Date(2002, 5, 2),
 */
  __pyx_tuple__162 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_2, __pyx_int_22); if (unlikely(!__pyx_tuple__162)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__162);
  __Pyx_GIVEREF(__pyx_tuple__162);

  /* "ultron/tradingday/Calendar.pyx":336
 *     Date(2002, 2, 21),
 *     Date(2002, 2, 22),
 *     Date(2002, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 2),
 *     Date(2002, 5, 3),
 */
  __pyx_tuple__163 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__163)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);

  /* "ultron/tradingday/Calendar.pyx":337
 *     Date(2002, 2, 22),
 *     Date(2002, 5, 1),
 *     Date(2002, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 3),
 *     Date(2002, 5, 6),
 */
  __pyx_tuple__164 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__164)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);

  /* "ultron/tradingday/Calendar.pyx":338
 *     Date(2002, 5, 1),
 *     Date(2002, 5, 2),
 *     Date(2002, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 6),
 *     Date(2002, 5, 7),
 */
  __pyx_tuple__165 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__165)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__165);
  __Pyx_GIVEREF(__pyx_tuple__165);

  /* "ultron/tradingday/Calendar.pyx":339
 *     Date(2002, 5, 2),
 *     Date(2002, 5, 3),
 *     Date(2002, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 7),
 *     Date(2002, 9, 30),
 */
  __pyx_tuple__166 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_5, __pyx_int_6); if (unlikely(!__pyx_tuple__166)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__166);
  __Pyx_GIVEREF(__pyx_tuple__166);

  /* "ultron/tradingday/Calendar.pyx":340
 *     Date(2002, 5, 3),
 *     Date(2002, 5, 6),
 *     Date(2002, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2002, 9, 30),
 *     Date(2002, 10, 1),
 */
  __pyx_tuple__167 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_5, __pyx_int_7); if (unlikely(!__pyx_tuple__167)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__167);
  __Pyx_GIVEREF(__pyx_tuple__167);

  /* "ultron/tradingday/Calendar.pyx":341
 *     Date(2002, 5, 6),
 *     Date(2002, 5, 7),
 *     Date(2002, 9, 30),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 1),
 *     Date(2002, 10, 2),
 */
  __pyx_tuple__168 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_9, __pyx_int_30); if (unlikely(!__pyx_tuple__168)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__168);
  __Pyx_GIVEREF(__pyx_tuple__168);

  /* "ultron/tradingday/Calendar.pyx":342
 *     Date(2002, 5, 7),
 *     Date(2002, 9, 30),
 *     Date(2002, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 2),
 *     Date(2002, 10, 3),
 */
  __pyx_tuple__169 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__169)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__169);
  __Pyx_GIVEREF(__pyx_tuple__169);

  /* "ultron/tradingday/Calendar.pyx":343
 *     Date(2002, 9, 30),
 *     Date(2002, 10, 1),
 *     Date(2002, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 3),
 *     Date(2002, 10, 4),
 */
  __pyx_tuple__170 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__170)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__170);
  __Pyx_GIVEREF(__pyx_tuple__170);

  /* "ultron/tradingday/Calendar.pyx":344
 *     Date(2002, 10, 1),
 *     Date(2002, 10, 2),
 *     Date(2002, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 4),
 *     Date(2002, 10, 7),
 */
  __pyx_tuple__171 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__171)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__171);
  __Pyx_GIVEREF(__pyx_tuple__171);

  /* "ultron/tradingday/Calendar.pyx":345
 *     Date(2002, 10, 2),
 *     Date(2002, 10, 3),
 *     Date(2002, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 7),
 *     Date(2003, 1, 1),
 */
  __pyx_tuple__172 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__172)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__172);
  __Pyx_GIVEREF(__pyx_tuple__172);

  /* "ultron/tradingday/Calendar.pyx":346
 *     Date(2002, 10, 3),
 *     Date(2002, 10, 4),
 *     Date(2002, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2003, 1, 1),
 *     Date(2003, 1, 30),
 */
  __pyx_tuple__173 = PyTuple_Pack(3, __pyx_int_2002, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__173)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__173);
  __Pyx_GIVEREF(__pyx_tuple__173);

  /* "ultron/tradingday/Calendar.pyx":347
 *     Date(2002, 10, 4),
 *     Date(2002, 10, 7),
 *     Date(2003, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2003, 1, 30),
 *     Date(2003, 1, 31),
 */
  __pyx_tuple__174 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__174)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__174);
  __Pyx_GIVEREF(__pyx_tuple__174);

  /* "ultron/tradingday/Calendar.pyx":348
 *     Date(2002, 10, 7),
 *     Date(2003, 1, 1),
 *     Date(2003, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2003, 1, 31),
 *     Date(2003, 2, 3),
 */
  __pyx_tuple__175 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_1, __pyx_int_30); if (unlikely(!__pyx_tuple__175)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__175);
  __Pyx_GIVEREF(__pyx_tuple__175);

  /* "ultron/tradingday/Calendar.pyx":349
 *     Date(2003, 1, 1),
 *     Date(2003, 1, 30),
 *     Date(2003, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 3),
 *     Date(2003, 2, 4),
 */
  __pyx_tuple__176 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_1, __pyx_int_31); if (unlikely(!__pyx_tuple__176)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__176);
  __Pyx_GIVEREF(__pyx_tuple__176);

  /* "ultron/tradingday/Calendar.pyx":350
 *     Date(2003, 1, 30),
 *     Date(2003, 1, 31),
 *     Date(2003, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 4),
 *     Date(2003, 2, 5),
 */
  __pyx_tuple__177 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__177)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__177);
  __Pyx_GIVEREF(__pyx_tuple__177);

  /* "ultron/tradingday/Calendar.pyx":351
 *     Date(2003, 1, 31),
 *     Date(2003, 2, 3),
 *     Date(2003, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 5),
 *     Date(2003, 2, 6),
 */
  __pyx_tuple__178 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__178)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__178);
  __Pyx_GIVEREF(__pyx_tuple__178);

  /* "ultron/tradingday/Calendar.pyx":352
 *     Date(2003, 2, 3),
 *     Date(2003, 2, 4),
 *     Date(2003, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 6),
 *     Date(2003, 2, 7),
 */
  __pyx_tuple__179 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_2, __pyx_int_5); if (unlikely(!__pyx_tuple__179)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);

  /* "ultron/tradingday/Calendar.pyx":353
 *     Date(2003, 2, 4),
 *     Date(2003, 2, 5),
 *     Date(2003, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 7),
 *     Date(2003, 5, 1),
 */
  __pyx_tuple__180 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_2, __pyx_int_6); if (unlikely(!__pyx_tuple__180)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__180);
  __Pyx_GIVEREF(__pyx_tuple__180);

  /* "ultron/tradingday/Calendar.pyx":354
 *     Date(2003, 2, 5),
 *     Date(2003, 2, 6),
 *     Date(2003, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 1),
 *     Date(2003, 5, 2),
 */
  __pyx_tuple__181 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__181)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__181);
  __Pyx_GIVEREF(__pyx_tuple__181);

  /* "ultron/tradingday/Calendar.pyx":355
 *     Date(2003, 2, 6),
 *     Date(2003, 2, 7),
 *     Date(2003, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 2),
 *     Date(2003, 5, 5),
 */
  __pyx_tuple__182 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__182)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__182);
  __Pyx_GIVEREF(__pyx_tuple__182);

  /* "ultron/tradingday/Calendar.pyx":356
 *     Date(2003, 2, 7),
 *     Date(2003, 5, 1),
 *     Date(2003, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 5),
 *     Date(2003, 5, 6),
 */
  __pyx_tuple__183 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__183)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);

  /* "ultron/tradingday/Calendar.pyx":357
 *     Date(2003, 5, 1),
 *     Date(2003, 5, 2),
 *     Date(2003, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 6),
 *     Date(2003, 5, 7),
 */
  __pyx_tuple__184 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_5, __pyx_int_5); if (unlikely(!__pyx_tuple__184)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__184);
  __Pyx_GIVEREF(__pyx_tuple__184);

  /* "ultron/tradingday/Calendar.pyx":358
 *     Date(2003, 5, 2),
 *     Date(2003, 5, 5),
 *     Date(2003, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 7),
 *     Date(2003, 5, 8),
 */
  __pyx_tuple__185 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_5, __pyx_int_6); if (unlikely(!__pyx_tuple__185)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__185);
  __Pyx_GIVEREF(__pyx_tuple__185);

  /* "ultron/tradingday/Calendar.pyx":359
 *     Date(2003, 5, 5),
 *     Date(2003, 5, 6),
 *     Date(2003, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 8),
 *     Date(2003, 5, 9),
 */
  __pyx_tuple__186 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_5, __pyx_int_7); if (unlikely(!__pyx_tuple__186)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__186);
  __Pyx_GIVEREF(__pyx_tuple__186);

  /* "ultron/tradingday/Calendar.pyx":360
 *     Date(2003, 5, 6),
 *     Date(2003, 5, 7),
 *     Date(2003, 5, 8),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 9),
 *     Date(2003, 10, 1),
 */
  __pyx_tuple__187 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_5, __pyx_int_8); if (unlikely(!__pyx_tuple__187)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__187);
  __Pyx_GIVEREF(__pyx_tuple__187);

  /* "ultron/tradingday/Calendar.pyx":361
 *     Date(2003, 5, 7),
 *     Date(2003, 5, 8),
 *     Date(2003, 5, 9),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 1),
 *     Date(2003, 10, 2),
 */
  __pyx_tuple__188 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_5, __pyx_int_9); if (unlikely(!__pyx_tuple__188)) __PYX_ERR(0, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__188);
  __Pyx_GIVEREF(__pyx_tuple__188);

  /* "ultron/tradingday/Calendar.pyx":362
 *     Date(2003, 5, 8),
 *     Date(2003, 5, 9),
 *     Date(2003, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 2),
 *     Date(2003, 10, 3),
 */
  __pyx_tuple__189 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__189)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__189);
  __Pyx_GIVEREF(__pyx_tuple__189);

  /* "ultron/tradingday/Calendar.pyx":363
 *     Date(2003, 5, 9),
 *     Date(2003, 10, 1),
 *     Date(2003, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 3),
 *     Date(2003, 10, 6),
 */
  __pyx_tuple__190 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__190)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__190);
  __Pyx_GIVEREF(__pyx_tuple__190);

  /* "ultron/tradingday/Calendar.pyx":364
 *     Date(2003, 10, 1),
 *     Date(2003, 10, 2),
 *     Date(2003, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 6),
 *     Date(2003, 10, 7),
 */
  __pyx_tuple__191 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__191)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__191);
  __Pyx_GIVEREF(__pyx_tuple__191);

  /* "ultron/tradingday/Calendar.pyx":365
 *     Date(2003, 10, 2),
 *     Date(2003, 10, 3),
 *     Date(2003, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 7),
 *     Date(2004, 1, 1),
 */
  __pyx_tuple__192 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__192)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__192);
  __Pyx_GIVEREF(__pyx_tuple__192);

  /* "ultron/tradingday/Calendar.pyx":366
 *     Date(2003, 10, 3),
 *     Date(2003, 10, 6),
 *     Date(2003, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 1),
 *     Date(2004, 1, 19),
 */
  __pyx_tuple__193 = PyTuple_Pack(3, __pyx_int_2003, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__193)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__193);
  __Pyx_GIVEREF(__pyx_tuple__193);

  /* "ultron/tradingday/Calendar.pyx":367
 *     Date(2003, 10, 6),
 *     Date(2003, 10, 7),
 *     Date(2004, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 19),
 *     Date(2004, 1, 20),
 */
  __pyx_tuple__194 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__194)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__194);
  __Pyx_GIVEREF(__pyx_tuple__194);

  /* "ultron/tradingday/Calendar.pyx":368
 *     Date(2003, 10, 7),
 *     Date(2004, 1, 1),
 *     Date(2004, 1, 19),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 20),
 *     Date(2004, 1, 21),
 */
  __pyx_tuple__195 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_1, __pyx_int_19); if (unlikely(!__pyx_tuple__195)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__195);
  __Pyx_GIVEREF(__pyx_tuple__195);

  /* "ultron/tradingday/Calendar.pyx":369
 *     Date(2004, 1, 1),
 *     Date(2004, 1, 19),
 *     Date(2004, 1, 20),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 21),
 *     Date(2004, 1, 22),
 */
  __pyx_tuple__196 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_1, __pyx_int_20); if (unlikely(!__pyx_tuple__196)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__196);
  __Pyx_GIVEREF(__pyx_tuple__196);

  /* "ultron/tradingday/Calendar.pyx":370
 *     Date(2004, 1, 19),
 *     Date(2004, 1, 20),
 *     Date(2004, 1, 21),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 22),
 *     Date(2004, 1, 23),
 */
  __pyx_tuple__197 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_1, __pyx_int_21); if (unlikely(!__pyx_tuple__197)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__197);
  __Pyx_GIVEREF(__pyx_tuple__197);

  /* "ultron/tradingday/Calendar.pyx":371
 *     Date(2004, 1, 20),
 *     Date(2004, 1, 21),
 *     Date(2004, 1, 22),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 23),
 *     Date(2004, 1, 26),
 */
  __pyx_tuple__198 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_1, __pyx_int_22); if (unlikely(!__pyx_tuple__198)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__198);
  __Pyx_GIVEREF(__pyx_tuple__198);

  /* "ultron/tradingday/Calendar.pyx":372
 *     Date(2004, 1, 21),
 *     Date(2004, 1, 22),
 *     Date(2004, 1, 23),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 26),
 *     Date(2004, 1, 27),
 */
  __pyx_tuple__199 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_1, __pyx_int_23); if (unlikely(!__pyx_tuple__199)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__199);
  __Pyx_GIVEREF(__pyx_tuple__199);

  /* "ultron/tradingday/Calendar.pyx":373
 *     Date(2004, 1, 22),
 *     Date(2004, 1, 23),
 *     Date(2004, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 27),
 *     Date(2004, 1, 28),
 */
  __pyx_tuple__200 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_1, __pyx_int_26); if (unlikely(!__pyx_tuple__200)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__200);
  __Pyx_GIVEREF(__pyx_tuple__200);

  /* "ultron/tradingday/Calendar.pyx":374
 *     Date(2004, 1, 23),
 *     Date(2004, 1, 26),
 *     Date(2004, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 28),
 *     Date(2004, 5, 3),
 */
  __pyx_tuple__201 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_1, __pyx_int_27); if (unlikely(!__pyx_tuple__201)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__201);
  __Pyx_GIVEREF(__pyx_tuple__201);

  /* "ultron/tradingday/Calendar.pyx":375
 *     Date(2004, 1, 26),
 *     Date(2004, 1, 27),
 *     Date(2004, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 3),
 *     Date(2004, 5, 4),
 */
  __pyx_tuple__202 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_1, __pyx_int_28); if (unlikely(!__pyx_tuple__202)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__202);
  __Pyx_GIVEREF(__pyx_tuple__202);

  /* "ultron/tradingday/Calendar.pyx":376
 *     Date(2004, 1, 27),
 *     Date(2004, 1, 28),
 *     Date(2004, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 4),
 *     Date(2004, 5, 5),
 */
  __pyx_tuple__203 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__203)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__203);
  __Pyx_GIVEREF(__pyx_tuple__203);

  /* "ultron/tradingday/Calendar.pyx":377
 *     Date(2004, 1, 28),
 *     Date(2004, 5, 3),
 *     Date(2004, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 5),
 *     Date(2004, 5, 6),
 */
  __pyx_tuple__204 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_5, __pyx_int_4); if (unlikely(!__pyx_tuple__204)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__204);
  __Pyx_GIVEREF(__pyx_tuple__204);

  /* "ultron/tradingday/Calendar.pyx":378
 *     Date(2004, 5, 3),
 *     Date(2004, 5, 4),
 *     Date(2004, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 6),
 *     Date(2004, 5, 7),
 */
  __pyx_tuple__205 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_5, __pyx_int_5); if (unlikely(!__pyx_tuple__205)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__205);
  __Pyx_GIVEREF(__pyx_tuple__205);

  /* "ultron/tradingday/Calendar.pyx":379
 *     Date(2004, 5, 4),
 *     Date(2004, 5, 5),
 *     Date(2004, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 7),
 *     Date(2004, 10, 1),
 */
  __pyx_tuple__206 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_5, __pyx_int_6); if (unlikely(!__pyx_tuple__206)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__206);
  __Pyx_GIVEREF(__pyx_tuple__206);

  /* "ultron/tradingday/Calendar.pyx":380
 *     Date(2004, 5, 5),
 *     Date(2004, 5, 6),
 *     Date(2004, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 1),
 *     Date(2004, 10, 4),
 */
  __pyx_tuple__207 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_5, __pyx_int_7); if (unlikely(!__pyx_tuple__207)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__207);
  __Pyx_GIVEREF(__pyx_tuple__207);

  /* "ultron/tradingday/Calendar.pyx":381
 *     Date(2004, 5, 6),
 *     Date(2004, 5, 7),
 *     Date(2004, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 4),
 *     Date(2004, 10, 5),
 */
  __pyx_tuple__208 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__208)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__208);
  __Pyx_GIVEREF(__pyx_tuple__208);

  /* "ultron/tradingday/Calendar.pyx":382
 *     Date(2004, 5, 7),
 *     Date(2004, 10, 1),
 *     Date(2004, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 5),
 *     Date(2004, 10, 6),
 */
  __pyx_tuple__209 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__209)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__209);
  __Pyx_GIVEREF(__pyx_tuple__209);

  /* "ultron/tradingday/Calendar.pyx":383
 *     Date(2004, 10, 1),
 *     Date(2004, 10, 4),
 *     Date(2004, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 6),
 *     Date(2004, 10, 7),
 */
  __pyx_tuple__210 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__210)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__210);
  __Pyx_GIVEREF(__pyx_tuple__210);

  /* "ultron/tradingday/Calendar.pyx":384
 *     Date(2004, 10, 4),
 *     Date(2004, 10, 5),
 *     Date(2004, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 7),
 *     Date(2005, 1, 3),
 */
  __pyx_tuple__211 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__211)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__211);
  __Pyx_GIVEREF(__pyx_tuple__211);

  /* "ultron/tradingday/Calendar.pyx":385
 *     Date(2004, 10, 5),
 *     Date(2004, 10, 6),
 *     Date(2004, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2005, 1, 3),
 *     Date(2005, 2, 7),
 */
  __pyx_tuple__212 = PyTuple_Pack(3, __pyx_int_2004, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__212)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__212);
  __Pyx_GIVEREF(__pyx_tuple__212);

  /* "ultron/tradingday/Calendar.pyx":386
 *     Date(2004, 10, 6),
 *     Date(2004, 10, 7),
 *     Date(2005, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 7),
 *     Date(2005, 2, 8),
 */
  __pyx_tuple__213 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_1, __pyx_int_3); if (unlikely(!__pyx_tuple__213)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__213);
  __Pyx_GIVEREF(__pyx_tuple__213);

  /* "ultron/tradingday/Calendar.pyx":387
 *     Date(2004, 10, 7),
 *     Date(2005, 1, 3),
 *     Date(2005, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 8),
 *     Date(2005, 2, 9),
 */
  __pyx_tuple__214 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__214)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__214);
  __Pyx_GIVEREF(__pyx_tuple__214);

  /* "ultron/tradingday/Calendar.pyx":388
 *     Date(2005, 1, 3),
 *     Date(2005, 2, 7),
 *     Date(2005, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 9),
 *     Date(2005, 2, 10),
 */
  __pyx_tuple__215 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_2, __pyx_int_8); if (unlikely(!__pyx_tuple__215)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__215);
  __Pyx_GIVEREF(__pyx_tuple__215);

  /* "ultron/tradingday/Calendar.pyx":389
 *     Date(2005, 2, 7),
 *     Date(2005, 2, 8),
 *     Date(2005, 2, 9),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 10),
 *     Date(2005, 2, 11),
 */
  __pyx_tuple__216 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_2, __pyx_int_9); if (unlikely(!__pyx_tuple__216)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__216);
  __Pyx_GIVEREF(__pyx_tuple__216);

  /* "ultron/tradingday/Calendar.pyx":390
 *     Date(2005, 2, 8),
 *     Date(2005, 2, 9),
 *     Date(2005, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 11),
 *     Date(2005, 2, 14),
 */
  __pyx_tuple__217 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_2, __pyx_int_10); if (unlikely(!__pyx_tuple__217)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__217);
  __Pyx_GIVEREF(__pyx_tuple__217);

  /* "ultron/tradingday/Calendar.pyx":391
 *     Date(2005, 2, 9),
 *     Date(2005, 2, 10),
 *     Date(2005, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 14),
 *     Date(2005, 2, 15),
 */
  __pyx_tuple__218 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__218)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__218);
  __Pyx_GIVEREF(__pyx_tuple__218);

  /* "ultron/tradingday/Calendar.pyx":392
 *     Date(2005, 2, 10),
 *     Date(2005, 2, 11),
 *     Date(2005, 2, 14),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 15),
 *     Date(2005, 5, 2),
 */
  __pyx_tuple__219 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_2, __pyx_int_14); if (unlikely(!__pyx_tuple__219)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__219);
  __Pyx_GIVEREF(__pyx_tuple__219);

  /* "ultron/tradingday/Calendar.pyx":393
 *     Date(2005, 2, 11),
 *     Date(2005, 2, 14),
 *     Date(2005, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 2),
 *     Date(2005, 5, 3),
 */
  __pyx_tuple__220 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_2, __pyx_int_15); if (unlikely(!__pyx_tuple__220)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__220);
  __Pyx_GIVEREF(__pyx_tuple__220);

  /* "ultron/tradingday/Calendar.pyx":394
 *     Date(2005, 2, 14),
 *     Date(2005, 2, 15),
 *     Date(2005, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 3),
 *     Date(2005, 5, 4),
 */
  __pyx_tuple__221 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__221)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__221);
  __Pyx_GIVEREF(__pyx_tuple__221);

  /* "ultron/tradingday/Calendar.pyx":395
 *     Date(2005, 2, 15),
 *     Date(2005, 5, 2),
 *     Date(2005, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 4),
 *     Date(2005, 5, 5),
 */
  __pyx_tuple__222 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__222)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__222);
  __Pyx_GIVEREF(__pyx_tuple__222);

  /* "ultron/tradingday/Calendar.pyx":396
 *     Date(2005, 5, 2),
 *     Date(2005, 5, 3),
 *     Date(2005, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 5),
 *     Date(2005, 5, 6),
 */
  __pyx_tuple__223 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_5, __pyx_int_4); if (unlikely(!__pyx_tuple__223)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__223);
  __Pyx_GIVEREF(__pyx_tuple__223);

  /* "ultron/tradingday/Calendar.pyx":397
 *     Date(2005, 5, 3),
 *     Date(2005, 5, 4),
 *     Date(2005, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 6),
 *     Date(2005, 10, 3),
 */
  __pyx_tuple__224 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_5, __pyx_int_5); if (unlikely(!__pyx_tuple__224)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__224);
  __Pyx_GIVEREF(__pyx_tuple__224);

  /* "ultron/tradingday/Calendar.pyx":398
 *     Date(2005, 5, 4),
 *     Date(2005, 5, 5),
 *     Date(2005, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 3),
 *     Date(2005, 10, 4),
 */
  __pyx_tuple__225 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_5, __pyx_int_6); if (unlikely(!__pyx_tuple__225)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__225);
  __Pyx_GIVEREF(__pyx_tuple__225);

  /* "ultron/tradingday/Calendar.pyx":399
 *     Date(2005, 5, 5),
 *     Date(2005, 5, 6),
 *     Date(2005, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 4),
 *     Date(2005, 10, 5),
 */
  __pyx_tuple__226 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__226)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__226);
  __Pyx_GIVEREF(__pyx_tuple__226);

  /* "ultron/tradingday/Calendar.pyx":400
 *     Date(2005, 5, 6),
 *     Date(2005, 10, 3),
 *     Date(2005, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 5),
 *     Date(2005, 10, 6),
 */
  __pyx_tuple__227 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__227)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__227);
  __Pyx_GIVEREF(__pyx_tuple__227);

  /* "ultron/tradingday/Calendar.pyx":401
 *     Date(2005, 10, 3),
 *     Date(2005, 10, 4),
 *     Date(2005, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 6),
 *     Date(2005, 10, 7),
 */
  __pyx_tuple__228 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__228)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__228);
  __Pyx_GIVEREF(__pyx_tuple__228);

  /* "ultron/tradingday/Calendar.pyx":402
 *     Date(2005, 10, 4),
 *     Date(2005, 10, 5),
 *     Date(2005, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 7),
 *     Date(2006, 1, 2),
 */
  __pyx_tuple__229 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__229)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__229);
  __Pyx_GIVEREF(__pyx_tuple__229);

  /* "ultron/tradingday/Calendar.pyx":403
 *     Date(2005, 10, 5),
 *     Date(2005, 10, 6),
 *     Date(2005, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 2),
 *     Date(2006, 1, 3),
 */
  __pyx_tuple__230 = PyTuple_Pack(3, __pyx_int_2005, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__230)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__230);
  __Pyx_GIVEREF(__pyx_tuple__230);

  /* "ultron/tradingday/Calendar.pyx":404
 *     Date(2005, 10, 6),
 *     Date(2005, 10, 7),
 *     Date(2006, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 3),
 *     Date(2006, 1, 26),
 */
  __pyx_tuple__231 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__231)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__231);
  __Pyx_GIVEREF(__pyx_tuple__231);

  /* "ultron/tradingday/Calendar.pyx":405
 *     Date(2005, 10, 7),
 *     Date(2006, 1, 2),
 *     Date(2006, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 26),
 *     Date(2006, 1, 27),
 */
  __pyx_tuple__232 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_1, __pyx_int_3); if (unlikely(!__pyx_tuple__232)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__232);
  __Pyx_GIVEREF(__pyx_tuple__232);

  /* "ultron/tradingday/Calendar.pyx":406
 *     Date(2006, 1, 2),
 *     Date(2006, 1, 3),
 *     Date(2006, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 27),
 *     Date(2006, 1, 30),
 */
  __pyx_tuple__233 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_1, __pyx_int_26); if (unlikely(!__pyx_tuple__233)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__233);
  __Pyx_GIVEREF(__pyx_tuple__233);

  /* "ultron/tradingday/Calendar.pyx":407
 *     Date(2006, 1, 3),
 *     Date(2006, 1, 26),
 *     Date(2006, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 30),
 *     Date(2006, 1, 31),
 */
  __pyx_tuple__234 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_1, __pyx_int_27); if (unlikely(!__pyx_tuple__234)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__234);
  __Pyx_GIVEREF(__pyx_tuple__234);

  /* "ultron/tradingday/Calendar.pyx":408
 *     Date(2006, 1, 26),
 *     Date(2006, 1, 27),
 *     Date(2006, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 31),
 *     Date(2006, 2, 1),
 */
  __pyx_tuple__235 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_1, __pyx_int_30); if (unlikely(!__pyx_tuple__235)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__235);
  __Pyx_GIVEREF(__pyx_tuple__235);

  /* "ultron/tradingday/Calendar.pyx":409
 *     Date(2006, 1, 27),
 *     Date(2006, 1, 30),
 *     Date(2006, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2006, 2, 1),
 *     Date(2006, 2, 2),
 */
  __pyx_tuple__236 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_1, __pyx_int_31); if (unlikely(!__pyx_tuple__236)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__236);
  __Pyx_GIVEREF(__pyx_tuple__236);

  /* "ultron/tradingday/Calendar.pyx":410
 *     Date(2006, 1, 30),
 *     Date(2006, 1, 31),
 *     Date(2006, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2006, 2, 2),
 *     Date(2006, 2, 3),
 */
  __pyx_tuple__237 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_2, __pyx_int_1); if (unlikely(!__pyx_tuple__237)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__237);
  __Pyx_GIVEREF(__pyx_tuple__237);

  /* "ultron/tradingday/Calendar.pyx":411
 *     Date(2006, 1, 31),
 *     Date(2006, 2, 1),
 *     Date(2006, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2006, 2, 3),
 *     Date(2006, 5, 1),
 */
  __pyx_tuple__238 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__238)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__238);
  __Pyx_GIVEREF(__pyx_tuple__238);

  /* "ultron/tradingday/Calendar.pyx":412
 *     Date(2006, 2, 1),
 *     Date(2006, 2, 2),
 *     Date(2006, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 1),
 *     Date(2006, 5, 2),
 */
  __pyx_tuple__239 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__239)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__239);
  __Pyx_GIVEREF(__pyx_tuple__239);

  /* "ultron/tradingday/Calendar.pyx":413
 *     Date(2006, 2, 2),
 *     Date(2006, 2, 3),
 *     Date(2006, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 2),
 *     Date(2006, 5, 3),
 */
  __pyx_tuple__240 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__240)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__240);
  __Pyx_GIVEREF(__pyx_tuple__240);

  /* "ultron/tradingday/Calendar.pyx":414
 *     Date(2006, 2, 3),
 *     Date(2006, 5, 1),
 *     Date(2006, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 3),
 *     Date(2006, 5, 4),
 */
  __pyx_tuple__241 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__241)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__241);
  __Pyx_GIVEREF(__pyx_tuple__241);

  /* "ultron/tradingday/Calendar.pyx":415
 *     Date(2006, 5, 1),
 *     Date(2006, 5, 2),
 *     Date(2006, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 4),
 *     Date(2006, 5, 5),
 */
  __pyx_tuple__242 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__242)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__242);
  __Pyx_GIVEREF(__pyx_tuple__242);

  /* "ultron/tradingday/Calendar.pyx":416
 *     Date(2006, 5, 2),
 *     Date(2006, 5, 3),
 *     Date(2006, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 5),
 *     Date(2006, 10, 2),
 */
  __pyx_tuple__243 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_5, __pyx_int_4); if (unlikely(!__pyx_tuple__243)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__243);
  __Pyx_GIVEREF(__pyx_tuple__243);

  /* "ultron/tradingday/Calendar.pyx":417
 *     Date(2006, 5, 3),
 *     Date(2006, 5, 4),
 *     Date(2006, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 2),
 *     Date(2006, 10, 3),
 */
  __pyx_tuple__244 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_5, __pyx_int_5); if (unlikely(!__pyx_tuple__244)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__244);
  __Pyx_GIVEREF(__pyx_tuple__244);

  /* "ultron/tradingday/Calendar.pyx":418
 *     Date(2006, 5, 4),
 *     Date(2006, 5, 5),
 *     Date(2006, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 3),
 *     Date(2006, 10, 4),
 */
  __pyx_tuple__245 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__245)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__245);
  __Pyx_GIVEREF(__pyx_tuple__245);

  /* "ultron/tradingday/Calendar.pyx":419
 *     Date(2006, 5, 5),
 *     Date(2006, 10, 2),
 *     Date(2006, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 4),
 *     Date(2006, 10, 5),
 */
  __pyx_tuple__246 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__246)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__246);
  __Pyx_GIVEREF(__pyx_tuple__246);

  /* "ultron/tradingday/Calendar.pyx":420
 *     Date(2006, 10, 2),
 *     Date(2006, 10, 3),
 *     Date(2006, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 5),
 *     Date(2006, 10, 6),
 */
  __pyx_tuple__247 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__247)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__247);
  __Pyx_GIVEREF(__pyx_tuple__247);

  /* "ultron/tradingday/Calendar.pyx":421
 *     Date(2006, 10, 3),
 *     Date(2006, 10, 4),
 *     Date(2006, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 6),
 *     Date(2007, 1, 1),
 */
  __pyx_tuple__248 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__248)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__248);
  __Pyx_GIVEREF(__pyx_tuple__248);

  /* "ultron/tradingday/Calendar.pyx":422
 *     Date(2006, 10, 4),
 *     Date(2006, 10, 5),
 *     Date(2006, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2007, 1, 1),
 *     Date(2007, 1, 2),
 */
  __pyx_tuple__249 = PyTuple_Pack(3, __pyx_int_2006, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__249)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__249);
  __Pyx_GIVEREF(__pyx_tuple__249);

  /* "ultron/tradingday/Calendar.pyx":423
 *     Date(2006, 10, 5),
 *     Date(2006, 10, 6),
 *     Date(2007, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2007, 1, 2),
 *     Date(2007, 1, 3),
 */
  __pyx_tuple__250 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__250)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__250);
  __Pyx_GIVEREF(__pyx_tuple__250);

  /* "ultron/tradingday/Calendar.pyx":424
 *     Date(2006, 10, 6),
 *     Date(2007, 1, 1),
 *     Date(2007, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2007, 1, 3),
 *     Date(2007, 2, 19),
 */
  __pyx_tuple__251 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__251)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__251);
  __Pyx_GIVEREF(__pyx_tuple__251);

  /* "ultron/tradingday/Calendar.pyx":425
 *     Date(2007, 1, 1),
 *     Date(2007, 1, 2),
 *     Date(2007, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 19),
 *     Date(2007, 2, 20),
 */
  __pyx_tuple__252 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_1, __pyx_int_3); if (unlikely(!__pyx_tuple__252)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__252);
  __Pyx_GIVEREF(__pyx_tuple__252);

  /* "ultron/tradingday/Calendar.pyx":426
 *     Date(2007, 1, 2),
 *     Date(2007, 1, 3),
 *     Date(2007, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 20),
 *     Date(2007, 2, 21),
 */
  __pyx_tuple__253 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_2, __pyx_int_19); if (unlikely(!__pyx_tuple__253)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__253);
  __Pyx_GIVEREF(__pyx_tuple__253);

  /* "ultron/tradingday/Calendar.pyx":427
 *     Date(2007, 1, 3),
 *     Date(2007, 2, 19),
 *     Date(2007, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 21),
 *     Date(2007, 2, 22),
 */
  __pyx_tuple__254 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_2, __pyx_int_20); if (unlikely(!__pyx_tuple__254)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__254);
  __Pyx_GIVEREF(__pyx_tuple__254);

  /* "ultron/tradingday/Calendar.pyx":428
 *     Date(2007, 2, 19),
 *     Date(2007, 2, 20),
 *     Date(2007, 2, 21),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 22),
 *     Date(2007, 2, 23),
 */
  __pyx_tuple__255 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_2, __pyx_int_21); if (unlikely(!__pyx_tuple__255)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__255);
  __Pyx_GIVEREF(__pyx_tuple__255);

  /* "ultron/tradingday/Calendar.pyx":429
 *     Date(2007, 2, 20),
 *     Date(2007, 2, 21),
 *     Date(2007, 2, 22),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 23),
 *     Date(2007, 5, 1),
 */
  __pyx_tuple__256 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_2, __pyx_int_22); if (unlikely(!__pyx_tuple__256)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__256);
  __Pyx_GIVEREF(__pyx_tuple__256);

  /* "ultron/tradingday/Calendar.pyx":430
 *     Date(2007, 2, 21),
 *     Date(2007, 2, 22),
 *     Date(2007, 2, 23),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 1),
 *     Date(2007, 5, 2),
 */
  __pyx_tuple__257 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_2, __pyx_int_23); if (unlikely(!__pyx_tuple__257)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__257);
  __Pyx_GIVEREF(__pyx_tuple__257);

  /* "ultron/tradingday/Calendar.pyx":431
 *     Date(2007, 2, 22),
 *     Date(2007, 2, 23),
 *     Date(2007, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 2),
 *     Date(2007, 5, 3),
 */
  __pyx_tuple__258 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__258)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__258);
  __Pyx_GIVEREF(__pyx_tuple__258);

  /* "ultron/tradingday/Calendar.pyx":432
 *     Date(2007, 2, 23),
 *     Date(2007, 5, 1),
 *     Date(2007, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 3),
 *     Date(2007, 5, 4),
 */
  __pyx_tuple__259 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__259)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__259);
  __Pyx_GIVEREF(__pyx_tuple__259);

  /* "ultron/tradingday/Calendar.pyx":433
 *     Date(2007, 5, 1),
 *     Date(2007, 5, 2),
 *     Date(2007, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 4),
 *     Date(2007, 5, 7),
 */
  __pyx_tuple__260 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__260)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__260);
  __Pyx_GIVEREF(__pyx_tuple__260);

  /* "ultron/tradingday/Calendar.pyx":434
 *     Date(2007, 5, 2),
 *     Date(2007, 5, 3),
 *     Date(2007, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 7),
 *     Date(2007, 10, 1),
 */
  __pyx_tuple__261 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_5, __pyx_int_4); if (unlikely(!__pyx_tuple__261)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__261);
  __Pyx_GIVEREF(__pyx_tuple__261);

  /* "ultron/tradingday/Calendar.pyx":435
 *     Date(2007, 5, 3),
 *     Date(2007, 5, 4),
 *     Date(2007, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 1),
 *     Date(2007, 10, 2),
 */
  __pyx_tuple__262 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_5, __pyx_int_7); if (unlikely(!__pyx_tuple__262)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__262);
  __Pyx_GIVEREF(__pyx_tuple__262);

  /* "ultron/tradingday/Calendar.pyx":436
 *     Date(2007, 5, 4),
 *     Date(2007, 5, 7),
 *     Date(2007, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 2),
 *     Date(2007, 10, 3),
 */
  __pyx_tuple__263 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__263)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__263);
  __Pyx_GIVEREF(__pyx_tuple__263);

  /* "ultron/tradingday/Calendar.pyx":437
 *     Date(2007, 5, 7),
 *     Date(2007, 10, 1),
 *     Date(2007, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 3),
 *     Date(2007, 10, 4),
 */
  __pyx_tuple__264 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__264)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__264);
  __Pyx_GIVEREF(__pyx_tuple__264);

  /* "ultron/tradingday/Calendar.pyx":438
 *     Date(2007, 10, 1),
 *     Date(2007, 10, 2),
 *     Date(2007, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 4),
 *     Date(2007, 10, 5),
 */
  __pyx_tuple__265 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__265)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__265);
  __Pyx_GIVEREF(__pyx_tuple__265);

  /* "ultron/tradingday/Calendar.pyx":439
 *     Date(2007, 10, 2),
 *     Date(2007, 10, 3),
 *     Date(2007, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 5),
 *     Date(2007, 12, 31),
 */
  __pyx_tuple__266 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__266)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__266);
  __Pyx_GIVEREF(__pyx_tuple__266);

  /* "ultron/tradingday/Calendar.pyx":440
 *     Date(2007, 10, 3),
 *     Date(2007, 10, 4),
 *     Date(2007, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2007, 12, 31),
 *     Date(2008, 1, 1),
 */
  __pyx_tuple__267 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__267)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__267);
  __Pyx_GIVEREF(__pyx_tuple__267);

  /* "ultron/tradingday/Calendar.pyx":441
 *     Date(2007, 10, 4),
 *     Date(2007, 10, 5),
 *     Date(2007, 12, 31),             # <<<<<<<<<<<<<<
 *     Date(2008, 1, 1),
 *     Date(2008, 2, 6),
 */
  __pyx_tuple__268 = PyTuple_Pack(3, __pyx_int_2007, __pyx_int_12, __pyx_int_31); if (unlikely(!__pyx_tuple__268)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__268);
  __Pyx_GIVEREF(__pyx_tuple__268);

  /* "ultron/tradingday/Calendar.pyx":442
 *     Date(2007, 10, 5),
 *     Date(2007, 12, 31),
 *     Date(2008, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 6),
 *     Date(2008, 2, 7),
 */
  __pyx_tuple__269 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__269)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__269);
  __Pyx_GIVEREF(__pyx_tuple__269);

  /* "ultron/tradingday/Calendar.pyx":443
 *     Date(2007, 12, 31),
 *     Date(2008, 1, 1),
 *     Date(2008, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 7),
 *     Date(2008, 2, 8),
 */
  __pyx_tuple__270 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_2, __pyx_int_6); if (unlikely(!__pyx_tuple__270)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__270);
  __Pyx_GIVEREF(__pyx_tuple__270);

  /* "ultron/tradingday/Calendar.pyx":444
 *     Date(2008, 1, 1),
 *     Date(2008, 2, 6),
 *     Date(2008, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 8),
 *     Date(2008, 2, 11),
 */
  __pyx_tuple__271 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__271)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__271);
  __Pyx_GIVEREF(__pyx_tuple__271);

  /* "ultron/tradingday/Calendar.pyx":445
 *     Date(2008, 2, 6),
 *     Date(2008, 2, 7),
 *     Date(2008, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 11),
 *     Date(2008, 2, 12),
 */
  __pyx_tuple__272 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_2, __pyx_int_8); if (unlikely(!__pyx_tuple__272)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__272);
  __Pyx_GIVEREF(__pyx_tuple__272);

  /* "ultron/tradingday/Calendar.pyx":446
 *     Date(2008, 2, 7),
 *     Date(2008, 2, 8),
 *     Date(2008, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 12),
 *     Date(2008, 4, 4),
 */
  __pyx_tuple__273 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__273)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__273);
  __Pyx_GIVEREF(__pyx_tuple__273);

  /* "ultron/tradingday/Calendar.pyx":447
 *     Date(2008, 2, 8),
 *     Date(2008, 2, 11),
 *     Date(2008, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2008, 4, 4),
 *     Date(2008, 5, 1),
 */
  __pyx_tuple__274 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_2, __pyx_int_12); if (unlikely(!__pyx_tuple__274)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__274);
  __Pyx_GIVEREF(__pyx_tuple__274);

  /* "ultron/tradingday/Calendar.pyx":448
 *     Date(2008, 2, 11),
 *     Date(2008, 2, 12),
 *     Date(2008, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2008, 5, 1),
 *     Date(2008, 5, 2),
 */
  __pyx_tuple__275 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_4, __pyx_int_4); if (unlikely(!__pyx_tuple__275)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__275);
  __Pyx_GIVEREF(__pyx_tuple__275);

  /* "ultron/tradingday/Calendar.pyx":449
 *     Date(2008, 2, 12),
 *     Date(2008, 4, 4),
 *     Date(2008, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2008, 5, 2),
 *     Date(2008, 6, 9),
 */
  __pyx_tuple__276 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__276)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__276);
  __Pyx_GIVEREF(__pyx_tuple__276);

  /* "ultron/tradingday/Calendar.pyx":450
 *     Date(2008, 4, 4),
 *     Date(2008, 5, 1),
 *     Date(2008, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2008, 6, 9),
 *     Date(2008, 9, 15),
 */
  __pyx_tuple__277 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__277)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__277);
  __Pyx_GIVEREF(__pyx_tuple__277);

  /* "ultron/tradingday/Calendar.pyx":451
 *     Date(2008, 5, 1),
 *     Date(2008, 5, 2),
 *     Date(2008, 6, 9),             # <<<<<<<<<<<<<<
 *     Date(2008, 9, 15),
 *     Date(2008, 9, 29),
 */
  __pyx_tuple__278 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_6, __pyx_int_9); if (unlikely(!__pyx_tuple__278)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__278);
  __Pyx_GIVEREF(__pyx_tuple__278);

  /* "ultron/tradingday/Calendar.pyx":452
 *     Date(2008, 5, 2),
 *     Date(2008, 6, 9),
 *     Date(2008, 9, 15),             # <<<<<<<<<<<<<<
 *     Date(2008, 9, 29),
 *     Date(2008, 9, 30),
 */
  __pyx_tuple__279 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_9, __pyx_int_15); if (unlikely(!__pyx_tuple__279)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__279);
  __Pyx_GIVEREF(__pyx_tuple__279);

  /* "ultron/tradingday/Calendar.pyx":453
 *     Date(2008, 6, 9),
 *     Date(2008, 9, 15),
 *     Date(2008, 9, 29),             # <<<<<<<<<<<<<<
 *     Date(2008, 9, 30),
 *     Date(2008, 10, 1),
 */
  __pyx_tuple__280 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_9, __pyx_int_29); if (unlikely(!__pyx_tuple__280)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__280);
  __Pyx_GIVEREF(__pyx_tuple__280);

  /* "ultron/tradingday/Calendar.pyx":454
 *     Date(2008, 9, 15),
 *     Date(2008, 9, 29),
 *     Date(2008, 9, 30),             # <<<<<<<<<<<<<<
 *     Date(2008, 10, 1),
 *     Date(2008, 10, 2),
 */
  __pyx_tuple__281 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_9, __pyx_int_30); if (unlikely(!__pyx_tuple__281)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__281);
  __Pyx_GIVEREF(__pyx_tuple__281);

  /* "ultron/tradingday/Calendar.pyx":455
 *     Date(2008, 9, 29),
 *     Date(2008, 9, 30),
 *     Date(2008, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2008, 10, 2),
 *     Date(2008, 10, 3),
 */
  __pyx_tuple__282 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__282)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__282);
  __Pyx_GIVEREF(__pyx_tuple__282);

  /* "ultron/tradingday/Calendar.pyx":456
 *     Date(2008, 9, 30),
 *     Date(2008, 10, 1),
 *     Date(2008, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2008, 10, 3),
 *     Date(2009, 1, 1),
 */
  __pyx_tuple__283 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__283)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__283);
  __Pyx_GIVEREF(__pyx_tuple__283);

  /* "ultron/tradingday/Calendar.pyx":457
 *     Date(2008, 10, 1),
 *     Date(2008, 10, 2),
 *     Date(2008, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 1),
 *     Date(2009, 1, 2),
 */
  __pyx_tuple__284 = PyTuple_Pack(3, __pyx_int_2008, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__284)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__284);
  __Pyx_GIVEREF(__pyx_tuple__284);

  /* "ultron/tradingday/Calendar.pyx":458
 *     Date(2008, 10, 2),
 *     Date(2008, 10, 3),
 *     Date(2009, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 2),
 *     Date(2009, 1, 26),
 */
  __pyx_tuple__285 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__285)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__285);
  __Pyx_GIVEREF(__pyx_tuple__285);

  /* "ultron/tradingday/Calendar.pyx":459
 *     Date(2008, 10, 3),
 *     Date(2009, 1, 1),
 *     Date(2009, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 26),
 *     Date(2009, 1, 27),
 */
  __pyx_tuple__286 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__286)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__286);
  __Pyx_GIVEREF(__pyx_tuple__286);

  /* "ultron/tradingday/Calendar.pyx":460
 *     Date(2009, 1, 1),
 *     Date(2009, 1, 2),
 *     Date(2009, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 27),
 *     Date(2009, 1, 28),
 */
  __pyx_tuple__287 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_1, __pyx_int_26); if (unlikely(!__pyx_tuple__287)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__287);
  __Pyx_GIVEREF(__pyx_tuple__287);

  /* "ultron/tradingday/Calendar.pyx":461
 *     Date(2009, 1, 2),
 *     Date(2009, 1, 26),
 *     Date(2009, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 28),
 *     Date(2009, 1, 29),
 */
  __pyx_tuple__288 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_1, __pyx_int_27); if (unlikely(!__pyx_tuple__288)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__288);
  __Pyx_GIVEREF(__pyx_tuple__288);

  /* "ultron/tradingday/Calendar.pyx":462
 *     Date(2009, 1, 26),
 *     Date(2009, 1, 27),
 *     Date(2009, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 29),
 *     Date(2009, 1, 30),
 */
  __pyx_tuple__289 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_1, __pyx_int_28); if (unlikely(!__pyx_tuple__289)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__289);
  __Pyx_GIVEREF(__pyx_tuple__289);

  /* "ultron/tradingday/Calendar.pyx":463
 *     Date(2009, 1, 27),
 *     Date(2009, 1, 28),
 *     Date(2009, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 30),
 *     Date(2009, 4, 6),
 */
  __pyx_tuple__290 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_1, __pyx_int_29); if (unlikely(!__pyx_tuple__290)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__290);
  __Pyx_GIVEREF(__pyx_tuple__290);

  /* "ultron/tradingday/Calendar.pyx":464
 *     Date(2009, 1, 28),
 *     Date(2009, 1, 29),
 *     Date(2009, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2009, 4, 6),
 *     Date(2009, 5, 1),
 */
  __pyx_tuple__291 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_1, __pyx_int_30); if (unlikely(!__pyx_tuple__291)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__291);
  __Pyx_GIVEREF(__pyx_tuple__291);

  /* "ultron/tradingday/Calendar.pyx":465
 *     Date(2009, 1, 29),
 *     Date(2009, 1, 30),
 *     Date(2009, 4, 6),             # <<<<<<<<<<<<<<
 *     Date(2009, 5, 1),
 *     Date(2009, 5, 28),
 */
  __pyx_tuple__292 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_4, __pyx_int_6); if (unlikely(!__pyx_tuple__292)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__292);
  __Pyx_GIVEREF(__pyx_tuple__292);

  /* "ultron/tradingday/Calendar.pyx":466
 *     Date(2009, 1, 30),
 *     Date(2009, 4, 6),
 *     Date(2009, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2009, 5, 28),
 *     Date(2009, 5, 29),
 */
  __pyx_tuple__293 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__293)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__293);
  __Pyx_GIVEREF(__pyx_tuple__293);

  /* "ultron/tradingday/Calendar.pyx":467
 *     Date(2009, 4, 6),
 *     Date(2009, 5, 1),
 *     Date(2009, 5, 28),             # <<<<<<<<<<<<<<
 *     Date(2009, 5, 29),
 *     Date(2009, 10, 1),
 */
  __pyx_tuple__294 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_5, __pyx_int_28); if (unlikely(!__pyx_tuple__294)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__294);
  __Pyx_GIVEREF(__pyx_tuple__294);

  /* "ultron/tradingday/Calendar.pyx":468
 *     Date(2009, 5, 1),
 *     Date(2009, 5, 28),
 *     Date(2009, 5, 29),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 1),
 *     Date(2009, 10, 2),
 */
  __pyx_tuple__295 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_5, __pyx_int_29); if (unlikely(!__pyx_tuple__295)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__295);
  __Pyx_GIVEREF(__pyx_tuple__295);

  /* "ultron/tradingday/Calendar.pyx":469
 *     Date(2009, 5, 28),
 *     Date(2009, 5, 29),
 *     Date(2009, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 2),
 *     Date(2009, 10, 5),
 */
  __pyx_tuple__296 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__296)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__296);
  __Pyx_GIVEREF(__pyx_tuple__296);

  /* "ultron/tradingday/Calendar.pyx":470
 *     Date(2009, 5, 29),
 *     Date(2009, 10, 1),
 *     Date(2009, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 5),
 *     Date(2009, 10, 6),
 */
  __pyx_tuple__297 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__297)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__297);
  __Pyx_GIVEREF(__pyx_tuple__297);

  /* "ultron/tradingday/Calendar.pyx":471
 *     Date(2009, 10, 1),
 *     Date(2009, 10, 2),
 *     Date(2009, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 6),
 *     Date(2009, 10, 7),
 */
  __pyx_tuple__298 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__298)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__298);
  __Pyx_GIVEREF(__pyx_tuple__298);

  /* "ultron/tradingday/Calendar.pyx":472
 *     Date(2009, 10, 2),
 *     Date(2009, 10, 5),
 *     Date(2009, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 7),
 *     Date(2009, 10, 8),
 */
  __pyx_tuple__299 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__299)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__299);
  __Pyx_GIVEREF(__pyx_tuple__299);

  /* "ultron/tradingday/Calendar.pyx":473
 *     Date(2009, 10, 5),
 *     Date(2009, 10, 6),
 *     Date(2009, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 8),
 *     Date(2010, 1, 1),
 */
  __pyx_tuple__300 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__300)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__300);
  __Pyx_GIVEREF(__pyx_tuple__300);

  /* "ultron/tradingday/Calendar.pyx":474
 *     Date(2009, 10, 6),
 *     Date(2009, 10, 7),
 *     Date(2009, 10, 8),             # <<<<<<<<<<<<<<
 *     Date(2010, 1, 1),
 *     Date(2010, 2, 15),
 */
  __pyx_tuple__301 = PyTuple_Pack(3, __pyx_int_2009, __pyx_int_10, __pyx_int_8); if (unlikely(!__pyx_tuple__301)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__301);
  __Pyx_GIVEREF(__pyx_tuple__301);

  /* "ultron/tradingday/Calendar.pyx":475
 *     Date(2009, 10, 7),
 *     Date(2009, 10, 8),
 *     Date(2010, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 15),
 *     Date(2010, 2, 16),
 */
  __pyx_tuple__302 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__302)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__302);
  __Pyx_GIVEREF(__pyx_tuple__302);

  /* "ultron/tradingday/Calendar.pyx":476
 *     Date(2009, 10, 8),
 *     Date(2010, 1, 1),
 *     Date(2010, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 16),
 *     Date(2010, 2, 17),
 */
  __pyx_tuple__303 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_2, __pyx_int_15); if (unlikely(!__pyx_tuple__303)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__303);
  __Pyx_GIVEREF(__pyx_tuple__303);

  /* "ultron/tradingday/Calendar.pyx":477
 *     Date(2010, 1, 1),
 *     Date(2010, 2, 15),
 *     Date(2010, 2, 16),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 17),
 *     Date(2010, 2, 18),
 */
  __pyx_tuple__304 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_2, __pyx_int_16); if (unlikely(!__pyx_tuple__304)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__304);
  __Pyx_GIVEREF(__pyx_tuple__304);

  /* "ultron/tradingday/Calendar.pyx":478
 *     Date(2010, 2, 15),
 *     Date(2010, 2, 16),
 *     Date(2010, 2, 17),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 18),
 *     Date(2010, 2, 19),
 */
  __pyx_tuple__305 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_2, __pyx_int_17); if (unlikely(!__pyx_tuple__305)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__305);
  __Pyx_GIVEREF(__pyx_tuple__305);

  /* "ultron/tradingday/Calendar.pyx":479
 *     Date(2010, 2, 16),
 *     Date(2010, 2, 17),
 *     Date(2010, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 19),
 *     Date(2010, 4, 5),
 */
  __pyx_tuple__306 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_2, __pyx_int_18); if (unlikely(!__pyx_tuple__306)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__306);
  __Pyx_GIVEREF(__pyx_tuple__306);

  /* "ultron/tradingday/Calendar.pyx":480
 *     Date(2010, 2, 17),
 *     Date(2010, 2, 18),
 *     Date(2010, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2010, 4, 5),
 *     Date(2010, 5, 3),
 */
  __pyx_tuple__307 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_2, __pyx_int_19); if (unlikely(!__pyx_tuple__307)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__307);
  __Pyx_GIVEREF(__pyx_tuple__307);

  /* "ultron/tradingday/Calendar.pyx":481
 *     Date(2010, 2, 18),
 *     Date(2010, 2, 19),
 *     Date(2010, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2010, 5, 3),
 *     Date(2010, 6, 14),
 */
  __pyx_tuple__308 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_4, __pyx_int_5); if (unlikely(!__pyx_tuple__308)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__308);
  __Pyx_GIVEREF(__pyx_tuple__308);

  /* "ultron/tradingday/Calendar.pyx":482
 *     Date(2010, 2, 19),
 *     Date(2010, 4, 5),
 *     Date(2010, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2010, 6, 14),
 *     Date(2010, 6, 15),
 */
  __pyx_tuple__309 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__309)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__309);
  __Pyx_GIVEREF(__pyx_tuple__309);

  /* "ultron/tradingday/Calendar.pyx":483
 *     Date(2010, 4, 5),
 *     Date(2010, 5, 3),
 *     Date(2010, 6, 14),             # <<<<<<<<<<<<<<
 *     Date(2010, 6, 15),
 *     Date(2010, 6, 16),
 */
  __pyx_tuple__310 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_6, __pyx_int_14); if (unlikely(!__pyx_tuple__310)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__310);
  __Pyx_GIVEREF(__pyx_tuple__310);

  /* "ultron/tradingday/Calendar.pyx":484
 *     Date(2010, 5, 3),
 *     Date(2010, 6, 14),
 *     Date(2010, 6, 15),             # <<<<<<<<<<<<<<
 *     Date(2010, 6, 16),
 *     Date(2010, 9, 22),
 */
  __pyx_tuple__311 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_6, __pyx_int_15); if (unlikely(!__pyx_tuple__311)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__311);
  __Pyx_GIVEREF(__pyx_tuple__311);

  /* "ultron/tradingday/Calendar.pyx":485
 *     Date(2010, 6, 14),
 *     Date(2010, 6, 15),
 *     Date(2010, 6, 16),             # <<<<<<<<<<<<<<
 *     Date(2010, 9, 22),
 *     Date(2010, 9, 23),
 */
  __pyx_tuple__312 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_6, __pyx_int_16); if (unlikely(!__pyx_tuple__312)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__312);
  __Pyx_GIVEREF(__pyx_tuple__312);

  /* "ultron/tradingday/Calendar.pyx":486
 *     Date(2010, 6, 15),
 *     Date(2010, 6, 16),
 *     Date(2010, 9, 22),             # <<<<<<<<<<<<<<
 *     Date(2010, 9, 23),
 *     Date(2010, 9, 24),
 */
  __pyx_tuple__313 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_9, __pyx_int_22); if (unlikely(!__pyx_tuple__313)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__313);
  __Pyx_GIVEREF(__pyx_tuple__313);

  /* "ultron/tradingday/Calendar.pyx":487
 *     Date(2010, 6, 16),
 *     Date(2010, 9, 22),
 *     Date(2010, 9, 23),             # <<<<<<<<<<<<<<
 *     Date(2010, 9, 24),
 *     Date(2010, 10, 1),
 */
  __pyx_tuple__314 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_9, __pyx_int_23); if (unlikely(!__pyx_tuple__314)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__314);
  __Pyx_GIVEREF(__pyx_tuple__314);

  /* "ultron/tradingday/Calendar.pyx":488
 *     Date(2010, 9, 22),
 *     Date(2010, 9, 23),
 *     Date(2010, 9, 24),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 1),
 *     Date(2010, 10, 4),
 */
  __pyx_tuple__315 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_9, __pyx_int_24); if (unlikely(!__pyx_tuple__315)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__315);
  __Pyx_GIVEREF(__pyx_tuple__315);

  /* "ultron/tradingday/Calendar.pyx":489
 *     Date(2010, 9, 23),
 *     Date(2010, 9, 24),
 *     Date(2010, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 4),
 *     Date(2010, 10, 5),
 */
  __pyx_tuple__316 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__316)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__316);
  __Pyx_GIVEREF(__pyx_tuple__316);

  /* "ultron/tradingday/Calendar.pyx":490
 *     Date(2010, 9, 24),
 *     Date(2010, 10, 1),
 *     Date(2010, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 5),
 *     Date(2010, 10, 6),
 */
  __pyx_tuple__317 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__317)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__317);
  __Pyx_GIVEREF(__pyx_tuple__317);

  /* "ultron/tradingday/Calendar.pyx":491
 *     Date(2010, 10, 1),
 *     Date(2010, 10, 4),
 *     Date(2010, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 6),
 *     Date(2010, 10, 7),
 */
  __pyx_tuple__318 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__318)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__318);
  __Pyx_GIVEREF(__pyx_tuple__318);

  /* "ultron/tradingday/Calendar.pyx":492
 *     Date(2010, 10, 4),
 *     Date(2010, 10, 5),
 *     Date(2010, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 7),
 *     Date(2011, 1, 3),
 */
  __pyx_tuple__319 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__319)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__319);
  __Pyx_GIVEREF(__pyx_tuple__319);

  /* "ultron/tradingday/Calendar.pyx":493
 *     Date(2010, 10, 5),
 *     Date(2010, 10, 6),
 *     Date(2010, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2011, 1, 3),
 *     Date(2011, 2, 2),
 */
  __pyx_tuple__320 = PyTuple_Pack(3, __pyx_int_2010, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__320)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__320);
  __Pyx_GIVEREF(__pyx_tuple__320);

  /* "ultron/tradingday/Calendar.pyx":494
 *     Date(2010, 10, 6),
 *     Date(2010, 10, 7),
 *     Date(2011, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 2),
 *     Date(2011, 2, 3),
 */
  __pyx_tuple__321 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_1, __pyx_int_3); if (unlikely(!__pyx_tuple__321)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__321);
  __Pyx_GIVEREF(__pyx_tuple__321);

  /* "ultron/tradingday/Calendar.pyx":495
 *     Date(2010, 10, 7),
 *     Date(2011, 1, 3),
 *     Date(2011, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 3),
 *     Date(2011, 2, 4),
 */
  __pyx_tuple__322 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__322)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__322);
  __Pyx_GIVEREF(__pyx_tuple__322);

  /* "ultron/tradingday/Calendar.pyx":496
 *     Date(2011, 1, 3),
 *     Date(2011, 2, 2),
 *     Date(2011, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 4),
 *     Date(2011, 2, 7),
 */
  __pyx_tuple__323 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__323)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__323);
  __Pyx_GIVEREF(__pyx_tuple__323);

  /* "ultron/tradingday/Calendar.pyx":497
 *     Date(2011, 2, 2),
 *     Date(2011, 2, 3),
 *     Date(2011, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 7),
 *     Date(2011, 2, 8),
 */
  __pyx_tuple__324 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__324)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__324);
  __Pyx_GIVEREF(__pyx_tuple__324);

  /* "ultron/tradingday/Calendar.pyx":498
 *     Date(2011, 2, 3),
 *     Date(2011, 2, 4),
 *     Date(2011, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 8),
 *     Date(2011, 4, 4),
 */
  __pyx_tuple__325 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__325)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__325);
  __Pyx_GIVEREF(__pyx_tuple__325);

  /* "ultron/tradingday/Calendar.pyx":499
 *     Date(2011, 2, 4),
 *     Date(2011, 2, 7),
 *     Date(2011, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2011, 4, 4),
 *     Date(2011, 4, 5),
 */
  __pyx_tuple__326 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_2, __pyx_int_8); if (unlikely(!__pyx_tuple__326)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__326);
  __Pyx_GIVEREF(__pyx_tuple__326);

  /* "ultron/tradingday/Calendar.pyx":500
 *     Date(2011, 2, 7),
 *     Date(2011, 2, 8),
 *     Date(2011, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2011, 4, 5),
 *     Date(2011, 5, 2),
 */
  __pyx_tuple__327 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_4, __pyx_int_4); if (unlikely(!__pyx_tuple__327)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__327);
  __Pyx_GIVEREF(__pyx_tuple__327);

  /* "ultron/tradingday/Calendar.pyx":501
 *     Date(2011, 2, 8),
 *     Date(2011, 4, 4),
 *     Date(2011, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2011, 5, 2),
 *     Date(2011, 6, 6),
 */
  __pyx_tuple__328 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_4, __pyx_int_5); if (unlikely(!__pyx_tuple__328)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__328);
  __Pyx_GIVEREF(__pyx_tuple__328);

  /* "ultron/tradingday/Calendar.pyx":502
 *     Date(2011, 4, 4),
 *     Date(2011, 4, 5),
 *     Date(2011, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2011, 6, 6),
 *     Date(2011, 9, 12),
 */
  __pyx_tuple__329 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__329)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__329);
  __Pyx_GIVEREF(__pyx_tuple__329);

  /* "ultron/tradingday/Calendar.pyx":503
 *     Date(2011, 4, 5),
 *     Date(2011, 5, 2),
 *     Date(2011, 6, 6),             # <<<<<<<<<<<<<<
 *     Date(2011, 9, 12),
 *     Date(2011, 10, 3),
 */
  __pyx_tuple__330 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_6, __pyx_int_6); if (unlikely(!__pyx_tuple__330)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__330);
  __Pyx_GIVEREF(__pyx_tuple__330);

  /* "ultron/tradingday/Calendar.pyx":504
 *     Date(2011, 5, 2),
 *     Date(2011, 6, 6),
 *     Date(2011, 9, 12),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 3),
 *     Date(2011, 10, 4),
 */
  __pyx_tuple__331 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_9, __pyx_int_12); if (unlikely(!__pyx_tuple__331)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__331);
  __Pyx_GIVEREF(__pyx_tuple__331);

  /* "ultron/tradingday/Calendar.pyx":505
 *     Date(2011, 6, 6),
 *     Date(2011, 9, 12),
 *     Date(2011, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 4),
 *     Date(2011, 10, 5),
 */
  __pyx_tuple__332 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__332)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__332);
  __Pyx_GIVEREF(__pyx_tuple__332);

  /* "ultron/tradingday/Calendar.pyx":506
 *     Date(2011, 9, 12),
 *     Date(2011, 10, 3),
 *     Date(2011, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 5),
 *     Date(2011, 10, 6),
 */
  __pyx_tuple__333 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__333)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__333);
  __Pyx_GIVEREF(__pyx_tuple__333);

  /* "ultron/tradingday/Calendar.pyx":507
 *     Date(2011, 10, 3),
 *     Date(2011, 10, 4),
 *     Date(2011, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 6),
 *     Date(2011, 10, 7),
 */
  __pyx_tuple__334 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__334)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__334);
  __Pyx_GIVEREF(__pyx_tuple__334);

  /* "ultron/tradingday/Calendar.pyx":508
 *     Date(2011, 10, 4),
 *     Date(2011, 10, 5),
 *     Date(2011, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 7),
 *     Date(2012, 1, 2),
 */
  __pyx_tuple__335 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__335)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__335);
  __Pyx_GIVEREF(__pyx_tuple__335);

  /* "ultron/tradingday/Calendar.pyx":509
 *     Date(2011, 10, 5),
 *     Date(2011, 10, 6),
 *     Date(2011, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 2),
 *     Date(2012, 1, 3),
 */
  __pyx_tuple__336 = PyTuple_Pack(3, __pyx_int_2011, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__336)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__336);
  __Pyx_GIVEREF(__pyx_tuple__336);

  /* "ultron/tradingday/Calendar.pyx":510
 *     Date(2011, 10, 6),
 *     Date(2011, 10, 7),
 *     Date(2012, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 3),
 *     Date(2012, 1, 23),
 */
  __pyx_tuple__337 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__337)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__337);
  __Pyx_GIVEREF(__pyx_tuple__337);

  /* "ultron/tradingday/Calendar.pyx":511
 *     Date(2011, 10, 7),
 *     Date(2012, 1, 2),
 *     Date(2012, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 23),
 *     Date(2012, 1, 24),
 */
  __pyx_tuple__338 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_1, __pyx_int_3); if (unlikely(!__pyx_tuple__338)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__338);
  __Pyx_GIVEREF(__pyx_tuple__338);

  /* "ultron/tradingday/Calendar.pyx":512
 *     Date(2012, 1, 2),
 *     Date(2012, 1, 3),
 *     Date(2012, 1, 23),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 24),
 *     Date(2012, 1, 25),
 */
  __pyx_tuple__339 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_1, __pyx_int_23); if (unlikely(!__pyx_tuple__339)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__339);
  __Pyx_GIVEREF(__pyx_tuple__339);

  /* "ultron/tradingday/Calendar.pyx":513
 *     Date(2012, 1, 3),
 *     Date(2012, 1, 23),
 *     Date(2012, 1, 24),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 25),
 *     Date(2012, 1, 26),
 */
  __pyx_tuple__340 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_1, __pyx_int_24); if (unlikely(!__pyx_tuple__340)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__340);
  __Pyx_GIVEREF(__pyx_tuple__340);

  /* "ultron/tradingday/Calendar.pyx":514
 *     Date(2012, 1, 23),
 *     Date(2012, 1, 24),
 *     Date(2012, 1, 25),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 26),
 *     Date(2012, 1, 27),
 */
  __pyx_tuple__341 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_1, __pyx_int_25); if (unlikely(!__pyx_tuple__341)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__341);
  __Pyx_GIVEREF(__pyx_tuple__341);

  /* "ultron/tradingday/Calendar.pyx":515
 *     Date(2012, 1, 24),
 *     Date(2012, 1, 25),
 *     Date(2012, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 27),
 *     Date(2012, 4, 2),
 */
  __pyx_tuple__342 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_1, __pyx_int_26); if (unlikely(!__pyx_tuple__342)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__342);
  __Pyx_GIVEREF(__pyx_tuple__342);

  /* "ultron/tradingday/Calendar.pyx":516
 *     Date(2012, 1, 25),
 *     Date(2012, 1, 26),
 *     Date(2012, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2012, 4, 2),
 *     Date(2012, 4, 3),
 */
  __pyx_tuple__343 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_1, __pyx_int_27); if (unlikely(!__pyx_tuple__343)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__343);
  __Pyx_GIVEREF(__pyx_tuple__343);

  /* "ultron/tradingday/Calendar.pyx":517
 *     Date(2012, 1, 26),
 *     Date(2012, 1, 27),
 *     Date(2012, 4, 2),             # <<<<<<<<<<<<<<
 *     Date(2012, 4, 3),
 *     Date(2012, 4, 4),
 */
  __pyx_tuple__344 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_4, __pyx_int_2); if (unlikely(!__pyx_tuple__344)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__344);
  __Pyx_GIVEREF(__pyx_tuple__344);

  /* "ultron/tradingday/Calendar.pyx":518
 *     Date(2012, 1, 27),
 *     Date(2012, 4, 2),
 *     Date(2012, 4, 3),             # <<<<<<<<<<<<<<
 *     Date(2012, 4, 4),
 *     Date(2012, 4, 30),
 */
  __pyx_tuple__345 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_4, __pyx_int_3); if (unlikely(!__pyx_tuple__345)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__345);
  __Pyx_GIVEREF(__pyx_tuple__345);

  /* "ultron/tradingday/Calendar.pyx":519
 *     Date(2012, 4, 2),
 *     Date(2012, 4, 3),
 *     Date(2012, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2012, 4, 30),
 *     Date(2012, 5, 1),
 */
  __pyx_tuple__346 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_4, __pyx_int_4); if (unlikely(!__pyx_tuple__346)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__346);
  __Pyx_GIVEREF(__pyx_tuple__346);

  /* "ultron/tradingday/Calendar.pyx":520
 *     Date(2012, 4, 3),
 *     Date(2012, 4, 4),
 *     Date(2012, 4, 30),             # <<<<<<<<<<<<<<
 *     Date(2012, 5, 1),
 *     Date(2012, 6, 22),
 */
  __pyx_tuple__347 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_4, __pyx_int_30); if (unlikely(!__pyx_tuple__347)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__347);
  __Pyx_GIVEREF(__pyx_tuple__347);

  /* "ultron/tradingday/Calendar.pyx":521
 *     Date(2012, 4, 4),
 *     Date(2012, 4, 30),
 *     Date(2012, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2012, 6, 22),
 *     Date(2012, 10, 1),
 */
  __pyx_tuple__348 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__348)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__348);
  __Pyx_GIVEREF(__pyx_tuple__348);

  /* "ultron/tradingday/Calendar.pyx":522
 *     Date(2012, 4, 30),
 *     Date(2012, 5, 1),
 *     Date(2012, 6, 22),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 1),
 *     Date(2012, 10, 2),
 */
  __pyx_tuple__349 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_6, __pyx_int_22); if (unlikely(!__pyx_tuple__349)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__349);
  __Pyx_GIVEREF(__pyx_tuple__349);

  /* "ultron/tradingday/Calendar.pyx":523
 *     Date(2012, 5, 1),
 *     Date(2012, 6, 22),
 *     Date(2012, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 2),
 *     Date(2012, 10, 3),
 */
  __pyx_tuple__350 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__350)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__350);
  __Pyx_GIVEREF(__pyx_tuple__350);

  /* "ultron/tradingday/Calendar.pyx":524
 *     Date(2012, 6, 22),
 *     Date(2012, 10, 1),
 *     Date(2012, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 3),
 *     Date(2012, 10, 4),
 */
  __pyx_tuple__351 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__351)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__351);
  __Pyx_GIVEREF(__pyx_tuple__351);

  /* "ultron/tradingday/Calendar.pyx":525
 *     Date(2012, 10, 1),
 *     Date(2012, 10, 2),
 *     Date(2012, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 4),
 *     Date(2012, 10, 5),
 */
  __pyx_tuple__352 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__352)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__352);
  __Pyx_GIVEREF(__pyx_tuple__352);

  /* "ultron/tradingday/Calendar.pyx":526
 *     Date(2012, 10, 2),
 *     Date(2012, 10, 3),
 *     Date(2012, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 5),
 *     Date(2013, 1, 1),
 */
  __pyx_tuple__353 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__353)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__353);
  __Pyx_GIVEREF(__pyx_tuple__353);

  /* "ultron/tradingday/Calendar.pyx":527
 *     Date(2012, 10, 3),
 *     Date(2012, 10, 4),
 *     Date(2012, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2013, 1, 1),
 *     Date(2013, 1, 2),
 */
  __pyx_tuple__354 = PyTuple_Pack(3, __pyx_int_2012, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__354)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__354);
  __Pyx_GIVEREF(__pyx_tuple__354);

  /* "ultron/tradingday/Calendar.pyx":528
 *     Date(2012, 10, 4),
 *     Date(2012, 10, 5),
 *     Date(2013, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2013, 1, 2),
 *     Date(2013, 1, 3),
 */
  __pyx_tuple__355 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__355)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__355);
  __Pyx_GIVEREF(__pyx_tuple__355);

  /* "ultron/tradingday/Calendar.pyx":529
 *     Date(2012, 10, 5),
 *     Date(2013, 1, 1),
 *     Date(2013, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2013, 1, 3),
 *     Date(2013, 2, 11),
 */
  __pyx_tuple__356 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__356)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__356);
  __Pyx_GIVEREF(__pyx_tuple__356);

  /* "ultron/tradingday/Calendar.pyx":530
 *     Date(2013, 1, 1),
 *     Date(2013, 1, 2),
 *     Date(2013, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 11),
 *     Date(2013, 2, 12),
 */
  __pyx_tuple__357 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_1, __pyx_int_3); if (unlikely(!__pyx_tuple__357)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__357);
  __Pyx_GIVEREF(__pyx_tuple__357);

  /* "ultron/tradingday/Calendar.pyx":531
 *     Date(2013, 1, 2),
 *     Date(2013, 1, 3),
 *     Date(2013, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 12),
 *     Date(2013, 2, 13),
 */
  __pyx_tuple__358 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__358)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__358);
  __Pyx_GIVEREF(__pyx_tuple__358);

  /* "ultron/tradingday/Calendar.pyx":532
 *     Date(2013, 1, 3),
 *     Date(2013, 2, 11),
 *     Date(2013, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 13),
 *     Date(2013, 2, 14),
 */
  __pyx_tuple__359 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_2, __pyx_int_12); if (unlikely(!__pyx_tuple__359)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__359);
  __Pyx_GIVEREF(__pyx_tuple__359);

  /* "ultron/tradingday/Calendar.pyx":533
 *     Date(2013, 2, 11),
 *     Date(2013, 2, 12),
 *     Date(2013, 2, 13),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 14),
 *     Date(2013, 2, 15),
 */
  __pyx_tuple__360 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_2, __pyx_int_13); if (unlikely(!__pyx_tuple__360)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__360);
  __Pyx_GIVEREF(__pyx_tuple__360);

  /* "ultron/tradingday/Calendar.pyx":534
 *     Date(2013, 2, 12),
 *     Date(2013, 2, 13),
 *     Date(2013, 2, 14),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 15),
 *     Date(2013, 4, 4),
 */
  __pyx_tuple__361 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_2, __pyx_int_14); if (unlikely(!__pyx_tuple__361)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__361);
  __Pyx_GIVEREF(__pyx_tuple__361);

  /* "ultron/tradingday/Calendar.pyx":535
 *     Date(2013, 2, 13),
 *     Date(2013, 2, 14),
 *     Date(2013, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2013, 4, 4),
 *     Date(2013, 4, 5),
 */
  __pyx_tuple__362 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_2, __pyx_int_15); if (unlikely(!__pyx_tuple__362)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__362);
  __Pyx_GIVEREF(__pyx_tuple__362);

  /* "ultron/tradingday/Calendar.pyx":536
 *     Date(2013, 2, 14),
 *     Date(2013, 2, 15),
 *     Date(2013, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2013, 4, 5),
 *     Date(2013, 4, 29),
 */
  __pyx_tuple__363 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_4, __pyx_int_4); if (unlikely(!__pyx_tuple__363)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__363);
  __Pyx_GIVEREF(__pyx_tuple__363);

  /* "ultron/tradingday/Calendar.pyx":537
 *     Date(2013, 2, 15),
 *     Date(2013, 4, 4),
 *     Date(2013, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2013, 4, 29),
 *     Date(2013, 4, 30),
 */
  __pyx_tuple__364 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_4, __pyx_int_5); if (unlikely(!__pyx_tuple__364)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__364);
  __Pyx_GIVEREF(__pyx_tuple__364);

  /* "ultron/tradingday/Calendar.pyx":538
 *     Date(2013, 4, 4),
 *     Date(2013, 4, 5),
 *     Date(2013, 4, 29),             # <<<<<<<<<<<<<<
 *     Date(2013, 4, 30),
 *     Date(2013, 5, 1),
 */
  __pyx_tuple__365 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_4, __pyx_int_29); if (unlikely(!__pyx_tuple__365)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__365);
  __Pyx_GIVEREF(__pyx_tuple__365);

  /* "ultron/tradingday/Calendar.pyx":539
 *     Date(2013, 4, 5),
 *     Date(2013, 4, 29),
 *     Date(2013, 4, 30),             # <<<<<<<<<<<<<<
 *     Date(2013, 5, 1),
 *     Date(2013, 6, 10),
 */
  __pyx_tuple__366 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_4, __pyx_int_30); if (unlikely(!__pyx_tuple__366)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__366);
  __Pyx_GIVEREF(__pyx_tuple__366);

  /* "ultron/tradingday/Calendar.pyx":540
 *     Date(2013, 4, 29),
 *     Date(2013, 4, 30),
 *     Date(2013, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2013, 6, 10),
 *     Date(2013, 6, 11),
 */
  __pyx_tuple__367 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__367)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__367);
  __Pyx_GIVEREF(__pyx_tuple__367);

  /* "ultron/tradingday/Calendar.pyx":541
 *     Date(2013, 4, 30),
 *     Date(2013, 5, 1),
 *     Date(2013, 6, 10),             # <<<<<<<<<<<<<<
 *     Date(2013, 6, 11),
 *     Date(2013, 6, 12),
 */
  __pyx_tuple__368 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_6, __pyx_int_10); if (unlikely(!__pyx_tuple__368)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__368);
  __Pyx_GIVEREF(__pyx_tuple__368);

  /* "ultron/tradingday/Calendar.pyx":542
 *     Date(2013, 5, 1),
 *     Date(2013, 6, 10),
 *     Date(2013, 6, 11),             # <<<<<<<<<<<<<<
 *     Date(2013, 6, 12),
 *     Date(2013, 9, 19),
 */
  __pyx_tuple__369 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_6, __pyx_int_11); if (unlikely(!__pyx_tuple__369)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__369);
  __Pyx_GIVEREF(__pyx_tuple__369);

  /* "ultron/tradingday/Calendar.pyx":543
 *     Date(2013, 6, 10),
 *     Date(2013, 6, 11),
 *     Date(2013, 6, 12),             # <<<<<<<<<<<<<<
 *     Date(2013, 9, 19),
 *     Date(2013, 9, 20),
 */
  __pyx_tuple__370 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_6, __pyx_int_12); if (unlikely(!__pyx_tuple__370)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__370);
  __Pyx_GIVEREF(__pyx_tuple__370);

  /* "ultron/tradingday/Calendar.pyx":544
 *     Date(2013, 6, 11),
 *     Date(2013, 6, 12),
 *     Date(2013, 9, 19),             # <<<<<<<<<<<<<<
 *     Date(2013, 9, 20),
 *     Date(2013, 10, 1),
 */
  __pyx_tuple__371 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_9, __pyx_int_19); if (unlikely(!__pyx_tuple__371)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__371);
  __Pyx_GIVEREF(__pyx_tuple__371);

  /* "ultron/tradingday/Calendar.pyx":545
 *     Date(2013, 6, 12),
 *     Date(2013, 9, 19),
 *     Date(2013, 9, 20),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 1),
 *     Date(2013, 10, 2),
 */
  __pyx_tuple__372 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_9, __pyx_int_20); if (unlikely(!__pyx_tuple__372)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__372);
  __Pyx_GIVEREF(__pyx_tuple__372);

  /* "ultron/tradingday/Calendar.pyx":546
 *     Date(2013, 9, 19),
 *     Date(2013, 9, 20),
 *     Date(2013, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 2),
 *     Date(2013, 10, 3),
 */
  __pyx_tuple__373 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__373)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__373);
  __Pyx_GIVEREF(__pyx_tuple__373);

  /* "ultron/tradingday/Calendar.pyx":547
 *     Date(2013, 9, 20),
 *     Date(2013, 10, 1),
 *     Date(2013, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 3),
 *     Date(2013, 10, 4),
 */
  __pyx_tuple__374 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__374)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__374);
  __Pyx_GIVEREF(__pyx_tuple__374);

  /* "ultron/tradingday/Calendar.pyx":548
 *     Date(2013, 10, 1),
 *     Date(2013, 10, 2),
 *     Date(2013, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 4),
 *     Date(2013, 10, 7),
 */
  __pyx_tuple__375 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__375)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__375);
  __Pyx_GIVEREF(__pyx_tuple__375);

  /* "ultron/tradingday/Calendar.pyx":549
 *     Date(2013, 10, 2),
 *     Date(2013, 10, 3),
 *     Date(2013, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 7),
 *     Date(2014, 1, 1),
 */
  __pyx_tuple__376 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__376)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__376);
  __Pyx_GIVEREF(__pyx_tuple__376);

  /* "ultron/tradingday/Calendar.pyx":550
 *     Date(2013, 10, 3),
 *     Date(2013, 10, 4),
 *     Date(2013, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2014, 1, 1),
 *     Date(2014, 1, 31),
 */
  __pyx_tuple__377 = PyTuple_Pack(3, __pyx_int_2013, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__377)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__377);
  __Pyx_GIVEREF(__pyx_tuple__377);

  /* "ultron/tradingday/Calendar.pyx":551
 *     Date(2013, 10, 4),
 *     Date(2013, 10, 7),
 *     Date(2014, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2014, 1, 31),
 *     Date(2014, 2, 3),
 */
  __pyx_tuple__378 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__378)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__378);
  __Pyx_GIVEREF(__pyx_tuple__378);

  /* "ultron/tradingday/Calendar.pyx":552
 *     Date(2013, 10, 7),
 *     Date(2014, 1, 1),
 *     Date(2014, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2014, 2, 3),
 *     Date(2014, 2, 4),
 */
  __pyx_tuple__379 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_1, __pyx_int_31); if (unlikely(!__pyx_tuple__379)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__379);
  __Pyx_GIVEREF(__pyx_tuple__379);

  /* "ultron/tradingday/Calendar.pyx":553
 *     Date(2014, 1, 1),
 *     Date(2014, 1, 31),
 *     Date(2014, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2014, 2, 4),
 *     Date(2014, 2, 5),
 */
  __pyx_tuple__380 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__380)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__380);
  __Pyx_GIVEREF(__pyx_tuple__380);

  /* "ultron/tradingday/Calendar.pyx":554
 *     Date(2014, 1, 31),
 *     Date(2014, 2, 3),
 *     Date(2014, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2014, 2, 5),
 *     Date(2014, 2, 6),
 */
  __pyx_tuple__381 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__381)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__381);
  __Pyx_GIVEREF(__pyx_tuple__381);

  /* "ultron/tradingday/Calendar.pyx":555
 *     Date(2014, 2, 3),
 *     Date(2014, 2, 4),
 *     Date(2014, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(2014, 2, 6),
 *     Date(2014, 4, 7),
 */
  __pyx_tuple__382 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_2, __pyx_int_5); if (unlikely(!__pyx_tuple__382)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__382);
  __Pyx_GIVEREF(__pyx_tuple__382);

  /* "ultron/tradingday/Calendar.pyx":556
 *     Date(2014, 2, 4),
 *     Date(2014, 2, 5),
 *     Date(2014, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(2014, 4, 7),
 *     Date(2014, 5, 1),
 */
  __pyx_tuple__383 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_2, __pyx_int_6); if (unlikely(!__pyx_tuple__383)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__383);
  __Pyx_GIVEREF(__pyx_tuple__383);

  /* "ultron/tradingday/Calendar.pyx":557
 *     Date(2014, 2, 5),
 *     Date(2014, 2, 6),
 *     Date(2014, 4, 7),             # <<<<<<<<<<<<<<
 *     Date(2014, 5, 1),
 *     Date(2014, 5, 2),
 */
  __pyx_tuple__384 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_4, __pyx_int_7); if (unlikely(!__pyx_tuple__384)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__384);
  __Pyx_GIVEREF(__pyx_tuple__384);

  /* "ultron/tradingday/Calendar.pyx":558
 *     Date(2014, 2, 6),
 *     Date(2014, 4, 7),
 *     Date(2014, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2014, 5, 2),
 *     Date(2014, 6, 2),
 */
  __pyx_tuple__385 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__385)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__385);
  __Pyx_GIVEREF(__pyx_tuple__385);

  /* "ultron/tradingday/Calendar.pyx":559
 *     Date(2014, 4, 7),
 *     Date(2014, 5, 1),
 *     Date(2014, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2014, 6, 2),
 *     Date(2014, 9, 8),
 */
  __pyx_tuple__386 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__386)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__386);
  __Pyx_GIVEREF(__pyx_tuple__386);

  /* "ultron/tradingday/Calendar.pyx":560
 *     Date(2014, 5, 1),
 *     Date(2014, 5, 2),
 *     Date(2014, 6, 2),             # <<<<<<<<<<<<<<
 *     Date(2014, 9, 8),
 *     Date(2014, 10, 1),
 */
  __pyx_tuple__387 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_6, __pyx_int_2); if (unlikely(!__pyx_tuple__387)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__387);
  __Pyx_GIVEREF(__pyx_tuple__387);

  /* "ultron/tradingday/Calendar.pyx":561
 *     Date(2014, 5, 2),
 *     Date(2014, 6, 2),
 *     Date(2014, 9, 8),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 1),
 *     Date(2014, 10, 2),
 */
  __pyx_tuple__388 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_9, __pyx_int_8); if (unlikely(!__pyx_tuple__388)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__388);
  __Pyx_GIVEREF(__pyx_tuple__388);

  /* "ultron/tradingday/Calendar.pyx":562
 *     Date(2014, 6, 2),
 *     Date(2014, 9, 8),
 *     Date(2014, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 2),
 *     Date(2014, 10, 3),
 */
  __pyx_tuple__389 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__389)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__389);
  __Pyx_GIVEREF(__pyx_tuple__389);

  /* "ultron/tradingday/Calendar.pyx":563
 *     Date(2014, 9, 8),
 *     Date(2014, 10, 1),
 *     Date(2014, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 3),
 *     Date(2014, 10, 6),
 */
  __pyx_tuple__390 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__390)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__390);
  __Pyx_GIVEREF(__pyx_tuple__390);

  /* "ultron/tradingday/Calendar.pyx":564
 *     Date(2014, 10, 1),
 *     Date(2014, 10, 2),
 *     Date(2014, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 6),
 *     Date(2014, 10, 7),
 */
  __pyx_tuple__391 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__391)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__391);
  __Pyx_GIVEREF(__pyx_tuple__391);

  /* "ultron/tradingday/Calendar.pyx":565
 *     Date(2014, 10, 2),
 *     Date(2014, 10, 3),
 *     Date(2014, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 7),
 *     Date(2015, 1, 1),
 */
  __pyx_tuple__392 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__392)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__392);
  __Pyx_GIVEREF(__pyx_tuple__392);

  /* "ultron/tradingday/Calendar.pyx":566
 *     Date(2014, 10, 3),
 *     Date(2014, 10, 6),
 *     Date(2014, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2015, 1, 1),
 *     Date(2015, 1, 2),
 */
  __pyx_tuple__393 = PyTuple_Pack(3, __pyx_int_2014, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__393)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__393);
  __Pyx_GIVEREF(__pyx_tuple__393);

  /* "ultron/tradingday/Calendar.pyx":567
 *     Date(2014, 10, 6),
 *     Date(2014, 10, 7),
 *     Date(2015, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2015, 1, 2),
 *     Date(2015, 2, 18),
 */
  __pyx_tuple__394 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__394)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__394);
  __Pyx_GIVEREF(__pyx_tuple__394);

  /* "ultron/tradingday/Calendar.pyx":568
 *     Date(2014, 10, 7),
 *     Date(2015, 1, 1),
 *     Date(2015, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 18),
 *     Date(2015, 2, 19),
 */
  __pyx_tuple__395 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__395)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__395);
  __Pyx_GIVEREF(__pyx_tuple__395);

  /* "ultron/tradingday/Calendar.pyx":569
 *     Date(2015, 1, 1),
 *     Date(2015, 1, 2),
 *     Date(2015, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 19),
 *     Date(2015, 2, 20),
 */
  __pyx_tuple__396 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_2, __pyx_int_18); if (unlikely(!__pyx_tuple__396)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__396);
  __Pyx_GIVEREF(__pyx_tuple__396);

  /* "ultron/tradingday/Calendar.pyx":570
 *     Date(2015, 1, 2),
 *     Date(2015, 2, 18),
 *     Date(2015, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 20),
 *     Date(2015, 2, 23),
 */
  __pyx_tuple__397 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_2, __pyx_int_19); if (unlikely(!__pyx_tuple__397)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__397);
  __Pyx_GIVEREF(__pyx_tuple__397);

  /* "ultron/tradingday/Calendar.pyx":571
 *     Date(2015, 2, 18),
 *     Date(2015, 2, 19),
 *     Date(2015, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 23),
 *     Date(2015, 2, 24),
 */
  __pyx_tuple__398 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_2, __pyx_int_20); if (unlikely(!__pyx_tuple__398)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__398);
  __Pyx_GIVEREF(__pyx_tuple__398);

  /* "ultron/tradingday/Calendar.pyx":572
 *     Date(2015, 2, 19),
 *     Date(2015, 2, 20),
 *     Date(2015, 2, 23),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 24),
 *     Date(2015, 4, 6),
 */
  __pyx_tuple__399 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_2, __pyx_int_23); if (unlikely(!__pyx_tuple__399)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__399);
  __Pyx_GIVEREF(__pyx_tuple__399);

  /* "ultron/tradingday/Calendar.pyx":573
 *     Date(2015, 2, 20),
 *     Date(2015, 2, 23),
 *     Date(2015, 2, 24),             # <<<<<<<<<<<<<<
 *     Date(2015, 4, 6),
 *     Date(2015, 5, 1),
 */
  __pyx_tuple__400 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_2, __pyx_int_24); if (unlikely(!__pyx_tuple__400)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__400);
  __Pyx_GIVEREF(__pyx_tuple__400);

  /* "ultron/tradingday/Calendar.pyx":574
 *     Date(2015, 2, 23),
 *     Date(2015, 2, 24),
 *     Date(2015, 4, 6),             # <<<<<<<<<<<<<<
 *     Date(2015, 5, 1),
 *     Date(2015, 6, 22),
 */
  __pyx_tuple__401 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_4, __pyx_int_6); if (unlikely(!__pyx_tuple__401)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__401);
  __Pyx_GIVEREF(__pyx_tuple__401);

  /* "ultron/tradingday/Calendar.pyx":575
 *     Date(2015, 2, 24),
 *     Date(2015, 4, 6),
 *     Date(2015, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2015, 6, 22),
 *     Date(2015, 9, 3),
 */
  __pyx_tuple__402 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__402)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__402);
  __Pyx_GIVEREF(__pyx_tuple__402);

  /* "ultron/tradingday/Calendar.pyx":576
 *     Date(2015, 4, 6),
 *     Date(2015, 5, 1),
 *     Date(2015, 6, 22),             # <<<<<<<<<<<<<<
 *     Date(2015, 9, 3),
 *     Date(2015, 9, 4),
 */
  __pyx_tuple__403 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_6, __pyx_int_22); if (unlikely(!__pyx_tuple__403)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__403);
  __Pyx_GIVEREF(__pyx_tuple__403);

  /* "ultron/tradingday/Calendar.pyx":577
 *     Date(2015, 5, 1),
 *     Date(2015, 6, 22),
 *     Date(2015, 9, 3),             # <<<<<<<<<<<<<<
 *     Date(2015, 9, 4),
 *     Date(2015, 10, 1),
 */
  __pyx_tuple__404 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_9, __pyx_int_3); if (unlikely(!__pyx_tuple__404)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__404);
  __Pyx_GIVEREF(__pyx_tuple__404);

  /* "ultron/tradingday/Calendar.pyx":578
 *     Date(2015, 6, 22),
 *     Date(2015, 9, 3),
 *     Date(2015, 9, 4),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 1),
 *     Date(2015, 10, 2),
 */
  __pyx_tuple__405 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_9, __pyx_int_4); if (unlikely(!__pyx_tuple__405)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__405);
  __Pyx_GIVEREF(__pyx_tuple__405);

  /* "ultron/tradingday/Calendar.pyx":579
 *     Date(2015, 9, 3),
 *     Date(2015, 9, 4),
 *     Date(2015, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 2),
 *     Date(2015, 10, 5),
 */
  __pyx_tuple__406 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__406)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__406);
  __Pyx_GIVEREF(__pyx_tuple__406);

  /* "ultron/tradingday/Calendar.pyx":580
 *     Date(2015, 9, 4),
 *     Date(2015, 10, 1),
 *     Date(2015, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 5),
 *     Date(2015, 10, 6),
 */
  __pyx_tuple__407 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__407)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__407);
  __Pyx_GIVEREF(__pyx_tuple__407);

  /* "ultron/tradingday/Calendar.pyx":581
 *     Date(2015, 10, 1),
 *     Date(2015, 10, 2),
 *     Date(2015, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 6),
 *     Date(2015, 10, 7),
 */
  __pyx_tuple__408 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__408)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__408);
  __Pyx_GIVEREF(__pyx_tuple__408);

  /* "ultron/tradingday/Calendar.pyx":582
 *     Date(2015, 10, 2),
 *     Date(2015, 10, 5),
 *     Date(2015, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 7),
 *     Date(2016, 1, 1),
 */
  __pyx_tuple__409 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__409)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__409);
  __Pyx_GIVEREF(__pyx_tuple__409);

  /* "ultron/tradingday/Calendar.pyx":583
 *     Date(2015, 10, 5),
 *     Date(2015, 10, 6),
 *     Date(2015, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2016, 1, 1),
 *     Date(2016, 2, 8),
 */
  __pyx_tuple__410 = PyTuple_Pack(3, __pyx_int_2015, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__410)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__410);
  __Pyx_GIVEREF(__pyx_tuple__410);

  /* "ultron/tradingday/Calendar.pyx":584
 *     Date(2015, 10, 6),
 *     Date(2015, 10, 7),
 *     Date(2016, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 8),
 *     Date(2016, 2, 9),
 */
  __pyx_tuple__411 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__411)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__411);
  __Pyx_GIVEREF(__pyx_tuple__411);

  /* "ultron/tradingday/Calendar.pyx":585
 *     Date(2015, 10, 7),
 *     Date(2016, 1, 1),
 *     Date(2016, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 9),
 *     Date(2016, 2, 10),
 */
  __pyx_tuple__412 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_2, __pyx_int_8); if (unlikely(!__pyx_tuple__412)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__412);
  __Pyx_GIVEREF(__pyx_tuple__412);

  /* "ultron/tradingday/Calendar.pyx":586
 *     Date(2016, 1, 1),
 *     Date(2016, 2, 8),
 *     Date(2016, 2, 9),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 10),
 *     Date(2016, 2, 11),
 */
  __pyx_tuple__413 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_2, __pyx_int_9); if (unlikely(!__pyx_tuple__413)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__413);
  __Pyx_GIVEREF(__pyx_tuple__413);

  /* "ultron/tradingday/Calendar.pyx":587
 *     Date(2016, 2, 8),
 *     Date(2016, 2, 9),
 *     Date(2016, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 11),
 *     Date(2016, 2, 12),
 */
  __pyx_tuple__414 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_2, __pyx_int_10); if (unlikely(!__pyx_tuple__414)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__414);
  __Pyx_GIVEREF(__pyx_tuple__414);

  /* "ultron/tradingday/Calendar.pyx":588
 *     Date(2016, 2, 9),
 *     Date(2016, 2, 10),
 *     Date(2016, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 12),
 *     Date(2016, 4, 4),
 */
  __pyx_tuple__415 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__415)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__415);
  __Pyx_GIVEREF(__pyx_tuple__415);

  /* "ultron/tradingday/Calendar.pyx":589
 *     Date(2016, 2, 10),
 *     Date(2016, 2, 11),
 *     Date(2016, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2016, 4, 4),
 *     Date(2016, 5, 1),
 */
  __pyx_tuple__416 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_2, __pyx_int_12); if (unlikely(!__pyx_tuple__416)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__416);
  __Pyx_GIVEREF(__pyx_tuple__416);

  /* "ultron/tradingday/Calendar.pyx":590
 *     Date(2016, 2, 11),
 *     Date(2016, 2, 12),
 *     Date(2016, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2016, 5, 1),
 *     Date(2016, 5, 2),
 */
  __pyx_tuple__417 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_4, __pyx_int_4); if (unlikely(!__pyx_tuple__417)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__417);
  __Pyx_GIVEREF(__pyx_tuple__417);

  /* "ultron/tradingday/Calendar.pyx":591
 *     Date(2016, 2, 12),
 *     Date(2016, 4, 4),
 *     Date(2016, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2016, 5, 2),
 *     Date(2016, 6, 9),
 */
  __pyx_tuple__418 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__418)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__418);
  __Pyx_GIVEREF(__pyx_tuple__418);

  /* "ultron/tradingday/Calendar.pyx":592
 *     Date(2016, 4, 4),
 *     Date(2016, 5, 1),
 *     Date(2016, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2016, 6, 9),
 *     Date(2016, 6, 10),
 */
  __pyx_tuple__419 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__419)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__419);
  __Pyx_GIVEREF(__pyx_tuple__419);

  /* "ultron/tradingday/Calendar.pyx":593
 *     Date(2016, 5, 1),
 *     Date(2016, 5, 2),
 *     Date(2016, 6, 9),             # <<<<<<<<<<<<<<
 *     Date(2016, 6, 10),
 *     Date(2016, 9, 15),
 */
  __pyx_tuple__420 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_6, __pyx_int_9); if (unlikely(!__pyx_tuple__420)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__420);
  __Pyx_GIVEREF(__pyx_tuple__420);

  /* "ultron/tradingday/Calendar.pyx":594
 *     Date(2016, 5, 2),
 *     Date(2016, 6, 9),
 *     Date(2016, 6, 10),             # <<<<<<<<<<<<<<
 *     Date(2016, 9, 15),
 *     Date(2016, 9, 16),
 */
  __pyx_tuple__421 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_6, __pyx_int_10); if (unlikely(!__pyx_tuple__421)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__421);
  __Pyx_GIVEREF(__pyx_tuple__421);

  /* "ultron/tradingday/Calendar.pyx":595
 *     Date(2016, 6, 9),
 *     Date(2016, 6, 10),
 *     Date(2016, 9, 15),             # <<<<<<<<<<<<<<
 *     Date(2016, 9, 16),
 *     Date(2016, 10, 3),
 */
  __pyx_tuple__422 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_9, __pyx_int_15); if (unlikely(!__pyx_tuple__422)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__422);
  __Pyx_GIVEREF(__pyx_tuple__422);

  /* "ultron/tradingday/Calendar.pyx":596
 *     Date(2016, 6, 10),
 *     Date(2016, 9, 15),
 *     Date(2016, 9, 16),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 3),
 *     Date(2016, 10, 4),
 */
  __pyx_tuple__423 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_9, __pyx_int_16); if (unlikely(!__pyx_tuple__423)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__423);
  __Pyx_GIVEREF(__pyx_tuple__423);

  /* "ultron/tradingday/Calendar.pyx":597
 *     Date(2016, 9, 15),
 *     Date(2016, 9, 16),
 *     Date(2016, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 4),
 *     Date(2016, 10, 5),
 */
  __pyx_tuple__424 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__424)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__424);
  __Pyx_GIVEREF(__pyx_tuple__424);

  /* "ultron/tradingday/Calendar.pyx":598
 *     Date(2016, 9, 16),
 *     Date(2016, 10, 3),
 *     Date(2016, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 5),
 *     Date(2016, 10, 6),
 */
  __pyx_tuple__425 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__425)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__425);
  __Pyx_GIVEREF(__pyx_tuple__425);

  /* "ultron/tradingday/Calendar.pyx":599
 *     Date(2016, 10, 3),
 *     Date(2016, 10, 4),
 *     Date(2016, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 6),
 *     Date(2016, 10, 7),
 */
  __pyx_tuple__426 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__426)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__426);
  __Pyx_GIVEREF(__pyx_tuple__426);

  /* "ultron/tradingday/Calendar.pyx":600
 *     Date(2016, 10, 4),
 *     Date(2016, 10, 5),
 *     Date(2016, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 7),
 *     Date(2017, 1, 2),
 */
  __pyx_tuple__427 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__427)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__427);
  __Pyx_GIVEREF(__pyx_tuple__427);

  /* "ultron/tradingday/Calendar.pyx":601
 *     Date(2016, 10, 5),
 *     Date(2016, 10, 6),
 *     Date(2016, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2017, 1, 2),
 *     Date(2017, 1, 27),
 */
  __pyx_tuple__428 = PyTuple_Pack(3, __pyx_int_2016, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__428)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__428);
  __Pyx_GIVEREF(__pyx_tuple__428);

  /* "ultron/tradingday/Calendar.pyx":602
 *     Date(2016, 10, 6),
 *     Date(2016, 10, 7),
 *     Date(2017, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2017, 1, 27),
 *     Date(2017, 1, 30),
 */
  __pyx_tuple__429 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__429)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__429);
  __Pyx_GIVEREF(__pyx_tuple__429);

  /* "ultron/tradingday/Calendar.pyx":603
 *     Date(2016, 10, 7),
 *     Date(2017, 1, 2),
 *     Date(2017, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2017, 1, 30),
 *     Date(2017, 1, 31),
 */
  __pyx_tuple__430 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_1, __pyx_int_27); if (unlikely(!__pyx_tuple__430)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__430);
  __Pyx_GIVEREF(__pyx_tuple__430);

  /* "ultron/tradingday/Calendar.pyx":604
 *     Date(2017, 1, 2),
 *     Date(2017, 1, 27),
 *     Date(2017, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2017, 1, 31),
 *     Date(2017, 2, 1),
 */
  __pyx_tuple__431 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_1, __pyx_int_30); if (unlikely(!__pyx_tuple__431)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__431);
  __Pyx_GIVEREF(__pyx_tuple__431);

  /* "ultron/tradingday/Calendar.pyx":605
 *     Date(2017, 1, 27),
 *     Date(2017, 1, 30),
 *     Date(2017, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2017, 2, 1),
 *     Date(2017, 2, 2),
 */
  __pyx_tuple__432 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_1, __pyx_int_31); if (unlikely(!__pyx_tuple__432)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__432);
  __Pyx_GIVEREF(__pyx_tuple__432);

  /* "ultron/tradingday/Calendar.pyx":606
 *     Date(2017, 1, 30),
 *     Date(2017, 1, 31),
 *     Date(2017, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2017, 2, 2),
 *     Date(2017, 4, 3),
 */
  __pyx_tuple__433 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_2, __pyx_int_1); if (unlikely(!__pyx_tuple__433)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__433);
  __Pyx_GIVEREF(__pyx_tuple__433);

  /* "ultron/tradingday/Calendar.pyx":607
 *     Date(2017, 1, 31),
 *     Date(2017, 2, 1),
 *     Date(2017, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2017, 4, 3),
 *     Date(2017, 4, 4),
 */
  __pyx_tuple__434 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__434)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__434);
  __Pyx_GIVEREF(__pyx_tuple__434);

  /* "ultron/tradingday/Calendar.pyx":608
 *     Date(2017, 2, 1),
 *     Date(2017, 2, 2),
 *     Date(2017, 4, 3),             # <<<<<<<<<<<<<<
 *     Date(2017, 4, 4),
 *     Date(2017, 5, 1),
 */
  __pyx_tuple__435 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_4, __pyx_int_3); if (unlikely(!__pyx_tuple__435)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__435);
  __Pyx_GIVEREF(__pyx_tuple__435);

  /* "ultron/tradingday/Calendar.pyx":609
 *     Date(2017, 2, 2),
 *     Date(2017, 4, 3),
 *     Date(2017, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2017, 5, 1),
 *     Date(2017, 5, 29),
 */
  __pyx_tuple__436 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_4, __pyx_int_4); if (unlikely(!__pyx_tuple__436)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__436);
  __Pyx_GIVEREF(__pyx_tuple__436);

  /* "ultron/tradingday/Calendar.pyx":610
 *     Date(2017, 4, 3),
 *     Date(2017, 4, 4),
 *     Date(2017, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2017, 5, 29),
 *     Date(2017, 5, 30),
 */
  __pyx_tuple__437 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__437)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__437);
  __Pyx_GIVEREF(__pyx_tuple__437);

  /* "ultron/tradingday/Calendar.pyx":611
 *     Date(2017, 4, 4),
 *     Date(2017, 5, 1),
 *     Date(2017, 5, 29),             # <<<<<<<<<<<<<<
 *     Date(2017, 5, 30),
 *     Date(2017, 10, 2),
 */
  __pyx_tuple__438 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_5, __pyx_int_29); if (unlikely(!__pyx_tuple__438)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__438);
  __Pyx_GIVEREF(__pyx_tuple__438);

  /* "ultron/tradingday/Calendar.pyx":612
 *     Date(2017, 5, 1),
 *     Date(2017, 5, 29),
 *     Date(2017, 5, 30),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 2),
 *     Date(2017, 10, 3),
 */
  __pyx_tuple__439 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_5, __pyx_int_30); if (unlikely(!__pyx_tuple__439)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__439);
  __Pyx_GIVEREF(__pyx_tuple__439);

  /* "ultron/tradingday/Calendar.pyx":613
 *     Date(2017, 5, 29),
 *     Date(2017, 5, 30),
 *     Date(2017, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 3),
 *     Date(2017, 10, 4),
 */
  __pyx_tuple__440 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__440)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__440);
  __Pyx_GIVEREF(__pyx_tuple__440);

  /* "ultron/tradingday/Calendar.pyx":614
 *     Date(2017, 5, 30),
 *     Date(2017, 10, 2),
 *     Date(2017, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 4),
 *     Date(2017, 10, 5),
 */
  __pyx_tuple__441 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__441)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__441);
  __Pyx_GIVEREF(__pyx_tuple__441);

  /* "ultron/tradingday/Calendar.pyx":615
 *     Date(2017, 10, 2),
 *     Date(2017, 10, 3),
 *     Date(2017, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 5),
 *     Date(2017, 10, 6),
 */
  __pyx_tuple__442 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__442)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__442);
  __Pyx_GIVEREF(__pyx_tuple__442);

  /* "ultron/tradingday/Calendar.pyx":616
 *     Date(2017, 10, 3),
 *     Date(2017, 10, 4),
 *     Date(2017, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 6),
 *     Date(2018, 1, 1),
 */
  __pyx_tuple__443 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__443)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__443);
  __Pyx_GIVEREF(__pyx_tuple__443);

  /* "ultron/tradingday/Calendar.pyx":617
 *     Date(2017, 10, 4),
 *     Date(2017, 10, 5),
 *     Date(2017, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2018, 1, 1),
 *     Date(2018, 2, 15),
 */
  __pyx_tuple__444 = PyTuple_Pack(3, __pyx_int_2017, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__444)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__444);
  __Pyx_GIVEREF(__pyx_tuple__444);

  /* "ultron/tradingday/Calendar.pyx":618
 *     Date(2017, 10, 5),
 *     Date(2017, 10, 6),
 *     Date(2018, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 15),
 *     Date(2018, 2, 16),
 */
  __pyx_tuple__445 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__445)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__445);
  __Pyx_GIVEREF(__pyx_tuple__445);

  /* "ultron/tradingday/Calendar.pyx":619
 *     Date(2017, 10, 6),
 *     Date(2018, 1, 1),
 *     Date(2018, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 16),
 *     Date(2018, 2, 19),
 */
  __pyx_tuple__446 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_2, __pyx_int_15); if (unlikely(!__pyx_tuple__446)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__446);
  __Pyx_GIVEREF(__pyx_tuple__446);

  /* "ultron/tradingday/Calendar.pyx":620
 *     Date(2018, 1, 1),
 *     Date(2018, 2, 15),
 *     Date(2018, 2, 16),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 19),
 *     Date(2018, 2, 20),
 */
  __pyx_tuple__447 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_2, __pyx_int_16); if (unlikely(!__pyx_tuple__447)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__447);
  __Pyx_GIVEREF(__pyx_tuple__447);

  /* "ultron/tradingday/Calendar.pyx":621
 *     Date(2018, 2, 15),
 *     Date(2018, 2, 16),
 *     Date(2018, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 20),
 *     Date(2018, 2, 21),
 */
  __pyx_tuple__448 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_2, __pyx_int_19); if (unlikely(!__pyx_tuple__448)) __PYX_ERR(0, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__448);
  __Pyx_GIVEREF(__pyx_tuple__448);

  /* "ultron/tradingday/Calendar.pyx":622
 *     Date(2018, 2, 16),
 *     Date(2018, 2, 19),
 *     Date(2018, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 21),
 *     Date(2018, 4, 5),
 */
  __pyx_tuple__449 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_2, __pyx_int_20); if (unlikely(!__pyx_tuple__449)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__449);
  __Pyx_GIVEREF(__pyx_tuple__449);

  /* "ultron/tradingday/Calendar.pyx":623
 *     Date(2018, 2, 19),
 *     Date(2018, 2, 20),
 *     Date(2018, 2, 21),             # <<<<<<<<<<<<<<
 *     Date(2018, 4, 5),
 *     Date(2018, 4, 6),
 */
  __pyx_tuple__450 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_2, __pyx_int_21); if (unlikely(!__pyx_tuple__450)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__450);
  __Pyx_GIVEREF(__pyx_tuple__450);

  /* "ultron/tradingday/Calendar.pyx":624
 *     Date(2018, 2, 20),
 *     Date(2018, 2, 21),
 *     Date(2018, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2018, 4, 6),
 *     Date(2018, 4, 30),
 */
  __pyx_tuple__451 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_4, __pyx_int_5); if (unlikely(!__pyx_tuple__451)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__451);
  __Pyx_GIVEREF(__pyx_tuple__451);

  /* "ultron/tradingday/Calendar.pyx":625
 *     Date(2018, 2, 21),
 *     Date(2018, 4, 5),
 *     Date(2018, 4, 6),             # <<<<<<<<<<<<<<
 *     Date(2018, 4, 30),
 *     Date(2018, 5, 1),
 */
  __pyx_tuple__452 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_4, __pyx_int_6); if (unlikely(!__pyx_tuple__452)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__452);
  __Pyx_GIVEREF(__pyx_tuple__452);

  /* "ultron/tradingday/Calendar.pyx":626
 *     Date(2018, 4, 5),
 *     Date(2018, 4, 6),
 *     Date(2018, 4, 30),             # <<<<<<<<<<<<<<
 *     Date(2018, 5, 1),
 *     Date(2018, 6, 18),
 */
  __pyx_tuple__453 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_4, __pyx_int_30); if (unlikely(!__pyx_tuple__453)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__453);
  __Pyx_GIVEREF(__pyx_tuple__453);

  /* "ultron/tradingday/Calendar.pyx":627
 *     Date(2018, 4, 6),
 *     Date(2018, 4, 30),
 *     Date(2018, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2018, 6, 18),
 *     Date(2018, 9, 24),
 */
  __pyx_tuple__454 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__454)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__454);
  __Pyx_GIVEREF(__pyx_tuple__454);

  /* "ultron/tradingday/Calendar.pyx":628
 *     Date(2018, 4, 30),
 *     Date(2018, 5, 1),
 *     Date(2018, 6, 18),             # <<<<<<<<<<<<<<
 *     Date(2018, 9, 24),
 *     Date(2018, 10, 1),
 */
  __pyx_tuple__455 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_6, __pyx_int_18); if (unlikely(!__pyx_tuple__455)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__455);
  __Pyx_GIVEREF(__pyx_tuple__455);

  /* "ultron/tradingday/Calendar.pyx":629
 *     Date(2018, 5, 1),
 *     Date(2018, 6, 18),
 *     Date(2018, 9, 24),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 1),
 *     Date(2018, 10, 2),
 */
  __pyx_tuple__456 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_9, __pyx_int_24); if (unlikely(!__pyx_tuple__456)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__456);
  __Pyx_GIVEREF(__pyx_tuple__456);

  /* "ultron/tradingday/Calendar.pyx":630
 *     Date(2018, 6, 18),
 *     Date(2018, 9, 24),
 *     Date(2018, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 2),
 *     Date(2018, 10, 3),
 */
  __pyx_tuple__457 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__457)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__457);
  __Pyx_GIVEREF(__pyx_tuple__457);

  /* "ultron/tradingday/Calendar.pyx":631
 *     Date(2018, 9, 24),
 *     Date(2018, 10, 1),
 *     Date(2018, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 3),
 *     Date(2018, 10, 4),
 */
  __pyx_tuple__458 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__458)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__458);
  __Pyx_GIVEREF(__pyx_tuple__458);

  /* "ultron/tradingday/Calendar.pyx":632
 *     Date(2018, 10, 1),
 *     Date(2018, 10, 2),
 *     Date(2018, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 4),
 *     Date(2018, 10, 5),
 */
  __pyx_tuple__459 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__459)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__459);
  __Pyx_GIVEREF(__pyx_tuple__459);

  /* "ultron/tradingday/Calendar.pyx":633
 *     Date(2018, 10, 2),
 *     Date(2018, 10, 3),
 *     Date(2018, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 5),
 *     Date(2018, 12, 31),
 */
  __pyx_tuple__460 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__460)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__460);
  __Pyx_GIVEREF(__pyx_tuple__460);

  /* "ultron/tradingday/Calendar.pyx":634
 *     Date(2018, 10, 3),
 *     Date(2018, 10, 4),
 *     Date(2018, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2018, 12, 31),
 *     Date(2019, 1, 1),
 */
  __pyx_tuple__461 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__461)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__461);
  __Pyx_GIVEREF(__pyx_tuple__461);

  /* "ultron/tradingday/Calendar.pyx":635
 *     Date(2018, 10, 4),
 *     Date(2018, 10, 5),
 *     Date(2018, 12, 31),             # <<<<<<<<<<<<<<
 *     Date(2019, 1, 1),
 *     Date(2019, 2, 4),
 */
  __pyx_tuple__462 = PyTuple_Pack(3, __pyx_int_2018, __pyx_int_12, __pyx_int_31); if (unlikely(!__pyx_tuple__462)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__462);
  __Pyx_GIVEREF(__pyx_tuple__462);

  /* "ultron/tradingday/Calendar.pyx":636
 *     Date(2018, 10, 5),
 *     Date(2018, 12, 31),
 *     Date(2019, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 4),
 *     Date(2019, 2, 5),
 */
  __pyx_tuple__463 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__463)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__463);
  __Pyx_GIVEREF(__pyx_tuple__463);

  /* "ultron/tradingday/Calendar.pyx":637
 *     Date(2018, 12, 31),
 *     Date(2019, 1, 1),
 *     Date(2019, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 5),
 *     Date(2019, 2, 6),
 */
  __pyx_tuple__464 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__464)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__464);
  __Pyx_GIVEREF(__pyx_tuple__464);

  /* "ultron/tradingday/Calendar.pyx":638
 *     Date(2019, 1, 1),
 *     Date(2019, 2, 4),
 *     Date(2019, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 6),
 *     Date(2019, 2, 7),
 */
  __pyx_tuple__465 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_2, __pyx_int_5); if (unlikely(!__pyx_tuple__465)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__465);
  __Pyx_GIVEREF(__pyx_tuple__465);

  /* "ultron/tradingday/Calendar.pyx":639
 *     Date(2019, 2, 4),
 *     Date(2019, 2, 5),
 *     Date(2019, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 7),
 *     Date(2019, 2, 8),
 */
  __pyx_tuple__466 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_2, __pyx_int_6); if (unlikely(!__pyx_tuple__466)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__466);
  __Pyx_GIVEREF(__pyx_tuple__466);

  /* "ultron/tradingday/Calendar.pyx":640
 *     Date(2019, 2, 5),
 *     Date(2019, 2, 6),
 *     Date(2019, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 8),
 *     Date(2019, 4, 5),
 */
  __pyx_tuple__467 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__467)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__467);
  __Pyx_GIVEREF(__pyx_tuple__467);

  /* "ultron/tradingday/Calendar.pyx":641
 *     Date(2019, 2, 6),
 *     Date(2019, 2, 7),
 *     Date(2019, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2019, 4, 5),
 *     Date(2019, 5, 1),
 */
  __pyx_tuple__468 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_2, __pyx_int_8); if (unlikely(!__pyx_tuple__468)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__468);
  __Pyx_GIVEREF(__pyx_tuple__468);

  /* "ultron/tradingday/Calendar.pyx":642
 *     Date(2019, 2, 7),
 *     Date(2019, 2, 8),
 *     Date(2019, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2019, 5, 1),
 *     Date(2019, 5, 2),
 */
  __pyx_tuple__469 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_4, __pyx_int_5); if (unlikely(!__pyx_tuple__469)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__469);
  __Pyx_GIVEREF(__pyx_tuple__469);

  /* "ultron/tradingday/Calendar.pyx":643
 *     Date(2019, 2, 8),
 *     Date(2019, 4, 5),
 *     Date(2019, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2019, 5, 2),
 *     Date(2019, 5, 3),
 */
  __pyx_tuple__470 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__470)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__470);
  __Pyx_GIVEREF(__pyx_tuple__470);

  /* "ultron/tradingday/Calendar.pyx":644
 *     Date(2019, 4, 5),
 *     Date(2019, 5, 1),
 *     Date(2019, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2019, 5, 3),
 *     Date(2019, 6, 7),
 */
  __pyx_tuple__471 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__471)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__471);
  __Pyx_GIVEREF(__pyx_tuple__471);

  /* "ultron/tradingday/Calendar.pyx":645
 *     Date(2019, 5, 1),
 *     Date(2019, 5, 2),
 *     Date(2019, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2019, 6, 7),
 *     Date(2019, 9, 13),
 */
  __pyx_tuple__472 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__472)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__472);
  __Pyx_GIVEREF(__pyx_tuple__472);

  /* "ultron/tradingday/Calendar.pyx":646
 *     Date(2019, 5, 2),
 *     Date(2019, 5, 3),
 *     Date(2019, 6, 7),             # <<<<<<<<<<<<<<
 *     Date(2019, 9, 13),
 *     Date(2019, 10, 1),
 */
  __pyx_tuple__473 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_6, __pyx_int_7); if (unlikely(!__pyx_tuple__473)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__473);
  __Pyx_GIVEREF(__pyx_tuple__473);

  /* "ultron/tradingday/Calendar.pyx":647
 *     Date(2019, 5, 3),
 *     Date(2019, 6, 7),
 *     Date(2019, 9, 13),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 1),
 *     Date(2019, 10, 2),
 */
  __pyx_tuple__474 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_9, __pyx_int_13); if (unlikely(!__pyx_tuple__474)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__474);
  __Pyx_GIVEREF(__pyx_tuple__474);

  /* "ultron/tradingday/Calendar.pyx":648
 *     Date(2019, 6, 7),
 *     Date(2019, 9, 13),
 *     Date(2019, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 2),
 *     Date(2019, 10, 3),
 */
  __pyx_tuple__475 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__475)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__475);
  __Pyx_GIVEREF(__pyx_tuple__475);

  /* "ultron/tradingday/Calendar.pyx":649
 *     Date(2019, 9, 13),
 *     Date(2019, 10, 1),
 *     Date(2019, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 3),
 *     Date(2019, 10, 4),
 */
  __pyx_tuple__476 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__476)) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__476);
  __Pyx_GIVEREF(__pyx_tuple__476);

  /* "ultron/tradingday/Calendar.pyx":650
 *     Date(2019, 10, 1),
 *     Date(2019, 10, 2),
 *     Date(2019, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 4),
 *     Date(2019, 10, 7),
 */
  __pyx_tuple__477 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__477)) __PYX_ERR(0, 650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__477);
  __Pyx_GIVEREF(__pyx_tuple__477);

  /* "ultron/tradingday/Calendar.pyx":651
 *     Date(2019, 10, 2),
 *     Date(2019, 10, 3),
 *     Date(2019, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 7),
 *     Date(2020, 1, 1),
 */
  __pyx_tuple__478 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__478)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__478);
  __Pyx_GIVEREF(__pyx_tuple__478);

  /* "ultron/tradingday/Calendar.pyx":652
 *     Date(2019, 10, 3),
 *     Date(2019, 10, 4),
 *     Date(2019, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 1),
 *     Date(2020, 1, 24),
 */
  __pyx_tuple__479 = PyTuple_Pack(3, __pyx_int_2019, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__479)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__479);
  __Pyx_GIVEREF(__pyx_tuple__479);

  /* "ultron/tradingday/Calendar.pyx":653
 *     Date(2019, 10, 4),
 *     Date(2019, 10, 7),
 *     Date(2020, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 24),
 *     Date(2020, 1, 27),
 */
  __pyx_tuple__480 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__480)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__480);
  __Pyx_GIVEREF(__pyx_tuple__480);

  /* "ultron/tradingday/Calendar.pyx":654
 *     Date(2019, 10, 7),
 *     Date(2020, 1, 1),
 *     Date(2020, 1, 24),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 27),
 *     Date(2020, 1, 28),
 */
  __pyx_tuple__481 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_1, __pyx_int_24); if (unlikely(!__pyx_tuple__481)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__481);
  __Pyx_GIVEREF(__pyx_tuple__481);

  /* "ultron/tradingday/Calendar.pyx":655
 *     Date(2020, 1, 1),
 *     Date(2020, 1, 24),
 *     Date(2020, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 28),
 *     Date(2020, 1, 29),
 */
  __pyx_tuple__482 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_1, __pyx_int_27); if (unlikely(!__pyx_tuple__482)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__482);
  __Pyx_GIVEREF(__pyx_tuple__482);

  /* "ultron/tradingday/Calendar.pyx":656
 *     Date(2020, 1, 24),
 *     Date(2020, 1, 27),
 *     Date(2020, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 29),
 *     Date(2020, 1, 30),
 */
  __pyx_tuple__483 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_1, __pyx_int_28); if (unlikely(!__pyx_tuple__483)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__483);
  __Pyx_GIVEREF(__pyx_tuple__483);

  /* "ultron/tradingday/Calendar.pyx":657
 *     Date(2020, 1, 27),
 *     Date(2020, 1, 28),
 *     Date(2020, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 30),
 *     Date(2020, 1, 31),
 */
  __pyx_tuple__484 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_1, __pyx_int_29); if (unlikely(!__pyx_tuple__484)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__484);
  __Pyx_GIVEREF(__pyx_tuple__484);

  /* "ultron/tradingday/Calendar.pyx":658
 *     Date(2020, 1, 28),
 *     Date(2020, 1, 29),
 *     Date(2020, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 31),
 *     Date(2020, 4, 6),
 */
  __pyx_tuple__485 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_1, __pyx_int_30); if (unlikely(!__pyx_tuple__485)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__485);
  __Pyx_GIVEREF(__pyx_tuple__485);

  /* "ultron/tradingday/Calendar.pyx":659
 *     Date(2020, 1, 29),
 *     Date(2020, 1, 30),
 *     Date(2020, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2020, 4, 6),
 *     Date(2020, 5, 1),
 */
  __pyx_tuple__486 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_1, __pyx_int_31); if (unlikely(!__pyx_tuple__486)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__486);
  __Pyx_GIVEREF(__pyx_tuple__486);

  /* "ultron/tradingday/Calendar.pyx":660
 *     Date(2020, 1, 30),
 *     Date(2020, 1, 31),
 *     Date(2020, 4, 6),             # <<<<<<<<<<<<<<
 *     Date(2020, 5, 1),
 *     Date(2020, 5, 4),
 */
  __pyx_tuple__487 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_4, __pyx_int_6); if (unlikely(!__pyx_tuple__487)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__487);
  __Pyx_GIVEREF(__pyx_tuple__487);

  /* "ultron/tradingday/Calendar.pyx":661
 *     Date(2020, 1, 31),
 *     Date(2020, 4, 6),
 *     Date(2020, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2020, 5, 4),
 *     Date(2020, 5, 5),
 */
  __pyx_tuple__488 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__488)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__488);
  __Pyx_GIVEREF(__pyx_tuple__488);

  /* "ultron/tradingday/Calendar.pyx":662
 *     Date(2020, 4, 6),
 *     Date(2020, 5, 1),
 *     Date(2020, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2020, 5, 5),
 *     Date(2020, 6, 25),
 */
  __pyx_tuple__489 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_5, __pyx_int_4); if (unlikely(!__pyx_tuple__489)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__489);
  __Pyx_GIVEREF(__pyx_tuple__489);

  /* "ultron/tradingday/Calendar.pyx":663
 *     Date(2020, 5, 1),
 *     Date(2020, 5, 4),
 *     Date(2020, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2020, 6, 25),
 *     Date(2020, 6, 26),
 */
  __pyx_tuple__490 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_5, __pyx_int_5); if (unlikely(!__pyx_tuple__490)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__490);
  __Pyx_GIVEREF(__pyx_tuple__490);

  /* "ultron/tradingday/Calendar.pyx":664
 *     Date(2020, 5, 4),
 *     Date(2020, 5, 5),
 *     Date(2020, 6, 25),             # <<<<<<<<<<<<<<
 *     Date(2020, 6, 26),
 *     Date(2020, 10, 1),
 */
  __pyx_tuple__491 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_6, __pyx_int_25); if (unlikely(!__pyx_tuple__491)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__491);
  __Pyx_GIVEREF(__pyx_tuple__491);

  /* "ultron/tradingday/Calendar.pyx":665
 *     Date(2020, 5, 5),
 *     Date(2020, 6, 25),
 *     Date(2020, 6, 26),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 1),
 *     Date(2020, 10, 2),
 */
  __pyx_tuple__492 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_6, __pyx_int_26); if (unlikely(!__pyx_tuple__492)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__492);
  __Pyx_GIVEREF(__pyx_tuple__492);

  /* "ultron/tradingday/Calendar.pyx":666
 *     Date(2020, 6, 25),
 *     Date(2020, 6, 26),
 *     Date(2020, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 2),
 *     Date(2020, 10, 5),
 */
  __pyx_tuple__493 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__493)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__493);
  __Pyx_GIVEREF(__pyx_tuple__493);

  /* "ultron/tradingday/Calendar.pyx":667
 *     Date(2020, 6, 26),
 *     Date(2020, 10, 1),
 *     Date(2020, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 5),
 *     Date(2020, 10, 6),
 */
  __pyx_tuple__494 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__494)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__494);
  __Pyx_GIVEREF(__pyx_tuple__494);

  /* "ultron/tradingday/Calendar.pyx":668
 *     Date(2020, 10, 1),
 *     Date(2020, 10, 2),
 *     Date(2020, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 6),
 *     Date(2020, 10, 7),
 */
  __pyx_tuple__495 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__495)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__495);
  __Pyx_GIVEREF(__pyx_tuple__495);

  /* "ultron/tradingday/Calendar.pyx":669
 *     Date(2020, 10, 2),
 *     Date(2020, 10, 5),
 *     Date(2020, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 7),
 *     Date(2020, 10, 8),
 */
  __pyx_tuple__496 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__496)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__496);
  __Pyx_GIVEREF(__pyx_tuple__496);

  /* "ultron/tradingday/Calendar.pyx":670
 *     Date(2020, 10, 5),
 *     Date(2020, 10, 6),
 *     Date(2020, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 8),
 *     Date(2021, 1, 1),
 */
  __pyx_tuple__497 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__497)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__497);
  __Pyx_GIVEREF(__pyx_tuple__497);

  /* "ultron/tradingday/Calendar.pyx":671
 *     Date(2020, 10, 6),
 *     Date(2020, 10, 7),
 *     Date(2020, 10, 8),             # <<<<<<<<<<<<<<
 *     Date(2021, 1, 1),
 *     Date(2021, 2, 11),
 */
  __pyx_tuple__498 = PyTuple_Pack(3, __pyx_int_2020, __pyx_int_10, __pyx_int_8); if (unlikely(!__pyx_tuple__498)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__498);
  __Pyx_GIVEREF(__pyx_tuple__498);

  /* "ultron/tradingday/Calendar.pyx":672
 *     Date(2020, 10, 7),
 *     Date(2020, 10, 8),
 *     Date(2021, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 11),
 *     Date(2021, 2, 12),
 */
  __pyx_tuple__499 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__499)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__499);
  __Pyx_GIVEREF(__pyx_tuple__499);

  /* "ultron/tradingday/Calendar.pyx":673
 *     Date(2020, 10, 8),
 *     Date(2021, 1, 1),
 *     Date(2021, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 12),
 *     Date(2021, 2, 15),
 */
  __pyx_tuple__500 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_2, __pyx_int_11); if (unlikely(!__pyx_tuple__500)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__500);
  __Pyx_GIVEREF(__pyx_tuple__500);

  /* "ultron/tradingday/Calendar.pyx":674
 *     Date(2021, 1, 1),
 *     Date(2021, 2, 11),
 *     Date(2021, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 15),
 *     Date(2021, 2, 16),
 */
  __pyx_tuple__501 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_2, __pyx_int_12); if (unlikely(!__pyx_tuple__501)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__501);
  __Pyx_GIVEREF(__pyx_tuple__501);

  /* "ultron/tradingday/Calendar.pyx":675
 *     Date(2021, 2, 11),
 *     Date(2021, 2, 12),
 *     Date(2021, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 16),
 *     Date(2021, 2, 17),
 */
  __pyx_tuple__502 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_2, __pyx_int_15); if (unlikely(!__pyx_tuple__502)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__502);
  __Pyx_GIVEREF(__pyx_tuple__502);

  /* "ultron/tradingday/Calendar.pyx":676
 *     Date(2021, 2, 12),
 *     Date(2021, 2, 15),
 *     Date(2021, 2, 16),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 17),
 *     Date(2021, 4, 5),
 */
  __pyx_tuple__503 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_2, __pyx_int_16); if (unlikely(!__pyx_tuple__503)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__503);
  __Pyx_GIVEREF(__pyx_tuple__503);

  /* "ultron/tradingday/Calendar.pyx":677
 *     Date(2021, 2, 15),
 *     Date(2021, 2, 16),
 *     Date(2021, 2, 17),             # <<<<<<<<<<<<<<
 *     Date(2021, 4, 5),
 *     Date(2021, 5, 3),
 */
  __pyx_tuple__504 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_2, __pyx_int_17); if (unlikely(!__pyx_tuple__504)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__504);
  __Pyx_GIVEREF(__pyx_tuple__504);

  /* "ultron/tradingday/Calendar.pyx":678
 *     Date(2021, 2, 16),
 *     Date(2021, 2, 17),
 *     Date(2021, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2021, 5, 3),
 *     Date(2021, 5, 4),
 */
  __pyx_tuple__505 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_4, __pyx_int_5); if (unlikely(!__pyx_tuple__505)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__505);
  __Pyx_GIVEREF(__pyx_tuple__505);

  /* "ultron/tradingday/Calendar.pyx":679
 *     Date(2021, 2, 17),
 *     Date(2021, 4, 5),
 *     Date(2021, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2021, 5, 4),
 *     Date(2021, 5, 5),
 */
  __pyx_tuple__506 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__506)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__506);
  __Pyx_GIVEREF(__pyx_tuple__506);

  /* "ultron/tradingday/Calendar.pyx":680
 *     Date(2021, 4, 5),
 *     Date(2021, 5, 3),
 *     Date(2021, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2021, 5, 5),
 *     Date(2021, 6, 14),
 */
  __pyx_tuple__507 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_5, __pyx_int_4); if (unlikely(!__pyx_tuple__507)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__507);
  __Pyx_GIVEREF(__pyx_tuple__507);

  /* "ultron/tradingday/Calendar.pyx":681
 *     Date(2021, 5, 3),
 *     Date(2021, 5, 4),
 *     Date(2021, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2021, 6, 14),
 *     Date(2021, 9, 20),
 */
  __pyx_tuple__508 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_5, __pyx_int_5); if (unlikely(!__pyx_tuple__508)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__508);
  __Pyx_GIVEREF(__pyx_tuple__508);

  /* "ultron/tradingday/Calendar.pyx":682
 *     Date(2021, 5, 4),
 *     Date(2021, 5, 5),
 *     Date(2021, 6, 14),             # <<<<<<<<<<<<<<
 *     Date(2021, 9, 20),
 *     Date(2021, 9, 21),
 */
  __pyx_tuple__509 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_6, __pyx_int_14); if (unlikely(!__pyx_tuple__509)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__509);
  __Pyx_GIVEREF(__pyx_tuple__509);

  /* "ultron/tradingday/Calendar.pyx":683
 *     Date(2021, 5, 5),
 *     Date(2021, 6, 14),
 *     Date(2021, 9, 20),             # <<<<<<<<<<<<<<
 *     Date(2021, 9, 21),
 *     Date(2021, 10, 1),
 */
  __pyx_tuple__510 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_9, __pyx_int_20); if (unlikely(!__pyx_tuple__510)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__510);
  __Pyx_GIVEREF(__pyx_tuple__510);

  /* "ultron/tradingday/Calendar.pyx":684
 *     Date(2021, 6, 14),
 *     Date(2021, 9, 20),
 *     Date(2021, 9, 21),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 1),
 *     Date(2021, 10, 4),
 */
  __pyx_tuple__511 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_9, __pyx_int_21); if (unlikely(!__pyx_tuple__511)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__511);
  __Pyx_GIVEREF(__pyx_tuple__511);

  /* "ultron/tradingday/Calendar.pyx":685
 *     Date(2021, 9, 20),
 *     Date(2021, 9, 21),
 *     Date(2021, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 4),
 *     Date(2021, 10, 5),
 */
  __pyx_tuple__512 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_10, __pyx_int_1); if (unlikely(!__pyx_tuple__512)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__512);
  __Pyx_GIVEREF(__pyx_tuple__512);

  /* "ultron/tradingday/Calendar.pyx":686
 *     Date(2021, 9, 21),
 *     Date(2021, 10, 1),
 *     Date(2021, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 5),
 *     Date(2021, 10, 6),
 */
  __pyx_tuple__513 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__513)) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__513);
  __Pyx_GIVEREF(__pyx_tuple__513);

  /* "ultron/tradingday/Calendar.pyx":687
 *     Date(2021, 10, 1),
 *     Date(2021, 10, 4),
 *     Date(2021, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 6),
 *     Date(2021, 10, 7),
 */
  __pyx_tuple__514 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__514)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__514);
  __Pyx_GIVEREF(__pyx_tuple__514);

  /* "ultron/tradingday/Calendar.pyx":688
 *     Date(2021, 10, 4),
 *     Date(2021, 10, 5),
 *     Date(2021, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 7),
 *     Date(2022, 1, 3),
 */
  __pyx_tuple__515 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__515)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__515);
  __Pyx_GIVEREF(__pyx_tuple__515);

  /* "ultron/tradingday/Calendar.pyx":689
 *     Date(2021, 10, 5),
 *     Date(2021, 10, 6),
 *     Date(2021, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2022, 1, 3),
 *     Date(2022, 1, 31),
 */
  __pyx_tuple__516 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__516)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__516);
  __Pyx_GIVEREF(__pyx_tuple__516);

  /* "ultron/tradingday/Calendar.pyx":690
 *     Date(2021, 10, 6),
 *     Date(2021, 10, 7),
 *     Date(2022, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 1, 31),
 *     Date(2022, 2, 1),
 */
  __pyx_tuple__517 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_1, __pyx_int_3); if (unlikely(!__pyx_tuple__517)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__517);
  __Pyx_GIVEREF(__pyx_tuple__517);

  /* "ultron/tradingday/Calendar.pyx":691
 *     Date(2021, 10, 7),
 *     Date(2022, 1, 3),
 *     Date(2022, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2022, 2, 1),
 *     Date(2022, 2, 2),
 */
  __pyx_tuple__518 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_1, __pyx_int_31); if (unlikely(!__pyx_tuple__518)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__518);
  __Pyx_GIVEREF(__pyx_tuple__518);

  /* "ultron/tradingday/Calendar.pyx":692
 *     Date(2022, 1, 3),
 *     Date(2022, 1, 31),
 *     Date(2022, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2022, 2, 2),
 *     Date(2022, 2, 3),
 */
  __pyx_tuple__519 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_2, __pyx_int_1); if (unlikely(!__pyx_tuple__519)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__519);
  __Pyx_GIVEREF(__pyx_tuple__519);

  /* "ultron/tradingday/Calendar.pyx":693
 *     Date(2022, 1, 31),
 *     Date(2022, 2, 1),
 *     Date(2022, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2022, 2, 3),
 *     Date(2022, 2, 4),
 */
  __pyx_tuple__520 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__520)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__520);
  __Pyx_GIVEREF(__pyx_tuple__520);

  /* "ultron/tradingday/Calendar.pyx":694
 *     Date(2022, 2, 1),
 *     Date(2022, 2, 2),
 *     Date(2022, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 2, 4),
 *     Date(2022, 4, 4),
 */
  __pyx_tuple__521 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__521)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__521);
  __Pyx_GIVEREF(__pyx_tuple__521);

  /* "ultron/tradingday/Calendar.pyx":695
 *     Date(2022, 2, 2),
 *     Date(2022, 2, 3),
 *     Date(2022, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2022, 4, 4),
 *     Date(2022, 4, 5),
 */
  __pyx_tuple__522 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__522)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__522);
  __Pyx_GIVEREF(__pyx_tuple__522);

  /* "ultron/tradingday/Calendar.pyx":696
 *     Date(2022, 2, 3),
 *     Date(2022, 2, 4),
 *     Date(2022, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2022, 4, 5),
 *     Date(2022, 5, 2),
 */
  __pyx_tuple__523 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_4, __pyx_int_4); if (unlikely(!__pyx_tuple__523)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__523);
  __Pyx_GIVEREF(__pyx_tuple__523);

  /* "ultron/tradingday/Calendar.pyx":697
 *     Date(2022, 2, 4),
 *     Date(2022, 4, 4),
 *     Date(2022, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2022, 5, 2),
 *     Date(2022, 5, 3),
 */
  __pyx_tuple__524 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_4, __pyx_int_5); if (unlikely(!__pyx_tuple__524)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__524);
  __Pyx_GIVEREF(__pyx_tuple__524);

  /* "ultron/tradingday/Calendar.pyx":698
 *     Date(2022, 4, 4),
 *     Date(2022, 4, 5),
 *     Date(2022, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2022, 5, 3),
 *     Date(2022, 5, 4),
 */
  __pyx_tuple__525 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__525)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__525);
  __Pyx_GIVEREF(__pyx_tuple__525);

  /* "ultron/tradingday/Calendar.pyx":699
 *     Date(2022, 4, 5),
 *     Date(2022, 5, 2),
 *     Date(2022, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 5, 4),
 *     Date(2022, 6, 3),
 */
  __pyx_tuple__526 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__526)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__526);
  __Pyx_GIVEREF(__pyx_tuple__526);

  /* "ultron/tradingday/Calendar.pyx":700
 *     Date(2022, 5, 2),
 *     Date(2022, 5, 3),
 *     Date(2022, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2022, 6, 3),
 *     Date(2022, 9, 12),
 */
  __pyx_tuple__527 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_5, __pyx_int_4); if (unlikely(!__pyx_tuple__527)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__527);
  __Pyx_GIVEREF(__pyx_tuple__527);

  /* "ultron/tradingday/Calendar.pyx":701
 *     Date(2022, 5, 3),
 *     Date(2022, 5, 4),
 *     Date(2022, 6, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 9, 12),
 *     Date(2022, 10, 3),
 */
  __pyx_tuple__528 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_6, __pyx_int_3); if (unlikely(!__pyx_tuple__528)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__528);
  __Pyx_GIVEREF(__pyx_tuple__528);

  /* "ultron/tradingday/Calendar.pyx":702
 *     Date(2022, 5, 4),
 *     Date(2022, 6, 3),
 *     Date(2022, 9, 12),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 3),
 *     Date(2022, 10, 4),
 */
  __pyx_tuple__529 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_9, __pyx_int_12); if (unlikely(!__pyx_tuple__529)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__529);
  __Pyx_GIVEREF(__pyx_tuple__529);

  /* "ultron/tradingday/Calendar.pyx":703
 *     Date(2022, 6, 3),
 *     Date(2022, 9, 12),
 *     Date(2022, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 4),
 *     Date(2022, 10, 5),
 */
  __pyx_tuple__530 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__530)) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__530);
  __Pyx_GIVEREF(__pyx_tuple__530);

  /* "ultron/tradingday/Calendar.pyx":704
 *     Date(2022, 9, 12),
 *     Date(2022, 10, 3),
 *     Date(2022, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 5),
 *     Date(2022, 10, 6),
 */
  __pyx_tuple__531 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__531)) __PYX_ERR(0, 704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__531);
  __Pyx_GIVEREF(__pyx_tuple__531);

  /* "ultron/tradingday/Calendar.pyx":705
 *     Date(2022, 10, 3),
 *     Date(2022, 10, 4),
 *     Date(2022, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 6),
 *     Date(2022, 10, 7),
 */
  __pyx_tuple__532 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__532)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__532);
  __Pyx_GIVEREF(__pyx_tuple__532);

  /* "ultron/tradingday/Calendar.pyx":706
 *     Date(2022, 10, 4),
 *     Date(2022, 10, 5),
 *     Date(2022, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 7),
 *     Date(2023, 1, 2),
 */
  __pyx_tuple__533 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__533)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__533);
  __Pyx_GIVEREF(__pyx_tuple__533);

  /* "ultron/tradingday/Calendar.pyx":707
 *     Date(2022, 10, 5),
 *     Date(2022, 10, 6),
 *     Date(2022, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 2),
 *     Date(2023, 1, 23),
 */
  __pyx_tuple__534 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__534)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__534);
  __Pyx_GIVEREF(__pyx_tuple__534);

  /* "ultron/tradingday/Calendar.pyx":708
 *     Date(2022, 10, 6),
 *     Date(2022, 10, 7),
 *     Date(2023, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 23),
 *     Date(2023, 1, 24),
 */
  __pyx_tuple__535 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__535)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__535);
  __Pyx_GIVEREF(__pyx_tuple__535);

  /* "ultron/tradingday/Calendar.pyx":709
 *     Date(2022, 10, 7),
 *     Date(2023, 1, 2),
 *     Date(2023, 1, 23),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 24),
 *     Date(2023, 1, 25),
 */
  __pyx_tuple__536 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_1, __pyx_int_23); if (unlikely(!__pyx_tuple__536)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__536);
  __Pyx_GIVEREF(__pyx_tuple__536);

  /* "ultron/tradingday/Calendar.pyx":710
 *     Date(2023, 1, 2),
 *     Date(2023, 1, 23),
 *     Date(2023, 1, 24),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 25),
 *     Date(2023, 1, 26),
 */
  __pyx_tuple__537 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_1, __pyx_int_24); if (unlikely(!__pyx_tuple__537)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__537);
  __Pyx_GIVEREF(__pyx_tuple__537);

  /* "ultron/tradingday/Calendar.pyx":711
 *     Date(2023, 1, 23),
 *     Date(2023, 1, 24),
 *     Date(2023, 1, 25),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 26),
 *     Date(2023, 1, 27),
 */
  __pyx_tuple__538 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_1, __pyx_int_25); if (unlikely(!__pyx_tuple__538)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__538);
  __Pyx_GIVEREF(__pyx_tuple__538);

  /* "ultron/tradingday/Calendar.pyx":712
 *     Date(2023, 1, 24),
 *     Date(2023, 1, 25),
 *     Date(2023, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 27),
 *     Date(2023, 4, 5),
 */
  __pyx_tuple__539 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_1, __pyx_int_26); if (unlikely(!__pyx_tuple__539)) __PYX_ERR(0, 712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__539);
  __Pyx_GIVEREF(__pyx_tuple__539);

  /* "ultron/tradingday/Calendar.pyx":713
 *     Date(2023, 1, 25),
 *     Date(2023, 1, 26),
 *     Date(2023, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2023, 4, 5),
 *     Date(2023, 5, 1),
 */
  __pyx_tuple__540 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_1, __pyx_int_27); if (unlikely(!__pyx_tuple__540)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__540);
  __Pyx_GIVEREF(__pyx_tuple__540);

  /* "ultron/tradingday/Calendar.pyx":714
 *     Date(2023, 1, 26),
 *     Date(2023, 1, 27),
 *     Date(2023, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2023, 5, 1),
 *     Date(2023, 5, 2),
 */
  __pyx_tuple__541 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_4, __pyx_int_5); if (unlikely(!__pyx_tuple__541)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__541);
  __Pyx_GIVEREF(__pyx_tuple__541);

  /* "ultron/tradingday/Calendar.pyx":715
 *     Date(2023, 1, 27),
 *     Date(2023, 4, 5),
 *     Date(2023, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2023, 5, 2),
 *     Date(2023, 5, 3),
 */
  __pyx_tuple__542 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__542)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__542);
  __Pyx_GIVEREF(__pyx_tuple__542);

  /* "ultron/tradingday/Calendar.pyx":716
 *     Date(2023, 4, 5),
 *     Date(2023, 5, 1),
 *     Date(2023, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2023, 5, 3),
 *     Date(2023, 6, 22),
 */
  __pyx_tuple__543 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_5, __pyx_int_2); if (unlikely(!__pyx_tuple__543)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__543);
  __Pyx_GIVEREF(__pyx_tuple__543);

  /* "ultron/tradingday/Calendar.pyx":717
 *     Date(2023, 5, 1),
 *     Date(2023, 5, 2),
 *     Date(2023, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2023, 6, 22),
 *     Date(2023, 6, 23),
 */
  __pyx_tuple__544 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_5, __pyx_int_3); if (unlikely(!__pyx_tuple__544)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__544);
  __Pyx_GIVEREF(__pyx_tuple__544);

  /* "ultron/tradingday/Calendar.pyx":718
 *     Date(2023, 5, 2),
 *     Date(2023, 5, 3),
 *     Date(2023, 6, 22),             # <<<<<<<<<<<<<<
 *     Date(2023, 6, 23),
 *     Date(2023, 9, 29),
 */
  __pyx_tuple__545 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_6, __pyx_int_22); if (unlikely(!__pyx_tuple__545)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__545);
  __Pyx_GIVEREF(__pyx_tuple__545);

  /* "ultron/tradingday/Calendar.pyx":719
 *     Date(2023, 5, 3),
 *     Date(2023, 6, 22),
 *     Date(2023, 6, 23),             # <<<<<<<<<<<<<<
 *     Date(2023, 9, 29),
 *     Date(2023, 10, 2),
 */
  __pyx_tuple__546 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_6, __pyx_int_23); if (unlikely(!__pyx_tuple__546)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__546);
  __Pyx_GIVEREF(__pyx_tuple__546);

  /* "ultron/tradingday/Calendar.pyx":720
 *     Date(2023, 6, 22),
 *     Date(2023, 6, 23),
 *     Date(2023, 9, 29),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 2),
 *     Date(2023, 10, 3),
 */
  __pyx_tuple__547 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_9, __pyx_int_29); if (unlikely(!__pyx_tuple__547)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__547);
  __Pyx_GIVEREF(__pyx_tuple__547);

  /* "ultron/tradingday/Calendar.pyx":721
 *     Date(2023, 6, 23),
 *     Date(2023, 9, 29),
 *     Date(2023, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 3),
 *     Date(2023, 10, 4),
 */
  __pyx_tuple__548 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_10, __pyx_int_2); if (unlikely(!__pyx_tuple__548)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__548);
  __Pyx_GIVEREF(__pyx_tuple__548);

  /* "ultron/tradingday/Calendar.pyx":722
 *     Date(2023, 9, 29),
 *     Date(2023, 10, 2),
 *     Date(2023, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 4),
 *     Date(2023, 10, 5),
 */
  __pyx_tuple__549 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_10, __pyx_int_3); if (unlikely(!__pyx_tuple__549)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__549);
  __Pyx_GIVEREF(__pyx_tuple__549);

  /* "ultron/tradingday/Calendar.pyx":723
 *     Date(2023, 10, 2),
 *     Date(2023, 10, 3),
 *     Date(2023, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 5),
 *     Date(2023, 10, 6)
 */
  __pyx_tuple__550 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_10, __pyx_int_4); if (unlikely(!__pyx_tuple__550)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__550);
  __Pyx_GIVEREF(__pyx_tuple__550);

  /* "ultron/tradingday/Calendar.pyx":724
 *     Date(2023, 10, 3),
 *     Date(2023, 10, 4),
 *     Date(2023, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 6)
 * }
 */
  __pyx_tuple__551 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_10, __pyx_int_5); if (unlikely(!__pyx_tuple__551)) __PYX_ERR(0, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__551);
  __Pyx_GIVEREF(__pyx_tuple__551);

  /* "ultron/tradingday/Calendar.pyx":725
 *     Date(2023, 10, 4),
 *     Date(2023, 10, 5),
 *     Date(2023, 10, 6)             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_tuple__552 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_10, __pyx_int_6); if (unlikely(!__pyx_tuple__552)) __PYX_ERR(0, 725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__552);
  __Pyx_GIVEREF(__pyx_tuple__552);

  /* "ultron/tradingday/Calendar.pyx":875
 *     Date.westernStyle(10, Months.October, 2020),
 *     # 2021
 *     Date(2021, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 20),
 *     Date(2021, 4, 25),
 */
  __pyx_tuple__553 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__553)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__553);
  __Pyx_GIVEREF(__pyx_tuple__553);

  /* "ultron/tradingday/Calendar.pyx":876
 *     # 2021
 *     Date(2021, 2, 7),
 *     Date(2021, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2021, 4, 25),
 *     Date(2021, 5, 8),
 */
  __pyx_tuple__554 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_2, __pyx_int_20); if (unlikely(!__pyx_tuple__554)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__554);
  __Pyx_GIVEREF(__pyx_tuple__554);

  /* "ultron/tradingday/Calendar.pyx":877
 *     Date(2021, 2, 7),
 *     Date(2021, 2, 20),
 *     Date(2021, 4, 25),             # <<<<<<<<<<<<<<
 *     Date(2021, 5, 8),
 *     Date(2021, 9, 18),
 */
  __pyx_tuple__555 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_4, __pyx_int_25); if (unlikely(!__pyx_tuple__555)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__555);
  __Pyx_GIVEREF(__pyx_tuple__555);

  /* "ultron/tradingday/Calendar.pyx":878
 *     Date(2021, 2, 20),
 *     Date(2021, 4, 25),
 *     Date(2021, 5, 8),             # <<<<<<<<<<<<<<
 *     Date(2021, 9, 18),
 *     Date(2021, 9, 26),
 */
  __pyx_tuple__556 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_5, __pyx_int_8); if (unlikely(!__pyx_tuple__556)) __PYX_ERR(0, 878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__556);
  __Pyx_GIVEREF(__pyx_tuple__556);

  /* "ultron/tradingday/Calendar.pyx":879
 *     Date(2021, 4, 25),
 *     Date(2021, 5, 8),
 *     Date(2021, 9, 18),             # <<<<<<<<<<<<<<
 *     Date(2021, 9, 26),
 *     Date(2021, 10, 9),
 */
  __pyx_tuple__557 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_9, __pyx_int_18); if (unlikely(!__pyx_tuple__557)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__557);
  __Pyx_GIVEREF(__pyx_tuple__557);

  /* "ultron/tradingday/Calendar.pyx":880
 *     Date(2021, 5, 8),
 *     Date(2021, 9, 18),
 *     Date(2021, 9, 26),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 9),
 *     # 2022
 */
  __pyx_tuple__558 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_9, __pyx_int_26); if (unlikely(!__pyx_tuple__558)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__558);
  __Pyx_GIVEREF(__pyx_tuple__558);

  /* "ultron/tradingday/Calendar.pyx":881
 *     Date(2021, 9, 18),
 *     Date(2021, 9, 26),
 *     Date(2021, 10, 9),             # <<<<<<<<<<<<<<
 *     # 2022
 *     Date(2022, 1, 29),
 */
  __pyx_tuple__559 = PyTuple_Pack(3, __pyx_int_2021, __pyx_int_10, __pyx_int_9); if (unlikely(!__pyx_tuple__559)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__559);
  __Pyx_GIVEREF(__pyx_tuple__559);

  /* "ultron/tradingday/Calendar.pyx":883
 *     Date(2021, 10, 9),
 *     # 2022
 *     Date(2022, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2022, 1, 30),
 *     Date(2022, 4, 2),
 */
  __pyx_tuple__560 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_1, __pyx_int_29); if (unlikely(!__pyx_tuple__560)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__560);
  __Pyx_GIVEREF(__pyx_tuple__560);

  /* "ultron/tradingday/Calendar.pyx":884
 *     # 2022
 *     Date(2022, 1, 29),
 *     Date(2022, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2022, 4, 2),
 *     Date(2022, 4, 24),
 */
  __pyx_tuple__561 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_1, __pyx_int_30); if (unlikely(!__pyx_tuple__561)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__561);
  __Pyx_GIVEREF(__pyx_tuple__561);

  /* "ultron/tradingday/Calendar.pyx":885
 *     Date(2022, 1, 29),
 *     Date(2022, 1, 30),
 *     Date(2022, 4, 2),             # <<<<<<<<<<<<<<
 *     Date(2022, 4, 24),
 *     Date(2022, 5, 7),
 */
  __pyx_tuple__562 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_4, __pyx_int_2); if (unlikely(!__pyx_tuple__562)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__562);
  __Pyx_GIVEREF(__pyx_tuple__562);

  /* "ultron/tradingday/Calendar.pyx":886
 *     Date(2022, 1, 30),
 *     Date(2022, 4, 2),
 *     Date(2022, 4, 24),             # <<<<<<<<<<<<<<
 *     Date(2022, 5, 7),
 *     Date(2022, 10, 8),
 */
  __pyx_tuple__563 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_4, __pyx_int_24); if (unlikely(!__pyx_tuple__563)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__563);
  __Pyx_GIVEREF(__pyx_tuple__563);

  /* "ultron/tradingday/Calendar.pyx":887
 *     Date(2022, 4, 2),
 *     Date(2022, 4, 24),
 *     Date(2022, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 8),
 *     Date(2022, 10, 9),
 */
  __pyx_tuple__564 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_5, __pyx_int_7); if (unlikely(!__pyx_tuple__564)) __PYX_ERR(0, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__564);
  __Pyx_GIVEREF(__pyx_tuple__564);

  /* "ultron/tradingday/Calendar.pyx":888
 *     Date(2022, 4, 24),
 *     Date(2022, 5, 7),
 *     Date(2022, 10, 8),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 9),
 * 
 */
  __pyx_tuple__565 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_10, __pyx_int_8); if (unlikely(!__pyx_tuple__565)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__565);
  __Pyx_GIVEREF(__pyx_tuple__565);

  /* "ultron/tradingday/Calendar.pyx":889
 *     Date(2022, 5, 7),
 *     Date(2022, 10, 8),
 *     Date(2022, 10, 9),             # <<<<<<<<<<<<<<
 * 
 *     # 2023
 */
  __pyx_tuple__566 = PyTuple_Pack(3, __pyx_int_2022, __pyx_int_10, __pyx_int_9); if (unlikely(!__pyx_tuple__566)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__566);
  __Pyx_GIVEREF(__pyx_tuple__566);

  /* "ultron/tradingday/Calendar.pyx":892
 * 
 *     # 2023
 *     Date(2023, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 29),
 *     Date(2023, 4, 23),
 */
  __pyx_tuple__567 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_1, __pyx_int_28); if (unlikely(!__pyx_tuple__567)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__567);
  __Pyx_GIVEREF(__pyx_tuple__567);

  /* "ultron/tradingday/Calendar.pyx":893
 *     # 2023
 *     Date(2023, 1, 28),
 *     Date(2023, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2023, 4, 23),
 *     Date(2023, 5, 6),
 */
  __pyx_tuple__568 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_1, __pyx_int_29); if (unlikely(!__pyx_tuple__568)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__568);
  __Pyx_GIVEREF(__pyx_tuple__568);

  /* "ultron/tradingday/Calendar.pyx":894
 *     Date(2023, 1, 28),
 *     Date(2023, 1, 29),
 *     Date(2023, 4, 23),             # <<<<<<<<<<<<<<
 *     Date(2023, 5, 6),
 *     Date(2023, 6, 25),
 */
  __pyx_tuple__569 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_4, __pyx_int_23); if (unlikely(!__pyx_tuple__569)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__569);
  __Pyx_GIVEREF(__pyx_tuple__569);

  /* "ultron/tradingday/Calendar.pyx":895
 *     Date(2023, 1, 29),
 *     Date(2023, 4, 23),
 *     Date(2023, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2023, 6, 25),
 *     Date(2023, 10, 7),
 */
  __pyx_tuple__570 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_5, __pyx_int_6); if (unlikely(!__pyx_tuple__570)) __PYX_ERR(0, 895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__570);
  __Pyx_GIVEREF(__pyx_tuple__570);

  /* "ultron/tradingday/Calendar.pyx":896
 *     Date(2023, 4, 23),
 *     Date(2023, 5, 6),
 *     Date(2023, 6, 25),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 7),
 *     Date(2023, 10, 8)
 */
  __pyx_tuple__571 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_6, __pyx_int_25); if (unlikely(!__pyx_tuple__571)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__571);
  __Pyx_GIVEREF(__pyx_tuple__571);

  /* "ultron/tradingday/Calendar.pyx":897
 *     Date(2023, 5, 6),
 *     Date(2023, 6, 25),
 *     Date(2023, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 8)
 * }
 */
  __pyx_tuple__572 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_10, __pyx_int_7); if (unlikely(!__pyx_tuple__572)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__572);
  __Pyx_GIVEREF(__pyx_tuple__572);

  /* "ultron/tradingday/Calendar.pyx":898
 *     Date(2023, 6, 25),
 *     Date(2023, 10, 7),
 *     Date(2023, 10, 8)             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_tuple__573 = PyTuple_Pack(3, __pyx_int_2023, __pyx_int_10, __pyx_int_8); if (unlikely(!__pyx_tuple__573)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__573);
  __Pyx_GIVEREF(__pyx_tuple__573);

  /* "(tree fragment)":1
 * def __pyx_unpickle_CalendarImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__574 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__574)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__574);
  __Pyx_GIVEREF(__pyx_tuple__574);
  __pyx_codeobj__575 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__574, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CalendarImpl, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__575)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__576 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__576)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__576);
  __Pyx_GIVEREF(__pyx_tuple__576);
  __pyx_codeobj__577 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__576, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ChinaSseImpl, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__577)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__578 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__578)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__578);
  __Pyx_GIVEREF(__pyx_tuple__578);
  __pyx_codeobj__579 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__578, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ChinaIBImpl, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__579)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__580 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__580)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__580);
  __Pyx_GIVEREF(__pyx_tuple__580);
  __pyx_codeobj__581 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__580, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_NullCalendar, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__581)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__582 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__582)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__582);
  __Pyx_GIVEREF(__pyx_tuple__582);
  __pyx_codeobj__583 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__582, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ChinaCFFEXImpl, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__583)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__584 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__584)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__584);
  __Pyx_GIVEREF(__pyx_tuple__584);
  __pyx_codeobj__585 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__584, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_WestenImpl, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__585)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__586 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__586)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__586);
  __Pyx_GIVEREF(__pyx_tuple__586);
  __pyx_codeobj__587 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__586, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_TargetImpl, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__587)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyString_Type_lower.type = (PyObject*)&PyString_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_11 = PyInt_FromLong(11); if (unlikely(!__pyx_int_11)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_13 = PyInt_FromLong(13); if (unlikely(!__pyx_int_13)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_14 = PyInt_FromLong(14); if (unlikely(!__pyx_int_14)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_17 = PyInt_FromLong(17); if (unlikely(!__pyx_int_17)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_18 = PyInt_FromLong(18); if (unlikely(!__pyx_int_18)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_19 = PyInt_FromLong(19); if (unlikely(!__pyx_int_19)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_20 = PyInt_FromLong(20); if (unlikely(!__pyx_int_20)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_21 = PyInt_FromLong(21); if (unlikely(!__pyx_int_21)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_22 = PyInt_FromLong(22); if (unlikely(!__pyx_int_22)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_23 = PyInt_FromLong(23); if (unlikely(!__pyx_int_23)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_24 = PyInt_FromLong(24); if (unlikely(!__pyx_int_24)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_25 = PyInt_FromLong(25); if (unlikely(!__pyx_int_25)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_26 = PyInt_FromLong(26); if (unlikely(!__pyx_int_26)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_27 = PyInt_FromLong(27); if (unlikely(!__pyx_int_27)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_28 = PyInt_FromLong(28); if (unlikely(!__pyx_int_28)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_29 = PyInt_FromLong(29); if (unlikely(!__pyx_int_29)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_30 = PyInt_FromLong(30); if (unlikely(!__pyx_int_30)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_31 = PyInt_FromLong(31); if (unlikely(!__pyx_int_31)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1991 = PyInt_FromLong(1991); if (unlikely(!__pyx_int_1991)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1992 = PyInt_FromLong(1992); if (unlikely(!__pyx_int_1992)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1993 = PyInt_FromLong(1993); if (unlikely(!__pyx_int_1993)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1994 = PyInt_FromLong(1994); if (unlikely(!__pyx_int_1994)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1995 = PyInt_FromLong(1995); if (unlikely(!__pyx_int_1995)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1996 = PyInt_FromLong(1996); if (unlikely(!__pyx_int_1996)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1997 = PyInt_FromLong(1997); if (unlikely(!__pyx_int_1997)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1998 = PyInt_FromLong(1998); if (unlikely(!__pyx_int_1998)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1999 = PyInt_FromLong(1999); if (unlikely(!__pyx_int_1999)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2000 = PyInt_FromLong(2000); if (unlikely(!__pyx_int_2000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2001 = PyInt_FromLong(2001); if (unlikely(!__pyx_int_2001)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2002 = PyInt_FromLong(2002); if (unlikely(!__pyx_int_2002)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2003 = PyInt_FromLong(2003); if (unlikely(!__pyx_int_2003)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2004 = PyInt_FromLong(2004); if (unlikely(!__pyx_int_2004)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2005 = PyInt_FromLong(2005); if (unlikely(!__pyx_int_2005)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2006 = PyInt_FromLong(2006); if (unlikely(!__pyx_int_2006)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2007 = PyInt_FromLong(2007); if (unlikely(!__pyx_int_2007)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2008 = PyInt_FromLong(2008); if (unlikely(!__pyx_int_2008)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2009 = PyInt_FromLong(2009); if (unlikely(!__pyx_int_2009)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2010 = PyInt_FromLong(2010); if (unlikely(!__pyx_int_2010)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2011 = PyInt_FromLong(2011); if (unlikely(!__pyx_int_2011)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2012 = PyInt_FromLong(2012); if (unlikely(!__pyx_int_2012)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2013 = PyInt_FromLong(2013); if (unlikely(!__pyx_int_2013)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2014 = PyInt_FromLong(2014); if (unlikely(!__pyx_int_2014)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2015 = PyInt_FromLong(2015); if (unlikely(!__pyx_int_2015)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2016 = PyInt_FromLong(2016); if (unlikely(!__pyx_int_2016)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2017 = PyInt_FromLong(2017); if (unlikely(!__pyx_int_2017)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2018 = PyInt_FromLong(2018); if (unlikely(!__pyx_int_2018)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2019 = PyInt_FromLong(2019); if (unlikely(!__pyx_int_2019)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2020 = PyInt_FromLong(2020); if (unlikely(!__pyx_int_2020)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2021 = PyInt_FromLong(2021); if (unlikely(!__pyx_int_2021)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2022 = PyInt_FromLong(2022); if (unlikely(!__pyx_int_2022)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2023 = PyInt_FromLong(2023); if (unlikely(!__pyx_int_2023)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_6ultron_10tradingday_8Calendar_sse_holDays = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_6ultron_10tradingday_8Calendar_sse_working_weekends = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_6ultron_10tradingday_8Calendar_ib_working_weekends = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_6ultron_10tradingday_8Calendar__sseImpl = ((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *)Py_None); Py_INCREF(Py_None);
  __pyx_v_6ultron_10tradingday_8Calendar__holDict = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6ultron_10tradingday_8Calendar_CalendarImpl = &__pyx_vtable_6ultron_10tradingday_8Calendar_CalendarImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_CalendarImpl.isBizDay = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *))__pyx_f_6ultron_10tradingday_8Calendar_12CalendarImpl_isBizDay;
  __pyx_vtable_6ultron_10tradingday_8Calendar_CalendarImpl.isWeekEnd = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, int))__pyx_f_6ultron_10tradingday_8Calendar_12CalendarImpl_isWeekEnd;
  if (PyType_Ready(&__pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl.tp_dictoffset && __pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl.tp_dict, __pyx_vtabptr_6ultron_10tradingday_8Calendar_CalendarImpl) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CalendarImpl, (PyObject *)&__pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl = &__pyx_type_6ultron_10tradingday_8Calendar_CalendarImpl;
  __pyx_vtabptr_6ultron_10tradingday_8Calendar_Calendar = &__pyx_vtable_6ultron_10tradingday_8Calendar_Calendar;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.isBizDay = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isBizDay;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.isHoliday = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isHoliday;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.isWeekEnd = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, int, int __pyx_skip_dispatch))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isWeekEnd;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.isEndOfMonth = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_isEndOfMonth;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.endOfMonth = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_endOfMonth;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.bizDaysBetween = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween *__pyx_optional_args))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_bizDaysBetween;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.adjustDate = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_adjustDate *__pyx_optional_args))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_adjustDate;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.advanceDate = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, struct __pyx_obj_6ultron_10tradingday_6Period_Period *, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_advanceDate *__pyx_optional_args))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_advanceDate;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.holDatesList = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch, struct __pyx_opt_args_6ultron_10tradingday_8Calendar_8Calendar_holDatesList *__pyx_optional_args))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_holDatesList;
  __pyx_vtable_6ultron_10tradingday_8Calendar_Calendar.bizDatesList = (PyObject *(*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_Calendar *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch))__pyx_f_6ultron_10tradingday_8Calendar_8Calendar_bizDatesList;
  if (PyType_Ready(&__pyx_type_6ultron_10tradingday_8Calendar_Calendar) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __pyx_type_6ultron_10tradingday_8Calendar_Calendar.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ultron_10tradingday_8Calendar_Calendar.tp_dictoffset && __pyx_type_6ultron_10tradingday_8Calendar_Calendar.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ultron_10tradingday_8Calendar_Calendar.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6ultron_10tradingday_8Calendar_Calendar.tp_dict, __pyx_vtabptr_6ultron_10tradingday_8Calendar_Calendar) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Calendar, (PyObject *)&__pyx_type_6ultron_10tradingday_8Calendar_Calendar) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __pyx_ptype_6ultron_10tradingday_8Calendar_Calendar = &__pyx_type_6ultron_10tradingday_8Calendar_Calendar;
  __pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaSseImpl = &__pyx_vtable_6ultron_10tradingday_8Calendar_ChinaSseImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaSseImpl.__pyx_base = *__pyx_vtabptr_6ultron_10tradingday_8Calendar_CalendarImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaSseImpl.__pyx_base.isBizDay = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *))__pyx_f_6ultron_10tradingday_8Calendar_12ChinaSseImpl_isBizDay;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaSseImpl.__pyx_base.isWeekEnd = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, int))__pyx_f_6ultron_10tradingday_8Calendar_12ChinaSseImpl_isWeekEnd;
  __pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl.tp_base = __pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl;
  if (PyType_Ready(&__pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl) < 0) __PYX_ERR(0, 735, __pyx_L1_error)
  __pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl.tp_dictoffset && __pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl.tp_dict, __pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaSseImpl) < 0) __PYX_ERR(0, 735, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ChinaSseImpl, (PyObject *)&__pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl) < 0) __PYX_ERR(0, 735, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl) < 0) __PYX_ERR(0, 735, __pyx_L1_error)
  __pyx_ptype_6ultron_10tradingday_8Calendar_ChinaSseImpl = &__pyx_type_6ultron_10tradingday_8Calendar_ChinaSseImpl;
  __pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaIBImpl = &__pyx_vtable_6ultron_10tradingday_8Calendar_ChinaIBImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaIBImpl.__pyx_base = *__pyx_vtabptr_6ultron_10tradingday_8Calendar_CalendarImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaIBImpl.__pyx_base.isBizDay = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *))__pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isBizDay__pyx_wrap_1;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaIBImpl.__pyx_base.isWeekEnd = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, int))__pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isWeekEnd__pyx_wrap_1;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaIBImpl.isBizDay = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *, int __pyx_skip_dispatch))__pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isBizDay;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaIBImpl.isWeekEnd = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaIBImpl *, int, int __pyx_skip_dispatch))__pyx_f_6ultron_10tradingday_8Calendar_11ChinaIBImpl_isWeekEnd;
  __pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl.tp_base = __pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl;
  if (PyType_Ready(&__pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl) < 0) __PYX_ERR(0, 903, __pyx_L1_error)
  __pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl.tp_dictoffset && __pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl.tp_dict, __pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaIBImpl) < 0) __PYX_ERR(0, 903, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ChinaIBImpl, (PyObject *)&__pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl) < 0) __PYX_ERR(0, 903, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl) < 0) __PYX_ERR(0, 903, __pyx_L1_error)
  __pyx_ptype_6ultron_10tradingday_8Calendar_ChinaIBImpl = &__pyx_type_6ultron_10tradingday_8Calendar_ChinaIBImpl;
  __pyx_vtabptr_6ultron_10tradingday_8Calendar_NullCalendar = &__pyx_vtable_6ultron_10tradingday_8Calendar_NullCalendar;
  __pyx_vtable_6ultron_10tradingday_8Calendar_NullCalendar.__pyx_base = *__pyx_vtabptr_6ultron_10tradingday_8Calendar_CalendarImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_NullCalendar.__pyx_base.isBizDay = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *))__pyx_f_6ultron_10tradingday_8Calendar_12NullCalendar_isBizDay;
  __pyx_vtable_6ultron_10tradingday_8Calendar_NullCalendar.__pyx_base.isWeekEnd = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, int))__pyx_f_6ultron_10tradingday_8Calendar_12NullCalendar_isWeekEnd;
  __pyx_type_6ultron_10tradingday_8Calendar_NullCalendar.tp_base = __pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl;
  if (PyType_Ready(&__pyx_type_6ultron_10tradingday_8Calendar_NullCalendar) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
  __pyx_type_6ultron_10tradingday_8Calendar_NullCalendar.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ultron_10tradingday_8Calendar_NullCalendar.tp_dictoffset && __pyx_type_6ultron_10tradingday_8Calendar_NullCalendar.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ultron_10tradingday_8Calendar_NullCalendar.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6ultron_10tradingday_8Calendar_NullCalendar.tp_dict, __pyx_vtabptr_6ultron_10tradingday_8Calendar_NullCalendar) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_NullCalendar, (PyObject *)&__pyx_type_6ultron_10tradingday_8Calendar_NullCalendar) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6ultron_10tradingday_8Calendar_NullCalendar) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
  __pyx_ptype_6ultron_10tradingday_8Calendar_NullCalendar = &__pyx_type_6ultron_10tradingday_8Calendar_NullCalendar;
  __pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl = &__pyx_vtable_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl.__pyx_base = *__pyx_vtabptr_6ultron_10tradingday_8Calendar_CalendarImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl.__pyx_base.isBizDay = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *))__pyx_f_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_isBizDay;
  __pyx_vtable_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl.__pyx_base.isWeekEnd = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, int))__pyx_f_6ultron_10tradingday_8Calendar_14ChinaCFFEXImpl_isWeekEnd;
  __pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl.tp_base = __pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl;
  if (PyType_Ready(&__pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl) < 0) __PYX_ERR(0, 931, __pyx_L1_error)
  __pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl.tp_dictoffset && __pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl.tp_dict, __pyx_vtabptr_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl) < 0) __PYX_ERR(0, 931, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ChinaCFFEXImpl, (PyObject *)&__pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl) < 0) __PYX_ERR(0, 931, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl) < 0) __PYX_ERR(0, 931, __pyx_L1_error)
  __pyx_ptype_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl = &__pyx_type_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl;
  __pyx_vtabptr_6ultron_10tradingday_8Calendar_WestenImpl = &__pyx_vtable_6ultron_10tradingday_8Calendar_WestenImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_WestenImpl.__pyx_base = *__pyx_vtabptr_6ultron_10tradingday_8Calendar_CalendarImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_WestenImpl.__pyx_base.isWeekEnd = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, int))__pyx_f_6ultron_10tradingday_8Calendar_10WestenImpl_isWeekEnd;
  __pyx_vtable_6ultron_10tradingday_8Calendar_WestenImpl.easterMonday = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_WestenImpl *, int))__pyx_f_6ultron_10tradingday_8Calendar_10WestenImpl_easterMonday;
  __pyx_type_6ultron_10tradingday_8Calendar_WestenImpl.tp_base = __pyx_ptype_6ultron_10tradingday_8Calendar_CalendarImpl;
  if (PyType_Ready(&__pyx_type_6ultron_10tradingday_8Calendar_WestenImpl) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  __pyx_type_6ultron_10tradingday_8Calendar_WestenImpl.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ultron_10tradingday_8Calendar_WestenImpl.tp_dictoffset && __pyx_type_6ultron_10tradingday_8Calendar_WestenImpl.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ultron_10tradingday_8Calendar_WestenImpl.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6ultron_10tradingday_8Calendar_WestenImpl.tp_dict, __pyx_vtabptr_6ultron_10tradingday_8Calendar_WestenImpl) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_WestenImpl, (PyObject *)&__pyx_type_6ultron_10tradingday_8Calendar_WestenImpl) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6ultron_10tradingday_8Calendar_WestenImpl) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  __pyx_ptype_6ultron_10tradingday_8Calendar_WestenImpl = &__pyx_type_6ultron_10tradingday_8Calendar_WestenImpl;
  __pyx_vtabptr_6ultron_10tradingday_8Calendar_TargetImpl = &__pyx_vtable_6ultron_10tradingday_8Calendar_TargetImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_TargetImpl.__pyx_base = *__pyx_vtabptr_6ultron_10tradingday_8Calendar_WestenImpl;
  __pyx_vtable_6ultron_10tradingday_8Calendar_TargetImpl.__pyx_base.__pyx_base.isBizDay = (int (*)(struct __pyx_obj_6ultron_10tradingday_8Calendar_CalendarImpl *, struct __pyx_obj_6ultron_10tradingday_4Date_Date *))__pyx_f_6ultron_10tradingday_8Calendar_10TargetImpl_isBizDay;
  __pyx_type_6ultron_10tradingday_8Calendar_TargetImpl.tp_base = __pyx_ptype_6ultron_10tradingday_8Calendar_WestenImpl;
  if (PyType_Ready(&__pyx_type_6ultron_10tradingday_8Calendar_TargetImpl) < 0) __PYX_ERR(0, 986, __pyx_L1_error)
  __pyx_type_6ultron_10tradingday_8Calendar_TargetImpl.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6ultron_10tradingday_8Calendar_TargetImpl.tp_dictoffset && __pyx_type_6ultron_10tradingday_8Calendar_TargetImpl.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6ultron_10tradingday_8Calendar_TargetImpl.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6ultron_10tradingday_8Calendar_TargetImpl.tp_dict, __pyx_vtabptr_6ultron_10tradingday_8Calendar_TargetImpl) < 0) __PYX_ERR(0, 986, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TargetImpl, (PyObject *)&__pyx_type_6ultron_10tradingday_8Calendar_TargetImpl) < 0) __PYX_ERR(0, 986, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6ultron_10tradingday_8Calendar_TargetImpl) < 0) __PYX_ERR(0, 986, __pyx_L1_error)
  __pyx_ptype_6ultron_10tradingday_8Calendar_TargetImpl = &__pyx_type_6ultron_10tradingday_8Calendar_TargetImpl;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ultron.tradingday.Date"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6ultron_10tradingday_4Date_Date = __Pyx_ImportType(__pyx_t_1, "ultron.tradingday.Date", "Date", sizeof(struct __pyx_obj_6ultron_10tradingday_4Date_Date), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6ultron_10tradingday_4Date_Date) __PYX_ERR(3, 3, __pyx_L1_error)
  __pyx_vtabptr_6ultron_10tradingday_4Date_Date = (struct __pyx_vtabstruct_6ultron_10tradingday_4Date_Date*)__Pyx_GetVtable(__pyx_ptype_6ultron_10tradingday_4Date_Date->tp_dict); if (unlikely(!__pyx_vtabptr_6ultron_10tradingday_4Date_Date)) __PYX_ERR(3, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ultron.tradingday.Period"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6ultron_10tradingday_6Period_Period = __Pyx_ImportType(__pyx_t_1, "ultron.tradingday.Period", "Period", sizeof(struct __pyx_obj_6ultron_10tradingday_6Period_Period), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6ultron_10tradingday_6Period_Period) __PYX_ERR(4, 3, __pyx_L1_error)
  __pyx_vtabptr_6ultron_10tradingday_6Period_Period = (struct __pyx_vtabstruct_6ultron_10tradingday_6Period_Period*)__Pyx_GetVtable(__pyx_ptype_6ultron_10tradingday_6Period_Period->tp_dict); if (unlikely(!__pyx_vtabptr_6ultron_10tradingday_6Period_Period)) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initCalendar(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initCalendar(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_Calendar(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_Calendar(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_Calendar(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyObject *__pyx_t_39 = NULL;
  PyObject *__pyx_t_40 = NULL;
  PyObject *__pyx_t_41 = NULL;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  PyObject *__pyx_t_45 = NULL;
  PyObject *__pyx_t_46 = NULL;
  PyObject *__pyx_t_47 = NULL;
  PyObject *__pyx_t_48 = NULL;
  PyObject *__pyx_t_49 = NULL;
  PyObject *__pyx_t_50 = NULL;
  PyObject *__pyx_t_51 = NULL;
  PyObject *__pyx_t_52 = NULL;
  PyObject *__pyx_t_53 = NULL;
  PyObject *__pyx_t_54 = NULL;
  PyObject *__pyx_t_55 = NULL;
  PyObject *__pyx_t_56 = NULL;
  PyObject *__pyx_t_57 = NULL;
  PyObject *__pyx_t_58 = NULL;
  PyObject *__pyx_t_59 = NULL;
  PyObject *__pyx_t_60 = NULL;
  PyObject *__pyx_t_61 = NULL;
  PyObject *__pyx_t_62 = NULL;
  PyObject *__pyx_t_63 = NULL;
  PyObject *__pyx_t_64 = NULL;
  PyObject *__pyx_t_65 = NULL;
  PyObject *__pyx_t_66 = NULL;
  PyObject *__pyx_t_67 = NULL;
  PyObject *__pyx_t_68 = NULL;
  PyObject *__pyx_t_69 = NULL;
  PyObject *__pyx_t_70 = NULL;
  PyObject *__pyx_t_71 = NULL;
  PyObject *__pyx_t_72 = NULL;
  PyObject *__pyx_t_73 = NULL;
  PyObject *__pyx_t_74 = NULL;
  PyObject *__pyx_t_75 = NULL;
  PyObject *__pyx_t_76 = NULL;
  PyObject *__pyx_t_77 = NULL;
  PyObject *__pyx_t_78 = NULL;
  PyObject *__pyx_t_79 = NULL;
  PyObject *__pyx_t_80 = NULL;
  PyObject *__pyx_t_81 = NULL;
  PyObject *__pyx_t_82 = NULL;
  PyObject *__pyx_t_83 = NULL;
  PyObject *__pyx_t_84 = NULL;
  PyObject *__pyx_t_85 = NULL;
  PyObject *__pyx_t_86 = NULL;
  PyObject *__pyx_t_87 = NULL;
  PyObject *__pyx_t_88 = NULL;
  PyObject *__pyx_t_89 = NULL;
  PyObject *__pyx_t_90 = NULL;
  PyObject *__pyx_t_91 = NULL;
  PyObject *__pyx_t_92 = NULL;
  PyObject *__pyx_t_93 = NULL;
  PyObject *__pyx_t_94 = NULL;
  PyObject *__pyx_t_95 = NULL;
  PyObject *__pyx_t_96 = NULL;
  PyObject *__pyx_t_97 = NULL;
  PyObject *__pyx_t_98 = NULL;
  PyObject *__pyx_t_99 = NULL;
  PyObject *__pyx_t_100 = NULL;
  PyObject *__pyx_t_101 = NULL;
  PyObject *__pyx_t_102 = NULL;
  PyObject *__pyx_t_103 = NULL;
  PyObject *__pyx_t_104 = NULL;
  PyObject *__pyx_t_105 = NULL;
  PyObject *__pyx_t_106 = NULL;
  PyObject *__pyx_t_107 = NULL;
  PyObject *__pyx_t_108 = NULL;
  PyObject *__pyx_t_109 = NULL;
  PyObject *__pyx_t_110 = NULL;
  PyObject *__pyx_t_111 = NULL;
  PyObject *__pyx_t_112 = NULL;
  PyObject *__pyx_t_113 = NULL;
  PyObject *__pyx_t_114 = NULL;
  PyObject *__pyx_t_115 = NULL;
  PyObject *__pyx_t_116 = NULL;
  PyObject *__pyx_t_117 = NULL;
  PyObject *__pyx_t_118 = NULL;
  PyObject *__pyx_t_119 = NULL;
  PyObject *__pyx_t_120 = NULL;
  PyObject *__pyx_t_121 = NULL;
  PyObject *__pyx_t_122 = NULL;
  PyObject *__pyx_t_123 = NULL;
  PyObject *__pyx_t_124 = NULL;
  PyObject *__pyx_t_125 = NULL;
  PyObject *__pyx_t_126 = NULL;
  PyObject *__pyx_t_127 = NULL;
  PyObject *__pyx_t_128 = NULL;
  PyObject *__pyx_t_129 = NULL;
  PyObject *__pyx_t_130 = NULL;
  PyObject *__pyx_t_131 = NULL;
  PyObject *__pyx_t_132 = NULL;
  PyObject *__pyx_t_133 = NULL;
  PyObject *__pyx_t_134 = NULL;
  PyObject *__pyx_t_135 = NULL;
  PyObject *__pyx_t_136 = NULL;
  PyObject *__pyx_t_137 = NULL;
  PyObject *__pyx_t_138 = NULL;
  PyObject *__pyx_t_139 = NULL;
  PyObject *__pyx_t_140 = NULL;
  PyObject *__pyx_t_141 = NULL;
  PyObject *__pyx_t_142 = NULL;
  PyObject *__pyx_t_143 = NULL;
  PyObject *__pyx_t_144 = NULL;
  PyObject *__pyx_t_145 = NULL;
  PyObject *__pyx_t_146 = NULL;
  PyObject *__pyx_t_147 = NULL;
  PyObject *__pyx_t_148 = NULL;
  PyObject *__pyx_t_149 = NULL;
  PyObject *__pyx_t_150 = NULL;
  PyObject *__pyx_t_151 = NULL;
  PyObject *__pyx_t_152 = NULL;
  PyObject *__pyx_t_153 = NULL;
  PyObject *__pyx_t_154 = NULL;
  PyObject *__pyx_t_155 = NULL;
  PyObject *__pyx_t_156 = NULL;
  PyObject *__pyx_t_157 = NULL;
  PyObject *__pyx_t_158 = NULL;
  PyObject *__pyx_t_159 = NULL;
  PyObject *__pyx_t_160 = NULL;
  PyObject *__pyx_t_161 = NULL;
  PyObject *__pyx_t_162 = NULL;
  PyObject *__pyx_t_163 = NULL;
  PyObject *__pyx_t_164 = NULL;
  PyObject *__pyx_t_165 = NULL;
  PyObject *__pyx_t_166 = NULL;
  PyObject *__pyx_t_167 = NULL;
  PyObject *__pyx_t_168 = NULL;
  PyObject *__pyx_t_169 = NULL;
  PyObject *__pyx_t_170 = NULL;
  PyObject *__pyx_t_171 = NULL;
  PyObject *__pyx_t_172 = NULL;
  PyObject *__pyx_t_173 = NULL;
  PyObject *__pyx_t_174 = NULL;
  PyObject *__pyx_t_175 = NULL;
  PyObject *__pyx_t_176 = NULL;
  PyObject *__pyx_t_177 = NULL;
  PyObject *__pyx_t_178 = NULL;
  PyObject *__pyx_t_179 = NULL;
  PyObject *__pyx_t_180 = NULL;
  PyObject *__pyx_t_181 = NULL;
  PyObject *__pyx_t_182 = NULL;
  PyObject *__pyx_t_183 = NULL;
  PyObject *__pyx_t_184 = NULL;
  PyObject *__pyx_t_185 = NULL;
  PyObject *__pyx_t_186 = NULL;
  PyObject *__pyx_t_187 = NULL;
  PyObject *__pyx_t_188 = NULL;
  PyObject *__pyx_t_189 = NULL;
  PyObject *__pyx_t_190 = NULL;
  PyObject *__pyx_t_191 = NULL;
  PyObject *__pyx_t_192 = NULL;
  PyObject *__pyx_t_193 = NULL;
  PyObject *__pyx_t_194 = NULL;
  PyObject *__pyx_t_195 = NULL;
  PyObject *__pyx_t_196 = NULL;
  PyObject *__pyx_t_197 = NULL;
  PyObject *__pyx_t_198 = NULL;
  PyObject *__pyx_t_199 = NULL;
  PyObject *__pyx_t_200 = NULL;
  PyObject *__pyx_t_201 = NULL;
  PyObject *__pyx_t_202 = NULL;
  PyObject *__pyx_t_203 = NULL;
  PyObject *__pyx_t_204 = NULL;
  PyObject *__pyx_t_205 = NULL;
  PyObject *__pyx_t_206 = NULL;
  PyObject *__pyx_t_207 = NULL;
  PyObject *__pyx_t_208 = NULL;
  PyObject *__pyx_t_209 = NULL;
  PyObject *__pyx_t_210 = NULL;
  PyObject *__pyx_t_211 = NULL;
  PyObject *__pyx_t_212 = NULL;
  PyObject *__pyx_t_213 = NULL;
  PyObject *__pyx_t_214 = NULL;
  PyObject *__pyx_t_215 = NULL;
  PyObject *__pyx_t_216 = NULL;
  PyObject *__pyx_t_217 = NULL;
  PyObject *__pyx_t_218 = NULL;
  PyObject *__pyx_t_219 = NULL;
  PyObject *__pyx_t_220 = NULL;
  PyObject *__pyx_t_221 = NULL;
  PyObject *__pyx_t_222 = NULL;
  PyObject *__pyx_t_223 = NULL;
  PyObject *__pyx_t_224 = NULL;
  PyObject *__pyx_t_225 = NULL;
  PyObject *__pyx_t_226 = NULL;
  PyObject *__pyx_t_227 = NULL;
  PyObject *__pyx_t_228 = NULL;
  PyObject *__pyx_t_229 = NULL;
  PyObject *__pyx_t_230 = NULL;
  PyObject *__pyx_t_231 = NULL;
  PyObject *__pyx_t_232 = NULL;
  PyObject *__pyx_t_233 = NULL;
  PyObject *__pyx_t_234 = NULL;
  PyObject *__pyx_t_235 = NULL;
  PyObject *__pyx_t_236 = NULL;
  PyObject *__pyx_t_237 = NULL;
  PyObject *__pyx_t_238 = NULL;
  PyObject *__pyx_t_239 = NULL;
  PyObject *__pyx_t_240 = NULL;
  PyObject *__pyx_t_241 = NULL;
  PyObject *__pyx_t_242 = NULL;
  PyObject *__pyx_t_243 = NULL;
  PyObject *__pyx_t_244 = NULL;
  PyObject *__pyx_t_245 = NULL;
  PyObject *__pyx_t_246 = NULL;
  PyObject *__pyx_t_247 = NULL;
  PyObject *__pyx_t_248 = NULL;
  PyObject *__pyx_t_249 = NULL;
  PyObject *__pyx_t_250 = NULL;
  PyObject *__pyx_t_251 = NULL;
  PyObject *__pyx_t_252 = NULL;
  PyObject *__pyx_t_253 = NULL;
  PyObject *__pyx_t_254 = NULL;
  PyObject *__pyx_t_255 = NULL;
  PyObject *__pyx_t_256 = NULL;
  PyObject *__pyx_t_257 = NULL;
  PyObject *__pyx_t_258 = NULL;
  PyObject *__pyx_t_259 = NULL;
  PyObject *__pyx_t_260 = NULL;
  PyObject *__pyx_t_261 = NULL;
  PyObject *__pyx_t_262 = NULL;
  PyObject *__pyx_t_263 = NULL;
  PyObject *__pyx_t_264 = NULL;
  PyObject *__pyx_t_265 = NULL;
  PyObject *__pyx_t_266 = NULL;
  PyObject *__pyx_t_267 = NULL;
  PyObject *__pyx_t_268 = NULL;
  PyObject *__pyx_t_269 = NULL;
  PyObject *__pyx_t_270 = NULL;
  PyObject *__pyx_t_271 = NULL;
  PyObject *__pyx_t_272 = NULL;
  PyObject *__pyx_t_273 = NULL;
  PyObject *__pyx_t_274 = NULL;
  PyObject *__pyx_t_275 = NULL;
  PyObject *__pyx_t_276 = NULL;
  PyObject *__pyx_t_277 = NULL;
  PyObject *__pyx_t_278 = NULL;
  PyObject *__pyx_t_279 = NULL;
  PyObject *__pyx_t_280 = NULL;
  PyObject *__pyx_t_281 = NULL;
  PyObject *__pyx_t_282 = NULL;
  PyObject *__pyx_t_283 = NULL;
  PyObject *__pyx_t_284 = NULL;
  PyObject *__pyx_t_285 = NULL;
  PyObject *__pyx_t_286 = NULL;
  PyObject *__pyx_t_287 = NULL;
  PyObject *__pyx_t_288 = NULL;
  PyObject *__pyx_t_289 = NULL;
  PyObject *__pyx_t_290 = NULL;
  PyObject *__pyx_t_291 = NULL;
  PyObject *__pyx_t_292 = NULL;
  PyObject *__pyx_t_293 = NULL;
  PyObject *__pyx_t_294 = NULL;
  PyObject *__pyx_t_295 = NULL;
  PyObject *__pyx_t_296 = NULL;
  PyObject *__pyx_t_297 = NULL;
  PyObject *__pyx_t_298 = NULL;
  PyObject *__pyx_t_299 = NULL;
  PyObject *__pyx_t_300 = NULL;
  PyObject *__pyx_t_301 = NULL;
  PyObject *__pyx_t_302 = NULL;
  PyObject *__pyx_t_303 = NULL;
  PyObject *__pyx_t_304 = NULL;
  PyObject *__pyx_t_305 = NULL;
  PyObject *__pyx_t_306 = NULL;
  PyObject *__pyx_t_307 = NULL;
  PyObject *__pyx_t_308 = NULL;
  PyObject *__pyx_t_309 = NULL;
  PyObject *__pyx_t_310 = NULL;
  PyObject *__pyx_t_311 = NULL;
  PyObject *__pyx_t_312 = NULL;
  PyObject *__pyx_t_313 = NULL;
  PyObject *__pyx_t_314 = NULL;
  PyObject *__pyx_t_315 = NULL;
  PyObject *__pyx_t_316 = NULL;
  PyObject *__pyx_t_317 = NULL;
  PyObject *__pyx_t_318 = NULL;
  PyObject *__pyx_t_319 = NULL;
  PyObject *__pyx_t_320 = NULL;
  PyObject *__pyx_t_321 = NULL;
  PyObject *__pyx_t_322 = NULL;
  PyObject *__pyx_t_323 = NULL;
  PyObject *__pyx_t_324 = NULL;
  PyObject *__pyx_t_325 = NULL;
  PyObject *__pyx_t_326 = NULL;
  PyObject *__pyx_t_327 = NULL;
  PyObject *__pyx_t_328 = NULL;
  PyObject *__pyx_t_329 = NULL;
  PyObject *__pyx_t_330 = NULL;
  PyObject *__pyx_t_331 = NULL;
  PyObject *__pyx_t_332 = NULL;
  PyObject *__pyx_t_333 = NULL;
  PyObject *__pyx_t_334 = NULL;
  PyObject *__pyx_t_335 = NULL;
  PyObject *__pyx_t_336 = NULL;
  PyObject *__pyx_t_337 = NULL;
  PyObject *__pyx_t_338 = NULL;
  PyObject *__pyx_t_339 = NULL;
  PyObject *__pyx_t_340 = NULL;
  PyObject *__pyx_t_341 = NULL;
  PyObject *__pyx_t_342 = NULL;
  PyObject *__pyx_t_343 = NULL;
  PyObject *__pyx_t_344 = NULL;
  PyObject *__pyx_t_345 = NULL;
  PyObject *__pyx_t_346 = NULL;
  PyObject *__pyx_t_347 = NULL;
  PyObject *__pyx_t_348 = NULL;
  PyObject *__pyx_t_349 = NULL;
  PyObject *__pyx_t_350 = NULL;
  PyObject *__pyx_t_351 = NULL;
  PyObject *__pyx_t_352 = NULL;
  PyObject *__pyx_t_353 = NULL;
  PyObject *__pyx_t_354 = NULL;
  PyObject *__pyx_t_355 = NULL;
  PyObject *__pyx_t_356 = NULL;
  PyObject *__pyx_t_357 = NULL;
  PyObject *__pyx_t_358 = NULL;
  PyObject *__pyx_t_359 = NULL;
  PyObject *__pyx_t_360 = NULL;
  PyObject *__pyx_t_361 = NULL;
  PyObject *__pyx_t_362 = NULL;
  PyObject *__pyx_t_363 = NULL;
  PyObject *__pyx_t_364 = NULL;
  PyObject *__pyx_t_365 = NULL;
  PyObject *__pyx_t_366 = NULL;
  PyObject *__pyx_t_367 = NULL;
  PyObject *__pyx_t_368 = NULL;
  PyObject *__pyx_t_369 = NULL;
  PyObject *__pyx_t_370 = NULL;
  PyObject *__pyx_t_371 = NULL;
  PyObject *__pyx_t_372 = NULL;
  PyObject *__pyx_t_373 = NULL;
  PyObject *__pyx_t_374 = NULL;
  PyObject *__pyx_t_375 = NULL;
  PyObject *__pyx_t_376 = NULL;
  PyObject *__pyx_t_377 = NULL;
  PyObject *__pyx_t_378 = NULL;
  PyObject *__pyx_t_379 = NULL;
  PyObject *__pyx_t_380 = NULL;
  PyObject *__pyx_t_381 = NULL;
  PyObject *__pyx_t_382 = NULL;
  PyObject *__pyx_t_383 = NULL;
  PyObject *__pyx_t_384 = NULL;
  PyObject *__pyx_t_385 = NULL;
  PyObject *__pyx_t_386 = NULL;
  PyObject *__pyx_t_387 = NULL;
  PyObject *__pyx_t_388 = NULL;
  PyObject *__pyx_t_389 = NULL;
  PyObject *__pyx_t_390 = NULL;
  PyObject *__pyx_t_391 = NULL;
  PyObject *__pyx_t_392 = NULL;
  PyObject *__pyx_t_393 = NULL;
  PyObject *__pyx_t_394 = NULL;
  PyObject *__pyx_t_395 = NULL;
  PyObject *__pyx_t_396 = NULL;
  PyObject *__pyx_t_397 = NULL;
  PyObject *__pyx_t_398 = NULL;
  PyObject *__pyx_t_399 = NULL;
  PyObject *__pyx_t_400 = NULL;
  PyObject *__pyx_t_401 = NULL;
  PyObject *__pyx_t_402 = NULL;
  PyObject *__pyx_t_403 = NULL;
  PyObject *__pyx_t_404 = NULL;
  PyObject *__pyx_t_405 = NULL;
  PyObject *__pyx_t_406 = NULL;
  PyObject *__pyx_t_407 = NULL;
  PyObject *__pyx_t_408 = NULL;
  PyObject *__pyx_t_409 = NULL;
  PyObject *__pyx_t_410 = NULL;
  PyObject *__pyx_t_411 = NULL;
  PyObject *__pyx_t_412 = NULL;
  PyObject *__pyx_t_413 = NULL;
  PyObject *__pyx_t_414 = NULL;
  PyObject *__pyx_t_415 = NULL;
  PyObject *__pyx_t_416 = NULL;
  PyObject *__pyx_t_417 = NULL;
  PyObject *__pyx_t_418 = NULL;
  PyObject *__pyx_t_419 = NULL;
  PyObject *__pyx_t_420 = NULL;
  PyObject *__pyx_t_421 = NULL;
  PyObject *__pyx_t_422 = NULL;
  PyObject *__pyx_t_423 = NULL;
  PyObject *__pyx_t_424 = NULL;
  PyObject *__pyx_t_425 = NULL;
  PyObject *__pyx_t_426 = NULL;
  PyObject *__pyx_t_427 = NULL;
  PyObject *__pyx_t_428 = NULL;
  PyObject *__pyx_t_429 = NULL;
  PyObject *__pyx_t_430 = NULL;
  PyObject *__pyx_t_431 = NULL;
  PyObject *__pyx_t_432 = NULL;
  PyObject *__pyx_t_433 = NULL;
  PyObject *__pyx_t_434 = NULL;
  PyObject *__pyx_t_435 = NULL;
  PyObject *__pyx_t_436 = NULL;
  PyObject *__pyx_t_437 = NULL;
  PyObject *__pyx_t_438 = NULL;
  PyObject *__pyx_t_439 = NULL;
  PyObject *__pyx_t_440 = NULL;
  PyObject *__pyx_t_441 = NULL;
  PyObject *__pyx_t_442 = NULL;
  PyObject *__pyx_t_443 = NULL;
  PyObject *__pyx_t_444 = NULL;
  PyObject *__pyx_t_445 = NULL;
  PyObject *__pyx_t_446 = NULL;
  PyObject *__pyx_t_447 = NULL;
  PyObject *__pyx_t_448 = NULL;
  PyObject *__pyx_t_449 = NULL;
  PyObject *__pyx_t_450 = NULL;
  PyObject *__pyx_t_451 = NULL;
  PyObject *__pyx_t_452 = NULL;
  PyObject *__pyx_t_453 = NULL;
  PyObject *__pyx_t_454 = NULL;
  PyObject *__pyx_t_455 = NULL;
  PyObject *__pyx_t_456 = NULL;
  PyObject *__pyx_t_457 = NULL;
  PyObject *__pyx_t_458 = NULL;
  PyObject *__pyx_t_459 = NULL;
  PyObject *__pyx_t_460 = NULL;
  PyObject *__pyx_t_461 = NULL;
  PyObject *__pyx_t_462 = NULL;
  PyObject *__pyx_t_463 = NULL;
  PyObject *__pyx_t_464 = NULL;
  PyObject *__pyx_t_465 = NULL;
  PyObject *__pyx_t_466 = NULL;
  PyObject *__pyx_t_467 = NULL;
  PyObject *__pyx_t_468 = NULL;
  PyObject *__pyx_t_469 = NULL;
  PyObject *__pyx_t_470 = NULL;
  PyObject *__pyx_t_471 = NULL;
  PyObject *__pyx_t_472 = NULL;
  PyObject *__pyx_t_473 = NULL;
  PyObject *__pyx_t_474 = NULL;
  PyObject *__pyx_t_475 = NULL;
  PyObject *__pyx_t_476 = NULL;
  PyObject *__pyx_t_477 = NULL;
  PyObject *__pyx_t_478 = NULL;
  PyObject *__pyx_t_479 = NULL;
  PyObject *__pyx_t_480 = NULL;
  PyObject *__pyx_t_481 = NULL;
  PyObject *__pyx_t_482 = NULL;
  PyObject *__pyx_t_483 = NULL;
  PyObject *__pyx_t_484 = NULL;
  PyObject *__pyx_t_485 = NULL;
  PyObject *__pyx_t_486 = NULL;
  PyObject *__pyx_t_487 = NULL;
  PyObject *__pyx_t_488 = NULL;
  PyObject *__pyx_t_489 = NULL;
  PyObject *__pyx_t_490 = NULL;
  PyObject *__pyx_t_491 = NULL;
  PyObject *__pyx_t_492 = NULL;
  PyObject *__pyx_t_493 = NULL;
  PyObject *__pyx_t_494 = NULL;
  PyObject *__pyx_t_495 = NULL;
  PyObject *__pyx_t_496 = NULL;
  PyObject *__pyx_t_497 = NULL;
  PyObject *__pyx_t_498 = NULL;
  PyObject *__pyx_t_499 = NULL;
  PyObject *__pyx_t_500 = NULL;
  PyObject *__pyx_t_501 = NULL;
  PyObject *__pyx_t_502 = NULL;
  PyObject *__pyx_t_503 = NULL;
  PyObject *__pyx_t_504 = NULL;
  PyObject *__pyx_t_505 = NULL;
  PyObject *__pyx_t_506 = NULL;
  PyObject *__pyx_t_507 = NULL;
  PyObject *__pyx_t_508 = NULL;
  PyObject *__pyx_t_509 = NULL;
  PyObject *__pyx_t_510 = NULL;
  PyObject *__pyx_t_511 = NULL;
  PyObject *__pyx_t_512 = NULL;
  PyObject *__pyx_t_513 = NULL;
  PyObject *__pyx_t_514 = NULL;
  PyObject *__pyx_t_515 = NULL;
  PyObject *__pyx_t_516 = NULL;
  PyObject *__pyx_t_517 = NULL;
  PyObject *__pyx_t_518 = NULL;
  PyObject *__pyx_t_519 = NULL;
  PyObject *__pyx_t_520 = NULL;
  PyObject *__pyx_t_521 = NULL;
  PyObject *__pyx_t_522 = NULL;
  PyObject *__pyx_t_523 = NULL;
  PyObject *__pyx_t_524 = NULL;
  PyObject *__pyx_t_525 = NULL;
  PyObject *__pyx_t_526 = NULL;
  PyObject *__pyx_t_527 = NULL;
  PyObject *__pyx_t_528 = NULL;
  PyObject *__pyx_t_529 = NULL;
  PyObject *__pyx_t_530 = NULL;
  PyObject *__pyx_t_531 = NULL;
  PyObject *__pyx_t_532 = NULL;
  PyObject *__pyx_t_533 = NULL;
  PyObject *__pyx_t_534 = NULL;
  PyObject *__pyx_t_535 = NULL;
  PyObject *__pyx_t_536 = NULL;
  PyObject *__pyx_t_537 = NULL;
  PyObject *__pyx_t_538 = NULL;
  PyObject *__pyx_t_539 = NULL;
  PyObject *__pyx_t_540 = NULL;
  PyObject *__pyx_t_541 = NULL;
  PyObject *__pyx_t_542 = NULL;
  PyObject *__pyx_t_543 = NULL;
  PyObject *__pyx_t_544 = NULL;
  PyObject *__pyx_t_545 = NULL;
  PyObject *__pyx_t_546 = NULL;
  PyObject *__pyx_t_547 = NULL;
  PyObject *__pyx_t_548 = NULL;
  PyObject *__pyx_t_549 = NULL;
  PyObject *__pyx_t_550 = NULL;
  static int __pyx_t_551[299];
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'Calendar' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_Calendar(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("Calendar", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ultron__tradingday__Calendar) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ultron.tradingday.Calendar")) {
      if (unlikely(PyDict_SetItemString(modules, "ultron.tradingday.Calendar", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ultron/tradingday/Calendar.pyx":61
 *         return wd
 * 
 *     cpdef adjustDate(self, Date d, int c=BizDayConventions.Following):             # <<<<<<<<<<<<<<
 * 
 *         cdef Date d1
 */
  __pyx_k_ = Following;
  __pyx_k_ = Following;

  /* "ultron/tradingday/Calendar.pyx":99
 *         return d1
 * 
 *     cpdef advanceDate(self, Date d, Period period, int c=BizDayConventions.Following, bint endOfMonth=False):             # <<<<<<<<<<<<<<
 * 
 *         cdef int n
 */
  __pyx_k__3 = Following;
  __pyx_k__3 = Following;

  /* "ultron/tradingday/Calendar.pyx":177
 * 
 * cdef set sse_holDays = {
 *     Date(1991, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1991, 2, 15),
 *     Date(1991, 2, 18),
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "ultron/tradingday/Calendar.pyx":178
 * cdef set sse_holDays = {
 *     Date(1991, 1, 1),
 *     Date(1991, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(1991, 2, 18),
 *     Date(1991, 5, 1),
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "ultron/tradingday/Calendar.pyx":179
 *     Date(1991, 1, 1),
 *     Date(1991, 2, 15),
 *     Date(1991, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(1991, 5, 1),
 *     Date(1991, 10, 1),
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "ultron/tradingday/Calendar.pyx":180
 *     Date(1991, 2, 15),
 *     Date(1991, 2, 18),
 *     Date(1991, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1991, 10, 1),
 *     Date(1991, 10, 2),
 */
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "ultron/tradingday/Calendar.pyx":181
 *     Date(1991, 2, 18),
 *     Date(1991, 5, 1),
 *     Date(1991, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1991, 10, 2),
 *     Date(1992, 1, 1),
 */
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "ultron/tradingday/Calendar.pyx":182
 *     Date(1991, 5, 1),
 *     Date(1991, 10, 1),
 *     Date(1991, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1992, 1, 1),
 *     Date(1992, 2, 4),
 */
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "ultron/tradingday/Calendar.pyx":183
 *     Date(1991, 10, 1),
 *     Date(1991, 10, 2),
 *     Date(1992, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1992, 2, 4),
 *     Date(1992, 2, 5),
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "ultron/tradingday/Calendar.pyx":184
 *     Date(1991, 10, 2),
 *     Date(1992, 1, 1),
 *     Date(1992, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(1992, 2, 5),
 *     Date(1992, 2, 6),
 */
  __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "ultron/tradingday/Calendar.pyx":185
 *     Date(1992, 1, 1),
 *     Date(1992, 2, 4),
 *     Date(1992, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(1992, 2, 6),
 *     Date(1992, 5, 1),
 */
  __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "ultron/tradingday/Calendar.pyx":186
 *     Date(1992, 2, 4),
 *     Date(1992, 2, 5),
 *     Date(1992, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(1992, 5, 1),
 *     Date(1992, 10, 1),
 */
  __pyx_t_10 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "ultron/tradingday/Calendar.pyx":187
 *     Date(1992, 2, 5),
 *     Date(1992, 2, 6),
 *     Date(1992, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1992, 10, 1),
 *     Date(1992, 10, 2),
 */
  __pyx_t_11 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "ultron/tradingday/Calendar.pyx":188
 *     Date(1992, 2, 6),
 *     Date(1992, 5, 1),
 *     Date(1992, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1992, 10, 2),
 *     Date(1993, 1, 1),
 */
  __pyx_t_12 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "ultron/tradingday/Calendar.pyx":189
 *     Date(1992, 5, 1),
 *     Date(1992, 10, 1),
 *     Date(1992, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1993, 1, 1),
 *     Date(1993, 1, 25),
 */
  __pyx_t_13 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "ultron/tradingday/Calendar.pyx":190
 *     Date(1992, 10, 1),
 *     Date(1992, 10, 2),
 *     Date(1993, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1993, 1, 25),
 *     Date(1993, 1, 26),
 */
  __pyx_t_14 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "ultron/tradingday/Calendar.pyx":191
 *     Date(1992, 10, 2),
 *     Date(1993, 1, 1),
 *     Date(1993, 1, 25),             # <<<<<<<<<<<<<<
 *     Date(1993, 1, 26),
 *     Date(1993, 10, 1),
 */
  __pyx_t_15 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);

  /* "ultron/tradingday/Calendar.pyx":192
 *     Date(1993, 1, 1),
 *     Date(1993, 1, 25),
 *     Date(1993, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(1993, 10, 1),
 *     Date(1994, 2, 7),
 */
  __pyx_t_16 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);

  /* "ultron/tradingday/Calendar.pyx":193
 *     Date(1993, 1, 25),
 *     Date(1993, 1, 26),
 *     Date(1993, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 7),
 *     Date(1994, 2, 8),
 */
  __pyx_t_17 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);

  /* "ultron/tradingday/Calendar.pyx":194
 *     Date(1993, 1, 26),
 *     Date(1993, 10, 1),
 *     Date(1994, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 8),
 *     Date(1994, 2, 9),
 */
  __pyx_t_18 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);

  /* "ultron/tradingday/Calendar.pyx":195
 *     Date(1993, 10, 1),
 *     Date(1994, 2, 7),
 *     Date(1994, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 9),
 *     Date(1994, 2, 10),
 */
  __pyx_t_19 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);

  /* "ultron/tradingday/Calendar.pyx":196
 *     Date(1994, 2, 7),
 *     Date(1994, 2, 8),
 *     Date(1994, 2, 9),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 10),
 *     Date(1994, 2, 11),
 */
  __pyx_t_20 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);

  /* "ultron/tradingday/Calendar.pyx":197
 *     Date(1994, 2, 8),
 *     Date(1994, 2, 9),
 *     Date(1994, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(1994, 2, 11),
 *     Date(1994, 5, 2),
 */
  __pyx_t_21 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);

  /* "ultron/tradingday/Calendar.pyx":198
 *     Date(1994, 2, 9),
 *     Date(1994, 2, 10),
 *     Date(1994, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(1994, 5, 2),
 *     Date(1994, 10, 3),
 */
  __pyx_t_22 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);

  /* "ultron/tradingday/Calendar.pyx":199
 *     Date(1994, 2, 10),
 *     Date(1994, 2, 11),
 *     Date(1994, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(1994, 10, 3),
 *     Date(1994, 10, 4),
 */
  __pyx_t_23 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);

  /* "ultron/tradingday/Calendar.pyx":200
 *     Date(1994, 2, 11),
 *     Date(1994, 5, 2),
 *     Date(1994, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(1994, 10, 4),
 *     Date(1995, 1, 2),
 */
  __pyx_t_24 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);

  /* "ultron/tradingday/Calendar.pyx":201
 *     Date(1994, 5, 2),
 *     Date(1994, 10, 3),
 *     Date(1994, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(1995, 1, 2),
 *     Date(1995, 1, 30),
 */
  __pyx_t_25 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);

  /* "ultron/tradingday/Calendar.pyx":202
 *     Date(1994, 10, 3),
 *     Date(1994, 10, 4),
 *     Date(1995, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(1995, 1, 30),
 *     Date(1995, 1, 31),
 */
  __pyx_t_26 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);

  /* "ultron/tradingday/Calendar.pyx":203
 *     Date(1994, 10, 4),
 *     Date(1995, 1, 2),
 *     Date(1995, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(1995, 1, 31),
 *     Date(1995, 2, 1),
 */
  __pyx_t_27 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);

  /* "ultron/tradingday/Calendar.pyx":204
 *     Date(1995, 1, 2),
 *     Date(1995, 1, 30),
 *     Date(1995, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(1995, 2, 1),
 *     Date(1995, 2, 2),
 */
  __pyx_t_28 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);

  /* "ultron/tradingday/Calendar.pyx":205
 *     Date(1995, 1, 30),
 *     Date(1995, 1, 31),
 *     Date(1995, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(1995, 2, 2),
 *     Date(1995, 2, 3),
 */
  __pyx_t_29 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);

  /* "ultron/tradingday/Calendar.pyx":206
 *     Date(1995, 1, 31),
 *     Date(1995, 2, 1),
 *     Date(1995, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(1995, 2, 3),
 *     Date(1995, 5, 1),
 */
  __pyx_t_30 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);

  /* "ultron/tradingday/Calendar.pyx":207
 *     Date(1995, 2, 1),
 *     Date(1995, 2, 2),
 *     Date(1995, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(1995, 5, 1),
 *     Date(1995, 10, 2),
 */
  __pyx_t_31 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);

  /* "ultron/tradingday/Calendar.pyx":208
 *     Date(1995, 2, 2),
 *     Date(1995, 2, 3),
 *     Date(1995, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1995, 10, 2),
 *     Date(1995, 10, 3),
 */
  __pyx_t_32 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);

  /* "ultron/tradingday/Calendar.pyx":209
 *     Date(1995, 2, 3),
 *     Date(1995, 5, 1),
 *     Date(1995, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1995, 10, 3),
 *     Date(1996, 1, 1),
 */
  __pyx_t_33 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);

  /* "ultron/tradingday/Calendar.pyx":210
 *     Date(1995, 5, 1),
 *     Date(1995, 10, 2),
 *     Date(1995, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(1996, 1, 1),
 *     Date(1996, 2, 19),
 */
  __pyx_t_34 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);

  /* "ultron/tradingday/Calendar.pyx":211
 *     Date(1995, 10, 2),
 *     Date(1995, 10, 3),
 *     Date(1996, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 19),
 *     Date(1996, 2, 20),
 */
  __pyx_t_35 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);

  /* "ultron/tradingday/Calendar.pyx":212
 *     Date(1995, 10, 3),
 *     Date(1996, 1, 1),
 *     Date(1996, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 20),
 *     Date(1996, 2, 21),
 */
  __pyx_t_36 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);

  /* "ultron/tradingday/Calendar.pyx":213
 *     Date(1996, 1, 1),
 *     Date(1996, 2, 19),
 *     Date(1996, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 21),
 *     Date(1996, 2, 22),
 */
  __pyx_t_37 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);

  /* "ultron/tradingday/Calendar.pyx":214
 *     Date(1996, 2, 19),
 *     Date(1996, 2, 20),
 *     Date(1996, 2, 21),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 22),
 *     Date(1996, 2, 23),
 */
  __pyx_t_38 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);

  /* "ultron/tradingday/Calendar.pyx":215
 *     Date(1996, 2, 20),
 *     Date(1996, 2, 21),
 *     Date(1996, 2, 22),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 23),
 *     Date(1996, 2, 26),
 */
  __pyx_t_39 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);

  /* "ultron/tradingday/Calendar.pyx":216
 *     Date(1996, 2, 21),
 *     Date(1996, 2, 22),
 *     Date(1996, 2, 23),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 26),
 *     Date(1996, 2, 27),
 */
  __pyx_t_40 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);

  /* "ultron/tradingday/Calendar.pyx":217
 *     Date(1996, 2, 22),
 *     Date(1996, 2, 23),
 *     Date(1996, 2, 26),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 27),
 *     Date(1996, 2, 28),
 */
  __pyx_t_41 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);

  /* "ultron/tradingday/Calendar.pyx":218
 *     Date(1996, 2, 23),
 *     Date(1996, 2, 26),
 *     Date(1996, 2, 27),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 28),
 *     Date(1996, 2, 29),
 */
  __pyx_t_42 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);

  /* "ultron/tradingday/Calendar.pyx":219
 *     Date(1996, 2, 26),
 *     Date(1996, 2, 27),
 *     Date(1996, 2, 28),             # <<<<<<<<<<<<<<
 *     Date(1996, 2, 29),
 *     Date(1996, 3, 1),
 */
  __pyx_t_43 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);

  /* "ultron/tradingday/Calendar.pyx":220
 *     Date(1996, 2, 27),
 *     Date(1996, 2, 28),
 *     Date(1996, 2, 29),             # <<<<<<<<<<<<<<
 *     Date(1996, 3, 1),
 *     Date(1996, 5, 1),
 */
  __pyx_t_44 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);

  /* "ultron/tradingday/Calendar.pyx":221
 *     Date(1996, 2, 28),
 *     Date(1996, 2, 29),
 *     Date(1996, 3, 1),             # <<<<<<<<<<<<<<
 *     Date(1996, 5, 1),
 *     Date(1996, 9, 30),
 */
  __pyx_t_45 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);

  /* "ultron/tradingday/Calendar.pyx":222
 *     Date(1996, 2, 29),
 *     Date(1996, 3, 1),
 *     Date(1996, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1996, 9, 30),
 *     Date(1996, 10, 1),
 */
  __pyx_t_46 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__49, NULL); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);

  /* "ultron/tradingday/Calendar.pyx":223
 *     Date(1996, 3, 1),
 *     Date(1996, 5, 1),
 *     Date(1996, 9, 30),             # <<<<<<<<<<<<<<
 *     Date(1996, 10, 1),
 *     Date(1996, 10, 2),
 */
  __pyx_t_47 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);

  /* "ultron/tradingday/Calendar.pyx":224
 *     Date(1996, 5, 1),
 *     Date(1996, 9, 30),
 *     Date(1996, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1996, 10, 2),
 *     Date(1997, 1, 1),
 */
  __pyx_t_48 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__51, NULL); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);

  /* "ultron/tradingday/Calendar.pyx":225
 *     Date(1996, 9, 30),
 *     Date(1996, 10, 1),
 *     Date(1996, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1997, 1, 1),
 *     Date(1997, 2, 3),
 */
  __pyx_t_49 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__52, NULL); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);

  /* "ultron/tradingday/Calendar.pyx":226
 *     Date(1996, 10, 1),
 *     Date(1996, 10, 2),
 *     Date(1997, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 3),
 *     Date(1997, 2, 4),
 */
  __pyx_t_50 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__53, NULL); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);

  /* "ultron/tradingday/Calendar.pyx":227
 *     Date(1996, 10, 2),
 *     Date(1997, 1, 1),
 *     Date(1997, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 4),
 *     Date(1997, 2, 5),
 */
  __pyx_t_51 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);

  /* "ultron/tradingday/Calendar.pyx":228
 *     Date(1997, 1, 1),
 *     Date(1997, 2, 3),
 *     Date(1997, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 5),
 *     Date(1997, 2, 6),
 */
  __pyx_t_52 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__55, NULL); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);

  /* "ultron/tradingday/Calendar.pyx":229
 *     Date(1997, 2, 3),
 *     Date(1997, 2, 4),
 *     Date(1997, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 6),
 *     Date(1997, 2, 7),
 */
  __pyx_t_53 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);

  /* "ultron/tradingday/Calendar.pyx":230
 *     Date(1997, 2, 4),
 *     Date(1997, 2, 5),
 *     Date(1997, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 7),
 *     Date(1997, 2, 10),
 */
  __pyx_t_54 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);

  /* "ultron/tradingday/Calendar.pyx":231
 *     Date(1997, 2, 5),
 *     Date(1997, 2, 6),
 *     Date(1997, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 10),
 *     Date(1997, 2, 11),
 */
  __pyx_t_55 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);

  /* "ultron/tradingday/Calendar.pyx":232
 *     Date(1997, 2, 6),
 *     Date(1997, 2, 7),
 *     Date(1997, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 11),
 *     Date(1997, 2, 12),
 */
  __pyx_t_56 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);

  /* "ultron/tradingday/Calendar.pyx":233
 *     Date(1997, 2, 7),
 *     Date(1997, 2, 10),
 *     Date(1997, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 12),
 *     Date(1997, 2, 13),
 */
  __pyx_t_57 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);

  /* "ultron/tradingday/Calendar.pyx":234
 *     Date(1997, 2, 10),
 *     Date(1997, 2, 11),
 *     Date(1997, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 13),
 *     Date(1997, 2, 14),
 */
  __pyx_t_58 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__61, NULL); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);

  /* "ultron/tradingday/Calendar.pyx":235
 *     Date(1997, 2, 11),
 *     Date(1997, 2, 12),
 *     Date(1997, 2, 13),             # <<<<<<<<<<<<<<
 *     Date(1997, 2, 14),
 *     Date(1997, 5, 1),
 */
  __pyx_t_59 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__62, NULL); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);

  /* "ultron/tradingday/Calendar.pyx":236
 *     Date(1997, 2, 12),
 *     Date(1997, 2, 13),
 *     Date(1997, 2, 14),             # <<<<<<<<<<<<<<
 *     Date(1997, 5, 1),
 *     Date(1997, 5, 2),
 */
  __pyx_t_60 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__63, NULL); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);

  /* "ultron/tradingday/Calendar.pyx":237
 *     Date(1997, 2, 13),
 *     Date(1997, 2, 14),
 *     Date(1997, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1997, 5, 2),
 *     Date(1997, 6, 30),
 */
  __pyx_t_61 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__64, NULL); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);

  /* "ultron/tradingday/Calendar.pyx":238
 *     Date(1997, 2, 14),
 *     Date(1997, 5, 1),
 *     Date(1997, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(1997, 6, 30),
 *     Date(1997, 7, 1),
 */
  __pyx_t_62 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);

  /* "ultron/tradingday/Calendar.pyx":239
 *     Date(1997, 5, 1),
 *     Date(1997, 5, 2),
 *     Date(1997, 6, 30),             # <<<<<<<<<<<<<<
 *     Date(1997, 7, 1),
 *     Date(1997, 10, 1),
 */
  __pyx_t_63 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__66, NULL); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);

  /* "ultron/tradingday/Calendar.pyx":240
 *     Date(1997, 5, 2),
 *     Date(1997, 6, 30),
 *     Date(1997, 7, 1),             # <<<<<<<<<<<<<<
 *     Date(1997, 10, 1),
 *     Date(1997, 10, 2),
 */
  __pyx_t_64 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__67, NULL); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);

  /* "ultron/tradingday/Calendar.pyx":241
 *     Date(1997, 6, 30),
 *     Date(1997, 7, 1),
 *     Date(1997, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1997, 10, 2),
 *     Date(1997, 10, 3),
 */
  __pyx_t_65 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__68, NULL); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);

  /* "ultron/tradingday/Calendar.pyx":242
 *     Date(1997, 7, 1),
 *     Date(1997, 10, 1),
 *     Date(1997, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1997, 10, 3),
 *     Date(1998, 1, 1),
 */
  __pyx_t_66 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__69, NULL); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);

  /* "ultron/tradingday/Calendar.pyx":243
 *     Date(1997, 10, 1),
 *     Date(1997, 10, 2),
 *     Date(1997, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 1),
 *     Date(1998, 1, 2),
 */
  __pyx_t_67 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__70, NULL); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);

  /* "ultron/tradingday/Calendar.pyx":244
 *     Date(1997, 10, 2),
 *     Date(1997, 10, 3),
 *     Date(1998, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 2),
 *     Date(1998, 1, 26),
 */
  __pyx_t_68 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__71, NULL); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);

  /* "ultron/tradingday/Calendar.pyx":245
 *     Date(1997, 10, 3),
 *     Date(1998, 1, 1),
 *     Date(1998, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 26),
 *     Date(1998, 1, 27),
 */
  __pyx_t_69 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__72, NULL); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);

  /* "ultron/tradingday/Calendar.pyx":246
 *     Date(1998, 1, 1),
 *     Date(1998, 1, 2),
 *     Date(1998, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 27),
 *     Date(1998, 1, 28),
 */
  __pyx_t_70 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__73, NULL); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);

  /* "ultron/tradingday/Calendar.pyx":247
 *     Date(1998, 1, 2),
 *     Date(1998, 1, 26),
 *     Date(1998, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 28),
 *     Date(1998, 1, 29),
 */
  __pyx_t_71 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__74, NULL); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);

  /* "ultron/tradingday/Calendar.pyx":248
 *     Date(1998, 1, 26),
 *     Date(1998, 1, 27),
 *     Date(1998, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 29),
 *     Date(1998, 1, 30),
 */
  __pyx_t_72 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__75, NULL); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);

  /* "ultron/tradingday/Calendar.pyx":249
 *     Date(1998, 1, 27),
 *     Date(1998, 1, 28),
 *     Date(1998, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(1998, 1, 30),
 *     Date(1998, 2, 2),
 */
  __pyx_t_73 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__76, NULL); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);

  /* "ultron/tradingday/Calendar.pyx":250
 *     Date(1998, 1, 28),
 *     Date(1998, 1, 29),
 *     Date(1998, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 2),
 *     Date(1998, 2, 3),
 */
  __pyx_t_74 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__77, NULL); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);

  /* "ultron/tradingday/Calendar.pyx":251
 *     Date(1998, 1, 29),
 *     Date(1998, 1, 30),
 *     Date(1998, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 3),
 *     Date(1998, 2, 4),
 */
  __pyx_t_75 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__78, NULL); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);

  /* "ultron/tradingday/Calendar.pyx":252
 *     Date(1998, 1, 30),
 *     Date(1998, 2, 2),
 *     Date(1998, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 4),
 *     Date(1998, 2, 5),
 */
  __pyx_t_76 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__79, NULL); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);

  /* "ultron/tradingday/Calendar.pyx":253
 *     Date(1998, 2, 2),
 *     Date(1998, 2, 3),
 *     Date(1998, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 5),
 *     Date(1998, 2, 6),
 */
  __pyx_t_77 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__80, NULL); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);

  /* "ultron/tradingday/Calendar.pyx":254
 *     Date(1998, 2, 3),
 *     Date(1998, 2, 4),
 *     Date(1998, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(1998, 2, 6),
 *     Date(1998, 5, 1),
 */
  __pyx_t_78 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__81, NULL); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);

  /* "ultron/tradingday/Calendar.pyx":255
 *     Date(1998, 2, 4),
 *     Date(1998, 2, 5),
 *     Date(1998, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(1998, 5, 1),
 *     Date(1998, 10, 1),
 */
  __pyx_t_79 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__82, NULL); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);

  /* "ultron/tradingday/Calendar.pyx":256
 *     Date(1998, 2, 5),
 *     Date(1998, 2, 6),
 *     Date(1998, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(1998, 10, 1),
 *     Date(1998, 10, 2),
 */
  __pyx_t_80 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);

  /* "ultron/tradingday/Calendar.pyx":257
 *     Date(1998, 2, 6),
 *     Date(1998, 5, 1),
 *     Date(1998, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1998, 10, 2),
 *     Date(1999, 1, 1),
 */
  __pyx_t_81 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__84, NULL); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);

  /* "ultron/tradingday/Calendar.pyx":258
 *     Date(1998, 5, 1),
 *     Date(1998, 10, 1),
 *     Date(1998, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(1999, 1, 1),
 *     Date(1999, 2, 10),
 */
  __pyx_t_82 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__85, NULL); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);

  /* "ultron/tradingday/Calendar.pyx":259
 *     Date(1998, 10, 1),
 *     Date(1998, 10, 2),
 *     Date(1999, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 10),
 *     Date(1999, 2, 11),
 */
  __pyx_t_83 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__86, NULL); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);

  /* "ultron/tradingday/Calendar.pyx":260
 *     Date(1998, 10, 2),
 *     Date(1999, 1, 1),
 *     Date(1999, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 11),
 *     Date(1999, 2, 12),
 */
  __pyx_t_84 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__87, NULL); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);

  /* "ultron/tradingday/Calendar.pyx":261
 *     Date(1999, 1, 1),
 *     Date(1999, 2, 10),
 *     Date(1999, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 12),
 *     Date(1999, 2, 15),
 */
  __pyx_t_85 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__88, NULL); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);

  /* "ultron/tradingday/Calendar.pyx":262
 *     Date(1999, 2, 10),
 *     Date(1999, 2, 11),
 *     Date(1999, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 15),
 *     Date(1999, 2, 16),
 */
  __pyx_t_86 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__89, NULL); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);

  /* "ultron/tradingday/Calendar.pyx":263
 *     Date(1999, 2, 11),
 *     Date(1999, 2, 12),
 *     Date(1999, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 16),
 *     Date(1999, 2, 17),
 */
  __pyx_t_87 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__90, NULL); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);

  /* "ultron/tradingday/Calendar.pyx":264
 *     Date(1999, 2, 12),
 *     Date(1999, 2, 15),
 *     Date(1999, 2, 16),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 17),
 *     Date(1999, 2, 18),
 */
  __pyx_t_88 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__91, NULL); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);

  /* "ultron/tradingday/Calendar.pyx":265
 *     Date(1999, 2, 15),
 *     Date(1999, 2, 16),
 *     Date(1999, 2, 17),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 18),
 *     Date(1999, 2, 19),
 */
  __pyx_t_89 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__92, NULL); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);

  /* "ultron/tradingday/Calendar.pyx":266
 *     Date(1999, 2, 16),
 *     Date(1999, 2, 17),
 *     Date(1999, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 19),
 *     Date(1999, 2, 22),
 */
  __pyx_t_90 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__93, NULL); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);

  /* "ultron/tradingday/Calendar.pyx":267
 *     Date(1999, 2, 17),
 *     Date(1999, 2, 18),
 *     Date(1999, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 22),
 *     Date(1999, 2, 23),
 */
  __pyx_t_91 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__94, NULL); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);

  /* "ultron/tradingday/Calendar.pyx":268
 *     Date(1999, 2, 18),
 *     Date(1999, 2, 19),
 *     Date(1999, 2, 22),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 23),
 *     Date(1999, 2, 24),
 */
  __pyx_t_92 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__95, NULL); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);

  /* "ultron/tradingday/Calendar.pyx":269
 *     Date(1999, 2, 19),
 *     Date(1999, 2, 22),
 *     Date(1999, 2, 23),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 24),
 *     Date(1999, 2, 25),
 */
  __pyx_t_93 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__96, NULL); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);

  /* "ultron/tradingday/Calendar.pyx":270
 *     Date(1999, 2, 22),
 *     Date(1999, 2, 23),
 *     Date(1999, 2, 24),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 25),
 *     Date(1999, 2, 26),
 */
  __pyx_t_94 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__97, NULL); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);

  /* "ultron/tradingday/Calendar.pyx":271
 *     Date(1999, 2, 23),
 *     Date(1999, 2, 24),
 *     Date(1999, 2, 25),             # <<<<<<<<<<<<<<
 *     Date(1999, 2, 26),
 *     Date(1999, 5, 3),
 */
  __pyx_t_95 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__98, NULL); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);

  /* "ultron/tradingday/Calendar.pyx":272
 *     Date(1999, 2, 24),
 *     Date(1999, 2, 25),
 *     Date(1999, 2, 26),             # <<<<<<<<<<<<<<
 *     Date(1999, 5, 3),
 *     Date(1999, 10, 1),
 */
  __pyx_t_96 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__99, NULL); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);

  /* "ultron/tradingday/Calendar.pyx":273
 *     Date(1999, 2, 25),
 *     Date(1999, 2, 26),
 *     Date(1999, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 1),
 *     Date(1999, 10, 4),
 */
  __pyx_t_97 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__100, NULL); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);

  /* "ultron/tradingday/Calendar.pyx":274
 *     Date(1999, 2, 26),
 *     Date(1999, 5, 3),
 *     Date(1999, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 4),
 *     Date(1999, 10, 5),
 */
  __pyx_t_98 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__101, NULL); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);

  /* "ultron/tradingday/Calendar.pyx":275
 *     Date(1999, 5, 3),
 *     Date(1999, 10, 1),
 *     Date(1999, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 5),
 *     Date(1999, 10, 6),
 */
  __pyx_t_99 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__102, NULL); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);

  /* "ultron/tradingday/Calendar.pyx":276
 *     Date(1999, 10, 1),
 *     Date(1999, 10, 4),
 *     Date(1999, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 6),
 *     Date(1999, 10, 7),
 */
  __pyx_t_100 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__103, NULL); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);

  /* "ultron/tradingday/Calendar.pyx":277
 *     Date(1999, 10, 4),
 *     Date(1999, 10, 5),
 *     Date(1999, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(1999, 10, 7),
 *     Date(1999, 12, 20),
 */
  __pyx_t_101 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__104, NULL); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);

  /* "ultron/tradingday/Calendar.pyx":278
 *     Date(1999, 10, 5),
 *     Date(1999, 10, 6),
 *     Date(1999, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(1999, 12, 20),
 *     Date(1999, 12, 31),
 */
  __pyx_t_102 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__105, NULL); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);

  /* "ultron/tradingday/Calendar.pyx":279
 *     Date(1999, 10, 6),
 *     Date(1999, 10, 7),
 *     Date(1999, 12, 20),             # <<<<<<<<<<<<<<
 *     Date(1999, 12, 31),
 *     Date(2000, 1, 3),
 */
  __pyx_t_103 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__106, NULL); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);

  /* "ultron/tradingday/Calendar.pyx":280
 *     Date(1999, 10, 7),
 *     Date(1999, 12, 20),
 *     Date(1999, 12, 31),             # <<<<<<<<<<<<<<
 *     Date(2000, 1, 3),
 *     Date(2000, 1, 31),
 */
  __pyx_t_104 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__107, NULL); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);

  /* "ultron/tradingday/Calendar.pyx":281
 *     Date(1999, 12, 20),
 *     Date(1999, 12, 31),
 *     Date(2000, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2000, 1, 31),
 *     Date(2000, 2, 1),
 */
  __pyx_t_105 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__108, NULL); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);

  /* "ultron/tradingday/Calendar.pyx":282
 *     Date(1999, 12, 31),
 *     Date(2000, 1, 3),
 *     Date(2000, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 1),
 *     Date(2000, 2, 2),
 */
  __pyx_t_106 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__109, NULL); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);

  /* "ultron/tradingday/Calendar.pyx":283
 *     Date(2000, 1, 3),
 *     Date(2000, 1, 31),
 *     Date(2000, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 2),
 *     Date(2000, 2, 3),
 */
  __pyx_t_107 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__110, NULL); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);

  /* "ultron/tradingday/Calendar.pyx":284
 *     Date(2000, 1, 31),
 *     Date(2000, 2, 1),
 *     Date(2000, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 3),
 *     Date(2000, 2, 4),
 */
  __pyx_t_108 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__111, NULL); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);

  /* "ultron/tradingday/Calendar.pyx":285
 *     Date(2000, 2, 1),
 *     Date(2000, 2, 2),
 *     Date(2000, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 4),
 *     Date(2000, 2, 7),
 */
  __pyx_t_109 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__112, NULL); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);

  /* "ultron/tradingday/Calendar.pyx":286
 *     Date(2000, 2, 2),
 *     Date(2000, 2, 3),
 *     Date(2000, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 7),
 *     Date(2000, 2, 8),
 */
  __pyx_t_110 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__113, NULL); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);

  /* "ultron/tradingday/Calendar.pyx":287
 *     Date(2000, 2, 3),
 *     Date(2000, 2, 4),
 *     Date(2000, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 8),
 *     Date(2000, 2, 9),
 */
  __pyx_t_111 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__114, NULL); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);

  /* "ultron/tradingday/Calendar.pyx":288
 *     Date(2000, 2, 4),
 *     Date(2000, 2, 7),
 *     Date(2000, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 9),
 *     Date(2000, 2, 10),
 */
  __pyx_t_112 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__115, NULL); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);

  /* "ultron/tradingday/Calendar.pyx":289
 *     Date(2000, 2, 7),
 *     Date(2000, 2, 8),
 *     Date(2000, 2, 9),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 10),
 *     Date(2000, 2, 11),
 */
  __pyx_t_113 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__116, NULL); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);

  /* "ultron/tradingday/Calendar.pyx":290
 *     Date(2000, 2, 8),
 *     Date(2000, 2, 9),
 *     Date(2000, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(2000, 2, 11),
 *     Date(2000, 5, 1),
 */
  __pyx_t_114 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__117, NULL); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);

  /* "ultron/tradingday/Calendar.pyx":291
 *     Date(2000, 2, 9),
 *     Date(2000, 2, 10),
 *     Date(2000, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 1),
 *     Date(2000, 5, 2),
 */
  __pyx_t_115 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__118, NULL); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);

  /* "ultron/tradingday/Calendar.pyx":292
 *     Date(2000, 2, 10),
 *     Date(2000, 2, 11),
 *     Date(2000, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 2),
 *     Date(2000, 5, 3),
 */
  __pyx_t_116 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__119, NULL); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);

  /* "ultron/tradingday/Calendar.pyx":293
 *     Date(2000, 2, 11),
 *     Date(2000, 5, 1),
 *     Date(2000, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 3),
 *     Date(2000, 5, 4),
 */
  __pyx_t_117 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__120, NULL); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);

  /* "ultron/tradingday/Calendar.pyx":294
 *     Date(2000, 5, 1),
 *     Date(2000, 5, 2),
 *     Date(2000, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 4),
 *     Date(2000, 5, 5),
 */
  __pyx_t_118 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__121, NULL); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);

  /* "ultron/tradingday/Calendar.pyx":295
 *     Date(2000, 5, 2),
 *     Date(2000, 5, 3),
 *     Date(2000, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2000, 5, 5),
 *     Date(2000, 10, 2),
 */
  __pyx_t_119 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__122, NULL); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);

  /* "ultron/tradingday/Calendar.pyx":296
 *     Date(2000, 5, 3),
 *     Date(2000, 5, 4),
 *     Date(2000, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 2),
 *     Date(2000, 10, 3),
 */
  __pyx_t_120 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__123, NULL); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);

  /* "ultron/tradingday/Calendar.pyx":297
 *     Date(2000, 5, 4),
 *     Date(2000, 5, 5),
 *     Date(2000, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 3),
 *     Date(2000, 10, 4),
 */
  __pyx_t_121 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__124, NULL); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);

  /* "ultron/tradingday/Calendar.pyx":298
 *     Date(2000, 5, 5),
 *     Date(2000, 10, 2),
 *     Date(2000, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 4),
 *     Date(2000, 10, 5),
 */
  __pyx_t_122 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__125, NULL); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);

  /* "ultron/tradingday/Calendar.pyx":299
 *     Date(2000, 10, 2),
 *     Date(2000, 10, 3),
 *     Date(2000, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 5),
 *     Date(2000, 10, 6),
 */
  __pyx_t_123 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__126, NULL); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);

  /* "ultron/tradingday/Calendar.pyx":300
 *     Date(2000, 10, 3),
 *     Date(2000, 10, 4),
 *     Date(2000, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2000, 10, 6),
 *     Date(2001, 1, 1),
 */
  __pyx_t_124 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__127, NULL); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);

  /* "ultron/tradingday/Calendar.pyx":301
 *     Date(2000, 10, 4),
 *     Date(2000, 10, 5),
 *     Date(2000, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 1),
 *     Date(2001, 1, 22),
 */
  __pyx_t_125 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__128, NULL); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);

  /* "ultron/tradingday/Calendar.pyx":302
 *     Date(2000, 10, 5),
 *     Date(2000, 10, 6),
 *     Date(2001, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 22),
 *     Date(2001, 1, 23),
 */
  __pyx_t_126 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__129, NULL); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);

  /* "ultron/tradingday/Calendar.pyx":303
 *     Date(2000, 10, 6),
 *     Date(2001, 1, 1),
 *     Date(2001, 1, 22),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 23),
 *     Date(2001, 1, 24),
 */
  __pyx_t_127 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__130, NULL); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);

  /* "ultron/tradingday/Calendar.pyx":304
 *     Date(2001, 1, 1),
 *     Date(2001, 1, 22),
 *     Date(2001, 1, 23),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 24),
 *     Date(2001, 1, 25),
 */
  __pyx_t_128 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__131, NULL); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);

  /* "ultron/tradingday/Calendar.pyx":305
 *     Date(2001, 1, 22),
 *     Date(2001, 1, 23),
 *     Date(2001, 1, 24),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 25),
 *     Date(2001, 1, 26),
 */
  __pyx_t_129 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__132, NULL); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);

  /* "ultron/tradingday/Calendar.pyx":306
 *     Date(2001, 1, 23),
 *     Date(2001, 1, 24),
 *     Date(2001, 1, 25),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 26),
 *     Date(2001, 1, 29),
 */
  __pyx_t_130 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__133, NULL); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);

  /* "ultron/tradingday/Calendar.pyx":307
 *     Date(2001, 1, 24),
 *     Date(2001, 1, 25),
 *     Date(2001, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 29),
 *     Date(2001, 1, 30),
 */
  __pyx_t_131 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__134, NULL); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);

  /* "ultron/tradingday/Calendar.pyx":308
 *     Date(2001, 1, 25),
 *     Date(2001, 1, 26),
 *     Date(2001, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 30),
 *     Date(2001, 1, 31),
 */
  __pyx_t_132 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__135, NULL); if (unlikely(!__pyx_t_132)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_132);

  /* "ultron/tradingday/Calendar.pyx":309
 *     Date(2001, 1, 26),
 *     Date(2001, 1, 29),
 *     Date(2001, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2001, 1, 31),
 *     Date(2001, 2, 1),
 */
  __pyx_t_133 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__136, NULL); if (unlikely(!__pyx_t_133)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_133);

  /* "ultron/tradingday/Calendar.pyx":310
 *     Date(2001, 1, 29),
 *     Date(2001, 1, 30),
 *     Date(2001, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2001, 2, 1),
 *     Date(2001, 2, 2),
 */
  __pyx_t_134 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__137, NULL); if (unlikely(!__pyx_t_134)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_134);

  /* "ultron/tradingday/Calendar.pyx":311
 *     Date(2001, 1, 30),
 *     Date(2001, 1, 31),
 *     Date(2001, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2001, 2, 2),
 *     Date(2001, 5, 1),
 */
  __pyx_t_135 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__138, NULL); if (unlikely(!__pyx_t_135)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_135);

  /* "ultron/tradingday/Calendar.pyx":312
 *     Date(2001, 1, 31),
 *     Date(2001, 2, 1),
 *     Date(2001, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 1),
 *     Date(2001, 5, 2),
 */
  __pyx_t_136 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__139, NULL); if (unlikely(!__pyx_t_136)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_136);

  /* "ultron/tradingday/Calendar.pyx":313
 *     Date(2001, 2, 1),
 *     Date(2001, 2, 2),
 *     Date(2001, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 2),
 *     Date(2001, 5, 3),
 */
  __pyx_t_137 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__140, NULL); if (unlikely(!__pyx_t_137)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_137);

  /* "ultron/tradingday/Calendar.pyx":314
 *     Date(2001, 2, 2),
 *     Date(2001, 5, 1),
 *     Date(2001, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 3),
 *     Date(2001, 5, 4),
 */
  __pyx_t_138 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__141, NULL); if (unlikely(!__pyx_t_138)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_138);

  /* "ultron/tradingday/Calendar.pyx":315
 *     Date(2001, 5, 1),
 *     Date(2001, 5, 2),
 *     Date(2001, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 4),
 *     Date(2001, 5, 7),
 */
  __pyx_t_139 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__142, NULL); if (unlikely(!__pyx_t_139)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_139);

  /* "ultron/tradingday/Calendar.pyx":316
 *     Date(2001, 5, 2),
 *     Date(2001, 5, 3),
 *     Date(2001, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2001, 5, 7),
 *     Date(2001, 10, 1),
 */
  __pyx_t_140 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__143, NULL); if (unlikely(!__pyx_t_140)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_140);

  /* "ultron/tradingday/Calendar.pyx":317
 *     Date(2001, 5, 3),
 *     Date(2001, 5, 4),
 *     Date(2001, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 1),
 *     Date(2001, 10, 2),
 */
  __pyx_t_141 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__144, NULL); if (unlikely(!__pyx_t_141)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_141);

  /* "ultron/tradingday/Calendar.pyx":318
 *     Date(2001, 5, 4),
 *     Date(2001, 5, 7),
 *     Date(2001, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 2),
 *     Date(2001, 10, 3),
 */
  __pyx_t_142 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__145, NULL); if (unlikely(!__pyx_t_142)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_142);

  /* "ultron/tradingday/Calendar.pyx":319
 *     Date(2001, 5, 7),
 *     Date(2001, 10, 1),
 *     Date(2001, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 3),
 *     Date(2001, 10, 4),
 */
  __pyx_t_143 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__146, NULL); if (unlikely(!__pyx_t_143)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_143);

  /* "ultron/tradingday/Calendar.pyx":320
 *     Date(2001, 10, 1),
 *     Date(2001, 10, 2),
 *     Date(2001, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 4),
 *     Date(2001, 10, 5),
 */
  __pyx_t_144 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__147, NULL); if (unlikely(!__pyx_t_144)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_144);

  /* "ultron/tradingday/Calendar.pyx":321
 *     Date(2001, 10, 2),
 *     Date(2001, 10, 3),
 *     Date(2001, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2001, 10, 5),
 *     Date(2002, 1, 1),
 */
  __pyx_t_145 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__148, NULL); if (unlikely(!__pyx_t_145)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_145);

  /* "ultron/tradingday/Calendar.pyx":322
 *     Date(2001, 10, 3),
 *     Date(2001, 10, 4),
 *     Date(2001, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2002, 1, 1),
 *     Date(2002, 1, 2),
 */
  __pyx_t_146 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__149, NULL); if (unlikely(!__pyx_t_146)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_146);

  /* "ultron/tradingday/Calendar.pyx":323
 *     Date(2001, 10, 4),
 *     Date(2001, 10, 5),
 *     Date(2002, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2002, 1, 2),
 *     Date(2002, 1, 3),
 */
  __pyx_t_147 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__150, NULL); if (unlikely(!__pyx_t_147)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_147);

  /* "ultron/tradingday/Calendar.pyx":324
 *     Date(2001, 10, 5),
 *     Date(2002, 1, 1),
 *     Date(2002, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2002, 1, 3),
 *     Date(2002, 2, 11),
 */
  __pyx_t_148 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__151, NULL); if (unlikely(!__pyx_t_148)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_148);

  /* "ultron/tradingday/Calendar.pyx":325
 *     Date(2002, 1, 1),
 *     Date(2002, 1, 2),
 *     Date(2002, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 11),
 *     Date(2002, 2, 12),
 */
  __pyx_t_149 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__152, NULL); if (unlikely(!__pyx_t_149)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_149);

  /* "ultron/tradingday/Calendar.pyx":326
 *     Date(2002, 1, 2),
 *     Date(2002, 1, 3),
 *     Date(2002, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 12),
 *     Date(2002, 2, 13),
 */
  __pyx_t_150 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__153, NULL); if (unlikely(!__pyx_t_150)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_150);

  /* "ultron/tradingday/Calendar.pyx":327
 *     Date(2002, 1, 3),
 *     Date(2002, 2, 11),
 *     Date(2002, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 13),
 *     Date(2002, 2, 14),
 */
  __pyx_t_151 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__154, NULL); if (unlikely(!__pyx_t_151)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_151);

  /* "ultron/tradingday/Calendar.pyx":328
 *     Date(2002, 2, 11),
 *     Date(2002, 2, 12),
 *     Date(2002, 2, 13),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 14),
 *     Date(2002, 2, 15),
 */
  __pyx_t_152 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__155, NULL); if (unlikely(!__pyx_t_152)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_152);

  /* "ultron/tradingday/Calendar.pyx":329
 *     Date(2002, 2, 12),
 *     Date(2002, 2, 13),
 *     Date(2002, 2, 14),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 15),
 *     Date(2002, 2, 18),
 */
  __pyx_t_153 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__156, NULL); if (unlikely(!__pyx_t_153)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_153);

  /* "ultron/tradingday/Calendar.pyx":330
 *     Date(2002, 2, 13),
 *     Date(2002, 2, 14),
 *     Date(2002, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 18),
 *     Date(2002, 2, 19),
 */
  __pyx_t_154 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__157, NULL); if (unlikely(!__pyx_t_154)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_154);

  /* "ultron/tradingday/Calendar.pyx":331
 *     Date(2002, 2, 14),
 *     Date(2002, 2, 15),
 *     Date(2002, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 19),
 *     Date(2002, 2, 20),
 */
  __pyx_t_155 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__158, NULL); if (unlikely(!__pyx_t_155)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_155);

  /* "ultron/tradingday/Calendar.pyx":332
 *     Date(2002, 2, 15),
 *     Date(2002, 2, 18),
 *     Date(2002, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 20),
 *     Date(2002, 2, 21),
 */
  __pyx_t_156 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__159, NULL); if (unlikely(!__pyx_t_156)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_156);

  /* "ultron/tradingday/Calendar.pyx":333
 *     Date(2002, 2, 18),
 *     Date(2002, 2, 19),
 *     Date(2002, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 21),
 *     Date(2002, 2, 22),
 */
  __pyx_t_157 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__160, NULL); if (unlikely(!__pyx_t_157)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_157);

  /* "ultron/tradingday/Calendar.pyx":334
 *     Date(2002, 2, 19),
 *     Date(2002, 2, 20),
 *     Date(2002, 2, 21),             # <<<<<<<<<<<<<<
 *     Date(2002, 2, 22),
 *     Date(2002, 5, 1),
 */
  __pyx_t_158 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__161, NULL); if (unlikely(!__pyx_t_158)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_158);

  /* "ultron/tradingday/Calendar.pyx":335
 *     Date(2002, 2, 20),
 *     Date(2002, 2, 21),
 *     Date(2002, 2, 22),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 1),
 *     Date(2002, 5, 2),
 */
  __pyx_t_159 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__162, NULL); if (unlikely(!__pyx_t_159)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_159);

  /* "ultron/tradingday/Calendar.pyx":336
 *     Date(2002, 2, 21),
 *     Date(2002, 2, 22),
 *     Date(2002, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 2),
 *     Date(2002, 5, 3),
 */
  __pyx_t_160 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__163, NULL); if (unlikely(!__pyx_t_160)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_160);

  /* "ultron/tradingday/Calendar.pyx":337
 *     Date(2002, 2, 22),
 *     Date(2002, 5, 1),
 *     Date(2002, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 3),
 *     Date(2002, 5, 6),
 */
  __pyx_t_161 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__164, NULL); if (unlikely(!__pyx_t_161)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_161);

  /* "ultron/tradingday/Calendar.pyx":338
 *     Date(2002, 5, 1),
 *     Date(2002, 5, 2),
 *     Date(2002, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 6),
 *     Date(2002, 5, 7),
 */
  __pyx_t_162 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__165, NULL); if (unlikely(!__pyx_t_162)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_162);

  /* "ultron/tradingday/Calendar.pyx":339
 *     Date(2002, 5, 2),
 *     Date(2002, 5, 3),
 *     Date(2002, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2002, 5, 7),
 *     Date(2002, 9, 30),
 */
  __pyx_t_163 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__166, NULL); if (unlikely(!__pyx_t_163)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_163);

  /* "ultron/tradingday/Calendar.pyx":340
 *     Date(2002, 5, 3),
 *     Date(2002, 5, 6),
 *     Date(2002, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2002, 9, 30),
 *     Date(2002, 10, 1),
 */
  __pyx_t_164 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__167, NULL); if (unlikely(!__pyx_t_164)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_164);

  /* "ultron/tradingday/Calendar.pyx":341
 *     Date(2002, 5, 6),
 *     Date(2002, 5, 7),
 *     Date(2002, 9, 30),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 1),
 *     Date(2002, 10, 2),
 */
  __pyx_t_165 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__168, NULL); if (unlikely(!__pyx_t_165)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_165);

  /* "ultron/tradingday/Calendar.pyx":342
 *     Date(2002, 5, 7),
 *     Date(2002, 9, 30),
 *     Date(2002, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 2),
 *     Date(2002, 10, 3),
 */
  __pyx_t_166 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__169, NULL); if (unlikely(!__pyx_t_166)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_166);

  /* "ultron/tradingday/Calendar.pyx":343
 *     Date(2002, 9, 30),
 *     Date(2002, 10, 1),
 *     Date(2002, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 3),
 *     Date(2002, 10, 4),
 */
  __pyx_t_167 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__170, NULL); if (unlikely(!__pyx_t_167)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_167);

  /* "ultron/tradingday/Calendar.pyx":344
 *     Date(2002, 10, 1),
 *     Date(2002, 10, 2),
 *     Date(2002, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 4),
 *     Date(2002, 10, 7),
 */
  __pyx_t_168 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__171, NULL); if (unlikely(!__pyx_t_168)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_168);

  /* "ultron/tradingday/Calendar.pyx":345
 *     Date(2002, 10, 2),
 *     Date(2002, 10, 3),
 *     Date(2002, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2002, 10, 7),
 *     Date(2003, 1, 1),
 */
  __pyx_t_169 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__172, NULL); if (unlikely(!__pyx_t_169)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_169);

  /* "ultron/tradingday/Calendar.pyx":346
 *     Date(2002, 10, 3),
 *     Date(2002, 10, 4),
 *     Date(2002, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2003, 1, 1),
 *     Date(2003, 1, 30),
 */
  __pyx_t_170 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__173, NULL); if (unlikely(!__pyx_t_170)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_170);

  /* "ultron/tradingday/Calendar.pyx":347
 *     Date(2002, 10, 4),
 *     Date(2002, 10, 7),
 *     Date(2003, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2003, 1, 30),
 *     Date(2003, 1, 31),
 */
  __pyx_t_171 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__174, NULL); if (unlikely(!__pyx_t_171)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_171);

  /* "ultron/tradingday/Calendar.pyx":348
 *     Date(2002, 10, 7),
 *     Date(2003, 1, 1),
 *     Date(2003, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2003, 1, 31),
 *     Date(2003, 2, 3),
 */
  __pyx_t_172 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__175, NULL); if (unlikely(!__pyx_t_172)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_172);

  /* "ultron/tradingday/Calendar.pyx":349
 *     Date(2003, 1, 1),
 *     Date(2003, 1, 30),
 *     Date(2003, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 3),
 *     Date(2003, 2, 4),
 */
  __pyx_t_173 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__176, NULL); if (unlikely(!__pyx_t_173)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_173);

  /* "ultron/tradingday/Calendar.pyx":350
 *     Date(2003, 1, 30),
 *     Date(2003, 1, 31),
 *     Date(2003, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 4),
 *     Date(2003, 2, 5),
 */
  __pyx_t_174 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__177, NULL); if (unlikely(!__pyx_t_174)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_174);

  /* "ultron/tradingday/Calendar.pyx":351
 *     Date(2003, 1, 31),
 *     Date(2003, 2, 3),
 *     Date(2003, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 5),
 *     Date(2003, 2, 6),
 */
  __pyx_t_175 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__178, NULL); if (unlikely(!__pyx_t_175)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_175);

  /* "ultron/tradingday/Calendar.pyx":352
 *     Date(2003, 2, 3),
 *     Date(2003, 2, 4),
 *     Date(2003, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 6),
 *     Date(2003, 2, 7),
 */
  __pyx_t_176 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__179, NULL); if (unlikely(!__pyx_t_176)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_176);

  /* "ultron/tradingday/Calendar.pyx":353
 *     Date(2003, 2, 4),
 *     Date(2003, 2, 5),
 *     Date(2003, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(2003, 2, 7),
 *     Date(2003, 5, 1),
 */
  __pyx_t_177 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__180, NULL); if (unlikely(!__pyx_t_177)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_177);

  /* "ultron/tradingday/Calendar.pyx":354
 *     Date(2003, 2, 5),
 *     Date(2003, 2, 6),
 *     Date(2003, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 1),
 *     Date(2003, 5, 2),
 */
  __pyx_t_178 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__181, NULL); if (unlikely(!__pyx_t_178)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_178);

  /* "ultron/tradingday/Calendar.pyx":355
 *     Date(2003, 2, 6),
 *     Date(2003, 2, 7),
 *     Date(2003, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 2),
 *     Date(2003, 5, 5),
 */
  __pyx_t_179 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__182, NULL); if (unlikely(!__pyx_t_179)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_179);

  /* "ultron/tradingday/Calendar.pyx":356
 *     Date(2003, 2, 7),
 *     Date(2003, 5, 1),
 *     Date(2003, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 5),
 *     Date(2003, 5, 6),
 */
  __pyx_t_180 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__183, NULL); if (unlikely(!__pyx_t_180)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_180);

  /* "ultron/tradingday/Calendar.pyx":357
 *     Date(2003, 5, 1),
 *     Date(2003, 5, 2),
 *     Date(2003, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 6),
 *     Date(2003, 5, 7),
 */
  __pyx_t_181 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__184, NULL); if (unlikely(!__pyx_t_181)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_181);

  /* "ultron/tradingday/Calendar.pyx":358
 *     Date(2003, 5, 2),
 *     Date(2003, 5, 5),
 *     Date(2003, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 7),
 *     Date(2003, 5, 8),
 */
  __pyx_t_182 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__185, NULL); if (unlikely(!__pyx_t_182)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_182);

  /* "ultron/tradingday/Calendar.pyx":359
 *     Date(2003, 5, 5),
 *     Date(2003, 5, 6),
 *     Date(2003, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 8),
 *     Date(2003, 5, 9),
 */
  __pyx_t_183 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__186, NULL); if (unlikely(!__pyx_t_183)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_183);

  /* "ultron/tradingday/Calendar.pyx":360
 *     Date(2003, 5, 6),
 *     Date(2003, 5, 7),
 *     Date(2003, 5, 8),             # <<<<<<<<<<<<<<
 *     Date(2003, 5, 9),
 *     Date(2003, 10, 1),
 */
  __pyx_t_184 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__187, NULL); if (unlikely(!__pyx_t_184)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_184);

  /* "ultron/tradingday/Calendar.pyx":361
 *     Date(2003, 5, 7),
 *     Date(2003, 5, 8),
 *     Date(2003, 5, 9),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 1),
 *     Date(2003, 10, 2),
 */
  __pyx_t_185 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__188, NULL); if (unlikely(!__pyx_t_185)) __PYX_ERR(0, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_185);

  /* "ultron/tradingday/Calendar.pyx":362
 *     Date(2003, 5, 8),
 *     Date(2003, 5, 9),
 *     Date(2003, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 2),
 *     Date(2003, 10, 3),
 */
  __pyx_t_186 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__189, NULL); if (unlikely(!__pyx_t_186)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_186);

  /* "ultron/tradingday/Calendar.pyx":363
 *     Date(2003, 5, 9),
 *     Date(2003, 10, 1),
 *     Date(2003, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 3),
 *     Date(2003, 10, 6),
 */
  __pyx_t_187 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__190, NULL); if (unlikely(!__pyx_t_187)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_187);

  /* "ultron/tradingday/Calendar.pyx":364
 *     Date(2003, 10, 1),
 *     Date(2003, 10, 2),
 *     Date(2003, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 6),
 *     Date(2003, 10, 7),
 */
  __pyx_t_188 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__191, NULL); if (unlikely(!__pyx_t_188)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_188);

  /* "ultron/tradingday/Calendar.pyx":365
 *     Date(2003, 10, 2),
 *     Date(2003, 10, 3),
 *     Date(2003, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2003, 10, 7),
 *     Date(2004, 1, 1),
 */
  __pyx_t_189 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__192, NULL); if (unlikely(!__pyx_t_189)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_189);

  /* "ultron/tradingday/Calendar.pyx":366
 *     Date(2003, 10, 3),
 *     Date(2003, 10, 6),
 *     Date(2003, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 1),
 *     Date(2004, 1, 19),
 */
  __pyx_t_190 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__193, NULL); if (unlikely(!__pyx_t_190)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_190);

  /* "ultron/tradingday/Calendar.pyx":367
 *     Date(2003, 10, 6),
 *     Date(2003, 10, 7),
 *     Date(2004, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 19),
 *     Date(2004, 1, 20),
 */
  __pyx_t_191 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__194, NULL); if (unlikely(!__pyx_t_191)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_191);

  /* "ultron/tradingday/Calendar.pyx":368
 *     Date(2003, 10, 7),
 *     Date(2004, 1, 1),
 *     Date(2004, 1, 19),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 20),
 *     Date(2004, 1, 21),
 */
  __pyx_t_192 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__195, NULL); if (unlikely(!__pyx_t_192)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_192);

  /* "ultron/tradingday/Calendar.pyx":369
 *     Date(2004, 1, 1),
 *     Date(2004, 1, 19),
 *     Date(2004, 1, 20),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 21),
 *     Date(2004, 1, 22),
 */
  __pyx_t_193 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__196, NULL); if (unlikely(!__pyx_t_193)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_193);

  /* "ultron/tradingday/Calendar.pyx":370
 *     Date(2004, 1, 19),
 *     Date(2004, 1, 20),
 *     Date(2004, 1, 21),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 22),
 *     Date(2004, 1, 23),
 */
  __pyx_t_194 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__197, NULL); if (unlikely(!__pyx_t_194)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_194);

  /* "ultron/tradingday/Calendar.pyx":371
 *     Date(2004, 1, 20),
 *     Date(2004, 1, 21),
 *     Date(2004, 1, 22),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 23),
 *     Date(2004, 1, 26),
 */
  __pyx_t_195 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__198, NULL); if (unlikely(!__pyx_t_195)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_195);

  /* "ultron/tradingday/Calendar.pyx":372
 *     Date(2004, 1, 21),
 *     Date(2004, 1, 22),
 *     Date(2004, 1, 23),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 26),
 *     Date(2004, 1, 27),
 */
  __pyx_t_196 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__199, NULL); if (unlikely(!__pyx_t_196)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_196);

  /* "ultron/tradingday/Calendar.pyx":373
 *     Date(2004, 1, 22),
 *     Date(2004, 1, 23),
 *     Date(2004, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 27),
 *     Date(2004, 1, 28),
 */
  __pyx_t_197 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__200, NULL); if (unlikely(!__pyx_t_197)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_197);

  /* "ultron/tradingday/Calendar.pyx":374
 *     Date(2004, 1, 23),
 *     Date(2004, 1, 26),
 *     Date(2004, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2004, 1, 28),
 *     Date(2004, 5, 3),
 */
  __pyx_t_198 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__201, NULL); if (unlikely(!__pyx_t_198)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_198);

  /* "ultron/tradingday/Calendar.pyx":375
 *     Date(2004, 1, 26),
 *     Date(2004, 1, 27),
 *     Date(2004, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 3),
 *     Date(2004, 5, 4),
 */
  __pyx_t_199 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__202, NULL); if (unlikely(!__pyx_t_199)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_199);

  /* "ultron/tradingday/Calendar.pyx":376
 *     Date(2004, 1, 27),
 *     Date(2004, 1, 28),
 *     Date(2004, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 4),
 *     Date(2004, 5, 5),
 */
  __pyx_t_200 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__203, NULL); if (unlikely(!__pyx_t_200)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_200);

  /* "ultron/tradingday/Calendar.pyx":377
 *     Date(2004, 1, 28),
 *     Date(2004, 5, 3),
 *     Date(2004, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 5),
 *     Date(2004, 5, 6),
 */
  __pyx_t_201 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__204, NULL); if (unlikely(!__pyx_t_201)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_201);

  /* "ultron/tradingday/Calendar.pyx":378
 *     Date(2004, 5, 3),
 *     Date(2004, 5, 4),
 *     Date(2004, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 6),
 *     Date(2004, 5, 7),
 */
  __pyx_t_202 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__205, NULL); if (unlikely(!__pyx_t_202)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_202);

  /* "ultron/tradingday/Calendar.pyx":379
 *     Date(2004, 5, 4),
 *     Date(2004, 5, 5),
 *     Date(2004, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2004, 5, 7),
 *     Date(2004, 10, 1),
 */
  __pyx_t_203 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__206, NULL); if (unlikely(!__pyx_t_203)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_203);

  /* "ultron/tradingday/Calendar.pyx":380
 *     Date(2004, 5, 5),
 *     Date(2004, 5, 6),
 *     Date(2004, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 1),
 *     Date(2004, 10, 4),
 */
  __pyx_t_204 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__207, NULL); if (unlikely(!__pyx_t_204)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_204);

  /* "ultron/tradingday/Calendar.pyx":381
 *     Date(2004, 5, 6),
 *     Date(2004, 5, 7),
 *     Date(2004, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 4),
 *     Date(2004, 10, 5),
 */
  __pyx_t_205 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__208, NULL); if (unlikely(!__pyx_t_205)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_205);

  /* "ultron/tradingday/Calendar.pyx":382
 *     Date(2004, 5, 7),
 *     Date(2004, 10, 1),
 *     Date(2004, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 5),
 *     Date(2004, 10, 6),
 */
  __pyx_t_206 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__209, NULL); if (unlikely(!__pyx_t_206)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_206);

  /* "ultron/tradingday/Calendar.pyx":383
 *     Date(2004, 10, 1),
 *     Date(2004, 10, 4),
 *     Date(2004, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 6),
 *     Date(2004, 10, 7),
 */
  __pyx_t_207 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__210, NULL); if (unlikely(!__pyx_t_207)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_207);

  /* "ultron/tradingday/Calendar.pyx":384
 *     Date(2004, 10, 4),
 *     Date(2004, 10, 5),
 *     Date(2004, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2004, 10, 7),
 *     Date(2005, 1, 3),
 */
  __pyx_t_208 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__211, NULL); if (unlikely(!__pyx_t_208)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_208);

  /* "ultron/tradingday/Calendar.pyx":385
 *     Date(2004, 10, 5),
 *     Date(2004, 10, 6),
 *     Date(2004, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2005, 1, 3),
 *     Date(2005, 2, 7),
 */
  __pyx_t_209 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__212, NULL); if (unlikely(!__pyx_t_209)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_209);

  /* "ultron/tradingday/Calendar.pyx":386
 *     Date(2004, 10, 6),
 *     Date(2004, 10, 7),
 *     Date(2005, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 7),
 *     Date(2005, 2, 8),
 */
  __pyx_t_210 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__213, NULL); if (unlikely(!__pyx_t_210)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_210);

  /* "ultron/tradingday/Calendar.pyx":387
 *     Date(2004, 10, 7),
 *     Date(2005, 1, 3),
 *     Date(2005, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 8),
 *     Date(2005, 2, 9),
 */
  __pyx_t_211 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__214, NULL); if (unlikely(!__pyx_t_211)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_211);

  /* "ultron/tradingday/Calendar.pyx":388
 *     Date(2005, 1, 3),
 *     Date(2005, 2, 7),
 *     Date(2005, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 9),
 *     Date(2005, 2, 10),
 */
  __pyx_t_212 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__215, NULL); if (unlikely(!__pyx_t_212)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_212);

  /* "ultron/tradingday/Calendar.pyx":389
 *     Date(2005, 2, 7),
 *     Date(2005, 2, 8),
 *     Date(2005, 2, 9),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 10),
 *     Date(2005, 2, 11),
 */
  __pyx_t_213 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__216, NULL); if (unlikely(!__pyx_t_213)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_213);

  /* "ultron/tradingday/Calendar.pyx":390
 *     Date(2005, 2, 8),
 *     Date(2005, 2, 9),
 *     Date(2005, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 11),
 *     Date(2005, 2, 14),
 */
  __pyx_t_214 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__217, NULL); if (unlikely(!__pyx_t_214)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_214);

  /* "ultron/tradingday/Calendar.pyx":391
 *     Date(2005, 2, 9),
 *     Date(2005, 2, 10),
 *     Date(2005, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 14),
 *     Date(2005, 2, 15),
 */
  __pyx_t_215 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__218, NULL); if (unlikely(!__pyx_t_215)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_215);

  /* "ultron/tradingday/Calendar.pyx":392
 *     Date(2005, 2, 10),
 *     Date(2005, 2, 11),
 *     Date(2005, 2, 14),             # <<<<<<<<<<<<<<
 *     Date(2005, 2, 15),
 *     Date(2005, 5, 2),
 */
  __pyx_t_216 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__219, NULL); if (unlikely(!__pyx_t_216)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_216);

  /* "ultron/tradingday/Calendar.pyx":393
 *     Date(2005, 2, 11),
 *     Date(2005, 2, 14),
 *     Date(2005, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 2),
 *     Date(2005, 5, 3),
 */
  __pyx_t_217 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__220, NULL); if (unlikely(!__pyx_t_217)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_217);

  /* "ultron/tradingday/Calendar.pyx":394
 *     Date(2005, 2, 14),
 *     Date(2005, 2, 15),
 *     Date(2005, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 3),
 *     Date(2005, 5, 4),
 */
  __pyx_t_218 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__221, NULL); if (unlikely(!__pyx_t_218)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_218);

  /* "ultron/tradingday/Calendar.pyx":395
 *     Date(2005, 2, 15),
 *     Date(2005, 5, 2),
 *     Date(2005, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 4),
 *     Date(2005, 5, 5),
 */
  __pyx_t_219 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__222, NULL); if (unlikely(!__pyx_t_219)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_219);

  /* "ultron/tradingday/Calendar.pyx":396
 *     Date(2005, 5, 2),
 *     Date(2005, 5, 3),
 *     Date(2005, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 5),
 *     Date(2005, 5, 6),
 */
  __pyx_t_220 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__223, NULL); if (unlikely(!__pyx_t_220)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_220);

  /* "ultron/tradingday/Calendar.pyx":397
 *     Date(2005, 5, 3),
 *     Date(2005, 5, 4),
 *     Date(2005, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2005, 5, 6),
 *     Date(2005, 10, 3),
 */
  __pyx_t_221 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__224, NULL); if (unlikely(!__pyx_t_221)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_221);

  /* "ultron/tradingday/Calendar.pyx":398
 *     Date(2005, 5, 4),
 *     Date(2005, 5, 5),
 *     Date(2005, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 3),
 *     Date(2005, 10, 4),
 */
  __pyx_t_222 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__225, NULL); if (unlikely(!__pyx_t_222)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_222);

  /* "ultron/tradingday/Calendar.pyx":399
 *     Date(2005, 5, 5),
 *     Date(2005, 5, 6),
 *     Date(2005, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 4),
 *     Date(2005, 10, 5),
 */
  __pyx_t_223 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__226, NULL); if (unlikely(!__pyx_t_223)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_223);

  /* "ultron/tradingday/Calendar.pyx":400
 *     Date(2005, 5, 6),
 *     Date(2005, 10, 3),
 *     Date(2005, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 5),
 *     Date(2005, 10, 6),
 */
  __pyx_t_224 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__227, NULL); if (unlikely(!__pyx_t_224)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_224);

  /* "ultron/tradingday/Calendar.pyx":401
 *     Date(2005, 10, 3),
 *     Date(2005, 10, 4),
 *     Date(2005, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 6),
 *     Date(2005, 10, 7),
 */
  __pyx_t_225 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__228, NULL); if (unlikely(!__pyx_t_225)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_225);

  /* "ultron/tradingday/Calendar.pyx":402
 *     Date(2005, 10, 4),
 *     Date(2005, 10, 5),
 *     Date(2005, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2005, 10, 7),
 *     Date(2006, 1, 2),
 */
  __pyx_t_226 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__229, NULL); if (unlikely(!__pyx_t_226)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_226);

  /* "ultron/tradingday/Calendar.pyx":403
 *     Date(2005, 10, 5),
 *     Date(2005, 10, 6),
 *     Date(2005, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 2),
 *     Date(2006, 1, 3),
 */
  __pyx_t_227 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__230, NULL); if (unlikely(!__pyx_t_227)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_227);

  /* "ultron/tradingday/Calendar.pyx":404
 *     Date(2005, 10, 6),
 *     Date(2005, 10, 7),
 *     Date(2006, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 3),
 *     Date(2006, 1, 26),
 */
  __pyx_t_228 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__231, NULL); if (unlikely(!__pyx_t_228)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_228);

  /* "ultron/tradingday/Calendar.pyx":405
 *     Date(2005, 10, 7),
 *     Date(2006, 1, 2),
 *     Date(2006, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 26),
 *     Date(2006, 1, 27),
 */
  __pyx_t_229 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__232, NULL); if (unlikely(!__pyx_t_229)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_229);

  /* "ultron/tradingday/Calendar.pyx":406
 *     Date(2006, 1, 2),
 *     Date(2006, 1, 3),
 *     Date(2006, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 27),
 *     Date(2006, 1, 30),
 */
  __pyx_t_230 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__233, NULL); if (unlikely(!__pyx_t_230)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_230);

  /* "ultron/tradingday/Calendar.pyx":407
 *     Date(2006, 1, 3),
 *     Date(2006, 1, 26),
 *     Date(2006, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 30),
 *     Date(2006, 1, 31),
 */
  __pyx_t_231 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__234, NULL); if (unlikely(!__pyx_t_231)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_231);

  /* "ultron/tradingday/Calendar.pyx":408
 *     Date(2006, 1, 26),
 *     Date(2006, 1, 27),
 *     Date(2006, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2006, 1, 31),
 *     Date(2006, 2, 1),
 */
  __pyx_t_232 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__235, NULL); if (unlikely(!__pyx_t_232)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_232);

  /* "ultron/tradingday/Calendar.pyx":409
 *     Date(2006, 1, 27),
 *     Date(2006, 1, 30),
 *     Date(2006, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2006, 2, 1),
 *     Date(2006, 2, 2),
 */
  __pyx_t_233 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__236, NULL); if (unlikely(!__pyx_t_233)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_233);

  /* "ultron/tradingday/Calendar.pyx":410
 *     Date(2006, 1, 30),
 *     Date(2006, 1, 31),
 *     Date(2006, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2006, 2, 2),
 *     Date(2006, 2, 3),
 */
  __pyx_t_234 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__237, NULL); if (unlikely(!__pyx_t_234)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_234);

  /* "ultron/tradingday/Calendar.pyx":411
 *     Date(2006, 1, 31),
 *     Date(2006, 2, 1),
 *     Date(2006, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2006, 2, 3),
 *     Date(2006, 5, 1),
 */
  __pyx_t_235 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__238, NULL); if (unlikely(!__pyx_t_235)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_235);

  /* "ultron/tradingday/Calendar.pyx":412
 *     Date(2006, 2, 1),
 *     Date(2006, 2, 2),
 *     Date(2006, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 1),
 *     Date(2006, 5, 2),
 */
  __pyx_t_236 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__239, NULL); if (unlikely(!__pyx_t_236)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_236);

  /* "ultron/tradingday/Calendar.pyx":413
 *     Date(2006, 2, 2),
 *     Date(2006, 2, 3),
 *     Date(2006, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 2),
 *     Date(2006, 5, 3),
 */
  __pyx_t_237 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__240, NULL); if (unlikely(!__pyx_t_237)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_237);

  /* "ultron/tradingday/Calendar.pyx":414
 *     Date(2006, 2, 3),
 *     Date(2006, 5, 1),
 *     Date(2006, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 3),
 *     Date(2006, 5, 4),
 */
  __pyx_t_238 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__241, NULL); if (unlikely(!__pyx_t_238)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_238);

  /* "ultron/tradingday/Calendar.pyx":415
 *     Date(2006, 5, 1),
 *     Date(2006, 5, 2),
 *     Date(2006, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 4),
 *     Date(2006, 5, 5),
 */
  __pyx_t_239 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__242, NULL); if (unlikely(!__pyx_t_239)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_239);

  /* "ultron/tradingday/Calendar.pyx":416
 *     Date(2006, 5, 2),
 *     Date(2006, 5, 3),
 *     Date(2006, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2006, 5, 5),
 *     Date(2006, 10, 2),
 */
  __pyx_t_240 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__243, NULL); if (unlikely(!__pyx_t_240)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_240);

  /* "ultron/tradingday/Calendar.pyx":417
 *     Date(2006, 5, 3),
 *     Date(2006, 5, 4),
 *     Date(2006, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 2),
 *     Date(2006, 10, 3),
 */
  __pyx_t_241 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__244, NULL); if (unlikely(!__pyx_t_241)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_241);

  /* "ultron/tradingday/Calendar.pyx":418
 *     Date(2006, 5, 4),
 *     Date(2006, 5, 5),
 *     Date(2006, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 3),
 *     Date(2006, 10, 4),
 */
  __pyx_t_242 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__245, NULL); if (unlikely(!__pyx_t_242)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_242);

  /* "ultron/tradingday/Calendar.pyx":419
 *     Date(2006, 5, 5),
 *     Date(2006, 10, 2),
 *     Date(2006, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 4),
 *     Date(2006, 10, 5),
 */
  __pyx_t_243 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__246, NULL); if (unlikely(!__pyx_t_243)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_243);

  /* "ultron/tradingday/Calendar.pyx":420
 *     Date(2006, 10, 2),
 *     Date(2006, 10, 3),
 *     Date(2006, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 5),
 *     Date(2006, 10, 6),
 */
  __pyx_t_244 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__247, NULL); if (unlikely(!__pyx_t_244)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_244);

  /* "ultron/tradingday/Calendar.pyx":421
 *     Date(2006, 10, 3),
 *     Date(2006, 10, 4),
 *     Date(2006, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2006, 10, 6),
 *     Date(2007, 1, 1),
 */
  __pyx_t_245 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__248, NULL); if (unlikely(!__pyx_t_245)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_245);

  /* "ultron/tradingday/Calendar.pyx":422
 *     Date(2006, 10, 4),
 *     Date(2006, 10, 5),
 *     Date(2006, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2007, 1, 1),
 *     Date(2007, 1, 2),
 */
  __pyx_t_246 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__249, NULL); if (unlikely(!__pyx_t_246)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_246);

  /* "ultron/tradingday/Calendar.pyx":423
 *     Date(2006, 10, 5),
 *     Date(2006, 10, 6),
 *     Date(2007, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2007, 1, 2),
 *     Date(2007, 1, 3),
 */
  __pyx_t_247 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__250, NULL); if (unlikely(!__pyx_t_247)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_247);

  /* "ultron/tradingday/Calendar.pyx":424
 *     Date(2006, 10, 6),
 *     Date(2007, 1, 1),
 *     Date(2007, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2007, 1, 3),
 *     Date(2007, 2, 19),
 */
  __pyx_t_248 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__251, NULL); if (unlikely(!__pyx_t_248)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_248);

  /* "ultron/tradingday/Calendar.pyx":425
 *     Date(2007, 1, 1),
 *     Date(2007, 1, 2),
 *     Date(2007, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 19),
 *     Date(2007, 2, 20),
 */
  __pyx_t_249 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__252, NULL); if (unlikely(!__pyx_t_249)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_249);

  /* "ultron/tradingday/Calendar.pyx":426
 *     Date(2007, 1, 2),
 *     Date(2007, 1, 3),
 *     Date(2007, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 20),
 *     Date(2007, 2, 21),
 */
  __pyx_t_250 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__253, NULL); if (unlikely(!__pyx_t_250)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_250);

  /* "ultron/tradingday/Calendar.pyx":427
 *     Date(2007, 1, 3),
 *     Date(2007, 2, 19),
 *     Date(2007, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 21),
 *     Date(2007, 2, 22),
 */
  __pyx_t_251 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__254, NULL); if (unlikely(!__pyx_t_251)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_251);

  /* "ultron/tradingday/Calendar.pyx":428
 *     Date(2007, 2, 19),
 *     Date(2007, 2, 20),
 *     Date(2007, 2, 21),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 22),
 *     Date(2007, 2, 23),
 */
  __pyx_t_252 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__255, NULL); if (unlikely(!__pyx_t_252)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_252);

  /* "ultron/tradingday/Calendar.pyx":429
 *     Date(2007, 2, 20),
 *     Date(2007, 2, 21),
 *     Date(2007, 2, 22),             # <<<<<<<<<<<<<<
 *     Date(2007, 2, 23),
 *     Date(2007, 5, 1),
 */
  __pyx_t_253 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__256, NULL); if (unlikely(!__pyx_t_253)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_253);

  /* "ultron/tradingday/Calendar.pyx":430
 *     Date(2007, 2, 21),
 *     Date(2007, 2, 22),
 *     Date(2007, 2, 23),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 1),
 *     Date(2007, 5, 2),
 */
  __pyx_t_254 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__257, NULL); if (unlikely(!__pyx_t_254)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_254);

  /* "ultron/tradingday/Calendar.pyx":431
 *     Date(2007, 2, 22),
 *     Date(2007, 2, 23),
 *     Date(2007, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 2),
 *     Date(2007, 5, 3),
 */
  __pyx_t_255 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__258, NULL); if (unlikely(!__pyx_t_255)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_255);

  /* "ultron/tradingday/Calendar.pyx":432
 *     Date(2007, 2, 23),
 *     Date(2007, 5, 1),
 *     Date(2007, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 3),
 *     Date(2007, 5, 4),
 */
  __pyx_t_256 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__259, NULL); if (unlikely(!__pyx_t_256)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_256);

  /* "ultron/tradingday/Calendar.pyx":433
 *     Date(2007, 5, 1),
 *     Date(2007, 5, 2),
 *     Date(2007, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 4),
 *     Date(2007, 5, 7),
 */
  __pyx_t_257 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__260, NULL); if (unlikely(!__pyx_t_257)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_257);

  /* "ultron/tradingday/Calendar.pyx":434
 *     Date(2007, 5, 2),
 *     Date(2007, 5, 3),
 *     Date(2007, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2007, 5, 7),
 *     Date(2007, 10, 1),
 */
  __pyx_t_258 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__261, NULL); if (unlikely(!__pyx_t_258)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_258);

  /* "ultron/tradingday/Calendar.pyx":435
 *     Date(2007, 5, 3),
 *     Date(2007, 5, 4),
 *     Date(2007, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 1),
 *     Date(2007, 10, 2),
 */
  __pyx_t_259 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__262, NULL); if (unlikely(!__pyx_t_259)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_259);

  /* "ultron/tradingday/Calendar.pyx":436
 *     Date(2007, 5, 4),
 *     Date(2007, 5, 7),
 *     Date(2007, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 2),
 *     Date(2007, 10, 3),
 */
  __pyx_t_260 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__263, NULL); if (unlikely(!__pyx_t_260)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_260);

  /* "ultron/tradingday/Calendar.pyx":437
 *     Date(2007, 5, 7),
 *     Date(2007, 10, 1),
 *     Date(2007, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 3),
 *     Date(2007, 10, 4),
 */
  __pyx_t_261 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__264, NULL); if (unlikely(!__pyx_t_261)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_261);

  /* "ultron/tradingday/Calendar.pyx":438
 *     Date(2007, 10, 1),
 *     Date(2007, 10, 2),
 *     Date(2007, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 4),
 *     Date(2007, 10, 5),
 */
  __pyx_t_262 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__265, NULL); if (unlikely(!__pyx_t_262)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_262);

  /* "ultron/tradingday/Calendar.pyx":439
 *     Date(2007, 10, 2),
 *     Date(2007, 10, 3),
 *     Date(2007, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2007, 10, 5),
 *     Date(2007, 12, 31),
 */
  __pyx_t_263 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__266, NULL); if (unlikely(!__pyx_t_263)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_263);

  /* "ultron/tradingday/Calendar.pyx":440
 *     Date(2007, 10, 3),
 *     Date(2007, 10, 4),
 *     Date(2007, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2007, 12, 31),
 *     Date(2008, 1, 1),
 */
  __pyx_t_264 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__267, NULL); if (unlikely(!__pyx_t_264)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_264);

  /* "ultron/tradingday/Calendar.pyx":441
 *     Date(2007, 10, 4),
 *     Date(2007, 10, 5),
 *     Date(2007, 12, 31),             # <<<<<<<<<<<<<<
 *     Date(2008, 1, 1),
 *     Date(2008, 2, 6),
 */
  __pyx_t_265 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__268, NULL); if (unlikely(!__pyx_t_265)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_265);

  /* "ultron/tradingday/Calendar.pyx":442
 *     Date(2007, 10, 5),
 *     Date(2007, 12, 31),
 *     Date(2008, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 6),
 *     Date(2008, 2, 7),
 */
  __pyx_t_266 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__269, NULL); if (unlikely(!__pyx_t_266)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_266);

  /* "ultron/tradingday/Calendar.pyx":443
 *     Date(2007, 12, 31),
 *     Date(2008, 1, 1),
 *     Date(2008, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 7),
 *     Date(2008, 2, 8),
 */
  __pyx_t_267 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__270, NULL); if (unlikely(!__pyx_t_267)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_267);

  /* "ultron/tradingday/Calendar.pyx":444
 *     Date(2008, 1, 1),
 *     Date(2008, 2, 6),
 *     Date(2008, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 8),
 *     Date(2008, 2, 11),
 */
  __pyx_t_268 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__271, NULL); if (unlikely(!__pyx_t_268)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_268);

  /* "ultron/tradingday/Calendar.pyx":445
 *     Date(2008, 2, 6),
 *     Date(2008, 2, 7),
 *     Date(2008, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 11),
 *     Date(2008, 2, 12),
 */
  __pyx_t_269 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__272, NULL); if (unlikely(!__pyx_t_269)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_269);

  /* "ultron/tradingday/Calendar.pyx":446
 *     Date(2008, 2, 7),
 *     Date(2008, 2, 8),
 *     Date(2008, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2008, 2, 12),
 *     Date(2008, 4, 4),
 */
  __pyx_t_270 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__273, NULL); if (unlikely(!__pyx_t_270)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_270);

  /* "ultron/tradingday/Calendar.pyx":447
 *     Date(2008, 2, 8),
 *     Date(2008, 2, 11),
 *     Date(2008, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2008, 4, 4),
 *     Date(2008, 5, 1),
 */
  __pyx_t_271 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__274, NULL); if (unlikely(!__pyx_t_271)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_271);

  /* "ultron/tradingday/Calendar.pyx":448
 *     Date(2008, 2, 11),
 *     Date(2008, 2, 12),
 *     Date(2008, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2008, 5, 1),
 *     Date(2008, 5, 2),
 */
  __pyx_t_272 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__275, NULL); if (unlikely(!__pyx_t_272)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_272);

  /* "ultron/tradingday/Calendar.pyx":449
 *     Date(2008, 2, 12),
 *     Date(2008, 4, 4),
 *     Date(2008, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2008, 5, 2),
 *     Date(2008, 6, 9),
 */
  __pyx_t_273 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__276, NULL); if (unlikely(!__pyx_t_273)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_273);

  /* "ultron/tradingday/Calendar.pyx":450
 *     Date(2008, 4, 4),
 *     Date(2008, 5, 1),
 *     Date(2008, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2008, 6, 9),
 *     Date(2008, 9, 15),
 */
  __pyx_t_274 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__277, NULL); if (unlikely(!__pyx_t_274)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_274);

  /* "ultron/tradingday/Calendar.pyx":451
 *     Date(2008, 5, 1),
 *     Date(2008, 5, 2),
 *     Date(2008, 6, 9),             # <<<<<<<<<<<<<<
 *     Date(2008, 9, 15),
 *     Date(2008, 9, 29),
 */
  __pyx_t_275 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__278, NULL); if (unlikely(!__pyx_t_275)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_275);

  /* "ultron/tradingday/Calendar.pyx":452
 *     Date(2008, 5, 2),
 *     Date(2008, 6, 9),
 *     Date(2008, 9, 15),             # <<<<<<<<<<<<<<
 *     Date(2008, 9, 29),
 *     Date(2008, 9, 30),
 */
  __pyx_t_276 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__279, NULL); if (unlikely(!__pyx_t_276)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_276);

  /* "ultron/tradingday/Calendar.pyx":453
 *     Date(2008, 6, 9),
 *     Date(2008, 9, 15),
 *     Date(2008, 9, 29),             # <<<<<<<<<<<<<<
 *     Date(2008, 9, 30),
 *     Date(2008, 10, 1),
 */
  __pyx_t_277 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__280, NULL); if (unlikely(!__pyx_t_277)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_277);

  /* "ultron/tradingday/Calendar.pyx":454
 *     Date(2008, 9, 15),
 *     Date(2008, 9, 29),
 *     Date(2008, 9, 30),             # <<<<<<<<<<<<<<
 *     Date(2008, 10, 1),
 *     Date(2008, 10, 2),
 */
  __pyx_t_278 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__281, NULL); if (unlikely(!__pyx_t_278)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_278);

  /* "ultron/tradingday/Calendar.pyx":455
 *     Date(2008, 9, 29),
 *     Date(2008, 9, 30),
 *     Date(2008, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2008, 10, 2),
 *     Date(2008, 10, 3),
 */
  __pyx_t_279 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__282, NULL); if (unlikely(!__pyx_t_279)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_279);

  /* "ultron/tradingday/Calendar.pyx":456
 *     Date(2008, 9, 30),
 *     Date(2008, 10, 1),
 *     Date(2008, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2008, 10, 3),
 *     Date(2009, 1, 1),
 */
  __pyx_t_280 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__283, NULL); if (unlikely(!__pyx_t_280)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_280);

  /* "ultron/tradingday/Calendar.pyx":457
 *     Date(2008, 10, 1),
 *     Date(2008, 10, 2),
 *     Date(2008, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 1),
 *     Date(2009, 1, 2),
 */
  __pyx_t_281 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__284, NULL); if (unlikely(!__pyx_t_281)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_281);

  /* "ultron/tradingday/Calendar.pyx":458
 *     Date(2008, 10, 2),
 *     Date(2008, 10, 3),
 *     Date(2009, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 2),
 *     Date(2009, 1, 26),
 */
  __pyx_t_282 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__285, NULL); if (unlikely(!__pyx_t_282)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_282);

  /* "ultron/tradingday/Calendar.pyx":459
 *     Date(2008, 10, 3),
 *     Date(2009, 1, 1),
 *     Date(2009, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 26),
 *     Date(2009, 1, 27),
 */
  __pyx_t_283 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__286, NULL); if (unlikely(!__pyx_t_283)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_283);

  /* "ultron/tradingday/Calendar.pyx":460
 *     Date(2009, 1, 1),
 *     Date(2009, 1, 2),
 *     Date(2009, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 27),
 *     Date(2009, 1, 28),
 */
  __pyx_t_284 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__287, NULL); if (unlikely(!__pyx_t_284)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_284);

  /* "ultron/tradingday/Calendar.pyx":461
 *     Date(2009, 1, 2),
 *     Date(2009, 1, 26),
 *     Date(2009, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 28),
 *     Date(2009, 1, 29),
 */
  __pyx_t_285 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__288, NULL); if (unlikely(!__pyx_t_285)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_285);

  /* "ultron/tradingday/Calendar.pyx":462
 *     Date(2009, 1, 26),
 *     Date(2009, 1, 27),
 *     Date(2009, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 29),
 *     Date(2009, 1, 30),
 */
  __pyx_t_286 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__289, NULL); if (unlikely(!__pyx_t_286)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_286);

  /* "ultron/tradingday/Calendar.pyx":463
 *     Date(2009, 1, 27),
 *     Date(2009, 1, 28),
 *     Date(2009, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2009, 1, 30),
 *     Date(2009, 4, 6),
 */
  __pyx_t_287 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__290, NULL); if (unlikely(!__pyx_t_287)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_287);

  /* "ultron/tradingday/Calendar.pyx":464
 *     Date(2009, 1, 28),
 *     Date(2009, 1, 29),
 *     Date(2009, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2009, 4, 6),
 *     Date(2009, 5, 1),
 */
  __pyx_t_288 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__291, NULL); if (unlikely(!__pyx_t_288)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_288);

  /* "ultron/tradingday/Calendar.pyx":465
 *     Date(2009, 1, 29),
 *     Date(2009, 1, 30),
 *     Date(2009, 4, 6),             # <<<<<<<<<<<<<<
 *     Date(2009, 5, 1),
 *     Date(2009, 5, 28),
 */
  __pyx_t_289 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__292, NULL); if (unlikely(!__pyx_t_289)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_289);

  /* "ultron/tradingday/Calendar.pyx":466
 *     Date(2009, 1, 30),
 *     Date(2009, 4, 6),
 *     Date(2009, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2009, 5, 28),
 *     Date(2009, 5, 29),
 */
  __pyx_t_290 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__293, NULL); if (unlikely(!__pyx_t_290)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_290);

  /* "ultron/tradingday/Calendar.pyx":467
 *     Date(2009, 4, 6),
 *     Date(2009, 5, 1),
 *     Date(2009, 5, 28),             # <<<<<<<<<<<<<<
 *     Date(2009, 5, 29),
 *     Date(2009, 10, 1),
 */
  __pyx_t_291 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__294, NULL); if (unlikely(!__pyx_t_291)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_291);

  /* "ultron/tradingday/Calendar.pyx":468
 *     Date(2009, 5, 1),
 *     Date(2009, 5, 28),
 *     Date(2009, 5, 29),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 1),
 *     Date(2009, 10, 2),
 */
  __pyx_t_292 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__295, NULL); if (unlikely(!__pyx_t_292)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_292);

  /* "ultron/tradingday/Calendar.pyx":469
 *     Date(2009, 5, 28),
 *     Date(2009, 5, 29),
 *     Date(2009, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 2),
 *     Date(2009, 10, 5),
 */
  __pyx_t_293 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__296, NULL); if (unlikely(!__pyx_t_293)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_293);

  /* "ultron/tradingday/Calendar.pyx":470
 *     Date(2009, 5, 29),
 *     Date(2009, 10, 1),
 *     Date(2009, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 5),
 *     Date(2009, 10, 6),
 */
  __pyx_t_294 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__297, NULL); if (unlikely(!__pyx_t_294)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_294);

  /* "ultron/tradingday/Calendar.pyx":471
 *     Date(2009, 10, 1),
 *     Date(2009, 10, 2),
 *     Date(2009, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 6),
 *     Date(2009, 10, 7),
 */
  __pyx_t_295 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__298, NULL); if (unlikely(!__pyx_t_295)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_295);

  /* "ultron/tradingday/Calendar.pyx":472
 *     Date(2009, 10, 2),
 *     Date(2009, 10, 5),
 *     Date(2009, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 7),
 *     Date(2009, 10, 8),
 */
  __pyx_t_296 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__299, NULL); if (unlikely(!__pyx_t_296)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_296);

  /* "ultron/tradingday/Calendar.pyx":473
 *     Date(2009, 10, 5),
 *     Date(2009, 10, 6),
 *     Date(2009, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2009, 10, 8),
 *     Date(2010, 1, 1),
 */
  __pyx_t_297 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__300, NULL); if (unlikely(!__pyx_t_297)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_297);

  /* "ultron/tradingday/Calendar.pyx":474
 *     Date(2009, 10, 6),
 *     Date(2009, 10, 7),
 *     Date(2009, 10, 8),             # <<<<<<<<<<<<<<
 *     Date(2010, 1, 1),
 *     Date(2010, 2, 15),
 */
  __pyx_t_298 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__301, NULL); if (unlikely(!__pyx_t_298)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_298);

  /* "ultron/tradingday/Calendar.pyx":475
 *     Date(2009, 10, 7),
 *     Date(2009, 10, 8),
 *     Date(2010, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 15),
 *     Date(2010, 2, 16),
 */
  __pyx_t_299 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__302, NULL); if (unlikely(!__pyx_t_299)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_299);

  /* "ultron/tradingday/Calendar.pyx":476
 *     Date(2009, 10, 8),
 *     Date(2010, 1, 1),
 *     Date(2010, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 16),
 *     Date(2010, 2, 17),
 */
  __pyx_t_300 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__303, NULL); if (unlikely(!__pyx_t_300)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_300);

  /* "ultron/tradingday/Calendar.pyx":477
 *     Date(2010, 1, 1),
 *     Date(2010, 2, 15),
 *     Date(2010, 2, 16),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 17),
 *     Date(2010, 2, 18),
 */
  __pyx_t_301 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__304, NULL); if (unlikely(!__pyx_t_301)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_301);

  /* "ultron/tradingday/Calendar.pyx":478
 *     Date(2010, 2, 15),
 *     Date(2010, 2, 16),
 *     Date(2010, 2, 17),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 18),
 *     Date(2010, 2, 19),
 */
  __pyx_t_302 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__305, NULL); if (unlikely(!__pyx_t_302)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_302);

  /* "ultron/tradingday/Calendar.pyx":479
 *     Date(2010, 2, 16),
 *     Date(2010, 2, 17),
 *     Date(2010, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(2010, 2, 19),
 *     Date(2010, 4, 5),
 */
  __pyx_t_303 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__306, NULL); if (unlikely(!__pyx_t_303)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_303);

  /* "ultron/tradingday/Calendar.pyx":480
 *     Date(2010, 2, 17),
 *     Date(2010, 2, 18),
 *     Date(2010, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2010, 4, 5),
 *     Date(2010, 5, 3),
 */
  __pyx_t_304 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__307, NULL); if (unlikely(!__pyx_t_304)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_304);

  /* "ultron/tradingday/Calendar.pyx":481
 *     Date(2010, 2, 18),
 *     Date(2010, 2, 19),
 *     Date(2010, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2010, 5, 3),
 *     Date(2010, 6, 14),
 */
  __pyx_t_305 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__308, NULL); if (unlikely(!__pyx_t_305)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_305);

  /* "ultron/tradingday/Calendar.pyx":482
 *     Date(2010, 2, 19),
 *     Date(2010, 4, 5),
 *     Date(2010, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2010, 6, 14),
 *     Date(2010, 6, 15),
 */
  __pyx_t_306 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__309, NULL); if (unlikely(!__pyx_t_306)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_306);

  /* "ultron/tradingday/Calendar.pyx":483
 *     Date(2010, 4, 5),
 *     Date(2010, 5, 3),
 *     Date(2010, 6, 14),             # <<<<<<<<<<<<<<
 *     Date(2010, 6, 15),
 *     Date(2010, 6, 16),
 */
  __pyx_t_307 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__310, NULL); if (unlikely(!__pyx_t_307)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_307);

  /* "ultron/tradingday/Calendar.pyx":484
 *     Date(2010, 5, 3),
 *     Date(2010, 6, 14),
 *     Date(2010, 6, 15),             # <<<<<<<<<<<<<<
 *     Date(2010, 6, 16),
 *     Date(2010, 9, 22),
 */
  __pyx_t_308 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__311, NULL); if (unlikely(!__pyx_t_308)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_308);

  /* "ultron/tradingday/Calendar.pyx":485
 *     Date(2010, 6, 14),
 *     Date(2010, 6, 15),
 *     Date(2010, 6, 16),             # <<<<<<<<<<<<<<
 *     Date(2010, 9, 22),
 *     Date(2010, 9, 23),
 */
  __pyx_t_309 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__312, NULL); if (unlikely(!__pyx_t_309)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_309);

  /* "ultron/tradingday/Calendar.pyx":486
 *     Date(2010, 6, 15),
 *     Date(2010, 6, 16),
 *     Date(2010, 9, 22),             # <<<<<<<<<<<<<<
 *     Date(2010, 9, 23),
 *     Date(2010, 9, 24),
 */
  __pyx_t_310 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__313, NULL); if (unlikely(!__pyx_t_310)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_310);

  /* "ultron/tradingday/Calendar.pyx":487
 *     Date(2010, 6, 16),
 *     Date(2010, 9, 22),
 *     Date(2010, 9, 23),             # <<<<<<<<<<<<<<
 *     Date(2010, 9, 24),
 *     Date(2010, 10, 1),
 */
  __pyx_t_311 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__314, NULL); if (unlikely(!__pyx_t_311)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_311);

  /* "ultron/tradingday/Calendar.pyx":488
 *     Date(2010, 9, 22),
 *     Date(2010, 9, 23),
 *     Date(2010, 9, 24),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 1),
 *     Date(2010, 10, 4),
 */
  __pyx_t_312 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__315, NULL); if (unlikely(!__pyx_t_312)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_312);

  /* "ultron/tradingday/Calendar.pyx":489
 *     Date(2010, 9, 23),
 *     Date(2010, 9, 24),
 *     Date(2010, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 4),
 *     Date(2010, 10, 5),
 */
  __pyx_t_313 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__316, NULL); if (unlikely(!__pyx_t_313)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_313);

  /* "ultron/tradingday/Calendar.pyx":490
 *     Date(2010, 9, 24),
 *     Date(2010, 10, 1),
 *     Date(2010, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 5),
 *     Date(2010, 10, 6),
 */
  __pyx_t_314 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__317, NULL); if (unlikely(!__pyx_t_314)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_314);

  /* "ultron/tradingday/Calendar.pyx":491
 *     Date(2010, 10, 1),
 *     Date(2010, 10, 4),
 *     Date(2010, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 6),
 *     Date(2010, 10, 7),
 */
  __pyx_t_315 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__318, NULL); if (unlikely(!__pyx_t_315)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_315);

  /* "ultron/tradingday/Calendar.pyx":492
 *     Date(2010, 10, 4),
 *     Date(2010, 10, 5),
 *     Date(2010, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2010, 10, 7),
 *     Date(2011, 1, 3),
 */
  __pyx_t_316 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__319, NULL); if (unlikely(!__pyx_t_316)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_316);

  /* "ultron/tradingday/Calendar.pyx":493
 *     Date(2010, 10, 5),
 *     Date(2010, 10, 6),
 *     Date(2010, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2011, 1, 3),
 *     Date(2011, 2, 2),
 */
  __pyx_t_317 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__320, NULL); if (unlikely(!__pyx_t_317)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_317);

  /* "ultron/tradingday/Calendar.pyx":494
 *     Date(2010, 10, 6),
 *     Date(2010, 10, 7),
 *     Date(2011, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 2),
 *     Date(2011, 2, 3),
 */
  __pyx_t_318 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__321, NULL); if (unlikely(!__pyx_t_318)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_318);

  /* "ultron/tradingday/Calendar.pyx":495
 *     Date(2010, 10, 7),
 *     Date(2011, 1, 3),
 *     Date(2011, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 3),
 *     Date(2011, 2, 4),
 */
  __pyx_t_319 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__322, NULL); if (unlikely(!__pyx_t_319)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_319);

  /* "ultron/tradingday/Calendar.pyx":496
 *     Date(2011, 1, 3),
 *     Date(2011, 2, 2),
 *     Date(2011, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 4),
 *     Date(2011, 2, 7),
 */
  __pyx_t_320 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__323, NULL); if (unlikely(!__pyx_t_320)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_320);

  /* "ultron/tradingday/Calendar.pyx":497
 *     Date(2011, 2, 2),
 *     Date(2011, 2, 3),
 *     Date(2011, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 7),
 *     Date(2011, 2, 8),
 */
  __pyx_t_321 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__324, NULL); if (unlikely(!__pyx_t_321)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_321);

  /* "ultron/tradingday/Calendar.pyx":498
 *     Date(2011, 2, 3),
 *     Date(2011, 2, 4),
 *     Date(2011, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2011, 2, 8),
 *     Date(2011, 4, 4),
 */
  __pyx_t_322 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__325, NULL); if (unlikely(!__pyx_t_322)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_322);

  /* "ultron/tradingday/Calendar.pyx":499
 *     Date(2011, 2, 4),
 *     Date(2011, 2, 7),
 *     Date(2011, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2011, 4, 4),
 *     Date(2011, 4, 5),
 */
  __pyx_t_323 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__326, NULL); if (unlikely(!__pyx_t_323)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_323);

  /* "ultron/tradingday/Calendar.pyx":500
 *     Date(2011, 2, 7),
 *     Date(2011, 2, 8),
 *     Date(2011, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2011, 4, 5),
 *     Date(2011, 5, 2),
 */
  __pyx_t_324 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__327, NULL); if (unlikely(!__pyx_t_324)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_324);

  /* "ultron/tradingday/Calendar.pyx":501
 *     Date(2011, 2, 8),
 *     Date(2011, 4, 4),
 *     Date(2011, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2011, 5, 2),
 *     Date(2011, 6, 6),
 */
  __pyx_t_325 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__328, NULL); if (unlikely(!__pyx_t_325)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_325);

  /* "ultron/tradingday/Calendar.pyx":502
 *     Date(2011, 4, 4),
 *     Date(2011, 4, 5),
 *     Date(2011, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2011, 6, 6),
 *     Date(2011, 9, 12),
 */
  __pyx_t_326 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__329, NULL); if (unlikely(!__pyx_t_326)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_326);

  /* "ultron/tradingday/Calendar.pyx":503
 *     Date(2011, 4, 5),
 *     Date(2011, 5, 2),
 *     Date(2011, 6, 6),             # <<<<<<<<<<<<<<
 *     Date(2011, 9, 12),
 *     Date(2011, 10, 3),
 */
  __pyx_t_327 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__330, NULL); if (unlikely(!__pyx_t_327)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_327);

  /* "ultron/tradingday/Calendar.pyx":504
 *     Date(2011, 5, 2),
 *     Date(2011, 6, 6),
 *     Date(2011, 9, 12),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 3),
 *     Date(2011, 10, 4),
 */
  __pyx_t_328 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__331, NULL); if (unlikely(!__pyx_t_328)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_328);

  /* "ultron/tradingday/Calendar.pyx":505
 *     Date(2011, 6, 6),
 *     Date(2011, 9, 12),
 *     Date(2011, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 4),
 *     Date(2011, 10, 5),
 */
  __pyx_t_329 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__332, NULL); if (unlikely(!__pyx_t_329)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_329);

  /* "ultron/tradingday/Calendar.pyx":506
 *     Date(2011, 9, 12),
 *     Date(2011, 10, 3),
 *     Date(2011, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 5),
 *     Date(2011, 10, 6),
 */
  __pyx_t_330 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__333, NULL); if (unlikely(!__pyx_t_330)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_330);

  /* "ultron/tradingday/Calendar.pyx":507
 *     Date(2011, 10, 3),
 *     Date(2011, 10, 4),
 *     Date(2011, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 6),
 *     Date(2011, 10, 7),
 */
  __pyx_t_331 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__334, NULL); if (unlikely(!__pyx_t_331)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_331);

  /* "ultron/tradingday/Calendar.pyx":508
 *     Date(2011, 10, 4),
 *     Date(2011, 10, 5),
 *     Date(2011, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2011, 10, 7),
 *     Date(2012, 1, 2),
 */
  __pyx_t_332 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__335, NULL); if (unlikely(!__pyx_t_332)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_332);

  /* "ultron/tradingday/Calendar.pyx":509
 *     Date(2011, 10, 5),
 *     Date(2011, 10, 6),
 *     Date(2011, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 2),
 *     Date(2012, 1, 3),
 */
  __pyx_t_333 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__336, NULL); if (unlikely(!__pyx_t_333)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_333);

  /* "ultron/tradingday/Calendar.pyx":510
 *     Date(2011, 10, 6),
 *     Date(2011, 10, 7),
 *     Date(2012, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 3),
 *     Date(2012, 1, 23),
 */
  __pyx_t_334 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__337, NULL); if (unlikely(!__pyx_t_334)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_334);

  /* "ultron/tradingday/Calendar.pyx":511
 *     Date(2011, 10, 7),
 *     Date(2012, 1, 2),
 *     Date(2012, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 23),
 *     Date(2012, 1, 24),
 */
  __pyx_t_335 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__338, NULL); if (unlikely(!__pyx_t_335)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_335);

  /* "ultron/tradingday/Calendar.pyx":512
 *     Date(2012, 1, 2),
 *     Date(2012, 1, 3),
 *     Date(2012, 1, 23),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 24),
 *     Date(2012, 1, 25),
 */
  __pyx_t_336 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__339, NULL); if (unlikely(!__pyx_t_336)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_336);

  /* "ultron/tradingday/Calendar.pyx":513
 *     Date(2012, 1, 3),
 *     Date(2012, 1, 23),
 *     Date(2012, 1, 24),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 25),
 *     Date(2012, 1, 26),
 */
  __pyx_t_337 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__340, NULL); if (unlikely(!__pyx_t_337)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_337);

  /* "ultron/tradingday/Calendar.pyx":514
 *     Date(2012, 1, 23),
 *     Date(2012, 1, 24),
 *     Date(2012, 1, 25),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 26),
 *     Date(2012, 1, 27),
 */
  __pyx_t_338 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__341, NULL); if (unlikely(!__pyx_t_338)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_338);

  /* "ultron/tradingday/Calendar.pyx":515
 *     Date(2012, 1, 24),
 *     Date(2012, 1, 25),
 *     Date(2012, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2012, 1, 27),
 *     Date(2012, 4, 2),
 */
  __pyx_t_339 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__342, NULL); if (unlikely(!__pyx_t_339)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_339);

  /* "ultron/tradingday/Calendar.pyx":516
 *     Date(2012, 1, 25),
 *     Date(2012, 1, 26),
 *     Date(2012, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2012, 4, 2),
 *     Date(2012, 4, 3),
 */
  __pyx_t_340 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__343, NULL); if (unlikely(!__pyx_t_340)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_340);

  /* "ultron/tradingday/Calendar.pyx":517
 *     Date(2012, 1, 26),
 *     Date(2012, 1, 27),
 *     Date(2012, 4, 2),             # <<<<<<<<<<<<<<
 *     Date(2012, 4, 3),
 *     Date(2012, 4, 4),
 */
  __pyx_t_341 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__344, NULL); if (unlikely(!__pyx_t_341)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_341);

  /* "ultron/tradingday/Calendar.pyx":518
 *     Date(2012, 1, 27),
 *     Date(2012, 4, 2),
 *     Date(2012, 4, 3),             # <<<<<<<<<<<<<<
 *     Date(2012, 4, 4),
 *     Date(2012, 4, 30),
 */
  __pyx_t_342 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__345, NULL); if (unlikely(!__pyx_t_342)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_342);

  /* "ultron/tradingday/Calendar.pyx":519
 *     Date(2012, 4, 2),
 *     Date(2012, 4, 3),
 *     Date(2012, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2012, 4, 30),
 *     Date(2012, 5, 1),
 */
  __pyx_t_343 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__346, NULL); if (unlikely(!__pyx_t_343)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_343);

  /* "ultron/tradingday/Calendar.pyx":520
 *     Date(2012, 4, 3),
 *     Date(2012, 4, 4),
 *     Date(2012, 4, 30),             # <<<<<<<<<<<<<<
 *     Date(2012, 5, 1),
 *     Date(2012, 6, 22),
 */
  __pyx_t_344 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__347, NULL); if (unlikely(!__pyx_t_344)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_344);

  /* "ultron/tradingday/Calendar.pyx":521
 *     Date(2012, 4, 4),
 *     Date(2012, 4, 30),
 *     Date(2012, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2012, 6, 22),
 *     Date(2012, 10, 1),
 */
  __pyx_t_345 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__348, NULL); if (unlikely(!__pyx_t_345)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_345);

  /* "ultron/tradingday/Calendar.pyx":522
 *     Date(2012, 4, 30),
 *     Date(2012, 5, 1),
 *     Date(2012, 6, 22),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 1),
 *     Date(2012, 10, 2),
 */
  __pyx_t_346 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__349, NULL); if (unlikely(!__pyx_t_346)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_346);

  /* "ultron/tradingday/Calendar.pyx":523
 *     Date(2012, 5, 1),
 *     Date(2012, 6, 22),
 *     Date(2012, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 2),
 *     Date(2012, 10, 3),
 */
  __pyx_t_347 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__350, NULL); if (unlikely(!__pyx_t_347)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_347);

  /* "ultron/tradingday/Calendar.pyx":524
 *     Date(2012, 6, 22),
 *     Date(2012, 10, 1),
 *     Date(2012, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 3),
 *     Date(2012, 10, 4),
 */
  __pyx_t_348 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__351, NULL); if (unlikely(!__pyx_t_348)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_348);

  /* "ultron/tradingday/Calendar.pyx":525
 *     Date(2012, 10, 1),
 *     Date(2012, 10, 2),
 *     Date(2012, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 4),
 *     Date(2012, 10, 5),
 */
  __pyx_t_349 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__352, NULL); if (unlikely(!__pyx_t_349)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_349);

  /* "ultron/tradingday/Calendar.pyx":526
 *     Date(2012, 10, 2),
 *     Date(2012, 10, 3),
 *     Date(2012, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2012, 10, 5),
 *     Date(2013, 1, 1),
 */
  __pyx_t_350 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__353, NULL); if (unlikely(!__pyx_t_350)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_350);

  /* "ultron/tradingday/Calendar.pyx":527
 *     Date(2012, 10, 3),
 *     Date(2012, 10, 4),
 *     Date(2012, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2013, 1, 1),
 *     Date(2013, 1, 2),
 */
  __pyx_t_351 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__354, NULL); if (unlikely(!__pyx_t_351)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_351);

  /* "ultron/tradingday/Calendar.pyx":528
 *     Date(2012, 10, 4),
 *     Date(2012, 10, 5),
 *     Date(2013, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2013, 1, 2),
 *     Date(2013, 1, 3),
 */
  __pyx_t_352 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__355, NULL); if (unlikely(!__pyx_t_352)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_352);

  /* "ultron/tradingday/Calendar.pyx":529
 *     Date(2012, 10, 5),
 *     Date(2013, 1, 1),
 *     Date(2013, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2013, 1, 3),
 *     Date(2013, 2, 11),
 */
  __pyx_t_353 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__356, NULL); if (unlikely(!__pyx_t_353)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_353);

  /* "ultron/tradingday/Calendar.pyx":530
 *     Date(2013, 1, 1),
 *     Date(2013, 1, 2),
 *     Date(2013, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 11),
 *     Date(2013, 2, 12),
 */
  __pyx_t_354 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__357, NULL); if (unlikely(!__pyx_t_354)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_354);

  /* "ultron/tradingday/Calendar.pyx":531
 *     Date(2013, 1, 2),
 *     Date(2013, 1, 3),
 *     Date(2013, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 12),
 *     Date(2013, 2, 13),
 */
  __pyx_t_355 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__358, NULL); if (unlikely(!__pyx_t_355)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_355);

  /* "ultron/tradingday/Calendar.pyx":532
 *     Date(2013, 1, 3),
 *     Date(2013, 2, 11),
 *     Date(2013, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 13),
 *     Date(2013, 2, 14),
 */
  __pyx_t_356 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__359, NULL); if (unlikely(!__pyx_t_356)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_356);

  /* "ultron/tradingday/Calendar.pyx":533
 *     Date(2013, 2, 11),
 *     Date(2013, 2, 12),
 *     Date(2013, 2, 13),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 14),
 *     Date(2013, 2, 15),
 */
  __pyx_t_357 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__360, NULL); if (unlikely(!__pyx_t_357)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_357);

  /* "ultron/tradingday/Calendar.pyx":534
 *     Date(2013, 2, 12),
 *     Date(2013, 2, 13),
 *     Date(2013, 2, 14),             # <<<<<<<<<<<<<<
 *     Date(2013, 2, 15),
 *     Date(2013, 4, 4),
 */
  __pyx_t_358 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__361, NULL); if (unlikely(!__pyx_t_358)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_358);

  /* "ultron/tradingday/Calendar.pyx":535
 *     Date(2013, 2, 13),
 *     Date(2013, 2, 14),
 *     Date(2013, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2013, 4, 4),
 *     Date(2013, 4, 5),
 */
  __pyx_t_359 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__362, NULL); if (unlikely(!__pyx_t_359)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_359);

  /* "ultron/tradingday/Calendar.pyx":536
 *     Date(2013, 2, 14),
 *     Date(2013, 2, 15),
 *     Date(2013, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2013, 4, 5),
 *     Date(2013, 4, 29),
 */
  __pyx_t_360 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__363, NULL); if (unlikely(!__pyx_t_360)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_360);

  /* "ultron/tradingday/Calendar.pyx":537
 *     Date(2013, 2, 15),
 *     Date(2013, 4, 4),
 *     Date(2013, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2013, 4, 29),
 *     Date(2013, 4, 30),
 */
  __pyx_t_361 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__364, NULL); if (unlikely(!__pyx_t_361)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_361);

  /* "ultron/tradingday/Calendar.pyx":538
 *     Date(2013, 4, 4),
 *     Date(2013, 4, 5),
 *     Date(2013, 4, 29),             # <<<<<<<<<<<<<<
 *     Date(2013, 4, 30),
 *     Date(2013, 5, 1),
 */
  __pyx_t_362 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__365, NULL); if (unlikely(!__pyx_t_362)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_362);

  /* "ultron/tradingday/Calendar.pyx":539
 *     Date(2013, 4, 5),
 *     Date(2013, 4, 29),
 *     Date(2013, 4, 30),             # <<<<<<<<<<<<<<
 *     Date(2013, 5, 1),
 *     Date(2013, 6, 10),
 */
  __pyx_t_363 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__366, NULL); if (unlikely(!__pyx_t_363)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_363);

  /* "ultron/tradingday/Calendar.pyx":540
 *     Date(2013, 4, 29),
 *     Date(2013, 4, 30),
 *     Date(2013, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2013, 6, 10),
 *     Date(2013, 6, 11),
 */
  __pyx_t_364 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__367, NULL); if (unlikely(!__pyx_t_364)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_364);

  /* "ultron/tradingday/Calendar.pyx":541
 *     Date(2013, 4, 30),
 *     Date(2013, 5, 1),
 *     Date(2013, 6, 10),             # <<<<<<<<<<<<<<
 *     Date(2013, 6, 11),
 *     Date(2013, 6, 12),
 */
  __pyx_t_365 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__368, NULL); if (unlikely(!__pyx_t_365)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_365);

  /* "ultron/tradingday/Calendar.pyx":542
 *     Date(2013, 5, 1),
 *     Date(2013, 6, 10),
 *     Date(2013, 6, 11),             # <<<<<<<<<<<<<<
 *     Date(2013, 6, 12),
 *     Date(2013, 9, 19),
 */
  __pyx_t_366 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__369, NULL); if (unlikely(!__pyx_t_366)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_366);

  /* "ultron/tradingday/Calendar.pyx":543
 *     Date(2013, 6, 10),
 *     Date(2013, 6, 11),
 *     Date(2013, 6, 12),             # <<<<<<<<<<<<<<
 *     Date(2013, 9, 19),
 *     Date(2013, 9, 20),
 */
  __pyx_t_367 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__370, NULL); if (unlikely(!__pyx_t_367)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_367);

  /* "ultron/tradingday/Calendar.pyx":544
 *     Date(2013, 6, 11),
 *     Date(2013, 6, 12),
 *     Date(2013, 9, 19),             # <<<<<<<<<<<<<<
 *     Date(2013, 9, 20),
 *     Date(2013, 10, 1),
 */
  __pyx_t_368 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__371, NULL); if (unlikely(!__pyx_t_368)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_368);

  /* "ultron/tradingday/Calendar.pyx":545
 *     Date(2013, 6, 12),
 *     Date(2013, 9, 19),
 *     Date(2013, 9, 20),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 1),
 *     Date(2013, 10, 2),
 */
  __pyx_t_369 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__372, NULL); if (unlikely(!__pyx_t_369)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_369);

  /* "ultron/tradingday/Calendar.pyx":546
 *     Date(2013, 9, 19),
 *     Date(2013, 9, 20),
 *     Date(2013, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 2),
 *     Date(2013, 10, 3),
 */
  __pyx_t_370 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__373, NULL); if (unlikely(!__pyx_t_370)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_370);

  /* "ultron/tradingday/Calendar.pyx":547
 *     Date(2013, 9, 20),
 *     Date(2013, 10, 1),
 *     Date(2013, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 3),
 *     Date(2013, 10, 4),
 */
  __pyx_t_371 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__374, NULL); if (unlikely(!__pyx_t_371)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_371);

  /* "ultron/tradingday/Calendar.pyx":548
 *     Date(2013, 10, 1),
 *     Date(2013, 10, 2),
 *     Date(2013, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 4),
 *     Date(2013, 10, 7),
 */
  __pyx_t_372 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__375, NULL); if (unlikely(!__pyx_t_372)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_372);

  /* "ultron/tradingday/Calendar.pyx":549
 *     Date(2013, 10, 2),
 *     Date(2013, 10, 3),
 *     Date(2013, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2013, 10, 7),
 *     Date(2014, 1, 1),
 */
  __pyx_t_373 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__376, NULL); if (unlikely(!__pyx_t_373)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_373);

  /* "ultron/tradingday/Calendar.pyx":550
 *     Date(2013, 10, 3),
 *     Date(2013, 10, 4),
 *     Date(2013, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2014, 1, 1),
 *     Date(2014, 1, 31),
 */
  __pyx_t_374 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__377, NULL); if (unlikely(!__pyx_t_374)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_374);

  /* "ultron/tradingday/Calendar.pyx":551
 *     Date(2013, 10, 4),
 *     Date(2013, 10, 7),
 *     Date(2014, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2014, 1, 31),
 *     Date(2014, 2, 3),
 */
  __pyx_t_375 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__378, NULL); if (unlikely(!__pyx_t_375)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_375);

  /* "ultron/tradingday/Calendar.pyx":552
 *     Date(2013, 10, 7),
 *     Date(2014, 1, 1),
 *     Date(2014, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2014, 2, 3),
 *     Date(2014, 2, 4),
 */
  __pyx_t_376 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__379, NULL); if (unlikely(!__pyx_t_376)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_376);

  /* "ultron/tradingday/Calendar.pyx":553
 *     Date(2014, 1, 1),
 *     Date(2014, 1, 31),
 *     Date(2014, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2014, 2, 4),
 *     Date(2014, 2, 5),
 */
  __pyx_t_377 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__380, NULL); if (unlikely(!__pyx_t_377)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_377);

  /* "ultron/tradingday/Calendar.pyx":554
 *     Date(2014, 1, 31),
 *     Date(2014, 2, 3),
 *     Date(2014, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2014, 2, 5),
 *     Date(2014, 2, 6),
 */
  __pyx_t_378 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__381, NULL); if (unlikely(!__pyx_t_378)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_378);

  /* "ultron/tradingday/Calendar.pyx":555
 *     Date(2014, 2, 3),
 *     Date(2014, 2, 4),
 *     Date(2014, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(2014, 2, 6),
 *     Date(2014, 4, 7),
 */
  __pyx_t_379 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__382, NULL); if (unlikely(!__pyx_t_379)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_379);

  /* "ultron/tradingday/Calendar.pyx":556
 *     Date(2014, 2, 4),
 *     Date(2014, 2, 5),
 *     Date(2014, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(2014, 4, 7),
 *     Date(2014, 5, 1),
 */
  __pyx_t_380 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__383, NULL); if (unlikely(!__pyx_t_380)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_380);

  /* "ultron/tradingday/Calendar.pyx":557
 *     Date(2014, 2, 5),
 *     Date(2014, 2, 6),
 *     Date(2014, 4, 7),             # <<<<<<<<<<<<<<
 *     Date(2014, 5, 1),
 *     Date(2014, 5, 2),
 */
  __pyx_t_381 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__384, NULL); if (unlikely(!__pyx_t_381)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_381);

  /* "ultron/tradingday/Calendar.pyx":558
 *     Date(2014, 2, 6),
 *     Date(2014, 4, 7),
 *     Date(2014, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2014, 5, 2),
 *     Date(2014, 6, 2),
 */
  __pyx_t_382 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__385, NULL); if (unlikely(!__pyx_t_382)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_382);

  /* "ultron/tradingday/Calendar.pyx":559
 *     Date(2014, 4, 7),
 *     Date(2014, 5, 1),
 *     Date(2014, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2014, 6, 2),
 *     Date(2014, 9, 8),
 */
  __pyx_t_383 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__386, NULL); if (unlikely(!__pyx_t_383)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_383);

  /* "ultron/tradingday/Calendar.pyx":560
 *     Date(2014, 5, 1),
 *     Date(2014, 5, 2),
 *     Date(2014, 6, 2),             # <<<<<<<<<<<<<<
 *     Date(2014, 9, 8),
 *     Date(2014, 10, 1),
 */
  __pyx_t_384 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__387, NULL); if (unlikely(!__pyx_t_384)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_384);

  /* "ultron/tradingday/Calendar.pyx":561
 *     Date(2014, 5, 2),
 *     Date(2014, 6, 2),
 *     Date(2014, 9, 8),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 1),
 *     Date(2014, 10, 2),
 */
  __pyx_t_385 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__388, NULL); if (unlikely(!__pyx_t_385)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_385);

  /* "ultron/tradingday/Calendar.pyx":562
 *     Date(2014, 6, 2),
 *     Date(2014, 9, 8),
 *     Date(2014, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 2),
 *     Date(2014, 10, 3),
 */
  __pyx_t_386 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__389, NULL); if (unlikely(!__pyx_t_386)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_386);

  /* "ultron/tradingday/Calendar.pyx":563
 *     Date(2014, 9, 8),
 *     Date(2014, 10, 1),
 *     Date(2014, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 3),
 *     Date(2014, 10, 6),
 */
  __pyx_t_387 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__390, NULL); if (unlikely(!__pyx_t_387)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_387);

  /* "ultron/tradingday/Calendar.pyx":564
 *     Date(2014, 10, 1),
 *     Date(2014, 10, 2),
 *     Date(2014, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 6),
 *     Date(2014, 10, 7),
 */
  __pyx_t_388 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__391, NULL); if (unlikely(!__pyx_t_388)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_388);

  /* "ultron/tradingday/Calendar.pyx":565
 *     Date(2014, 10, 2),
 *     Date(2014, 10, 3),
 *     Date(2014, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2014, 10, 7),
 *     Date(2015, 1, 1),
 */
  __pyx_t_389 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__392, NULL); if (unlikely(!__pyx_t_389)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_389);

  /* "ultron/tradingday/Calendar.pyx":566
 *     Date(2014, 10, 3),
 *     Date(2014, 10, 6),
 *     Date(2014, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2015, 1, 1),
 *     Date(2015, 1, 2),
 */
  __pyx_t_390 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__393, NULL); if (unlikely(!__pyx_t_390)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_390);

  /* "ultron/tradingday/Calendar.pyx":567
 *     Date(2014, 10, 6),
 *     Date(2014, 10, 7),
 *     Date(2015, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2015, 1, 2),
 *     Date(2015, 2, 18),
 */
  __pyx_t_391 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__394, NULL); if (unlikely(!__pyx_t_391)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_391);

  /* "ultron/tradingday/Calendar.pyx":568
 *     Date(2014, 10, 7),
 *     Date(2015, 1, 1),
 *     Date(2015, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 18),
 *     Date(2015, 2, 19),
 */
  __pyx_t_392 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__395, NULL); if (unlikely(!__pyx_t_392)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_392);

  /* "ultron/tradingday/Calendar.pyx":569
 *     Date(2015, 1, 1),
 *     Date(2015, 1, 2),
 *     Date(2015, 2, 18),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 19),
 *     Date(2015, 2, 20),
 */
  __pyx_t_393 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__396, NULL); if (unlikely(!__pyx_t_393)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_393);

  /* "ultron/tradingday/Calendar.pyx":570
 *     Date(2015, 1, 2),
 *     Date(2015, 2, 18),
 *     Date(2015, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 20),
 *     Date(2015, 2, 23),
 */
  __pyx_t_394 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__397, NULL); if (unlikely(!__pyx_t_394)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_394);

  /* "ultron/tradingday/Calendar.pyx":571
 *     Date(2015, 2, 18),
 *     Date(2015, 2, 19),
 *     Date(2015, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 23),
 *     Date(2015, 2, 24),
 */
  __pyx_t_395 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__398, NULL); if (unlikely(!__pyx_t_395)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_395);

  /* "ultron/tradingday/Calendar.pyx":572
 *     Date(2015, 2, 19),
 *     Date(2015, 2, 20),
 *     Date(2015, 2, 23),             # <<<<<<<<<<<<<<
 *     Date(2015, 2, 24),
 *     Date(2015, 4, 6),
 */
  __pyx_t_396 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__399, NULL); if (unlikely(!__pyx_t_396)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_396);

  /* "ultron/tradingday/Calendar.pyx":573
 *     Date(2015, 2, 20),
 *     Date(2015, 2, 23),
 *     Date(2015, 2, 24),             # <<<<<<<<<<<<<<
 *     Date(2015, 4, 6),
 *     Date(2015, 5, 1),
 */
  __pyx_t_397 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__400, NULL); if (unlikely(!__pyx_t_397)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_397);

  /* "ultron/tradingday/Calendar.pyx":574
 *     Date(2015, 2, 23),
 *     Date(2015, 2, 24),
 *     Date(2015, 4, 6),             # <<<<<<<<<<<<<<
 *     Date(2015, 5, 1),
 *     Date(2015, 6, 22),
 */
  __pyx_t_398 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__401, NULL); if (unlikely(!__pyx_t_398)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_398);

  /* "ultron/tradingday/Calendar.pyx":575
 *     Date(2015, 2, 24),
 *     Date(2015, 4, 6),
 *     Date(2015, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2015, 6, 22),
 *     Date(2015, 9, 3),
 */
  __pyx_t_399 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__402, NULL); if (unlikely(!__pyx_t_399)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_399);

  /* "ultron/tradingday/Calendar.pyx":576
 *     Date(2015, 4, 6),
 *     Date(2015, 5, 1),
 *     Date(2015, 6, 22),             # <<<<<<<<<<<<<<
 *     Date(2015, 9, 3),
 *     Date(2015, 9, 4),
 */
  __pyx_t_400 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__403, NULL); if (unlikely(!__pyx_t_400)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_400);

  /* "ultron/tradingday/Calendar.pyx":577
 *     Date(2015, 5, 1),
 *     Date(2015, 6, 22),
 *     Date(2015, 9, 3),             # <<<<<<<<<<<<<<
 *     Date(2015, 9, 4),
 *     Date(2015, 10, 1),
 */
  __pyx_t_401 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__404, NULL); if (unlikely(!__pyx_t_401)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_401);

  /* "ultron/tradingday/Calendar.pyx":578
 *     Date(2015, 6, 22),
 *     Date(2015, 9, 3),
 *     Date(2015, 9, 4),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 1),
 *     Date(2015, 10, 2),
 */
  __pyx_t_402 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__405, NULL); if (unlikely(!__pyx_t_402)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_402);

  /* "ultron/tradingday/Calendar.pyx":579
 *     Date(2015, 9, 3),
 *     Date(2015, 9, 4),
 *     Date(2015, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 2),
 *     Date(2015, 10, 5),
 */
  __pyx_t_403 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__406, NULL); if (unlikely(!__pyx_t_403)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_403);

  /* "ultron/tradingday/Calendar.pyx":580
 *     Date(2015, 9, 4),
 *     Date(2015, 10, 1),
 *     Date(2015, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 5),
 *     Date(2015, 10, 6),
 */
  __pyx_t_404 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__407, NULL); if (unlikely(!__pyx_t_404)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_404);

  /* "ultron/tradingday/Calendar.pyx":581
 *     Date(2015, 10, 1),
 *     Date(2015, 10, 2),
 *     Date(2015, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 6),
 *     Date(2015, 10, 7),
 */
  __pyx_t_405 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__408, NULL); if (unlikely(!__pyx_t_405)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_405);

  /* "ultron/tradingday/Calendar.pyx":582
 *     Date(2015, 10, 2),
 *     Date(2015, 10, 5),
 *     Date(2015, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2015, 10, 7),
 *     Date(2016, 1, 1),
 */
  __pyx_t_406 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__409, NULL); if (unlikely(!__pyx_t_406)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_406);

  /* "ultron/tradingday/Calendar.pyx":583
 *     Date(2015, 10, 5),
 *     Date(2015, 10, 6),
 *     Date(2015, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2016, 1, 1),
 *     Date(2016, 2, 8),
 */
  __pyx_t_407 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__410, NULL); if (unlikely(!__pyx_t_407)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_407);

  /* "ultron/tradingday/Calendar.pyx":584
 *     Date(2015, 10, 6),
 *     Date(2015, 10, 7),
 *     Date(2016, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 8),
 *     Date(2016, 2, 9),
 */
  __pyx_t_408 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__411, NULL); if (unlikely(!__pyx_t_408)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_408);

  /* "ultron/tradingday/Calendar.pyx":585
 *     Date(2015, 10, 7),
 *     Date(2016, 1, 1),
 *     Date(2016, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 9),
 *     Date(2016, 2, 10),
 */
  __pyx_t_409 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__412, NULL); if (unlikely(!__pyx_t_409)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_409);

  /* "ultron/tradingday/Calendar.pyx":586
 *     Date(2016, 1, 1),
 *     Date(2016, 2, 8),
 *     Date(2016, 2, 9),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 10),
 *     Date(2016, 2, 11),
 */
  __pyx_t_410 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__413, NULL); if (unlikely(!__pyx_t_410)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_410);

  /* "ultron/tradingday/Calendar.pyx":587
 *     Date(2016, 2, 8),
 *     Date(2016, 2, 9),
 *     Date(2016, 2, 10),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 11),
 *     Date(2016, 2, 12),
 */
  __pyx_t_411 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__414, NULL); if (unlikely(!__pyx_t_411)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_411);

  /* "ultron/tradingday/Calendar.pyx":588
 *     Date(2016, 2, 9),
 *     Date(2016, 2, 10),
 *     Date(2016, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2016, 2, 12),
 *     Date(2016, 4, 4),
 */
  __pyx_t_412 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__415, NULL); if (unlikely(!__pyx_t_412)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_412);

  /* "ultron/tradingday/Calendar.pyx":589
 *     Date(2016, 2, 10),
 *     Date(2016, 2, 11),
 *     Date(2016, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2016, 4, 4),
 *     Date(2016, 5, 1),
 */
  __pyx_t_413 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__416, NULL); if (unlikely(!__pyx_t_413)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_413);

  /* "ultron/tradingday/Calendar.pyx":590
 *     Date(2016, 2, 11),
 *     Date(2016, 2, 12),
 *     Date(2016, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2016, 5, 1),
 *     Date(2016, 5, 2),
 */
  __pyx_t_414 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__417, NULL); if (unlikely(!__pyx_t_414)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_414);

  /* "ultron/tradingday/Calendar.pyx":591
 *     Date(2016, 2, 12),
 *     Date(2016, 4, 4),
 *     Date(2016, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2016, 5, 2),
 *     Date(2016, 6, 9),
 */
  __pyx_t_415 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__418, NULL); if (unlikely(!__pyx_t_415)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_415);

  /* "ultron/tradingday/Calendar.pyx":592
 *     Date(2016, 4, 4),
 *     Date(2016, 5, 1),
 *     Date(2016, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2016, 6, 9),
 *     Date(2016, 6, 10),
 */
  __pyx_t_416 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__419, NULL); if (unlikely(!__pyx_t_416)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_416);

  /* "ultron/tradingday/Calendar.pyx":593
 *     Date(2016, 5, 1),
 *     Date(2016, 5, 2),
 *     Date(2016, 6, 9),             # <<<<<<<<<<<<<<
 *     Date(2016, 6, 10),
 *     Date(2016, 9, 15),
 */
  __pyx_t_417 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__420, NULL); if (unlikely(!__pyx_t_417)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_417);

  /* "ultron/tradingday/Calendar.pyx":594
 *     Date(2016, 5, 2),
 *     Date(2016, 6, 9),
 *     Date(2016, 6, 10),             # <<<<<<<<<<<<<<
 *     Date(2016, 9, 15),
 *     Date(2016, 9, 16),
 */
  __pyx_t_418 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__421, NULL); if (unlikely(!__pyx_t_418)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_418);

  /* "ultron/tradingday/Calendar.pyx":595
 *     Date(2016, 6, 9),
 *     Date(2016, 6, 10),
 *     Date(2016, 9, 15),             # <<<<<<<<<<<<<<
 *     Date(2016, 9, 16),
 *     Date(2016, 10, 3),
 */
  __pyx_t_419 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__422, NULL); if (unlikely(!__pyx_t_419)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_419);

  /* "ultron/tradingday/Calendar.pyx":596
 *     Date(2016, 6, 10),
 *     Date(2016, 9, 15),
 *     Date(2016, 9, 16),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 3),
 *     Date(2016, 10, 4),
 */
  __pyx_t_420 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__423, NULL); if (unlikely(!__pyx_t_420)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_420);

  /* "ultron/tradingday/Calendar.pyx":597
 *     Date(2016, 9, 15),
 *     Date(2016, 9, 16),
 *     Date(2016, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 4),
 *     Date(2016, 10, 5),
 */
  __pyx_t_421 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__424, NULL); if (unlikely(!__pyx_t_421)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_421);

  /* "ultron/tradingday/Calendar.pyx":598
 *     Date(2016, 9, 16),
 *     Date(2016, 10, 3),
 *     Date(2016, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 5),
 *     Date(2016, 10, 6),
 */
  __pyx_t_422 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__425, NULL); if (unlikely(!__pyx_t_422)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_422);

  /* "ultron/tradingday/Calendar.pyx":599
 *     Date(2016, 10, 3),
 *     Date(2016, 10, 4),
 *     Date(2016, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 6),
 *     Date(2016, 10, 7),
 */
  __pyx_t_423 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__426, NULL); if (unlikely(!__pyx_t_423)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_423);

  /* "ultron/tradingday/Calendar.pyx":600
 *     Date(2016, 10, 4),
 *     Date(2016, 10, 5),
 *     Date(2016, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2016, 10, 7),
 *     Date(2017, 1, 2),
 */
  __pyx_t_424 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__427, NULL); if (unlikely(!__pyx_t_424)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_424);

  /* "ultron/tradingday/Calendar.pyx":601
 *     Date(2016, 10, 5),
 *     Date(2016, 10, 6),
 *     Date(2016, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2017, 1, 2),
 *     Date(2017, 1, 27),
 */
  __pyx_t_425 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__428, NULL); if (unlikely(!__pyx_t_425)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);

  /* "ultron/tradingday/Calendar.pyx":602
 *     Date(2016, 10, 6),
 *     Date(2016, 10, 7),
 *     Date(2017, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2017, 1, 27),
 *     Date(2017, 1, 30),
 */
  __pyx_t_426 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__429, NULL); if (unlikely(!__pyx_t_426)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_426);

  /* "ultron/tradingday/Calendar.pyx":603
 *     Date(2016, 10, 7),
 *     Date(2017, 1, 2),
 *     Date(2017, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2017, 1, 30),
 *     Date(2017, 1, 31),
 */
  __pyx_t_427 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__430, NULL); if (unlikely(!__pyx_t_427)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_427);

  /* "ultron/tradingday/Calendar.pyx":604
 *     Date(2017, 1, 2),
 *     Date(2017, 1, 27),
 *     Date(2017, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2017, 1, 31),
 *     Date(2017, 2, 1),
 */
  __pyx_t_428 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__431, NULL); if (unlikely(!__pyx_t_428)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_428);

  /* "ultron/tradingday/Calendar.pyx":605
 *     Date(2017, 1, 27),
 *     Date(2017, 1, 30),
 *     Date(2017, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2017, 2, 1),
 *     Date(2017, 2, 2),
 */
  __pyx_t_429 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__432, NULL); if (unlikely(!__pyx_t_429)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_429);

  /* "ultron/tradingday/Calendar.pyx":606
 *     Date(2017, 1, 30),
 *     Date(2017, 1, 31),
 *     Date(2017, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2017, 2, 2),
 *     Date(2017, 4, 3),
 */
  __pyx_t_430 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__433, NULL); if (unlikely(!__pyx_t_430)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_430);

  /* "ultron/tradingday/Calendar.pyx":607
 *     Date(2017, 1, 31),
 *     Date(2017, 2, 1),
 *     Date(2017, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2017, 4, 3),
 *     Date(2017, 4, 4),
 */
  __pyx_t_431 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__434, NULL); if (unlikely(!__pyx_t_431)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_431);

  /* "ultron/tradingday/Calendar.pyx":608
 *     Date(2017, 2, 1),
 *     Date(2017, 2, 2),
 *     Date(2017, 4, 3),             # <<<<<<<<<<<<<<
 *     Date(2017, 4, 4),
 *     Date(2017, 5, 1),
 */
  __pyx_t_432 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__435, NULL); if (unlikely(!__pyx_t_432)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_432);

  /* "ultron/tradingday/Calendar.pyx":609
 *     Date(2017, 2, 2),
 *     Date(2017, 4, 3),
 *     Date(2017, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2017, 5, 1),
 *     Date(2017, 5, 29),
 */
  __pyx_t_433 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__436, NULL); if (unlikely(!__pyx_t_433)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_433);

  /* "ultron/tradingday/Calendar.pyx":610
 *     Date(2017, 4, 3),
 *     Date(2017, 4, 4),
 *     Date(2017, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2017, 5, 29),
 *     Date(2017, 5, 30),
 */
  __pyx_t_434 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__437, NULL); if (unlikely(!__pyx_t_434)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_434);

  /* "ultron/tradingday/Calendar.pyx":611
 *     Date(2017, 4, 4),
 *     Date(2017, 5, 1),
 *     Date(2017, 5, 29),             # <<<<<<<<<<<<<<
 *     Date(2017, 5, 30),
 *     Date(2017, 10, 2),
 */
  __pyx_t_435 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__438, NULL); if (unlikely(!__pyx_t_435)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_435);

  /* "ultron/tradingday/Calendar.pyx":612
 *     Date(2017, 5, 1),
 *     Date(2017, 5, 29),
 *     Date(2017, 5, 30),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 2),
 *     Date(2017, 10, 3),
 */
  __pyx_t_436 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__439, NULL); if (unlikely(!__pyx_t_436)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_436);

  /* "ultron/tradingday/Calendar.pyx":613
 *     Date(2017, 5, 29),
 *     Date(2017, 5, 30),
 *     Date(2017, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 3),
 *     Date(2017, 10, 4),
 */
  __pyx_t_437 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__440, NULL); if (unlikely(!__pyx_t_437)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_437);

  /* "ultron/tradingday/Calendar.pyx":614
 *     Date(2017, 5, 30),
 *     Date(2017, 10, 2),
 *     Date(2017, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 4),
 *     Date(2017, 10, 5),
 */
  __pyx_t_438 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__441, NULL); if (unlikely(!__pyx_t_438)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_438);

  /* "ultron/tradingday/Calendar.pyx":615
 *     Date(2017, 10, 2),
 *     Date(2017, 10, 3),
 *     Date(2017, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 5),
 *     Date(2017, 10, 6),
 */
  __pyx_t_439 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__442, NULL); if (unlikely(!__pyx_t_439)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_439);

  /* "ultron/tradingday/Calendar.pyx":616
 *     Date(2017, 10, 3),
 *     Date(2017, 10, 4),
 *     Date(2017, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2017, 10, 6),
 *     Date(2018, 1, 1),
 */
  __pyx_t_440 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__443, NULL); if (unlikely(!__pyx_t_440)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_440);

  /* "ultron/tradingday/Calendar.pyx":617
 *     Date(2017, 10, 4),
 *     Date(2017, 10, 5),
 *     Date(2017, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2018, 1, 1),
 *     Date(2018, 2, 15),
 */
  __pyx_t_441 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__444, NULL); if (unlikely(!__pyx_t_441)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_441);

  /* "ultron/tradingday/Calendar.pyx":618
 *     Date(2017, 10, 5),
 *     Date(2017, 10, 6),
 *     Date(2018, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 15),
 *     Date(2018, 2, 16),
 */
  __pyx_t_442 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__445, NULL); if (unlikely(!__pyx_t_442)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_442);

  /* "ultron/tradingday/Calendar.pyx":619
 *     Date(2017, 10, 6),
 *     Date(2018, 1, 1),
 *     Date(2018, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 16),
 *     Date(2018, 2, 19),
 */
  __pyx_t_443 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__446, NULL); if (unlikely(!__pyx_t_443)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_443);

  /* "ultron/tradingday/Calendar.pyx":620
 *     Date(2018, 1, 1),
 *     Date(2018, 2, 15),
 *     Date(2018, 2, 16),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 19),
 *     Date(2018, 2, 20),
 */
  __pyx_t_444 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__447, NULL); if (unlikely(!__pyx_t_444)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_444);

  /* "ultron/tradingday/Calendar.pyx":621
 *     Date(2018, 2, 15),
 *     Date(2018, 2, 16),
 *     Date(2018, 2, 19),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 20),
 *     Date(2018, 2, 21),
 */
  __pyx_t_445 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__448, NULL); if (unlikely(!__pyx_t_445)) __PYX_ERR(0, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_445);

  /* "ultron/tradingday/Calendar.pyx":622
 *     Date(2018, 2, 16),
 *     Date(2018, 2, 19),
 *     Date(2018, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2018, 2, 21),
 *     Date(2018, 4, 5),
 */
  __pyx_t_446 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__449, NULL); if (unlikely(!__pyx_t_446)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_446);

  /* "ultron/tradingday/Calendar.pyx":623
 *     Date(2018, 2, 19),
 *     Date(2018, 2, 20),
 *     Date(2018, 2, 21),             # <<<<<<<<<<<<<<
 *     Date(2018, 4, 5),
 *     Date(2018, 4, 6),
 */
  __pyx_t_447 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__450, NULL); if (unlikely(!__pyx_t_447)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_447);

  /* "ultron/tradingday/Calendar.pyx":624
 *     Date(2018, 2, 20),
 *     Date(2018, 2, 21),
 *     Date(2018, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2018, 4, 6),
 *     Date(2018, 4, 30),
 */
  __pyx_t_448 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__451, NULL); if (unlikely(!__pyx_t_448)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_448);

  /* "ultron/tradingday/Calendar.pyx":625
 *     Date(2018, 2, 21),
 *     Date(2018, 4, 5),
 *     Date(2018, 4, 6),             # <<<<<<<<<<<<<<
 *     Date(2018, 4, 30),
 *     Date(2018, 5, 1),
 */
  __pyx_t_449 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__452, NULL); if (unlikely(!__pyx_t_449)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_449);

  /* "ultron/tradingday/Calendar.pyx":626
 *     Date(2018, 4, 5),
 *     Date(2018, 4, 6),
 *     Date(2018, 4, 30),             # <<<<<<<<<<<<<<
 *     Date(2018, 5, 1),
 *     Date(2018, 6, 18),
 */
  __pyx_t_450 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__453, NULL); if (unlikely(!__pyx_t_450)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_450);

  /* "ultron/tradingday/Calendar.pyx":627
 *     Date(2018, 4, 6),
 *     Date(2018, 4, 30),
 *     Date(2018, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2018, 6, 18),
 *     Date(2018, 9, 24),
 */
  __pyx_t_451 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__454, NULL); if (unlikely(!__pyx_t_451)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_451);

  /* "ultron/tradingday/Calendar.pyx":628
 *     Date(2018, 4, 30),
 *     Date(2018, 5, 1),
 *     Date(2018, 6, 18),             # <<<<<<<<<<<<<<
 *     Date(2018, 9, 24),
 *     Date(2018, 10, 1),
 */
  __pyx_t_452 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__455, NULL); if (unlikely(!__pyx_t_452)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_452);

  /* "ultron/tradingday/Calendar.pyx":629
 *     Date(2018, 5, 1),
 *     Date(2018, 6, 18),
 *     Date(2018, 9, 24),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 1),
 *     Date(2018, 10, 2),
 */
  __pyx_t_453 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__456, NULL); if (unlikely(!__pyx_t_453)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_453);

  /* "ultron/tradingday/Calendar.pyx":630
 *     Date(2018, 6, 18),
 *     Date(2018, 9, 24),
 *     Date(2018, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 2),
 *     Date(2018, 10, 3),
 */
  __pyx_t_454 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__457, NULL); if (unlikely(!__pyx_t_454)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_454);

  /* "ultron/tradingday/Calendar.pyx":631
 *     Date(2018, 9, 24),
 *     Date(2018, 10, 1),
 *     Date(2018, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 3),
 *     Date(2018, 10, 4),
 */
  __pyx_t_455 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__458, NULL); if (unlikely(!__pyx_t_455)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_455);

  /* "ultron/tradingday/Calendar.pyx":632
 *     Date(2018, 10, 1),
 *     Date(2018, 10, 2),
 *     Date(2018, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 4),
 *     Date(2018, 10, 5),
 */
  __pyx_t_456 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__459, NULL); if (unlikely(!__pyx_t_456)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_456);

  /* "ultron/tradingday/Calendar.pyx":633
 *     Date(2018, 10, 2),
 *     Date(2018, 10, 3),
 *     Date(2018, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2018, 10, 5),
 *     Date(2018, 12, 31),
 */
  __pyx_t_457 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__460, NULL); if (unlikely(!__pyx_t_457)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_457);

  /* "ultron/tradingday/Calendar.pyx":634
 *     Date(2018, 10, 3),
 *     Date(2018, 10, 4),
 *     Date(2018, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2018, 12, 31),
 *     Date(2019, 1, 1),
 */
  __pyx_t_458 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__461, NULL); if (unlikely(!__pyx_t_458)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_458);

  /* "ultron/tradingday/Calendar.pyx":635
 *     Date(2018, 10, 4),
 *     Date(2018, 10, 5),
 *     Date(2018, 12, 31),             # <<<<<<<<<<<<<<
 *     Date(2019, 1, 1),
 *     Date(2019, 2, 4),
 */
  __pyx_t_459 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__462, NULL); if (unlikely(!__pyx_t_459)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_459);

  /* "ultron/tradingday/Calendar.pyx":636
 *     Date(2018, 10, 5),
 *     Date(2018, 12, 31),
 *     Date(2019, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 4),
 *     Date(2019, 2, 5),
 */
  __pyx_t_460 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__463, NULL); if (unlikely(!__pyx_t_460)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_460);

  /* "ultron/tradingday/Calendar.pyx":637
 *     Date(2018, 12, 31),
 *     Date(2019, 1, 1),
 *     Date(2019, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 5),
 *     Date(2019, 2, 6),
 */
  __pyx_t_461 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__464, NULL); if (unlikely(!__pyx_t_461)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_461);

  /* "ultron/tradingday/Calendar.pyx":638
 *     Date(2019, 1, 1),
 *     Date(2019, 2, 4),
 *     Date(2019, 2, 5),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 6),
 *     Date(2019, 2, 7),
 */
  __pyx_t_462 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__465, NULL); if (unlikely(!__pyx_t_462)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_462);

  /* "ultron/tradingday/Calendar.pyx":639
 *     Date(2019, 2, 4),
 *     Date(2019, 2, 5),
 *     Date(2019, 2, 6),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 7),
 *     Date(2019, 2, 8),
 */
  __pyx_t_463 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__466, NULL); if (unlikely(!__pyx_t_463)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_463);

  /* "ultron/tradingday/Calendar.pyx":640
 *     Date(2019, 2, 5),
 *     Date(2019, 2, 6),
 *     Date(2019, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2019, 2, 8),
 *     Date(2019, 4, 5),
 */
  __pyx_t_464 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__467, NULL); if (unlikely(!__pyx_t_464)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_464);

  /* "ultron/tradingday/Calendar.pyx":641
 *     Date(2019, 2, 6),
 *     Date(2019, 2, 7),
 *     Date(2019, 2, 8),             # <<<<<<<<<<<<<<
 *     Date(2019, 4, 5),
 *     Date(2019, 5, 1),
 */
  __pyx_t_465 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__468, NULL); if (unlikely(!__pyx_t_465)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_465);

  /* "ultron/tradingday/Calendar.pyx":642
 *     Date(2019, 2, 7),
 *     Date(2019, 2, 8),
 *     Date(2019, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2019, 5, 1),
 *     Date(2019, 5, 2),
 */
  __pyx_t_466 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__469, NULL); if (unlikely(!__pyx_t_466)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_466);

  /* "ultron/tradingday/Calendar.pyx":643
 *     Date(2019, 2, 8),
 *     Date(2019, 4, 5),
 *     Date(2019, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2019, 5, 2),
 *     Date(2019, 5, 3),
 */
  __pyx_t_467 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__470, NULL); if (unlikely(!__pyx_t_467)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_467);

  /* "ultron/tradingday/Calendar.pyx":644
 *     Date(2019, 4, 5),
 *     Date(2019, 5, 1),
 *     Date(2019, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2019, 5, 3),
 *     Date(2019, 6, 7),
 */
  __pyx_t_468 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__471, NULL); if (unlikely(!__pyx_t_468)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_468);

  /* "ultron/tradingday/Calendar.pyx":645
 *     Date(2019, 5, 1),
 *     Date(2019, 5, 2),
 *     Date(2019, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2019, 6, 7),
 *     Date(2019, 9, 13),
 */
  __pyx_t_469 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__472, NULL); if (unlikely(!__pyx_t_469)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_469);

  /* "ultron/tradingday/Calendar.pyx":646
 *     Date(2019, 5, 2),
 *     Date(2019, 5, 3),
 *     Date(2019, 6, 7),             # <<<<<<<<<<<<<<
 *     Date(2019, 9, 13),
 *     Date(2019, 10, 1),
 */
  __pyx_t_470 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__473, NULL); if (unlikely(!__pyx_t_470)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_470);

  /* "ultron/tradingday/Calendar.pyx":647
 *     Date(2019, 5, 3),
 *     Date(2019, 6, 7),
 *     Date(2019, 9, 13),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 1),
 *     Date(2019, 10, 2),
 */
  __pyx_t_471 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__474, NULL); if (unlikely(!__pyx_t_471)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_471);

  /* "ultron/tradingday/Calendar.pyx":648
 *     Date(2019, 6, 7),
 *     Date(2019, 9, 13),
 *     Date(2019, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 2),
 *     Date(2019, 10, 3),
 */
  __pyx_t_472 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__475, NULL); if (unlikely(!__pyx_t_472)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_472);

  /* "ultron/tradingday/Calendar.pyx":649
 *     Date(2019, 9, 13),
 *     Date(2019, 10, 1),
 *     Date(2019, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 3),
 *     Date(2019, 10, 4),
 */
  __pyx_t_473 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__476, NULL); if (unlikely(!__pyx_t_473)) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_473);

  /* "ultron/tradingday/Calendar.pyx":650
 *     Date(2019, 10, 1),
 *     Date(2019, 10, 2),
 *     Date(2019, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 4),
 *     Date(2019, 10, 7),
 */
  __pyx_t_474 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__477, NULL); if (unlikely(!__pyx_t_474)) __PYX_ERR(0, 650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_474);

  /* "ultron/tradingday/Calendar.pyx":651
 *     Date(2019, 10, 2),
 *     Date(2019, 10, 3),
 *     Date(2019, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2019, 10, 7),
 *     Date(2020, 1, 1),
 */
  __pyx_t_475 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__478, NULL); if (unlikely(!__pyx_t_475)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_475);

  /* "ultron/tradingday/Calendar.pyx":652
 *     Date(2019, 10, 3),
 *     Date(2019, 10, 4),
 *     Date(2019, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 1),
 *     Date(2020, 1, 24),
 */
  __pyx_t_476 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__479, NULL); if (unlikely(!__pyx_t_476)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_476);

  /* "ultron/tradingday/Calendar.pyx":653
 *     Date(2019, 10, 4),
 *     Date(2019, 10, 7),
 *     Date(2020, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 24),
 *     Date(2020, 1, 27),
 */
  __pyx_t_477 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__480, NULL); if (unlikely(!__pyx_t_477)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_477);

  /* "ultron/tradingday/Calendar.pyx":654
 *     Date(2019, 10, 7),
 *     Date(2020, 1, 1),
 *     Date(2020, 1, 24),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 27),
 *     Date(2020, 1, 28),
 */
  __pyx_t_478 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__481, NULL); if (unlikely(!__pyx_t_478)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_478);

  /* "ultron/tradingday/Calendar.pyx":655
 *     Date(2020, 1, 1),
 *     Date(2020, 1, 24),
 *     Date(2020, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 28),
 *     Date(2020, 1, 29),
 */
  __pyx_t_479 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__482, NULL); if (unlikely(!__pyx_t_479)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_479);

  /* "ultron/tradingday/Calendar.pyx":656
 *     Date(2020, 1, 24),
 *     Date(2020, 1, 27),
 *     Date(2020, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 29),
 *     Date(2020, 1, 30),
 */
  __pyx_t_480 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__483, NULL); if (unlikely(!__pyx_t_480)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_480);

  /* "ultron/tradingday/Calendar.pyx":657
 *     Date(2020, 1, 27),
 *     Date(2020, 1, 28),
 *     Date(2020, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 30),
 *     Date(2020, 1, 31),
 */
  __pyx_t_481 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__484, NULL); if (unlikely(!__pyx_t_481)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_481);

  /* "ultron/tradingday/Calendar.pyx":658
 *     Date(2020, 1, 28),
 *     Date(2020, 1, 29),
 *     Date(2020, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2020, 1, 31),
 *     Date(2020, 4, 6),
 */
  __pyx_t_482 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__485, NULL); if (unlikely(!__pyx_t_482)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_482);

  /* "ultron/tradingday/Calendar.pyx":659
 *     Date(2020, 1, 29),
 *     Date(2020, 1, 30),
 *     Date(2020, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2020, 4, 6),
 *     Date(2020, 5, 1),
 */
  __pyx_t_483 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__486, NULL); if (unlikely(!__pyx_t_483)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_483);

  /* "ultron/tradingday/Calendar.pyx":660
 *     Date(2020, 1, 30),
 *     Date(2020, 1, 31),
 *     Date(2020, 4, 6),             # <<<<<<<<<<<<<<
 *     Date(2020, 5, 1),
 *     Date(2020, 5, 4),
 */
  __pyx_t_484 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__487, NULL); if (unlikely(!__pyx_t_484)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_484);

  /* "ultron/tradingday/Calendar.pyx":661
 *     Date(2020, 1, 31),
 *     Date(2020, 4, 6),
 *     Date(2020, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2020, 5, 4),
 *     Date(2020, 5, 5),
 */
  __pyx_t_485 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__488, NULL); if (unlikely(!__pyx_t_485)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_485);

  /* "ultron/tradingday/Calendar.pyx":662
 *     Date(2020, 4, 6),
 *     Date(2020, 5, 1),
 *     Date(2020, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2020, 5, 5),
 *     Date(2020, 6, 25),
 */
  __pyx_t_486 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__489, NULL); if (unlikely(!__pyx_t_486)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_486);

  /* "ultron/tradingday/Calendar.pyx":663
 *     Date(2020, 5, 1),
 *     Date(2020, 5, 4),
 *     Date(2020, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2020, 6, 25),
 *     Date(2020, 6, 26),
 */
  __pyx_t_487 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__490, NULL); if (unlikely(!__pyx_t_487)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_487);

  /* "ultron/tradingday/Calendar.pyx":664
 *     Date(2020, 5, 4),
 *     Date(2020, 5, 5),
 *     Date(2020, 6, 25),             # <<<<<<<<<<<<<<
 *     Date(2020, 6, 26),
 *     Date(2020, 10, 1),
 */
  __pyx_t_488 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__491, NULL); if (unlikely(!__pyx_t_488)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_488);

  /* "ultron/tradingday/Calendar.pyx":665
 *     Date(2020, 5, 5),
 *     Date(2020, 6, 25),
 *     Date(2020, 6, 26),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 1),
 *     Date(2020, 10, 2),
 */
  __pyx_t_489 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__492, NULL); if (unlikely(!__pyx_t_489)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_489);

  /* "ultron/tradingday/Calendar.pyx":666
 *     Date(2020, 6, 25),
 *     Date(2020, 6, 26),
 *     Date(2020, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 2),
 *     Date(2020, 10, 5),
 */
  __pyx_t_490 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__493, NULL); if (unlikely(!__pyx_t_490)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_490);

  /* "ultron/tradingday/Calendar.pyx":667
 *     Date(2020, 6, 26),
 *     Date(2020, 10, 1),
 *     Date(2020, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 5),
 *     Date(2020, 10, 6),
 */
  __pyx_t_491 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__494, NULL); if (unlikely(!__pyx_t_491)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_491);

  /* "ultron/tradingday/Calendar.pyx":668
 *     Date(2020, 10, 1),
 *     Date(2020, 10, 2),
 *     Date(2020, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 6),
 *     Date(2020, 10, 7),
 */
  __pyx_t_492 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__495, NULL); if (unlikely(!__pyx_t_492)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_492);

  /* "ultron/tradingday/Calendar.pyx":669
 *     Date(2020, 10, 2),
 *     Date(2020, 10, 5),
 *     Date(2020, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 7),
 *     Date(2020, 10, 8),
 */
  __pyx_t_493 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__496, NULL); if (unlikely(!__pyx_t_493)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_493);

  /* "ultron/tradingday/Calendar.pyx":670
 *     Date(2020, 10, 5),
 *     Date(2020, 10, 6),
 *     Date(2020, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2020, 10, 8),
 *     Date(2021, 1, 1),
 */
  __pyx_t_494 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__497, NULL); if (unlikely(!__pyx_t_494)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_494);

  /* "ultron/tradingday/Calendar.pyx":671
 *     Date(2020, 10, 6),
 *     Date(2020, 10, 7),
 *     Date(2020, 10, 8),             # <<<<<<<<<<<<<<
 *     Date(2021, 1, 1),
 *     Date(2021, 2, 11),
 */
  __pyx_t_495 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__498, NULL); if (unlikely(!__pyx_t_495)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_495);

  /* "ultron/tradingday/Calendar.pyx":672
 *     Date(2020, 10, 7),
 *     Date(2020, 10, 8),
 *     Date(2021, 1, 1),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 11),
 *     Date(2021, 2, 12),
 */
  __pyx_t_496 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__499, NULL); if (unlikely(!__pyx_t_496)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_496);

  /* "ultron/tradingday/Calendar.pyx":673
 *     Date(2020, 10, 8),
 *     Date(2021, 1, 1),
 *     Date(2021, 2, 11),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 12),
 *     Date(2021, 2, 15),
 */
  __pyx_t_497 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__500, NULL); if (unlikely(!__pyx_t_497)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_497);

  /* "ultron/tradingday/Calendar.pyx":674
 *     Date(2021, 1, 1),
 *     Date(2021, 2, 11),
 *     Date(2021, 2, 12),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 15),
 *     Date(2021, 2, 16),
 */
  __pyx_t_498 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__501, NULL); if (unlikely(!__pyx_t_498)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_498);

  /* "ultron/tradingday/Calendar.pyx":675
 *     Date(2021, 2, 11),
 *     Date(2021, 2, 12),
 *     Date(2021, 2, 15),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 16),
 *     Date(2021, 2, 17),
 */
  __pyx_t_499 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__502, NULL); if (unlikely(!__pyx_t_499)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_499);

  /* "ultron/tradingday/Calendar.pyx":676
 *     Date(2021, 2, 12),
 *     Date(2021, 2, 15),
 *     Date(2021, 2, 16),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 17),
 *     Date(2021, 4, 5),
 */
  __pyx_t_500 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__503, NULL); if (unlikely(!__pyx_t_500)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_500);

  /* "ultron/tradingday/Calendar.pyx":677
 *     Date(2021, 2, 15),
 *     Date(2021, 2, 16),
 *     Date(2021, 2, 17),             # <<<<<<<<<<<<<<
 *     Date(2021, 4, 5),
 *     Date(2021, 5, 3),
 */
  __pyx_t_501 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__504, NULL); if (unlikely(!__pyx_t_501)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_501);

  /* "ultron/tradingday/Calendar.pyx":678
 *     Date(2021, 2, 16),
 *     Date(2021, 2, 17),
 *     Date(2021, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2021, 5, 3),
 *     Date(2021, 5, 4),
 */
  __pyx_t_502 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__505, NULL); if (unlikely(!__pyx_t_502)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_502);

  /* "ultron/tradingday/Calendar.pyx":679
 *     Date(2021, 2, 17),
 *     Date(2021, 4, 5),
 *     Date(2021, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2021, 5, 4),
 *     Date(2021, 5, 5),
 */
  __pyx_t_503 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__506, NULL); if (unlikely(!__pyx_t_503)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_503);

  /* "ultron/tradingday/Calendar.pyx":680
 *     Date(2021, 4, 5),
 *     Date(2021, 5, 3),
 *     Date(2021, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2021, 5, 5),
 *     Date(2021, 6, 14),
 */
  __pyx_t_504 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__507, NULL); if (unlikely(!__pyx_t_504)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_504);

  /* "ultron/tradingday/Calendar.pyx":681
 *     Date(2021, 5, 3),
 *     Date(2021, 5, 4),
 *     Date(2021, 5, 5),             # <<<<<<<<<<<<<<
 *     Date(2021, 6, 14),
 *     Date(2021, 9, 20),
 */
  __pyx_t_505 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__508, NULL); if (unlikely(!__pyx_t_505)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_505);

  /* "ultron/tradingday/Calendar.pyx":682
 *     Date(2021, 5, 4),
 *     Date(2021, 5, 5),
 *     Date(2021, 6, 14),             # <<<<<<<<<<<<<<
 *     Date(2021, 9, 20),
 *     Date(2021, 9, 21),
 */
  __pyx_t_506 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__509, NULL); if (unlikely(!__pyx_t_506)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_506);

  /* "ultron/tradingday/Calendar.pyx":683
 *     Date(2021, 5, 5),
 *     Date(2021, 6, 14),
 *     Date(2021, 9, 20),             # <<<<<<<<<<<<<<
 *     Date(2021, 9, 21),
 *     Date(2021, 10, 1),
 */
  __pyx_t_507 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__510, NULL); if (unlikely(!__pyx_t_507)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_507);

  /* "ultron/tradingday/Calendar.pyx":684
 *     Date(2021, 6, 14),
 *     Date(2021, 9, 20),
 *     Date(2021, 9, 21),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 1),
 *     Date(2021, 10, 4),
 */
  __pyx_t_508 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__511, NULL); if (unlikely(!__pyx_t_508)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_508);

  /* "ultron/tradingday/Calendar.pyx":685
 *     Date(2021, 9, 20),
 *     Date(2021, 9, 21),
 *     Date(2021, 10, 1),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 4),
 *     Date(2021, 10, 5),
 */
  __pyx_t_509 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__512, NULL); if (unlikely(!__pyx_t_509)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_509);

  /* "ultron/tradingday/Calendar.pyx":686
 *     Date(2021, 9, 21),
 *     Date(2021, 10, 1),
 *     Date(2021, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 5),
 *     Date(2021, 10, 6),
 */
  __pyx_t_510 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__513, NULL); if (unlikely(!__pyx_t_510)) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_510);

  /* "ultron/tradingday/Calendar.pyx":687
 *     Date(2021, 10, 1),
 *     Date(2021, 10, 4),
 *     Date(2021, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 6),
 *     Date(2021, 10, 7),
 */
  __pyx_t_511 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__514, NULL); if (unlikely(!__pyx_t_511)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_511);

  /* "ultron/tradingday/Calendar.pyx":688
 *     Date(2021, 10, 4),
 *     Date(2021, 10, 5),
 *     Date(2021, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 7),
 *     Date(2022, 1, 3),
 */
  __pyx_t_512 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__515, NULL); if (unlikely(!__pyx_t_512)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_512);

  /* "ultron/tradingday/Calendar.pyx":689
 *     Date(2021, 10, 5),
 *     Date(2021, 10, 6),
 *     Date(2021, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2022, 1, 3),
 *     Date(2022, 1, 31),
 */
  __pyx_t_513 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__516, NULL); if (unlikely(!__pyx_t_513)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_513);

  /* "ultron/tradingday/Calendar.pyx":690
 *     Date(2021, 10, 6),
 *     Date(2021, 10, 7),
 *     Date(2022, 1, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 1, 31),
 *     Date(2022, 2, 1),
 */
  __pyx_t_514 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__517, NULL); if (unlikely(!__pyx_t_514)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_514);

  /* "ultron/tradingday/Calendar.pyx":691
 *     Date(2021, 10, 7),
 *     Date(2022, 1, 3),
 *     Date(2022, 1, 31),             # <<<<<<<<<<<<<<
 *     Date(2022, 2, 1),
 *     Date(2022, 2, 2),
 */
  __pyx_t_515 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__518, NULL); if (unlikely(!__pyx_t_515)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_515);

  /* "ultron/tradingday/Calendar.pyx":692
 *     Date(2022, 1, 3),
 *     Date(2022, 1, 31),
 *     Date(2022, 2, 1),             # <<<<<<<<<<<<<<
 *     Date(2022, 2, 2),
 *     Date(2022, 2, 3),
 */
  __pyx_t_516 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__519, NULL); if (unlikely(!__pyx_t_516)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_516);

  /* "ultron/tradingday/Calendar.pyx":693
 *     Date(2022, 1, 31),
 *     Date(2022, 2, 1),
 *     Date(2022, 2, 2),             # <<<<<<<<<<<<<<
 *     Date(2022, 2, 3),
 *     Date(2022, 2, 4),
 */
  __pyx_t_517 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__520, NULL); if (unlikely(!__pyx_t_517)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_517);

  /* "ultron/tradingday/Calendar.pyx":694
 *     Date(2022, 2, 1),
 *     Date(2022, 2, 2),
 *     Date(2022, 2, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 2, 4),
 *     Date(2022, 4, 4),
 */
  __pyx_t_518 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__521, NULL); if (unlikely(!__pyx_t_518)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_518);

  /* "ultron/tradingday/Calendar.pyx":695
 *     Date(2022, 2, 2),
 *     Date(2022, 2, 3),
 *     Date(2022, 2, 4),             # <<<<<<<<<<<<<<
 *     Date(2022, 4, 4),
 *     Date(2022, 4, 5),
 */
  __pyx_t_519 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__522, NULL); if (unlikely(!__pyx_t_519)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_519);

  /* "ultron/tradingday/Calendar.pyx":696
 *     Date(2022, 2, 3),
 *     Date(2022, 2, 4),
 *     Date(2022, 4, 4),             # <<<<<<<<<<<<<<
 *     Date(2022, 4, 5),
 *     Date(2022, 5, 2),
 */
  __pyx_t_520 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__523, NULL); if (unlikely(!__pyx_t_520)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_520);

  /* "ultron/tradingday/Calendar.pyx":697
 *     Date(2022, 2, 4),
 *     Date(2022, 4, 4),
 *     Date(2022, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2022, 5, 2),
 *     Date(2022, 5, 3),
 */
  __pyx_t_521 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__524, NULL); if (unlikely(!__pyx_t_521)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_521);

  /* "ultron/tradingday/Calendar.pyx":698
 *     Date(2022, 4, 4),
 *     Date(2022, 4, 5),
 *     Date(2022, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2022, 5, 3),
 *     Date(2022, 5, 4),
 */
  __pyx_t_522 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__525, NULL); if (unlikely(!__pyx_t_522)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_522);

  /* "ultron/tradingday/Calendar.pyx":699
 *     Date(2022, 4, 5),
 *     Date(2022, 5, 2),
 *     Date(2022, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 5, 4),
 *     Date(2022, 6, 3),
 */
  __pyx_t_523 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__526, NULL); if (unlikely(!__pyx_t_523)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_523);

  /* "ultron/tradingday/Calendar.pyx":700
 *     Date(2022, 5, 2),
 *     Date(2022, 5, 3),
 *     Date(2022, 5, 4),             # <<<<<<<<<<<<<<
 *     Date(2022, 6, 3),
 *     Date(2022, 9, 12),
 */
  __pyx_t_524 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__527, NULL); if (unlikely(!__pyx_t_524)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_524);

  /* "ultron/tradingday/Calendar.pyx":701
 *     Date(2022, 5, 3),
 *     Date(2022, 5, 4),
 *     Date(2022, 6, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 9, 12),
 *     Date(2022, 10, 3),
 */
  __pyx_t_525 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__528, NULL); if (unlikely(!__pyx_t_525)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_525);

  /* "ultron/tradingday/Calendar.pyx":702
 *     Date(2022, 5, 4),
 *     Date(2022, 6, 3),
 *     Date(2022, 9, 12),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 3),
 *     Date(2022, 10, 4),
 */
  __pyx_t_526 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__529, NULL); if (unlikely(!__pyx_t_526)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_526);

  /* "ultron/tradingday/Calendar.pyx":703
 *     Date(2022, 6, 3),
 *     Date(2022, 9, 12),
 *     Date(2022, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 4),
 *     Date(2022, 10, 5),
 */
  __pyx_t_527 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__530, NULL); if (unlikely(!__pyx_t_527)) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_527);

  /* "ultron/tradingday/Calendar.pyx":704
 *     Date(2022, 9, 12),
 *     Date(2022, 10, 3),
 *     Date(2022, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 5),
 *     Date(2022, 10, 6),
 */
  __pyx_t_528 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__531, NULL); if (unlikely(!__pyx_t_528)) __PYX_ERR(0, 704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_528);

  /* "ultron/tradingday/Calendar.pyx":705
 *     Date(2022, 10, 3),
 *     Date(2022, 10, 4),
 *     Date(2022, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 6),
 *     Date(2022, 10, 7),
 */
  __pyx_t_529 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__532, NULL); if (unlikely(!__pyx_t_529)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_529);

  /* "ultron/tradingday/Calendar.pyx":706
 *     Date(2022, 10, 4),
 *     Date(2022, 10, 5),
 *     Date(2022, 10, 6),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 7),
 *     Date(2023, 1, 2),
 */
  __pyx_t_530 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__533, NULL); if (unlikely(!__pyx_t_530)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_530);

  /* "ultron/tradingday/Calendar.pyx":707
 *     Date(2022, 10, 5),
 *     Date(2022, 10, 6),
 *     Date(2022, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 2),
 *     Date(2023, 1, 23),
 */
  __pyx_t_531 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__534, NULL); if (unlikely(!__pyx_t_531)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_531);

  /* "ultron/tradingday/Calendar.pyx":708
 *     Date(2022, 10, 6),
 *     Date(2022, 10, 7),
 *     Date(2023, 1, 2),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 23),
 *     Date(2023, 1, 24),
 */
  __pyx_t_532 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__535, NULL); if (unlikely(!__pyx_t_532)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_532);

  /* "ultron/tradingday/Calendar.pyx":709
 *     Date(2022, 10, 7),
 *     Date(2023, 1, 2),
 *     Date(2023, 1, 23),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 24),
 *     Date(2023, 1, 25),
 */
  __pyx_t_533 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__536, NULL); if (unlikely(!__pyx_t_533)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_533);

  /* "ultron/tradingday/Calendar.pyx":710
 *     Date(2023, 1, 2),
 *     Date(2023, 1, 23),
 *     Date(2023, 1, 24),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 25),
 *     Date(2023, 1, 26),
 */
  __pyx_t_534 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__537, NULL); if (unlikely(!__pyx_t_534)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_534);

  /* "ultron/tradingday/Calendar.pyx":711
 *     Date(2023, 1, 23),
 *     Date(2023, 1, 24),
 *     Date(2023, 1, 25),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 26),
 *     Date(2023, 1, 27),
 */
  __pyx_t_535 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__538, NULL); if (unlikely(!__pyx_t_535)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_535);

  /* "ultron/tradingday/Calendar.pyx":712
 *     Date(2023, 1, 24),
 *     Date(2023, 1, 25),
 *     Date(2023, 1, 26),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 27),
 *     Date(2023, 4, 5),
 */
  __pyx_t_536 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__539, NULL); if (unlikely(!__pyx_t_536)) __PYX_ERR(0, 712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_536);

  /* "ultron/tradingday/Calendar.pyx":713
 *     Date(2023, 1, 25),
 *     Date(2023, 1, 26),
 *     Date(2023, 1, 27),             # <<<<<<<<<<<<<<
 *     Date(2023, 4, 5),
 *     Date(2023, 5, 1),
 */
  __pyx_t_537 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__540, NULL); if (unlikely(!__pyx_t_537)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_537);

  /* "ultron/tradingday/Calendar.pyx":714
 *     Date(2023, 1, 26),
 *     Date(2023, 1, 27),
 *     Date(2023, 4, 5),             # <<<<<<<<<<<<<<
 *     Date(2023, 5, 1),
 *     Date(2023, 5, 2),
 */
  __pyx_t_538 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__541, NULL); if (unlikely(!__pyx_t_538)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_538);

  /* "ultron/tradingday/Calendar.pyx":715
 *     Date(2023, 1, 27),
 *     Date(2023, 4, 5),
 *     Date(2023, 5, 1),             # <<<<<<<<<<<<<<
 *     Date(2023, 5, 2),
 *     Date(2023, 5, 3),
 */
  __pyx_t_539 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__542, NULL); if (unlikely(!__pyx_t_539)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_539);

  /* "ultron/tradingday/Calendar.pyx":716
 *     Date(2023, 4, 5),
 *     Date(2023, 5, 1),
 *     Date(2023, 5, 2),             # <<<<<<<<<<<<<<
 *     Date(2023, 5, 3),
 *     Date(2023, 6, 22),
 */
  __pyx_t_540 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__543, NULL); if (unlikely(!__pyx_t_540)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_540);

  /* "ultron/tradingday/Calendar.pyx":717
 *     Date(2023, 5, 1),
 *     Date(2023, 5, 2),
 *     Date(2023, 5, 3),             # <<<<<<<<<<<<<<
 *     Date(2023, 6, 22),
 *     Date(2023, 6, 23),
 */
  __pyx_t_541 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__544, NULL); if (unlikely(!__pyx_t_541)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_541);

  /* "ultron/tradingday/Calendar.pyx":718
 *     Date(2023, 5, 2),
 *     Date(2023, 5, 3),
 *     Date(2023, 6, 22),             # <<<<<<<<<<<<<<
 *     Date(2023, 6, 23),
 *     Date(2023, 9, 29),
 */
  __pyx_t_542 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__545, NULL); if (unlikely(!__pyx_t_542)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_542);

  /* "ultron/tradingday/Calendar.pyx":719
 *     Date(2023, 5, 3),
 *     Date(2023, 6, 22),
 *     Date(2023, 6, 23),             # <<<<<<<<<<<<<<
 *     Date(2023, 9, 29),
 *     Date(2023, 10, 2),
 */
  __pyx_t_543 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__546, NULL); if (unlikely(!__pyx_t_543)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_543);

  /* "ultron/tradingday/Calendar.pyx":720
 *     Date(2023, 6, 22),
 *     Date(2023, 6, 23),
 *     Date(2023, 9, 29),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 2),
 *     Date(2023, 10, 3),
 */
  __pyx_t_544 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__547, NULL); if (unlikely(!__pyx_t_544)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_544);

  /* "ultron/tradingday/Calendar.pyx":721
 *     Date(2023, 6, 23),
 *     Date(2023, 9, 29),
 *     Date(2023, 10, 2),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 3),
 *     Date(2023, 10, 4),
 */
  __pyx_t_545 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__548, NULL); if (unlikely(!__pyx_t_545)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_545);

  /* "ultron/tradingday/Calendar.pyx":722
 *     Date(2023, 9, 29),
 *     Date(2023, 10, 2),
 *     Date(2023, 10, 3),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 4),
 *     Date(2023, 10, 5),
 */
  __pyx_t_546 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__549, NULL); if (unlikely(!__pyx_t_546)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_546);

  /* "ultron/tradingday/Calendar.pyx":723
 *     Date(2023, 10, 2),
 *     Date(2023, 10, 3),
 *     Date(2023, 10, 4),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 5),
 *     Date(2023, 10, 6)
 */
  __pyx_t_547 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__550, NULL); if (unlikely(!__pyx_t_547)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_547);

  /* "ultron/tradingday/Calendar.pyx":724
 *     Date(2023, 10, 3),
 *     Date(2023, 10, 4),
 *     Date(2023, 10, 5),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 6)
 * }
 */
  __pyx_t_548 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__551, NULL); if (unlikely(!__pyx_t_548)) __PYX_ERR(0, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_548);

  /* "ultron/tradingday/Calendar.pyx":725
 *     Date(2023, 10, 4),
 *     Date(2023, 10, 5),
 *     Date(2023, 10, 6)             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_t_549 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__552, NULL); if (unlikely(!__pyx_t_549)) __PYX_ERR(0, 725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_549);
  __pyx_t_550 = PySet_New(0); if (unlikely(!__pyx_t_550)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_550);
  if (PySet_Add(__pyx_t_550, __pyx_t_1) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_2) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_3) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_4) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_5) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_6) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_7) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_8) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_9) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_10) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_11) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_12) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_13) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_14) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_15) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_16) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_17) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_18) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_19) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_20) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_21) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_22) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_23) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_24) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_25) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_26) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_27) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_28) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_29) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_30) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_31) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_32) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_33) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_34) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_35) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_36) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_37) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_38) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_39) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_40) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_41) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_42) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_43) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_44) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_45) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_46) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_47) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_48) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_49) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_50) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_51) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_52) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_53) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_54) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_55) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_56) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_57) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_58) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_59) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_60) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_61) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_62) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_63) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_64) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_65) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_66) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_67) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_68) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_69) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_70) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_71) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_72) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_73) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_74) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_75) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_76) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_77) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_78) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_79) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_80) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_81) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_82) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_83) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_84) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_85) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_86) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_87) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_88) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_89) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_90) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_91) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_92) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_93) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_94) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_95) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_96) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_97) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_98) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_99) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_100) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_101) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_102) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_103) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_104) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_105) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_106) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_107) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_108) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_109) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_110) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_111) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_112) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_113) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_114) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_115) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_116) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_117) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_118) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_119) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_120) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_121) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_122) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_123) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_124) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_125) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_126) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_127) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_128) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_129) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_130) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_131) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_132) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_132); __pyx_t_132 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_133) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_133); __pyx_t_133 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_134) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_134); __pyx_t_134 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_135) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_135); __pyx_t_135 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_136) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_136); __pyx_t_136 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_137) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_137); __pyx_t_137 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_138) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_138); __pyx_t_138 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_139) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_139); __pyx_t_139 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_140) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_140); __pyx_t_140 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_141) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_141); __pyx_t_141 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_142) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_142); __pyx_t_142 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_143) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_143); __pyx_t_143 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_144) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_144); __pyx_t_144 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_145) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_145); __pyx_t_145 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_146) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_146); __pyx_t_146 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_147) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_147); __pyx_t_147 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_148) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_148); __pyx_t_148 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_149) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_149); __pyx_t_149 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_150) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_150); __pyx_t_150 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_151) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_151); __pyx_t_151 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_152) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_152); __pyx_t_152 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_153) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_153); __pyx_t_153 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_154) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_154); __pyx_t_154 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_155) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_155); __pyx_t_155 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_156) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_156); __pyx_t_156 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_157) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_157); __pyx_t_157 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_158) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_158); __pyx_t_158 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_159) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_159); __pyx_t_159 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_160) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_160); __pyx_t_160 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_161) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_161); __pyx_t_161 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_162) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_162); __pyx_t_162 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_163) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_163); __pyx_t_163 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_164) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_164); __pyx_t_164 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_165) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_165); __pyx_t_165 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_166) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_166); __pyx_t_166 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_167) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_167); __pyx_t_167 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_168) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_168); __pyx_t_168 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_169) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_169); __pyx_t_169 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_170) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_170); __pyx_t_170 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_171) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_171); __pyx_t_171 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_172) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_172); __pyx_t_172 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_173) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_173); __pyx_t_173 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_174) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_174); __pyx_t_174 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_175) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_175); __pyx_t_175 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_176) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_176); __pyx_t_176 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_177) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_177); __pyx_t_177 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_178) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_178); __pyx_t_178 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_179) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_179); __pyx_t_179 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_180) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_180); __pyx_t_180 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_181) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_181); __pyx_t_181 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_182) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_182); __pyx_t_182 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_183) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_183); __pyx_t_183 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_184) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_184); __pyx_t_184 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_185) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_185); __pyx_t_185 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_186) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_186); __pyx_t_186 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_187) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_187); __pyx_t_187 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_188) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_188); __pyx_t_188 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_189) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_189); __pyx_t_189 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_190) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_190); __pyx_t_190 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_191) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_191); __pyx_t_191 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_192) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_192); __pyx_t_192 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_193) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_193); __pyx_t_193 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_194) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_194); __pyx_t_194 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_195) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_195); __pyx_t_195 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_196) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_196); __pyx_t_196 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_197) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_197); __pyx_t_197 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_198) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_198); __pyx_t_198 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_199) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_199); __pyx_t_199 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_200) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_200); __pyx_t_200 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_201) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_201); __pyx_t_201 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_202) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_202); __pyx_t_202 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_203) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_203); __pyx_t_203 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_204) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_204); __pyx_t_204 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_205) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_205); __pyx_t_205 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_206) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_206); __pyx_t_206 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_207) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_207); __pyx_t_207 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_208) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_208); __pyx_t_208 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_209) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_209); __pyx_t_209 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_210) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_210); __pyx_t_210 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_211) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_211); __pyx_t_211 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_212) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_212); __pyx_t_212 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_213) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_213); __pyx_t_213 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_214) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_214); __pyx_t_214 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_215) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_215); __pyx_t_215 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_216) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_216); __pyx_t_216 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_217) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_217); __pyx_t_217 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_218) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_218); __pyx_t_218 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_219) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_219); __pyx_t_219 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_220) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_220); __pyx_t_220 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_221) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_221); __pyx_t_221 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_222) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_222); __pyx_t_222 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_223) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_223); __pyx_t_223 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_224) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_224); __pyx_t_224 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_225) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_225); __pyx_t_225 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_226) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_226); __pyx_t_226 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_227) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_227); __pyx_t_227 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_228) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_228); __pyx_t_228 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_229) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_229); __pyx_t_229 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_230) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_230); __pyx_t_230 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_231) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_231); __pyx_t_231 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_232) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_232); __pyx_t_232 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_233) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_233); __pyx_t_233 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_234) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_234); __pyx_t_234 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_235) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_235); __pyx_t_235 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_236) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_236); __pyx_t_236 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_237) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_237); __pyx_t_237 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_238) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_238); __pyx_t_238 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_239) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_239); __pyx_t_239 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_240) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_240); __pyx_t_240 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_241) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_241); __pyx_t_241 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_242) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_242); __pyx_t_242 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_243) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_243); __pyx_t_243 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_244) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_244); __pyx_t_244 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_245) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_245); __pyx_t_245 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_246) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_246); __pyx_t_246 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_247) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_247); __pyx_t_247 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_248) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_248); __pyx_t_248 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_249) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_249); __pyx_t_249 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_250) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_250); __pyx_t_250 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_251) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_251); __pyx_t_251 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_252) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_252); __pyx_t_252 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_253) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_253); __pyx_t_253 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_254) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_254); __pyx_t_254 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_255) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_255); __pyx_t_255 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_256) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_256); __pyx_t_256 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_257) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_257); __pyx_t_257 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_258) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_258); __pyx_t_258 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_259) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_259); __pyx_t_259 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_260) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_260); __pyx_t_260 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_261) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_261); __pyx_t_261 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_262) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_262); __pyx_t_262 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_263) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_263); __pyx_t_263 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_264) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_264); __pyx_t_264 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_265) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_265); __pyx_t_265 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_266) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_266); __pyx_t_266 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_267) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_267); __pyx_t_267 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_268) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_268); __pyx_t_268 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_269) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_269); __pyx_t_269 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_270) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_270); __pyx_t_270 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_271) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_271); __pyx_t_271 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_272) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_272); __pyx_t_272 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_273) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_273); __pyx_t_273 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_274) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_274); __pyx_t_274 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_275) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_275); __pyx_t_275 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_276) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_276); __pyx_t_276 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_277) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_277); __pyx_t_277 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_278) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_278); __pyx_t_278 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_279) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_279); __pyx_t_279 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_280) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_280); __pyx_t_280 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_281) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_281); __pyx_t_281 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_282) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_282); __pyx_t_282 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_283) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_283); __pyx_t_283 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_284) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_284); __pyx_t_284 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_285) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_285); __pyx_t_285 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_286) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_286); __pyx_t_286 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_287) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_287); __pyx_t_287 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_288) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_288); __pyx_t_288 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_289) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_289); __pyx_t_289 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_290) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_290); __pyx_t_290 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_291) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_291); __pyx_t_291 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_292) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_292); __pyx_t_292 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_293) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_293); __pyx_t_293 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_294) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_294); __pyx_t_294 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_295) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_295); __pyx_t_295 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_296) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_296); __pyx_t_296 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_297) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_297); __pyx_t_297 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_298) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_298); __pyx_t_298 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_299) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_299); __pyx_t_299 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_300) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_300); __pyx_t_300 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_301) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_301); __pyx_t_301 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_302) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_302); __pyx_t_302 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_303) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_303); __pyx_t_303 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_304) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_304); __pyx_t_304 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_305) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_305); __pyx_t_305 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_306) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_306); __pyx_t_306 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_307) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_307); __pyx_t_307 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_308) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_308); __pyx_t_308 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_309) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_309); __pyx_t_309 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_310) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_310); __pyx_t_310 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_311) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_311); __pyx_t_311 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_312) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_312); __pyx_t_312 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_313) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_313); __pyx_t_313 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_314) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_314); __pyx_t_314 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_315) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_315); __pyx_t_315 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_316) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_316); __pyx_t_316 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_317) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_317); __pyx_t_317 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_318) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_318); __pyx_t_318 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_319) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_319); __pyx_t_319 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_320) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_320); __pyx_t_320 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_321) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_321); __pyx_t_321 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_322) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_322); __pyx_t_322 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_323) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_323); __pyx_t_323 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_324) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_324); __pyx_t_324 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_325) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_325); __pyx_t_325 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_326) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_326); __pyx_t_326 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_327) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_327); __pyx_t_327 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_328) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_328); __pyx_t_328 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_329) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_329); __pyx_t_329 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_330) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_330); __pyx_t_330 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_331) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_331); __pyx_t_331 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_332) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_332); __pyx_t_332 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_333) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_333); __pyx_t_333 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_334) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_334); __pyx_t_334 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_335) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_335); __pyx_t_335 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_336) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_336); __pyx_t_336 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_337) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_337); __pyx_t_337 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_338) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_338); __pyx_t_338 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_339) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_339); __pyx_t_339 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_340) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_340); __pyx_t_340 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_341) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_341); __pyx_t_341 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_342) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_342); __pyx_t_342 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_343) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_343); __pyx_t_343 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_344) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_344); __pyx_t_344 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_345) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_345); __pyx_t_345 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_346) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_346); __pyx_t_346 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_347) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_347); __pyx_t_347 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_348) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_348); __pyx_t_348 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_349) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_349); __pyx_t_349 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_350) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_350); __pyx_t_350 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_351) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_351); __pyx_t_351 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_352) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_352); __pyx_t_352 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_353) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_353); __pyx_t_353 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_354) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_354); __pyx_t_354 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_355) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_355); __pyx_t_355 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_356) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_356); __pyx_t_356 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_357) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_357); __pyx_t_357 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_358) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_358); __pyx_t_358 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_359) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_359); __pyx_t_359 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_360) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_360); __pyx_t_360 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_361) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_361); __pyx_t_361 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_362) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_362); __pyx_t_362 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_363) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_363); __pyx_t_363 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_364) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_364); __pyx_t_364 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_365) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_365); __pyx_t_365 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_366) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_366); __pyx_t_366 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_367) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_367); __pyx_t_367 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_368) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_368); __pyx_t_368 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_369) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_369); __pyx_t_369 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_370) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_370); __pyx_t_370 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_371) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_371); __pyx_t_371 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_372) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_372); __pyx_t_372 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_373) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_373); __pyx_t_373 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_374) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_374); __pyx_t_374 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_375) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_375); __pyx_t_375 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_376) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_376); __pyx_t_376 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_377) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_377); __pyx_t_377 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_378) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_378); __pyx_t_378 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_379) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_379); __pyx_t_379 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_380) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_380); __pyx_t_380 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_381) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_381); __pyx_t_381 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_382) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_382); __pyx_t_382 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_383) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_383); __pyx_t_383 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_384) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_384); __pyx_t_384 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_385) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_385); __pyx_t_385 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_386) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_386); __pyx_t_386 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_387) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_387); __pyx_t_387 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_388) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_388); __pyx_t_388 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_389) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_389); __pyx_t_389 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_390) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_390); __pyx_t_390 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_391) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_391); __pyx_t_391 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_392) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_392); __pyx_t_392 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_393) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_393); __pyx_t_393 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_394) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_394); __pyx_t_394 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_395) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_395); __pyx_t_395 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_396) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_396); __pyx_t_396 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_397) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_397); __pyx_t_397 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_398) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_398); __pyx_t_398 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_399) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_399); __pyx_t_399 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_400) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_400); __pyx_t_400 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_401) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_401); __pyx_t_401 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_402) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_402); __pyx_t_402 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_403) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_403); __pyx_t_403 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_404) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_404); __pyx_t_404 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_405) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_405); __pyx_t_405 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_406) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_406); __pyx_t_406 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_407) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_407); __pyx_t_407 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_408) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_408); __pyx_t_408 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_409) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_409); __pyx_t_409 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_410) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_410); __pyx_t_410 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_411) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_411); __pyx_t_411 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_412) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_412); __pyx_t_412 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_413) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_413); __pyx_t_413 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_414) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_414); __pyx_t_414 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_415) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_415); __pyx_t_415 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_416) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_416); __pyx_t_416 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_417) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_417); __pyx_t_417 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_418) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_418); __pyx_t_418 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_419) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_419); __pyx_t_419 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_420) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_420); __pyx_t_420 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_421) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_421); __pyx_t_421 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_422) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_422); __pyx_t_422 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_423) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_423); __pyx_t_423 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_424) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_424); __pyx_t_424 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_425) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_425); __pyx_t_425 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_426) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_426); __pyx_t_426 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_427) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_427); __pyx_t_427 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_428) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_428); __pyx_t_428 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_429) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_429); __pyx_t_429 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_430) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_430); __pyx_t_430 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_431) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_431); __pyx_t_431 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_432) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_432); __pyx_t_432 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_433) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_433); __pyx_t_433 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_434) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_434); __pyx_t_434 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_435) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_435); __pyx_t_435 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_436) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_436); __pyx_t_436 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_437) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_437); __pyx_t_437 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_438) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_438); __pyx_t_438 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_439) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_439); __pyx_t_439 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_440) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_440); __pyx_t_440 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_441) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_441); __pyx_t_441 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_442) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_442); __pyx_t_442 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_443) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_443); __pyx_t_443 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_444) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_444); __pyx_t_444 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_445) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_445); __pyx_t_445 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_446) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_446); __pyx_t_446 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_447) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_447); __pyx_t_447 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_448) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_448); __pyx_t_448 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_449) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_449); __pyx_t_449 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_450) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_450); __pyx_t_450 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_451) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_451); __pyx_t_451 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_452) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_452); __pyx_t_452 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_453) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_453); __pyx_t_453 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_454) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_454); __pyx_t_454 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_455) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_455); __pyx_t_455 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_456) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_456); __pyx_t_456 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_457) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_457); __pyx_t_457 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_458) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_458); __pyx_t_458 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_459) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_459); __pyx_t_459 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_460) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_460); __pyx_t_460 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_461) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_461); __pyx_t_461 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_462) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_462); __pyx_t_462 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_463) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_463); __pyx_t_463 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_464) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_464); __pyx_t_464 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_465) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_465); __pyx_t_465 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_466) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_466); __pyx_t_466 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_467) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_467); __pyx_t_467 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_468) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_468); __pyx_t_468 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_469) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_469); __pyx_t_469 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_470) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_470); __pyx_t_470 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_471) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_471); __pyx_t_471 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_472) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_472); __pyx_t_472 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_473) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_473); __pyx_t_473 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_474) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_474); __pyx_t_474 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_475) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_475); __pyx_t_475 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_476) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_476); __pyx_t_476 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_477) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_477); __pyx_t_477 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_478) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_478); __pyx_t_478 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_479) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_479); __pyx_t_479 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_480) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_480); __pyx_t_480 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_481) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_481); __pyx_t_481 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_482) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_482); __pyx_t_482 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_483) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_483); __pyx_t_483 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_484) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_484); __pyx_t_484 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_485) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_485); __pyx_t_485 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_486) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_486); __pyx_t_486 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_487) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_487); __pyx_t_487 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_488) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_488); __pyx_t_488 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_489) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_489); __pyx_t_489 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_490) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_490); __pyx_t_490 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_491) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_491); __pyx_t_491 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_492) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_492); __pyx_t_492 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_493) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_493); __pyx_t_493 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_494) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_494); __pyx_t_494 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_495) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_495); __pyx_t_495 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_496) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_496); __pyx_t_496 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_497) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_497); __pyx_t_497 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_498) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_498); __pyx_t_498 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_499) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_499); __pyx_t_499 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_500) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_500); __pyx_t_500 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_501) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_501); __pyx_t_501 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_502) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_502); __pyx_t_502 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_503) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_503); __pyx_t_503 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_504) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_504); __pyx_t_504 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_505) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_505); __pyx_t_505 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_506) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_506); __pyx_t_506 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_507) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_507); __pyx_t_507 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_508) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_508); __pyx_t_508 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_509) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_509); __pyx_t_509 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_510) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_510); __pyx_t_510 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_511) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_511); __pyx_t_511 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_512) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_512); __pyx_t_512 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_513) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_513); __pyx_t_513 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_514) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_514); __pyx_t_514 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_515) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_515); __pyx_t_515 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_516) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_516); __pyx_t_516 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_517) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_517); __pyx_t_517 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_518) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_518); __pyx_t_518 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_519) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_519); __pyx_t_519 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_520) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_520); __pyx_t_520 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_521) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_521); __pyx_t_521 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_522) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_522); __pyx_t_522 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_523) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_523); __pyx_t_523 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_524) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_524); __pyx_t_524 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_525) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_525); __pyx_t_525 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_526) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_526); __pyx_t_526 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_527) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_527); __pyx_t_527 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_528) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_528); __pyx_t_528 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_529) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_529); __pyx_t_529 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_530) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_530); __pyx_t_530 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_531) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_531); __pyx_t_531 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_532) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_532); __pyx_t_532 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_533) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_533); __pyx_t_533 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_534) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_534); __pyx_t_534 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_535) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_535); __pyx_t_535 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_536) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_536); __pyx_t_536 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_537) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_537); __pyx_t_537 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_538) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_538); __pyx_t_538 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_539) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_539); __pyx_t_539 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_540) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_540); __pyx_t_540 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_541) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_541); __pyx_t_541 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_542) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_542); __pyx_t_542 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_543) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_543); __pyx_t_543 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_544) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_544); __pyx_t_544 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_545) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_545); __pyx_t_545 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_546) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_546); __pyx_t_546 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_547) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_547); __pyx_t_547 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_548) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_548); __pyx_t_548 = 0;
  if (PySet_Add(__pyx_t_550, __pyx_t_549) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_549); __pyx_t_549 = 0;
  __Pyx_XGOTREF(__pyx_v_6ultron_10tradingday_8Calendar_sse_holDays);
  __Pyx_DECREF_SET(__pyx_v_6ultron_10tradingday_8Calendar_sse_holDays, ((PyObject*)__pyx_t_550));
  __Pyx_GIVEREF(__pyx_t_550);
  __pyx_t_550 = 0;

  /* "ultron/tradingday/Calendar.pyx":731
 * cdef set sse_working_weekends = {
 *     # 1992
 *     Date.westernStyle(4, Months.October, 1992)             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_t_550 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_550)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_550);
  __pyx_t_549 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_549)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_549);
  __pyx_t_548 = PyTuple_New(3); if (unlikely(!__pyx_t_548)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_548);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_548, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_t_549);
  PyTuple_SET_ITEM(__pyx_t_548, 1, __pyx_t_549);
  __Pyx_INCREF(__pyx_int_1992);
  __Pyx_GIVEREF(__pyx_int_1992);
  PyTuple_SET_ITEM(__pyx_t_548, 2, __pyx_int_1992);
  __pyx_t_549 = 0;
  __pyx_t_549 = __Pyx_PyObject_Call(__pyx_t_550, __pyx_t_548, NULL); if (unlikely(!__pyx_t_549)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_549);
  __Pyx_DECREF(__pyx_t_550); __pyx_t_550 = 0;
  __Pyx_DECREF(__pyx_t_548); __pyx_t_548 = 0;
  __pyx_t_548 = PySet_New(0); if (unlikely(!__pyx_t_548)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_548);
  if (PySet_Add(__pyx_t_548, __pyx_t_549) < 0) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_549); __pyx_t_549 = 0;
  __Pyx_XGOTREF(__pyx_v_6ultron_10tradingday_8Calendar_sse_working_weekends);
  __Pyx_DECREF_SET(__pyx_v_6ultron_10tradingday_8Calendar_sse_working_weekends, ((PyObject*)__pyx_t_548));
  __Pyx_GIVEREF(__pyx_t_548);
  __pyx_t_548 = 0;

  /* "ultron/tradingday/Calendar.pyx":754
 * cdef set ib_working_weekends = {
 *     # 1992
 *     Date.westernStyle(4, Months.October, 1992),             # <<<<<<<<<<<<<<
 *     # 2005
 *     Date.westernStyle(5, Months.February, 2005),
 */
  __pyx_t_548 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_548)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_548);
  __pyx_t_549 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_549)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_549);
  __pyx_t_550 = PyTuple_New(3); if (unlikely(!__pyx_t_550)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_550);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_550, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_t_549);
  PyTuple_SET_ITEM(__pyx_t_550, 1, __pyx_t_549);
  __Pyx_INCREF(__pyx_int_1992);
  __Pyx_GIVEREF(__pyx_int_1992);
  PyTuple_SET_ITEM(__pyx_t_550, 2, __pyx_int_1992);
  __pyx_t_549 = 0;
  __pyx_t_549 = __Pyx_PyObject_Call(__pyx_t_548, __pyx_t_550, NULL); if (unlikely(!__pyx_t_549)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_549);
  __Pyx_DECREF(__pyx_t_548); __pyx_t_548 = 0;
  __Pyx_DECREF(__pyx_t_550); __pyx_t_550 = 0;

  /* "ultron/tradingday/Calendar.pyx":756
 *     Date.westernStyle(4, Months.October, 1992),
 *     # 2005
 *     Date.westernStyle(5, Months.February, 2005),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(6, Months.February, 2005),
 *     Date.westernStyle(30, Months.April, 2005),
 */
  __pyx_t_550 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_550)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_550);
  __pyx_t_548 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_548)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_548);
  __pyx_t_547 = PyTuple_New(3); if (unlikely(!__pyx_t_547)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_547);
  __Pyx_INCREF(__pyx_int_5);
  __Pyx_GIVEREF(__pyx_int_5);
  PyTuple_SET_ITEM(__pyx_t_547, 0, __pyx_int_5);
  __Pyx_GIVEREF(__pyx_t_548);
  PyTuple_SET_ITEM(__pyx_t_547, 1, __pyx_t_548);
  __Pyx_INCREF(__pyx_int_2005);
  __Pyx_GIVEREF(__pyx_int_2005);
  PyTuple_SET_ITEM(__pyx_t_547, 2, __pyx_int_2005);
  __pyx_t_548 = 0;
  __pyx_t_548 = __Pyx_PyObject_Call(__pyx_t_550, __pyx_t_547, NULL); if (unlikely(!__pyx_t_548)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_548);
  __Pyx_DECREF(__pyx_t_550); __pyx_t_550 = 0;
  __Pyx_DECREF(__pyx_t_547); __pyx_t_547 = 0;

  /* "ultron/tradingday/Calendar.pyx":757
 *     # 2005
 *     Date.westernStyle(5, Months.February, 2005),
 *     Date.westernStyle(6, Months.February, 2005),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(30, Months.April, 2005),
 *     Date.westernStyle(8, Months.May, 2005),
 */
  __pyx_t_547 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_547)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_547);
  __pyx_t_550 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_550)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_550);
  __pyx_t_546 = PyTuple_New(3); if (unlikely(!__pyx_t_546)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_546);
  __Pyx_INCREF(__pyx_int_6);
  __Pyx_GIVEREF(__pyx_int_6);
  PyTuple_SET_ITEM(__pyx_t_546, 0, __pyx_int_6);
  __Pyx_GIVEREF(__pyx_t_550);
  PyTuple_SET_ITEM(__pyx_t_546, 1, __pyx_t_550);
  __Pyx_INCREF(__pyx_int_2005);
  __Pyx_GIVEREF(__pyx_int_2005);
  PyTuple_SET_ITEM(__pyx_t_546, 2, __pyx_int_2005);
  __pyx_t_550 = 0;
  __pyx_t_550 = __Pyx_PyObject_Call(__pyx_t_547, __pyx_t_546, NULL); if (unlikely(!__pyx_t_550)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_550);
  __Pyx_DECREF(__pyx_t_547); __pyx_t_547 = 0;
  __Pyx_DECREF(__pyx_t_546); __pyx_t_546 = 0;

  /* "ultron/tradingday/Calendar.pyx":758
 *     Date.westernStyle(5, Months.February, 2005),
 *     Date.westernStyle(6, Months.February, 2005),
 *     Date.westernStyle(30, Months.April, 2005),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(8, Months.May, 2005),
 *     Date.westernStyle(8, Months.October, 2005),
 */
  __pyx_t_546 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_546)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_546);
  __pyx_t_547 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_547)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_547);
  __pyx_t_545 = PyTuple_New(3); if (unlikely(!__pyx_t_545)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_545);
  __Pyx_INCREF(__pyx_int_30);
  __Pyx_GIVEREF(__pyx_int_30);
  PyTuple_SET_ITEM(__pyx_t_545, 0, __pyx_int_30);
  __Pyx_GIVEREF(__pyx_t_547);
  PyTuple_SET_ITEM(__pyx_t_545, 1, __pyx_t_547);
  __Pyx_INCREF(__pyx_int_2005);
  __Pyx_GIVEREF(__pyx_int_2005);
  PyTuple_SET_ITEM(__pyx_t_545, 2, __pyx_int_2005);
  __pyx_t_547 = 0;
  __pyx_t_547 = __Pyx_PyObject_Call(__pyx_t_546, __pyx_t_545, NULL); if (unlikely(!__pyx_t_547)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_547);
  __Pyx_DECREF(__pyx_t_546); __pyx_t_546 = 0;
  __Pyx_DECREF(__pyx_t_545); __pyx_t_545 = 0;

  /* "ultron/tradingday/Calendar.pyx":759
 *     Date.westernStyle(6, Months.February, 2005),
 *     Date.westernStyle(30, Months.April, 2005),
 *     Date.westernStyle(8, Months.May, 2005),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(8, Months.October, 2005),
 *     Date.westernStyle(9, Months.October, 2005),
 */
  __pyx_t_545 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_545)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_545);
  __pyx_t_546 = __Pyx_PyInt_From_enum__Months(May); if (unlikely(!__pyx_t_546)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_546);
  __pyx_t_544 = PyTuple_New(3); if (unlikely(!__pyx_t_544)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_544);
  __Pyx_INCREF(__pyx_int_8);
  __Pyx_GIVEREF(__pyx_int_8);
  PyTuple_SET_ITEM(__pyx_t_544, 0, __pyx_int_8);
  __Pyx_GIVEREF(__pyx_t_546);
  PyTuple_SET_ITEM(__pyx_t_544, 1, __pyx_t_546);
  __Pyx_INCREF(__pyx_int_2005);
  __Pyx_GIVEREF(__pyx_int_2005);
  PyTuple_SET_ITEM(__pyx_t_544, 2, __pyx_int_2005);
  __pyx_t_546 = 0;
  __pyx_t_546 = __Pyx_PyObject_Call(__pyx_t_545, __pyx_t_544, NULL); if (unlikely(!__pyx_t_546)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_546);
  __Pyx_DECREF(__pyx_t_545); __pyx_t_545 = 0;
  __Pyx_DECREF(__pyx_t_544); __pyx_t_544 = 0;

  /* "ultron/tradingday/Calendar.pyx":760
 *     Date.westernStyle(30, Months.April, 2005),
 *     Date.westernStyle(8, Months.May, 2005),
 *     Date.westernStyle(8, Months.October, 2005),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(9, Months.October, 2005),
 *     Date.westernStyle(31, Months.December, 2005),
 */
  __pyx_t_544 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_544)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_544);
  __pyx_t_545 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_545)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_545);
  __pyx_t_543 = PyTuple_New(3); if (unlikely(!__pyx_t_543)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_543);
  __Pyx_INCREF(__pyx_int_8);
  __Pyx_GIVEREF(__pyx_int_8);
  PyTuple_SET_ITEM(__pyx_t_543, 0, __pyx_int_8);
  __Pyx_GIVEREF(__pyx_t_545);
  PyTuple_SET_ITEM(__pyx_t_543, 1, __pyx_t_545);
  __Pyx_INCREF(__pyx_int_2005);
  __Pyx_GIVEREF(__pyx_int_2005);
  PyTuple_SET_ITEM(__pyx_t_543, 2, __pyx_int_2005);
  __pyx_t_545 = 0;
  __pyx_t_545 = __Pyx_PyObject_Call(__pyx_t_544, __pyx_t_543, NULL); if (unlikely(!__pyx_t_545)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_545);
  __Pyx_DECREF(__pyx_t_544); __pyx_t_544 = 0;
  __Pyx_DECREF(__pyx_t_543); __pyx_t_543 = 0;

  /* "ultron/tradingday/Calendar.pyx":761
 *     Date.westernStyle(8, Months.May, 2005),
 *     Date.westernStyle(8, Months.October, 2005),
 *     Date.westernStyle(9, Months.October, 2005),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(31, Months.December, 2005),
 *     # 2006
 */
  __pyx_t_543 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_543)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_543);
  __pyx_t_544 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_544)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_544);
  __pyx_t_542 = PyTuple_New(3); if (unlikely(!__pyx_t_542)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_542);
  __Pyx_INCREF(__pyx_int_9);
  __Pyx_GIVEREF(__pyx_int_9);
  PyTuple_SET_ITEM(__pyx_t_542, 0, __pyx_int_9);
  __Pyx_GIVEREF(__pyx_t_544);
  PyTuple_SET_ITEM(__pyx_t_542, 1, __pyx_t_544);
  __Pyx_INCREF(__pyx_int_2005);
  __Pyx_GIVEREF(__pyx_int_2005);
  PyTuple_SET_ITEM(__pyx_t_542, 2, __pyx_int_2005);
  __pyx_t_544 = 0;
  __pyx_t_544 = __Pyx_PyObject_Call(__pyx_t_543, __pyx_t_542, NULL); if (unlikely(!__pyx_t_544)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_544);
  __Pyx_DECREF(__pyx_t_543); __pyx_t_543 = 0;
  __Pyx_DECREF(__pyx_t_542); __pyx_t_542 = 0;

  /* "ultron/tradingday/Calendar.pyx":762
 *     Date.westernStyle(8, Months.October, 2005),
 *     Date.westernStyle(9, Months.October, 2005),
 *     Date.westernStyle(31, Months.December, 2005),             # <<<<<<<<<<<<<<
 *     # 2006
 *     Date.westernStyle(28, Months.January, 2006),
 */
  __pyx_t_542 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_542)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_542);
  __pyx_t_543 = __Pyx_PyInt_From_enum__Months(December); if (unlikely(!__pyx_t_543)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_543);
  __pyx_t_541 = PyTuple_New(3); if (unlikely(!__pyx_t_541)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_541);
  __Pyx_INCREF(__pyx_int_31);
  __Pyx_GIVEREF(__pyx_int_31);
  PyTuple_SET_ITEM(__pyx_t_541, 0, __pyx_int_31);
  __Pyx_GIVEREF(__pyx_t_543);
  PyTuple_SET_ITEM(__pyx_t_541, 1, __pyx_t_543);
  __Pyx_INCREF(__pyx_int_2005);
  __Pyx_GIVEREF(__pyx_int_2005);
  PyTuple_SET_ITEM(__pyx_t_541, 2, __pyx_int_2005);
  __pyx_t_543 = 0;
  __pyx_t_543 = __Pyx_PyObject_Call(__pyx_t_542, __pyx_t_541, NULL); if (unlikely(!__pyx_t_543)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_543);
  __Pyx_DECREF(__pyx_t_542); __pyx_t_542 = 0;
  __Pyx_DECREF(__pyx_t_541); __pyx_t_541 = 0;

  /* "ultron/tradingday/Calendar.pyx":764
 *     Date.westernStyle(31, Months.December, 2005),
 *     # 2006
 *     Date.westernStyle(28, Months.January, 2006),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.April, 2006),
 *     Date.westernStyle(30, Months.April, 2006),
 */
  __pyx_t_541 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_541)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_541);
  __pyx_t_542 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_542)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_542);
  __pyx_t_540 = PyTuple_New(3); if (unlikely(!__pyx_t_540)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_540);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_540, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_542);
  PyTuple_SET_ITEM(__pyx_t_540, 1, __pyx_t_542);
  __Pyx_INCREF(__pyx_int_2006);
  __Pyx_GIVEREF(__pyx_int_2006);
  PyTuple_SET_ITEM(__pyx_t_540, 2, __pyx_int_2006);
  __pyx_t_542 = 0;
  __pyx_t_542 = __Pyx_PyObject_Call(__pyx_t_541, __pyx_t_540, NULL); if (unlikely(!__pyx_t_542)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_542);
  __Pyx_DECREF(__pyx_t_541); __pyx_t_541 = 0;
  __Pyx_DECREF(__pyx_t_540); __pyx_t_540 = 0;

  /* "ultron/tradingday/Calendar.pyx":765
 *     # 2006
 *     Date.westernStyle(28, Months.January, 2006),
 *     Date.westernStyle(29, Months.April, 2006),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(30, Months.April, 2006),
 *     Date.westernStyle(30, Months.September, 2006),
 */
  __pyx_t_540 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_540)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_540);
  __pyx_t_541 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_541)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_541);
  __pyx_t_539 = PyTuple_New(3); if (unlikely(!__pyx_t_539)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_539);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_539, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_541);
  PyTuple_SET_ITEM(__pyx_t_539, 1, __pyx_t_541);
  __Pyx_INCREF(__pyx_int_2006);
  __Pyx_GIVEREF(__pyx_int_2006);
  PyTuple_SET_ITEM(__pyx_t_539, 2, __pyx_int_2006);
  __pyx_t_541 = 0;
  __pyx_t_541 = __Pyx_PyObject_Call(__pyx_t_540, __pyx_t_539, NULL); if (unlikely(!__pyx_t_541)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_541);
  __Pyx_DECREF(__pyx_t_540); __pyx_t_540 = 0;
  __Pyx_DECREF(__pyx_t_539); __pyx_t_539 = 0;

  /* "ultron/tradingday/Calendar.pyx":766
 *     Date.westernStyle(28, Months.January, 2006),
 *     Date.westernStyle(29, Months.April, 2006),
 *     Date.westernStyle(30, Months.April, 2006),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(30, Months.September, 2006),
 *     Date.westernStyle(30, Months.December, 2006),
 */
  __pyx_t_539 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_539)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_539);
  __pyx_t_540 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_540)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_540);
  __pyx_t_538 = PyTuple_New(3); if (unlikely(!__pyx_t_538)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_538);
  __Pyx_INCREF(__pyx_int_30);
  __Pyx_GIVEREF(__pyx_int_30);
  PyTuple_SET_ITEM(__pyx_t_538, 0, __pyx_int_30);
  __Pyx_GIVEREF(__pyx_t_540);
  PyTuple_SET_ITEM(__pyx_t_538, 1, __pyx_t_540);
  __Pyx_INCREF(__pyx_int_2006);
  __Pyx_GIVEREF(__pyx_int_2006);
  PyTuple_SET_ITEM(__pyx_t_538, 2, __pyx_int_2006);
  __pyx_t_540 = 0;
  __pyx_t_540 = __Pyx_PyObject_Call(__pyx_t_539, __pyx_t_538, NULL); if (unlikely(!__pyx_t_540)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_540);
  __Pyx_DECREF(__pyx_t_539); __pyx_t_539 = 0;
  __Pyx_DECREF(__pyx_t_538); __pyx_t_538 = 0;

  /* "ultron/tradingday/Calendar.pyx":767
 *     Date.westernStyle(29, Months.April, 2006),
 *     Date.westernStyle(30, Months.April, 2006),
 *     Date.westernStyle(30, Months.September, 2006),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(30, Months.December, 2006),
 *     Date.westernStyle(31, Months.December, 2006),
 */
  __pyx_t_538 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_538)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_538);
  __pyx_t_539 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_539)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_539);
  __pyx_t_537 = PyTuple_New(3); if (unlikely(!__pyx_t_537)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_537);
  __Pyx_INCREF(__pyx_int_30);
  __Pyx_GIVEREF(__pyx_int_30);
  PyTuple_SET_ITEM(__pyx_t_537, 0, __pyx_int_30);
  __Pyx_GIVEREF(__pyx_t_539);
  PyTuple_SET_ITEM(__pyx_t_537, 1, __pyx_t_539);
  __Pyx_INCREF(__pyx_int_2006);
  __Pyx_GIVEREF(__pyx_int_2006);
  PyTuple_SET_ITEM(__pyx_t_537, 2, __pyx_int_2006);
  __pyx_t_539 = 0;
  __pyx_t_539 = __Pyx_PyObject_Call(__pyx_t_538, __pyx_t_537, NULL); if (unlikely(!__pyx_t_539)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_539);
  __Pyx_DECREF(__pyx_t_538); __pyx_t_538 = 0;
  __Pyx_DECREF(__pyx_t_537); __pyx_t_537 = 0;

  /* "ultron/tradingday/Calendar.pyx":768
 *     Date.westernStyle(30, Months.April, 2006),
 *     Date.westernStyle(30, Months.September, 2006),
 *     Date.westernStyle(30, Months.December, 2006),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(31, Months.December, 2006),
 *     # 2007
 */
  __pyx_t_537 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_537)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_537);
  __pyx_t_538 = __Pyx_PyInt_From_enum__Months(December); if (unlikely(!__pyx_t_538)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_538);
  __pyx_t_536 = PyTuple_New(3); if (unlikely(!__pyx_t_536)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_536);
  __Pyx_INCREF(__pyx_int_30);
  __Pyx_GIVEREF(__pyx_int_30);
  PyTuple_SET_ITEM(__pyx_t_536, 0, __pyx_int_30);
  __Pyx_GIVEREF(__pyx_t_538);
  PyTuple_SET_ITEM(__pyx_t_536, 1, __pyx_t_538);
  __Pyx_INCREF(__pyx_int_2006);
  __Pyx_GIVEREF(__pyx_int_2006);
  PyTuple_SET_ITEM(__pyx_t_536, 2, __pyx_int_2006);
  __pyx_t_538 = 0;
  __pyx_t_538 = __Pyx_PyObject_Call(__pyx_t_537, __pyx_t_536, NULL); if (unlikely(!__pyx_t_538)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_538);
  __Pyx_DECREF(__pyx_t_537); __pyx_t_537 = 0;
  __Pyx_DECREF(__pyx_t_536); __pyx_t_536 = 0;

  /* "ultron/tradingday/Calendar.pyx":769
 *     Date.westernStyle(30, Months.September, 2006),
 *     Date.westernStyle(30, Months.December, 2006),
 *     Date.westernStyle(31, Months.December, 2006),             # <<<<<<<<<<<<<<
 *     # 2007
 *     Date.westernStyle(17, Months.February, 2007),
 */
  __pyx_t_536 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_536)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_536);
  __pyx_t_537 = __Pyx_PyInt_From_enum__Months(December); if (unlikely(!__pyx_t_537)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_537);
  __pyx_t_535 = PyTuple_New(3); if (unlikely(!__pyx_t_535)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_535);
  __Pyx_INCREF(__pyx_int_31);
  __Pyx_GIVEREF(__pyx_int_31);
  PyTuple_SET_ITEM(__pyx_t_535, 0, __pyx_int_31);
  __Pyx_GIVEREF(__pyx_t_537);
  PyTuple_SET_ITEM(__pyx_t_535, 1, __pyx_t_537);
  __Pyx_INCREF(__pyx_int_2006);
  __Pyx_GIVEREF(__pyx_int_2006);
  PyTuple_SET_ITEM(__pyx_t_535, 2, __pyx_int_2006);
  __pyx_t_537 = 0;
  __pyx_t_537 = __Pyx_PyObject_Call(__pyx_t_536, __pyx_t_535, NULL); if (unlikely(!__pyx_t_537)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_537);
  __Pyx_DECREF(__pyx_t_536); __pyx_t_536 = 0;
  __Pyx_DECREF(__pyx_t_535); __pyx_t_535 = 0;

  /* "ultron/tradingday/Calendar.pyx":771
 *     Date.westernStyle(31, Months.December, 2006),
 *     # 2007
 *     Date.westernStyle(17, Months.February, 2007),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(25, Months.February, 2007),
 *     Date.westernStyle(28, Months.April, 2007),
 */
  __pyx_t_535 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_535)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_535);
  __pyx_t_536 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_536)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_536);
  __pyx_t_534 = PyTuple_New(3); if (unlikely(!__pyx_t_534)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_534);
  __Pyx_INCREF(__pyx_int_17);
  __Pyx_GIVEREF(__pyx_int_17);
  PyTuple_SET_ITEM(__pyx_t_534, 0, __pyx_int_17);
  __Pyx_GIVEREF(__pyx_t_536);
  PyTuple_SET_ITEM(__pyx_t_534, 1, __pyx_t_536);
  __Pyx_INCREF(__pyx_int_2007);
  __Pyx_GIVEREF(__pyx_int_2007);
  PyTuple_SET_ITEM(__pyx_t_534, 2, __pyx_int_2007);
  __pyx_t_536 = 0;
  __pyx_t_536 = __Pyx_PyObject_Call(__pyx_t_535, __pyx_t_534, NULL); if (unlikely(!__pyx_t_536)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_536);
  __Pyx_DECREF(__pyx_t_535); __pyx_t_535 = 0;
  __Pyx_DECREF(__pyx_t_534); __pyx_t_534 = 0;

  /* "ultron/tradingday/Calendar.pyx":772
 *     # 2007
 *     Date.westernStyle(17, Months.February, 2007),
 *     Date.westernStyle(25, Months.February, 2007),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(28, Months.April, 2007),
 *     Date.westernStyle(29, Months.April, 2007),
 */
  __pyx_t_534 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_534)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_534);
  __pyx_t_535 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_535)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_535);
  __pyx_t_533 = PyTuple_New(3); if (unlikely(!__pyx_t_533)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_533);
  __Pyx_INCREF(__pyx_int_25);
  __Pyx_GIVEREF(__pyx_int_25);
  PyTuple_SET_ITEM(__pyx_t_533, 0, __pyx_int_25);
  __Pyx_GIVEREF(__pyx_t_535);
  PyTuple_SET_ITEM(__pyx_t_533, 1, __pyx_t_535);
  __Pyx_INCREF(__pyx_int_2007);
  __Pyx_GIVEREF(__pyx_int_2007);
  PyTuple_SET_ITEM(__pyx_t_533, 2, __pyx_int_2007);
  __pyx_t_535 = 0;
  __pyx_t_535 = __Pyx_PyObject_Call(__pyx_t_534, __pyx_t_533, NULL); if (unlikely(!__pyx_t_535)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_535);
  __Pyx_DECREF(__pyx_t_534); __pyx_t_534 = 0;
  __Pyx_DECREF(__pyx_t_533); __pyx_t_533 = 0;

  /* "ultron/tradingday/Calendar.pyx":773
 *     Date.westernStyle(17, Months.February, 2007),
 *     Date.westernStyle(25, Months.February, 2007),
 *     Date.westernStyle(28, Months.April, 2007),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.April, 2007),
 *     Date.westernStyle(29, Months.September, 2007),
 */
  __pyx_t_533 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_533)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_533);
  __pyx_t_534 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_534)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_534);
  __pyx_t_532 = PyTuple_New(3); if (unlikely(!__pyx_t_532)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_532);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_532, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_534);
  PyTuple_SET_ITEM(__pyx_t_532, 1, __pyx_t_534);
  __Pyx_INCREF(__pyx_int_2007);
  __Pyx_GIVEREF(__pyx_int_2007);
  PyTuple_SET_ITEM(__pyx_t_532, 2, __pyx_int_2007);
  __pyx_t_534 = 0;
  __pyx_t_534 = __Pyx_PyObject_Call(__pyx_t_533, __pyx_t_532, NULL); if (unlikely(!__pyx_t_534)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_534);
  __Pyx_DECREF(__pyx_t_533); __pyx_t_533 = 0;
  __Pyx_DECREF(__pyx_t_532); __pyx_t_532 = 0;

  /* "ultron/tradingday/Calendar.pyx":774
 *     Date.westernStyle(25, Months.February, 2007),
 *     Date.westernStyle(28, Months.April, 2007),
 *     Date.westernStyle(29, Months.April, 2007),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.September, 2007),
 *     Date.westernStyle(30, Months.September, 2007),
 */
  __pyx_t_532 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_532)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_532);
  __pyx_t_533 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_533)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_533);
  __pyx_t_531 = PyTuple_New(3); if (unlikely(!__pyx_t_531)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_531);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_531, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_533);
  PyTuple_SET_ITEM(__pyx_t_531, 1, __pyx_t_533);
  __Pyx_INCREF(__pyx_int_2007);
  __Pyx_GIVEREF(__pyx_int_2007);
  PyTuple_SET_ITEM(__pyx_t_531, 2, __pyx_int_2007);
  __pyx_t_533 = 0;
  __pyx_t_533 = __Pyx_PyObject_Call(__pyx_t_532, __pyx_t_531, NULL); if (unlikely(!__pyx_t_533)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_533);
  __Pyx_DECREF(__pyx_t_532); __pyx_t_532 = 0;
  __Pyx_DECREF(__pyx_t_531); __pyx_t_531 = 0;

  /* "ultron/tradingday/Calendar.pyx":775
 *     Date.westernStyle(28, Months.April, 2007),
 *     Date.westernStyle(29, Months.April, 2007),
 *     Date.westernStyle(29, Months.September, 2007),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(30, Months.September, 2007),
 *     Date.westernStyle(29, Months.December, 2007),
 */
  __pyx_t_531 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_531)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_531);
  __pyx_t_532 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_532)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_532);
  __pyx_t_530 = PyTuple_New(3); if (unlikely(!__pyx_t_530)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_530);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_530, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_532);
  PyTuple_SET_ITEM(__pyx_t_530, 1, __pyx_t_532);
  __Pyx_INCREF(__pyx_int_2007);
  __Pyx_GIVEREF(__pyx_int_2007);
  PyTuple_SET_ITEM(__pyx_t_530, 2, __pyx_int_2007);
  __pyx_t_532 = 0;
  __pyx_t_532 = __Pyx_PyObject_Call(__pyx_t_531, __pyx_t_530, NULL); if (unlikely(!__pyx_t_532)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_532);
  __Pyx_DECREF(__pyx_t_531); __pyx_t_531 = 0;
  __Pyx_DECREF(__pyx_t_530); __pyx_t_530 = 0;

  /* "ultron/tradingday/Calendar.pyx":776
 *     Date.westernStyle(29, Months.April, 2007),
 *     Date.westernStyle(29, Months.September, 2007),
 *     Date.westernStyle(30, Months.September, 2007),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.December, 2007),
 *     # 2008
 */
  __pyx_t_530 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_530)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_530);
  __pyx_t_531 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_531)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_531);
  __pyx_t_529 = PyTuple_New(3); if (unlikely(!__pyx_t_529)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_529);
  __Pyx_INCREF(__pyx_int_30);
  __Pyx_GIVEREF(__pyx_int_30);
  PyTuple_SET_ITEM(__pyx_t_529, 0, __pyx_int_30);
  __Pyx_GIVEREF(__pyx_t_531);
  PyTuple_SET_ITEM(__pyx_t_529, 1, __pyx_t_531);
  __Pyx_INCREF(__pyx_int_2007);
  __Pyx_GIVEREF(__pyx_int_2007);
  PyTuple_SET_ITEM(__pyx_t_529, 2, __pyx_int_2007);
  __pyx_t_531 = 0;
  __pyx_t_531 = __Pyx_PyObject_Call(__pyx_t_530, __pyx_t_529, NULL); if (unlikely(!__pyx_t_531)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_531);
  __Pyx_DECREF(__pyx_t_530); __pyx_t_530 = 0;
  __Pyx_DECREF(__pyx_t_529); __pyx_t_529 = 0;

  /* "ultron/tradingday/Calendar.pyx":777
 *     Date.westernStyle(29, Months.September, 2007),
 *     Date.westernStyle(30, Months.September, 2007),
 *     Date.westernStyle(29, Months.December, 2007),             # <<<<<<<<<<<<<<
 *     # 2008
 *     Date.westernStyle(2, Months.February, 2008),
 */
  __pyx_t_529 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_529)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_529);
  __pyx_t_530 = __Pyx_PyInt_From_enum__Months(December); if (unlikely(!__pyx_t_530)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_530);
  __pyx_t_528 = PyTuple_New(3); if (unlikely(!__pyx_t_528)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_528);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_528, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_530);
  PyTuple_SET_ITEM(__pyx_t_528, 1, __pyx_t_530);
  __Pyx_INCREF(__pyx_int_2007);
  __Pyx_GIVEREF(__pyx_int_2007);
  PyTuple_SET_ITEM(__pyx_t_528, 2, __pyx_int_2007);
  __pyx_t_530 = 0;
  __pyx_t_530 = __Pyx_PyObject_Call(__pyx_t_529, __pyx_t_528, NULL); if (unlikely(!__pyx_t_530)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_530);
  __Pyx_DECREF(__pyx_t_529); __pyx_t_529 = 0;
  __Pyx_DECREF(__pyx_t_528); __pyx_t_528 = 0;

  /* "ultron/tradingday/Calendar.pyx":779
 *     Date.westernStyle(29, Months.December, 2007),
 *     # 2008
 *     Date.westernStyle(2, Months.February, 2008),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(3, Months.February, 2008),
 *     Date.westernStyle(4, Months.May, 2008),
 */
  __pyx_t_528 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_528)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_528);
  __pyx_t_529 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_529)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_529);
  __pyx_t_527 = PyTuple_New(3); if (unlikely(!__pyx_t_527)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_527);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_527, 0, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_t_529);
  PyTuple_SET_ITEM(__pyx_t_527, 1, __pyx_t_529);
  __Pyx_INCREF(__pyx_int_2008);
  __Pyx_GIVEREF(__pyx_int_2008);
  PyTuple_SET_ITEM(__pyx_t_527, 2, __pyx_int_2008);
  __pyx_t_529 = 0;
  __pyx_t_529 = __Pyx_PyObject_Call(__pyx_t_528, __pyx_t_527, NULL); if (unlikely(!__pyx_t_529)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_529);
  __Pyx_DECREF(__pyx_t_528); __pyx_t_528 = 0;
  __Pyx_DECREF(__pyx_t_527); __pyx_t_527 = 0;

  /* "ultron/tradingday/Calendar.pyx":780
 *     # 2008
 *     Date.westernStyle(2, Months.February, 2008),
 *     Date.westernStyle(3, Months.February, 2008),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(4, Months.May, 2008),
 *     Date.westernStyle(27, Months.September, 2008),
 */
  __pyx_t_527 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_527)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_527);
  __pyx_t_528 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_528)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_528);
  __pyx_t_526 = PyTuple_New(3); if (unlikely(!__pyx_t_526)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_526);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_526, 0, __pyx_int_3);
  __Pyx_GIVEREF(__pyx_t_528);
  PyTuple_SET_ITEM(__pyx_t_526, 1, __pyx_t_528);
  __Pyx_INCREF(__pyx_int_2008);
  __Pyx_GIVEREF(__pyx_int_2008);
  PyTuple_SET_ITEM(__pyx_t_526, 2, __pyx_int_2008);
  __pyx_t_528 = 0;
  __pyx_t_528 = __Pyx_PyObject_Call(__pyx_t_527, __pyx_t_526, NULL); if (unlikely(!__pyx_t_528)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_528);
  __Pyx_DECREF(__pyx_t_527); __pyx_t_527 = 0;
  __Pyx_DECREF(__pyx_t_526); __pyx_t_526 = 0;

  /* "ultron/tradingday/Calendar.pyx":781
 *     Date.westernStyle(2, Months.February, 2008),
 *     Date.westernStyle(3, Months.February, 2008),
 *     Date.westernStyle(4, Months.May, 2008),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(27, Months.September, 2008),
 *     Date.westernStyle(28, Months.September, 2008),
 */
  __pyx_t_526 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_526)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_526);
  __pyx_t_527 = __Pyx_PyInt_From_enum__Months(May); if (unlikely(!__pyx_t_527)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_527);
  __pyx_t_525 = PyTuple_New(3); if (unlikely(!__pyx_t_525)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_525);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_525, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_t_527);
  PyTuple_SET_ITEM(__pyx_t_525, 1, __pyx_t_527);
  __Pyx_INCREF(__pyx_int_2008);
  __Pyx_GIVEREF(__pyx_int_2008);
  PyTuple_SET_ITEM(__pyx_t_525, 2, __pyx_int_2008);
  __pyx_t_527 = 0;
  __pyx_t_527 = __Pyx_PyObject_Call(__pyx_t_526, __pyx_t_525, NULL); if (unlikely(!__pyx_t_527)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_527);
  __Pyx_DECREF(__pyx_t_526); __pyx_t_526 = 0;
  __Pyx_DECREF(__pyx_t_525); __pyx_t_525 = 0;

  /* "ultron/tradingday/Calendar.pyx":782
 *     Date.westernStyle(3, Months.February, 2008),
 *     Date.westernStyle(4, Months.May, 2008),
 *     Date.westernStyle(27, Months.September, 2008),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(28, Months.September, 2008),
 *     # 2009
 */
  __pyx_t_525 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_525)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_525);
  __pyx_t_526 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_526)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_526);
  __pyx_t_524 = PyTuple_New(3); if (unlikely(!__pyx_t_524)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_524);
  __Pyx_INCREF(__pyx_int_27);
  __Pyx_GIVEREF(__pyx_int_27);
  PyTuple_SET_ITEM(__pyx_t_524, 0, __pyx_int_27);
  __Pyx_GIVEREF(__pyx_t_526);
  PyTuple_SET_ITEM(__pyx_t_524, 1, __pyx_t_526);
  __Pyx_INCREF(__pyx_int_2008);
  __Pyx_GIVEREF(__pyx_int_2008);
  PyTuple_SET_ITEM(__pyx_t_524, 2, __pyx_int_2008);
  __pyx_t_526 = 0;
  __pyx_t_526 = __Pyx_PyObject_Call(__pyx_t_525, __pyx_t_524, NULL); if (unlikely(!__pyx_t_526)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_526);
  __Pyx_DECREF(__pyx_t_525); __pyx_t_525 = 0;
  __Pyx_DECREF(__pyx_t_524); __pyx_t_524 = 0;

  /* "ultron/tradingday/Calendar.pyx":783
 *     Date.westernStyle(4, Months.May, 2008),
 *     Date.westernStyle(27, Months.September, 2008),
 *     Date.westernStyle(28, Months.September, 2008),             # <<<<<<<<<<<<<<
 *     # 2009
 *     Date.westernStyle(4, Months.January, 2009),
 */
  __pyx_t_524 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_524)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_524);
  __pyx_t_525 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_525)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_525);
  __pyx_t_523 = PyTuple_New(3); if (unlikely(!__pyx_t_523)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_523);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_523, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_525);
  PyTuple_SET_ITEM(__pyx_t_523, 1, __pyx_t_525);
  __Pyx_INCREF(__pyx_int_2008);
  __Pyx_GIVEREF(__pyx_int_2008);
  PyTuple_SET_ITEM(__pyx_t_523, 2, __pyx_int_2008);
  __pyx_t_525 = 0;
  __pyx_t_525 = __Pyx_PyObject_Call(__pyx_t_524, __pyx_t_523, NULL); if (unlikely(!__pyx_t_525)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_525);
  __Pyx_DECREF(__pyx_t_524); __pyx_t_524 = 0;
  __Pyx_DECREF(__pyx_t_523); __pyx_t_523 = 0;

  /* "ultron/tradingday/Calendar.pyx":785
 *     Date.westernStyle(28, Months.September, 2008),
 *     # 2009
 *     Date.westernStyle(4, Months.January, 2009),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(24, Months.January, 2009),
 *     Date.westernStyle(1, Months.February, 2009),
 */
  __pyx_t_523 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_523)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_523);
  __pyx_t_524 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_524)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_524);
  __pyx_t_522 = PyTuple_New(3); if (unlikely(!__pyx_t_522)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_522);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_522, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_t_524);
  PyTuple_SET_ITEM(__pyx_t_522, 1, __pyx_t_524);
  __Pyx_INCREF(__pyx_int_2009);
  __Pyx_GIVEREF(__pyx_int_2009);
  PyTuple_SET_ITEM(__pyx_t_522, 2, __pyx_int_2009);
  __pyx_t_524 = 0;
  __pyx_t_524 = __Pyx_PyObject_Call(__pyx_t_523, __pyx_t_522, NULL); if (unlikely(!__pyx_t_524)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_524);
  __Pyx_DECREF(__pyx_t_523); __pyx_t_523 = 0;
  __Pyx_DECREF(__pyx_t_522); __pyx_t_522 = 0;

  /* "ultron/tradingday/Calendar.pyx":786
 *     # 2009
 *     Date.westernStyle(4, Months.January, 2009),
 *     Date.westernStyle(24, Months.January, 2009),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(1, Months.February, 2009),
 *     Date.westernStyle(31, Months.May, 2009),
 */
  __pyx_t_522 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_522)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_522);
  __pyx_t_523 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_523)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_523);
  __pyx_t_521 = PyTuple_New(3); if (unlikely(!__pyx_t_521)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_521);
  __Pyx_INCREF(__pyx_int_24);
  __Pyx_GIVEREF(__pyx_int_24);
  PyTuple_SET_ITEM(__pyx_t_521, 0, __pyx_int_24);
  __Pyx_GIVEREF(__pyx_t_523);
  PyTuple_SET_ITEM(__pyx_t_521, 1, __pyx_t_523);
  __Pyx_INCREF(__pyx_int_2009);
  __Pyx_GIVEREF(__pyx_int_2009);
  PyTuple_SET_ITEM(__pyx_t_521, 2, __pyx_int_2009);
  __pyx_t_523 = 0;
  __pyx_t_523 = __Pyx_PyObject_Call(__pyx_t_522, __pyx_t_521, NULL); if (unlikely(!__pyx_t_523)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_523);
  __Pyx_DECREF(__pyx_t_522); __pyx_t_522 = 0;
  __Pyx_DECREF(__pyx_t_521); __pyx_t_521 = 0;

  /* "ultron/tradingday/Calendar.pyx":787
 *     Date.westernStyle(4, Months.January, 2009),
 *     Date.westernStyle(24, Months.January, 2009),
 *     Date.westernStyle(1, Months.February, 2009),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(31, Months.May, 2009),
 *     Date.westernStyle(27, Months.September, 2009),
 */
  __pyx_t_521 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_521)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_521);
  __pyx_t_522 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_522)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_522);
  __pyx_t_520 = PyTuple_New(3); if (unlikely(!__pyx_t_520)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_520);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_t_520, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_t_522);
  PyTuple_SET_ITEM(__pyx_t_520, 1, __pyx_t_522);
  __Pyx_INCREF(__pyx_int_2009);
  __Pyx_GIVEREF(__pyx_int_2009);
  PyTuple_SET_ITEM(__pyx_t_520, 2, __pyx_int_2009);
  __pyx_t_522 = 0;
  __pyx_t_522 = __Pyx_PyObject_Call(__pyx_t_521, __pyx_t_520, NULL); if (unlikely(!__pyx_t_522)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_522);
  __Pyx_DECREF(__pyx_t_521); __pyx_t_521 = 0;
  __Pyx_DECREF(__pyx_t_520); __pyx_t_520 = 0;

  /* "ultron/tradingday/Calendar.pyx":788
 *     Date.westernStyle(24, Months.January, 2009),
 *     Date.westernStyle(1, Months.February, 2009),
 *     Date.westernStyle(31, Months.May, 2009),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(27, Months.September, 2009),
 *     Date.westernStyle(10, Months.October, 2009),
 */
  __pyx_t_520 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_520)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_520);
  __pyx_t_521 = __Pyx_PyInt_From_enum__Months(May); if (unlikely(!__pyx_t_521)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_521);
  __pyx_t_519 = PyTuple_New(3); if (unlikely(!__pyx_t_519)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_519);
  __Pyx_INCREF(__pyx_int_31);
  __Pyx_GIVEREF(__pyx_int_31);
  PyTuple_SET_ITEM(__pyx_t_519, 0, __pyx_int_31);
  __Pyx_GIVEREF(__pyx_t_521);
  PyTuple_SET_ITEM(__pyx_t_519, 1, __pyx_t_521);
  __Pyx_INCREF(__pyx_int_2009);
  __Pyx_GIVEREF(__pyx_int_2009);
  PyTuple_SET_ITEM(__pyx_t_519, 2, __pyx_int_2009);
  __pyx_t_521 = 0;
  __pyx_t_521 = __Pyx_PyObject_Call(__pyx_t_520, __pyx_t_519, NULL); if (unlikely(!__pyx_t_521)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_521);
  __Pyx_DECREF(__pyx_t_520); __pyx_t_520 = 0;
  __Pyx_DECREF(__pyx_t_519); __pyx_t_519 = 0;

  /* "ultron/tradingday/Calendar.pyx":789
 *     Date.westernStyle(1, Months.February, 2009),
 *     Date.westernStyle(31, Months.May, 2009),
 *     Date.westernStyle(27, Months.September, 2009),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(10, Months.October, 2009),
 *     # 2010
 */
  __pyx_t_519 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_519)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_519);
  __pyx_t_520 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_520)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_520);
  __pyx_t_518 = PyTuple_New(3); if (unlikely(!__pyx_t_518)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_518);
  __Pyx_INCREF(__pyx_int_27);
  __Pyx_GIVEREF(__pyx_int_27);
  PyTuple_SET_ITEM(__pyx_t_518, 0, __pyx_int_27);
  __Pyx_GIVEREF(__pyx_t_520);
  PyTuple_SET_ITEM(__pyx_t_518, 1, __pyx_t_520);
  __Pyx_INCREF(__pyx_int_2009);
  __Pyx_GIVEREF(__pyx_int_2009);
  PyTuple_SET_ITEM(__pyx_t_518, 2, __pyx_int_2009);
  __pyx_t_520 = 0;
  __pyx_t_520 = __Pyx_PyObject_Call(__pyx_t_519, __pyx_t_518, NULL); if (unlikely(!__pyx_t_520)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_520);
  __Pyx_DECREF(__pyx_t_519); __pyx_t_519 = 0;
  __Pyx_DECREF(__pyx_t_518); __pyx_t_518 = 0;

  /* "ultron/tradingday/Calendar.pyx":790
 *     Date.westernStyle(31, Months.May, 2009),
 *     Date.westernStyle(27, Months.September, 2009),
 *     Date.westernStyle(10, Months.October, 2009),             # <<<<<<<<<<<<<<
 *     # 2010
 *     Date.westernStyle(20, Months.February, 2010),
 */
  __pyx_t_518 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_518)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_518);
  __pyx_t_519 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_519)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_519);
  __pyx_t_517 = PyTuple_New(3); if (unlikely(!__pyx_t_517)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_517);
  __Pyx_INCREF(__pyx_int_10);
  __Pyx_GIVEREF(__pyx_int_10);
  PyTuple_SET_ITEM(__pyx_t_517, 0, __pyx_int_10);
  __Pyx_GIVEREF(__pyx_t_519);
  PyTuple_SET_ITEM(__pyx_t_517, 1, __pyx_t_519);
  __Pyx_INCREF(__pyx_int_2009);
  __Pyx_GIVEREF(__pyx_int_2009);
  PyTuple_SET_ITEM(__pyx_t_517, 2, __pyx_int_2009);
  __pyx_t_519 = 0;
  __pyx_t_519 = __Pyx_PyObject_Call(__pyx_t_518, __pyx_t_517, NULL); if (unlikely(!__pyx_t_519)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_519);
  __Pyx_DECREF(__pyx_t_518); __pyx_t_518 = 0;
  __Pyx_DECREF(__pyx_t_517); __pyx_t_517 = 0;

  /* "ultron/tradingday/Calendar.pyx":792
 *     Date.westernStyle(10, Months.October, 2009),
 *     # 2010
 *     Date.westernStyle(20, Months.February, 2010),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(21, Months.February, 2010),
 *     Date.westernStyle(12, Months.June, 2010),
 */
  __pyx_t_517 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_517)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_517);
  __pyx_t_518 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_518)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_518);
  __pyx_t_516 = PyTuple_New(3); if (unlikely(!__pyx_t_516)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_516);
  __Pyx_INCREF(__pyx_int_20);
  __Pyx_GIVEREF(__pyx_int_20);
  PyTuple_SET_ITEM(__pyx_t_516, 0, __pyx_int_20);
  __Pyx_GIVEREF(__pyx_t_518);
  PyTuple_SET_ITEM(__pyx_t_516, 1, __pyx_t_518);
  __Pyx_INCREF(__pyx_int_2010);
  __Pyx_GIVEREF(__pyx_int_2010);
  PyTuple_SET_ITEM(__pyx_t_516, 2, __pyx_int_2010);
  __pyx_t_518 = 0;
  __pyx_t_518 = __Pyx_PyObject_Call(__pyx_t_517, __pyx_t_516, NULL); if (unlikely(!__pyx_t_518)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_518);
  __Pyx_DECREF(__pyx_t_517); __pyx_t_517 = 0;
  __Pyx_DECREF(__pyx_t_516); __pyx_t_516 = 0;

  /* "ultron/tradingday/Calendar.pyx":793
 *     # 2010
 *     Date.westernStyle(20, Months.February, 2010),
 *     Date.westernStyle(21, Months.February, 2010),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(12, Months.June, 2010),
 *     Date.westernStyle(13, Months.June, 2010),
 */
  __pyx_t_516 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_516)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_516);
  __pyx_t_517 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_517)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_517);
  __pyx_t_515 = PyTuple_New(3); if (unlikely(!__pyx_t_515)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_515);
  __Pyx_INCREF(__pyx_int_21);
  __Pyx_GIVEREF(__pyx_int_21);
  PyTuple_SET_ITEM(__pyx_t_515, 0, __pyx_int_21);
  __Pyx_GIVEREF(__pyx_t_517);
  PyTuple_SET_ITEM(__pyx_t_515, 1, __pyx_t_517);
  __Pyx_INCREF(__pyx_int_2010);
  __Pyx_GIVEREF(__pyx_int_2010);
  PyTuple_SET_ITEM(__pyx_t_515, 2, __pyx_int_2010);
  __pyx_t_517 = 0;
  __pyx_t_517 = __Pyx_PyObject_Call(__pyx_t_516, __pyx_t_515, NULL); if (unlikely(!__pyx_t_517)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_517);
  __Pyx_DECREF(__pyx_t_516); __pyx_t_516 = 0;
  __Pyx_DECREF(__pyx_t_515); __pyx_t_515 = 0;

  /* "ultron/tradingday/Calendar.pyx":794
 *     Date.westernStyle(20, Months.February, 2010),
 *     Date.westernStyle(21, Months.February, 2010),
 *     Date.westernStyle(12, Months.June, 2010),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(13, Months.June, 2010),
 *     Date.westernStyle(19, Months.September, 2010),
 */
  __pyx_t_515 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_515)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_515);
  __pyx_t_516 = __Pyx_PyInt_From_enum__Months(June); if (unlikely(!__pyx_t_516)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_516);
  __pyx_t_514 = PyTuple_New(3); if (unlikely(!__pyx_t_514)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_514);
  __Pyx_INCREF(__pyx_int_12);
  __Pyx_GIVEREF(__pyx_int_12);
  PyTuple_SET_ITEM(__pyx_t_514, 0, __pyx_int_12);
  __Pyx_GIVEREF(__pyx_t_516);
  PyTuple_SET_ITEM(__pyx_t_514, 1, __pyx_t_516);
  __Pyx_INCREF(__pyx_int_2010);
  __Pyx_GIVEREF(__pyx_int_2010);
  PyTuple_SET_ITEM(__pyx_t_514, 2, __pyx_int_2010);
  __pyx_t_516 = 0;
  __pyx_t_516 = __Pyx_PyObject_Call(__pyx_t_515, __pyx_t_514, NULL); if (unlikely(!__pyx_t_516)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_516);
  __Pyx_DECREF(__pyx_t_515); __pyx_t_515 = 0;
  __Pyx_DECREF(__pyx_t_514); __pyx_t_514 = 0;

  /* "ultron/tradingday/Calendar.pyx":795
 *     Date.westernStyle(21, Months.February, 2010),
 *     Date.westernStyle(12, Months.June, 2010),
 *     Date.westernStyle(13, Months.June, 2010),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(19, Months.September, 2010),
 *     Date.westernStyle(25, Months.September, 2010),
 */
  __pyx_t_514 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_514)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_514);
  __pyx_t_515 = __Pyx_PyInt_From_enum__Months(June); if (unlikely(!__pyx_t_515)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_515);
  __pyx_t_513 = PyTuple_New(3); if (unlikely(!__pyx_t_513)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_513);
  __Pyx_INCREF(__pyx_int_13);
  __Pyx_GIVEREF(__pyx_int_13);
  PyTuple_SET_ITEM(__pyx_t_513, 0, __pyx_int_13);
  __Pyx_GIVEREF(__pyx_t_515);
  PyTuple_SET_ITEM(__pyx_t_513, 1, __pyx_t_515);
  __Pyx_INCREF(__pyx_int_2010);
  __Pyx_GIVEREF(__pyx_int_2010);
  PyTuple_SET_ITEM(__pyx_t_513, 2, __pyx_int_2010);
  __pyx_t_515 = 0;
  __pyx_t_515 = __Pyx_PyObject_Call(__pyx_t_514, __pyx_t_513, NULL); if (unlikely(!__pyx_t_515)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_515);
  __Pyx_DECREF(__pyx_t_514); __pyx_t_514 = 0;
  __Pyx_DECREF(__pyx_t_513); __pyx_t_513 = 0;

  /* "ultron/tradingday/Calendar.pyx":796
 *     Date.westernStyle(12, Months.June, 2010),
 *     Date.westernStyle(13, Months.June, 2010),
 *     Date.westernStyle(19, Months.September, 2010),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(25, Months.September, 2010),
 *     Date.westernStyle(26, Months.September, 2010),
 */
  __pyx_t_513 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_513)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_513);
  __pyx_t_514 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_514)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_514);
  __pyx_t_512 = PyTuple_New(3); if (unlikely(!__pyx_t_512)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_512);
  __Pyx_INCREF(__pyx_int_19);
  __Pyx_GIVEREF(__pyx_int_19);
  PyTuple_SET_ITEM(__pyx_t_512, 0, __pyx_int_19);
  __Pyx_GIVEREF(__pyx_t_514);
  PyTuple_SET_ITEM(__pyx_t_512, 1, __pyx_t_514);
  __Pyx_INCREF(__pyx_int_2010);
  __Pyx_GIVEREF(__pyx_int_2010);
  PyTuple_SET_ITEM(__pyx_t_512, 2, __pyx_int_2010);
  __pyx_t_514 = 0;
  __pyx_t_514 = __Pyx_PyObject_Call(__pyx_t_513, __pyx_t_512, NULL); if (unlikely(!__pyx_t_514)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_514);
  __Pyx_DECREF(__pyx_t_513); __pyx_t_513 = 0;
  __Pyx_DECREF(__pyx_t_512); __pyx_t_512 = 0;

  /* "ultron/tradingday/Calendar.pyx":797
 *     Date.westernStyle(13, Months.June, 2010),
 *     Date.westernStyle(19, Months.September, 2010),
 *     Date.westernStyle(25, Months.September, 2010),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(26, Months.September, 2010),
 *     Date.westernStyle(9, Months.October, 2010),
 */
  __pyx_t_512 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_512)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_512);
  __pyx_t_513 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_513)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_513);
  __pyx_t_511 = PyTuple_New(3); if (unlikely(!__pyx_t_511)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_511);
  __Pyx_INCREF(__pyx_int_25);
  __Pyx_GIVEREF(__pyx_int_25);
  PyTuple_SET_ITEM(__pyx_t_511, 0, __pyx_int_25);
  __Pyx_GIVEREF(__pyx_t_513);
  PyTuple_SET_ITEM(__pyx_t_511, 1, __pyx_t_513);
  __Pyx_INCREF(__pyx_int_2010);
  __Pyx_GIVEREF(__pyx_int_2010);
  PyTuple_SET_ITEM(__pyx_t_511, 2, __pyx_int_2010);
  __pyx_t_513 = 0;
  __pyx_t_513 = __Pyx_PyObject_Call(__pyx_t_512, __pyx_t_511, NULL); if (unlikely(!__pyx_t_513)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_513);
  __Pyx_DECREF(__pyx_t_512); __pyx_t_512 = 0;
  __Pyx_DECREF(__pyx_t_511); __pyx_t_511 = 0;

  /* "ultron/tradingday/Calendar.pyx":798
 *     Date.westernStyle(19, Months.September, 2010),
 *     Date.westernStyle(25, Months.September, 2010),
 *     Date.westernStyle(26, Months.September, 2010),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(9, Months.October, 2010),
 *     # 2011
 */
  __pyx_t_511 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_511)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_511);
  __pyx_t_512 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_512)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_512);
  __pyx_t_510 = PyTuple_New(3); if (unlikely(!__pyx_t_510)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_510);
  __Pyx_INCREF(__pyx_int_26);
  __Pyx_GIVEREF(__pyx_int_26);
  PyTuple_SET_ITEM(__pyx_t_510, 0, __pyx_int_26);
  __Pyx_GIVEREF(__pyx_t_512);
  PyTuple_SET_ITEM(__pyx_t_510, 1, __pyx_t_512);
  __Pyx_INCREF(__pyx_int_2010);
  __Pyx_GIVEREF(__pyx_int_2010);
  PyTuple_SET_ITEM(__pyx_t_510, 2, __pyx_int_2010);
  __pyx_t_512 = 0;
  __pyx_t_512 = __Pyx_PyObject_Call(__pyx_t_511, __pyx_t_510, NULL); if (unlikely(!__pyx_t_512)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_512);
  __Pyx_DECREF(__pyx_t_511); __pyx_t_511 = 0;
  __Pyx_DECREF(__pyx_t_510); __pyx_t_510 = 0;

  /* "ultron/tradingday/Calendar.pyx":799
 *     Date.westernStyle(25, Months.September, 2010),
 *     Date.westernStyle(26, Months.September, 2010),
 *     Date.westernStyle(9, Months.October, 2010),             # <<<<<<<<<<<<<<
 *     # 2011
 *     Date.westernStyle(30, Months.January, 2011),
 */
  __pyx_t_510 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_510)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_510);
  __pyx_t_511 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_511)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_511);
  __pyx_t_509 = PyTuple_New(3); if (unlikely(!__pyx_t_509)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_509);
  __Pyx_INCREF(__pyx_int_9);
  __Pyx_GIVEREF(__pyx_int_9);
  PyTuple_SET_ITEM(__pyx_t_509, 0, __pyx_int_9);
  __Pyx_GIVEREF(__pyx_t_511);
  PyTuple_SET_ITEM(__pyx_t_509, 1, __pyx_t_511);
  __Pyx_INCREF(__pyx_int_2010);
  __Pyx_GIVEREF(__pyx_int_2010);
  PyTuple_SET_ITEM(__pyx_t_509, 2, __pyx_int_2010);
  __pyx_t_511 = 0;
  __pyx_t_511 = __Pyx_PyObject_Call(__pyx_t_510, __pyx_t_509, NULL); if (unlikely(!__pyx_t_511)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_511);
  __Pyx_DECREF(__pyx_t_510); __pyx_t_510 = 0;
  __Pyx_DECREF(__pyx_t_509); __pyx_t_509 = 0;

  /* "ultron/tradingday/Calendar.pyx":801
 *     Date.westernStyle(9, Months.October, 2010),
 *     # 2011
 *     Date.westernStyle(30, Months.January, 2011),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(12, Months.February, 2011),
 *     Date.westernStyle(2, Months.April, 2011),
 */
  __pyx_t_509 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_509)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_509);
  __pyx_t_510 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_510)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_510);
  __pyx_t_508 = PyTuple_New(3); if (unlikely(!__pyx_t_508)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_508);
  __Pyx_INCREF(__pyx_int_30);
  __Pyx_GIVEREF(__pyx_int_30);
  PyTuple_SET_ITEM(__pyx_t_508, 0, __pyx_int_30);
  __Pyx_GIVEREF(__pyx_t_510);
  PyTuple_SET_ITEM(__pyx_t_508, 1, __pyx_t_510);
  __Pyx_INCREF(__pyx_int_2011);
  __Pyx_GIVEREF(__pyx_int_2011);
  PyTuple_SET_ITEM(__pyx_t_508, 2, __pyx_int_2011);
  __pyx_t_510 = 0;
  __pyx_t_510 = __Pyx_PyObject_Call(__pyx_t_509, __pyx_t_508, NULL); if (unlikely(!__pyx_t_510)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_510);
  __Pyx_DECREF(__pyx_t_509); __pyx_t_509 = 0;
  __Pyx_DECREF(__pyx_t_508); __pyx_t_508 = 0;

  /* "ultron/tradingday/Calendar.pyx":802
 *     # 2011
 *     Date.westernStyle(30, Months.January, 2011),
 *     Date.westernStyle(12, Months.February, 2011),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(2, Months.April, 2011),
 *     Date.westernStyle(8, Months.October, 2011),
 */
  __pyx_t_508 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_508)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_508);
  __pyx_t_509 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_509)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_509);
  __pyx_t_507 = PyTuple_New(3); if (unlikely(!__pyx_t_507)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_507);
  __Pyx_INCREF(__pyx_int_12);
  __Pyx_GIVEREF(__pyx_int_12);
  PyTuple_SET_ITEM(__pyx_t_507, 0, __pyx_int_12);
  __Pyx_GIVEREF(__pyx_t_509);
  PyTuple_SET_ITEM(__pyx_t_507, 1, __pyx_t_509);
  __Pyx_INCREF(__pyx_int_2011);
  __Pyx_GIVEREF(__pyx_int_2011);
  PyTuple_SET_ITEM(__pyx_t_507, 2, __pyx_int_2011);
  __pyx_t_509 = 0;
  __pyx_t_509 = __Pyx_PyObject_Call(__pyx_t_508, __pyx_t_507, NULL); if (unlikely(!__pyx_t_509)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_509);
  __Pyx_DECREF(__pyx_t_508); __pyx_t_508 = 0;
  __Pyx_DECREF(__pyx_t_507); __pyx_t_507 = 0;

  /* "ultron/tradingday/Calendar.pyx":803
 *     Date.westernStyle(30, Months.January, 2011),
 *     Date.westernStyle(12, Months.February, 2011),
 *     Date.westernStyle(2, Months.April, 2011),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(8, Months.October, 2011),
 *     Date.westernStyle(9, Months.October, 2011),
 */
  __pyx_t_507 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_507)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_507);
  __pyx_t_508 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_508)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_508);
  __pyx_t_506 = PyTuple_New(3); if (unlikely(!__pyx_t_506)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_506);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_506, 0, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_t_508);
  PyTuple_SET_ITEM(__pyx_t_506, 1, __pyx_t_508);
  __Pyx_INCREF(__pyx_int_2011);
  __Pyx_GIVEREF(__pyx_int_2011);
  PyTuple_SET_ITEM(__pyx_t_506, 2, __pyx_int_2011);
  __pyx_t_508 = 0;
  __pyx_t_508 = __Pyx_PyObject_Call(__pyx_t_507, __pyx_t_506, NULL); if (unlikely(!__pyx_t_508)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_508);
  __Pyx_DECREF(__pyx_t_507); __pyx_t_507 = 0;
  __Pyx_DECREF(__pyx_t_506); __pyx_t_506 = 0;

  /* "ultron/tradingday/Calendar.pyx":804
 *     Date.westernStyle(12, Months.February, 2011),
 *     Date.westernStyle(2, Months.April, 2011),
 *     Date.westernStyle(8, Months.October, 2011),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(9, Months.October, 2011),
 *     Date.westernStyle(31, Months.December, 2011),
 */
  __pyx_t_506 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_506)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_506);
  __pyx_t_507 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_507)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_507);
  __pyx_t_505 = PyTuple_New(3); if (unlikely(!__pyx_t_505)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_505);
  __Pyx_INCREF(__pyx_int_8);
  __Pyx_GIVEREF(__pyx_int_8);
  PyTuple_SET_ITEM(__pyx_t_505, 0, __pyx_int_8);
  __Pyx_GIVEREF(__pyx_t_507);
  PyTuple_SET_ITEM(__pyx_t_505, 1, __pyx_t_507);
  __Pyx_INCREF(__pyx_int_2011);
  __Pyx_GIVEREF(__pyx_int_2011);
  PyTuple_SET_ITEM(__pyx_t_505, 2, __pyx_int_2011);
  __pyx_t_507 = 0;
  __pyx_t_507 = __Pyx_PyObject_Call(__pyx_t_506, __pyx_t_505, NULL); if (unlikely(!__pyx_t_507)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_507);
  __Pyx_DECREF(__pyx_t_506); __pyx_t_506 = 0;
  __Pyx_DECREF(__pyx_t_505); __pyx_t_505 = 0;

  /* "ultron/tradingday/Calendar.pyx":805
 *     Date.westernStyle(2, Months.April, 2011),
 *     Date.westernStyle(8, Months.October, 2011),
 *     Date.westernStyle(9, Months.October, 2011),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(31, Months.December, 2011),
 *     # 2012
 */
  __pyx_t_505 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_505)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_505);
  __pyx_t_506 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_506)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_506);
  __pyx_t_504 = PyTuple_New(3); if (unlikely(!__pyx_t_504)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_504);
  __Pyx_INCREF(__pyx_int_9);
  __Pyx_GIVEREF(__pyx_int_9);
  PyTuple_SET_ITEM(__pyx_t_504, 0, __pyx_int_9);
  __Pyx_GIVEREF(__pyx_t_506);
  PyTuple_SET_ITEM(__pyx_t_504, 1, __pyx_t_506);
  __Pyx_INCREF(__pyx_int_2011);
  __Pyx_GIVEREF(__pyx_int_2011);
  PyTuple_SET_ITEM(__pyx_t_504, 2, __pyx_int_2011);
  __pyx_t_506 = 0;
  __pyx_t_506 = __Pyx_PyObject_Call(__pyx_t_505, __pyx_t_504, NULL); if (unlikely(!__pyx_t_506)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_506);
  __Pyx_DECREF(__pyx_t_505); __pyx_t_505 = 0;
  __Pyx_DECREF(__pyx_t_504); __pyx_t_504 = 0;

  /* "ultron/tradingday/Calendar.pyx":806
 *     Date.westernStyle(8, Months.October, 2011),
 *     Date.westernStyle(9, Months.October, 2011),
 *     Date.westernStyle(31, Months.December, 2011),             # <<<<<<<<<<<<<<
 *     # 2012
 *     Date.westernStyle(21, Months.January, 2012),
 */
  __pyx_t_504 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_504)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_504);
  __pyx_t_505 = __Pyx_PyInt_From_enum__Months(December); if (unlikely(!__pyx_t_505)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_505);
  __pyx_t_503 = PyTuple_New(3); if (unlikely(!__pyx_t_503)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_503);
  __Pyx_INCREF(__pyx_int_31);
  __Pyx_GIVEREF(__pyx_int_31);
  PyTuple_SET_ITEM(__pyx_t_503, 0, __pyx_int_31);
  __Pyx_GIVEREF(__pyx_t_505);
  PyTuple_SET_ITEM(__pyx_t_503, 1, __pyx_t_505);
  __Pyx_INCREF(__pyx_int_2011);
  __Pyx_GIVEREF(__pyx_int_2011);
  PyTuple_SET_ITEM(__pyx_t_503, 2, __pyx_int_2011);
  __pyx_t_505 = 0;
  __pyx_t_505 = __Pyx_PyObject_Call(__pyx_t_504, __pyx_t_503, NULL); if (unlikely(!__pyx_t_505)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_505);
  __Pyx_DECREF(__pyx_t_504); __pyx_t_504 = 0;
  __Pyx_DECREF(__pyx_t_503); __pyx_t_503 = 0;

  /* "ultron/tradingday/Calendar.pyx":808
 *     Date.westernStyle(31, Months.December, 2011),
 *     # 2012
 *     Date.westernStyle(21, Months.January, 2012),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.January, 2012),
 *     Date.westernStyle(31, Months.March, 2012),
 */
  __pyx_t_503 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_503)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_503);
  __pyx_t_504 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_504)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_504);
  __pyx_t_502 = PyTuple_New(3); if (unlikely(!__pyx_t_502)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_502);
  __Pyx_INCREF(__pyx_int_21);
  __Pyx_GIVEREF(__pyx_int_21);
  PyTuple_SET_ITEM(__pyx_t_502, 0, __pyx_int_21);
  __Pyx_GIVEREF(__pyx_t_504);
  PyTuple_SET_ITEM(__pyx_t_502, 1, __pyx_t_504);
  __Pyx_INCREF(__pyx_int_2012);
  __Pyx_GIVEREF(__pyx_int_2012);
  PyTuple_SET_ITEM(__pyx_t_502, 2, __pyx_int_2012);
  __pyx_t_504 = 0;
  __pyx_t_504 = __Pyx_PyObject_Call(__pyx_t_503, __pyx_t_502, NULL); if (unlikely(!__pyx_t_504)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_504);
  __Pyx_DECREF(__pyx_t_503); __pyx_t_503 = 0;
  __Pyx_DECREF(__pyx_t_502); __pyx_t_502 = 0;

  /* "ultron/tradingday/Calendar.pyx":809
 *     # 2012
 *     Date.westernStyle(21, Months.January, 2012),
 *     Date.westernStyle(29, Months.January, 2012),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(31, Months.March, 2012),
 *     Date.westernStyle(1, Months.April, 2012),
 */
  __pyx_t_502 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_502)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_502);
  __pyx_t_503 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_503)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_503);
  __pyx_t_501 = PyTuple_New(3); if (unlikely(!__pyx_t_501)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_501);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_501, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_503);
  PyTuple_SET_ITEM(__pyx_t_501, 1, __pyx_t_503);
  __Pyx_INCREF(__pyx_int_2012);
  __Pyx_GIVEREF(__pyx_int_2012);
  PyTuple_SET_ITEM(__pyx_t_501, 2, __pyx_int_2012);
  __pyx_t_503 = 0;
  __pyx_t_503 = __Pyx_PyObject_Call(__pyx_t_502, __pyx_t_501, NULL); if (unlikely(!__pyx_t_503)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_503);
  __Pyx_DECREF(__pyx_t_502); __pyx_t_502 = 0;
  __Pyx_DECREF(__pyx_t_501); __pyx_t_501 = 0;

  /* "ultron/tradingday/Calendar.pyx":810
 *     Date.westernStyle(21, Months.January, 2012),
 *     Date.westernStyle(29, Months.January, 2012),
 *     Date.westernStyle(31, Months.March, 2012),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(1, Months.April, 2012),
 *     Date.westernStyle(28, Months.April, 2012),
 */
  __pyx_t_501 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_501)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_501);
  __pyx_t_502 = __Pyx_PyInt_From_enum__Months(March); if (unlikely(!__pyx_t_502)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_502);
  __pyx_t_500 = PyTuple_New(3); if (unlikely(!__pyx_t_500)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_500);
  __Pyx_INCREF(__pyx_int_31);
  __Pyx_GIVEREF(__pyx_int_31);
  PyTuple_SET_ITEM(__pyx_t_500, 0, __pyx_int_31);
  __Pyx_GIVEREF(__pyx_t_502);
  PyTuple_SET_ITEM(__pyx_t_500, 1, __pyx_t_502);
  __Pyx_INCREF(__pyx_int_2012);
  __Pyx_GIVEREF(__pyx_int_2012);
  PyTuple_SET_ITEM(__pyx_t_500, 2, __pyx_int_2012);
  __pyx_t_502 = 0;
  __pyx_t_502 = __Pyx_PyObject_Call(__pyx_t_501, __pyx_t_500, NULL); if (unlikely(!__pyx_t_502)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_502);
  __Pyx_DECREF(__pyx_t_501); __pyx_t_501 = 0;
  __Pyx_DECREF(__pyx_t_500); __pyx_t_500 = 0;

  /* "ultron/tradingday/Calendar.pyx":811
 *     Date.westernStyle(29, Months.January, 2012),
 *     Date.westernStyle(31, Months.March, 2012),
 *     Date.westernStyle(1, Months.April, 2012),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(28, Months.April, 2012),
 *     Date.westernStyle(29, Months.September, 2012),
 */
  __pyx_t_500 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_500)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_500);
  __pyx_t_501 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_501)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_501);
  __pyx_t_499 = PyTuple_New(3); if (unlikely(!__pyx_t_499)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_499);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_t_499, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_t_501);
  PyTuple_SET_ITEM(__pyx_t_499, 1, __pyx_t_501);
  __Pyx_INCREF(__pyx_int_2012);
  __Pyx_GIVEREF(__pyx_int_2012);
  PyTuple_SET_ITEM(__pyx_t_499, 2, __pyx_int_2012);
  __pyx_t_501 = 0;
  __pyx_t_501 = __Pyx_PyObject_Call(__pyx_t_500, __pyx_t_499, NULL); if (unlikely(!__pyx_t_501)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_501);
  __Pyx_DECREF(__pyx_t_500); __pyx_t_500 = 0;
  __Pyx_DECREF(__pyx_t_499); __pyx_t_499 = 0;

  /* "ultron/tradingday/Calendar.pyx":812
 *     Date.westernStyle(31, Months.March, 2012),
 *     Date.westernStyle(1, Months.April, 2012),
 *     Date.westernStyle(28, Months.April, 2012),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.September, 2012),
 *     # 2013
 */
  __pyx_t_499 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_499)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_499);
  __pyx_t_500 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_500)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_500);
  __pyx_t_498 = PyTuple_New(3); if (unlikely(!__pyx_t_498)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_498);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_498, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_500);
  PyTuple_SET_ITEM(__pyx_t_498, 1, __pyx_t_500);
  __Pyx_INCREF(__pyx_int_2012);
  __Pyx_GIVEREF(__pyx_int_2012);
  PyTuple_SET_ITEM(__pyx_t_498, 2, __pyx_int_2012);
  __pyx_t_500 = 0;
  __pyx_t_500 = __Pyx_PyObject_Call(__pyx_t_499, __pyx_t_498, NULL); if (unlikely(!__pyx_t_500)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_500);
  __Pyx_DECREF(__pyx_t_499); __pyx_t_499 = 0;
  __Pyx_DECREF(__pyx_t_498); __pyx_t_498 = 0;

  /* "ultron/tradingday/Calendar.pyx":813
 *     Date.westernStyle(1, Months.April, 2012),
 *     Date.westernStyle(28, Months.April, 2012),
 *     Date.westernStyle(29, Months.September, 2012),             # <<<<<<<<<<<<<<
 *     # 2013
 *     Date.westernStyle(5, Months.January, 2013),
 */
  __pyx_t_498 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_498)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_498);
  __pyx_t_499 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_499)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_499);
  __pyx_t_497 = PyTuple_New(3); if (unlikely(!__pyx_t_497)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_497);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_497, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_499);
  PyTuple_SET_ITEM(__pyx_t_497, 1, __pyx_t_499);
  __Pyx_INCREF(__pyx_int_2012);
  __Pyx_GIVEREF(__pyx_int_2012);
  PyTuple_SET_ITEM(__pyx_t_497, 2, __pyx_int_2012);
  __pyx_t_499 = 0;
  __pyx_t_499 = __Pyx_PyObject_Call(__pyx_t_498, __pyx_t_497, NULL); if (unlikely(!__pyx_t_499)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_499);
  __Pyx_DECREF(__pyx_t_498); __pyx_t_498 = 0;
  __Pyx_DECREF(__pyx_t_497); __pyx_t_497 = 0;

  /* "ultron/tradingday/Calendar.pyx":815
 *     Date.westernStyle(29, Months.September, 2012),
 *     # 2013
 *     Date.westernStyle(5, Months.January, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(6, Months.January, 2013),
 *     Date.westernStyle(16, Months.February, 2013),
 */
  __pyx_t_497 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_497)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_497);
  __pyx_t_498 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_498)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_498);
  __pyx_t_496 = PyTuple_New(3); if (unlikely(!__pyx_t_496)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_496);
  __Pyx_INCREF(__pyx_int_5);
  __Pyx_GIVEREF(__pyx_int_5);
  PyTuple_SET_ITEM(__pyx_t_496, 0, __pyx_int_5);
  __Pyx_GIVEREF(__pyx_t_498);
  PyTuple_SET_ITEM(__pyx_t_496, 1, __pyx_t_498);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_496, 2, __pyx_int_2013);
  __pyx_t_498 = 0;
  __pyx_t_498 = __Pyx_PyObject_Call(__pyx_t_497, __pyx_t_496, NULL); if (unlikely(!__pyx_t_498)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_498);
  __Pyx_DECREF(__pyx_t_497); __pyx_t_497 = 0;
  __Pyx_DECREF(__pyx_t_496); __pyx_t_496 = 0;

  /* "ultron/tradingday/Calendar.pyx":816
 *     # 2013
 *     Date.westernStyle(5, Months.January, 2013),
 *     Date.westernStyle(6, Months.January, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(16, Months.February, 2013),
 *     Date.westernStyle(17, Months.February, 2013),
 */
  __pyx_t_496 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_496)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_496);
  __pyx_t_497 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_497)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_497);
  __pyx_t_495 = PyTuple_New(3); if (unlikely(!__pyx_t_495)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_495);
  __Pyx_INCREF(__pyx_int_6);
  __Pyx_GIVEREF(__pyx_int_6);
  PyTuple_SET_ITEM(__pyx_t_495, 0, __pyx_int_6);
  __Pyx_GIVEREF(__pyx_t_497);
  PyTuple_SET_ITEM(__pyx_t_495, 1, __pyx_t_497);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_495, 2, __pyx_int_2013);
  __pyx_t_497 = 0;
  __pyx_t_497 = __Pyx_PyObject_Call(__pyx_t_496, __pyx_t_495, NULL); if (unlikely(!__pyx_t_497)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_497);
  __Pyx_DECREF(__pyx_t_496); __pyx_t_496 = 0;
  __Pyx_DECREF(__pyx_t_495); __pyx_t_495 = 0;

  /* "ultron/tradingday/Calendar.pyx":817
 *     Date.westernStyle(5, Months.January, 2013),
 *     Date.westernStyle(6, Months.January, 2013),
 *     Date.westernStyle(16, Months.February, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(17, Months.February, 2013),
 *     Date.westernStyle(7, Months.April, 2013),
 */
  __pyx_t_495 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_495)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_495);
  __pyx_t_496 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_496)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_496);
  __pyx_t_494 = PyTuple_New(3); if (unlikely(!__pyx_t_494)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_494);
  __Pyx_INCREF(__pyx_int_16);
  __Pyx_GIVEREF(__pyx_int_16);
  PyTuple_SET_ITEM(__pyx_t_494, 0, __pyx_int_16);
  __Pyx_GIVEREF(__pyx_t_496);
  PyTuple_SET_ITEM(__pyx_t_494, 1, __pyx_t_496);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_494, 2, __pyx_int_2013);
  __pyx_t_496 = 0;
  __pyx_t_496 = __Pyx_PyObject_Call(__pyx_t_495, __pyx_t_494, NULL); if (unlikely(!__pyx_t_496)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_496);
  __Pyx_DECREF(__pyx_t_495); __pyx_t_495 = 0;
  __Pyx_DECREF(__pyx_t_494); __pyx_t_494 = 0;

  /* "ultron/tradingday/Calendar.pyx":818
 *     Date.westernStyle(6, Months.January, 2013),
 *     Date.westernStyle(16, Months.February, 2013),
 *     Date.westernStyle(17, Months.February, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(7, Months.April, 2013),
 *     Date.westernStyle(27, Months.April, 2013),
 */
  __pyx_t_494 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_494)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_494);
  __pyx_t_495 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_495)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_495);
  __pyx_t_493 = PyTuple_New(3); if (unlikely(!__pyx_t_493)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_493);
  __Pyx_INCREF(__pyx_int_17);
  __Pyx_GIVEREF(__pyx_int_17);
  PyTuple_SET_ITEM(__pyx_t_493, 0, __pyx_int_17);
  __Pyx_GIVEREF(__pyx_t_495);
  PyTuple_SET_ITEM(__pyx_t_493, 1, __pyx_t_495);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_493, 2, __pyx_int_2013);
  __pyx_t_495 = 0;
  __pyx_t_495 = __Pyx_PyObject_Call(__pyx_t_494, __pyx_t_493, NULL); if (unlikely(!__pyx_t_495)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_495);
  __Pyx_DECREF(__pyx_t_494); __pyx_t_494 = 0;
  __Pyx_DECREF(__pyx_t_493); __pyx_t_493 = 0;

  /* "ultron/tradingday/Calendar.pyx":819
 *     Date.westernStyle(16, Months.February, 2013),
 *     Date.westernStyle(17, Months.February, 2013),
 *     Date.westernStyle(7, Months.April, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(27, Months.April, 2013),
 *     Date.westernStyle(28, Months.April, 2013),
 */
  __pyx_t_493 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_493)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_493);
  __pyx_t_494 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_494)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_494);
  __pyx_t_492 = PyTuple_New(3); if (unlikely(!__pyx_t_492)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_492);
  __Pyx_INCREF(__pyx_int_7);
  __Pyx_GIVEREF(__pyx_int_7);
  PyTuple_SET_ITEM(__pyx_t_492, 0, __pyx_int_7);
  __Pyx_GIVEREF(__pyx_t_494);
  PyTuple_SET_ITEM(__pyx_t_492, 1, __pyx_t_494);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_492, 2, __pyx_int_2013);
  __pyx_t_494 = 0;
  __pyx_t_494 = __Pyx_PyObject_Call(__pyx_t_493, __pyx_t_492, NULL); if (unlikely(!__pyx_t_494)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_494);
  __Pyx_DECREF(__pyx_t_493); __pyx_t_493 = 0;
  __Pyx_DECREF(__pyx_t_492); __pyx_t_492 = 0;

  /* "ultron/tradingday/Calendar.pyx":820
 *     Date.westernStyle(17, Months.February, 2013),
 *     Date.westernStyle(7, Months.April, 2013),
 *     Date.westernStyle(27, Months.April, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(28, Months.April, 2013),
 *     Date.westernStyle(8, Months.June, 2013),
 */
  __pyx_t_492 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_492)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_492);
  __pyx_t_493 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_493)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_493);
  __pyx_t_491 = PyTuple_New(3); if (unlikely(!__pyx_t_491)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_491);
  __Pyx_INCREF(__pyx_int_27);
  __Pyx_GIVEREF(__pyx_int_27);
  PyTuple_SET_ITEM(__pyx_t_491, 0, __pyx_int_27);
  __Pyx_GIVEREF(__pyx_t_493);
  PyTuple_SET_ITEM(__pyx_t_491, 1, __pyx_t_493);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_491, 2, __pyx_int_2013);
  __pyx_t_493 = 0;
  __pyx_t_493 = __Pyx_PyObject_Call(__pyx_t_492, __pyx_t_491, NULL); if (unlikely(!__pyx_t_493)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_493);
  __Pyx_DECREF(__pyx_t_492); __pyx_t_492 = 0;
  __Pyx_DECREF(__pyx_t_491); __pyx_t_491 = 0;

  /* "ultron/tradingday/Calendar.pyx":821
 *     Date.westernStyle(7, Months.April, 2013),
 *     Date.westernStyle(27, Months.April, 2013),
 *     Date.westernStyle(28, Months.April, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(8, Months.June, 2013),
 *     Date.westernStyle(9, Months.June, 2013),
 */
  __pyx_t_491 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_491)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_491);
  __pyx_t_492 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_492)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_492);
  __pyx_t_490 = PyTuple_New(3); if (unlikely(!__pyx_t_490)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_490);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_490, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_492);
  PyTuple_SET_ITEM(__pyx_t_490, 1, __pyx_t_492);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_490, 2, __pyx_int_2013);
  __pyx_t_492 = 0;
  __pyx_t_492 = __Pyx_PyObject_Call(__pyx_t_491, __pyx_t_490, NULL); if (unlikely(!__pyx_t_492)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_492);
  __Pyx_DECREF(__pyx_t_491); __pyx_t_491 = 0;
  __Pyx_DECREF(__pyx_t_490); __pyx_t_490 = 0;

  /* "ultron/tradingday/Calendar.pyx":822
 *     Date.westernStyle(27, Months.April, 2013),
 *     Date.westernStyle(28, Months.April, 2013),
 *     Date.westernStyle(8, Months.June, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(9, Months.June, 2013),
 *     Date.westernStyle(22, Months.September, 2013),
 */
  __pyx_t_490 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_490)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_490);
  __pyx_t_491 = __Pyx_PyInt_From_enum__Months(June); if (unlikely(!__pyx_t_491)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_491);
  __pyx_t_489 = PyTuple_New(3); if (unlikely(!__pyx_t_489)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_489);
  __Pyx_INCREF(__pyx_int_8);
  __Pyx_GIVEREF(__pyx_int_8);
  PyTuple_SET_ITEM(__pyx_t_489, 0, __pyx_int_8);
  __Pyx_GIVEREF(__pyx_t_491);
  PyTuple_SET_ITEM(__pyx_t_489, 1, __pyx_t_491);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_489, 2, __pyx_int_2013);
  __pyx_t_491 = 0;
  __pyx_t_491 = __Pyx_PyObject_Call(__pyx_t_490, __pyx_t_489, NULL); if (unlikely(!__pyx_t_491)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_491);
  __Pyx_DECREF(__pyx_t_490); __pyx_t_490 = 0;
  __Pyx_DECREF(__pyx_t_489); __pyx_t_489 = 0;

  /* "ultron/tradingday/Calendar.pyx":823
 *     Date.westernStyle(28, Months.April, 2013),
 *     Date.westernStyle(8, Months.June, 2013),
 *     Date.westernStyle(9, Months.June, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(22, Months.September, 2013),
 *     Date.westernStyle(29, Months.September, 2013),
 */
  __pyx_t_489 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_489)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_489);
  __pyx_t_490 = __Pyx_PyInt_From_enum__Months(June); if (unlikely(!__pyx_t_490)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_490);
  __pyx_t_488 = PyTuple_New(3); if (unlikely(!__pyx_t_488)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_488);
  __Pyx_INCREF(__pyx_int_9);
  __Pyx_GIVEREF(__pyx_int_9);
  PyTuple_SET_ITEM(__pyx_t_488, 0, __pyx_int_9);
  __Pyx_GIVEREF(__pyx_t_490);
  PyTuple_SET_ITEM(__pyx_t_488, 1, __pyx_t_490);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_488, 2, __pyx_int_2013);
  __pyx_t_490 = 0;
  __pyx_t_490 = __Pyx_PyObject_Call(__pyx_t_489, __pyx_t_488, NULL); if (unlikely(!__pyx_t_490)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_490);
  __Pyx_DECREF(__pyx_t_489); __pyx_t_489 = 0;
  __Pyx_DECREF(__pyx_t_488); __pyx_t_488 = 0;

  /* "ultron/tradingday/Calendar.pyx":824
 *     Date.westernStyle(8, Months.June, 2013),
 *     Date.westernStyle(9, Months.June, 2013),
 *     Date.westernStyle(22, Months.September, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.September, 2013),
 *     Date.westernStyle(12, Months.October, 2013),
 */
  __pyx_t_488 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_488)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_488);
  __pyx_t_489 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_489)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_489);
  __pyx_t_487 = PyTuple_New(3); if (unlikely(!__pyx_t_487)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_487);
  __Pyx_INCREF(__pyx_int_22);
  __Pyx_GIVEREF(__pyx_int_22);
  PyTuple_SET_ITEM(__pyx_t_487, 0, __pyx_int_22);
  __Pyx_GIVEREF(__pyx_t_489);
  PyTuple_SET_ITEM(__pyx_t_487, 1, __pyx_t_489);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_487, 2, __pyx_int_2013);
  __pyx_t_489 = 0;
  __pyx_t_489 = __Pyx_PyObject_Call(__pyx_t_488, __pyx_t_487, NULL); if (unlikely(!__pyx_t_489)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_489);
  __Pyx_DECREF(__pyx_t_488); __pyx_t_488 = 0;
  __Pyx_DECREF(__pyx_t_487); __pyx_t_487 = 0;

  /* "ultron/tradingday/Calendar.pyx":825
 *     Date.westernStyle(9, Months.June, 2013),
 *     Date.westernStyle(22, Months.September, 2013),
 *     Date.westernStyle(29, Months.September, 2013),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(12, Months.October, 2013),
 *     # 2014
 */
  __pyx_t_487 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_487)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_487);
  __pyx_t_488 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_488)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_488);
  __pyx_t_486 = PyTuple_New(3); if (unlikely(!__pyx_t_486)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_486);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_486, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_488);
  PyTuple_SET_ITEM(__pyx_t_486, 1, __pyx_t_488);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_486, 2, __pyx_int_2013);
  __pyx_t_488 = 0;
  __pyx_t_488 = __Pyx_PyObject_Call(__pyx_t_487, __pyx_t_486, NULL); if (unlikely(!__pyx_t_488)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_488);
  __Pyx_DECREF(__pyx_t_487); __pyx_t_487 = 0;
  __Pyx_DECREF(__pyx_t_486); __pyx_t_486 = 0;

  /* "ultron/tradingday/Calendar.pyx":826
 *     Date.westernStyle(22, Months.September, 2013),
 *     Date.westernStyle(29, Months.September, 2013),
 *     Date.westernStyle(12, Months.October, 2013),             # <<<<<<<<<<<<<<
 *     # 2014
 *     Date.westernStyle(26, Months.January, 2014),
 */
  __pyx_t_486 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_486)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_486);
  __pyx_t_487 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_487)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_487);
  __pyx_t_485 = PyTuple_New(3); if (unlikely(!__pyx_t_485)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_485);
  __Pyx_INCREF(__pyx_int_12);
  __Pyx_GIVEREF(__pyx_int_12);
  PyTuple_SET_ITEM(__pyx_t_485, 0, __pyx_int_12);
  __Pyx_GIVEREF(__pyx_t_487);
  PyTuple_SET_ITEM(__pyx_t_485, 1, __pyx_t_487);
  __Pyx_INCREF(__pyx_int_2013);
  __Pyx_GIVEREF(__pyx_int_2013);
  PyTuple_SET_ITEM(__pyx_t_485, 2, __pyx_int_2013);
  __pyx_t_487 = 0;
  __pyx_t_487 = __Pyx_PyObject_Call(__pyx_t_486, __pyx_t_485, NULL); if (unlikely(!__pyx_t_487)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_487);
  __Pyx_DECREF(__pyx_t_486); __pyx_t_486 = 0;
  __Pyx_DECREF(__pyx_t_485); __pyx_t_485 = 0;

  /* "ultron/tradingday/Calendar.pyx":828
 *     Date.westernStyle(12, Months.October, 2013),
 *     # 2014
 *     Date.westernStyle(26, Months.January, 2014),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(8, Months.February, 2014),
 *     Date.westernStyle(4, Months.May, 2014),
 */
  __pyx_t_485 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_485)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_485);
  __pyx_t_486 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_486)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_486);
  __pyx_t_484 = PyTuple_New(3); if (unlikely(!__pyx_t_484)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_484);
  __Pyx_INCREF(__pyx_int_26);
  __Pyx_GIVEREF(__pyx_int_26);
  PyTuple_SET_ITEM(__pyx_t_484, 0, __pyx_int_26);
  __Pyx_GIVEREF(__pyx_t_486);
  PyTuple_SET_ITEM(__pyx_t_484, 1, __pyx_t_486);
  __Pyx_INCREF(__pyx_int_2014);
  __Pyx_GIVEREF(__pyx_int_2014);
  PyTuple_SET_ITEM(__pyx_t_484, 2, __pyx_int_2014);
  __pyx_t_486 = 0;
  __pyx_t_486 = __Pyx_PyObject_Call(__pyx_t_485, __pyx_t_484, NULL); if (unlikely(!__pyx_t_486)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_486);
  __Pyx_DECREF(__pyx_t_485); __pyx_t_485 = 0;
  __Pyx_DECREF(__pyx_t_484); __pyx_t_484 = 0;

  /* "ultron/tradingday/Calendar.pyx":829
 *     # 2014
 *     Date.westernStyle(26, Months.January, 2014),
 *     Date.westernStyle(8, Months.February, 2014),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(4, Months.May, 2014),
 *     Date.westernStyle(28, Months.September, 2014),
 */
  __pyx_t_484 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_484)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_484);
  __pyx_t_485 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_485)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_485);
  __pyx_t_483 = PyTuple_New(3); if (unlikely(!__pyx_t_483)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_483);
  __Pyx_INCREF(__pyx_int_8);
  __Pyx_GIVEREF(__pyx_int_8);
  PyTuple_SET_ITEM(__pyx_t_483, 0, __pyx_int_8);
  __Pyx_GIVEREF(__pyx_t_485);
  PyTuple_SET_ITEM(__pyx_t_483, 1, __pyx_t_485);
  __Pyx_INCREF(__pyx_int_2014);
  __Pyx_GIVEREF(__pyx_int_2014);
  PyTuple_SET_ITEM(__pyx_t_483, 2, __pyx_int_2014);
  __pyx_t_485 = 0;
  __pyx_t_485 = __Pyx_PyObject_Call(__pyx_t_484, __pyx_t_483, NULL); if (unlikely(!__pyx_t_485)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_485);
  __Pyx_DECREF(__pyx_t_484); __pyx_t_484 = 0;
  __Pyx_DECREF(__pyx_t_483); __pyx_t_483 = 0;

  /* "ultron/tradingday/Calendar.pyx":830
 *     Date.westernStyle(26, Months.January, 2014),
 *     Date.westernStyle(8, Months.February, 2014),
 *     Date.westernStyle(4, Months.May, 2014),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(28, Months.September, 2014),
 *     Date.westernStyle(11, Months.October, 2014),
 */
  __pyx_t_483 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_483)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_483);
  __pyx_t_484 = __Pyx_PyInt_From_enum__Months(May); if (unlikely(!__pyx_t_484)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_484);
  __pyx_t_482 = PyTuple_New(3); if (unlikely(!__pyx_t_482)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_482);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_482, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_t_484);
  PyTuple_SET_ITEM(__pyx_t_482, 1, __pyx_t_484);
  __Pyx_INCREF(__pyx_int_2014);
  __Pyx_GIVEREF(__pyx_int_2014);
  PyTuple_SET_ITEM(__pyx_t_482, 2, __pyx_int_2014);
  __pyx_t_484 = 0;
  __pyx_t_484 = __Pyx_PyObject_Call(__pyx_t_483, __pyx_t_482, NULL); if (unlikely(!__pyx_t_484)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_484);
  __Pyx_DECREF(__pyx_t_483); __pyx_t_483 = 0;
  __Pyx_DECREF(__pyx_t_482); __pyx_t_482 = 0;

  /* "ultron/tradingday/Calendar.pyx":831
 *     Date.westernStyle(8, Months.February, 2014),
 *     Date.westernStyle(4, Months.May, 2014),
 *     Date.westernStyle(28, Months.September, 2014),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(11, Months.October, 2014),
 *     # 2015
 */
  __pyx_t_482 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_482)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_482);
  __pyx_t_483 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_483)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_483);
  __pyx_t_481 = PyTuple_New(3); if (unlikely(!__pyx_t_481)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_481);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_481, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_483);
  PyTuple_SET_ITEM(__pyx_t_481, 1, __pyx_t_483);
  __Pyx_INCREF(__pyx_int_2014);
  __Pyx_GIVEREF(__pyx_int_2014);
  PyTuple_SET_ITEM(__pyx_t_481, 2, __pyx_int_2014);
  __pyx_t_483 = 0;
  __pyx_t_483 = __Pyx_PyObject_Call(__pyx_t_482, __pyx_t_481, NULL); if (unlikely(!__pyx_t_483)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_483);
  __Pyx_DECREF(__pyx_t_482); __pyx_t_482 = 0;
  __Pyx_DECREF(__pyx_t_481); __pyx_t_481 = 0;

  /* "ultron/tradingday/Calendar.pyx":832
 *     Date.westernStyle(4, Months.May, 2014),
 *     Date.westernStyle(28, Months.September, 2014),
 *     Date.westernStyle(11, Months.October, 2014),             # <<<<<<<<<<<<<<
 *     # 2015
 *     Date.westernStyle(4, Months.January, 2015),
 */
  __pyx_t_481 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_481)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_481);
  __pyx_t_482 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_482)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_482);
  __pyx_t_480 = PyTuple_New(3); if (unlikely(!__pyx_t_480)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_480);
  __Pyx_INCREF(__pyx_int_11);
  __Pyx_GIVEREF(__pyx_int_11);
  PyTuple_SET_ITEM(__pyx_t_480, 0, __pyx_int_11);
  __Pyx_GIVEREF(__pyx_t_482);
  PyTuple_SET_ITEM(__pyx_t_480, 1, __pyx_t_482);
  __Pyx_INCREF(__pyx_int_2014);
  __Pyx_GIVEREF(__pyx_int_2014);
  PyTuple_SET_ITEM(__pyx_t_480, 2, __pyx_int_2014);
  __pyx_t_482 = 0;
  __pyx_t_482 = __Pyx_PyObject_Call(__pyx_t_481, __pyx_t_480, NULL); if (unlikely(!__pyx_t_482)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_482);
  __Pyx_DECREF(__pyx_t_481); __pyx_t_481 = 0;
  __Pyx_DECREF(__pyx_t_480); __pyx_t_480 = 0;

  /* "ultron/tradingday/Calendar.pyx":834
 *     Date.westernStyle(11, Months.October, 2014),
 *     # 2015
 *     Date.westernStyle(4, Months.January, 2015),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(15, Months.February, 2015),
 *     Date.westernStyle(28, Months.February, 2015),
 */
  __pyx_t_480 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_480)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_480);
  __pyx_t_481 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_481)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_481);
  __pyx_t_479 = PyTuple_New(3); if (unlikely(!__pyx_t_479)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_479);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_479, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_t_481);
  PyTuple_SET_ITEM(__pyx_t_479, 1, __pyx_t_481);
  __Pyx_INCREF(__pyx_int_2015);
  __Pyx_GIVEREF(__pyx_int_2015);
  PyTuple_SET_ITEM(__pyx_t_479, 2, __pyx_int_2015);
  __pyx_t_481 = 0;
  __pyx_t_481 = __Pyx_PyObject_Call(__pyx_t_480, __pyx_t_479, NULL); if (unlikely(!__pyx_t_481)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_481);
  __Pyx_DECREF(__pyx_t_480); __pyx_t_480 = 0;
  __Pyx_DECREF(__pyx_t_479); __pyx_t_479 = 0;

  /* "ultron/tradingday/Calendar.pyx":835
 *     # 2015
 *     Date.westernStyle(4, Months.January, 2015),
 *     Date.westernStyle(15, Months.February, 2015),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(28, Months.February, 2015),
 *     Date.westernStyle(6, Months.September, 2015),
 */
  __pyx_t_479 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_479)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_479);
  __pyx_t_480 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_480)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_480);
  __pyx_t_478 = PyTuple_New(3); if (unlikely(!__pyx_t_478)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_478);
  __Pyx_INCREF(__pyx_int_15);
  __Pyx_GIVEREF(__pyx_int_15);
  PyTuple_SET_ITEM(__pyx_t_478, 0, __pyx_int_15);
  __Pyx_GIVEREF(__pyx_t_480);
  PyTuple_SET_ITEM(__pyx_t_478, 1, __pyx_t_480);
  __Pyx_INCREF(__pyx_int_2015);
  __Pyx_GIVEREF(__pyx_int_2015);
  PyTuple_SET_ITEM(__pyx_t_478, 2, __pyx_int_2015);
  __pyx_t_480 = 0;
  __pyx_t_480 = __Pyx_PyObject_Call(__pyx_t_479, __pyx_t_478, NULL); if (unlikely(!__pyx_t_480)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_480);
  __Pyx_DECREF(__pyx_t_479); __pyx_t_479 = 0;
  __Pyx_DECREF(__pyx_t_478); __pyx_t_478 = 0;

  /* "ultron/tradingday/Calendar.pyx":836
 *     Date.westernStyle(4, Months.January, 2015),
 *     Date.westernStyle(15, Months.February, 2015),
 *     Date.westernStyle(28, Months.February, 2015),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(6, Months.September, 2015),
 *     Date.westernStyle(10, Months.October, 2015),
 */
  __pyx_t_478 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_478)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_478);
  __pyx_t_479 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_479)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_479);
  __pyx_t_477 = PyTuple_New(3); if (unlikely(!__pyx_t_477)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_477);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_477, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_479);
  PyTuple_SET_ITEM(__pyx_t_477, 1, __pyx_t_479);
  __Pyx_INCREF(__pyx_int_2015);
  __Pyx_GIVEREF(__pyx_int_2015);
  PyTuple_SET_ITEM(__pyx_t_477, 2, __pyx_int_2015);
  __pyx_t_479 = 0;
  __pyx_t_479 = __Pyx_PyObject_Call(__pyx_t_478, __pyx_t_477, NULL); if (unlikely(!__pyx_t_479)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_479);
  __Pyx_DECREF(__pyx_t_478); __pyx_t_478 = 0;
  __Pyx_DECREF(__pyx_t_477); __pyx_t_477 = 0;

  /* "ultron/tradingday/Calendar.pyx":837
 *     Date.westernStyle(15, Months.February, 2015),
 *     Date.westernStyle(28, Months.February, 2015),
 *     Date.westernStyle(6, Months.September, 2015),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(10, Months.October, 2015),
 *     # 2016
 */
  __pyx_t_477 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_477)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_477);
  __pyx_t_478 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_478)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_478);
  __pyx_t_476 = PyTuple_New(3); if (unlikely(!__pyx_t_476)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_476);
  __Pyx_INCREF(__pyx_int_6);
  __Pyx_GIVEREF(__pyx_int_6);
  PyTuple_SET_ITEM(__pyx_t_476, 0, __pyx_int_6);
  __Pyx_GIVEREF(__pyx_t_478);
  PyTuple_SET_ITEM(__pyx_t_476, 1, __pyx_t_478);
  __Pyx_INCREF(__pyx_int_2015);
  __Pyx_GIVEREF(__pyx_int_2015);
  PyTuple_SET_ITEM(__pyx_t_476, 2, __pyx_int_2015);
  __pyx_t_478 = 0;
  __pyx_t_478 = __Pyx_PyObject_Call(__pyx_t_477, __pyx_t_476, NULL); if (unlikely(!__pyx_t_478)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_478);
  __Pyx_DECREF(__pyx_t_477); __pyx_t_477 = 0;
  __Pyx_DECREF(__pyx_t_476); __pyx_t_476 = 0;

  /* "ultron/tradingday/Calendar.pyx":838
 *     Date.westernStyle(28, Months.February, 2015),
 *     Date.westernStyle(6, Months.September, 2015),
 *     Date.westernStyle(10, Months.October, 2015),             # <<<<<<<<<<<<<<
 *     # 2016
 *     Date.westernStyle(6, Months.February, 2016),
 */
  __pyx_t_476 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_476)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_476);
  __pyx_t_477 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_477)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_477);
  __pyx_t_475 = PyTuple_New(3); if (unlikely(!__pyx_t_475)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_475);
  __Pyx_INCREF(__pyx_int_10);
  __Pyx_GIVEREF(__pyx_int_10);
  PyTuple_SET_ITEM(__pyx_t_475, 0, __pyx_int_10);
  __Pyx_GIVEREF(__pyx_t_477);
  PyTuple_SET_ITEM(__pyx_t_475, 1, __pyx_t_477);
  __Pyx_INCREF(__pyx_int_2015);
  __Pyx_GIVEREF(__pyx_int_2015);
  PyTuple_SET_ITEM(__pyx_t_475, 2, __pyx_int_2015);
  __pyx_t_477 = 0;
  __pyx_t_477 = __Pyx_PyObject_Call(__pyx_t_476, __pyx_t_475, NULL); if (unlikely(!__pyx_t_477)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_477);
  __Pyx_DECREF(__pyx_t_476); __pyx_t_476 = 0;
  __Pyx_DECREF(__pyx_t_475); __pyx_t_475 = 0;

  /* "ultron/tradingday/Calendar.pyx":840
 *     Date.westernStyle(10, Months.October, 2015),
 *     # 2016
 *     Date.westernStyle(6, Months.February, 2016),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(14, Months.February, 2016),
 *     Date.westernStyle(12, Months.June, 2016),
 */
  __pyx_t_475 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_475)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_475);
  __pyx_t_476 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_476)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_476);
  __pyx_t_474 = PyTuple_New(3); if (unlikely(!__pyx_t_474)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_474);
  __Pyx_INCREF(__pyx_int_6);
  __Pyx_GIVEREF(__pyx_int_6);
  PyTuple_SET_ITEM(__pyx_t_474, 0, __pyx_int_6);
  __Pyx_GIVEREF(__pyx_t_476);
  PyTuple_SET_ITEM(__pyx_t_474, 1, __pyx_t_476);
  __Pyx_INCREF(__pyx_int_2016);
  __Pyx_GIVEREF(__pyx_int_2016);
  PyTuple_SET_ITEM(__pyx_t_474, 2, __pyx_int_2016);
  __pyx_t_476 = 0;
  __pyx_t_476 = __Pyx_PyObject_Call(__pyx_t_475, __pyx_t_474, NULL); if (unlikely(!__pyx_t_476)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_476);
  __Pyx_DECREF(__pyx_t_475); __pyx_t_475 = 0;
  __Pyx_DECREF(__pyx_t_474); __pyx_t_474 = 0;

  /* "ultron/tradingday/Calendar.pyx":841
 *     # 2016
 *     Date.westernStyle(6, Months.February, 2016),
 *     Date.westernStyle(14, Months.February, 2016),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(12, Months.June, 2016),
 *     Date.westernStyle(18, Months.September, 2016),
 */
  __pyx_t_474 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_474)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_474);
  __pyx_t_475 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_475)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_475);
  __pyx_t_473 = PyTuple_New(3); if (unlikely(!__pyx_t_473)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_473);
  __Pyx_INCREF(__pyx_int_14);
  __Pyx_GIVEREF(__pyx_int_14);
  PyTuple_SET_ITEM(__pyx_t_473, 0, __pyx_int_14);
  __Pyx_GIVEREF(__pyx_t_475);
  PyTuple_SET_ITEM(__pyx_t_473, 1, __pyx_t_475);
  __Pyx_INCREF(__pyx_int_2016);
  __Pyx_GIVEREF(__pyx_int_2016);
  PyTuple_SET_ITEM(__pyx_t_473, 2, __pyx_int_2016);
  __pyx_t_475 = 0;
  __pyx_t_475 = __Pyx_PyObject_Call(__pyx_t_474, __pyx_t_473, NULL); if (unlikely(!__pyx_t_475)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_475);
  __Pyx_DECREF(__pyx_t_474); __pyx_t_474 = 0;
  __Pyx_DECREF(__pyx_t_473); __pyx_t_473 = 0;

  /* "ultron/tradingday/Calendar.pyx":842
 *     Date.westernStyle(6, Months.February, 2016),
 *     Date.westernStyle(14, Months.February, 2016),
 *     Date.westernStyle(12, Months.June, 2016),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(18, Months.September, 2016),
 *     Date.westernStyle(8, Months.October, 2016),
 */
  __pyx_t_473 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_473)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_473);
  __pyx_t_474 = __Pyx_PyInt_From_enum__Months(June); if (unlikely(!__pyx_t_474)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_474);
  __pyx_t_472 = PyTuple_New(3); if (unlikely(!__pyx_t_472)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_472);
  __Pyx_INCREF(__pyx_int_12);
  __Pyx_GIVEREF(__pyx_int_12);
  PyTuple_SET_ITEM(__pyx_t_472, 0, __pyx_int_12);
  __Pyx_GIVEREF(__pyx_t_474);
  PyTuple_SET_ITEM(__pyx_t_472, 1, __pyx_t_474);
  __Pyx_INCREF(__pyx_int_2016);
  __Pyx_GIVEREF(__pyx_int_2016);
  PyTuple_SET_ITEM(__pyx_t_472, 2, __pyx_int_2016);
  __pyx_t_474 = 0;
  __pyx_t_474 = __Pyx_PyObject_Call(__pyx_t_473, __pyx_t_472, NULL); if (unlikely(!__pyx_t_474)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_474);
  __Pyx_DECREF(__pyx_t_473); __pyx_t_473 = 0;
  __Pyx_DECREF(__pyx_t_472); __pyx_t_472 = 0;

  /* "ultron/tradingday/Calendar.pyx":843
 *     Date.westernStyle(14, Months.February, 2016),
 *     Date.westernStyle(12, Months.June, 2016),
 *     Date.westernStyle(18, Months.September, 2016),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(8, Months.October, 2016),
 *     Date.westernStyle(9, Months.October, 2016),
 */
  __pyx_t_472 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_472)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_472);
  __pyx_t_473 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_473)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_473);
  __pyx_t_471 = PyTuple_New(3); if (unlikely(!__pyx_t_471)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_471);
  __Pyx_INCREF(__pyx_int_18);
  __Pyx_GIVEREF(__pyx_int_18);
  PyTuple_SET_ITEM(__pyx_t_471, 0, __pyx_int_18);
  __Pyx_GIVEREF(__pyx_t_473);
  PyTuple_SET_ITEM(__pyx_t_471, 1, __pyx_t_473);
  __Pyx_INCREF(__pyx_int_2016);
  __Pyx_GIVEREF(__pyx_int_2016);
  PyTuple_SET_ITEM(__pyx_t_471, 2, __pyx_int_2016);
  __pyx_t_473 = 0;
  __pyx_t_473 = __Pyx_PyObject_Call(__pyx_t_472, __pyx_t_471, NULL); if (unlikely(!__pyx_t_473)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_473);
  __Pyx_DECREF(__pyx_t_472); __pyx_t_472 = 0;
  __Pyx_DECREF(__pyx_t_471); __pyx_t_471 = 0;

  /* "ultron/tradingday/Calendar.pyx":844
 *     Date.westernStyle(12, Months.June, 2016),
 *     Date.westernStyle(18, Months.September, 2016),
 *     Date.westernStyle(8, Months.October, 2016),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(9, Months.October, 2016),
 *     # 2017
 */
  __pyx_t_471 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_471)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_471);
  __pyx_t_472 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_472)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_472);
  __pyx_t_470 = PyTuple_New(3); if (unlikely(!__pyx_t_470)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_470);
  __Pyx_INCREF(__pyx_int_8);
  __Pyx_GIVEREF(__pyx_int_8);
  PyTuple_SET_ITEM(__pyx_t_470, 0, __pyx_int_8);
  __Pyx_GIVEREF(__pyx_t_472);
  PyTuple_SET_ITEM(__pyx_t_470, 1, __pyx_t_472);
  __Pyx_INCREF(__pyx_int_2016);
  __Pyx_GIVEREF(__pyx_int_2016);
  PyTuple_SET_ITEM(__pyx_t_470, 2, __pyx_int_2016);
  __pyx_t_472 = 0;
  __pyx_t_472 = __Pyx_PyObject_Call(__pyx_t_471, __pyx_t_470, NULL); if (unlikely(!__pyx_t_472)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_472);
  __Pyx_DECREF(__pyx_t_471); __pyx_t_471 = 0;
  __Pyx_DECREF(__pyx_t_470); __pyx_t_470 = 0;

  /* "ultron/tradingday/Calendar.pyx":845
 *     Date.westernStyle(18, Months.September, 2016),
 *     Date.westernStyle(8, Months.October, 2016),
 *     Date.westernStyle(9, Months.October, 2016),             # <<<<<<<<<<<<<<
 *     # 2017
 *     Date.westernStyle(22, Months.January, 2017),
 */
  __pyx_t_470 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_470)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_470);
  __pyx_t_471 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_471)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_471);
  __pyx_t_469 = PyTuple_New(3); if (unlikely(!__pyx_t_469)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_469);
  __Pyx_INCREF(__pyx_int_9);
  __Pyx_GIVEREF(__pyx_int_9);
  PyTuple_SET_ITEM(__pyx_t_469, 0, __pyx_int_9);
  __Pyx_GIVEREF(__pyx_t_471);
  PyTuple_SET_ITEM(__pyx_t_469, 1, __pyx_t_471);
  __Pyx_INCREF(__pyx_int_2016);
  __Pyx_GIVEREF(__pyx_int_2016);
  PyTuple_SET_ITEM(__pyx_t_469, 2, __pyx_int_2016);
  __pyx_t_471 = 0;
  __pyx_t_471 = __Pyx_PyObject_Call(__pyx_t_470, __pyx_t_469, NULL); if (unlikely(!__pyx_t_471)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_471);
  __Pyx_DECREF(__pyx_t_470); __pyx_t_470 = 0;
  __Pyx_DECREF(__pyx_t_469); __pyx_t_469 = 0;

  /* "ultron/tradingday/Calendar.pyx":847
 *     Date.westernStyle(9, Months.October, 2016),
 *     # 2017
 *     Date.westernStyle(22, Months.January, 2017),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(4, Months.February, 2017),
 *     Date.westernStyle(1, Months.April, 2017),
 */
  __pyx_t_469 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_469)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_469);
  __pyx_t_470 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_470)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_470);
  __pyx_t_468 = PyTuple_New(3); if (unlikely(!__pyx_t_468)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_468);
  __Pyx_INCREF(__pyx_int_22);
  __Pyx_GIVEREF(__pyx_int_22);
  PyTuple_SET_ITEM(__pyx_t_468, 0, __pyx_int_22);
  __Pyx_GIVEREF(__pyx_t_470);
  PyTuple_SET_ITEM(__pyx_t_468, 1, __pyx_t_470);
  __Pyx_INCREF(__pyx_int_2017);
  __Pyx_GIVEREF(__pyx_int_2017);
  PyTuple_SET_ITEM(__pyx_t_468, 2, __pyx_int_2017);
  __pyx_t_470 = 0;
  __pyx_t_470 = __Pyx_PyObject_Call(__pyx_t_469, __pyx_t_468, NULL); if (unlikely(!__pyx_t_470)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_470);
  __Pyx_DECREF(__pyx_t_469); __pyx_t_469 = 0;
  __Pyx_DECREF(__pyx_t_468); __pyx_t_468 = 0;

  /* "ultron/tradingday/Calendar.pyx":848
 *     # 2017
 *     Date.westernStyle(22, Months.January, 2017),
 *     Date.westernStyle(4, Months.February, 2017),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(1, Months.April, 2017),
 *     Date.westernStyle(27, Months.May, 2017),
 */
  __pyx_t_468 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_468)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_468);
  __pyx_t_469 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_469)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_469);
  __pyx_t_467 = PyTuple_New(3); if (unlikely(!__pyx_t_467)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_467);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_467, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_t_469);
  PyTuple_SET_ITEM(__pyx_t_467, 1, __pyx_t_469);
  __Pyx_INCREF(__pyx_int_2017);
  __Pyx_GIVEREF(__pyx_int_2017);
  PyTuple_SET_ITEM(__pyx_t_467, 2, __pyx_int_2017);
  __pyx_t_469 = 0;
  __pyx_t_469 = __Pyx_PyObject_Call(__pyx_t_468, __pyx_t_467, NULL); if (unlikely(!__pyx_t_469)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_469);
  __Pyx_DECREF(__pyx_t_468); __pyx_t_468 = 0;
  __Pyx_DECREF(__pyx_t_467); __pyx_t_467 = 0;

  /* "ultron/tradingday/Calendar.pyx":849
 *     Date.westernStyle(22, Months.January, 2017),
 *     Date.westernStyle(4, Months.February, 2017),
 *     Date.westernStyle(1, Months.April, 2017),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(27, Months.May, 2017),
 *     Date.westernStyle(30, Months.September, 2017),
 */
  __pyx_t_467 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_467)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_467);
  __pyx_t_468 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_468)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_468);
  __pyx_t_466 = PyTuple_New(3); if (unlikely(!__pyx_t_466)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_466);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_t_466, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_t_468);
  PyTuple_SET_ITEM(__pyx_t_466, 1, __pyx_t_468);
  __Pyx_INCREF(__pyx_int_2017);
  __Pyx_GIVEREF(__pyx_int_2017);
  PyTuple_SET_ITEM(__pyx_t_466, 2, __pyx_int_2017);
  __pyx_t_468 = 0;
  __pyx_t_468 = __Pyx_PyObject_Call(__pyx_t_467, __pyx_t_466, NULL); if (unlikely(!__pyx_t_468)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_468);
  __Pyx_DECREF(__pyx_t_467); __pyx_t_467 = 0;
  __Pyx_DECREF(__pyx_t_466); __pyx_t_466 = 0;

  /* "ultron/tradingday/Calendar.pyx":850
 *     Date.westernStyle(4, Months.February, 2017),
 *     Date.westernStyle(1, Months.April, 2017),
 *     Date.westernStyle(27, Months.May, 2017),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(30, Months.September, 2017),
 *     # 2018
 */
  __pyx_t_466 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_466)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_466);
  __pyx_t_467 = __Pyx_PyInt_From_enum__Months(May); if (unlikely(!__pyx_t_467)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_467);
  __pyx_t_465 = PyTuple_New(3); if (unlikely(!__pyx_t_465)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_465);
  __Pyx_INCREF(__pyx_int_27);
  __Pyx_GIVEREF(__pyx_int_27);
  PyTuple_SET_ITEM(__pyx_t_465, 0, __pyx_int_27);
  __Pyx_GIVEREF(__pyx_t_467);
  PyTuple_SET_ITEM(__pyx_t_465, 1, __pyx_t_467);
  __Pyx_INCREF(__pyx_int_2017);
  __Pyx_GIVEREF(__pyx_int_2017);
  PyTuple_SET_ITEM(__pyx_t_465, 2, __pyx_int_2017);
  __pyx_t_467 = 0;
  __pyx_t_467 = __Pyx_PyObject_Call(__pyx_t_466, __pyx_t_465, NULL); if (unlikely(!__pyx_t_467)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_467);
  __Pyx_DECREF(__pyx_t_466); __pyx_t_466 = 0;
  __Pyx_DECREF(__pyx_t_465); __pyx_t_465 = 0;

  /* "ultron/tradingday/Calendar.pyx":851
 *     Date.westernStyle(1, Months.April, 2017),
 *     Date.westernStyle(27, Months.May, 2017),
 *     Date.westernStyle(30, Months.September, 2017),             # <<<<<<<<<<<<<<
 *     # 2018
 *     Date.westernStyle(11, Months.February, 2018),
 */
  __pyx_t_465 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_465)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_465);
  __pyx_t_466 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_466)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_466);
  __pyx_t_464 = PyTuple_New(3); if (unlikely(!__pyx_t_464)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_464);
  __Pyx_INCREF(__pyx_int_30);
  __Pyx_GIVEREF(__pyx_int_30);
  PyTuple_SET_ITEM(__pyx_t_464, 0, __pyx_int_30);
  __Pyx_GIVEREF(__pyx_t_466);
  PyTuple_SET_ITEM(__pyx_t_464, 1, __pyx_t_466);
  __Pyx_INCREF(__pyx_int_2017);
  __Pyx_GIVEREF(__pyx_int_2017);
  PyTuple_SET_ITEM(__pyx_t_464, 2, __pyx_int_2017);
  __pyx_t_466 = 0;
  __pyx_t_466 = __Pyx_PyObject_Call(__pyx_t_465, __pyx_t_464, NULL); if (unlikely(!__pyx_t_466)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_466);
  __Pyx_DECREF(__pyx_t_465); __pyx_t_465 = 0;
  __Pyx_DECREF(__pyx_t_464); __pyx_t_464 = 0;

  /* "ultron/tradingday/Calendar.pyx":853
 *     Date.westernStyle(30, Months.September, 2017),
 *     # 2018
 *     Date.westernStyle(11, Months.February, 2018),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(24, Months.February, 2018),
 *     Date.westernStyle(8, Months.April, 2018),
 */
  __pyx_t_464 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_464)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_464);
  __pyx_t_465 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_465)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_465);
  __pyx_t_463 = PyTuple_New(3); if (unlikely(!__pyx_t_463)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_463);
  __Pyx_INCREF(__pyx_int_11);
  __Pyx_GIVEREF(__pyx_int_11);
  PyTuple_SET_ITEM(__pyx_t_463, 0, __pyx_int_11);
  __Pyx_GIVEREF(__pyx_t_465);
  PyTuple_SET_ITEM(__pyx_t_463, 1, __pyx_t_465);
  __Pyx_INCREF(__pyx_int_2018);
  __Pyx_GIVEREF(__pyx_int_2018);
  PyTuple_SET_ITEM(__pyx_t_463, 2, __pyx_int_2018);
  __pyx_t_465 = 0;
  __pyx_t_465 = __Pyx_PyObject_Call(__pyx_t_464, __pyx_t_463, NULL); if (unlikely(!__pyx_t_465)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_465);
  __Pyx_DECREF(__pyx_t_464); __pyx_t_464 = 0;
  __Pyx_DECREF(__pyx_t_463); __pyx_t_463 = 0;

  /* "ultron/tradingday/Calendar.pyx":854
 *     # 2018
 *     Date.westernStyle(11, Months.February, 2018),
 *     Date.westernStyle(24, Months.February, 2018),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(8, Months.April, 2018),
 *     Date.westernStyle(28, Months.April, 2018),
 */
  __pyx_t_463 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_463)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_463);
  __pyx_t_464 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_464)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_464);
  __pyx_t_462 = PyTuple_New(3); if (unlikely(!__pyx_t_462)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_462);
  __Pyx_INCREF(__pyx_int_24);
  __Pyx_GIVEREF(__pyx_int_24);
  PyTuple_SET_ITEM(__pyx_t_462, 0, __pyx_int_24);
  __Pyx_GIVEREF(__pyx_t_464);
  PyTuple_SET_ITEM(__pyx_t_462, 1, __pyx_t_464);
  __Pyx_INCREF(__pyx_int_2018);
  __Pyx_GIVEREF(__pyx_int_2018);
  PyTuple_SET_ITEM(__pyx_t_462, 2, __pyx_int_2018);
  __pyx_t_464 = 0;
  __pyx_t_464 = __Pyx_PyObject_Call(__pyx_t_463, __pyx_t_462, NULL); if (unlikely(!__pyx_t_464)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_464);
  __Pyx_DECREF(__pyx_t_463); __pyx_t_463 = 0;
  __Pyx_DECREF(__pyx_t_462); __pyx_t_462 = 0;

  /* "ultron/tradingday/Calendar.pyx":855
 *     Date.westernStyle(11, Months.February, 2018),
 *     Date.westernStyle(24, Months.February, 2018),
 *     Date.westernStyle(8, Months.April, 2018),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(28, Months.April, 2018),
 *     Date.westernStyle(29, Months.September, 2018),
 */
  __pyx_t_462 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_462)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_462);
  __pyx_t_463 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_463)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_463);
  __pyx_t_461 = PyTuple_New(3); if (unlikely(!__pyx_t_461)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_461);
  __Pyx_INCREF(__pyx_int_8);
  __Pyx_GIVEREF(__pyx_int_8);
  PyTuple_SET_ITEM(__pyx_t_461, 0, __pyx_int_8);
  __Pyx_GIVEREF(__pyx_t_463);
  PyTuple_SET_ITEM(__pyx_t_461, 1, __pyx_t_463);
  __Pyx_INCREF(__pyx_int_2018);
  __Pyx_GIVEREF(__pyx_int_2018);
  PyTuple_SET_ITEM(__pyx_t_461, 2, __pyx_int_2018);
  __pyx_t_463 = 0;
  __pyx_t_463 = __Pyx_PyObject_Call(__pyx_t_462, __pyx_t_461, NULL); if (unlikely(!__pyx_t_463)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_463);
  __Pyx_DECREF(__pyx_t_462); __pyx_t_462 = 0;
  __Pyx_DECREF(__pyx_t_461); __pyx_t_461 = 0;

  /* "ultron/tradingday/Calendar.pyx":856
 *     Date.westernStyle(24, Months.February, 2018),
 *     Date.westernStyle(8, Months.April, 2018),
 *     Date.westernStyle(28, Months.April, 2018),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.September, 2018),
 *     Date.westernStyle(30, Months.September, 2018),
 */
  __pyx_t_461 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_461)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_461);
  __pyx_t_462 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_462)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_462);
  __pyx_t_460 = PyTuple_New(3); if (unlikely(!__pyx_t_460)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_460);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_460, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_462);
  PyTuple_SET_ITEM(__pyx_t_460, 1, __pyx_t_462);
  __Pyx_INCREF(__pyx_int_2018);
  __Pyx_GIVEREF(__pyx_int_2018);
  PyTuple_SET_ITEM(__pyx_t_460, 2, __pyx_int_2018);
  __pyx_t_462 = 0;
  __pyx_t_462 = __Pyx_PyObject_Call(__pyx_t_461, __pyx_t_460, NULL); if (unlikely(!__pyx_t_462)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_462);
  __Pyx_DECREF(__pyx_t_461); __pyx_t_461 = 0;
  __Pyx_DECREF(__pyx_t_460); __pyx_t_460 = 0;

  /* "ultron/tradingday/Calendar.pyx":857
 *     Date.westernStyle(8, Months.April, 2018),
 *     Date.westernStyle(28, Months.April, 2018),
 *     Date.westernStyle(29, Months.September, 2018),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(30, Months.September, 2018),
 *     Date.westernStyle(29, Months.December, 2018),
 */
  __pyx_t_460 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_460)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_460);
  __pyx_t_461 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_461)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_461);
  __pyx_t_459 = PyTuple_New(3); if (unlikely(!__pyx_t_459)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_459);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_459, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_461);
  PyTuple_SET_ITEM(__pyx_t_459, 1, __pyx_t_461);
  __Pyx_INCREF(__pyx_int_2018);
  __Pyx_GIVEREF(__pyx_int_2018);
  PyTuple_SET_ITEM(__pyx_t_459, 2, __pyx_int_2018);
  __pyx_t_461 = 0;
  __pyx_t_461 = __Pyx_PyObject_Call(__pyx_t_460, __pyx_t_459, NULL); if (unlikely(!__pyx_t_461)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_461);
  __Pyx_DECREF(__pyx_t_460); __pyx_t_460 = 0;
  __Pyx_DECREF(__pyx_t_459); __pyx_t_459 = 0;

  /* "ultron/tradingday/Calendar.pyx":858
 *     Date.westernStyle(28, Months.April, 2018),
 *     Date.westernStyle(29, Months.September, 2018),
 *     Date.westernStyle(30, Months.September, 2018),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.December, 2018),
 *     # 2019
 */
  __pyx_t_459 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_459)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_459);
  __pyx_t_460 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_460)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_460);
  __pyx_t_458 = PyTuple_New(3); if (unlikely(!__pyx_t_458)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_458);
  __Pyx_INCREF(__pyx_int_30);
  __Pyx_GIVEREF(__pyx_int_30);
  PyTuple_SET_ITEM(__pyx_t_458, 0, __pyx_int_30);
  __Pyx_GIVEREF(__pyx_t_460);
  PyTuple_SET_ITEM(__pyx_t_458, 1, __pyx_t_460);
  __Pyx_INCREF(__pyx_int_2018);
  __Pyx_GIVEREF(__pyx_int_2018);
  PyTuple_SET_ITEM(__pyx_t_458, 2, __pyx_int_2018);
  __pyx_t_460 = 0;
  __pyx_t_460 = __Pyx_PyObject_Call(__pyx_t_459, __pyx_t_458, NULL); if (unlikely(!__pyx_t_460)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_460);
  __Pyx_DECREF(__pyx_t_459); __pyx_t_459 = 0;
  __Pyx_DECREF(__pyx_t_458); __pyx_t_458 = 0;

  /* "ultron/tradingday/Calendar.pyx":859
 *     Date.westernStyle(29, Months.September, 2018),
 *     Date.westernStyle(30, Months.September, 2018),
 *     Date.westernStyle(29, Months.December, 2018),             # <<<<<<<<<<<<<<
 *     # 2019
 *     Date.westernStyle(2, Months.February, 2019),
 */
  __pyx_t_458 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_458)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_458);
  __pyx_t_459 = __Pyx_PyInt_From_enum__Months(December); if (unlikely(!__pyx_t_459)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_459);
  __pyx_t_457 = PyTuple_New(3); if (unlikely(!__pyx_t_457)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_457);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_457, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_459);
  PyTuple_SET_ITEM(__pyx_t_457, 1, __pyx_t_459);
  __Pyx_INCREF(__pyx_int_2018);
  __Pyx_GIVEREF(__pyx_int_2018);
  PyTuple_SET_ITEM(__pyx_t_457, 2, __pyx_int_2018);
  __pyx_t_459 = 0;
  __pyx_t_459 = __Pyx_PyObject_Call(__pyx_t_458, __pyx_t_457, NULL); if (unlikely(!__pyx_t_459)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_459);
  __Pyx_DECREF(__pyx_t_458); __pyx_t_458 = 0;
  __Pyx_DECREF(__pyx_t_457); __pyx_t_457 = 0;

  /* "ultron/tradingday/Calendar.pyx":861
 *     Date.westernStyle(29, Months.December, 2018),
 *     # 2019
 *     Date.westernStyle(2, Months.February, 2019),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(3, Months.February, 2019),
 *     Date.westernStyle(28, Months.April, 2019),
 */
  __pyx_t_457 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_457)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_457);
  __pyx_t_458 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_458)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_458);
  __pyx_t_456 = PyTuple_New(3); if (unlikely(!__pyx_t_456)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_456);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_456, 0, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_t_458);
  PyTuple_SET_ITEM(__pyx_t_456, 1, __pyx_t_458);
  __Pyx_INCREF(__pyx_int_2019);
  __Pyx_GIVEREF(__pyx_int_2019);
  PyTuple_SET_ITEM(__pyx_t_456, 2, __pyx_int_2019);
  __pyx_t_458 = 0;
  __pyx_t_458 = __Pyx_PyObject_Call(__pyx_t_457, __pyx_t_456, NULL); if (unlikely(!__pyx_t_458)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_458);
  __Pyx_DECREF(__pyx_t_457); __pyx_t_457 = 0;
  __Pyx_DECREF(__pyx_t_456); __pyx_t_456 = 0;

  /* "ultron/tradingday/Calendar.pyx":862
 *     # 2019
 *     Date.westernStyle(2, Months.February, 2019),
 *     Date.westernStyle(3, Months.February, 2019),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(28, Months.April, 2019),
 *     Date.westernStyle(5, Months.May, 2019),
 */
  __pyx_t_456 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_456)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_456);
  __pyx_t_457 = __Pyx_PyInt_From_enum__Months(February); if (unlikely(!__pyx_t_457)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_457);
  __pyx_t_455 = PyTuple_New(3); if (unlikely(!__pyx_t_455)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_455);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_455, 0, __pyx_int_3);
  __Pyx_GIVEREF(__pyx_t_457);
  PyTuple_SET_ITEM(__pyx_t_455, 1, __pyx_t_457);
  __Pyx_INCREF(__pyx_int_2019);
  __Pyx_GIVEREF(__pyx_int_2019);
  PyTuple_SET_ITEM(__pyx_t_455, 2, __pyx_int_2019);
  __pyx_t_457 = 0;
  __pyx_t_457 = __Pyx_PyObject_Call(__pyx_t_456, __pyx_t_455, NULL); if (unlikely(!__pyx_t_457)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_457);
  __Pyx_DECREF(__pyx_t_456); __pyx_t_456 = 0;
  __Pyx_DECREF(__pyx_t_455); __pyx_t_455 = 0;

  /* "ultron/tradingday/Calendar.pyx":863
 *     Date.westernStyle(2, Months.February, 2019),
 *     Date.westernStyle(3, Months.February, 2019),
 *     Date.westernStyle(28, Months.April, 2019),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(5, Months.May, 2019),
 *     Date.westernStyle(29, Months.September, 2019),
 */
  __pyx_t_455 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_455)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_455);
  __pyx_t_456 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_456)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_456);
  __pyx_t_454 = PyTuple_New(3); if (unlikely(!__pyx_t_454)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_454);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_454, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_456);
  PyTuple_SET_ITEM(__pyx_t_454, 1, __pyx_t_456);
  __Pyx_INCREF(__pyx_int_2019);
  __Pyx_GIVEREF(__pyx_int_2019);
  PyTuple_SET_ITEM(__pyx_t_454, 2, __pyx_int_2019);
  __pyx_t_456 = 0;
  __pyx_t_456 = __Pyx_PyObject_Call(__pyx_t_455, __pyx_t_454, NULL); if (unlikely(!__pyx_t_456)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_456);
  __Pyx_DECREF(__pyx_t_455); __pyx_t_455 = 0;
  __Pyx_DECREF(__pyx_t_454); __pyx_t_454 = 0;

  /* "ultron/tradingday/Calendar.pyx":864
 *     Date.westernStyle(3, Months.February, 2019),
 *     Date.westernStyle(28, Months.April, 2019),
 *     Date.westernStyle(5, Months.May, 2019),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(29, Months.September, 2019),
 *     Date.westernStyle(12, Months.October, 2019),
 */
  __pyx_t_454 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_454)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_454);
  __pyx_t_455 = __Pyx_PyInt_From_enum__Months(May); if (unlikely(!__pyx_t_455)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_455);
  __pyx_t_453 = PyTuple_New(3); if (unlikely(!__pyx_t_453)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_453);
  __Pyx_INCREF(__pyx_int_5);
  __Pyx_GIVEREF(__pyx_int_5);
  PyTuple_SET_ITEM(__pyx_t_453, 0, __pyx_int_5);
  __Pyx_GIVEREF(__pyx_t_455);
  PyTuple_SET_ITEM(__pyx_t_453, 1, __pyx_t_455);
  __Pyx_INCREF(__pyx_int_2019);
  __Pyx_GIVEREF(__pyx_int_2019);
  PyTuple_SET_ITEM(__pyx_t_453, 2, __pyx_int_2019);
  __pyx_t_455 = 0;
  __pyx_t_455 = __Pyx_PyObject_Call(__pyx_t_454, __pyx_t_453, NULL); if (unlikely(!__pyx_t_455)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_455);
  __Pyx_DECREF(__pyx_t_454); __pyx_t_454 = 0;
  __Pyx_DECREF(__pyx_t_453); __pyx_t_453 = 0;

  /* "ultron/tradingday/Calendar.pyx":865
 *     Date.westernStyle(28, Months.April, 2019),
 *     Date.westernStyle(5, Months.May, 2019),
 *     Date.westernStyle(29, Months.September, 2019),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(12, Months.October, 2019),
 *     # 2020
 */
  __pyx_t_453 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_453)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_453);
  __pyx_t_454 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_454)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_454);
  __pyx_t_452 = PyTuple_New(3); if (unlikely(!__pyx_t_452)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_452);
  __Pyx_INCREF(__pyx_int_29);
  __Pyx_GIVEREF(__pyx_int_29);
  PyTuple_SET_ITEM(__pyx_t_452, 0, __pyx_int_29);
  __Pyx_GIVEREF(__pyx_t_454);
  PyTuple_SET_ITEM(__pyx_t_452, 1, __pyx_t_454);
  __Pyx_INCREF(__pyx_int_2019);
  __Pyx_GIVEREF(__pyx_int_2019);
  PyTuple_SET_ITEM(__pyx_t_452, 2, __pyx_int_2019);
  __pyx_t_454 = 0;
  __pyx_t_454 = __Pyx_PyObject_Call(__pyx_t_453, __pyx_t_452, NULL); if (unlikely(!__pyx_t_454)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_454);
  __Pyx_DECREF(__pyx_t_453); __pyx_t_453 = 0;
  __Pyx_DECREF(__pyx_t_452); __pyx_t_452 = 0;

  /* "ultron/tradingday/Calendar.pyx":866
 *     Date.westernStyle(5, Months.May, 2019),
 *     Date.westernStyle(29, Months.September, 2019),
 *     Date.westernStyle(12, Months.October, 2019),             # <<<<<<<<<<<<<<
 *     # 2020
 *     Date.westernStyle(19, Months.January, 2020),
 */
  __pyx_t_452 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_452)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_452);
  __pyx_t_453 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_453)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_453);
  __pyx_t_451 = PyTuple_New(3); if (unlikely(!__pyx_t_451)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_451);
  __Pyx_INCREF(__pyx_int_12);
  __Pyx_GIVEREF(__pyx_int_12);
  PyTuple_SET_ITEM(__pyx_t_451, 0, __pyx_int_12);
  __Pyx_GIVEREF(__pyx_t_453);
  PyTuple_SET_ITEM(__pyx_t_451, 1, __pyx_t_453);
  __Pyx_INCREF(__pyx_int_2019);
  __Pyx_GIVEREF(__pyx_int_2019);
  PyTuple_SET_ITEM(__pyx_t_451, 2, __pyx_int_2019);
  __pyx_t_453 = 0;
  __pyx_t_453 = __Pyx_PyObject_Call(__pyx_t_452, __pyx_t_451, NULL); if (unlikely(!__pyx_t_453)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_453);
  __Pyx_DECREF(__pyx_t_452); __pyx_t_452 = 0;
  __Pyx_DECREF(__pyx_t_451); __pyx_t_451 = 0;

  /* "ultron/tradingday/Calendar.pyx":868
 *     Date.westernStyle(12, Months.October, 2019),
 *     # 2020
 *     Date.westernStyle(19, Months.January, 2020),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(26, Months.April, 2020),
 *     Date.westernStyle(9, Months.May, 2020),
 */
  __pyx_t_451 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_451)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_451);
  __pyx_t_452 = __Pyx_PyInt_From_enum__Months(January); if (unlikely(!__pyx_t_452)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_452);
  __pyx_t_450 = PyTuple_New(3); if (unlikely(!__pyx_t_450)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_450);
  __Pyx_INCREF(__pyx_int_19);
  __Pyx_GIVEREF(__pyx_int_19);
  PyTuple_SET_ITEM(__pyx_t_450, 0, __pyx_int_19);
  __Pyx_GIVEREF(__pyx_t_452);
  PyTuple_SET_ITEM(__pyx_t_450, 1, __pyx_t_452);
  __Pyx_INCREF(__pyx_int_2020);
  __Pyx_GIVEREF(__pyx_int_2020);
  PyTuple_SET_ITEM(__pyx_t_450, 2, __pyx_int_2020);
  __pyx_t_452 = 0;
  __pyx_t_452 = __Pyx_PyObject_Call(__pyx_t_451, __pyx_t_450, NULL); if (unlikely(!__pyx_t_452)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_452);
  __Pyx_DECREF(__pyx_t_451); __pyx_t_451 = 0;
  __Pyx_DECREF(__pyx_t_450); __pyx_t_450 = 0;

  /* "ultron/tradingday/Calendar.pyx":869
 *     # 2020
 *     Date.westernStyle(19, Months.January, 2020),
 *     Date.westernStyle(26, Months.April, 2020),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(9, Months.May, 2020),
 *     Date.westernStyle(28, Months.June, 2020),
 */
  __pyx_t_450 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_450)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_450);
  __pyx_t_451 = __Pyx_PyInt_From_enum__Months(April); if (unlikely(!__pyx_t_451)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_451);
  __pyx_t_449 = PyTuple_New(3); if (unlikely(!__pyx_t_449)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_449);
  __Pyx_INCREF(__pyx_int_26);
  __Pyx_GIVEREF(__pyx_int_26);
  PyTuple_SET_ITEM(__pyx_t_449, 0, __pyx_int_26);
  __Pyx_GIVEREF(__pyx_t_451);
  PyTuple_SET_ITEM(__pyx_t_449, 1, __pyx_t_451);
  __Pyx_INCREF(__pyx_int_2020);
  __Pyx_GIVEREF(__pyx_int_2020);
  PyTuple_SET_ITEM(__pyx_t_449, 2, __pyx_int_2020);
  __pyx_t_451 = 0;
  __pyx_t_451 = __Pyx_PyObject_Call(__pyx_t_450, __pyx_t_449, NULL); if (unlikely(!__pyx_t_451)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_451);
  __Pyx_DECREF(__pyx_t_450); __pyx_t_450 = 0;
  __Pyx_DECREF(__pyx_t_449); __pyx_t_449 = 0;

  /* "ultron/tradingday/Calendar.pyx":870
 *     Date.westernStyle(19, Months.January, 2020),
 *     Date.westernStyle(26, Months.April, 2020),
 *     Date.westernStyle(9, Months.May, 2020),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(28, Months.June, 2020),
 *     Date.westernStyle(27, Months.September, 2020),
 */
  __pyx_t_449 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_449)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_449);
  __pyx_t_450 = __Pyx_PyInt_From_enum__Months(May); if (unlikely(!__pyx_t_450)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_450);
  __pyx_t_448 = PyTuple_New(3); if (unlikely(!__pyx_t_448)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_448);
  __Pyx_INCREF(__pyx_int_9);
  __Pyx_GIVEREF(__pyx_int_9);
  PyTuple_SET_ITEM(__pyx_t_448, 0, __pyx_int_9);
  __Pyx_GIVEREF(__pyx_t_450);
  PyTuple_SET_ITEM(__pyx_t_448, 1, __pyx_t_450);
  __Pyx_INCREF(__pyx_int_2020);
  __Pyx_GIVEREF(__pyx_int_2020);
  PyTuple_SET_ITEM(__pyx_t_448, 2, __pyx_int_2020);
  __pyx_t_450 = 0;
  __pyx_t_450 = __Pyx_PyObject_Call(__pyx_t_449, __pyx_t_448, NULL); if (unlikely(!__pyx_t_450)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_450);
  __Pyx_DECREF(__pyx_t_449); __pyx_t_449 = 0;
  __Pyx_DECREF(__pyx_t_448); __pyx_t_448 = 0;

  /* "ultron/tradingday/Calendar.pyx":871
 *     Date.westernStyle(26, Months.April, 2020),
 *     Date.westernStyle(9, Months.May, 2020),
 *     Date.westernStyle(28, Months.June, 2020),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(27, Months.September, 2020),
 *     Date.westernStyle(10, Months.October, 2020),
 */
  __pyx_t_448 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_448)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_448);
  __pyx_t_449 = __Pyx_PyInt_From_enum__Months(June); if (unlikely(!__pyx_t_449)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_449);
  __pyx_t_447 = PyTuple_New(3); if (unlikely(!__pyx_t_447)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_447);
  __Pyx_INCREF(__pyx_int_28);
  __Pyx_GIVEREF(__pyx_int_28);
  PyTuple_SET_ITEM(__pyx_t_447, 0, __pyx_int_28);
  __Pyx_GIVEREF(__pyx_t_449);
  PyTuple_SET_ITEM(__pyx_t_447, 1, __pyx_t_449);
  __Pyx_INCREF(__pyx_int_2020);
  __Pyx_GIVEREF(__pyx_int_2020);
  PyTuple_SET_ITEM(__pyx_t_447, 2, __pyx_int_2020);
  __pyx_t_449 = 0;
  __pyx_t_449 = __Pyx_PyObject_Call(__pyx_t_448, __pyx_t_447, NULL); if (unlikely(!__pyx_t_449)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_449);
  __Pyx_DECREF(__pyx_t_448); __pyx_t_448 = 0;
  __Pyx_DECREF(__pyx_t_447); __pyx_t_447 = 0;

  /* "ultron/tradingday/Calendar.pyx":872
 *     Date.westernStyle(9, Months.May, 2020),
 *     Date.westernStyle(28, Months.June, 2020),
 *     Date.westernStyle(27, Months.September, 2020),             # <<<<<<<<<<<<<<
 *     Date.westernStyle(10, Months.October, 2020),
 *     # 2021
 */
  __pyx_t_447 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_447)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_447);
  __pyx_t_448 = __Pyx_PyInt_From_enum__Months(September); if (unlikely(!__pyx_t_448)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_448);
  __pyx_t_446 = PyTuple_New(3); if (unlikely(!__pyx_t_446)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_446);
  __Pyx_INCREF(__pyx_int_27);
  __Pyx_GIVEREF(__pyx_int_27);
  PyTuple_SET_ITEM(__pyx_t_446, 0, __pyx_int_27);
  __Pyx_GIVEREF(__pyx_t_448);
  PyTuple_SET_ITEM(__pyx_t_446, 1, __pyx_t_448);
  __Pyx_INCREF(__pyx_int_2020);
  __Pyx_GIVEREF(__pyx_int_2020);
  PyTuple_SET_ITEM(__pyx_t_446, 2, __pyx_int_2020);
  __pyx_t_448 = 0;
  __pyx_t_448 = __Pyx_PyObject_Call(__pyx_t_447, __pyx_t_446, NULL); if (unlikely(!__pyx_t_448)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_448);
  __Pyx_DECREF(__pyx_t_447); __pyx_t_447 = 0;
  __Pyx_DECREF(__pyx_t_446); __pyx_t_446 = 0;

  /* "ultron/tradingday/Calendar.pyx":873
 *     Date.westernStyle(28, Months.June, 2020),
 *     Date.westernStyle(27, Months.September, 2020),
 *     Date.westernStyle(10, Months.October, 2020),             # <<<<<<<<<<<<<<
 *     # 2021
 *     Date(2021, 2, 7),
 */
  __pyx_t_446 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_n_s_westernStyle); if (unlikely(!__pyx_t_446)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_446);
  __pyx_t_447 = __Pyx_PyInt_From_enum__Months(October); if (unlikely(!__pyx_t_447)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_447);
  __pyx_t_445 = PyTuple_New(3); if (unlikely(!__pyx_t_445)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_445);
  __Pyx_INCREF(__pyx_int_10);
  __Pyx_GIVEREF(__pyx_int_10);
  PyTuple_SET_ITEM(__pyx_t_445, 0, __pyx_int_10);
  __Pyx_GIVEREF(__pyx_t_447);
  PyTuple_SET_ITEM(__pyx_t_445, 1, __pyx_t_447);
  __Pyx_INCREF(__pyx_int_2020);
  __Pyx_GIVEREF(__pyx_int_2020);
  PyTuple_SET_ITEM(__pyx_t_445, 2, __pyx_int_2020);
  __pyx_t_447 = 0;
  __pyx_t_447 = __Pyx_PyObject_Call(__pyx_t_446, __pyx_t_445, NULL); if (unlikely(!__pyx_t_447)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_447);
  __Pyx_DECREF(__pyx_t_446); __pyx_t_446 = 0;
  __Pyx_DECREF(__pyx_t_445); __pyx_t_445 = 0;

  /* "ultron/tradingday/Calendar.pyx":875
 *     Date.westernStyle(10, Months.October, 2020),
 *     # 2021
 *     Date(2021, 2, 7),             # <<<<<<<<<<<<<<
 *     Date(2021, 2, 20),
 *     Date(2021, 4, 25),
 */
  __pyx_t_445 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__553, NULL); if (unlikely(!__pyx_t_445)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_445);

  /* "ultron/tradingday/Calendar.pyx":876
 *     # 2021
 *     Date(2021, 2, 7),
 *     Date(2021, 2, 20),             # <<<<<<<<<<<<<<
 *     Date(2021, 4, 25),
 *     Date(2021, 5, 8),
 */
  __pyx_t_446 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__554, NULL); if (unlikely(!__pyx_t_446)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_446);

  /* "ultron/tradingday/Calendar.pyx":877
 *     Date(2021, 2, 7),
 *     Date(2021, 2, 20),
 *     Date(2021, 4, 25),             # <<<<<<<<<<<<<<
 *     Date(2021, 5, 8),
 *     Date(2021, 9, 18),
 */
  __pyx_t_444 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__555, NULL); if (unlikely(!__pyx_t_444)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_444);

  /* "ultron/tradingday/Calendar.pyx":878
 *     Date(2021, 2, 20),
 *     Date(2021, 4, 25),
 *     Date(2021, 5, 8),             # <<<<<<<<<<<<<<
 *     Date(2021, 9, 18),
 *     Date(2021, 9, 26),
 */
  __pyx_t_443 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__556, NULL); if (unlikely(!__pyx_t_443)) __PYX_ERR(0, 878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_443);

  /* "ultron/tradingday/Calendar.pyx":879
 *     Date(2021, 4, 25),
 *     Date(2021, 5, 8),
 *     Date(2021, 9, 18),             # <<<<<<<<<<<<<<
 *     Date(2021, 9, 26),
 *     Date(2021, 10, 9),
 */
  __pyx_t_442 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__557, NULL); if (unlikely(!__pyx_t_442)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_442);

  /* "ultron/tradingday/Calendar.pyx":880
 *     Date(2021, 5, 8),
 *     Date(2021, 9, 18),
 *     Date(2021, 9, 26),             # <<<<<<<<<<<<<<
 *     Date(2021, 10, 9),
 *     # 2022
 */
  __pyx_t_441 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__558, NULL); if (unlikely(!__pyx_t_441)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_441);

  /* "ultron/tradingday/Calendar.pyx":881
 *     Date(2021, 9, 18),
 *     Date(2021, 9, 26),
 *     Date(2021, 10, 9),             # <<<<<<<<<<<<<<
 *     # 2022
 *     Date(2022, 1, 29),
 */
  __pyx_t_440 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__559, NULL); if (unlikely(!__pyx_t_440)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_440);

  /* "ultron/tradingday/Calendar.pyx":883
 *     Date(2021, 10, 9),
 *     # 2022
 *     Date(2022, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2022, 1, 30),
 *     Date(2022, 4, 2),
 */
  __pyx_t_439 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__560, NULL); if (unlikely(!__pyx_t_439)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_439);

  /* "ultron/tradingday/Calendar.pyx":884
 *     # 2022
 *     Date(2022, 1, 29),
 *     Date(2022, 1, 30),             # <<<<<<<<<<<<<<
 *     Date(2022, 4, 2),
 *     Date(2022, 4, 24),
 */
  __pyx_t_438 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__561, NULL); if (unlikely(!__pyx_t_438)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_438);

  /* "ultron/tradingday/Calendar.pyx":885
 *     Date(2022, 1, 29),
 *     Date(2022, 1, 30),
 *     Date(2022, 4, 2),             # <<<<<<<<<<<<<<
 *     Date(2022, 4, 24),
 *     Date(2022, 5, 7),
 */
  __pyx_t_437 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__562, NULL); if (unlikely(!__pyx_t_437)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_437);

  /* "ultron/tradingday/Calendar.pyx":886
 *     Date(2022, 1, 30),
 *     Date(2022, 4, 2),
 *     Date(2022, 4, 24),             # <<<<<<<<<<<<<<
 *     Date(2022, 5, 7),
 *     Date(2022, 10, 8),
 */
  __pyx_t_436 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__563, NULL); if (unlikely(!__pyx_t_436)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_436);

  /* "ultron/tradingday/Calendar.pyx":887
 *     Date(2022, 4, 2),
 *     Date(2022, 4, 24),
 *     Date(2022, 5, 7),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 8),
 *     Date(2022, 10, 9),
 */
  __pyx_t_435 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__564, NULL); if (unlikely(!__pyx_t_435)) __PYX_ERR(0, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_435);

  /* "ultron/tradingday/Calendar.pyx":888
 *     Date(2022, 4, 24),
 *     Date(2022, 5, 7),
 *     Date(2022, 10, 8),             # <<<<<<<<<<<<<<
 *     Date(2022, 10, 9),
 * 
 */
  __pyx_t_434 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__565, NULL); if (unlikely(!__pyx_t_434)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_434);

  /* "ultron/tradingday/Calendar.pyx":889
 *     Date(2022, 5, 7),
 *     Date(2022, 10, 8),
 *     Date(2022, 10, 9),             # <<<<<<<<<<<<<<
 * 
 *     # 2023
 */
  __pyx_t_433 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__566, NULL); if (unlikely(!__pyx_t_433)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_433);

  /* "ultron/tradingday/Calendar.pyx":892
 * 
 *     # 2023
 *     Date(2023, 1, 28),             # <<<<<<<<<<<<<<
 *     Date(2023, 1, 29),
 *     Date(2023, 4, 23),
 */
  __pyx_t_432 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__567, NULL); if (unlikely(!__pyx_t_432)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_432);

  /* "ultron/tradingday/Calendar.pyx":893
 *     # 2023
 *     Date(2023, 1, 28),
 *     Date(2023, 1, 29),             # <<<<<<<<<<<<<<
 *     Date(2023, 4, 23),
 *     Date(2023, 5, 6),
 */
  __pyx_t_431 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__568, NULL); if (unlikely(!__pyx_t_431)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_431);

  /* "ultron/tradingday/Calendar.pyx":894
 *     Date(2023, 1, 28),
 *     Date(2023, 1, 29),
 *     Date(2023, 4, 23),             # <<<<<<<<<<<<<<
 *     Date(2023, 5, 6),
 *     Date(2023, 6, 25),
 */
  __pyx_t_430 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__569, NULL); if (unlikely(!__pyx_t_430)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_430);

  /* "ultron/tradingday/Calendar.pyx":895
 *     Date(2023, 1, 29),
 *     Date(2023, 4, 23),
 *     Date(2023, 5, 6),             # <<<<<<<<<<<<<<
 *     Date(2023, 6, 25),
 *     Date(2023, 10, 7),
 */
  __pyx_t_429 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__570, NULL); if (unlikely(!__pyx_t_429)) __PYX_ERR(0, 895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_429);

  /* "ultron/tradingday/Calendar.pyx":896
 *     Date(2023, 4, 23),
 *     Date(2023, 5, 6),
 *     Date(2023, 6, 25),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 7),
 *     Date(2023, 10, 8)
 */
  __pyx_t_428 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__571, NULL); if (unlikely(!__pyx_t_428)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_428);

  /* "ultron/tradingday/Calendar.pyx":897
 *     Date(2023, 5, 6),
 *     Date(2023, 6, 25),
 *     Date(2023, 10, 7),             # <<<<<<<<<<<<<<
 *     Date(2023, 10, 8)
 * }
 */
  __pyx_t_427 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__572, NULL); if (unlikely(!__pyx_t_427)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_427);

  /* "ultron/tradingday/Calendar.pyx":898
 *     Date(2023, 6, 25),
 *     Date(2023, 10, 7),
 *     Date(2023, 10, 8)             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_t_426 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6ultron_10tradingday_4Date_Date), __pyx_tuple__573, NULL); if (unlikely(!__pyx_t_426)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_426);
  __pyx_t_425 = PySet_New(0); if (unlikely(!__pyx_t_425)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PySet_Add(__pyx_t_425, __pyx_t_549) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_549); __pyx_t_549 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_548) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_548); __pyx_t_548 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_550) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_550); __pyx_t_550 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_547) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_547); __pyx_t_547 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_546) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_546); __pyx_t_546 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_545) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_545); __pyx_t_545 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_544) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_544); __pyx_t_544 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_543) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_543); __pyx_t_543 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_542) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_542); __pyx_t_542 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_541) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_541); __pyx_t_541 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_540) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_540); __pyx_t_540 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_539) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_539); __pyx_t_539 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_538) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_538); __pyx_t_538 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_537) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_537); __pyx_t_537 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_536) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_536); __pyx_t_536 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_535) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_535); __pyx_t_535 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_534) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_534); __pyx_t_534 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_533) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_533); __pyx_t_533 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_532) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_532); __pyx_t_532 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_531) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_531); __pyx_t_531 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_530) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_530); __pyx_t_530 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_529) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_529); __pyx_t_529 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_528) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_528); __pyx_t_528 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_527) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_527); __pyx_t_527 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_526) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_526); __pyx_t_526 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_525) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_525); __pyx_t_525 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_524) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_524); __pyx_t_524 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_523) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_523); __pyx_t_523 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_522) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_522); __pyx_t_522 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_521) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_521); __pyx_t_521 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_520) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_520); __pyx_t_520 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_519) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_519); __pyx_t_519 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_518) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_518); __pyx_t_518 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_517) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_517); __pyx_t_517 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_516) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_516); __pyx_t_516 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_515) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_515); __pyx_t_515 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_514) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_514); __pyx_t_514 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_513) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_513); __pyx_t_513 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_512) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_512); __pyx_t_512 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_511) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_511); __pyx_t_511 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_510) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_510); __pyx_t_510 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_509) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_509); __pyx_t_509 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_508) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_508); __pyx_t_508 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_507) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_507); __pyx_t_507 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_506) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_506); __pyx_t_506 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_505) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_505); __pyx_t_505 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_504) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_504); __pyx_t_504 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_503) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_503); __pyx_t_503 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_502) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_502); __pyx_t_502 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_501) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_501); __pyx_t_501 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_500) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_500); __pyx_t_500 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_499) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_499); __pyx_t_499 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_498) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_498); __pyx_t_498 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_497) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_497); __pyx_t_497 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_496) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_496); __pyx_t_496 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_495) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_495); __pyx_t_495 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_494) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_494); __pyx_t_494 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_493) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_493); __pyx_t_493 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_492) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_492); __pyx_t_492 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_491) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_491); __pyx_t_491 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_490) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_490); __pyx_t_490 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_489) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_489); __pyx_t_489 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_488) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_488); __pyx_t_488 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_487) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_487); __pyx_t_487 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_486) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_486); __pyx_t_486 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_485) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_485); __pyx_t_485 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_484) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_484); __pyx_t_484 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_483) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_483); __pyx_t_483 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_482) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_482); __pyx_t_482 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_481) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_481); __pyx_t_481 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_480) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_480); __pyx_t_480 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_479) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_479); __pyx_t_479 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_478) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_478); __pyx_t_478 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_477) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_477); __pyx_t_477 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_476) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_476); __pyx_t_476 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_475) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_475); __pyx_t_475 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_474) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_474); __pyx_t_474 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_473) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_473); __pyx_t_473 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_472) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_472); __pyx_t_472 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_471) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_471); __pyx_t_471 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_470) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_470); __pyx_t_470 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_469) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_469); __pyx_t_469 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_468) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_468); __pyx_t_468 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_467) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_467); __pyx_t_467 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_466) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_466); __pyx_t_466 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_465) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_465); __pyx_t_465 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_464) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_464); __pyx_t_464 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_463) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_463); __pyx_t_463 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_462) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_462); __pyx_t_462 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_461) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_461); __pyx_t_461 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_460) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_460); __pyx_t_460 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_459) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_459); __pyx_t_459 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_458) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_458); __pyx_t_458 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_457) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_457); __pyx_t_457 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_456) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_456); __pyx_t_456 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_455) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_455); __pyx_t_455 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_454) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_454); __pyx_t_454 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_453) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_453); __pyx_t_453 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_452) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_452); __pyx_t_452 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_451) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_451); __pyx_t_451 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_450) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_450); __pyx_t_450 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_449) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_449); __pyx_t_449 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_448) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_448); __pyx_t_448 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_447) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_447); __pyx_t_447 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_445) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_445); __pyx_t_445 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_446) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_446); __pyx_t_446 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_444) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_444); __pyx_t_444 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_443) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_443); __pyx_t_443 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_442) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_442); __pyx_t_442 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_441) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_441); __pyx_t_441 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_440) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_440); __pyx_t_440 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_439) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_439); __pyx_t_439 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_438) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_438); __pyx_t_438 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_437) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_437); __pyx_t_437 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_436) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_436); __pyx_t_436 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_435) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_435); __pyx_t_435 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_434) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_434); __pyx_t_434 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_433) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_433); __pyx_t_433 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_432) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_432); __pyx_t_432 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_431) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_431); __pyx_t_431 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_430) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_430); __pyx_t_430 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_429) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_429); __pyx_t_429 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_428) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_428); __pyx_t_428 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_427) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_427); __pyx_t_427 = 0;
  if (PySet_Add(__pyx_t_425, __pyx_t_426) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_426); __pyx_t_426 = 0;
  __Pyx_XGOTREF(__pyx_v_6ultron_10tradingday_8Calendar_ib_working_weekends);
  __Pyx_DECREF_SET(__pyx_v_6ultron_10tradingday_8Calendar_ib_working_weekends, ((PyObject*)__pyx_t_425));
  __Pyx_GIVEREF(__pyx_t_425);
  __pyx_t_425 = 0;

  /* "ultron/tradingday/Calendar.pyx":901
 * }
 * 
 * cdef ChinaSseImpl _sseImpl = ChinaSseImpl()             # <<<<<<<<<<<<<<
 * 
 * cdef class ChinaIBImpl(CalendarImpl):
 */
  __pyx_t_425 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaSseImpl)); if (unlikely(!__pyx_t_425)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  __Pyx_XGOTREF(((PyObject *)__pyx_v_6ultron_10tradingday_8Calendar__sseImpl));
  __Pyx_DECREF_SET(__pyx_v_6ultron_10tradingday_8Calendar__sseImpl, ((struct __pyx_obj_6ultron_10tradingday_8Calendar_ChinaSseImpl *)__pyx_t_425));
  __Pyx_GIVEREF(__pyx_t_425);
  __pyx_t_425 = 0;

  /* "ultron/tradingday/Calendar.pyx":946
 * 
 * cdef int EasterMonday[299]
 * EasterMonday[:] = [             # <<<<<<<<<<<<<<
 *     98, 90, 103, 95, 114, 106, 91, 111, 102,  # 1901-1909
 *     87, 107, 99, 83, 103, 95, 115, 99, 91, 111,  # 1910-1919
 */
  __pyx_t_551[0] = 98;
  __pyx_t_551[1] = 90;
  __pyx_t_551[2] = 0x67;
  __pyx_t_551[3] = 95;
  __pyx_t_551[4] = 0x72;
  __pyx_t_551[5] = 0x6A;
  __pyx_t_551[6] = 91;
  __pyx_t_551[7] = 0x6F;
  __pyx_t_551[8] = 0x66;
  __pyx_t_551[9] = 87;
  __pyx_t_551[10] = 0x6B;
  __pyx_t_551[11] = 99;
  __pyx_t_551[12] = 83;
  __pyx_t_551[13] = 0x67;
  __pyx_t_551[14] = 95;
  __pyx_t_551[15] = 0x73;
  __pyx_t_551[16] = 99;
  __pyx_t_551[17] = 91;
  __pyx_t_551[18] = 0x6F;
  __pyx_t_551[19] = 96;
  __pyx_t_551[20] = 87;
  __pyx_t_551[21] = 0x6B;
  __pyx_t_551[22] = 92;
  __pyx_t_551[23] = 0x70;
  __pyx_t_551[24] = 0x67;
  __pyx_t_551[25] = 95;
  __pyx_t_551[26] = 0x6C;
  __pyx_t_551[27] = 0x64;
  __pyx_t_551[28] = 91;
  __pyx_t_551[29] = 0x6F;
  __pyx_t_551[30] = 96;
  __pyx_t_551[31] = 88;
  __pyx_t_551[32] = 0x6B;
  __pyx_t_551[33] = 92;
  __pyx_t_551[34] = 0x70;
  __pyx_t_551[35] = 0x68;
  __pyx_t_551[36] = 88;
  __pyx_t_551[37] = 0x6C;
  __pyx_t_551[38] = 0x64;
  __pyx_t_551[39] = 85;
  __pyx_t_551[40] = 0x68;
  __pyx_t_551[41] = 96;
  __pyx_t_551[42] = 0x74;
  __pyx_t_551[43] = 0x65;
  __pyx_t_551[44] = 92;
  __pyx_t_551[45] = 0x70;
  __pyx_t_551[46] = 97;
  __pyx_t_551[47] = 89;
  __pyx_t_551[48] = 0x6C;
  __pyx_t_551[49] = 0x64;
  __pyx_t_551[50] = 85;
  __pyx_t_551[51] = 0x69;
  __pyx_t_551[52] = 96;
  __pyx_t_551[53] = 0x6D;
  __pyx_t_551[54] = 0x65;
  __pyx_t_551[55] = 93;
  __pyx_t_551[56] = 0x70;
  __pyx_t_551[57] = 97;
  __pyx_t_551[58] = 89;
  __pyx_t_551[59] = 0x6D;
  __pyx_t_551[60] = 93;
  __pyx_t_551[61] = 0x71;
  __pyx_t_551[62] = 0x69;
  __pyx_t_551[63] = 90;
  __pyx_t_551[64] = 0x6D;
  __pyx_t_551[65] = 0x65;
  __pyx_t_551[66] = 86;
  __pyx_t_551[67] = 0x6A;
  __pyx_t_551[68] = 97;
  __pyx_t_551[69] = 89;
  __pyx_t_551[70] = 0x66;
  __pyx_t_551[71] = 94;
  __pyx_t_551[72] = 0x71;
  __pyx_t_551[73] = 0x69;
  __pyx_t_551[74] = 90;
  __pyx_t_551[75] = 0x6E;
  __pyx_t_551[76] = 0x65;
  __pyx_t_551[77] = 86;
  __pyx_t_551[78] = 0x6A;
  __pyx_t_551[79] = 98;
  __pyx_t_551[80] = 0x6E;
  __pyx_t_551[81] = 0x66;
  __pyx_t_551[82] = 94;
  __pyx_t_551[83] = 0x72;
  __pyx_t_551[84] = 98;
  __pyx_t_551[85] = 90;
  __pyx_t_551[86] = 0x6E;
  __pyx_t_551[87] = 95;
  __pyx_t_551[88] = 86;
  __pyx_t_551[89] = 0x6A;
  __pyx_t_551[90] = 91;
  __pyx_t_551[91] = 0x6F;
  __pyx_t_551[92] = 0x66;
  __pyx_t_551[93] = 94;
  __pyx_t_551[94] = 0x6B;
  __pyx_t_551[95] = 99;
  __pyx_t_551[96] = 90;
  __pyx_t_551[97] = 0x67;
  __pyx_t_551[98] = 95;
  __pyx_t_551[99] = 0x73;
  __pyx_t_551[100] = 0x6A;
  __pyx_t_551[101] = 91;
  __pyx_t_551[102] = 0x6F;
  __pyx_t_551[103] = 0x67;
  __pyx_t_551[104] = 87;
  __pyx_t_551[105] = 0x6B;
  __pyx_t_551[106] = 99;
  __pyx_t_551[107] = 84;
  __pyx_t_551[108] = 0x67;
  __pyx_t_551[109] = 95;
  __pyx_t_551[110] = 0x73;
  __pyx_t_551[111] = 0x64;
  __pyx_t_551[112] = 91;
  __pyx_t_551[113] = 0x6F;
  __pyx_t_551[114] = 96;
  __pyx_t_551[115] = 88;
  __pyx_t_551[116] = 0x6B;
  __pyx_t_551[117] = 92;
  __pyx_t_551[118] = 0x70;
  __pyx_t_551[119] = 0x68;
  __pyx_t_551[120] = 95;
  __pyx_t_551[121] = 0x6C;
  __pyx_t_551[122] = 0x64;
  __pyx_t_551[123] = 92;
  __pyx_t_551[124] = 0x6F;
  __pyx_t_551[125] = 96;
  __pyx_t_551[126] = 88;
  __pyx_t_551[127] = 0x6C;
  __pyx_t_551[128] = 92;
  __pyx_t_551[129] = 0x70;
  __pyx_t_551[130] = 0x68;
  __pyx_t_551[131] = 89;
  __pyx_t_551[132] = 0x6C;
  __pyx_t_551[133] = 0x64;
  __pyx_t_551[134] = 85;
  __pyx_t_551[135] = 0x69;
  __pyx_t_551[136] = 96;
  __pyx_t_551[137] = 0x74;
  __pyx_t_551[138] = 0x65;
  __pyx_t_551[139] = 93;
  __pyx_t_551[140] = 0x70;
  __pyx_t_551[141] = 97;
  __pyx_t_551[142] = 89;
  __pyx_t_551[143] = 0x6D;
  __pyx_t_551[144] = 0x64;
  __pyx_t_551[145] = 85;
  __pyx_t_551[146] = 0x69;
  __pyx_t_551[147] = 97;
  __pyx_t_551[148] = 0x6D;
  __pyx_t_551[149] = 0x65;
  __pyx_t_551[150] = 93;
  __pyx_t_551[151] = 0x71;
  __pyx_t_551[152] = 97;
  __pyx_t_551[153] = 89;
  __pyx_t_551[154] = 0x6D;
  __pyx_t_551[155] = 94;
  __pyx_t_551[156] = 0x71;
  __pyx_t_551[157] = 0x69;
  __pyx_t_551[158] = 90;
  __pyx_t_551[159] = 0x6E;
  __pyx_t_551[160] = 0x65;
  __pyx_t_551[161] = 86;
  __pyx_t_551[162] = 0x6A;
  __pyx_t_551[163] = 98;
  __pyx_t_551[164] = 89;
  __pyx_t_551[165] = 0x66;
  __pyx_t_551[166] = 94;
  __pyx_t_551[167] = 0x72;
  __pyx_t_551[168] = 0x69;
  __pyx_t_551[169] = 90;
  __pyx_t_551[170] = 0x6E;
  __pyx_t_551[171] = 0x66;
  __pyx_t_551[172] = 86;
  __pyx_t_551[173] = 0x6A;
  __pyx_t_551[174] = 98;
  __pyx_t_551[175] = 0x6F;
  __pyx_t_551[176] = 0x66;
  __pyx_t_551[177] = 94;
  __pyx_t_551[178] = 0x72;
  __pyx_t_551[179] = 99;
  __pyx_t_551[180] = 90;
  __pyx_t_551[181] = 0x6E;
  __pyx_t_551[182] = 95;
  __pyx_t_551[183] = 87;
  __pyx_t_551[184] = 0x6A;
  __pyx_t_551[185] = 91;
  __pyx_t_551[186] = 0x6F;
  __pyx_t_551[187] = 0x67;
  __pyx_t_551[188] = 94;
  __pyx_t_551[189] = 0x6B;
  __pyx_t_551[190] = 99;
  __pyx_t_551[191] = 91;
  __pyx_t_551[192] = 0x67;
  __pyx_t_551[193] = 95;
  __pyx_t_551[194] = 0x73;
  __pyx_t_551[195] = 0x6B;
  __pyx_t_551[196] = 91;
  __pyx_t_551[197] = 0x6F;
  __pyx_t_551[198] = 0x67;
  __pyx_t_551[199] = 88;
  __pyx_t_551[200] = 0x6C;
  __pyx_t_551[201] = 0x64;
  __pyx_t_551[202] = 85;
  __pyx_t_551[203] = 0x69;
  __pyx_t_551[204] = 96;
  __pyx_t_551[205] = 0x6D;
  __pyx_t_551[206] = 0x65;
  __pyx_t_551[207] = 93;
  __pyx_t_551[208] = 0x70;
  __pyx_t_551[209] = 97;
  __pyx_t_551[210] = 89;
  __pyx_t_551[211] = 0x6D;
  __pyx_t_551[212] = 93;
  __pyx_t_551[213] = 0x71;
  __pyx_t_551[214] = 0x69;
  __pyx_t_551[215] = 90;
  __pyx_t_551[216] = 0x6D;
  __pyx_t_551[217] = 0x65;
  __pyx_t_551[218] = 86;
  __pyx_t_551[219] = 0x6A;
  __pyx_t_551[220] = 97;
  __pyx_t_551[221] = 89;
  __pyx_t_551[222] = 0x66;
  __pyx_t_551[223] = 94;
  __pyx_t_551[224] = 0x71;
  __pyx_t_551[225] = 0x69;
  __pyx_t_551[226] = 90;
  __pyx_t_551[227] = 0x6E;
  __pyx_t_551[228] = 0x65;
  __pyx_t_551[229] = 86;
  __pyx_t_551[230] = 0x6A;
  __pyx_t_551[231] = 98;
  __pyx_t_551[232] = 0x6E;
  __pyx_t_551[233] = 0x66;
  __pyx_t_551[234] = 94;
  __pyx_t_551[235] = 0x72;
  __pyx_t_551[236] = 98;
  __pyx_t_551[237] = 90;
  __pyx_t_551[238] = 0x6E;
  __pyx_t_551[239] = 95;
  __pyx_t_551[240] = 86;
  __pyx_t_551[241] = 0x6A;
  __pyx_t_551[242] = 91;
  __pyx_t_551[243] = 0x6F;
  __pyx_t_551[244] = 0x66;
  __pyx_t_551[245] = 94;
  __pyx_t_551[246] = 0x6B;
  __pyx_t_551[247] = 99;
  __pyx_t_551[248] = 90;
  __pyx_t_551[249] = 0x67;
  __pyx_t_551[250] = 95;
  __pyx_t_551[251] = 0x73;
  __pyx_t_551[252] = 0x6A;
  __pyx_t_551[253] = 91;
  __pyx_t_551[254] = 0x6F;
  __pyx_t_551[255] = 0x67;
  __pyx_t_551[256] = 87;
  __pyx_t_551[257] = 0x6B;
  __pyx_t_551[258] = 99;
  __pyx_t_551[259] = 84;
  __pyx_t_551[260] = 0x67;
  __pyx_t_551[261] = 95;
  __pyx_t_551[262] = 0x73;
  __pyx_t_551[263] = 0x64;
  __pyx_t_551[264] = 91;
  __pyx_t_551[265] = 0x6F;
  __pyx_t_551[266] = 96;
  __pyx_t_551[267] = 88;
  __pyx_t_551[268] = 0x6B;
  __pyx_t_551[269] = 92;
  __pyx_t_551[270] = 0x70;
  __pyx_t_551[271] = 0x68;
  __pyx_t_551[272] = 95;
  __pyx_t_551[273] = 0x6C;
  __pyx_t_551[274] = 0x64;
  __pyx_t_551[275] = 92;
  __pyx_t_551[276] = 0x6F;
  __pyx_t_551[277] = 96;
  __pyx_t_551[278] = 88;
  __pyx_t_551[279] = 0x6C;
  __pyx_t_551[280] = 92;
  __pyx_t_551[281] = 0x70;
  __pyx_t_551[282] = 0x68;
  __pyx_t_551[283] = 89;
  __pyx_t_551[284] = 0x6C;
  __pyx_t_551[285] = 0x64;
  __pyx_t_551[286] = 85;
  __pyx_t_551[287] = 0x69;
  __pyx_t_551[288] = 96;
  __pyx_t_551[289] = 0x74;
  __pyx_t_551[290] = 0x65;
  __pyx_t_551[291] = 93;
  __pyx_t_551[292] = 0x70;
  __pyx_t_551[293] = 97;
  __pyx_t_551[294] = 89;
  __pyx_t_551[295] = 0x6D;
  __pyx_t_551[296] = 0x64;
  __pyx_t_551[297] = 85;
  __pyx_t_551[298] = 0x69;
  if (unlikely((0x12B) != (299))) {
    PyErr_Format(PyExc_ValueError, "Assignment to slice of wrong length, expected %" CYTHON_FORMAT_SSIZE_T "d, got %" CYTHON_FORMAT_SSIZE_T "d", (Py_ssize_t)(299), (Py_ssize_t)(0x12B));
    __PYX_ERR(0, 946, __pyx_L1_error)
  }
  memcpy(&(__pyx_v_6ultron_10tradingday_8Calendar_EasterMonday[0]), __pyx_t_551, sizeof(__pyx_v_6ultron_10tradingday_8Calendar_EasterMonday[0]) * (299));

  /* "ultron/tradingday/Calendar.pyx":1013
 *             return isinstance(right, TargetImpl)
 * 
 * cdef dict _holDict = {'china.sse': ChinaSseImpl,             # <<<<<<<<<<<<<<
 *                       'china.ib': ChinaIBImpl,
 *                       'china.cffex': ChinaCFFEXImpl,
 */
  __pyx_t_425 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_425)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PyDict_SetItem(__pyx_t_425, __pyx_kp_s_china_sse, ((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaSseImpl)) < 0) __PYX_ERR(0, 1013, __pyx_L1_error)

  /* "ultron/tradingday/Calendar.pyx":1014
 * 
 * cdef dict _holDict = {'china.sse': ChinaSseImpl,
 *                       'china.ib': ChinaIBImpl,             # <<<<<<<<<<<<<<
 *                       'china.cffex': ChinaCFFEXImpl,
 *                       'target': TargetImpl,
 */
  if (PyDict_SetItem(__pyx_t_425, __pyx_kp_s_china_ib, ((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaIBImpl)) < 0) __PYX_ERR(0, 1013, __pyx_L1_error)

  /* "ultron/tradingday/Calendar.pyx":1015
 * cdef dict _holDict = {'china.sse': ChinaSseImpl,
 *                       'china.ib': ChinaIBImpl,
 *                       'china.cffex': ChinaCFFEXImpl,             # <<<<<<<<<<<<<<
 *                       'target': TargetImpl,
 *                       'null': NullCalendar,
 */
  if (PyDict_SetItem(__pyx_t_425, __pyx_kp_s_china_cffex, ((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_ChinaCFFEXImpl)) < 0) __PYX_ERR(0, 1013, __pyx_L1_error)

  /* "ultron/tradingday/Calendar.pyx":1016
 *                       'china.ib': ChinaIBImpl,
 *                       'china.cffex': ChinaCFFEXImpl,
 *                       'target': TargetImpl,             # <<<<<<<<<<<<<<
 *                       'null': NullCalendar,
 *                       'nullcalendar': NullCalendar}
 */
  if (PyDict_SetItem(__pyx_t_425, __pyx_n_s_target, ((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_TargetImpl)) < 0) __PYX_ERR(0, 1013, __pyx_L1_error)

  /* "ultron/tradingday/Calendar.pyx":1017
 *                       'china.cffex': ChinaCFFEXImpl,
 *                       'target': TargetImpl,
 *                       'null': NullCalendar,             # <<<<<<<<<<<<<<
 *                       'nullcalendar': NullCalendar}
 */
  if (PyDict_SetItem(__pyx_t_425, __pyx_n_s_null, ((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_NullCalendar)) < 0) __PYX_ERR(0, 1013, __pyx_L1_error)

  /* "ultron/tradingday/Calendar.pyx":1018
 *                       'target': TargetImpl,
 *                       'null': NullCalendar,
 *                       'nullcalendar': NullCalendar}             # <<<<<<<<<<<<<<
 */
  if (PyDict_SetItem(__pyx_t_425, __pyx_n_s_nullcalendar, ((PyObject *)__pyx_ptype_6ultron_10tradingday_8Calendar_NullCalendar)) < 0) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_6ultron_10tradingday_8Calendar__holDict);
  __Pyx_DECREF_SET(__pyx_v_6ultron_10tradingday_8Calendar__holDict, ((PyObject*)__pyx_t_425));
  __Pyx_GIVEREF(__pyx_t_425);
  __pyx_t_425 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CalendarImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_425 = PyCFunction_NewEx(&__pyx_mdef_6ultron_10tradingday_8Calendar_1__pyx_unpickle_CalendarImpl, NULL, __pyx_n_s_ultron_tradingday_Calendar); if (unlikely(!__pyx_t_425)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CalendarImpl, __pyx_t_425) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_425); __pyx_t_425 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_CalendarImpl__set_state(<CalendarImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CalendarImpl__set_state(CalendarImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_425 = PyCFunction_NewEx(&__pyx_mdef_6ultron_10tradingday_8Calendar_3__pyx_unpickle_ChinaSseImpl, NULL, __pyx_n_s_ultron_tradingday_Calendar); if (unlikely(!__pyx_t_425)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ChinaSseImpl, __pyx_t_425) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_425); __pyx_t_425 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ChinaIBImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_425 = PyCFunction_NewEx(&__pyx_mdef_6ultron_10tradingday_8Calendar_5__pyx_unpickle_ChinaIBImpl, NULL, __pyx_n_s_ultron_tradingday_Calendar); if (unlikely(!__pyx_t_425)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ChinaIBImpl, __pyx_t_425) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_425); __pyx_t_425 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_ChinaIBImpl__set_state(<ChinaIBImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaIBImpl__set_state(ChinaIBImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_425 = PyCFunction_NewEx(&__pyx_mdef_6ultron_10tradingday_8Calendar_7__pyx_unpickle_NullCalendar, NULL, __pyx_n_s_ultron_tradingday_Calendar); if (unlikely(!__pyx_t_425)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_NullCalendar, __pyx_t_425) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_425); __pyx_t_425 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ChinaCFFEXImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_425 = PyCFunction_NewEx(&__pyx_mdef_6ultron_10tradingday_8Calendar_9__pyx_unpickle_ChinaCFFEXImpl, NULL, __pyx_n_s_ultron_tradingday_Calendar); if (unlikely(!__pyx_t_425)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ChinaCFFEXImpl, __pyx_t_425) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_425); __pyx_t_425 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_ChinaCFFEXImpl__set_state(<ChinaCFFEXImpl> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChinaCFFEXImpl__set_state(ChinaCFFEXImpl __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_425 = PyCFunction_NewEx(&__pyx_mdef_6ultron_10tradingday_8Calendar_11__pyx_unpickle_WestenImpl, NULL, __pyx_n_s_ultron_tradingday_Calendar); if (unlikely(!__pyx_t_425)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_WestenImpl, __pyx_t_425) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_425); __pyx_t_425 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TargetImpl(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_425 = PyCFunction_NewEx(&__pyx_mdef_6ultron_10tradingday_8Calendar_13__pyx_unpickle_TargetImpl, NULL, __pyx_n_s_ultron_tradingday_Calendar); if (unlikely(!__pyx_t_425)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_TargetImpl, __pyx_t_425) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_425); __pyx_t_425 = 0;

  /* "ultron/tradingday/Calendar.pyx":1
 * # -*- coding: utf-8 -*-             # <<<<<<<<<<<<<<
 * 
 * from ultron.tradingday.Enums._Weekdays cimport Weekdays
 */
  __pyx_t_425 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_425)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_425);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_425) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_425); __pyx_t_425 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  __Pyx_XDECREF(__pyx_t_39);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_XDECREF(__pyx_t_42);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_XDECREF(__pyx_t_44);
  __Pyx_XDECREF(__pyx_t_45);
  __Pyx_XDECREF(__pyx_t_46);
  __Pyx_XDECREF(__pyx_t_47);
  __Pyx_XDECREF(__pyx_t_48);
  __Pyx_XDECREF(__pyx_t_49);
  __Pyx_XDECREF(__pyx_t_50);
  __Pyx_XDECREF(__pyx_t_51);
  __Pyx_XDECREF(__pyx_t_52);
  __Pyx_XDECREF(__pyx_t_53);
  __Pyx_XDECREF(__pyx_t_54);
  __Pyx_XDECREF(__pyx_t_55);
  __Pyx_XDECREF(__pyx_t_56);
  __Pyx_XDECREF(__pyx_t_57);
  __Pyx_XDECREF(__pyx_t_58);
  __Pyx_XDECREF(__pyx_t_59);
  __Pyx_XDECREF(__pyx_t_60);
  __Pyx_XDECREF(__pyx_t_61);
  __Pyx_XDECREF(__pyx_t_62);
  __Pyx_XDECREF(__pyx_t_63);
  __Pyx_XDECREF(__pyx_t_64);
  __Pyx_XDECREF(__pyx_t_65);
  __Pyx_XDECREF(__pyx_t_66);
  __Pyx_XDECREF(__pyx_t_67);
  __Pyx_XDECREF(__pyx_t_68);
  __Pyx_XDECREF(__pyx_t_69);
  __Pyx_XDECREF(__pyx_t_70);
  __Pyx_XDECREF(__pyx_t_71);
  __Pyx_XDECREF(__pyx_t_72);
  __Pyx_XDECREF(__pyx_t_73);
  __Pyx_XDECREF(__pyx_t_74);
  __Pyx_XDECREF(__pyx_t_75);
  __Pyx_XDECREF(__pyx_t_76);
  __Pyx_XDECREF(__pyx_t_77);
  __Pyx_XDECREF(__pyx_t_78);
  __Pyx_XDECREF(__pyx_t_79);
  __Pyx_XDECREF(__pyx_t_80);
  __Pyx_XDECREF(__pyx_t_81);
  __Pyx_XDECREF(__pyx_t_82);
  __Pyx_XDECREF(__pyx_t_83);
  __Pyx_XDECREF(__pyx_t_84);
  __Pyx_XDECREF(__pyx_t_85);
  __Pyx_XDECREF(__pyx_t_86);
  __Pyx_XDECREF(__pyx_t_87);
  __Pyx_XDECREF(__pyx_t_88);
  __Pyx_XDECREF(__pyx_t_89);
  __Pyx_XDECREF(__pyx_t_90);
  __Pyx_XDECREF(__pyx_t_91);
  __Pyx_XDECREF(__pyx_t_92);
  __Pyx_XDECREF(__pyx_t_93);
  __Pyx_XDECREF(__pyx_t_94);
  __Pyx_XDECREF(__pyx_t_95);
  __Pyx_XDECREF(__pyx_t_96);
  __Pyx_XDECREF(__pyx_t_97);
  __Pyx_XDECREF(__pyx_t_98);
  __Pyx_XDECREF(__pyx_t_99);
  __Pyx_XDECREF(__pyx_t_100);
  __Pyx_XDECREF(__pyx_t_101);
  __Pyx_XDECREF(__pyx_t_102);
  __Pyx_XDECREF(__pyx_t_103);
  __Pyx_XDECREF(__pyx_t_104);
  __Pyx_XDECREF(__pyx_t_105);
  __Pyx_XDECREF(__pyx_t_106);
  __Pyx_XDECREF(__pyx_t_107);
  __Pyx_XDECREF(__pyx_t_108);
  __Pyx_XDECREF(__pyx_t_109);
  __Pyx_XDECREF(__pyx_t_110);
  __Pyx_XDECREF(__pyx_t_111);
  __Pyx_XDECREF(__pyx_t_112);
  __Pyx_XDECREF(__pyx_t_113);
  __Pyx_XDECREF(__pyx_t_114);
  __Pyx_XDECREF(__pyx_t_115);
  __Pyx_XDECREF(__pyx_t_116);
  __Pyx_XDECREF(__pyx_t_117);
  __Pyx_XDECREF(__pyx_t_118);
  __Pyx_XDECREF(__pyx_t_119);
  __Pyx_XDECREF(__pyx_t_120);
  __Pyx_XDECREF(__pyx_t_121);
  __Pyx_XDECREF(__pyx_t_122);
  __Pyx_XDECREF(__pyx_t_123);
  __Pyx_XDECREF(__pyx_t_124);
  __Pyx_XDECREF(__pyx_t_125);
  __Pyx_XDECREF(__pyx_t_126);
  __Pyx_XDECREF(__pyx_t_127);
  __Pyx_XDECREF(__pyx_t_128);
  __Pyx_XDECREF(__pyx_t_129);
  __Pyx_XDECREF(__pyx_t_130);
  __Pyx_XDECREF(__pyx_t_131);
  __Pyx_XDECREF(__pyx_t_132);
  __Pyx_XDECREF(__pyx_t_133);
  __Pyx_XDECREF(__pyx_t_134);
  __Pyx_XDECREF(__pyx_t_135);
  __Pyx_XDECREF(__pyx_t_136);
  __Pyx_XDECREF(__pyx_t_137);
  __Pyx_XDECREF(__pyx_t_138);
  __Pyx_XDECREF(__pyx_t_139);
  __Pyx_XDECREF(__pyx_t_140);
  __Pyx_XDECREF(__pyx_t_141);
  __Pyx_XDECREF(__pyx_t_142);
  __Pyx_XDECREF(__pyx_t_143);
  __Pyx_XDECREF(__pyx_t_144);
  __Pyx_XDECREF(__pyx_t_145);
  __Pyx_XDECREF(__pyx_t_146);
  __Pyx_XDECREF(__pyx_t_147);
  __Pyx_XDECREF(__pyx_t_148);
  __Pyx_XDECREF(__pyx_t_149);
  __Pyx_XDECREF(__pyx_t_150);
  __Pyx_XDECREF(__pyx_t_151);
  __Pyx_XDECREF(__pyx_t_152);
  __Pyx_XDECREF(__pyx_t_153);
  __Pyx_XDECREF(__pyx_t_154);
  __Pyx_XDECREF(__pyx_t_155);
  __Pyx_XDECREF(__pyx_t_156);
  __Pyx_XDECREF(__pyx_t_157);
  __Pyx_XDECREF(__pyx_t_158);
  __Pyx_XDECREF(__pyx_t_159);
  __Pyx_XDECREF(__pyx_t_160);
  __Pyx_XDECREF(__pyx_t_161);
  __Pyx_XDECREF(__pyx_t_162);
  __Pyx_XDECREF(__pyx_t_163);
  __Pyx_XDECREF(__pyx_t_164);
  __Pyx_XDECREF(__pyx_t_165);
  __Pyx_XDECREF(__pyx_t_166);
  __Pyx_XDECREF(__pyx_t_167);
  __Pyx_XDECREF(__pyx_t_168);
  __Pyx_XDECREF(__pyx_t_169);
  __Pyx_XDECREF(__pyx_t_170);
  __Pyx_XDECREF(__pyx_t_171);
  __Pyx_XDECREF(__pyx_t_172);
  __Pyx_XDECREF(__pyx_t_173);
  __Pyx_XDECREF(__pyx_t_174);
  __Pyx_XDECREF(__pyx_t_175);
  __Pyx_XDECREF(__pyx_t_176);
  __Pyx_XDECREF(__pyx_t_177);
  __Pyx_XDECREF(__pyx_t_178);
  __Pyx_XDECREF(__pyx_t_179);
  __Pyx_XDECREF(__pyx_t_180);
  __Pyx_XDECREF(__pyx_t_181);
  __Pyx_XDECREF(__pyx_t_182);
  __Pyx_XDECREF(__pyx_t_183);
  __Pyx_XDECREF(__pyx_t_184);
  __Pyx_XDECREF(__pyx_t_185);
  __Pyx_XDECREF(__pyx_t_186);
  __Pyx_XDECREF(__pyx_t_187);
  __Pyx_XDECREF(__pyx_t_188);
  __Pyx_XDECREF(__pyx_t_189);
  __Pyx_XDECREF(__pyx_t_190);
  __Pyx_XDECREF(__pyx_t_191);
  __Pyx_XDECREF(__pyx_t_192);
  __Pyx_XDECREF(__pyx_t_193);
  __Pyx_XDECREF(__pyx_t_194);
  __Pyx_XDECREF(__pyx_t_195);
  __Pyx_XDECREF(__pyx_t_196);
  __Pyx_XDECREF(__pyx_t_197);
  __Pyx_XDECREF(__pyx_t_198);
  __Pyx_XDECREF(__pyx_t_199);
  __Pyx_XDECREF(__pyx_t_200);
  __Pyx_XDECREF(__pyx_t_201);
  __Pyx_XDECREF(__pyx_t_202);
  __Pyx_XDECREF(__pyx_t_203);
  __Pyx_XDECREF(__pyx_t_204);
  __Pyx_XDECREF(__pyx_t_205);
  __Pyx_XDECREF(__pyx_t_206);
  __Pyx_XDECREF(__pyx_t_207);
  __Pyx_XDECREF(__pyx_t_208);
  __Pyx_XDECREF(__pyx_t_209);
  __Pyx_XDECREF(__pyx_t_210);
  __Pyx_XDECREF(__pyx_t_211);
  __Pyx_XDECREF(__pyx_t_212);
  __Pyx_XDECREF(__pyx_t_213);
  __Pyx_XDECREF(__pyx_t_214);
  __Pyx_XDECREF(__pyx_t_215);
  __Pyx_XDECREF(__pyx_t_216);
  __Pyx_XDECREF(__pyx_t_217);
  __Pyx_XDECREF(__pyx_t_218);
  __Pyx_XDECREF(__pyx_t_219);
  __Pyx_XDECREF(__pyx_t_220);
  __Pyx_XDECREF(__pyx_t_221);
  __Pyx_XDECREF(__pyx_t_222);
  __Pyx_XDECREF(__pyx_t_223);
  __Pyx_XDECREF(__pyx_t_224);
  __Pyx_XDECREF(__pyx_t_225);
  __Pyx_XDECREF(__pyx_t_226);
  __Pyx_XDECREF(__pyx_t_227);
  __Pyx_XDECREF(__pyx_t_228);
  __Pyx_XDECREF(__pyx_t_229);
  __Pyx_XDECREF(__pyx_t_230);
  __Pyx_XDECREF(__pyx_t_231);
  __Pyx_XDECREF(__pyx_t_232);
  __Pyx_XDECREF(__pyx_t_233);
  __Pyx_XDECREF(__pyx_t_234);
  __Pyx_XDECREF(__pyx_t_235);
  __Pyx_XDECREF(__pyx_t_236);
  __Pyx_XDECREF(__pyx_t_237);
  __Pyx_XDECREF(__pyx_t_238);
  __Pyx_XDECREF(__pyx_t_239);
  __Pyx_XDECREF(__pyx_t_240);
  __Pyx_XDECREF(__pyx_t_241);
  __Pyx_XDECREF(__pyx_t_242);
  __Pyx_XDECREF(__pyx_t_243);
  __Pyx_XDECREF(__pyx_t_244);
  __Pyx_XDECREF(__pyx_t_245);
  __Pyx_XDECREF(__pyx_t_246);
  __Pyx_XDECREF(__pyx_t_247);
  __Pyx_XDECREF(__pyx_t_248);
  __Pyx_XDECREF(__pyx_t_249);
  __Pyx_XDECREF(__pyx_t_250);
  __Pyx_XDECREF(__pyx_t_251);
  __Pyx_XDECREF(__pyx_t_252);
  __Pyx_XDECREF(__pyx_t_253);
  __Pyx_XDECREF(__pyx_t_254);
  __Pyx_XDECREF(__pyx_t_255);
  __Pyx_XDECREF(__pyx_t_256);
  __Pyx_XDECREF(__pyx_t_257);
  __Pyx_XDECREF(__pyx_t_258);
  __Pyx_XDECREF(__pyx_t_259);
  __Pyx_XDECREF(__pyx_t_260);
  __Pyx_XDECREF(__pyx_t_261);
  __Pyx_XDECREF(__pyx_t_262);
  __Pyx_XDECREF(__pyx_t_263);
  __Pyx_XDECREF(__pyx_t_264);
  __Pyx_XDECREF(__pyx_t_265);
  __Pyx_XDECREF(__pyx_t_266);
  __Pyx_XDECREF(__pyx_t_267);
  __Pyx_XDECREF(__pyx_t_268);
  __Pyx_XDECREF(__pyx_t_269);
  __Pyx_XDECREF(__pyx_t_270);
  __Pyx_XDECREF(__pyx_t_271);
  __Pyx_XDECREF(__pyx_t_272);
  __Pyx_XDECREF(__pyx_t_273);
  __Pyx_XDECREF(__pyx_t_274);
  __Pyx_XDECREF(__pyx_t_275);
  __Pyx_XDECREF(__pyx_t_276);
  __Pyx_XDECREF(__pyx_t_277);
  __Pyx_XDECREF(__pyx_t_278);
  __Pyx_XDECREF(__pyx_t_279);
  __Pyx_XDECREF(__pyx_t_280);
  __Pyx_XDECREF(__pyx_t_281);
  __Pyx_XDECREF(__pyx_t_282);
  __Pyx_XDECREF(__pyx_t_283);
  __Pyx_XDECREF(__pyx_t_284);
  __Pyx_XDECREF(__pyx_t_285);
  __Pyx_XDECREF(__pyx_t_286);
  __Pyx_XDECREF(__pyx_t_287);
  __Pyx_XDECREF(__pyx_t_288);
  __Pyx_XDECREF(__pyx_t_289);
  __Pyx_XDECREF(__pyx_t_290);
  __Pyx_XDECREF(__pyx_t_291);
  __Pyx_XDECREF(__pyx_t_292);
  __Pyx_XDECREF(__pyx_t_293);
  __Pyx_XDECREF(__pyx_t_294);
  __Pyx_XDECREF(__pyx_t_295);
  __Pyx_XDECREF(__pyx_t_296);
  __Pyx_XDECREF(__pyx_t_297);
  __Pyx_XDECREF(__pyx_t_298);
  __Pyx_XDECREF(__pyx_t_299);
  __Pyx_XDECREF(__pyx_t_300);
  __Pyx_XDECREF(__pyx_t_301);
  __Pyx_XDECREF(__pyx_t_302);
  __Pyx_XDECREF(__pyx_t_303);
  __Pyx_XDECREF(__pyx_t_304);
  __Pyx_XDECREF(__pyx_t_305);
  __Pyx_XDECREF(__pyx_t_306);
  __Pyx_XDECREF(__pyx_t_307);
  __Pyx_XDECREF(__pyx_t_308);
  __Pyx_XDECREF(__pyx_t_309);
  __Pyx_XDECREF(__pyx_t_310);
  __Pyx_XDECREF(__pyx_t_311);
  __Pyx_XDECREF(__pyx_t_312);
  __Pyx_XDECREF(__pyx_t_313);
  __Pyx_XDECREF(__pyx_t_314);
  __Pyx_XDECREF(__pyx_t_315);
  __Pyx_XDECREF(__pyx_t_316);
  __Pyx_XDECREF(__pyx_t_317);
  __Pyx_XDECREF(__pyx_t_318);
  __Pyx_XDECREF(__pyx_t_319);
  __Pyx_XDECREF(__pyx_t_320);
  __Pyx_XDECREF(__pyx_t_321);
  __Pyx_XDECREF(__pyx_t_322);
  __Pyx_XDECREF(__pyx_t_323);
  __Pyx_XDECREF(__pyx_t_324);
  __Pyx_XDECREF(__pyx_t_325);
  __Pyx_XDECREF(__pyx_t_326);
  __Pyx_XDECREF(__pyx_t_327);
  __Pyx_XDECREF(__pyx_t_328);
  __Pyx_XDECREF(__pyx_t_329);
  __Pyx_XDECREF(__pyx_t_330);
  __Pyx_XDECREF(__pyx_t_331);
  __Pyx_XDECREF(__pyx_t_332);
  __Pyx_XDECREF(__pyx_t_333);
  __Pyx_XDECREF(__pyx_t_334);
  __Pyx_XDECREF(__pyx_t_335);
  __Pyx_XDECREF(__pyx_t_336);
  __Pyx_XDECREF(__pyx_t_337);
  __Pyx_XDECREF(__pyx_t_338);
  __Pyx_XDECREF(__pyx_t_339);
  __Pyx_XDECREF(__pyx_t_340);
  __Pyx_XDECREF(__pyx_t_341);
  __Pyx_XDECREF(__pyx_t_342);
  __Pyx_XDECREF(__pyx_t_343);
  __Pyx_XDECREF(__pyx_t_344);
  __Pyx_XDECREF(__pyx_t_345);
  __Pyx_XDECREF(__pyx_t_346);
  __Pyx_XDECREF(__pyx_t_347);
  __Pyx_XDECREF(__pyx_t_348);
  __Pyx_XDECREF(__pyx_t_349);
  __Pyx_XDECREF(__pyx_t_350);
  __Pyx_XDECREF(__pyx_t_351);
  __Pyx_XDECREF(__pyx_t_352);
  __Pyx_XDECREF(__pyx_t_353);
  __Pyx_XDECREF(__pyx_t_354);
  __Pyx_XDECREF(__pyx_t_355);
  __Pyx_XDECREF(__pyx_t_356);
  __Pyx_XDECREF(__pyx_t_357);
  __Pyx_XDECREF(__pyx_t_358);
  __Pyx_XDECREF(__pyx_t_359);
  __Pyx_XDECREF(__pyx_t_360);
  __Pyx_XDECREF(__pyx_t_361);
  __Pyx_XDECREF(__pyx_t_362);
  __Pyx_XDECREF(__pyx_t_363);
  __Pyx_XDECREF(__pyx_t_364);
  __Pyx_XDECREF(__pyx_t_365);
  __Pyx_XDECREF(__pyx_t_366);
  __Pyx_XDECREF(__pyx_t_367);
  __Pyx_XDECREF(__pyx_t_368);
  __Pyx_XDECREF(__pyx_t_369);
  __Pyx_XDECREF(__pyx_t_370);
  __Pyx_XDECREF(__pyx_t_371);
  __Pyx_XDECREF(__pyx_t_372);
  __Pyx_XDECREF(__pyx_t_373);
  __Pyx_XDECREF(__pyx_t_374);
  __Pyx_XDECREF(__pyx_t_375);
  __Pyx_XDECREF(__pyx_t_376);
  __Pyx_XDECREF(__pyx_t_377);
  __Pyx_XDECREF(__pyx_t_378);
  __Pyx_XDECREF(__pyx_t_379);
  __Pyx_XDECREF(__pyx_t_380);
  __Pyx_XDECREF(__pyx_t_381);
  __Pyx_XDECREF(__pyx_t_382);
  __Pyx_XDECREF(__pyx_t_383);
  __Pyx_XDECREF(__pyx_t_384);
  __Pyx_XDECREF(__pyx_t_385);
  __Pyx_XDECREF(__pyx_t_386);
  __Pyx_XDECREF(__pyx_t_387);
  __Pyx_XDECREF(__pyx_t_388);
  __Pyx_XDECREF(__pyx_t_389);
  __Pyx_XDECREF(__pyx_t_390);
  __Pyx_XDECREF(__pyx_t_391);
  __Pyx_XDECREF(__pyx_t_392);
  __Pyx_XDECREF(__pyx_t_393);
  __Pyx_XDECREF(__pyx_t_394);
  __Pyx_XDECREF(__pyx_t_395);
  __Pyx_XDECREF(__pyx_t_396);
  __Pyx_XDECREF(__pyx_t_397);
  __Pyx_XDECREF(__pyx_t_398);
  __Pyx_XDECREF(__pyx_t_399);
  __Pyx_XDECREF(__pyx_t_400);
  __Pyx_XDECREF(__pyx_t_401);
  __Pyx_XDECREF(__pyx_t_402);
  __Pyx_XDECREF(__pyx_t_403);
  __Pyx_XDECREF(__pyx_t_404);
  __Pyx_XDECREF(__pyx_t_405);
  __Pyx_XDECREF(__pyx_t_406);
  __Pyx_XDECREF(__pyx_t_407);
  __Pyx_XDECREF(__pyx_t_408);
  __Pyx_XDECREF(__pyx_t_409);
  __Pyx_XDECREF(__pyx_t_410);
  __Pyx_XDECREF(__pyx_t_411);
  __Pyx_XDECREF(__pyx_t_412);
  __Pyx_XDECREF(__pyx_t_413);
  __Pyx_XDECREF(__pyx_t_414);
  __Pyx_XDECREF(__pyx_t_415);
  __Pyx_XDECREF(__pyx_t_416);
  __Pyx_XDECREF(__pyx_t_417);
  __Pyx_XDECREF(__pyx_t_418);
  __Pyx_XDECREF(__pyx_t_419);
  __Pyx_XDECREF(__pyx_t_420);
  __Pyx_XDECREF(__pyx_t_421);
  __Pyx_XDECREF(__pyx_t_422);
  __Pyx_XDECREF(__pyx_t_423);
  __Pyx_XDECREF(__pyx_t_424);
  __Pyx_XDECREF(__pyx_t_425);
  __Pyx_XDECREF(__pyx_t_426);
  __Pyx_XDECREF(__pyx_t_427);
  __Pyx_XDECREF(__pyx_t_428);
  __Pyx_XDECREF(__pyx_t_429);
  __Pyx_XDECREF(__pyx_t_430);
  __Pyx_XDECREF(__pyx_t_431);
  __Pyx_XDECREF(__pyx_t_432);
  __Pyx_XDECREF(__pyx_t_433);
  __Pyx_XDECREF(__pyx_t_434);
  __Pyx_XDECREF(__pyx_t_435);
  __Pyx_XDECREF(__pyx_t_436);
  __Pyx_XDECREF(__pyx_t_437);
  __Pyx_XDECREF(__pyx_t_438);
  __Pyx_XDECREF(__pyx_t_439);
  __Pyx_XDECREF(__pyx_t_440);
  __Pyx_XDECREF(__pyx_t_441);
  __Pyx_XDECREF(__pyx_t_442);
  __Pyx_XDECREF(__pyx_t_443);
  __Pyx_XDECREF(__pyx_t_444);
  __Pyx_XDECREF(__pyx_t_445);
  __Pyx_XDECREF(__pyx_t_446);
  __Pyx_XDECREF(__pyx_t_447);
  __Pyx_XDECREF(__pyx_t_448);
  __Pyx_XDECREF(__pyx_t_449);
  __Pyx_XDECREF(__pyx_t_450);
  __Pyx_XDECREF(__pyx_t_451);
  __Pyx_XDECREF(__pyx_t_452);
  __Pyx_XDECREF(__pyx_t_453);
  __Pyx_XDECREF(__pyx_t_454);
  __Pyx_XDECREF(__pyx_t_455);
  __Pyx_XDECREF(__pyx_t_456);
  __Pyx_XDECREF(__pyx_t_457);
  __Pyx_XDECREF(__pyx_t_458);
  __Pyx_XDECREF(__pyx_t_459);
  __Pyx_XDECREF(__pyx_t_460);
  __Pyx_XDECREF(__pyx_t_461);
  __Pyx_XDECREF(__pyx_t_462);
  __Pyx_XDECREF(__pyx_t_463);
  __Pyx_XDECREF(__pyx_t_464);
  __Pyx_XDECREF(__pyx_t_465);
  __Pyx_XDECREF(__pyx_t_466);
  __Pyx_XDECREF(__pyx_t_467);
  __Pyx_XDECREF(__pyx_t_468);
  __Pyx_XDECREF(__pyx_t_469);
  __Pyx_XDECREF(__pyx_t_470);
  __Pyx_XDECREF(__pyx_t_471);
  __Pyx_XDECREF(__pyx_t_472);
  __Pyx_XDECREF(__pyx_t_473);
  __Pyx_XDECREF(__pyx_t_474);
  __Pyx_XDECREF(__pyx_t_475);
  __Pyx_XDECREF(__pyx_t_476);
  __Pyx_XDECREF(__pyx_t_477);
  __Pyx_XDECREF(__pyx_t_478);
  __Pyx_XDECREF(__pyx_t_479);
  __Pyx_XDECREF(__pyx_t_480);
  __Pyx_XDECREF(__pyx_t_481);
  __Pyx_XDECREF(__pyx_t_482);
  __Pyx_XDECREF(__pyx_t_483);
  __Pyx_XDECREF(__pyx_t_484);
  __Pyx_XDECREF(__pyx_t_485);
  __Pyx_XDECREF(__pyx_t_486);
  __Pyx_XDECREF(__pyx_t_487);
  __Pyx_XDECREF(__pyx_t_488);
  __Pyx_XDECREF(__pyx_t_489);
  __Pyx_XDECREF(__pyx_t_490);
  __Pyx_XDECREF(__pyx_t_491);
  __Pyx_XDECREF(__pyx_t_492);
  __Pyx_XDECREF(__pyx_t_493);
  __Pyx_XDECREF(__pyx_t_494);
  __Pyx_XDECREF(__pyx_t_495);
  __Pyx_XDECREF(__pyx_t_496);
  __Pyx_XDECREF(__pyx_t_497);
  __Pyx_XDECREF(__pyx_t_498);
  __Pyx_XDECREF(__pyx_t_499);
  __Pyx_XDECREF(__pyx_t_500);
  __Pyx_XDECREF(__pyx_t_501);
  __Pyx_XDECREF(__pyx_t_502);
  __Pyx_XDECREF(__pyx_t_503);
  __Pyx_XDECREF(__pyx_t_504);
  __Pyx_XDECREF(__pyx_t_505);
  __Pyx_XDECREF(__pyx_t_506);
  __Pyx_XDECREF(__pyx_t_507);
  __Pyx_XDECREF(__pyx_t_508);
  __Pyx_XDECREF(__pyx_t_509);
  __Pyx_XDECREF(__pyx_t_510);
  __Pyx_XDECREF(__pyx_t_511);
  __Pyx_XDECREF(__pyx_t_512);
  __Pyx_XDECREF(__pyx_t_513);
  __Pyx_XDECREF(__pyx_t_514);
  __Pyx_XDECREF(__pyx_t_515);
  __Pyx_XDECREF(__pyx_t_516);
  __Pyx_XDECREF(__pyx_t_517);
  __Pyx_XDECREF(__pyx_t_518);
  __Pyx_XDECREF(__pyx_t_519);
  __Pyx_XDECREF(__pyx_t_520);
  __Pyx_XDECREF(__pyx_t_521);
  __Pyx_XDECREF(__pyx_t_522);
  __Pyx_XDECREF(__pyx_t_523);
  __Pyx_XDECREF(__pyx_t_524);
  __Pyx_XDECREF(__pyx_t_525);
  __Pyx_XDECREF(__pyx_t_526);
  __Pyx_XDECREF(__pyx_t_527);
  __Pyx_XDECREF(__pyx_t_528);
  __Pyx_XDECREF(__pyx_t_529);
  __Pyx_XDECREF(__pyx_t_530);
  __Pyx_XDECREF(__pyx_t_531);
  __Pyx_XDECREF(__pyx_t_532);
  __Pyx_XDECREF(__pyx_t_533);
  __Pyx_XDECREF(__pyx_t_534);
  __Pyx_XDECREF(__pyx_t_535);
  __Pyx_XDECREF(__pyx_t_536);
  __Pyx_XDECREF(__pyx_t_537);
  __Pyx_XDECREF(__pyx_t_538);
  __Pyx_XDECREF(__pyx_t_539);
  __Pyx_XDECREF(__pyx_t_540);
  __Pyx_XDECREF(__pyx_t_541);
  __Pyx_XDECREF(__pyx_t_542);
  __Pyx_XDECREF(__pyx_t_543);
  __Pyx_XDECREF(__pyx_t_544);
  __Pyx_XDECREF(__pyx_t_545);
  __Pyx_XDECREF(__pyx_t_546);
  __Pyx_XDECREF(__pyx_t_547);
  __Pyx_XDECREF(__pyx_t_548);
  __Pyx_XDECREF(__pyx_t_549);
  __Pyx_XDECREF(__pyx_t_550);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init ultron.tradingday.Calendar", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init ultron.tradingday.Calendar");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return dict ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
        dictptr = (offset > 0) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (!dict || tp_dict_version != __PYX_GET_DICT_VERSION(dict))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if (likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__Months(enum Months value) {
    const enum Months neg_one = (enum Months) ((enum Months) 0 - (enum Months) 1), const_zero = (enum Months) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum Months) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum Months) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum Months) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum Months) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum Months) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum Months),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
